/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include/numpy/ufuncobject.h",
            "ttcr/Cell.h",
            "ttcr/Grid2D.h",
            "ttcr/Grid2Drc.h",
            "ttcr/Grid2Drcfs.h",
            "ttcr/Grid2Drcsp.h",
            "ttcr/Grid2Drn.h",
            "ttcr/Grid2Drnfs.h",
            "ttcr/Grid2Drnsp.h",
            "ttcr/Grid3D.h",
            "ttcr/Grid3Drc.h",
            "ttcr/Grid3Drcdsp.h",
            "ttcr/Grid3Drcfs.h",
            "ttcr/Grid3Drcsp.h",
            "ttcr/Grid3Drn.h",
            "ttcr/Grid3Drndsp.h",
            "ttcr/Grid3Drnfs.h",
            "ttcr/Grid3Drnsp.h",
            "ttcr/Node2Dc.h",
            "ttcr/Node2Dcsp.h",
            "ttcr/Node2Dn.h",
            "ttcr/Node2Dnsp.h",
            "ttcr/Node3Dc.h",
            "ttcr/Node3Dcsp.h",
            "ttcr/Node3Dn.h",
            "ttcr/Node3Dnsp.h",
            "ttcr/ttcr_t.h",
            "ttcrpy/typedefs.h",
            "ttcrpy/verbose.h"
        ],
        "extra_compile_args": [
            "-std=c++11",
            "-stdlib=libc++",
            "-O3"
        ],
        "include_dirs": [
            "ttcrpy",
            "./ttcrpy",
            "ttcr",
            "boost_1_72_0",
            "eigen-3.3.7",
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include"
        ],
        "language": "c++",
        "name": "ttcrpy.rgrid",
        "sources": [
            "ttcrpy/rgrid.pyx",
            "ttcrpy/verbose.cpp"
        ]
    },
    "module_name": "ttcrpy.rgrid"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ttcrpy__rgrid
#define __PYX_HAVE_API__ttcrpy__rgrid
/* Early includes */
#include <string.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <string>
#include <vector>
#include <stdint.h>
#include <math.h>
#include "ttcr_t.h"
#include "Node3Dn.h"
#include "Node3Dnsp.h"
#include "Node3Dc.h"
#include "Node3Dcsp.h"
#include "Cell.h"
#include "Node2Dn.h"
#include "Node2Dc.h"
#include "Node2Dnsp.h"
#include "Node2Dcsp.h"
#include "typedefs.h"
#include "Grid3D.h"
#include "Grid3Drn.h"
#include "Grid3Drc.h"
#include "Grid3Drnfs.h"
#include "Grid3Drnsp.h"
#include "Grid3Drndsp.h"
#include "Grid3Drcfs.h"
#include "Grid3Drcsp.h"
#include "Grid3Drcdsp.h"
#include "Grid2D.h"
#include "Grid2Drc.h"
#include "Grid2Drn.h"
#include "Grid2Drcsp.h"
#include "Grid2Drcfs.h"
#include "Grid2Drnsp.h"
#include "Grid2Drnfs.h"
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "verbose.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "ttcrpy/rgrid.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_6ttcrpy_5rgrid_Grid3d;
struct __pyx_obj_6ttcrpy_5rgrid_Grid2d;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "ttcrpy/rgrid.pyx":49
 *     setVerbose(v)
 * 
 * cdef class Grid3d:             # <<<<<<<<<<<<<<
 *     """
 *     class to perform raytracing with 3D rectilinear grids
 */
struct __pyx_obj_6ttcrpy_5rgrid_Grid3d {
  PyObject_HEAD
  struct __pyx_vtabstruct_6ttcrpy_5rgrid_Grid3d *__pyx_vtab;
  std::vector<double>  _x;
  std::vector<double>  _y;
  std::vector<double>  _z;
  double _dx;
  double _dy;
  double _dz;
  bool cell_slowness;
  size_t _n_threads;
  char method;
  bool tt_from_rp;
  bool interp_vel;
  double eps;
  int maxit;
  bool weno;
  uint32_t nsnx;
  uint32_t nsny;
  uint32_t nsnz;
  uint32_t n_secondary;
  uint32_t n_tertiary;
  double radius_tertiary;
  ttcr::Grid3D<double,uint32_t>  *grid;
};


/* "ttcrpy/rgrid.pyx":1727
 * 
 * 
 * cdef class Grid2d:             # <<<<<<<<<<<<<<
 *     """
 *     class to perform raytracing with 2D rectilinear grids
 */
struct __pyx_obj_6ttcrpy_5rgrid_Grid2d {
  PyObject_HEAD
  std::vector<double>  _x;
  std::vector<double>  _z;
  double _dx;
  double _dz;
  bool cell_slowness;
  size_t _n_threads;
  char method;
  char iso;
  double eps;
  int maxit;
  bool weno;
  bool rotated_template;
  uint32_t nsnx;
  uint32_t nsnz;
  ttcr::Grid2D<double,uint32_t,ttcr::sxz<double> >  *grid;
};



/* "ttcrpy/rgrid.pyx":49
 *     setVerbose(v)
 * 
 * cdef class Grid3d:             # <<<<<<<<<<<<<<
 *     """
 *     class to perform raytracing with 3D rectilinear grids
 */

struct __pyx_vtabstruct_6ttcrpy_5rgrid_Grid3d {
  uint32_t (*_f2c_ind)(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *, uint32_t);
};
static struct __pyx_vtabstruct_6ttcrpy_5rgrid_Grid3d *__pyx_vtabptr_6ttcrpy_5rgrid_Grid3d;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_AddCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyUnicodeContains.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_ContainsTF(PyObject* substring, PyObject* text, int eq) {
    int result = PyUnicode_Contains(text, substring);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static uint32_t __pyx_f_6ttcrpy_5rgrid_6Grid3d__f2c_ind(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, uint32_t __pyx_v_ind); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libcpp.string' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libcpp' */

/* Module declarations from 'ttcrpy.common' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'ttcrpy.rgrid' */
static PyTypeObject *__pyx_ptype_6ttcrpy_5rgrid_Grid3d = 0;
static PyTypeObject *__pyx_ptype_6ttcrpy_5rgrid_Grid2d = 0;
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_double_t = { "double_t", NULL, sizeof(__pyx_t_5numpy_double_t), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ttcrpy.rgrid"
extern int __pyx_module_is_main_ttcrpy__rgrid;
int __pyx_module_is_main_ttcrpy__rgrid = 0;

/* Implementation of 'ttcrpy.rgrid' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_F[] = "F";
static const char __pyx_k_L[] = "L";
static const char __pyx_k_T[] = "T";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_z[] = "z";
static const char __pyx_k_Rx[] = "Rx";
static const char __pyx_k_Tx[] = "Tx";
static const char __pyx_k_dl[] = "dl";
static const char __pyx_k_dx[] = "dx";
static const char __pyx_k_dz[] = "dz";
static const char __pyx_k_ix[] = "ix";
static const char __pyx_k_iy[] = "iy";
static const char __pyx_k_iz[] = "iz";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_nt[] = "nt";
static const char __pyx_k_sp[] = "sp";
static const char __pyx_k_sy[] = "sy";
static const char __pyx_k_sz[] = "sz";
static const char __pyx_k_up[] = "up";
static const char __pyx_k_x1[] = "x1";
static const char __pyx_k_x2[] = "x2";
static const char __pyx_k_xc[] = "xc";
static const char __pyx_k_xe[] = "xe";
static const char __pyx_k_xi[] = "xi";
static const char __pyx_k_xr[] = "xr";
static const char __pyx_k_xs[] = "xs";
static const char __pyx_k_y1[] = "y1";
static const char __pyx_k_y2[] = "y2";
static const char __pyx_k_yc[] = "yc";
static const char __pyx_k_ye[] = "ye";
static const char __pyx_k_yi[] = "yi";
static const char __pyx_k_z1[] = "z1";
static const char __pyx_k_z2[] = "z2";
static const char __pyx_k_zc[] = "zc";
static const char __pyx_k_ze[] = "ze";
static const char __pyx_k_zi[] = "zi";
static const char __pyx_k_zr[] = "zr";
static const char __pyx_k_zs[] = "zs";
static const char __pyx_k_FSM[] = "FSM";
static const char __pyx_k_SPM[] = "SPM";
static const char __pyx_k_Vp0[] = "Vp0";
static const char __pyx_k_Vs0[] = "Vs0";
static const char __pyx_k__34[] = "_";
static const char __pyx_k__61[] = "*";
static const char __pyx_k_abs[] = "abs";
static const char __pyx_k_b_y[] = "b_y";
static const char __pyx_k_b_z[] = "b_z";
static const char __pyx_k_dim[] = "dim";
static const char __pyx_k_dlx[] = "dlx";
static const char __pyx_k_dly[] = "dly";
static const char __pyx_k_dlz[] = "dlz";
static const char __pyx_k_eps[] = "eps";
static const char __pyx_k_grx[] = "grx";
static const char __pyx_k_gry[] = "gry";
static const char __pyx_k_grz[] = "grz";
static const char __pyx_k_ind[] = "ind";
static const char __pyx_k_iso[] = "iso";
static const char __pyx_k_m_y[] = "m_y";
static const char __pyx_k_m_z[] = "m_z";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_nTx[] = "nTx";
static const char __pyx_k_rcv[] = "rcv";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_vtk[] = "vtk";
static const char __pyx_k_vtp[] = ".vtp";
static const char __pyx_k_vtr[] = ".vtr";
static const char __pyx_k_DSPM[] = "DSPM";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dtmp[] = "dtmp";
static const char __pyx_k_hypo[] = "hypo";
static const char __pyx_k_indc[] = "indc";
static const char __pyx_k_kron[] = "kron";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_nsnx[] = "nsnx";
static const char __pyx_k_nsny[] = "nsny";
static const char __pyx_k_nsnz[] = "nsnz";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_rays[] = "rays";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tile[] = "tile";
static const char __pyx_k_up_y[] = "up_y";
static const char __pyx_k_up_z[] = "up_z";
static const char __pyx_k_weno[] = "weno";
static const char __pyx_k_SetId[] = "SetId";
static const char __pyx_k_aniso[] = "aniso";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_delta[] = "delta";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_gamma[] = "gamma";
static const char __pyx_k_iCell[] = "iCell";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_maxit[] = "maxit";
static const char __pyx_k_n_grx[] = "n_grx";
static const char __pyx_k_n_gry[] = "n_gry";
static const char __pyx_k_n_grz[] = "n_grz";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_small[] = "small";
static const char __pyx_k_theta[] = "theta";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_Grid2d[] = "Grid2d";
static const char __pyx_k_Grid3d[] = "Grid3d";
static const char __pyx_k_Update[] = "Update";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_data_p[] = "data_p";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_hstack[] = "hstack";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_nCells[] = "nCells";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_set_xi[] = "set_xi";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_unique[] = "unique";
static const char __pyx_k_vstack[] = "vstack";
static const char __pyx_k_vti_sh[] = "vti_sh";
static const char __pyx_k_SetName[] = "SetName";
static const char __pyx_k_builder[] = "builder";
static const char __pyx_k_centers[] = "centers";
static const char __pyx_k_epsilon[] = "epsilon";
static const char __pyx_k_flatten[] = "flatten";
static const char __pyx_k_lowness[] = "lowness";
static const char __pyx_k_nonzero[] = "nonzero";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_set_Vp0[] = "set_Vp0";
static const char __pyx_k_set_Vs0[] = "set_Vs0";
static const char __pyx_k_vti_psv[] = "vti_psv";
static const char __pyx_k_AddArray[] = "AddArray";
static const char __pyx_k_GetArray[] = "GetArray";
static const char __pyx_k_HasArray[] = "HasArray";
static const char __pyx_k_SetLines[] = "SetLines";
static const char __pyx_k_Slowness[] = "Slowness";
static const char __pyx_k_Velocity[] = "Velocity";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_indptr_p[] = "indptr_p";
static const char __pyx_k_slowness[] = "slowness";
static const char __pyx_k_velocity[] = "velocity";
static const char __pyx_k_vtk_util[] = "vtk.util";
static const char __pyx_k_GetOutput[] = "GetOutput";
static const char __pyx_k_SetPoints[] = "SetPoints";
static const char __pyx_k_SetTuple1[] = "SetTuple1";
static const char __pyx_k_compute_L[] = "compute_L";
static const char __pyx_k_compute_M[] = "compute_M";
static const char __pyx_k_indices_p[] = "indices_p";
static const char __pyx_k_n_threads[] = "n_threads";
static const char __pyx_k_rebuild2d[] = "_rebuild2d";
static const char __pyx_k_rebuild3d[] = "_rebuild3d";
static const char __pyx_k_set_delta[] = "set_delta";
static const char __pyx_k_set_gamma[] = "set_gamma";
static const char __pyx_k_thread_no[] = "thread_no";
static const char __pyx_k_vtkPoints[] = "vtkPoints";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_csr_matrix[] = "csr_matrix";
static const char __pyx_k_elliptical[] = "elliptical";
static const char __pyx_k_interp_vel[] = "interp_vel";
static const char __pyx_k_is_outside[] = "is_outside";
static const char __pyx_k_n_tertiary[] = "n_tertiary";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_tt_from_rp[] = "tt_from_rp";
static const char __pyx_k_GetCellData[] = "GetCellData";
static const char __pyx_k_GetPointIds[] = "GetPointIds";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_InsertPoint[] = "InsertPoint";
static const char __pyx_k_SetFileName[] = "SetFileName";
static const char __pyx_k_n_secondary[] = "n_secondary";
static const char __pyx_k_return_rays[] = "return_rays";
static const char __pyx_k_set_epsilon[] = "set_epsilon";
static const char __pyx_k_set_verbose[] = "set_verbose";
static const char __pyx_k_vtkPolyData[] = "vtkPolyData";
static const char __pyx_k_vtkPolyLine[] = "vtkPolyLine";
static const char __pyx_k_GetPointData[] = "GetPointData";
static const char __pyx_k_SetInputData[] = "SetInputData";
static const char __pyx_k_numpy_to_vtk[] = "numpy_to_vtk";
static const char __pyx_k_scipy_sparse[] = "scipy.sparse";
static const char __pyx_k_set_slowness[] = "set_slowness";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_ttcrpy_rgrid[] = "ttcrpy.rgrid";
static const char __pyx_k_vtkCellArray[] = "vtkCellArray";
static const char __pyx_k_vtk_to_numpy[] = "vtk_to_numpy";
static const char __pyx_k_SetDimensions[] = "SetDimensions";
static const char __pyx_k_aggregate_src[] = "aggregate_src";
static const char __pyx_k_cell_slowness[] = "cell_slowness";
static const char __pyx_k_numpy_support[] = "numpy_support";
static const char __pyx_k_save_raypaths[] = "_save_raypaths";
static const char __pyx_k_InsertNextCell[] = "InsertNextCell";
static const char __pyx_k_SetNumberOfIds[] = "SetNumberOfIds";
static const char __pyx_k_set_tilt_angle[] = "set_tilt_angle";
static const char __pyx_k_vtkDoubleArray[] = "vtkDoubleArray";
static const char __pyx_k_GetXCoordinates[] = "GetXCoordinates";
static const char __pyx_k_GetYCoordinates[] = "GetYCoordinates";
static const char __pyx_k_GetZCoordinates[] = "GetZCoordinates";
static const char __pyx_k_P_wave_velocity[] = "P-wave velocity";
static const char __pyx_k_SetXCoordinates[] = "SetXCoordinates";
static const char __pyx_k_SetYCoordinates[] = "SetYCoordinates";
static const char __pyx_k_SetZCoordinates[] = "SetZCoordinates";
static const char __pyx_k_radius_tertiary[] = "radius_tertiary";
static const char __pyx_k_rotated_template[] = "rotated_template";
static const char __pyx_k_ttcrpy_rgrid_pyx[] = "ttcrpy/rgrid.pyx";
static const char __pyx_k_SetNumberOfPoints[] = "SetNumberOfPoints";
static const char __pyx_k_SetNumberOfTuples[] = "SetNumberOfTuples";
static const char __pyx_k_tilted_elliptical[] = "tilted_elliptical";
static const char __pyx_k_v_has_wrong_shape[] = "v has wrong shape";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_constructor_params[] = "constructor_params";
static const char __pyx_k_vtkRectilinearGrid[] = "vtkRectilinearGrid";
static const char __pyx_k_xi_has_wrong_shape[] = "xi has wrong shape";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_SetDataModeToBinary[] = "SetDataModeToBinary";
static const char __pyx_k_get_number_of_cells[] = "get_number_of_cells";
static const char __pyx_k_get_number_of_nodes[] = "get_number_of_nodes";
static const char __pyx_k_Method_0_s_undefined[] = "Method {0:s} undefined";
static const char __pyx_k_vtkXMLPolyDataWriter[] = "vtkXMLPolyDataWriter";
static const char __pyx_k_Receiver_outside_grid[] = "Receiver outside grid";
static const char __pyx_k_SetNumberOfComponents[] = "SetNumberOfComponents";
static const char __pyx_k_theta_has_wrong_shape[] = "theta has wrong shape";
static const char __pyx_k_hypo_should_be_npts_x_5[] = "hypo should be npts x 5";
static const char __pyx_k_v_vector_has_wrong_size[] = "v vector has wrong size";
static const char __pyx_k_Slowness_has_wrong_shape[] = "Slowness has wrong shape";
static const char __pyx_k_velocity_has_wrong_shape[] = "velocity has wrong shape";
static const char __pyx_k_xi_vector_has_wrong_size[] = "xi vector has wrong size";
static const char __pyx_k_Source_point_outside_grid[] = "Source point outside grid";
static const char __pyx_k_data_kernel_straight_rays[] = "data_kernel_straight_rays";
static const char __pyx_k_v_must_be_1D_or_3D_ndarray[] = "v must be 1D or 3D ndarray";
static const char __pyx_k_theta_vector_has_wrong_size[] = "theta vector has wrong size";
static const char __pyx_k_vtkXMLRectilinearGridReader[] = "vtkXMLRectilinearGridReader";
static const char __pyx_k_vtkXMLRectilinearGridWriter[] = "vtkXMLRectilinearGridWriter";
static const char __pyx_k_xi_must_be_1D_or_3D_ndarray[] = "xi must be 1D or 3D ndarray";
static const char __pyx_k_FSM_Grid_cells_must_be_cubic[] = "FSM: Grid cells must be cubic";
static const char __pyx_k_Field_0_s_has_incorrect_ndim[] = "Field {0:s} has incorrect ndim";
static const char __pyx_k_Field_0_s_has_incorrect_size[] = "Field {0:s} has incorrect size";
static const char __pyx_k_Field_0_s_has_incorrect_shape[] = "Field {0:s} has incorrect shape";
static const char __pyx_k_Slowness_vector_has_wrong_size[] = "Slowness vector has wrong size";
static const char __pyx_k_theta_must_be_1D_or_3D_ndarray[] = "theta must be 1D or 3D ndarray";
static const char __pyx_k_velocity_vector_has_wrong_size[] = "velocity vector has wrong size";
static const char __pyx_k_File_should_contain_slowness_or[] = "File should contain slowness or velocity data";
static const char __pyx_k_Raytracing_on_rectilinear_grids[] = "\nRaytracing on rectilinear grids\n\nThis module contains two classes to perform traveltime computation and\nraytracing on rectilinear grids:\n    - `Grid2d` for 2D media\n    - `Grid3d` for 3D media\n\nTwo general algorithms are implemented\n    - the Shortest-Path Method\n    - the Fast-Sweeping Method\nand `Grid3d` also support the Dynamic Shortest-Path Method\n\nSlowness model can be defined in two ways:\n    1) slowness constant within the voxels of the grid (the default)\n    2) slowness defined at nodes of the grid\n\nThis code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )\n";
static const char __pyx_k_compute_M_not_defined_for_grids[] = "compute_M not defined for grids with slowness defined for cells";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_src_and_rcv_should_be_ndata_x_2[] = "src and rcv should be ndata x 2";
static const char __pyx_k_src_and_rcv_should_be_ndata_x_3[] = "src and rcv should be ndata x 3";
static const char __pyx_k_Anisotropic_raytracing_implement[] = "Anisotropic raytracing implemented only for SPM";
static const char __pyx_k_Anisotropy_model_not_implemented[] = "Anisotropy model not implemented";
static const char __pyx_k_Cannot_aggregate_source_with_DSP[] = "Cannot aggregate source with DSPM raytracing";
static const char __pyx_k_Field_0_s_has_incorrect_ndim_1_d[] = "Field {0:s} has incorrect ndim ({1:d})";
static const char __pyx_k_Slowness_must_be_1D_or_3D_ndarra[] = "Slowness must be 1D or 3D ndarray";
static const char __pyx_k_Thread_number_is_larger_than_num[] = "Thread number is larger than number of threads";
static const char __pyx_k_Velocity_data_point_outside_grid[] = "Velocity data point outside grid";
static const char __pyx_k_compute_L_and_compute_M_are_mutu[] = "compute_L and compute_M are mutually exclusive";
static const char __pyx_k_compute_L_defined_only_for_grids[] = "compute_L defined only for grids with slowness defined for cells";
static const char __pyx_k_compute_L_not_implemented_for_VT[] = "compute_L not implemented for VTI media";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_order_value_not_valid_1_or_2_acc[] = "order value not valid (1 or 2 accepted)";
static const char __pyx_k_source_and_rcv_should_be_2D_arra[] = "source and rcv should be 2D arrays";
static const char __pyx_k_source_should_be_either_nsrc_x_2[] = "source should be either nsrc x 2 or 3";
static const char __pyx_k_source_should_be_either_nsrc_x_3[] = "source should be either nsrc x 3, 4 or 5";
static const char __pyx_k_src_and_rcv_should_be_of_equal_s[] = "src and rcv should be of equal size";
static const char __pyx_k_velocity_must_be_1D_or_3D_ndarra[] = "velocity must be 1D or 3D ndarray";
static PyObject *__pyx_n_s_AddArray;
static PyObject *__pyx_kp_u_Anisotropic_raytracing_implement;
static PyObject *__pyx_kp_u_Anisotropy_model_not_implemented;
static PyObject *__pyx_kp_u_Cannot_aggregate_source_with_DSP;
static PyObject *__pyx_n_u_DSPM;
static PyObject *__pyx_n_u_F;
static PyObject *__pyx_n_u_FSM;
static PyObject *__pyx_kp_u_FSM_Grid_cells_must_be_cubic;
static PyObject *__pyx_kp_u_Field_0_s_has_incorrect_ndim;
static PyObject *__pyx_kp_u_Field_0_s_has_incorrect_ndim_1_d;
static PyObject *__pyx_kp_u_Field_0_s_has_incorrect_shape;
static PyObject *__pyx_kp_u_Field_0_s_has_incorrect_size;
static PyObject *__pyx_kp_u_File_should_contain_slowness_or;
static PyObject *__pyx_n_s_GetArray;
static PyObject *__pyx_n_s_GetCellData;
static PyObject *__pyx_n_s_GetOutput;
static PyObject *__pyx_n_s_GetPointData;
static PyObject *__pyx_n_s_GetPointIds;
static PyObject *__pyx_n_s_GetXCoordinates;
static PyObject *__pyx_n_s_GetYCoordinates;
static PyObject *__pyx_n_s_GetZCoordinates;
static PyObject *__pyx_n_s_Grid2d;
static PyObject *__pyx_n_s_Grid3d;
static PyObject *__pyx_n_s_HasArray;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_InsertNextCell;
static PyObject *__pyx_n_s_InsertPoint;
static PyObject *__pyx_n_s_L;
static PyObject *__pyx_kp_u_Method_0_s_undefined;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_kp_u_P_wave_velocity;
static PyObject *__pyx_kp_u_Receiver_outside_grid;
static PyObject *__pyx_n_s_Rx;
static PyObject *__pyx_n_u_SPM;
static PyObject *__pyx_n_s_SetDataModeToBinary;
static PyObject *__pyx_n_s_SetDimensions;
static PyObject *__pyx_n_s_SetFileName;
static PyObject *__pyx_n_s_SetId;
static PyObject *__pyx_n_s_SetInputData;
static PyObject *__pyx_n_s_SetLines;
static PyObject *__pyx_n_s_SetName;
static PyObject *__pyx_n_s_SetNumberOfComponents;
static PyObject *__pyx_n_s_SetNumberOfIds;
static PyObject *__pyx_n_s_SetNumberOfPoints;
static PyObject *__pyx_n_s_SetNumberOfTuples;
static PyObject *__pyx_n_s_SetPoints;
static PyObject *__pyx_n_s_SetTuple1;
static PyObject *__pyx_n_s_SetXCoordinates;
static PyObject *__pyx_n_s_SetYCoordinates;
static PyObject *__pyx_n_s_SetZCoordinates;
static PyObject *__pyx_n_u_Slowness;
static PyObject *__pyx_kp_u_Slowness_has_wrong_shape;
static PyObject *__pyx_kp_u_Slowness_must_be_1D_or_3D_ndarra;
static PyObject *__pyx_kp_u_Slowness_vector_has_wrong_size;
static PyObject *__pyx_kp_u_Source_point_outside_grid;
static PyObject *__pyx_n_s_T;
static PyObject *__pyx_kp_u_Thread_number_is_larger_than_num;
static PyObject *__pyx_n_s_Tx;
static PyObject *__pyx_n_s_Update;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_u_Velocity;
static PyObject *__pyx_kp_u_Velocity_data_point_outside_grid;
static PyObject *__pyx_n_s_Vp0;
static PyObject *__pyx_n_s_Vs0;
static PyObject *__pyx_n_u__34;
static PyObject *__pyx_n_s__61;
static PyObject *__pyx_n_s_abs;
static PyObject *__pyx_n_s_aggregate_src;
static PyObject *__pyx_n_s_aniso;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_b_y;
static PyObject *__pyx_n_s_b_z;
static PyObject *__pyx_n_s_builder;
static PyObject *__pyx_n_s_cell_slowness;
static PyObject *__pyx_n_s_centers;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_compute_L;
static PyObject *__pyx_kp_u_compute_L_and_compute_M_are_mutu;
static PyObject *__pyx_kp_u_compute_L_defined_only_for_grids;
static PyObject *__pyx_kp_u_compute_L_not_implemented_for_VT;
static PyObject *__pyx_n_s_compute_M;
static PyObject *__pyx_kp_u_compute_M_not_defined_for_grids;
static PyObject *__pyx_n_s_constructor_params;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_csr_matrix;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_kernel_straight_rays;
static PyObject *__pyx_n_s_data_p;
static PyObject *__pyx_n_s_delta;
static PyObject *__pyx_n_s_dim;
static PyObject *__pyx_n_s_dl;
static PyObject *__pyx_n_s_dlx;
static PyObject *__pyx_n_s_dly;
static PyObject *__pyx_n_s_dlz;
static PyObject *__pyx_n_s_dtmp;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dx;
static PyObject *__pyx_n_s_dz;
static PyObject *__pyx_n_u_elliptical;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_eps;
static PyObject *__pyx_n_s_epsilon;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_flatten;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_gamma;
static PyObject *__pyx_n_s_get_number_of_cells;
static PyObject *__pyx_n_s_get_number_of_nodes;
static PyObject *__pyx_n_s_grx;
static PyObject *__pyx_n_s_gry;
static PyObject *__pyx_n_s_grz;
static PyObject *__pyx_n_s_hstack;
static PyObject *__pyx_n_s_hypo;
static PyObject *__pyx_kp_u_hypo_should_be_npts_x_5;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_iCell;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_ind;
static PyObject *__pyx_n_s_indc;
static PyObject *__pyx_n_s_indices_p;
static PyObject *__pyx_n_s_indptr_p;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_interp_vel;
static PyObject *__pyx_n_s_is_outside;
static PyObject *__pyx_n_u_iso;
static PyObject *__pyx_n_s_ix;
static PyObject *__pyx_n_s_iy;
static PyObject *__pyx_n_s_iz;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_kron;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_u_lowness;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_m_y;
static PyObject *__pyx_n_s_m_z;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_maxit;
static PyObject *__pyx_n_s_method;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nCells;
static PyObject *__pyx_n_s_nTx;
static PyObject *__pyx_n_s_n_grx;
static PyObject *__pyx_n_s_n_gry;
static PyObject *__pyx_n_s_n_grz;
static PyObject *__pyx_n_s_n_secondary;
static PyObject *__pyx_n_s_n_tertiary;
static PyObject *__pyx_n_s_n_threads;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_nonzero;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nsnx;
static PyObject *__pyx_n_s_nsny;
static PyObject *__pyx_n_s_nsnz;
static PyObject *__pyx_n_s_nt;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_numpy_support;
static PyObject *__pyx_n_s_numpy_to_vtk;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_kp_u_order_value_not_valid_1_or_2_acc;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_radius_tertiary;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rays;
static PyObject *__pyx_n_s_rcv;
static PyObject *__pyx_n_s_reader;
static PyObject *__pyx_n_s_rebuild2d;
static PyObject *__pyx_n_s_rebuild3d;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_return_rays;
static PyObject *__pyx_n_s_rotated_template;
static PyObject *__pyx_n_s_save_raypaths;
static PyObject *__pyx_n_s_scipy_sparse;
static PyObject *__pyx_n_s_set_Vp0;
static PyObject *__pyx_n_s_set_Vs0;
static PyObject *__pyx_n_s_set_delta;
static PyObject *__pyx_n_s_set_epsilon;
static PyObject *__pyx_n_s_set_gamma;
static PyObject *__pyx_n_s_set_slowness;
static PyObject *__pyx_n_s_set_tilt_angle;
static PyObject *__pyx_n_s_set_verbose;
static PyObject *__pyx_n_s_set_xi;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_slowness;
static PyObject *__pyx_n_u_slowness;
static PyObject *__pyx_n_s_small;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_kp_u_source_and_rcv_should_be_2D_arra;
static PyObject *__pyx_kp_u_source_should_be_either_nsrc_x_2;
static PyObject *__pyx_kp_u_source_should_be_either_nsrc_x_3;
static PyObject *__pyx_n_s_sp;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_ndata_x_2;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_ndata_x_3;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_of_equal_s;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_sy;
static PyObject *__pyx_n_s_sz;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theta;
static PyObject *__pyx_kp_u_theta_has_wrong_shape;
static PyObject *__pyx_kp_u_theta_must_be_1D_or_3D_ndarray;
static PyObject *__pyx_kp_u_theta_vector_has_wrong_size;
static PyObject *__pyx_n_s_thread_no;
static PyObject *__pyx_n_s_tile;
static PyObject *__pyx_n_u_tilted_elliptical;
static PyObject *__pyx_n_s_tt_from_rp;
static PyObject *__pyx_n_s_ttcrpy_rgrid;
static PyObject *__pyx_kp_s_ttcrpy_rgrid_pyx;
static PyObject *__pyx_n_s_unique;
static PyObject *__pyx_n_s_up;
static PyObject *__pyx_n_s_up_y;
static PyObject *__pyx_n_s_up_z;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_kp_u_v_has_wrong_shape;
static PyObject *__pyx_kp_u_v_must_be_1D_or_3D_ndarray;
static PyObject *__pyx_kp_u_v_vector_has_wrong_size;
static PyObject *__pyx_n_u_velocity;
static PyObject *__pyx_kp_u_velocity_has_wrong_shape;
static PyObject *__pyx_kp_u_velocity_must_be_1D_or_3D_ndarra;
static PyObject *__pyx_kp_u_velocity_vector_has_wrong_size;
static PyObject *__pyx_n_s_vstack;
static PyObject *__pyx_n_u_vti_psv;
static PyObject *__pyx_n_u_vti_sh;
static PyObject *__pyx_n_s_vtk;
static PyObject *__pyx_n_s_vtkCellArray;
static PyObject *__pyx_n_s_vtkDoubleArray;
static PyObject *__pyx_n_s_vtkPoints;
static PyObject *__pyx_n_s_vtkPolyData;
static PyObject *__pyx_n_s_vtkPolyLine;
static PyObject *__pyx_n_s_vtkRectilinearGrid;
static PyObject *__pyx_n_s_vtkXMLPolyDataWriter;
static PyObject *__pyx_n_s_vtkXMLRectilinearGridReader;
static PyObject *__pyx_n_s_vtkXMLRectilinearGridWriter;
static PyObject *__pyx_n_s_vtk_to_numpy;
static PyObject *__pyx_n_s_vtk_util;
static PyObject *__pyx_kp_u_vtp;
static PyObject *__pyx_kp_u_vtr;
static PyObject *__pyx_n_s_weno;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_x1;
static PyObject *__pyx_n_s_x2;
static PyObject *__pyx_n_s_xc;
static PyObject *__pyx_n_s_xe;
static PyObject *__pyx_n_s_xi;
static PyObject *__pyx_kp_u_xi_has_wrong_shape;
static PyObject *__pyx_kp_u_xi_must_be_1D_or_3D_ndarray;
static PyObject *__pyx_kp_u_xi_vector_has_wrong_size;
static PyObject *__pyx_n_s_xr;
static PyObject *__pyx_n_s_xs;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_y1;
static PyObject *__pyx_n_s_y2;
static PyObject *__pyx_n_s_yc;
static PyObject *__pyx_n_s_ye;
static PyObject *__pyx_n_s_yi;
static PyObject *__pyx_n_s_z;
static PyObject *__pyx_n_s_z1;
static PyObject *__pyx_n_s_z2;
static PyObject *__pyx_n_s_zc;
static PyObject *__pyx_n_s_ze;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_n_s_zi;
static PyObject *__pyx_n_s_zr;
static PyObject *__pyx_n_s_zs;
static PyObject *__pyx_pf_6ttcrpy_5rgrid_set_verbose(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static int __pyx_pf_6ttcrpy_5rgrid_6Grid3d___cinit__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_weno, uint32_t __pyx_v_nsnx, uint32_t __pyx_v_nsny, uint32_t __pyx_v_nsnz, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary); /* proto */
static void __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2__dealloc__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_4__reduce__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1x___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1y___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1z___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dx___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dy___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dz___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_5shape___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_7nparams___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_6set_use_thread_pool(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_use_thread_pool); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_8get_number_of_nodes(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_10get_number_of_cells(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_12get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_thread_no); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_14ind(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_16indc(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_18is_outside(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyArrayObject *__pyx_v_pts); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_20get_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_22set_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_24set_velocity(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_velocity); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_26compute_D(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_coord); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_28compute_K(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_30get_s0(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_hypo, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_32raytrace(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_compute_M, PyObject *__pyx_v_return_rays); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_34to_vtk(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_36_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_38builder(PyObject *__pyx_v_filename, PyObject *__pyx_v_n_threads, PyObject *__pyx_v_method, PyObject *__pyx_v_tt_from_rp, PyObject *__pyx_v_interp_vel, PyObject *__pyx_v_eps, PyObject *__pyx_v_maxit, PyObject *__pyx_v_weno, PyObject *__pyx_v_nsnx, PyObject *__pyx_v_nsny, PyObject *__pyx_v_nsnz, PyObject *__pyx_v_n_secondary, PyObject *__pyx_v_n_tertiary, PyObject *__pyx_v_radius_tertiary); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays(PyArrayObject *__pyx_v_Tx, PyArrayObject *__pyx_v_Rx, PyArrayObject *__pyx_v_grx, PyArrayObject *__pyx_v_gry, PyArrayObject *__pyx_v_grz, PyObject *__pyx_v_centers); /* proto */
static int __pyx_pf_6ttcrpy_5rgrid_6Grid2d___cinit__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_z, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, PyObject *__pyx_v_aniso, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_weno, bool __pyx_v_rotated_template, uint32_t __pyx_v_nsnx, uint32_t __pyx_v_nsnz); /* proto */
static void __pyx_pf_6ttcrpy_5rgrid_6Grid2d_2__dealloc__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_4__reduce__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_1x___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_1z___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dx___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dz___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_5shape___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_7nparams___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_6get_number_of_nodes(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_8get_number_of_cells(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_10get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_thread_no); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_12is_outside(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyArrayObject *__pyx_v_pts); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_14get_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_16set_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_18set_velocity(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_velocity); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_20set_xi(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_xi); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_22set_tilt_angle(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_theta); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_24set_Vp0(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_26set_Vs0(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_28set_delta(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_30set_epsilon(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_32set_gamma(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_34compute_K(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_order); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_36raytrace(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_xi, PyObject *__pyx_v_theta, PyObject *__pyx_v_Vp0, PyObject *__pyx_v_Vs0, PyObject *__pyx_v_delta, PyObject *__pyx_v_epsilon, PyObject *__pyx_v_gamma, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_return_rays); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_38to_vtk(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_40_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays(PyArrayObject *__pyx_v_Tx, PyArrayObject *__pyx_v_Rx, PyArrayObject *__pyx_v_grx, PyArrayObject *__pyx_v_grz, PyObject *__pyx_v_aniso); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_2_rebuild3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y, PyObject *__pyx_v_z, PyObject *__pyx_v_constructor_params); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5rgrid_4_rebuild2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_z, PyObject *__pyx_v_constructor_params); /* proto */
static PyObject *__pyx_tp_new_6ttcrpy_5rgrid_Grid3d(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ttcrpy_5rgrid_Grid2d(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_1_;
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_1_eneg_6;
static PyObject *__pyx_float_1_eneg_15;
static PyObject *__pyx_float_0_000001;
static PyObject *__pyx_float_neg_2_;
static PyObject *__pyx_float_neg_0_5;
static PyObject *__pyx_float_neg_2_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__3;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__17;
static PyObject *__pyx_slice__24;
static PyObject *__pyx_slice__37;
static PyObject *__pyx_slice__38;
static PyObject *__pyx_slice__55;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__73;
/* Late includes */

/* "ttcrpy/rgrid.pyx":38
 *     void setVerbose(int)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """
 *     Set verbosity level for C++ code
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_1set_verbose(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_set_verbose[] = "\n    Set verbosity level for C++ code\n\n    Parameters\n    ----------\n    v: int\n        verbosity level\n    ";
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_1set_verbose = {"set_verbose", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_1set_verbose, METH_O, __pyx_doc_6ttcrpy_5rgrid_set_verbose};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_1set_verbose(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_verbose (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_set_verbose(__pyx_self, ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_set_verbose(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_verbose", 0);

  /* "ttcrpy/rgrid.pyx":47
 *         verbosity level
 *     """
 *     setVerbose(v)             # <<<<<<<<<<<<<<
 * 
 * cdef class Grid3d:
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_v); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L1_error)
  ttcr::setVerbose(__pyx_t_1);

  /* "ttcrpy/rgrid.pyx":38
 *     void setVerbose(int)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """
 *     Set verbosity level for C++ code
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.set_verbose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":143
 *     cdef Grid3D[double, uint32_t]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=1] x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.double_t, ndim=1] y,
 *                   np.ndarray[np.double_t, ndim=1] z,
 */

/* Python wrapper */
static int __pyx_pw_6ttcrpy_5rgrid_6Grid3d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ttcrpy_5rgrid_6Grid3d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_z = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  bool __pyx_v_tt_from_rp;
  bool __pyx_v_interp_vel;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  bool __pyx_v_weno;
  uint32_t __pyx_v_nsnx;
  uint32_t __pyx_v_nsny;
  uint32_t __pyx_v_nsnz;
  uint32_t __pyx_v_n_secondary;
  uint32_t __pyx_v_n_tertiary;
  double __pyx_v_radius_tertiary;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_tt_from_rp,&__pyx_n_s_interp_vel,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_weno,&__pyx_n_s_nsnx,&__pyx_n_s_nsny,&__pyx_n_s_nsnz,&__pyx_n_s_n_secondary,&__pyx_n_s_n_tertiary,&__pyx_n_s_radius_tertiary,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[5] = ((PyObject*)__pyx_n_u_FSM);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, 1); __PYX_ERR(0, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, 2); __PYX_ERR(0, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt_from_rp);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interp_vel);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_weno);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnx);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsny);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnz);
          if (value) { values[13] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[14] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_tertiary);
          if (value) { values[15] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius_tertiary);
          if (value) { values[16] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 143, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_z = ((PyArrayObject *)values[2]);
    if (values[3]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[4]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[5]);
    if (values[6]) {
      __pyx_v_tt_from_rp = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_tt_from_rp == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L3_error)
    } else {
      __pyx_v_tt_from_rp = ((bool)1);
    }
    if (values[7]) {
      __pyx_v_interp_vel = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_interp_vel == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L3_error)
    } else {
      __pyx_v_interp_vel = ((bool)0);
    }
    if (values[8]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[9]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[10]) {
      __pyx_v_weno = __Pyx_PyObject_IsTrue(values[10]); if (unlikely((__pyx_v_weno == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
    } else {
      __pyx_v_weno = ((bool)1);
    }
    if (values[11]) {
      __pyx_v_nsnx = __Pyx_PyInt_As_uint32_t(values[11]); if (unlikely((__pyx_v_nsnx == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L3_error)
    } else {
      __pyx_v_nsnx = ((uint32_t)5);
    }
    if (values[12]) {
      __pyx_v_nsny = __Pyx_PyInt_As_uint32_t(values[12]); if (unlikely((__pyx_v_nsny == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L3_error)
    } else {
      __pyx_v_nsny = ((uint32_t)5);
    }
    if (values[13]) {
      __pyx_v_nsnz = __Pyx_PyInt_As_uint32_t(values[13]); if (unlikely((__pyx_v_nsnz == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L3_error)
    } else {
      __pyx_v_nsnz = ((uint32_t)5);
    }
    if (values[14]) {
      __pyx_v_n_secondary = __Pyx_PyInt_As_uint32_t(values[14]); if (unlikely((__pyx_v_n_secondary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 151, __pyx_L3_error)
    } else {
      __pyx_v_n_secondary = ((uint32_t)2);
    }
    if (values[15]) {
      __pyx_v_n_tertiary = __Pyx_PyInt_As_uint32_t(values[15]); if (unlikely((__pyx_v_n_tertiary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 151, __pyx_L3_error)
    } else {
      __pyx_v_n_tertiary = ((uint32_t)2);
    }
    if (values[16]) {
      __pyx_v_radius_tertiary = __pyx_PyFloat_AsDouble(values[16]); if (unlikely((__pyx_v_radius_tertiary == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 152, __pyx_L3_error)
    } else {
      __pyx_v_radius_tertiary = ((double)1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 143, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 143, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 144, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 145, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 147, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d___cinit__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_nsnx, __pyx_v_nsny, __pyx_v_nsnz, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_radius_tertiary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ttcrpy_5rgrid_6Grid3d___cinit__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_weno, uint32_t __pyx_v_nsnx, uint32_t __pyx_v_nsny, uint32_t __pyx_v_nsnz, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary) {
  uint32_t __pyx_v_nx;
  uint32_t __pyx_v_ny;
  uint32_t __pyx_v_nz;
  double __pyx_v_xmin;
  double __pyx_v_ymin;
  double __pyx_v_zmin;
  PyObject *__pyx_v_val = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  double __pyx_t_13;
  ttcr::Grid3Drcfs<double,uint32_t>  *__pyx_t_14;
  ttcr::Grid3Drcsp<double,uint32_t,ttcr::Cell<double,ttcr::Node3Dcsp<double,uint32_t> ,ttcr::sxyz<double> > >  *__pyx_t_15;
  ttcr::Grid3Drcdsp<double,uint32_t,ttcr::Cell<double,ttcr::Node3Dc<double,uint32_t> ,ttcr::sxyz<double> > >  *__pyx_t_16;
  ttcr::Grid3Drnfs<double,uint32_t>  *__pyx_t_17;
  ttcr::Grid3Drnsp<double,uint32_t>  *__pyx_t_18;
  ttcr::Grid3Drndsp<double,uint32_t>  *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 143, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 143, __pyx_L1_error)
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 143, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "ttcrpy/rgrid.pyx":154
 *                   double radius_tertiary=1.0):
 * 
 *         cdef uint32_t nx = x.size-1             # <<<<<<<<<<<<<<
 *         cdef uint32_t ny = y.size-1
 *         cdef uint32_t nz = z.size-1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_x), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nx = __pyx_t_3;

  /* "ttcrpy/rgrid.pyx":155
 * 
 *         cdef uint32_t nx = x.size-1
 *         cdef uint32_t ny = y.size-1             # <<<<<<<<<<<<<<
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ny = __pyx_t_3;

  /* "ttcrpy/rgrid.pyx":156
 *         cdef uint32_t nx = x.size-1
 *         cdef uint32_t ny = y.size-1
 *         cdef uint32_t nz = z.size-1             # <<<<<<<<<<<<<<
 *         self._dx = x[1] - x[0]
 *         self._dy = y[1] - y[0]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_z), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nz = __pyx_t_3;

  /* "ttcrpy/rgrid.pyx":157
 *         cdef uint32_t ny = y.size-1
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]             # <<<<<<<<<<<<<<
 *         self._dy = y[1] - y[0]
 *         self._dz = z[1] - z[0]
 */
  __pyx_t_4 = 1;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 157, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 157, __pyx_L1_error)
  }
  __pyx_v_self->_dx = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_x.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides)));

  /* "ttcrpy/rgrid.pyx":158
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]
 *         self._dy = y[1] - y[0]             # <<<<<<<<<<<<<<
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]
 */
  __pyx_t_6 = 1;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_y.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_y.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 158, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_y.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_y.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 158, __pyx_L1_error)
  }
  __pyx_v_self->_dy = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_y.diminfo[0].strides)));

  /* "ttcrpy/rgrid.pyx":159
 *         self._dx = x[1] - x[0]
 *         self._dy = y[1] - y[0]
 *         self._dz = z[1] - z[0]             # <<<<<<<<<<<<<<
 *         cdef double xmin = x[0]
 *         cdef double ymin = y[0]
 */
  __pyx_t_4 = 1;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 159, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 159, __pyx_L1_error)
  }
  __pyx_v_self->_dz = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_z.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides)));

  /* "ttcrpy/rgrid.pyx":160
 *         self._dy = y[1] - y[0]
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]             # <<<<<<<<<<<<<<
 *         cdef double ymin = y[0]
 *         cdef double zmin = z[0]
 */
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 160, __pyx_L1_error)
  }
  __pyx_v_xmin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides));

  /* "ttcrpy/rgrid.pyx":161
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]
 *         cdef double ymin = y[0]             # <<<<<<<<<<<<<<
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness
 */
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_y.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_y.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 161, __pyx_L1_error)
  }
  __pyx_v_ymin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides));

  /* "ttcrpy/rgrid.pyx":162
 *         cdef double xmin = x[0]
 *         cdef double ymin = y[0]
 *         cdef double zmin = z[0]             # <<<<<<<<<<<<<<
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 */
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 162, __pyx_L1_error)
  }
  __pyx_v_zmin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides));

  /* "ttcrpy/rgrid.pyx":163
 *         cdef double ymin = y[0]
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness             # <<<<<<<<<<<<<<
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp
 */
  __pyx_v_self->cell_slowness = __pyx_v_cell_slowness;

  /* "ttcrpy/rgrid.pyx":164
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads             # <<<<<<<<<<<<<<
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel
 */
  __pyx_v_self->_n_threads = __pyx_v_n_threads;

  /* "ttcrpy/rgrid.pyx":165
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp             # <<<<<<<<<<<<<<
 *         self.interp_vel = interp_vel
 *         self.eps = eps
 */
  __pyx_v_self->tt_from_rp = __pyx_v_tt_from_rp;

  /* "ttcrpy/rgrid.pyx":166
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel             # <<<<<<<<<<<<<<
 *         self.eps = eps
 *         self.maxit = maxit
 */
  __pyx_v_self->interp_vel = __pyx_v_interp_vel;

  /* "ttcrpy/rgrid.pyx":167
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel
 *         self.eps = eps             # <<<<<<<<<<<<<<
 *         self.maxit = maxit
 *         self.weno = weno
 */
  __pyx_v_self->eps = __pyx_v_eps;

  /* "ttcrpy/rgrid.pyx":168
 *         self.interp_vel = interp_vel
 *         self.eps = eps
 *         self.maxit = maxit             # <<<<<<<<<<<<<<
 *         self.weno = weno
 *         self.nsnx = nsnx
 */
  __pyx_v_self->maxit = __pyx_v_maxit;

  /* "ttcrpy/rgrid.pyx":169
 *         self.eps = eps
 *         self.maxit = maxit
 *         self.weno = weno             # <<<<<<<<<<<<<<
 *         self.nsnx = nsnx
 *         self.nsny = nsny
 */
  __pyx_v_self->weno = __pyx_v_weno;

  /* "ttcrpy/rgrid.pyx":170
 *         self.maxit = maxit
 *         self.weno = weno
 *         self.nsnx = nsnx             # <<<<<<<<<<<<<<
 *         self.nsny = nsny
 *         self.nsnz = nsnz
 */
  __pyx_v_self->nsnx = __pyx_v_nsnx;

  /* "ttcrpy/rgrid.pyx":171
 *         self.weno = weno
 *         self.nsnx = nsnx
 *         self.nsny = nsny             # <<<<<<<<<<<<<<
 *         self.nsnz = nsnz
 *         self.n_secondary = n_secondary
 */
  __pyx_v_self->nsny = __pyx_v_nsny;

  /* "ttcrpy/rgrid.pyx":172
 *         self.nsnx = nsnx
 *         self.nsny = nsny
 *         self.nsnz = nsnz             # <<<<<<<<<<<<<<
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary
 */
  __pyx_v_self->nsnz = __pyx_v_nsnz;

  /* "ttcrpy/rgrid.pyx":173
 *         self.nsny = nsny
 *         self.nsnz = nsnz
 *         self.n_secondary = n_secondary             # <<<<<<<<<<<<<<
 *         self.n_tertiary = n_tertiary
 *         self.radius_tertiary = radius_tertiary
 */
  __pyx_v_self->n_secondary = __pyx_v_n_secondary;

  /* "ttcrpy/rgrid.pyx":174
 *         self.nsnz = nsnz
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary             # <<<<<<<<<<<<<<
 *         self.radius_tertiary = radius_tertiary
 * 
 */
  __pyx_v_self->n_tertiary = __pyx_v_n_tertiary;

  /* "ttcrpy/rgrid.pyx":175
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary
 *         self.radius_tertiary = radius_tertiary             # <<<<<<<<<<<<<<
 * 
 *         if method == 'FSM':
 */
  __pyx_v_self->radius_tertiary = __pyx_v_radius_tertiary;

  /* "ttcrpy/rgrid.pyx":177
 *         self.radius_tertiary = radius_tertiary
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:
 *                 raise ValueError('FSM: Grid cells must be cubic')
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "ttcrpy/rgrid.pyx":178
 * 
 *         if method == 'FSM':
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:             # <<<<<<<<<<<<<<
 *                 raise ValueError('FSM: Grid cells must be cubic')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_abs); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_dx - __pyx_v_self->_dy)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_2, __pyx_float_0_000001, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (!__pyx_t_7) {
    } else {
      __pyx_t_8 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_abs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->_dx - __pyx_v_self->_dz)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_9, __pyx_float_0_000001, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_t_7;
    __pyx_L5_bool_binop_done:;
    if (unlikely(__pyx_t_8)) {

      /* "ttcrpy/rgrid.pyx":179
 *         if method == 'FSM':
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:
 *                 raise ValueError('FSM: Grid cells must be cubic')             # <<<<<<<<<<<<<<
 * 
 *         for val in x:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 179, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":178
 * 
 *         if method == 'FSM':
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:             # <<<<<<<<<<<<<<
 *                 raise ValueError('FSM: Grid cells must be cubic')
 * 
 */
    }

    /* "ttcrpy/rgrid.pyx":177
 *         self.radius_tertiary = radius_tertiary
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:
 *                 raise ValueError('FSM: Grid cells must be cubic')
 */
  }

  /* "ttcrpy/rgrid.pyx":181
 *                 raise ValueError('FSM: Grid cells must be cubic')
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             self._x.push_back(val)
 *         for val in y:
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_x))) || PyTuple_CheckExact(((PyObject *)__pyx_v_x))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_x); __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 181, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_12(__pyx_t_1);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 181, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":182
 * 
 *         for val in x:
 *             self._x.push_back(val)             # <<<<<<<<<<<<<<
 *         for val in y:
 *             self._y.push_back(val)
 */
    __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
    try {
      __pyx_v_self->_x.push_back(__pyx_t_13);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 182, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":181
 *                 raise ValueError('FSM: Grid cells must be cubic')
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             self._x.push_back(val)
 *         for val in y:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":183
 *         for val in x:
 *             self._x.push_back(val)
 *         for val in y:             # <<<<<<<<<<<<<<
 *             self._y.push_back(val)
 *         for val in z:
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_y))) || PyTuple_CheckExact(((PyObject *)__pyx_v_y))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_y); __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 183, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_12(__pyx_t_1);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 183, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":184
 *             self._x.push_back(val)
 *         for val in y:
 *             self._y.push_back(val)             # <<<<<<<<<<<<<<
 *         for val in z:
 *             self._z.push_back(val)
 */
    __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
    try {
      __pyx_v_self->_y.push_back(__pyx_t_13);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 184, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":183
 *         for val in x:
 *             self._x.push_back(val)
 *         for val in y:             # <<<<<<<<<<<<<<
 *             self._y.push_back(val)
 *         for val in z:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":185
 *         for val in y:
 *             self._y.push_back(val)
 *         for val in z:             # <<<<<<<<<<<<<<
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_z))) || PyTuple_CheckExact(((PyObject *)__pyx_v_z))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_z); __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 185, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_9); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_12(__pyx_t_1);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 185, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":186
 *             self._y.push_back(val)
 *         for val in z:
 *             self._z.push_back(val)             # <<<<<<<<<<<<<<
 *         self._x.shrink_to_fit()
 *         self._y.shrink_to_fit()
 */
    __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 186, __pyx_L1_error)
    try {
      __pyx_v_self->_z.push_back(__pyx_t_13);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 186, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":185
 *         for val in y:
 *             self._y.push_back(val)
 *         for val in z:             # <<<<<<<<<<<<<<
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":187
 *         for val in z:
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()             # <<<<<<<<<<<<<<
 *         self._y.shrink_to_fit()
 *         self._z.shrink_to_fit()
 */
  __pyx_v_self->_x.shrink_to_fit();

  /* "ttcrpy/rgrid.pyx":188
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 *         self._y.shrink_to_fit()             # <<<<<<<<<<<<<<
 *         self._z.shrink_to_fit()
 * 
 */
  __pyx_v_self->_y.shrink_to_fit();

  /* "ttcrpy/rgrid.pyx":189
 *         self._x.shrink_to_fit()
 *         self._y.shrink_to_fit()
 *         self._z.shrink_to_fit()             # <<<<<<<<<<<<<<
 * 
 *         if cell_slowness:
 */
  __pyx_v_self->_z.shrink_to_fit();

  /* "ttcrpy/rgrid.pyx":191
 *         self._z.shrink_to_fit()
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
  __pyx_t_8 = (__pyx_v_cell_slowness != 0);
  if (__pyx_t_8) {

    /* "ttcrpy/rgrid.pyx":192
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drcfs[double,uint32_t](nx, ny, nz, self._dx,
 */
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 192, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (__pyx_t_7) {

      /* "ttcrpy/rgrid.pyx":193
 *         if cell_slowness:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drcfs[double,uint32_t](nx, ny, nz, self._dx,
 *                                                             xmin, ymin, zmin,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/rgrid.pyx":194
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drcfs[double,uint32_t](nx, ny, nz, self._dx,             # <<<<<<<<<<<<<<
 *                                                             xmin, ymin, zmin,
 *                                                             eps, maxit, weno,
 */
      try {
        __pyx_t_14 = new ttcr::Grid3Drcfs<double,uint32_t> (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 194, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_14;

      /* "ttcrpy/rgrid.pyx":192
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drcfs[double,uint32_t](nx, ny, nz, self._dx,
 */
      goto __pyx_L14;
    }

    /* "ttcrpy/rgrid.pyx":199
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Drcsp[double,uint32_t,Cell[double,Node3Dcsp[double,uint32_t],sxyz[double]]](nx, ny, nz,
 */
    __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 199, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "ttcrpy/rgrid.pyx":200
 *                                                             n_threads)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drcsp[double,uint32_t,Cell[double,Node3Dcsp[double,uint32_t],sxyz[double]]](nx, ny, nz,
 *                                                             self._dx, self._dy, self._dz,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/rgrid.pyx":201
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid3Drcsp[double,uint32_t,Cell[double,Node3Dcsp[double,uint32_t],sxyz[double]]](nx, ny, nz,             # <<<<<<<<<<<<<<
 *                                                             self._dx, self._dy, self._dz,
 *                                                             xmin, ymin, zmin,
 */
      try {
        __pyx_t_15 = new ttcr::Grid3Drcsp<double,uint32_t,ttcr::Cell<double,ttcr::Node3Dcsp<double,uint32_t> ,ttcr::sxyz<double> > > (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dy, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsny, __pyx_v_nsnz, __pyx_v_tt_from_rp, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 201, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_15;

      /* "ttcrpy/rgrid.pyx":199
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Drcsp[double,uint32_t,Cell[double,Node3Dcsp[double,uint32_t],sxyz[double]]](nx, ny, nz,
 */
      goto __pyx_L14;
    }

    /* "ttcrpy/rgrid.pyx":207
 *                                                             tt_from_rp,
 *                                                             n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drcdsp[double,uint32_t,Cell[double,Node3Dc[double,uint32_t],sxyz[double]]](nx, ny, nz,
 */
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_DSPM, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (likely(__pyx_t_7)) {

      /* "ttcrpy/rgrid.pyx":208
 *                                                             n_threads)
 *             elif method == 'DSPM':
 *                 self.method = b'd'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drcdsp[double,uint32_t,Cell[double,Node3Dc[double,uint32_t],sxyz[double]]](nx, ny, nz,
 *                                                            self._dx, self._dy, self._dz,
 */
      __pyx_v_self->method = 'd';

      /* "ttcrpy/rgrid.pyx":209
 *             elif method == 'DSPM':
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drcdsp[double,uint32_t,Cell[double,Node3Dc[double,uint32_t],sxyz[double]]](nx, ny, nz,             # <<<<<<<<<<<<<<
 *                                                            self._dx, self._dy, self._dz,
 *                                                            xmin, ymin, zmin,
 */
      try {
        __pyx_t_16 = new ttcr::Grid3Drcdsp<double,uint32_t,ttcr::Cell<double,ttcr::Node3Dc<double,uint32_t> ,ttcr::sxyz<double> > > (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dy, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_n_secondary, __pyx_v_tt_from_rp, __pyx_v_n_tertiary, __pyx_v_radius_tertiary, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 209, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_16;

      /* "ttcrpy/rgrid.pyx":207
 *                                                             tt_from_rp,
 *                                                             n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drcdsp[double,uint32_t,Cell[double,Node3Dc[double,uint32_t],sxyz[double]]](nx, ny, nz,
 */
      goto __pyx_L14;
    }

    /* "ttcrpy/rgrid.pyx":216
 *                                                            n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 *         else:
 *             if method == 'FSM':
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 216, __pyx_L1_error)
    }
    __pyx_L14:;

    /* "ttcrpy/rgrid.pyx":191
 *         self._z.shrink_to_fit()
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
    goto __pyx_L13;
  }

  /* "ttcrpy/rgrid.pyx":218
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drnfs[double,uint32_t](nx, ny, nz, self._dx,
 */
  /*else*/ {
    __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 218, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "ttcrpy/rgrid.pyx":219
 *         else:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drnfs[double,uint32_t](nx, ny, nz, self._dx,
 *                                                             xmin, ymin, zmin,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/rgrid.pyx":220
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drnfs[double,uint32_t](nx, ny, nz, self._dx,             # <<<<<<<<<<<<<<
 *                                                             xmin, ymin, zmin,
 *                                                             eps, maxit, weno,
 */
      try {
        __pyx_t_17 = new ttcr::Grid3Drnfs<double,uint32_t> (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 220, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_17;

      /* "ttcrpy/rgrid.pyx":218
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Drnfs[double,uint32_t](nx, ny, nz, self._dx,
 */
      goto __pyx_L15;
    }

    /* "ttcrpy/rgrid.pyx":225
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Drnsp[double,uint32_t](nx, ny, nz,
 */
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 225, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (__pyx_t_7) {

      /* "ttcrpy/rgrid.pyx":226
 *                                                             n_threads)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drnsp[double,uint32_t](nx, ny, nz,
 *                                                             self._dx, self._dy, self._dz,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/rgrid.pyx":227
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid3Drnsp[double,uint32_t](nx, ny, nz,             # <<<<<<<<<<<<<<
 *                                                             self._dx, self._dy, self._dz,
 *                                                             xmin, ymin, zmin,
 */
      try {
        __pyx_t_18 = new ttcr::Grid3Drnsp<double,uint32_t> (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dy, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsny, __pyx_v_nsnz, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 227, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_18;

      /* "ttcrpy/rgrid.pyx":225
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Drnsp[double,uint32_t](nx, ny, nz,
 */
      goto __pyx_L15;
    }

    /* "ttcrpy/rgrid.pyx":233
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drndsp[double,uint32_t](nx, ny, nz,
 */
    __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_DSPM, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 233, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (likely(__pyx_t_8)) {

      /* "ttcrpy/rgrid.pyx":234
 *                                                             n_threads)
 *             elif method == 'DSPM':
 *                 self.method = b'd'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Drndsp[double,uint32_t](nx, ny, nz,
 *                                                              self._dx, self._dy, self._dz,
 */
      __pyx_v_self->method = 'd';

      /* "ttcrpy/rgrid.pyx":235
 *             elif method == 'DSPM':
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drndsp[double,uint32_t](nx, ny, nz,             # <<<<<<<<<<<<<<
 *                                                              self._dx, self._dy, self._dz,
 *                                                              xmin, ymin, zmin,
 */
      try {
        __pyx_t_19 = new ttcr::Grid3Drndsp<double,uint32_t> (__pyx_v_nx, __pyx_v_ny, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dy, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_ymin, __pyx_v_zmin, __pyx_v_n_secondary, __pyx_v_tt_from_rp, __pyx_v_n_tertiary, __pyx_v_radius_tertiary, __pyx_v_interp_vel, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 235, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_19;

      /* "ttcrpy/rgrid.pyx":233
 *                                                             tt_from_rp, interp_vel,
 *                                                             n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Drndsp[double,uint32_t](nx, ny, nz,
 */
      goto __pyx_L15;
    }

    /* "ttcrpy/rgrid.pyx":242
 *                                                              interp_vel, n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_9 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 242, __pyx_L1_error)
    }
    __pyx_L15:;
  }
  __pyx_L13:;

  /* "ttcrpy/rgrid.pyx":143
 *     cdef Grid3D[double, uint32_t]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=1] x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.double_t, ndim=1] y,
 *                   np.ndarray[np.double_t, ndim=1] z,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":244
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

/* Python wrapper */
static void __pyx_pw_6ttcrpy_5rgrid_6Grid3d_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6ttcrpy_5rgrid_6Grid3d_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2__dealloc__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2__dealloc__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ttcrpy/rgrid.pyx":245
 * 
 *     def __dealloc__(self):
 *         del self.grid             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  delete __pyx_v_self->grid;

  /* "ttcrpy/rgrid.pyx":244
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ttcrpy/rgrid.pyx":247
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_4__reduce__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_4__reduce__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_constructor_params = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ttcrpy/rgrid.pyx":248
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
  switch (__pyx_v_self->method) {
    case 'f':

    /* "ttcrpy/rgrid.pyx":249
 *     def __reduce__(self):
 *         if self.method == b'f':
 *             method = 'FSM'             # <<<<<<<<<<<<<<
 *         elif self.method == b's':
 *             method = 'SPM'
 */
    __Pyx_INCREF(__pyx_n_u_FSM);
    __pyx_v_method = __pyx_n_u_FSM;

    /* "ttcrpy/rgrid.pyx":248
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
    break;
    case 's':

    /* "ttcrpy/rgrid.pyx":251
 *             method = 'FSM'
 *         elif self.method == b's':
 *             method = 'SPM'             # <<<<<<<<<<<<<<
 *         elif self.method == b'd':
 *             method = 'DSPM'
 */
    __Pyx_INCREF(__pyx_n_u_SPM);
    __pyx_v_method = __pyx_n_u_SPM;

    /* "ttcrpy/rgrid.pyx":250
 *         if self.method == b'f':
 *             method = 'FSM'
 *         elif self.method == b's':             # <<<<<<<<<<<<<<
 *             method = 'SPM'
 *         elif self.method == b'd':
 */
    break;
    case 'd':

    /* "ttcrpy/rgrid.pyx":253
 *             method = 'SPM'
 *         elif self.method == b'd':
 *             method = 'DSPM'             # <<<<<<<<<<<<<<
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 */
    __Pyx_INCREF(__pyx_n_u_DSPM);
    __pyx_v_method = __pyx_n_u_DSPM;

    /* "ttcrpy/rgrid.pyx":252
 *         elif self.method == b's':
 *             method = 'SPM'
 *         elif self.method == b'd':             # <<<<<<<<<<<<<<
 *             method = 'DSPM'
 * 
 */
    break;
    default: break;
  }

  /* "ttcrpy/rgrid.pyx":255
 *             method = 'DSPM'
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,             # <<<<<<<<<<<<<<
 *                               self.tt_from_rp, self.interp_vel, self.eps,
 *                               self.maxit, self.weno, self.nsnx, self.nsny,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->cell_slowness); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_method)) { __Pyx_RaiseUnboundLocalError("method"); __PYX_ERR(0, 255, __pyx_L1_error) }

  /* "ttcrpy/rgrid.pyx":256
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 *                               self.tt_from_rp, self.interp_vel, self.eps,             # <<<<<<<<<<<<<<
 *                               self.maxit, self.weno, self.nsnx, self.nsny,
 *                               self.nsnz, self.n_secondary, self.n_tertiary,
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->tt_from_rp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->interp_vel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->eps); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "ttcrpy/rgrid.pyx":257
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 *                               self.tt_from_rp, self.interp_vel, self.eps,
 *                               self.maxit, self.weno, self.nsnx, self.nsny,             # <<<<<<<<<<<<<<
 *                               self.nsnz, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_self->maxit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->weno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->nsnx); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->nsny); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "ttcrpy/rgrid.pyx":258
 *                               self.tt_from_rp, self.interp_vel, self.eps,
 *                               self.maxit, self.weno, self.nsnx, self.nsny,
 *                               self.nsnz, self.n_secondary, self.n_tertiary,             # <<<<<<<<<<<<<<
 *                               self.radius_tertiary)
 *         return (_rebuild3d, (self.x, self.y, self.z, constructor_params))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->nsnz); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->n_secondary); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->n_tertiary); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "ttcrpy/rgrid.pyx":259
 *                               self.maxit, self.weno, self.nsnx, self.nsny,
 *                               self.nsnz, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)             # <<<<<<<<<<<<<<
 *         return (_rebuild3d, (self.x, self.y, self.z, constructor_params))
 * 
 */
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->radius_tertiary); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "ttcrpy/rgrid.pyx":255
 *             method = 'DSPM'
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,             # <<<<<<<<<<<<<<
 *                               self.tt_from_rp, self.interp_vel, self.eps,
 *                               self.maxit, self.weno, self.nsnx, self.nsny,
 */
  __pyx_t_14 = PyTuple_New(14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_v_method);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_14, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_14, 4, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_14, 5, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_14, 6, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_14, 7, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_14, 8, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_14, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_14, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_14, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_14, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_14, 13, __pyx_t_13);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_v_constructor_params = ((PyObject*)__pyx_t_14);
  __pyx_t_14 = 0;

  /* "ttcrpy/rgrid.pyx":260
 *                               self.nsnz, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)
 *         return (_rebuild3d, (self.x, self.y, self.z, constructor_params))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_rebuild3d); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_x); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_y); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_z); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = PyTuple_New(4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_11);
  __Pyx_INCREF(__pyx_v_constructor_params);
  __Pyx_GIVEREF(__pyx_v_constructor_params);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_v_constructor_params);
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_10);
  __pyx_t_14 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":247
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_constructor_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":263
 * 
 *     @property
 *     def x(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1x_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1x_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_1x___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1x___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_v_tmp = NULL;
  int __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<double> ::size_type __pyx_t_5;
  std::vector<double> ::size_type __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":265
 *     def x(self):
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))             # <<<<<<<<<<<<<<
 *         cdef int n
 *         for n in range(self._x.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tmp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":267
 *         tmp = np.empty((self._x.size(),))
 *         cdef int n
 *         for n in range(self._x.size()):             # <<<<<<<<<<<<<<
 *             tmp[n] = self._x[n]
 *         return tmp
 */
  __pyx_t_5 = __pyx_v_self->_x.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/rgrid.pyx":268
 *         cdef int n
 *         for n in range(self._x.size()):
 *             tmp[n] = self._x[n]             # <<<<<<<<<<<<<<
 *         return tmp
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_x[__pyx_v_n])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tmp, __pyx_v_n, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":269
 *         for n in range(self._x.size()):
 *             tmp[n] = self._x[n]
 *         return tmp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_r = __pyx_v_tmp;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":263
 * 
 *     @property
 *     def x(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.x.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":272
 * 
 *     @property
 *     def y(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along y"""
 *         tmp = np.empty((self._y.size(),))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1y_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1y_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_1y___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1y___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_v_tmp = NULL;
  int __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<double> ::size_type __pyx_t_5;
  std::vector<double> ::size_type __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":274
 *     def y(self):
 *         """np.ndarray: node coordinates along y"""
 *         tmp = np.empty((self._y.size(),))             # <<<<<<<<<<<<<<
 *         cdef int n
 *         for n in range(self._y.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tmp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":276
 *         tmp = np.empty((self._y.size(),))
 *         cdef int n
 *         for n in range(self._y.size()):             # <<<<<<<<<<<<<<
 *             tmp[n] = self._y[n]
 *         return tmp
 */
  __pyx_t_5 = __pyx_v_self->_y.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/rgrid.pyx":277
 *         cdef int n
 *         for n in range(self._y.size()):
 *             tmp[n] = self._y[n]             # <<<<<<<<<<<<<<
 *         return tmp
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_y[__pyx_v_n])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tmp, __pyx_v_n, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":278
 *         for n in range(self._y.size()):
 *             tmp[n] = self._y[n]
 *         return tmp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_r = __pyx_v_tmp;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":272
 * 
 *     @property
 *     def y(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along y"""
 *         tmp = np.empty((self._y.size(),))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.y.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":281
 * 
 *     @property
 *     def z(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1z_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1z_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_1z___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_1z___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_v_tmp = NULL;
  int __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<double> ::size_type __pyx_t_5;
  std::vector<double> ::size_type __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":283
 *     def z(self):
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))             # <<<<<<<<<<<<<<
 *         cdef int n
 *         for n in range(self._z.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tmp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":285
 *         tmp = np.empty((self._z.size(),))
 *         cdef int n
 *         for n in range(self._z.size()):             # <<<<<<<<<<<<<<
 *             tmp[n] = self._z[n]
 *         return tmp
 */
  __pyx_t_5 = __pyx_v_self->_z.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/rgrid.pyx":286
 *         cdef int n
 *         for n in range(self._z.size()):
 *             tmp[n] = self._z[n]             # <<<<<<<<<<<<<<
 *         return tmp
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_z[__pyx_v_n])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tmp, __pyx_v_n, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":287
 *         for n in range(self._z.size()):
 *             tmp[n] = self._z[n]
 *         return tmp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_r = __pyx_v_tmp;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":281
 * 
 *     @property
 *     def z(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.z.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":290
 * 
 *     @property
 *     def dx(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dx
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dx_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dx___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dx___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":292
 *     def dx(self):
 *         """float: node separation along x"""
 *         return self._dx             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":290
 * 
 *     @property
 *     def dx(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dx
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.dx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":295
 * 
 *     @property
 *     def dy(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along y"""
 *         return self._dy
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dy_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dy___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dy___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":297
 *     def dy(self):
 *         """float: node separation along y"""
 *         return self._dy             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":295
 * 
 *     @property
 *     def dy(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along y"""
 *         return self._dy
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.dy.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":300
 * 
 *     @property
 *     def dz(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along z"""
 *         return self._dz
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dz_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dz_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dz___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_2dz___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":302
 *     def dz(self):
 *         """float: node separation along z"""
 *         return self._dz             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":300
 * 
 *     @property
 *     def dz(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along z"""
 *         return self._dz
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.dz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":305
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_5shape___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_5shape___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":307
 *     def shape(self):
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":308
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 *             return (self._x.size()-1, self._y.size()-1, self._z.size()-1)             # <<<<<<<<<<<<<<
 *         else:
 *             return (self._x.size(), self._y.size(), self._z.size())
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_y.size() - 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":307
 *     def shape(self):
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":310
 *             return (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *         else:
 *             return (self._x.size(), self._y.size(), self._z.size())             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":305
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":313
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_9n_threads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_9n_threads_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_9n_threads___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":315
 *     def n_threads(self):
 *         """int: number of threads for raytracing"""
 *         return self._n_threads             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":313
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.n_threads.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":318
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_7nparams_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_7nparams_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_7nparams___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_7nparams___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":320
 *     def nparams(self):
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":321
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 *             return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._x.size() * self._y.size() * self._z.size()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((((__pyx_v_self->_x.size() - 1) * (__pyx_v_self->_y.size() - 1)) * (__pyx_v_self->_z.size() - 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":320
 *     def nparams(self):
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":323
 *             return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)
 *         else:
 *             return self._x.size() * self._y.size() * self._z.size()             # <<<<<<<<<<<<<<
 * 
 *     def set_use_thread_pool(self, use_thread_pool):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->_x.size() * __pyx_v_self->_y.size()) * __pyx_v_self->_z.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":318
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.nparams.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":325
 *             return self._x.size() * self._y.size() * self._z.size()
 * 
 *     def set_use_thread_pool(self, use_thread_pool):             # <<<<<<<<<<<<<<
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_7set_use_thread_pool(PyObject *__pyx_v_self, PyObject *__pyx_v_use_thread_pool); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_6set_use_thread_pool[] = "bool: use thread pool instead of parallel loop";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_7set_use_thread_pool(PyObject *__pyx_v_self, PyObject *__pyx_v_use_thread_pool) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_use_thread_pool (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_6set_use_thread_pool(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), ((PyObject *)__pyx_v_use_thread_pool));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_6set_use_thread_pool(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_use_thread_pool) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_use_thread_pool", 0);

  /* "ttcrpy/rgrid.pyx":327
 *     def set_use_thread_pool(self, use_thread_pool):
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_nodes(self):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_use_thread_pool); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 327, __pyx_L1_error)
  __pyx_v_self->grid->setUsePool(__pyx_t_1);

  /* "ttcrpy/rgrid.pyx":325
 *             return self._x.size() * self._y.size() * self._z.size()
 * 
 *     def set_use_thread_pool(self, use_thread_pool):             # <<<<<<<<<<<<<<
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.set_use_thread_pool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":329
 *         self.grid.setUsePool(use_thread_pool)
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_9get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_8get_number_of_nodes[] = "\n        Returns\n        -------\n        int:\n            number of nodes in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_9get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_nodes (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_8get_number_of_nodes(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_8get_number_of_nodes(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_nodes", 0);

  /* "ttcrpy/rgrid.pyx":336
 *             number of nodes in grid
 *         """
 *         return self._x.size() * self._y.size() * self._z.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_cells(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->_x.size() * __pyx_v_self->_y.size()) * __pyx_v_self->_z.size())); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":329
 *         self.grid.setUsePool(use_thread_pool)
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_number_of_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":338
 *         return self._x.size() * self._y.size() * self._z.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_11get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_10get_number_of_cells[] = "\n        Returns\n        -------\n        int:\n            number of cells in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_11get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_cells (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_10get_number_of_cells(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_10get_number_of_cells(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_cells", 0);

  /* "ttcrpy/rgrid.pyx":345
 *             number of cells in grid
 *         """
 *         return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)             # <<<<<<<<<<<<<<
 * 
 *     def get_grid_traveltimes(self, thread_no=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((((__pyx_v_self->_x.size() - 1) * (__pyx_v_self->_y.size() - 1)) * (__pyx_v_self->_z.size() - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":338
 *         return self._x.size() * self._y.size() * self._z.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_number_of_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":347
 *         return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_13get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_12get_grid_traveltimes[] = "\n        get_grid_traveltimes(thread_no=0)\n\n        Obtain traveltimes computed at primary grid nodes\n\n        Parameters\n        ----------\n        thread_no : int\n            thread used to computed traveltimes (default is 0)\n\n        Returns\n        -------\n        tt: np ndarray, shape (nx, ny, nz)\n            traveltimes\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_13get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_thread_no = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_grid_traveltimes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_thread_no,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_grid_traveltimes") < 0)) __PYX_ERR(0, 347, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_thread_no = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_grid_traveltimes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 347, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_12get_grid_traveltimes(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_thread_no);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_12get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_thread_no) {
  std::vector<double>  __pyx_v_tmp;
  int __pyx_v_n;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<double> ::size_type __pyx_t_7;
  std::vector<double> ::size_type __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_grid_traveltimes", 0);

  /* "ttcrpy/rgrid.pyx":363
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_thread_no, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/rgrid.pyx":364
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 364, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":363
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  }

  /* "ttcrpy/rgrid.pyx":367
 *         cdef vector[double] tmp
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)             # <<<<<<<<<<<<<<
 *         tt = np.empty((tmp.size(),), order='F')
 *         for n in range(tmp.size()):
 */
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)
  try {
    __pyx_v_self->grid->getTT(__pyx_v_tmp, __pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 367, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":368
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),), order='F')             # <<<<<<<<<<<<<<
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_tmp.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_tt = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "ttcrpy/rgrid.pyx":369
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),), order='F')
 *         for n in range(tmp.size()):             # <<<<<<<<<<<<<<
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._y.size(), self._z.size())
 */
  __pyx_t_7 = __pyx_v_tmp.size();
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/rgrid.pyx":370
 *         tt = np.empty((tmp.size(),), order='F')
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]             # <<<<<<<<<<<<<<
 *         shape = (self._x.size(), self._y.size(), self._z.size())
 *         return tt.reshape(shape)
 */
    __pyx_t_6 = PyFloat_FromDouble((__pyx_v_tmp[__pyx_v_n])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_n, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "ttcrpy/rgrid.pyx":371
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._y.size(), self._z.size())             # <<<<<<<<<<<<<<
 *         return tt.reshape(shape)
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_2 = 0;
  __pyx_v_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":372
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._y.size(), self._z.size())
 *         return tt.reshape(shape)             # <<<<<<<<<<<<<<
 * 
 *     def ind(self, int i, int j, int k):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tt, __pyx_n_s_reshape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_shape);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":347
 *         return (self._x.size()-1) * (self._y.size()-1) * (self._z.size()-1)
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":374
 *         return tt.reshape(shape)
 * 
 *     def ind(self, int i, int j, int k):             # <<<<<<<<<<<<<<
 *         """
 *         ind(i, j, k)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_15ind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_14ind[] = "\n        ind(i, j, k)\n\n        Return node index\n\n        Parameters\n        ----------\n        i : int\n            index of node along x\n        j : int\n            index of node along y\n        k : int\n            index of node along z\n\n        Returns\n        -------\n        int:\n            node index for a \"flattened\" grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_15ind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ind (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_j,&__pyx_n_s_k,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ind", 1, 3, 3, 1); __PYX_ERR(0, 374, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ind", 1, 3, 3, 2); __PYX_ERR(0, 374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ind") < 0)) __PYX_ERR(0, 374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
    __pyx_v_j = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_j == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ind", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.ind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_14ind(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_i, __pyx_v_j, __pyx_v_k);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_14ind(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ind", 0);

  /* "ttcrpy/rgrid.pyx":394
 *             node index for a "flattened" grid
 *         """
 *         return (i*self._y.size() + j)*self._z.size() + k             # <<<<<<<<<<<<<<
 * 
 *     def indc(self, int i, int j, int k):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((((__pyx_v_i * __pyx_v_self->_y.size()) + __pyx_v_j) * __pyx_v_self->_z.size()) + __pyx_v_k)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":374
 *         return tt.reshape(shape)
 * 
 *     def ind(self, int i, int j, int k):             # <<<<<<<<<<<<<<
 *         """
 *         ind(i, j, k)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.ind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":396
 *         return (i*self._y.size() + j)*self._z.size() + k
 * 
 *     def indc(self, int i, int j, int k):             # <<<<<<<<<<<<<<
 *         """
 *         return cell index
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_17indc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_16indc[] = "\n        return cell index\n\n        Parameters\n        ----------\n        i : int\n            index of cell along x\n        j : int\n            index of cell along y\n        k : int\n            index of cell along z\n\n        Returns\n        -------\n        int:\n            cell index for a \"flattened\" grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_17indc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("indc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_j,&__pyx_n_s_k,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("indc", 1, 3, 3, 1); __PYX_ERR(0, 396, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("indc", 1, 3, 3, 2); __PYX_ERR(0, 396, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "indc") < 0)) __PYX_ERR(0, 396, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L3_error)
    __pyx_v_j = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_j == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("indc", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 396, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.indc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_16indc(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_i, __pyx_v_j, __pyx_v_k);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_16indc(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("indc", 0);

  /* "ttcrpy/rgrid.pyx":414
 *             cell index for a "flattened" grid
 *         """
 *         return (i*(self._y.size()-1) + j)*(self._z.size()-1) + k             # <<<<<<<<<<<<<<
 * 
 *     cdef uint32_t _f2c_ind(self, uint32_t ind):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((((__pyx_v_i * (__pyx_v_self->_y.size() - 1)) + __pyx_v_j) * (__pyx_v_self->_z.size() - 1)) + __pyx_v_k)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":396
 *         return (i*self._y.size() + j)*self._z.size() + k
 * 
 *     def indc(self, int i, int j, int k):             # <<<<<<<<<<<<<<
 *         """
 *         return cell index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.indc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":416
 *         return (i*(self._y.size()-1) + j)*(self._z.size()-1) + k
 * 
 *     cdef uint32_t _f2c_ind(self, uint32_t ind):             # <<<<<<<<<<<<<<
 *         """Convert cell index from 'F' order to 'C' order"""
 *         cdef uint32_t i, j, k
 */

static uint32_t __pyx_f_6ttcrpy_5rgrid_6Grid3d__f2c_ind(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, uint32_t __pyx_v_ind) {
  uint32_t __pyx_v_i;
  uint32_t __pyx_v_j;
  uint32_t __pyx_v_k;
  uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<double> ::size_type __pyx_t_1;
  std::vector<double> ::size_type __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_f2c_ind", 0);

  /* "ttcrpy/rgrid.pyx":419
 *         """Convert cell index from 'F' order to 'C' order"""
 *         cdef uint32_t i, j, k
 *         k = <uint32_t>(ind / ((self._x.size()-1) * (self._y.size()-1)))             # <<<<<<<<<<<<<<
 *         j = <uint32_t>((ind - k * (self._x.size()-1) * (self._y.size()-1)) / (self._x.size()-1))
 *         i = <uint32_t>(ind - ((k * (self._y.size()-1) + j) * (self._x.size()-1)))
 */
  __pyx_t_1 = ((__pyx_v_self->_x.size() - 1) * (__pyx_v_self->_y.size() - 1));
  if (unlikely(__pyx_t_1 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 419, __pyx_L1_error)
  }
  __pyx_v_k = ((uint32_t)(((double)__pyx_v_ind) / ((double)__pyx_t_1)));

  /* "ttcrpy/rgrid.pyx":420
 *         cdef uint32_t i, j, k
 *         k = <uint32_t>(ind / ((self._x.size()-1) * (self._y.size()-1)))
 *         j = <uint32_t>((ind - k * (self._x.size()-1) * (self._y.size()-1)) / (self._x.size()-1))             # <<<<<<<<<<<<<<
 *         i = <uint32_t>(ind - ((k * (self._y.size()-1) + j) * (self._x.size()-1)))
 *         return (i * (self._y.size()-1) + j) * (self._z.size()-1) + k
 */
  __pyx_t_1 = (__pyx_v_ind - ((__pyx_v_k * (__pyx_v_self->_x.size() - 1)) * (__pyx_v_self->_y.size() - 1)));
  __pyx_t_2 = (__pyx_v_self->_x.size() - 1);
  if (unlikely(__pyx_t_2 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 420, __pyx_L1_error)
  }
  __pyx_v_j = ((uint32_t)(((double)__pyx_t_1) / ((double)__pyx_t_2)));

  /* "ttcrpy/rgrid.pyx":421
 *         k = <uint32_t>(ind / ((self._x.size()-1) * (self._y.size()-1)))
 *         j = <uint32_t>((ind - k * (self._x.size()-1) * (self._y.size()-1)) / (self._x.size()-1))
 *         i = <uint32_t>(ind - ((k * (self._y.size()-1) + j) * (self._x.size()-1)))             # <<<<<<<<<<<<<<
 *         return (i * (self._y.size()-1) + j) * (self._z.size()-1) + k
 * 
 */
  __pyx_v_i = ((uint32_t)(__pyx_v_ind - (((__pyx_v_k * (__pyx_v_self->_y.size() - 1)) + __pyx_v_j) * (__pyx_v_self->_x.size() - 1))));

  /* "ttcrpy/rgrid.pyx":422
 *         j = <uint32_t>((ind - k * (self._x.size()-1) * (self._y.size()-1)) / (self._x.size()-1))
 *         i = <uint32_t>(ind - ((k * (self._y.size()-1) + j) * (self._x.size()-1)))
 *         return (i * (self._y.size()-1) + j) * (self._z.size()-1) + k             # <<<<<<<<<<<<<<
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):
 */
  __pyx_r = ((((__pyx_v_i * (__pyx_v_self->_y.size() - 1)) + __pyx_v_j) * (__pyx_v_self->_z.size() - 1)) + __pyx_v_k);
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":416
 *         return (i*(self._y.size()-1) + j)*(self._z.size()-1) + k
 * 
 *     cdef uint32_t _f2c_ind(self, uint32_t ind):             # <<<<<<<<<<<<<<
 *         """Convert cell index from 'F' order to 'C' order"""
 *         cdef uint32_t i, j, k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ttcrpy.rgrid.Grid3d._f2c_ind", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":424
 *         return (i * (self._y.size()-1) + j) * (self._z.size()-1) + k
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_19is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_18is_outside[] = "\n        is_outside(pts)\n\n        Check if points are outside grid\n\n        Parameters\n        ----------\n        pts : np ndarray, shape(npts, 3)\n            coordinates of points to check\n\n        Returns\n        -------\n        bool:\n            True if at least one point outside grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_19is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_outside (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pts), __pyx_ptype_5numpy_ndarray, 1, "pts", 0))) __PYX_ERR(0, 424, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_18is_outside(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), ((PyArrayObject *)__pyx_v_pts));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_18is_outside(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyArrayObject *__pyx_v_pts) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pts;
  __Pyx_Buffer __pyx_pybuffer_pts;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_outside", 0);
  __pyx_pybuffer_pts.pybuffer.buf = NULL;
  __pyx_pybuffer_pts.refcount = 0;
  __pyx_pybuffernd_pts.data = NULL;
  __pyx_pybuffernd_pts.rcbuffer = &__pyx_pybuffer_pts;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pts.rcbuffer->pybuffer, (PyObject*)__pyx_v_pts, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 424, __pyx_L1_error)
  }
  __pyx_pybuffernd_pts.diminfo[0].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pts.diminfo[0].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pts.diminfo[1].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pts.diminfo[1].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[1];

  /* "ttcrpy/rgrid.pyx":440
 *             True if at least one point outside grid
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or             # <<<<<<<<<<<<<<
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_min); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_x.front()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_x.back()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ttcrpy/rgrid.pyx":441
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or             # <<<<<<<<<<<<<<
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_min); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_y.front()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_y.back()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ttcrpy/rgrid.pyx":442
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )             # <<<<<<<<<<<<<<
 * 
 *     def get_slowness(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_z.front()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 442, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_z.back()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":424
 *         return (i * (self._y.size()-1) + j) * (self._z.size()-1) + k
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.is_outside", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":444
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 * 
 *     def get_slowness(self):             # <<<<<<<<<<<<<<
 *         """Returns slowness of grid
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_21get_slowness(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_20get_slowness[] = "Returns slowness of grid\n\n        Returns\n        -------\n        slowness : np ndarray, shape (nx, ny, nz)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_21get_slowness(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_20get_slowness(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_20get_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  int __pyx_v_i;
  std::vector<double>  __pyx_v_slown;
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_ny = NULL;
  PyObject *__pyx_v_nz = NULL;
  PyObject *__pyx_v_slowness = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  std::vector<double> ::size_type __pyx_t_7;
  std::vector<double> ::size_type __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slowness", 0);

  /* "ttcrpy/rgrid.pyx":453
 *         cdef int i
 *         cdef vector[double] slown
 *         self.grid.getSlowness(slown)             # <<<<<<<<<<<<<<
 *         nx, ny, nz = self.shape
 *         slowness = np.ndarray((nx*ny*nz,), order='F')
 */
  try {
    __pyx_v_self->grid->getSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 453, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":454
 *         cdef vector[double] slown
 *         self.grid.getSlowness(slown)
 *         nx, ny, nz = self.shape             # <<<<<<<<<<<<<<
 *         slowness = np.ndarray((nx*ny*nz,), order='F')
 *         for i in range(slown.size()):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 454, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 454, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 454, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_nx = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_ny = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_nz = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "ttcrpy/rgrid.pyx":455
 *         self.grid.getSlowness(slown)
 *         nx, ny, nz = self.shape
 *         slowness = np.ndarray((nx*ny*nz,), order='F')             # <<<<<<<<<<<<<<
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]
 */
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 455, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_slowness = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":456
 *         nx, ny, nz = self.shape
 *         slowness = np.ndarray((nx*ny*nz,), order='F')
 *         for i in range(slown.size()):             # <<<<<<<<<<<<<<
 *             slowness[i] = slown[i]
 *         return slowness.reshape((nx, ny, nz))
 */
  __pyx_t_7 = __pyx_v_slown.size();
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ttcrpy/rgrid.pyx":457
 *         slowness = np.ndarray((nx*ny*nz,), order='F')
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]             # <<<<<<<<<<<<<<
 *         return slowness.reshape((nx, ny, nz))
 * 
 */
    __pyx_t_3 = PyFloat_FromDouble((__pyx_v_slown[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_slowness, __pyx_v_i, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":458
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]
 *         return slowness.reshape((nx, ny, nz))             # <<<<<<<<<<<<<<
 * 
 *     def set_slowness(self, slowness):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_reshape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_nx);
  __Pyx_GIVEREF(__pyx_v_nx);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
  __Pyx_INCREF(__pyx_v_ny);
  __Pyx_GIVEREF(__pyx_v_ny);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_ny);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_nz);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":444
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 * 
 *     def get_slowness(self):             # <<<<<<<<<<<<<<
 *         """Returns slowness of grid
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_ny);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":460
 *         return slowness.reshape((nx, ny, nz))
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_23set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_22set_slowness[] = "\n        set_slowness(slowness)\n\n        Assign slowness to grid\n\n        Parameters\n        ----------\n        slowness : np ndarray, shape (nx, ny, nz)\n            slowness may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_23set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_22set_slowness(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), ((PyObject *)__pyx_v_slowness));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_22set_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_ny = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_slowness", 0);

  /* "ttcrpy/rgrid.pyx":471
 *             slowness may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":472
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             ny = self._y.size()-1
 *             nz = self._z.size()-1
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":473
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_y.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ny = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":474
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":471
 *             slowness may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":476
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             ny = self._y.size()
 *             nz = self._z.size()
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":477
 *         else:
 *             nx = self._x.size()
 *             ny = self._y.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if slowness.size != nx*ny*nz:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ny = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":478
 *             nx = self._x.size()
 *             ny = self._y.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if slowness.size != nx*ny*nz:
 *             raise ValueError('Slowness vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":479
 *             ny = self._y.size()
 *             nz = self._z.size()
 *         if slowness.size != nx*ny*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":480
 *             nz = self._z.size()
 *         if slowness.size != nx*ny*nz:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 480, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":479
 *             ny = self._y.size()
 *             nz = self._z.size()
 *         if slowness.size != nx*ny*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":484
 *         cdef vector[double] slown
 *         cdef int i
 *         if slowness.ndim == 3:             # <<<<<<<<<<<<<<
 *             if slowness.shape != (nx, ny, nz):
 *                 raise ValueError('Slowness has wrong shape')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":485
 *         cdef int i
 *         if slowness.ndim == 3:
 *             if slowness.shape != (nx, ny, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten('F')
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_ny);
    __Pyx_GIVEREF(__pyx_v_ny);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_ny);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":486
 *         if slowness.ndim == 3:
 *             if slowness.shape != (nx, ny, nz):
 *                 raise ValueError('Slowness has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = slowness.flatten('F')
 *             for i in range(nx*ny*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 486, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":485
 *         cdef int i
 *         if slowness.ndim == 3:
 *             if slowness.shape != (nx, ny, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten('F')
 */
    }

    /* "ttcrpy/rgrid.pyx":487
 *             if slowness.shape != (nx, ny, nz):
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten('F')             # <<<<<<<<<<<<<<
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_n_u_F) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_F);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":488
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten('F')
 *             for i in range(nx*ny*nz):             # <<<<<<<<<<<<<<
 *                 slown.push_back(tmp[i])
 *         elif slowness.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":489
 *             tmp = slowness.flatten('F')
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif slowness.ndim == 1:
 *             # slowness is in 'C' order and we must pass it in 'F' order
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_v_slown.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 489, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":484
 *         cdef vector[double] slown
 *         cdef int i
 *         if slowness.ndim == 3:             # <<<<<<<<<<<<<<
 *             if slowness.shape != (nx, ny, nz):
 *                 raise ValueError('Slowness has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":490
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])
 *         elif slowness.ndim == 1:             # <<<<<<<<<<<<<<
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = slowness.reshape((nx, ny, nz)).flatten('F')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":492
 *         elif slowness.ndim == 1:
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = slowness.reshape((nx, ny, nz)).flatten('F')             # <<<<<<<<<<<<<<
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_ny);
    __Pyx_GIVEREF(__pyx_v_ny);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_ny);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_u_F) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_F);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":493
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = slowness.reshape((nx, ny, nz)).flatten('F')
 *             for i in range(nx*ny*nz):             # <<<<<<<<<<<<<<
 *                 slown.push_back(tmp[i])
 *         else:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":494
 *             tmp = slowness.reshape((nx, ny, nz)).flatten('F')
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_slown.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 494, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":490
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(tmp[i])
 *         elif slowness.ndim == 1:             # <<<<<<<<<<<<<<
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = slowness.reshape((nx, ny, nz)).flatten('F')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":496
 *                 slown.push_back(tmp[i])
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 496, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":497
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def set_velocity(self, velocity):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 497, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":460
 *         return slowness.reshape((nx, ny, nz))
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.set_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_ny);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":499
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(slowness)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_25set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_24set_velocity[] = "\n        set_velocity(slowness)\n\n        Assign velocity to grid\n\n        Parameters\n        ----------\n        velocity : np ndarray, shape (nx, ny, nz)\n            velocity may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_25set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_velocity (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_24set_velocity(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), ((PyObject *)__pyx_v_velocity));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_24set_velocity(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_ny = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_velocity", 0);

  /* "ttcrpy/rgrid.pyx":510
 *             velocity may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":511
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             ny = self._y.size()-1
 *             nz = self._z.size()-1
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":512
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_y.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ny = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":513
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":510
 *             velocity may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             ny = self._y.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":515
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             ny = self._y.size()
 *             nz = self._z.size()
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":516
 *         else:
 *             nx = self._x.size()
 *             ny = self._y.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if velocity.size != nx*ny*nz:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ny = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":517
 *             nx = self._x.size()
 *             ny = self._y.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if velocity.size != nx*ny*nz:
 *             raise ValueError('velocity vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":518
 *             ny = self._y.size()
 *             nz = self._z.size()
 *         if velocity.size != nx*ny*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":519
 *             nz = self._z.size()
 *         if velocity.size != nx*ny*nz:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 519, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":518
 *             ny = self._y.size()
 *             nz = self._z.size()
 *         if velocity.size != nx*ny*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":523
 *         cdef vector[double] slown
 *         cdef int i
 *         if velocity.ndim == 3:             # <<<<<<<<<<<<<<
 *             if velocity.shape != (nx, ny, nz):
 *                 raise ValueError('velocity has wrong shape')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":524
 *         cdef int i
 *         if velocity.ndim == 3:
 *             if velocity.shape != (nx, ny, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten('F')
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_ny);
    __Pyx_GIVEREF(__pyx_v_ny);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_ny);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":525
 *         if velocity.ndim == 3:
 *             if velocity.shape != (nx, ny, nz):
 *                 raise ValueError('velocity has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = velocity.flatten('F')
 *             for i in range(nx*ny*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 525, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":524
 *         cdef int i
 *         if velocity.ndim == 3:
 *             if velocity.shape != (nx, ny, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten('F')
 */
    }

    /* "ttcrpy/rgrid.pyx":526
 *             if velocity.shape != (nx, ny, nz):
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten('F')             # <<<<<<<<<<<<<<
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_n_u_F) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_F);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":527
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten('F')
 *             for i in range(nx*ny*nz):             # <<<<<<<<<<<<<<
 *                 slown.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":528
 *             tmp = velocity.flatten('F')
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])             # <<<<<<<<<<<<<<
 *         elif velocity.ndim == 1:
 *             # slowness is in 'C' order and we must pass it in 'F' order
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_3, 1., 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_slown.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 528, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":523
 *         cdef vector[double] slown
 *         cdef int i
 *         if velocity.ndim == 3:             # <<<<<<<<<<<<<<
 *             if velocity.shape != (nx, ny, nz):
 *                 raise ValueError('velocity has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":529
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:             # <<<<<<<<<<<<<<
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = velocity.reshape((nx, ny, nz)).flatten('F')
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":531
 *         elif velocity.ndim == 1:
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = velocity.reshape((nx, ny, nz)).flatten('F')             # <<<<<<<<<<<<<<
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_ny);
    __Pyx_GIVEREF(__pyx_v_ny);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_ny);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_n_u_F) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_F);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":532
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = velocity.reshape((nx, ny, nz)).flatten('F')
 *             for i in range(nx*ny*nz):             # <<<<<<<<<<<<<<
 *                 slown.push_back(1./tmp[i])
 *         else:
 */
    __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":533
 *             tmp = velocity.reshape((nx, ny, nz)).flatten('F')
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_4, 1., 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_v_slown.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 533, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":529
 *             for i in range(nx*ny*nz):
 *                 slown.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:             # <<<<<<<<<<<<<<
 *             # slowness is in 'C' order and we must pass it in 'F' order
 *             tmp = velocity.reshape((nx, ny, nz)).flatten('F')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":535
 *                 slown.push_back(1./tmp[i])
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 535, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":536
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def compute_D(self, coord):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 536, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":499
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(slowness)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.set_velocity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_ny);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":538
 *         self.grid.setSlowness(slown)
 * 
 *     def compute_D(self, coord):             # <<<<<<<<<<<<<<
 *         """
 *         compute_D(coord)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_27compute_D(PyObject *__pyx_v_self, PyObject *__pyx_v_coord); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_26compute_D[] = "\n        compute_D(coord)\n\n        Return matrix of interpolation weights for velocity data points\n        constraint\n\n        Parameters\n        ----------\n        coord : np.ndarray, shape (npts, 3)\n            coordinates of data points\n\n        Returns\n        -------\n        D : scipy csr_matrix, shape (npts, nparams)\n            Matrix of interpolation weights\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_27compute_D(PyObject *__pyx_v_self, PyObject *__pyx_v_coord) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_D (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_26compute_D(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), ((PyObject *)__pyx_v_coord));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_26compute_D(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_coord) {
  Py_ssize_t __pyx_v_n;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_k;
  Py_ssize_t __pyx_v_i1;
  Py_ssize_t __pyx_v_i2;
  Py_ssize_t __pyx_v_j1;
  Py_ssize_t __pyx_v_j2;
  Py_ssize_t __pyx_v_k1;
  Py_ssize_t __pyx_v_k2;
  Py_ssize_t __pyx_v_ii;
  PyObject *__pyx_v_ivec = NULL;
  PyObject *__pyx_v_jvec = NULL;
  PyObject *__pyx_v_vec = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_D", 0);

  /* "ttcrpy/rgrid.pyx":555
 *             Matrix of interpolation weights
 *         """
 *         if self.is_outside(coord):             # <<<<<<<<<<<<<<
 *             raise ValueError('Velocity data point outside grid')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_coord) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_coord);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/rgrid.pyx":556
 *         """
 *         if self.is_outside(coord):
 *             raise ValueError('Velocity data point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         # D is npts x nparams
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 556, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":555
 *             Matrix of interpolation weights
 *         """
 *         if self.is_outside(coord):             # <<<<<<<<<<<<<<
 *             raise ValueError('Velocity data point outside grid')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":562
 *         cdef Py_ssize_t n, i, j, k, i1, i2, j1, j2, k1, k2, ii
 * 
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             # for each point in coord, we have 1 values in D
 *             ivec = np.arange(coord.shape[0], dtype=np.int64)
 */
  __pyx_t_4 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":564
 *         if self.cell_slowness:
 *             # for each point in coord, we have 1 values in D
 *             ivec = np.arange(coord.shape[0], dtype=np.int64)             # <<<<<<<<<<<<<<
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.ones(ivec.shape)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_ivec = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "ttcrpy/rgrid.pyx":565
 *             # for each point in coord, we have 1 values in D
 *             ivec = np.arange(coord.shape[0], dtype=np.int64)
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)             # <<<<<<<<<<<<<<
 *             vec = np.ones(ivec.shape)
 *             for n in np.arange(coord.shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_ivec, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_jvec = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":566
 *             ivec = np.arange(coord.shape[0], dtype=np.int64)
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.ones(ivec.shape)             # <<<<<<<<<<<<<<
 *             for n in np.arange(coord.shape[0]):
 *                 i = int((coord[n,0]-self.x[0])/self._dx )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_ivec, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_vec = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":567
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.ones(ivec.shape)
 *             for n in np.arange(coord.shape[0]):             # <<<<<<<<<<<<<<
 *                 i = int((coord[n,0]-self.x[0])/self._dx )
 *                 j = int((coord[n,1]-self.y[0])/self._dx )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 567, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_8(__pyx_t_6);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 567, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_n = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":568
 *             vec = np.ones(ivec.shape)
 *             for n in np.arange(coord.shape[0]):
 *                 i = int((coord[n,0]-self.x[0])/self._dx )             # <<<<<<<<<<<<<<
 *                 j = int((coord[n,1]-self.y[0])/self._dx )
 *                 k = int((coord[n,2]-self.z[0])/self._dx )
 */
      __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Subtract(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_i = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":569
 *             for n in np.arange(coord.shape[0]):
 *                 i = int((coord[n,0]-self.x[0])/self._dx )
 *                 j = int((coord[n,1]-self.y[0])/self._dx )             # <<<<<<<<<<<<<<
 *                 k = int((coord[n,2]-self.z[0])/self._dx )
 * 
 */
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Subtract(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_j = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":570
 *                 i = int((coord[n,0]-self.x[0])/self._dx )
 *                 j = int((coord[n,1]-self.y[0])/self._dx )
 *                 k = int((coord[n,2]-self.z[0])/self._dx )             # <<<<<<<<<<<<<<
 * 
 *                 jvec[n] = self.indc(i,j,k)
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_2);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_k = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":572
 *                 k = int((coord[n,2]-self.z[0])/self._dx )
 * 
 *                 jvec[n] = self.indc(i,j,k)             # <<<<<<<<<<<<<<
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_indc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = PyInt_FromSsize_t(__pyx_v_k); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_t_1, __pyx_t_2, __pyx_t_10};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_t_1, __pyx_t_2, __pyx_t_10};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_12, __pyx_t_10);
        __pyx_t_1 = 0;
        __pyx_t_2 = 0;
        __pyx_t_10 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_SetItemInt(__pyx_v_jvec, __pyx_v_n, __pyx_t_5, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":567
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.ones(ivec.shape)
 *             for n in np.arange(coord.shape[0]):             # <<<<<<<<<<<<<<
 *                 i = int((coord[n,0]-self.x[0])/self._dx )
 *                 j = int((coord[n,1]-self.y[0])/self._dx )
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "ttcrpy/rgrid.pyx":574
 *                 jvec[n] = self.indc(i,j,k)
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),             # <<<<<<<<<<<<<<
 *                                  shape=(coord.shape[0], self.get_number_of_cells()))
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_ivec);
    __Pyx_GIVEREF(__pyx_v_ivec);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_ivec);
    __Pyx_INCREF(__pyx_v_jvec);
    __Pyx_GIVEREF(__pyx_v_jvec);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_jvec);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_vec);
    __Pyx_GIVEREF(__pyx_v_vec);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_vec);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":575
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),
 *                                  shape=(coord.shape[0], self.get_number_of_cells()))             # <<<<<<<<<<<<<<
 *         else:
 *             # for each point in coord, we have 8 values in D
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_13, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_13 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_13);
    __pyx_t_10 = 0;
    __pyx_t_13 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":574
 *                 jvec[n] = self.indc(i,j,k)
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),             # <<<<<<<<<<<<<<
 *                                  shape=(coord.shape[0], self.get_number_of_cells()))
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":562
 *         cdef Py_ssize_t n, i, j, k, i1, i2, j1, j2, k1, k2, ii
 * 
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             # for each point in coord, we have 1 values in D
 *             ivec = np.arange(coord.shape[0], dtype=np.int64)
 */
  }

  /* "ttcrpy/rgrid.pyx":578
 *         else:
 *             # for each point in coord, we have 8 values in D
 *             ivec = np.kron(np.arange(coord.shape[0], dtype=np.int64),np.ones(8, dtype=np.int64))             # <<<<<<<<<<<<<<
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.zeros(ivec.shape)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_kron); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_13);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_int64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__14, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_1, __pyx_t_10};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_1, __pyx_t_10};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_12, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_12, __pyx_t_10);
      __pyx_t_1 = 0;
      __pyx_t_10 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_ivec = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":579
 *             # for each point in coord, we have 8 values in D
 *             ivec = np.kron(np.arange(coord.shape[0], dtype=np.int64),np.ones(8, dtype=np.int64))
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)             # <<<<<<<<<<<<<<
 *             vec = np.zeros(ivec.shape)
 *             for n in np.arange(coord.shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ivec, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_int64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_jvec = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":580
 *             ivec = np.kron(np.arange(coord.shape[0], dtype=np.int64),np.ones(8, dtype=np.int64))
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.zeros(ivec.shape)             # <<<<<<<<<<<<<<
 *             for n in np.arange(coord.shape[0]):
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ivec, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_vec = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":581
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.zeros(ivec.shape)
 *             for n in np.arange(coord.shape[0]):             # <<<<<<<<<<<<<<
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )
 *                 i2 = i1 + 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 581, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 581, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_n = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":582
 *             vec = np.zeros(ivec.shape)
 *             for n in np.arange(coord.shape[0]):
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )             # <<<<<<<<<<<<<<
 *                 i2 = i1 + 1
 *                 j1 = int(1.e-6 + (coord[n,1]-self._y[0])/self._dy )
 */
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble((__pyx_v_self->_x[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyFloat_AddCObj(__pyx_float_1_eneg_6, __pyx_t_1, 1.e-6, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_i1 = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":583
 *             for n in np.arange(coord.shape[0]):
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )
 *                 i2 = i1 + 1             # <<<<<<<<<<<<<<
 *                 j1 = int(1.e-6 + (coord[n,1]-self._y[0])/self._dy )
 *                 j2 = j1 + 1
 */
      __pyx_v_i2 = (__pyx_v_i1 + 1);

      /* "ttcrpy/rgrid.pyx":584
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )
 *                 i2 = i1 + 1
 *                 j1 = int(1.e-6 + (coord[n,1]-self._y[0])/self._dy )             # <<<<<<<<<<<<<<
 *                 j2 = j1 + 1
 *                 k1 = int(1.e-6 + (coord[n,2]-self._z[0])/self._dz )
 */
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble((__pyx_v_self->_y[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->_dy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyFloat_AddCObj(__pyx_float_1_eneg_6, __pyx_t_1, 1.e-6, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_j1 = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":585
 *                 i2 = i1 + 1
 *                 j1 = int(1.e-6 + (coord[n,1]-self._y[0])/self._dy )
 *                 j2 = j1 + 1             # <<<<<<<<<<<<<<
 *                 k1 = int(1.e-6 + (coord[n,2]-self._z[0])/self._dz )
 *                 k2 = k1 + 1
 */
      __pyx_v_j2 = (__pyx_v_j1 + 1);

      /* "ttcrpy/rgrid.pyx":586
 *                 j1 = int(1.e-6 + (coord[n,1]-self._y[0])/self._dy )
 *                 j2 = j1 + 1
 *                 k1 = int(1.e-6 + (coord[n,2]-self._z[0])/self._dz )             # <<<<<<<<<<<<<<
 *                 k2 = k1 + 1
 * 
 */
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_2);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble((__pyx_v_self->_z[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->_dz); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyFloat_AddCObj(__pyx_float_1_eneg_6, __pyx_t_1, 1.e-6, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_k1 = __pyx_t_9;

      /* "ttcrpy/rgrid.pyx":587
 *                 j2 = j1 + 1
 *                 k1 = int(1.e-6 + (coord[n,2]-self._z[0])/self._dz )
 *                 k2 = k1 + 1             # <<<<<<<<<<<<<<
 * 
 *                 ii = 0
 */
      __pyx_v_k2 = (__pyx_v_k1 + 1);

      /* "ttcrpy/rgrid.pyx":589
 *                 k2 = k1 + 1
 * 
 *                 ii = 0             # <<<<<<<<<<<<<<
 *                 for i in (i1, i2):
 *                     for j in (j1, j2):
 */
      __pyx_v_ii = 0;

      /* "ttcrpy/rgrid.pyx":590
 * 
 *                 ii = 0
 *                 for i in (i1, i2):             # <<<<<<<<<<<<<<
 *                     for j in (j1, j2):
 *                         for k in (k1, k2):
 */
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_i1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_i2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
      __pyx_t_1 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (__pyx_t_9 >= 2) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 590, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __pyx_t_14 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_14 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 590, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_i = __pyx_t_14;

        /* "ttcrpy/rgrid.pyx":591
 *                 ii = 0
 *                 for i in (i1, i2):
 *                     for j in (j1, j2):             # <<<<<<<<<<<<<<
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 */
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_j1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_j2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
        __pyx_t_3 = 0;
        __pyx_t_1 = 0;
        __pyx_t_1 = __pyx_t_10; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        for (;;) {
          if (__pyx_t_14 >= 2) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_10); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 591, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 591, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_10); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_v_j = __pyx_t_15;

          /* "ttcrpy/rgrid.pyx":592
 *                 for i in (i1, i2):
 *                     for j in (j1, j2):
 *                         for k in (k1, k2):             # <<<<<<<<<<<<<<
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 */
          __pyx_t_10 = PyInt_FromSsize_t(__pyx_v_k1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_k2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_3);
          __pyx_t_10 = 0;
          __pyx_t_3 = 0;
          __pyx_t_3 = __pyx_t_13; __Pyx_INCREF(__pyx_t_3); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          for (;;) {
            if (__pyx_t_15 >= 2) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_15); __Pyx_INCREF(__pyx_t_13); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 592, __pyx_L1_error)
            #else
            __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 592, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            #endif
            __pyx_t_16 = __Pyx_PyIndex_AsSsize_t(__pyx_t_13); if (unlikely((__pyx_t_16 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 592, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_v_k = __pyx_t_16;

            /* "ttcrpy/rgrid.pyx":593
 *                     for j in (j1, j2):
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)             # <<<<<<<<<<<<<<
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ind); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 593, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 593, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = PyInt_FromSsize_t(__pyx_v_j); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 593, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_17 = PyInt_FromSsize_t(__pyx_v_k); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 593, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_18 = NULL;
            __pyx_t_12 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_18)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_18);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_12 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_5, __pyx_t_11, __pyx_t_17};
              __pyx_t_13 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 593, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_5, __pyx_t_11, __pyx_t_17};
              __pyx_t_13 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 593, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            } else
            #endif
            {
              __pyx_t_19 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 593, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_19);
              if (__pyx_t_18) {
                __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_18); __pyx_t_18 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_12, __pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_11);
              PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_12, __pyx_t_11);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_12, __pyx_t_17);
              __pyx_t_5 = 0;
              __pyx_t_11 = 0;
              __pyx_t_17 = 0;
              __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_19, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 593, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_16 = ((__pyx_v_n * 8) + __pyx_v_ii);
            if (unlikely(__Pyx_SetItemInt(__pyx_v_jvec, __pyx_t_16, __pyx_t_13, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 593, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "ttcrpy/rgrid.pyx":594
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *             # <<<<<<<<<<<<<<
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 */
            __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_abs); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_10);
            PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_10);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_int_0);
            __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = PyFloat_FromDouble((__pyx_v_self->_x[__pyx_v_i])); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_11 = PyNumber_Subtract(__pyx_t_10, __pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
              __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_19);
              if (likely(__pyx_t_17)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                __Pyx_INCREF(__pyx_t_17);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_19, function);
              }
            }
            __pyx_t_13 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_17, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_t_11);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __pyx_t_19 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_11 = __Pyx_PyNumber_Divide(__pyx_t_13, __pyx_t_19); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __pyx_t_19 = __Pyx_PyFloat_SubtractCObj(__pyx_float_1_, __pyx_t_11, 1., 0, 0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "ttcrpy/rgrid.pyx":595
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *             # <<<<<<<<<<<<<<
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 *                             ii += 1
 */
            __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_abs); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_13);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_1);
            __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_self->_y[__pyx_v_j])); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_5 = PyNumber_Subtract(__pyx_t_13, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
              }
            }
            __pyx_t_11 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_10, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_5);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = PyFloat_FromDouble(__pyx_v_self->_dy); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_11, __pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = __Pyx_PyFloat_SubtractCObj(__pyx_float_1_, __pyx_t_5, 1., 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "ttcrpy/rgrid.pyx":594
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *             # <<<<<<<<<<<<<<
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 */
            __pyx_t_5 = PyNumber_Multiply(__pyx_t_19, __pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

            /* "ttcrpy/rgrid.pyx":596
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))             # <<<<<<<<<<<<<<
 *                             ii += 1
 * 
 */
            __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_np); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_abs); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __pyx_t_19 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_19);
            PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_19);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_2);
            __pyx_t_19 = 0;
            __pyx_t_19 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_10); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_self->_z[__pyx_v_k])); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_13 = PyNumber_Subtract(__pyx_t_19, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_11, function);
              }
            }
            __pyx_t_17 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_13);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = PyFloat_FromDouble(__pyx_v_self->_dz); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_13 = __Pyx_PyNumber_Divide(__pyx_t_17, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = __Pyx_PyFloat_SubtractCObj(__pyx_float_1_, __pyx_t_13, 1., 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "ttcrpy/rgrid.pyx":595
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *             # <<<<<<<<<<<<<<
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 *                             ii += 1
 */
            __pyx_t_13 = PyNumber_Multiply(__pyx_t_5, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 595, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "ttcrpy/rgrid.pyx":594
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *             # <<<<<<<<<<<<<<
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 */
            __pyx_t_16 = ((__pyx_v_n * 8) + __pyx_v_ii);
            if (unlikely(__Pyx_SetItemInt(__pyx_v_vec, __pyx_t_16, __pyx_t_13, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "ttcrpy/rgrid.pyx":597
 *                                            (1. - np.abs(coord[n,1]-self._y[j])/self._dy) *
 *                                            (1. - np.abs(coord[n,2]-self._z[k])/self._dz))
 *                             ii += 1             # <<<<<<<<<<<<<<
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),
 */
            __pyx_v_ii = (__pyx_v_ii + 1);

            /* "ttcrpy/rgrid.pyx":592
 *                 for i in (i1, i2):
 *                     for j in (j1, j2):
 *                         for k in (k1, k2):             # <<<<<<<<<<<<<<
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 *                             vec[n*8+ii] = ((1. - np.abs(coord[n,0]-self._x[i])/self._dx) *
 */
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ttcrpy/rgrid.pyx":591
 *                 ii = 0
 *                 for i in (i1, i2):
 *                     for j in (j1, j2):             # <<<<<<<<<<<<<<
 *                         for k in (k1, k2):
 *                             jvec[n*8+ii] = self.ind(i,j,k)
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":590
 * 
 *                 ii = 0
 *                 for i in (i1, i2):             # <<<<<<<<<<<<<<
 *                     for j in (j1, j2):
 *                         for k in (k1, k2):
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "ttcrpy/rgrid.pyx":581
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.zeros(ivec.shape)
 *             for n in np.arange(coord.shape[0]):             # <<<<<<<<<<<<<<
 *                 i1 = int(1.e-6 + (coord[n,0]-self._x[0])/self._dx )
 *                 i2 = i1 + 1
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":599
 *                             ii += 1
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),             # <<<<<<<<<<<<<<
 *                                  shape=(coord.shape[0], self.get_number_of_nodes()))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_ivec);
    __Pyx_GIVEREF(__pyx_v_ivec);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ivec);
    __Pyx_INCREF(__pyx_v_jvec);
    __Pyx_GIVEREF(__pyx_v_jvec);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_jvec);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_vec);
    __Pyx_GIVEREF(__pyx_v_vec);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_vec);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":600
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),
 *                                  shape=(coord.shape[0], self.get_number_of_nodes()))             # <<<<<<<<<<<<<<
 * 
 *     def compute_K(self):
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_3);
    __pyx_t_13 = 0;
    __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_11) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "ttcrpy/rgrid.pyx":599
 *                             ii += 1
 * 
 *             return sp.csr_matrix((vec, (ivec,jvec)),             # <<<<<<<<<<<<<<
 *                                  shape=(coord.shape[0], self.get_number_of_nodes()))
 * 
 */
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":538
 *         self.grid.setSlowness(slown)
 * 
 *     def compute_D(self, coord):             # <<<<<<<<<<<<<<
 *         """
 *         compute_D(coord)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.compute_D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ivec);
  __Pyx_XDECREF(__pyx_v_jvec);
  __Pyx_XDECREF(__pyx_v_vec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":602
 *                                  shape=(coord.shape[0], self.get_number_of_nodes()))
 * 
 *     def compute_K(self):             # <<<<<<<<<<<<<<
 *         """
 *         Compute smoothing matrices (2nd order derivative)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_29compute_K(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_28compute_K[] = "\n        Compute smoothing matrices (2nd order derivative)\n\n        Returns\n        -------\n        Kx, Ky, Kz : :obj:`tuple` of :obj:`csr_matrix`\n            matrices for derivatives along x, y, & z\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_29compute_K(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_K (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_28compute_K(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_28compute_K(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_k;
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_ny = NULL;
  PyObject *__pyx_v_nz = NULL;
  PyObject *__pyx_v_iK = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_jK = NULL;
  PyObject *__pyx_v_Kx = NULL;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_Ky = NULL;
  PyObject *__pyx_v_Kz = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  Py_ssize_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_K", 0);

  /* "ttcrpy/rgrid.pyx":617
 *         cdef Py_ssize_t i, j, k
 * 
 *         nx, ny, nz = self.shape             # <<<<<<<<<<<<<<
 *         # Kx
 *         iK = np.kron(np.arange(nx*ny*nz, dtype=np.int64), np.ones(3,dtype=np.int64))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 617, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 617, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 617, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_nx = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_ny = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_nz = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "ttcrpy/rgrid.pyx":619
 *         nx, ny, nz = self.shape
 *         # Kx
 *         iK = np.kron(np.arange(nx*ny*nz, dtype=np.int64), np.ones(3,dtype=np.int64))             # <<<<<<<<<<<<<<
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dx*self._dx)
 *         # i=0 -> forward op
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_kron); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_Multiply(__pyx_t_4, __pyx_v_nz); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ones); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__15, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_8, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_8, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_8 = 0;
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_iK = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":620
 *         # Kx
 *         iK = np.kron(np.arange(nx*ny*nz, dtype=np.int64), np.ones(3,dtype=np.int64))
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dx*self._dx)             # <<<<<<<<<<<<<<
 *         # i=0 -> forward op
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_tile); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyList_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_float_1_);
  __Pyx_INCREF(__pyx_float_neg_2_);
  __Pyx_GIVEREF(__pyx_float_neg_2_);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_float_neg_2_);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_7, 2, __pyx_float_1_);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_9, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_3 = 0;
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->_dx * __pyx_v_self->_dx)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_val = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":622
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dx*self._dx)
 *         # i=0 -> forward op
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(ny*nz, 2*ny*nz,dtype=np.int64),
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vstack); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_ny, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":623
 *         # i=0 -> forward op
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),
 *                         np.arange(ny*nz, 2*ny*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_ny, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int64); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":624
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),
 *                         np.arange(ny*nz, 2*ny*nz,dtype=np.int64),
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()             # <<<<<<<<<<<<<<
 * 
 *         for i in np.arange(1,nx-1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_int_3, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int64); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":622
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dx*self._dx)
 *         # i=0 -> forward op
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(ny*nz, 2*ny*nz,dtype=np.int64),
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_12);
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_4 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_12, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":624
 *         jK = np.vstack((np.arange(ny*nz,dtype=np.int64),
 *                         np.arange(ny*nz, 2*ny*nz,dtype=np.int64),
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()             # <<<<<<<<<<<<<<
 * 
 *         for i in np.arange(1,nx-1):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_jK = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":626
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()
 * 
 *         for i in np.arange(1,nx-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_int_1, __pyx_t_4};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_int_1, __pyx_t_4};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_12 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_12, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_7 = __pyx_t_5; __Pyx_INCREF(__pyx_t_7); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_14 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 626, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_5); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 626, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 626, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_5); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 626, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 626, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_14(__pyx_t_7);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 626, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_i = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":627
 * 
 *         for i in np.arange(1,nx-1):
 *             jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "ttcrpy/rgrid.pyx":628
 *         for i in np.arange(1,nx-1):
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_vstack); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyInt_FromSsize_t((__pyx_v_i - 1)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_11, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_3, __pyx_v_ny); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
    __pyx_t_11 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int64); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_16) < 0) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, __pyx_t_3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":629
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))
 *         # i=nx-1 -> backward op
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_ny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_i + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyNumber_Multiply(__pyx_t_2, __pyx_v_ny); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_11, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int64); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":630
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_i + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_2, __pyx_v_ny); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyInt_FromSsize_t((__pyx_v_i + 2)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_8, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
    __pyx_t_2 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_18) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "ttcrpy/rgrid.pyx":628
 *         for i in np.arange(1,nx-1):
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))
 */
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_16);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_17);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_18);
    __pyx_t_16 = 0;
    __pyx_t_17 = 0;
    __pyx_t_18 = 0;
    __pyx_t_18 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_18)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_18);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_1 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_18, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "ttcrpy/rgrid.pyx":630
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_flatten); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_12 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":627
 * 
 *         for i in np.arange(1,nx-1):
 *             jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 *                                        np.arange(i*ny*nz, (i+1)*ny*nz,dtype=np.int64),
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_12);
    __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_12, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":626
 *                         np.arange(2*ny*nz, 3*ny*nz,dtype=np.int64))).T.flatten()
 * 
 *         for i in np.arange(1,nx-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((i-1)*ny*nz,i*ny*nz, dtype=np.int64),
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":632
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":633
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_vstack); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_arange); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_18 = PyNumber_Multiply(__pyx_t_12, __pyx_v_ny); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = PyNumber_Multiply(__pyx_t_18, __pyx_v_nz); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_17 = PyNumber_Multiply(__pyx_t_18, __pyx_v_ny); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_18);
  __pyx_t_12 = 0;
  __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_int64); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_dtype, __pyx_t_16) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_17, __pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "ttcrpy/rgrid.pyx":634
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_arange); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_8 = PyNumber_Multiply(__pyx_t_18, __pyx_v_ny); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = PyNumber_Multiply(__pyx_t_8, __pyx_v_ny); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyNumber_Multiply(__pyx_t_12, __pyx_v_nz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_8);
  __pyx_t_18 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_int64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_12, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "ttcrpy/rgrid.pyx":635
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 * 
 *         Kx = sp.csr_matrix((val,(iK,jK)))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_arange); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_17 = PyNumber_Multiply(__pyx_t_8, __pyx_v_ny); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_18);
  __pyx_t_8 = 0;
  __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int64); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_17, __pyx_t_18); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "ttcrpy/rgrid.pyx":633
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))
 */
  __pyx_t_18 = PyTuple_New(3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_t_11);
  __pyx_t_16 = 0;
  __pyx_t_3 = 0;
  __pyx_t_11 = 0;
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_18);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "ttcrpy/rgrid.pyx":635
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 * 
 *         Kx = sp.csr_matrix((val,(iK,jK)))
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_flatten); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":632
 *                                        np.arange((i+1)*ny*nz, (i+2)*ny*nz,dtype=np.int64))).T.flatten()))
 *         # i=nx-1 -> backward op
 *         jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                         np.vstack((np.arange((nx-3)*ny*nz, (nx-2)*ny*nz, dtype=np.int64),
 *                                    np.arange((nx-2)*ny*nz, (nx-1)*ny*nz,dtype=np.int64),
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_jK);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":637
 *                                    np.arange((nx-1)*ny*nz, nx*ny*nz,dtype=np.int64))).T.flatten()))
 * 
 *         Kx = sp.csr_matrix((val,(iK,jK)))             # <<<<<<<<<<<<<<
 * 
 *         # j=0 -> forward op
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_iK);
  __Pyx_GIVEREF(__pyx_v_iK);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_iK);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_jK);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_val);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_Kx = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":641
 *         # j=0 -> forward op
 * 
 *         jK = np.vstack((np.arange(nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(nz,2*nz,dtype=np.int64),
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_vstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nz);
  __pyx_t_18 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_5, __pyx_t_18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "ttcrpy/rgrid.pyx":642
 * 
 *         jK = np.vstack((np.arange(nz,dtype=np.int64),
 *                         np.arange(nz,2*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_arange); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = PyNumber_Multiply(__pyx_int_2, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_nz);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_18);
  __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int64); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_dtype, __pyx_t_16) < 0) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, __pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "ttcrpy/rgrid.pyx":643
 *         jK = np.vstack((np.arange(nz,dtype=np.int64),
 *                         np.arange(nz,2*nz,dtype=np.int64),
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()             # <<<<<<<<<<<<<<
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_arange); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = PyNumber_Multiply(__pyx_int_2, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_5 = PyNumber_Multiply(__pyx_int_3, __pyx_v_nz); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_5);
  __pyx_t_18 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_int64); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_11, __pyx_t_5); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":641
 *         # j=0 -> forward op
 * 
 *         jK = np.vstack((np.arange(nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                         np.arange(nz,2*nz,dtype=np.int64),
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_17);
  __pyx_t_3 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_17 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_17, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ttcrpy/rgrid.pyx":643
 *         jK = np.vstack((np.arange(nz,dtype=np.int64),
 *                         np.arange(nz,2*nz,dtype=np.int64),
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()             # <<<<<<<<<<<<<<
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_flatten); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":644
 *                         np.arange(nz,2*nz,dtype=np.int64),
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)             # <<<<<<<<<<<<<<
 *         for j in np.arange(1,ny-1):
 *             jK = np.hstack((jK,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_tile); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_float_1_);
  __Pyx_INCREF(__pyx_float_neg_2_);
  __Pyx_GIVEREF(__pyx_float_neg_2_);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_float_neg_2_);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_float_1_);
  __pyx_t_16 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_17);
    if (likely(__pyx_t_16)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_16);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_17, function);
    }
  }
  __pyx_t_1 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_16, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_5 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_t_1, __pyx_t_5};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_t_1, __pyx_t_5};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_16 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__pyx_t_17) {
      __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_17); __pyx_t_17 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_9, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_9, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_16, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->_dy * __pyx_v_self->_dy)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_16);
  __pyx_t_16 = 0;

  /* "ttcrpy/rgrid.pyx":645
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_1, __pyx_t_4};
    __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_1, __pyx_t_4};
    __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_9, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(PyList_CheckExact(__pyx_t_16)) || PyTuple_CheckExact(__pyx_t_16)) {
    __pyx_t_7 = __pyx_t_16; __Pyx_INCREF(__pyx_t_7); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_14 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 645, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_16 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_16); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
        #else
        __pyx_t_16 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_16 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_16); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
        #else
        __pyx_t_16 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
      }
    } else {
      __pyx_t_16 = __pyx_t_14(__pyx_t_7);
      if (unlikely(!__pyx_t_16)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 645, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_16);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_16); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_v_j = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":646
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):
 *             jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":647
 *         for j in np.arange(1,ny-1):
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_np); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_vstack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_np); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_arange); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = PyInt_FromSsize_t((__pyx_v_j - 1)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_10 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = PyInt_FromSsize_t(__pyx_v_j); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_18 = PyNumber_Multiply(__pyx_t_17, __pyx_v_nz); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_18);
    __pyx_t_10 = 0;
    __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_int64); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_17, __pyx_t_18); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "ttcrpy/rgrid.pyx":648
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))
 *         # j=ny-1 -> backward op
 */
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_arange); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = PyInt_FromSsize_t(__pyx_v_j); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_11 = PyNumber_Multiply(__pyx_t_18, __pyx_v_nz); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = PyInt_FromSsize_t((__pyx_v_j + 1)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_10 = PyNumber_Multiply(__pyx_t_18, __pyx_v_nz); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_10);
    __pyx_t_11 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_18, __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "ttcrpy/rgrid.pyx":649
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_arange); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyInt_FromSsize_t((__pyx_v_j + 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_17 = PyNumber_Multiply(__pyx_t_10, __pyx_v_nz); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyInt_FromSsize_t((__pyx_v_j + 2)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = PyNumber_Multiply(__pyx_t_10, __pyx_v_nz); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_17);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_11);
    __pyx_t_17 = 0;
    __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_np); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "ttcrpy/rgrid.pyx":647
 *         for j in np.arange(1,ny-1):
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))
 */
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_2);
    __pyx_t_12 = 0;
    __pyx_t_8 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":649
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_flatten); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":646
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):
 *             jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 *                                        np.arange(j*nz,(j+1)*nz,dtype=np.int64),
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_16 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "ttcrpy/rgrid.pyx":645
 *                         np.arange(2*nz,3*nz,dtype=np.int64))).T.flatten()
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dy*self._dy)
 *         for j in np.arange(1,ny-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,
 *                             np.vstack((np.arange((j-1)*nz,j*nz,dtype=np.int64),
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":651
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_np); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "ttcrpy/rgrid.pyx":652
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vstack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyNumber_Multiply(__pyx_t_1, __pyx_v_nz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8);
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "ttcrpy/rgrid.pyx":653
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 *         tmp = jK.copy()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_arange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_8, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __pyx_t_11 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":654
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_ny, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_ny, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int64); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_18) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":652
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),             # <<<<<<<<<<<<<<
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_18);
  __pyx_t_12 = 0;
  __pyx_t_10 = 0;
  __pyx_t_18 = 0;
  __pyx_t_18 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_18)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_18);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_18, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":654
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))             # <<<<<<<<<<<<<<
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_flatten); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_16 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":651
 *                                        np.arange((j+1)*nz,(j+2)*nz,dtype=np.int64))).T.flatten()))
 *         # j=ny-1 -> backward op
 *         jK = np.hstack((jK,             # <<<<<<<<<<<<<<
 *                         np.vstack((np.arange((ny-3)*nz,(ny-2)*nz,dtype=np.int64),
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_jK);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_16)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_16);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_7 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_16, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":655
 *                                    np.arange((ny-2)*nz,(ny-1)*nz,dtype=np.int64),
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 *         tmp = jK.copy()             # <<<<<<<<<<<<<<
 *         for i in np.arange(1,nx):
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_jK, __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_tmp = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":656
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_int_1, __pyx_v_nx};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_int_1, __pyx_v_nx};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  {
    __pyx_t_16 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_9, __pyx_v_nx);
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_16, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_5 = __pyx_t_7; __Pyx_INCREF(__pyx_t_5); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_14 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 656, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_7); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 656, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_7); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 656, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_14(__pyx_t_5);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 656, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_7); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_i = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":657
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):
 *             jK = np.hstack((jK, i*ny*nz+tmp))             # <<<<<<<<<<<<<<
 * 
 *         Ky = sp.csr_matrix((val,(iK,jK)))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_np); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_16, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_16, __pyx_v_tmp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_16);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":656
 *                                    np.arange((ny-1)*nz,ny*nz,dtype=np.int64))).T.flatten()))
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":659
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 *         Ky = sp.csr_matrix((val,(iK,jK)))             # <<<<<<<<<<<<<<
 * 
 *         # k=0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_sp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_iK);
  __Pyx_GIVEREF(__pyx_v_iK);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_iK);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_jK);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_val);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_16);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_Ky = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":662
 * 
 *         # k=0
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dz*self._dz)             # <<<<<<<<<<<<<<
 *         jK = np.arange(3,dtype=np.int64)
 *         for k in np.arange(1,nz-1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_tile); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyList_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_float_1_);
  __Pyx_INCREF(__pyx_float_neg_2_);
  __Pyx_GIVEREF(__pyx_float_neg_2_);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_float_neg_2_);
  __Pyx_INCREF(__pyx_float_1_);
  __Pyx_GIVEREF(__pyx_float_1_);
  PyList_SET_ITEM(__pyx_t_7, 2, __pyx_float_1_);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_ny); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_t_7};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_t_7};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_9, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = PyFloat_FromDouble((__pyx_v_self->_dz * __pyx_v_self->_dz)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":663
 *         # k=0
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dz*self._dz)
 *         jK = np.arange(3,dtype=np.int64)             # <<<<<<<<<<<<<<
 *         for k in np.arange(1,nz-1):
 *             jK = np.hstack((jK,(k-1)+np.arange(3,dtype=np.int64)))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_tuple__15, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":664
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dz*self._dz)
 *         jK = np.arange(3,dtype=np.int64)
 *         for k in np.arange(1,nz-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,(k-1)+np.arange(3,dtype=np.int64)))
 *         # k=nz-1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_1, __pyx_t_2};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_1, __pyx_t_2};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_16 = __pyx_t_7; __Pyx_INCREF(__pyx_t_16); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_16 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_14 = Py_TYPE(__pyx_t_16)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 664, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_16))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_16)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_16, __pyx_t_13); __Pyx_INCREF(__pyx_t_7); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 664, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_16, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_16)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_16, __pyx_t_13); __Pyx_INCREF(__pyx_t_7); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 664, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_16, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_14(__pyx_t_16);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 664, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_7); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_k = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":665
 *         jK = np.arange(3,dtype=np.int64)
 *         for k in np.arange(1,nz-1):
 *             jK = np.hstack((jK,(k-1)+np.arange(3,dtype=np.int64)))             # <<<<<<<<<<<<<<
 *         # k=nz-1
 *         jK = np.hstack((jK, (nz-3)+np.arange(3,dtype=np.int64)))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_hstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_k - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_np); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_int64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__15, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":664
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dz*self._dz)
 *         jK = np.arange(3,dtype=np.int64)
 *         for k in np.arange(1,nz-1):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK,(k-1)+np.arange(3,dtype=np.int64)))
 *         # k=nz-1
 */
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "ttcrpy/rgrid.pyx":667
 *             jK = np.hstack((jK,(k-1)+np.arange(3,dtype=np.int64)))
 *         # k=nz-1
 *         jK = np.hstack((jK, (nz-3)+np.arange(3,dtype=np.int64)))             # <<<<<<<<<<<<<<
 * 
 *         tmp = jK.copy()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_hstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_arange); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__15, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyNumber_Add(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_jK);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_10);
  __pyx_t_10 = 0;
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_16 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_16);
  __pyx_t_16 = 0;

  /* "ttcrpy/rgrid.pyx":669
 *         jK = np.hstack((jK, (nz-3)+np.arange(3,dtype=np.int64)))
 * 
 *         tmp = jK.copy()             # <<<<<<<<<<<<<<
 *         for j in np.arange(1,ny):
 *             jK = np.hstack((jK, j*nz+tmp))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_jK, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_16 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_tmp, __pyx_t_16);
  __pyx_t_16 = 0;

  /* "ttcrpy/rgrid.pyx":670
 * 
 *         tmp = jK.copy()
 *         for j in np.arange(1,ny):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, j*nz+tmp))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_int_1, __pyx_v_ny};
    __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_16);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_int_1, __pyx_v_ny};
    __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_16);
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_INCREF(__pyx_v_ny);
    __Pyx_GIVEREF(__pyx_v_ny);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_ny);
    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_16)) || PyTuple_CheckExact(__pyx_t_16)) {
    __pyx_t_3 = __pyx_t_16; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 670, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_16 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_16); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
        #else
        __pyx_t_16 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_16 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_16); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
        #else
        __pyx_t_16 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
      }
    } else {
      __pyx_t_16 = __pyx_t_14(__pyx_t_3);
      if (unlikely(!__pyx_t_16)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 670, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_16);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_16); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_v_j = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":671
 *         tmp = jK.copy()
 *         for j in np.arange(1,ny):
 *             jK = np.hstack((jK, j*nz+tmp))             # <<<<<<<<<<<<<<
 * 
 *         tmp = jK.copy()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_hstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyInt_FromSsize_t(__pyx_v_j); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_10, __pyx_v_nz); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Add(__pyx_t_7, __pyx_v_tmp); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_16 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "ttcrpy/rgrid.pyx":670
 * 
 *         tmp = jK.copy()
 *         for j in np.arange(1,ny):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, j*nz+tmp))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":673
 *             jK = np.hstack((jK, j*nz+tmp))
 * 
 *         tmp = jK.copy()             # <<<<<<<<<<<<<<
 *         for i in np.arange(1,nx):
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 */
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_jK, __pyx_n_s_copy); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF_SET(__pyx_v_tmp, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":674
 * 
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_np); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_16)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_16);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_int_1, __pyx_v_nx};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_int_1, __pyx_v_nx};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_16) {
      __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_16); __pyx_t_16 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_int_1);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_nx);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 674, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 674, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 674, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_14(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 674, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_i = __pyx_t_15;

    /* "ttcrpy/rgrid.pyx":675
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):
 *             jK = np.hstack((jK, i*ny*nz+tmp))             # <<<<<<<<<<<<<<
 * 
 *         Kz = sp.csr_matrix((val,(iK,jK)))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_hstack); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = PyNumber_Multiply(__pyx_t_7, __pyx_v_ny); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_10, __pyx_v_nz); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Add(__pyx_t_7, __pyx_v_tmp); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_jK);
    __Pyx_GIVEREF(__pyx_v_jK);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_jK);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_16);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_16, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_10, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF_SET(__pyx_v_jK, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":674
 * 
 *         tmp = jK.copy()
 *         for i in np.arange(1,nx):             # <<<<<<<<<<<<<<
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":677
 *             jK = np.hstack((jK, i*ny*nz+tmp))
 * 
 *         Kz = sp.csr_matrix((val,(iK,jK)))             # <<<<<<<<<<<<<<
 * 
 *         return Kx, Ky, Kz
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_iK);
  __Pyx_GIVEREF(__pyx_v_iK);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_iK);
  __Pyx_INCREF(__pyx_v_jK);
  __Pyx_GIVEREF(__pyx_v_jK);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_jK);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_val);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_v_Kz = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":679
 *         Kz = sp.csr_matrix((val,(iK,jK)))
 * 
 *         return Kx, Ky, Kz             # <<<<<<<<<<<<<<
 * 
 *     def get_s0(self, hypo, slowness=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_Kx);
  __Pyx_GIVEREF(__pyx_v_Kx);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_Kx);
  __Pyx_INCREF(__pyx_v_Ky);
  __Pyx_GIVEREF(__pyx_v_Ky);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_Ky);
  __Pyx_INCREF(__pyx_v_Kz);
  __Pyx_GIVEREF(__pyx_v_Kz);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_Kz);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":602
 *                                  shape=(coord.shape[0], self.get_number_of_nodes()))
 * 
 *     def compute_K(self):             # <<<<<<<<<<<<<<
 *         """
 *         Compute smoothing matrices (2nd order derivative)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.compute_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_ny);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_iK);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_jK);
  __Pyx_XDECREF(__pyx_v_Kx);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_Ky);
  __Pyx_XDECREF(__pyx_v_Kz);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":681
 *         return Kx, Ky, Kz
 * 
 *     def get_s0(self, hypo, slowness=None):             # <<<<<<<<<<<<<<
 *         """
 *         get_s0(hypo, slowness=None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_31get_s0(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_30get_s0[] = "\n        get_s0(hypo, slowness=None)\n\n        Return slowness at source points\n\n        Parameters\n        ----------\n        hypo : np.ndarray with 5 columns\n            hypo holds source information, i.e.\n                - 1st column is event ID number\n                - 2nd column is origin time\n                - 3rd column is source easting\n                - 4th column is source northing\n                - 5th column is source elevation\n\n        slowness : np ndarray, shape (nx, ny, nz) (optional)\n            slowness at grid nodes or cells (depending on cell_slowness)\n            slowness may also have been flattened (with default 'C' order)\n\n        Returns\n        -------\n        s0 : np.ndarray\n            slowness at source points\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_31get_s0(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hypo = 0;
  PyObject *__pyx_v_slowness = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_s0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hypo,&__pyx_n_s_slowness,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hypo)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_s0") < 0)) __PYX_ERR(0, 681, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_hypo = values[0];
    __pyx_v_slowness = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_s0", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 681, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_s0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_30get_s0(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_hypo, __pyx_v_slowness);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_30get_s0(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_hypo, PyObject *__pyx_v_slowness) {
  Py_ssize_t __pyx_v_n;
  Py_ssize_t __pyx_v_nn;
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_evID = NULL;
  PyObject *__pyx_v_eid = NULL;
  Py_ssize_t __pyx_v_nTx;
  PyObject *__pyx_v_i0 = NULL;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vTx;
  PyObject *__pyx_v_iTx = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_ii = NULL;
  PyObject *__pyx_v_s0 = NULL;
  double __pyx_v_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  long __pyx_t_11;
  long __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  double __pyx_t_14;
  double __pyx_t_15;
  double __pyx_t_16;
  ttcr::sxyz<double>  __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_t_19;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_20;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_s0", 0);

  /* "ttcrpy/rgrid.pyx":709
 *         cdef Py_ssize_t n, nn
 * 
 *         if hypo.shape[1] != 5:             # <<<<<<<<<<<<<<
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_hypo, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/rgrid.pyx":710
 * 
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')             # <<<<<<<<<<<<<<
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 710, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":709
 *         cdef Py_ssize_t n, nn
 * 
 *         if hypo.shape[1] != 5:             # <<<<<<<<<<<<<<
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 */
  }

  /* "ttcrpy/rgrid.pyx":711
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]             # <<<<<<<<<<<<<<
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_hypo, __pyx_tuple__18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_src = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":712
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]             # <<<<<<<<<<<<<<
 *         eid = np.sort(np.unique(evID))
 *         nTx = len(eid)
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_hypo, __pyx_tuple__4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_evID = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":713
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))             # <<<<<<<<<<<<<<
 *         nTx = len(eid)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sort); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_unique); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_v_evID) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_evID);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_eid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":714
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 *         nTx = len(eid)             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_eid); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_nTx = __pyx_t_7;

  /* "ttcrpy/rgrid.pyx":716
 *         nTx = len(eid)
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  __pyx_t_3 = (__pyx_v_slowness != Py_None);
  __pyx_t_8 = (__pyx_t_3 != 0);
  if (__pyx_t_8) {

    /* "ttcrpy/rgrid.pyx":717
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":716
 *         nTx = len(eid)
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":719
 *             self.set_slowness(slowness)
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_nTx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_i0 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":720
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/rgrid.pyx":721
 *         i0 = np.empty((nTx,), dtype=np.int64)
 *         for n in range(nTx):
 *             for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_5); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":722
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0[n] = nn
 *                     break
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_evID, __pyx_v_nn, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_8) {

        /* "ttcrpy/rgrid.pyx":723
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_nn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_i0, __pyx_v_n, __pyx_t_1, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 723, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":724
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn
 *                     break             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 */
        goto __pyx_L8_break;

        /* "ttcrpy/rgrid.pyx":722
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0[n] = nn
 *                     break
 */
      }
    }
    __pyx_L8_break:;
  }

  /* "ttcrpy/rgrid.pyx":727
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iTx = []
 */
  try {
    __pyx_v_vTx.resize(__pyx_v_nTx);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 727, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":729
 *         vTx.resize(nTx)
 * 
 *         iTx = []             # <<<<<<<<<<<<<<
 *         i0 = 0
 *         for n in range(nTx):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_iTx = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":730
 * 
 *         iTx = []
 *         i0 = 0             # <<<<<<<<<<<<<<
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_i0, __pyx_int_0);

  /* "ttcrpy/rgrid.pyx":731
 *         iTx = []
 *         i0 = 0
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/rgrid.pyx":732
 *         i0 = 0
 *         for n in range(nTx):
 *             for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":733
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0 = nn
 *                     break
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_evID, __pyx_v_nn, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_8) {

        /* "ttcrpy/rgrid.pyx":734
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn             # <<<<<<<<<<<<<<
 *                     break
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 */
        __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_nn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 734, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_i0, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/rgrid.pyx":735
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn
 *                     break             # <<<<<<<<<<<<<<
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 */
        goto __pyx_L13_break;

        /* "ttcrpy/rgrid.pyx":733
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0 = nn
 *                     break
 */
      }
    }
    __pyx_L13_break:;

    /* "ttcrpy/rgrid.pyx":736
 *                     i0 = nn
 *                     break
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))             # <<<<<<<<<<<<<<
 * 
 *         for i in eid:
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 736, __pyx_L1_error)
    }
    try {
      (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_17);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 736, __pyx_L1_error)
    }
  }

  /* "ttcrpy/rgrid.pyx":738
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 *         for i in eid:             # <<<<<<<<<<<<<<
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])
 */
  if (likely(PyList_CheckExact(__pyx_v_eid)) || PyTuple_CheckExact(__pyx_v_eid)) {
    __pyx_t_2 = __pyx_v_eid; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_18 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_eid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 738, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_18)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 738, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 738, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 738, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 738, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_18(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 738, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":739
 * 
 *         for i in eid:
 *             ii = evID == i             # <<<<<<<<<<<<<<
 *             iTx.append(np.nonzero(ii)[0])
 * 
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_evID, __pyx_v_i, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_ii, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":740
 *         for i in eid:
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])             # <<<<<<<<<<<<<<
 * 
 *         s0 = np.zeros((src.shape[0],))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_ii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_ii);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iTx, __pyx_t_4); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":738
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 *         for i in eid:             # <<<<<<<<<<<<<<
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":742
 *             iTx.append(np.nonzero(ii)[0])
 * 
 *         s0 = np.zeros((src.shape[0],))             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_s0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":744
 *         s0 = np.zeros((src.shape[0],))
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             s = 0.0
 *             for nn in range(vTx[n].size()):
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/rgrid.pyx":745
 * 
 *         for n in range(nTx):
 *             s = 0.0             # <<<<<<<<<<<<<<
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 */
    __pyx_v_s = 0.0;

    /* "ttcrpy/rgrid.pyx":746
 *         for n in range(nTx):
 *             s = 0.0
 *             for nn in range(vTx[n].size()):             # <<<<<<<<<<<<<<
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()
 */
    __pyx_t_20 = (__pyx_v_vTx[__pyx_v_n]).size();
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_21; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":747
 *             s = 0.0
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])             # <<<<<<<<<<<<<<
 *             s0[iTx[n]] = s/vTx[n].size()
 *         return s0
 */
      try {
        __pyx_t_16 = __pyx_v_self->grid->computeSlowness(((__pyx_v_vTx[__pyx_v_n])[__pyx_v_nn]));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 747, __pyx_L1_error)
      }
      __pyx_v_s = (__pyx_v_s + __pyx_t_16);
    }

    /* "ttcrpy/rgrid.pyx":748
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()             # <<<<<<<<<<<<<<
 *         return s0
 * 
 */
    __pyx_t_20 = (__pyx_v_vTx[__pyx_v_n]).size();
    if (unlikely(__pyx_t_20 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 748, __pyx_L1_error)
    }
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_s / ((double)__pyx_t_20))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iTx, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(PyObject_SetItem(__pyx_v_s0, __pyx_t_5, __pyx_t_2) < 0)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/rgrid.pyx":749
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()
 *         return s0             # <<<<<<<<<<<<<<
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_s0);
  __pyx_r = __pyx_v_s0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":681
 *         return Kx, Ky, Kz
 * 
 *     def get_s0(self, hypo, slowness=None):             # <<<<<<<<<<<<<<
 *         """
 *         get_s0(hypo, slowness=None)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.get_s0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_evID);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_i0);
  __Pyx_XDECREF(__pyx_v_iTx);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_ii);
  __Pyx_XDECREF(__pyx_v_s0);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":751
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, compute_M=False,
 *                  return_rays=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_33raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_32raytrace[] = "\n        raytrace(source, rcv, slowness=None, thread_no=None,\n                 aggregate_src=False, compute_L=False, compute_M=False,\n                 return_rays=False) -> tt, rays, M, L\n\n        Perform raytracing\n\n        Parameters\n        ----------\n        source : 2D np.ndarray with 3, 4 or 5 columns\n            see notes below\n        rcv : 2D np.ndarray with 3 columns\n            Columns correspond to x, y and z coordinates\n        slowness : np ndarray, shape (nx, ny, nz) (None by default)\n            slowness at grid nodes or cells (depending on cell_slowness)\n            slowness may also have been flattened (with default 'C' order)\n            if None, slowness must have been assigned previously\n        thread_no : int (None by default)\n            Perform calculations in thread number \"thread_no\"\n            if None, attempt to run in parallel if warranted by number of\n            sources and value of n_threads in constructor\n        aggregate_src : bool (False by default)\n            if True, all source coordinates belong to a single event\n        compute_L : bool (False by default)\n            Compute matrices of partial derivative of travel time w/r to slowness\n        compute_M : bool (False by default)\n            Compute matrices of partial derivative of travel time w/r to velocity\n            Note : compute_M and compute_L are mutually exclusive\n        return_rays : bool (False by default)\n            Return raypaths\n\n        Returns\n        -------\n        tt : np.ndarray\n            travel times for the appropriate source-rcv  (see Notes below)\n        rays : :obj:`list` of :obj:`np.ndarray`\n            Coordinates of segments forming raypaths (if return_rays is True)\n        M : :obj:`list` of :obj:`csr_matrix`\n            matrices of partial derivative of travel time w/r to velocity.\n            the number of matrices is equal to the number of sources\n        L : scipy csr_matrix\n            Matrix"" of partial derivative of travel time w/r to slowness.\n            if input argument source has 5 columns, L is a list of matrices and\n            the number of matrices is equal to the number of sources\n            otherwise, L is a single csr_matrix\n\n        Notes\n        -----\n        If source has 3 columns:\n            - Columns correspond to x, y and z coordinates\n            - Origin time (t0) is 0 for all points\n        If source has 4 columns:\n            - 1st column corresponds to origin times\n            - 2nd, 3rd & 4th columns correspond to x, y and z coordinates\n        If source has 5 columns:\n            - 1st column corresponds to event ID\n            - 2nd column corresponds to origin times\n            - 3rd, 4th & 5th columns correspond to x, y and z coordinates\n\n        For the latter case (5 columns), source and rcv should contain the same\n        number of rows, each row corresponding to a source-receiver pair.\n        For the 2 other cases, source and rcv can contain the same number of\n        rows, each row corresponding to a source-receiver pair, or the number\n        of rows may differ if aggregate_src is True or if all rows in source\n        are identical.\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_33raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_rcv = 0;
  PyObject *__pyx_v_slowness = 0;
  PyObject *__pyx_v_thread_no = 0;
  PyObject *__pyx_v_aggregate_src = 0;
  PyObject *__pyx_v_compute_L = 0;
  PyObject *__pyx_v_compute_M = 0;
  PyObject *__pyx_v_return_rays = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("raytrace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_rcv,&__pyx_n_s_slowness,&__pyx_n_s_thread_no,&__pyx_n_s_aggregate_src,&__pyx_n_s_compute_L,&__pyx_n_s_compute_M,&__pyx_n_s_return_rays,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);

    /* "ttcrpy/rgrid.pyx":752
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 *                  aggregate_src=False, compute_L=False, compute_M=False,             # <<<<<<<<<<<<<<
 *                  return_rays=False):
 *         """
 */
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)Py_False);
    values[6] = ((PyObject *)Py_False);

    /* "ttcrpy/rgrid.pyx":753
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 *                  aggregate_src=False, compute_L=False, compute_M=False,
 *                  return_rays=False):             # <<<<<<<<<<<<<<
 *         """
 *         raytrace(source, rcv, slowness=None, thread_no=None,
 */
    values[7] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rcv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 8, 1); __PYX_ERR(0, 751, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aggregate_src);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_L);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_M);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_rays);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "raytrace") < 0)) __PYX_ERR(0, 751, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_rcv = values[1];
    __pyx_v_slowness = values[2];
    __pyx_v_thread_no = values[3];
    __pyx_v_aggregate_src = values[4];
    __pyx_v_compute_L = values[5];
    __pyx_v_compute_M = values[6];
    __pyx_v_return_rays = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 751, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_32raytrace(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_source, __pyx_v_rcv, __pyx_v_slowness, __pyx_v_thread_no, __pyx_v_aggregate_src, __pyx_v_compute_L, __pyx_v_compute_M, __pyx_v_return_rays);

  /* "ttcrpy/rgrid.pyx":751
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, compute_M=False,
 *                  return_rays=False):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_32raytrace(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_compute_M, PyObject *__pyx_v_return_rays) {
  PyObject *__pyx_v_evID = NULL;
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_t0 = NULL;
  PyObject *__pyx_v_eid = NULL;
  PyObject *__pyx_v_nTx = NULL;
  PyObject *__pyx_v_Tx = NULL;
  PyObject *__pyx_v_tmp = NULL;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vTx;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vRx;
  std::vector<std::vector<double> >  __pyx_v_vt0;
  std::vector<std::vector<double> >  __pyx_v_vtt;
  std::vector<std::vector<std::vector<ttcr::sxyz<double> > > >  __pyx_v_r_data;
  std::vector<std::vector<std::vector<ttcr::siv<double> > > >  __pyx_v_l_data;
  std::vector<std::vector<std::vector<ttcr::sijv<double> > > >  __pyx_v_m_data;
  size_t __pyx_v_thread_nb;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_n;
  int __pyx_v_nn;
  int __pyx_v_MM;
  int __pyx_v_NN;
  PyObject *__pyx_v_iRx = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_ind = NULL;
  long __pyx_v_i0;
  PyObject *__pyx_v_ii = NULL;
  PyObject *__pyx_v_tt = NULL;
  std::vector<double> ::size_type __pyx_v_nt;
  PyObject *__pyx_v_rays = NULL;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_v_n2;
  PyObject *__pyx_v_L = NULL;
  PyObject *__pyx_v_nnz = NULL;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_v_ni;
  PyObject *__pyx_v_indptr = NULL;
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_itmp = NULL;
  PyObject *__pyx_v_M = NULL;
  CYTHON_UNUSED int __pyx_7genexpr__pyx_v_n;
  CYTHON_UNUSED int __pyx_8genexpr1__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_9;
  std::vector<std::vector<double> > ::size_type __pyx_t_10;
  std::vector<std::vector<std::vector<ttcr::siv<double> > > > ::size_type __pyx_t_11;
  std::vector<std::vector<std::vector<ttcr::sijv<double> > > > ::size_type __pyx_t_12;
  std::vector<std::vector<std::vector<ttcr::sxyz<double> > > > ::size_type __pyx_t_13;
  double __pyx_t_14;
  double __pyx_t_15;
  double __pyx_t_16;
  ttcr::sxyz<double>  __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  std::vector<double> ::size_type __pyx_t_19;
  int __pyx_t_20;
  long __pyx_t_21;
  long __pyx_t_22;
  int __pyx_t_23;
  long __pyx_t_24;
  long __pyx_t_25;
  int __pyx_t_26;
  size_t __pyx_t_27;
  std::vector<double> ::size_type __pyx_t_28;
  std::vector<double> ::size_type __pyx_t_29;
  std::vector<double> ::size_type __pyx_t_30;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_31;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_32;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_33;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_34;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_35;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_36;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_37;
  std::vector<std::vector<ttcr::siv<double> > > ::size_type __pyx_t_38;
  std::vector<std::vector<ttcr::siv<double> > > ::size_type __pyx_t_39;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_40;
  int __pyx_t_41;
  int __pyx_t_42;
  int __pyx_t_43;
  int __pyx_t_44;
  int __pyx_t_45;
  std::vector<ttcr::siv<double> > ::size_type __pyx_t_46;
  std::vector<ttcr::siv<double> > ::size_type __pyx_t_47;
  int __pyx_t_48;
  PyObject *__pyx_t_49 = NULL;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_50;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_51;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_52;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_53;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raytrace", 0);

  /* "ttcrpy/rgrid.pyx":823
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":824
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if self.method == b'd' and aggregate_src:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 824, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":823
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":826
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.method == b'd' and aggregate_src:             # <<<<<<<<<<<<<<
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 */
  __pyx_t_4 = ((__pyx_v_self->method == 'd') != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 826, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":827
 * 
 *         if self.method == b'd' and aggregate_src:
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and compute_M:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 827, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":826
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.method == b'd' and aggregate_src:             # <<<<<<<<<<<<<<
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":829
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 *         if compute_L and compute_M:             # <<<<<<<<<<<<<<
 *             raise ValueError('compute_L and compute_M are mutually exclusive')
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 829, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 829, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":830
 * 
 *         if compute_L and compute_M:
 *             raise ValueError('compute_L and compute_M are mutually exclusive')             # <<<<<<<<<<<<<<
 * 
 *         if self.cell_slowness and compute_M:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 830, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":829
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 *         if compute_L and compute_M:             # <<<<<<<<<<<<<<
 *             raise ValueError('compute_L and compute_M are mutually exclusive')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":832
 *             raise ValueError('compute_L and compute_M are mutually exclusive')
 * 
 *         if self.cell_slowness and compute_M:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')
 * 
 */
  __pyx_t_4 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 832, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L13_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":833
 * 
 *         if self.cell_slowness and compute_M:
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and not self.cell_slowness:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 833, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":832
 *             raise ValueError('compute_L and compute_M are mutually exclusive')
 * 
 *         if self.cell_slowness and compute_M:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":835
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 835, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_v_self->cell_slowness != 0)) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":836
 * 
 *         if compute_L and not self.cell_slowness:
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')             # <<<<<<<<<<<<<<
 * 
 *         evID = None
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 836, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":835
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":838
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 *         evID = None             # <<<<<<<<<<<<<<
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_evID = Py_None;

  /* "ttcrpy/rgrid.pyx":839
 * 
 *         evID = None
 *         if source.shape[1] == 5:             # <<<<<<<<<<<<<<
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":840
 *         evID = None
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]             # <<<<<<<<<<<<<<
 *             t0 = source[:,1]
 *             evID = source[:,0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__18); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":841
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]
 *             t0 = source[:,1]             # <<<<<<<<<<<<<<
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_t0 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":842
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 *             evID = source[:,0]             # <<<<<<<<<<<<<<
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_evID, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":843
 *             t0 = source[:,1]
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))             # <<<<<<<<<<<<<<
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sort); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_evID) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_evID);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_eid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":844
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 3:
 *             src = source
 */
    __pyx_t_8 = PyObject_Length(__pyx_v_eid); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 844, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":839
 * 
 *         evID = None
 *         if source.shape[1] == 5:             # <<<<<<<<<<<<<<
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 */
    goto __pyx_L18;
  }

  /* "ttcrpy/rgrid.pyx":845
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":846
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:
 *             src = source             # <<<<<<<<<<<<<<
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 */
    __Pyx_INCREF(__pyx_v_source);
    __pyx_v_src = __pyx_v_source;

    /* "ttcrpy/rgrid.pyx":847
 *         elif source.shape[1] == 3:
 *             src = source
 *             Tx = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 847, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_Tx = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":848
 *             src = source
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))             # <<<<<<<<<<<<<<
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
    __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_t0 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":849
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":845
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
    goto __pyx_L18;
  }

  /* "ttcrpy/rgrid.pyx":850
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:             # <<<<<<<<<<<<<<
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":851
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":852
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 852, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":853
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]             # <<<<<<<<<<<<<<
 *             Tx = tmp[:,1:4]
 *             t0 = tmp[:,0]
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmp, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_nTx = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":854
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]             # <<<<<<<<<<<<<<
 *             t0 = tmp[:,0]
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__25); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_Tx = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":855
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]
 *             t0 = tmp[:,0]             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_t0 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":850
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:             # <<<<<<<<<<<<<<
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 */
    goto __pyx_L18;
  }

  /* "ttcrpy/rgrid.pyx":857
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 857, __pyx_L1_error)
  }
  __pyx_L18:;

  /* "ttcrpy/rgrid.pyx":859
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L20_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L20_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":860
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 *             raise ValueError('src and rcv should be ndata x 3')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 860, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":859
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":862
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_src);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":863
 * 
 *         if self.is_outside(src):
 *             raise ValueError('Source point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(rcv):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 863, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":862
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":865
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_rcv) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_rcv);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":866
 * 
 *         if self.is_outside(rcv):
 *             raise ValueError('Receiver outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 866, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":865
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":868
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  __pyx_t_1 = (__pyx_v_slowness != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":869
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":868
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":883
 *         cdef int i, j, k, n, nn, MM, NN
 * 
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 */
  __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 883, __pyx_L1_error)
  try {
    __pyx_v_vTx.resize(__pyx_t_9);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 883, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":884
 * 
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 */
  __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 884, __pyx_L1_error)
  try {
    __pyx_v_vRx.resize(__pyx_t_9);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 884, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":885
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)             # <<<<<<<<<<<<<<
 *         vtt.resize(nTx)
 *         if compute_L:
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 885, __pyx_L1_error)
  try {
    __pyx_v_vt0.resize(__pyx_t_10);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 885, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":886
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)             # <<<<<<<<<<<<<<
 *         if compute_L:
 *             l_data.resize(nTx)
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 886, __pyx_L1_error)
  try {
    __pyx_v_vtt.resize(__pyx_t_10);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 886, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":887
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             l_data.resize(nTx)
 *         if compute_M:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 887, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":888
 *         vtt.resize(nTx)
 *         if compute_L:
 *             l_data.resize(nTx)             # <<<<<<<<<<<<<<
 *         if compute_M:
 *             m_data.resize(nTx)
 */
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L1_error)
    try {
      __pyx_v_l_data.resize(__pyx_t_11);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 888, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":887
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             l_data.resize(nTx)
 *         if compute_M:
 */
  }

  /* "ttcrpy/rgrid.pyx":889
 *         if compute_L:
 *             l_data.resize(nTx)
 *         if compute_M:             # <<<<<<<<<<<<<<
 *             m_data.resize(nTx)
 *         if return_rays:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 889, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":890
 *             l_data.resize(nTx)
 *         if compute_M:
 *             m_data.resize(nTx)             # <<<<<<<<<<<<<<
 *         if return_rays:
 *             r_data.resize(nTx)
 */
    __pyx_t_12 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_12 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L1_error)
    try {
      __pyx_v_m_data.resize(__pyx_t_12);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 890, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":889
 *         if compute_L:
 *             l_data.resize(nTx)
 *         if compute_M:             # <<<<<<<<<<<<<<
 *             m_data.resize(nTx)
 *         if return_rays:
 */
  }

  /* "ttcrpy/rgrid.pyx":891
 *         if compute_M:
 *             m_data.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 891, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":892
 *             m_data.resize(nTx)
 *         if return_rays:
 *             r_data.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iRx = []
 */
    __pyx_t_13 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_13 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 892, __pyx_L1_error)
    try {
      __pyx_v_r_data.resize(__pyx_t_13);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 892, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":891
 *         if compute_M:
 *             m_data.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":894
 *             r_data.resize(nTx)
 * 
 *         iRx = []             # <<<<<<<<<<<<<<
 *         if evID is None:
 *             if nTx == 1:
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_iRx = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":895
 * 
 *         iRx = []
 *         if evID is None:             # <<<<<<<<<<<<<<
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 */
  __pyx_t_4 = (__pyx_v_evID == Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":896
 *         iRx = []
 *         if evID is None:
 *             if nTx == 1:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 */
    __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":897
 *         if evID is None:
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__30); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__31); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 897, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[0]).push_back(__pyx_t_17);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 897, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":898
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 */
      if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
        __pyx_t_3 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
        __pyx_t_18 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_18 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 898, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_18)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 898, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 898, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_18(__pyx_t_3);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 898, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/rgrid.pyx":899
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        try {
          __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_16, __pyx_t_15, __pyx_t_14);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 899, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[0]).push_back(__pyx_t_17);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 899, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":898
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "ttcrpy/rgrid.pyx":900
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])             # <<<<<<<<<<<<<<
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_t0, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        (__pyx_v_vt0[0]).push_back(__pyx_t_14);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 900, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":901
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 901, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 901, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = __Pyx_PyInt_As_size_t(__pyx_t_5); if (unlikely((__pyx_t_19 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 901, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        (__pyx_v_vtt[0]).resize(__pyx_t_19);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 901, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":902
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *             elif aggregate_src:
 *                 for t in Tx:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_5); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":896
 *         iRx = []
 *         if evID is None:
 *             if nTx == 1:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 */
      goto __pyx_L29;
    }

    /* "ttcrpy/rgrid.pyx":903
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:             # <<<<<<<<<<<<<<
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 903, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":904
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 *                 for t in Tx:             # <<<<<<<<<<<<<<
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 */
      if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 904, __pyx_L1_error) }
      if (likely(PyList_CheckExact(__pyx_v_Tx)) || PyTuple_CheckExact(__pyx_v_Tx)) {
        __pyx_t_5 = __pyx_v_Tx; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_18 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_Tx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_18 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 904, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_18)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 904, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 904, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_18(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 904, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":905
 *             elif aggregate_src:
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))             # <<<<<<<<<<<<<<
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 905, __pyx_L1_error)
        }
        try {
          (__pyx_v_vTx[0]).push_back(__pyx_t_17);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 905, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":904
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 *                 for t in Tx:             # <<<<<<<<<<<<<<
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":906
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:             # <<<<<<<<<<<<<<
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 */
      if (likely(PyList_CheckExact(__pyx_v_t0)) || PyTuple_CheckExact(__pyx_v_t0)) {
        __pyx_t_5 = __pyx_v_t0; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_18 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_t0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 906, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_18 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 906, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_18)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 906, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 906, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_18(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 906, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":907
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 *                     vt0[0].push_back(t)             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_v_t); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 907, __pyx_L1_error)
        try {
          (__pyx_v_vt0[0]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 907, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":906
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:             # <<<<<<<<<<<<<<
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":908
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 */
      if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
        __pyx_t_5 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_18 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 908, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_18 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 908, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_18)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_18(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 908, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":909
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_16, __pyx_t_15, __pyx_t_14);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 909, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[0]).push_back(__pyx_t_17);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 909, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":908
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":910
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *                 nTx = 1
 *                 iRx.append(np.arange(rcv.shape[0]))
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_19 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_19 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        (__pyx_v_vtt[0]).resize(__pyx_t_19);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 910, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":911
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1             # <<<<<<<<<<<<<<
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 */
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_DECREF_SET(__pyx_v_nTx, __pyx_int_1);

      /* "ttcrpy/rgrid.pyx":912
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1
 *                 iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *             else:
 *                 if src.shape != rcv.shape:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_2); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":903
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:             # <<<<<<<<<<<<<<
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 */
      goto __pyx_L29;
    }

    /* "ttcrpy/rgrid.pyx":914
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 *                 if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_t_1)) {

        /* "ttcrpy/rgrid.pyx":915
 *             else:
 *                 if src.shape != rcv.shape:
 *                     raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *                 for n in range(nTx):
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 915, __pyx_L1_error)

        /* "ttcrpy/rgrid.pyx":914
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 *                 if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 */
      }

      /* "ttcrpy/rgrid.pyx":917
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":918
 * 
 *                 for n in range(nTx):
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3             # <<<<<<<<<<<<<<
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sum); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 918, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_slice__3);
        __Pyx_GIVEREF(__pyx_slice__3);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice__3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_v_src, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_1) < 0) __PYX_ERR(0, 918, __pyx_L1_error)
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ind, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":919
 *                 for n in range(nTx):
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])             # <<<<<<<<<<<<<<
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_ind) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ind);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_3); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":920
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))             # <<<<<<<<<<<<<<
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:
 */
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 920, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 920, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_1);
        __Pyx_GIVEREF(__pyx_int_1);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 920, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 920, __pyx_L1_error)
        }
        try {
          (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_17);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 920, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":921
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])             # <<<<<<<<<<<<<<
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_t0, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 921, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 921, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":922
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())
 */
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_ind);
        __Pyx_GIVEREF(__pyx_v_ind);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_ind);
        __Pyx_INCREF(__pyx_slice__3);
        __Pyx_GIVEREF(__pyx_slice__3);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_slice__3);
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
          __pyx_t_18 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 922, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_18 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 922, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_18)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 922, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 922, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_18(__pyx_t_3);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 922, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":923
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 */
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          try {
            __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_16, __pyx_t_15, __pyx_t_14);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 923, __pyx_L1_error)
          }
          try {
            (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_17);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 923, __pyx_L1_error)
          }

          /* "ttcrpy/rgrid.pyx":922
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":924
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 *         else:
 *             if src.shape != rcv.shape:
 */
        try {
          (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 924, __pyx_L1_error)
        }
      }
    }
    __pyx_L29:;

    /* "ttcrpy/rgrid.pyx":895
 * 
 *         iRx = []
 *         if evID is None:             # <<<<<<<<<<<<<<
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 */
    goto __pyx_L28;
  }

  /* "ttcrpy/rgrid.pyx":926
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":927
 *         else:
 *             if src.shape != rcv.shape:
 *                 raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *             i0 = 0
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 927, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 927, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":926
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
    }

    /* "ttcrpy/rgrid.pyx":929
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 *             i0 = 0             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 */
    __pyx_v_i0 = 0;

    /* "ttcrpy/rgrid.pyx":930
 * 
 *             i0 = 0
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:
 */
    __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
      __pyx_v_n = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":931
 *             i0 = 0
 *             for n in range(nTx):
 *                 for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_24 = __Pyx_PyInt_As_long(__pyx_t_5); if (unlikely((__pyx_t_24 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_25 = __pyx_t_24;
      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
        __pyx_v_nn = __pyx_t_26;

        /* "ttcrpy/rgrid.pyx":932
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                         i0 = nn
 *                         break
 */
        if (unlikely(!__pyx_v_eid)) { __Pyx_RaiseUnboundLocalError("eid"); __PYX_ERR(0, 932, __pyx_L1_error) }
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 932, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_evID, __pyx_v_nn, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 932, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 932, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_1) {

          /* "ttcrpy/rgrid.pyx":933
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn             # <<<<<<<<<<<<<<
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 */
          __pyx_v_i0 = __pyx_v_nn;

          /* "ttcrpy/rgrid.pyx":934
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn
 *                         break             # <<<<<<<<<<<<<<
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 *                 vt0[n].push_back(t0[i0])
 */
          goto __pyx_L47_break;

          /* "ttcrpy/rgrid.pyx":932
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                         i0 = nn
 *                         break
 */
        }
      }
      __pyx_L47_break:;

      /* "ttcrpy/rgrid.pyx":935
 *                         i0 = nn
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))             # <<<<<<<<<<<<<<
 *                 vt0[n].push_back(t0[i0])
 * 
 */
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 935, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_17);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 935, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":936
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 *                 vt0[n].push_back(t0[i0])             # <<<<<<<<<<<<<<
 * 
 *             for i in eid:
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_t0, __pyx_v_i0, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_16);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 936, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":938
 *                 vt0[n].push_back(t0[i0])
 * 
 *             for i in eid:             # <<<<<<<<<<<<<<
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])
 */
    if (unlikely(!__pyx_v_eid)) { __Pyx_RaiseUnboundLocalError("eid"); __PYX_ERR(0, 938, __pyx_L1_error) }
    if (likely(PyList_CheckExact(__pyx_v_eid)) || PyTuple_CheckExact(__pyx_v_eid)) {
      __pyx_t_3 = __pyx_v_eid; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
      __pyx_t_18 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_eid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_18 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 938, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_18)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 938, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 938, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_18(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 938, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_23 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_23 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_i = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":939
 * 
 *             for i in eid:
 *                 ii = evID == i             # <<<<<<<<<<<<<<
 *                 iRx.append(np.nonzero(ii)[0])
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_evID, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ii, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":940
 *             for i in eid:
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nTx):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_v_ii) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_ii);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "ttcrpy/rgrid.pyx":938
 *                 vt0[n].push_back(t0[i0])
 * 
 *             for i in eid:             # <<<<<<<<<<<<<<
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":942
 *                 iRx.append(np.nonzero(ii)[0])
 * 
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 */
    __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
      __pyx_v_n = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":943
 * 
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())
 */
      __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_slice__3);
      __Pyx_GIVEREF(__pyx_slice__3);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_slice__3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_7 = __pyx_t_3; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
        __pyx_t_18 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_18 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 943, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_18)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 943, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 943, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_18(__pyx_t_7);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 943, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":944
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vtt[n].resize(vRx[n].size())
 * 
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_16, __pyx_t_15, __pyx_t_14);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 944, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_17);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 944, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":943
 * 
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())
 */
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "ttcrpy/rgrid.pyx":945
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 * 
 *         # for n in range(nTx):
 */
      try {
        (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 945, __pyx_L1_error)
      }
    }
  }
  __pyx_L28:;

  /* "ttcrpy/rgrid.pyx":960
 *         #     print('  size of tt = ',vtt[n].size())
 * 
 *         tt = np.zeros((rcv.shape[0],))             # <<<<<<<<<<<<<<
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and compute_M==False and return_rays==False:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_tt = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "ttcrpy/rgrid.pyx":961
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 for n in range(nTx):
 */
  __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_nTx, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L56_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_self->_n_threads == 1) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L56_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":962
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L59_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_compute_M, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L59_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_L59_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":963
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_M and return_rays:
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":964
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_M and return_rays:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 964, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":962
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
      goto __pyx_L58;
    }

    /* "ttcrpy/rgrid.pyx":965
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 965, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L64_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 965, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L64_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":966
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_M and return_rays:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L and return_rays:
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":967
 *             elif compute_M and return_rays:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), (__pyx_v_m_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 967, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":965
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 */
      goto __pyx_L58;
    }

    /* "ttcrpy/rgrid.pyx":968
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L68_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L68_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":969
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":970
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 970, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":968
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 */
      goto __pyx_L58;
    }

    /* "ttcrpy/rgrid.pyx":971
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 971, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":972
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             elif compute_M:
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":973
 *             elif compute_L:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_M:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 973, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":971
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
      goto __pyx_L58;
    }

    /* "ttcrpy/rgrid.pyx":974
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             elif compute_M:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 974, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":975
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             elif compute_M:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 *             else:
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":976
 *             elif compute_M:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_m_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 976, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":974
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             elif compute_M:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 */
      goto __pyx_L58;
    }

    /* "ttcrpy/rgrid.pyx":978
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 *             else:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 */
    /*else*/ {
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":979
 *             else:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)             # <<<<<<<<<<<<<<
 * 
 *         elif thread_no is not None:
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 979, __pyx_L1_error)
        }
      }
    }
    __pyx_L58:;

    /* "ttcrpy/rgrid.pyx":961
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 for n in range(nTx):
 */
    goto __pyx_L55;
  }

  /* "ttcrpy/rgrid.pyx":981
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  __pyx_t_1 = (__pyx_v_thread_no != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":983
 *         elif thread_no is not None:
 *             # we should be here for just one event
 *             assert nTx == 1             # <<<<<<<<<<<<<<
 *             # normally we should not need to compute M or L
 *             assert compute_L is False
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_5 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 983, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 983, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 983, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/rgrid.pyx":985
 *             assert nTx == 1
 *             # normally we should not need to compute M or L
 *             assert compute_L is False             # <<<<<<<<<<<<<<
 *             assert compute_M is False
 *             thread_nb = thread_no
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_4 = (__pyx_v_compute_L == Py_False);
      if (unlikely(!(__pyx_t_4 != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 985, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/rgrid.pyx":986
 *             # normally we should not need to compute M or L
 *             assert compute_L is False
 *             assert compute_M is False             # <<<<<<<<<<<<<<
 *             thread_nb = thread_no
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_4 = (__pyx_v_compute_M == Py_False);
      if (unlikely(!(__pyx_t_4 != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 986, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/rgrid.pyx":987
 *             assert compute_L is False
 *             assert compute_M is False
 *             thread_nb = thread_no             # <<<<<<<<<<<<<<
 * 
 *             if return_rays:
 */
    __pyx_t_27 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_27 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L1_error)
    __pyx_v_thread_nb = __pyx_t_27;

    /* "ttcrpy/rgrid.pyx":989
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 989, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":990
 * 
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), (__pyx_v_r_data[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 990, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":991
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 */
      __pyx_t_28 = (__pyx_v_vtt[0]).size();
      __pyx_t_29 = __pyx_t_28;
      for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
        __pyx_v_nt = __pyx_t_30;

        /* "ttcrpy/rgrid.pyx":992
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 */
        __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_5, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }

      /* "ttcrpy/rgrid.pyx":993
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_rays = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":994
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 *                 for n2 in range(vRx.size()):             # <<<<<<<<<<<<<<
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):
 */
      __pyx_t_31 = __pyx_v_vRx.size();
      __pyx_t_32 = __pyx_t_31;
      for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
        __pyx_v_n2 = __pyx_t_33;

        /* "ttcrpy/rgrid.pyx":995
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[0])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
        __Pyx_INCREF(__pyx_int_3);
        __Pyx_GIVEREF(__pyx_int_3);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
        __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/rgrid.pyx":996
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y
 */
        __pyx_t_34 = ((__pyx_v_r_data[0])[__pyx_v_n2]).size();
        __pyx_t_35 = __pyx_t_34;
        for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_35; __pyx_t_23+=1) {
          __pyx_v_nn = __pyx_t_23;

          /* "ttcrpy/rgrid.pyx":997
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
          __pyx_t_3 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_2, __pyx_t_5) < 0)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":998
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y             # <<<<<<<<<<<<<<
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
          __pyx_t_2 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 998, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":999
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z             # <<<<<<<<<<<<<<
 *                     rays.append(r)
 *                 return tt, rays
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
          __Pyx_INCREF(__pyx_int_2);
          __Pyx_GIVEREF(__pyx_int_2);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
          __pyx_t_3 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_2, __pyx_t_5) < 0)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }

        /* "ttcrpy/rgrid.pyx":1000
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)             # <<<<<<<<<<<<<<
 *                 return tt, rays
 * 
 */
        __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_rays, __pyx_v_r); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1000, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1001
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)
 *                 return tt, rays             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_tt);
      __Pyx_GIVEREF(__pyx_v_tt);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_tt);
      __Pyx_INCREF(__pyx_v_rays);
      __Pyx_GIVEREF(__pyx_v_rays);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_rays);
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "ttcrpy/rgrid.pyx":989
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    }

    /* "ttcrpy/rgrid.pyx":1004
 * 
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1004, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1005
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 return tt
 */
      __pyx_t_28 = (__pyx_v_vtt[0]).size();
      __pyx_t_29 = __pyx_t_28;
      for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
        __pyx_v_nt = __pyx_t_30;

        /* "ttcrpy/rgrid.pyx":1006
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 return tt
 * 
 */
        __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1006, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_5, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 1006, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }

      /* "ttcrpy/rgrid.pyx":1007
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 return tt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_tt);
      __pyx_r = __pyx_v_tt;
      goto __pyx_L0;
    }

    /* "ttcrpy/rgrid.pyx":981
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  }

  /* "ttcrpy/rgrid.pyx":1010
 * 
 *         else:
 *             if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_M and return_rays:
 */
  /*else*/ {
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L88_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_compute_M, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L88_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_L88_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":1011
 *         else:
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)             # <<<<<<<<<<<<<<
 *             elif compute_M and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1011, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1010
 * 
 *         else:
 *             if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_M and return_rays:
 */
      goto __pyx_L87;
    }

    /* "ttcrpy/rgrid.pyx":1012
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L and return_rays:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L91_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L91_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":1013
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_M and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data, __pyx_v_m_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1013, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1012
 *             if compute_L==False and compute_M==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L and return_rays:
 */
      goto __pyx_L87;
    }

    /* "ttcrpy/rgrid.pyx":1014
 *             elif compute_M and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L93_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L93_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":1015
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data, __pyx_v_l_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1015, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1014
 *             elif compute_M and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 */
      goto __pyx_L87;
    }

    /* "ttcrpy/rgrid.pyx":1016
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             elif compute_M:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1016, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":1017
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)             # <<<<<<<<<<<<<<
 *             elif compute_M:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_l_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1017, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1016
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             elif compute_M:
 */
      goto __pyx_L87;
    }

    /* "ttcrpy/rgrid.pyx":1018
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             elif compute_M:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1018, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":1019
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             elif compute_M:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_m_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1019, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":1018
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             elif compute_M:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 *             else:
 */
      goto __pyx_L87;
    }

    /* "ttcrpy/rgrid.pyx":1021
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1021, __pyx_L1_error)
      }
    }
    __pyx_L87:;
  }
  __pyx_L55:;

  /* "ttcrpy/rgrid.pyx":1023
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 */
  __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1023, __pyx_L1_error)
  __pyx_t_22 = __pyx_t_21;
  for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
    __pyx_v_n = __pyx_t_23;

    /* "ttcrpy/rgrid.pyx":1024
 * 
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 */
    __pyx_t_28 = (__pyx_v_vtt[__pyx_v_n]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
      __pyx_v_nt = __pyx_t_30;

      /* "ttcrpy/rgrid.pyx":1025
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]             # <<<<<<<<<<<<<<
 * 
 *         if return_rays:
 */
      __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_vtt[__pyx_v_n])[__pyx_v_nt])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tt, __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }

  /* "ttcrpy/rgrid.pyx":1027
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":1028
 * 
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 */
    { /* enter inner scope */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1028, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1028, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1028, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_7genexpr__pyx_v_n = __pyx_t_23;
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1028, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 1028, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    } /* exit inner scope */
    __pyx_v_rays = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":1029
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 */
    __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1029, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
      __pyx_v_n = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":1030
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 */
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1030, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_34 = (__pyx_v_vRx[__pyx_v_n]).size();
        __pyx_t_35 = __pyx_t_34;
        for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_35; __pyx_t_26+=1) {
          __pyx_8genexpr1__pyx_v_i = __pyx_t_26;
          __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_float_0_0);
          __Pyx_GIVEREF(__pyx_float_0_0);
          PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 1030, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":1031
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):             # <<<<<<<<<<<<<<
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):
 */
      __pyx_t_34 = (__pyx_v_vRx[__pyx_v_n]).size();
      __pyx_t_35 = __pyx_t_34;
      for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_35; __pyx_t_31+=1) {
        __pyx_v_n2 = __pyx_t_31;

        /* "ttcrpy/rgrid.pyx":1032
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
        __Pyx_INCREF(__pyx_int_3);
        __Pyx_GIVEREF(__pyx_int_3);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_3);
        __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_v_r, __pyx_v_n2, __pyx_t_5, std::vector<std::vector<ttcr::sxyz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 1032, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "ttcrpy/rgrid.pyx":1033
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 */
        __pyx_t_36 = ((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size();
        __pyx_t_37 = __pyx_t_36;
        for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_37; __pyx_t_26+=1) {
          __pyx_v_nn = __pyx_t_26;

          /* "ttcrpy/rgrid.pyx":1034
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, std::vector<std::vector<ttcr::sxyz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
          __pyx_t_7 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_5) < 0)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":1035
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, std::vector<std::vector<ttcr::sxyz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1035, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_1);
          __pyx_t_3 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_7, __pyx_t_5) < 0)) __PYX_ERR(0, 1035, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":1036
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]
 */
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, std::vector<std::vector<ttcr::sxyz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_2);
          __Pyx_GIVEREF(__pyx_int_2);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_2);
          __pyx_t_2 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }

      /* "ttcrpy/rgrid.pyx":1037
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 */
      __pyx_t_28 = (__pyx_v_vtt[__pyx_v_n]).size();
      __pyx_t_29 = __pyx_t_28;
      for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
        __pyx_v_nt = __pyx_t_30;

        /* "ttcrpy/rgrid.pyx":1038
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]             # <<<<<<<<<<<<<<
 * 
 *         if compute_L:
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_rays, __pyx_t_7, __pyx_t_5) < 0)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }

    /* "ttcrpy/rgrid.pyx":1027
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  }

  /* "ttcrpy/rgrid.pyx":1040
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L:             # <<<<<<<<<<<<<<
 * 
 *             # we build an array of matrices, for each event
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":1043
 * 
 *             # we build an array of matrices, for each event
 *             L = []             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 nnz = 0
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_L = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":1044
 *             # we build an array of matrices, for each event
 *             L = []
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):
 */
    __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
      __pyx_v_n = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":1045
 *             L = []
 *             for n in range(nTx):
 *                 nnz = 0             # <<<<<<<<<<<<<<
 *                 for ni in range(l_data[n].size()):
 *                     nnz += l_data[n][ni].size()
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_nnz, __pyx_int_0);

      /* "ttcrpy/rgrid.pyx":1046
 *             for n in range(nTx):
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):             # <<<<<<<<<<<<<<
 *                     nnz += l_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 */
      __pyx_t_38 = (__pyx_v_l_data[__pyx_v_n]).size();
      __pyx_t_39 = __pyx_t_38;
      for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
        __pyx_v_ni = __pyx_t_40;

        /* "ttcrpy/rgrid.pyx":1047
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):
 *                     nnz += l_data[n][ni].size()             # <<<<<<<<<<<<<<
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 */
        __pyx_t_5 = __Pyx_PyInt_FromSize_t(((__pyx_v_l_data[__pyx_v_n])[__pyx_v_ni]).size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1047, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_nnz, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1047, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_7);
        __pyx_t_7 = 0;
      }

      /* "ttcrpy/rgrid.pyx":1048
 *                 for ni in range(l_data[n].size()):
 *                     nnz += l_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(((__pyx_v_vRx[__pyx_v_n]).size() + 1)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indptr, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "ttcrpy/rgrid.pyx":1049
 *                     nnz += l_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 val = np.empty((nnz,))
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_nnz);
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indices, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1050
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *                 k = 0
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_nnz);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1052
 *                 val = np.empty((nnz,))
 * 
 *                 k = 0             # <<<<<<<<<<<<<<
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_cells()
 */
      __pyx_v_k = 0;

      /* "ttcrpy/rgrid.pyx":1053
 * 
 *                 k = 0
 *                 MM = vRx[n].size()             # <<<<<<<<<<<<<<
 *                 NN = self.get_number_of_cells()
 *                 for i in range(MM):
 */
      __pyx_v_MM = (__pyx_v_vRx[__pyx_v_n]).size();

      /* "ttcrpy/rgrid.pyx":1054
 *                 k = 0
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_cells()             # <<<<<<<<<<<<<<
 *                 for i in range(MM):
 *                     indptr[i] = k
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_26 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_26 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_NN = __pyx_t_26;

      /* "ttcrpy/rgrid.pyx":1055
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_cells()
 *                 for i in range(MM):             # <<<<<<<<<<<<<<
 *                     indptr[i] = k
 *                     for j in range(NN):
 */
      __pyx_t_26 = __pyx_v_MM;
      __pyx_t_41 = __pyx_t_26;
      for (__pyx_t_42 = 0; __pyx_t_42 < __pyx_t_41; __pyx_t_42+=1) {
        __pyx_v_i = __pyx_t_42;

        /* "ttcrpy/rgrid.pyx":1056
 *                 NN = self.get_number_of_cells()
 *                 for i in range(MM):
 *                     indptr[i] = k             # <<<<<<<<<<<<<<
 *                     for j in range(NN):
 *                         for nn in range(l_data[n][i].size()):
 */
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1056, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":1057
 *                 for i in range(MM):
 *                     indptr[i] = k
 *                     for j in range(NN):             # <<<<<<<<<<<<<<
 *                         for nn in range(l_data[n][i].size()):
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:
 */
        __pyx_t_43 = __pyx_v_NN;
        __pyx_t_44 = __pyx_t_43;
        for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_44; __pyx_t_45+=1) {
          __pyx_v_j = __pyx_t_45;

          /* "ttcrpy/rgrid.pyx":1058
 *                     indptr[i] = k
 *                     for j in range(NN):
 *                         for nn in range(l_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:
 *                                 indices[k] = j
 */
          __pyx_t_46 = ((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i]).size();
          __pyx_t_47 = __pyx_t_46;
          for (__pyx_t_48 = 0; __pyx_t_48 < __pyx_t_47; __pyx_t_48+=1) {
            __pyx_v_nn = __pyx_t_48;

            /* "ttcrpy/rgrid.pyx":1059
 *                     for j in range(NN):
 *                         for nn in range(l_data[n][i].size()):
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = l_data[n][i][nn].v
 */
            __pyx_t_4 = ((((struct __pyx_vtabstruct_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self->__pyx_vtab)->_f2c_ind(__pyx_v_self, (((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).i) == __pyx_v_j) != 0);
            if (__pyx_t_4) {

              /* "ttcrpy/rgrid.pyx":1060
 *                         for nn in range(l_data[n][i].size()):
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:
 *                                 indices[k] = j             # <<<<<<<<<<<<<<
 *                                 val[k] = l_data[n][i][nn].v
 *                                 k += 1
 */
              __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (unlikely(__Pyx_SetItemInt(__pyx_v_indices, __pyx_v_k, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1060, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "ttcrpy/rgrid.pyx":1061
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:
 *                                 indices[k] = j
 *                                 val[k] = l_data[n][i][nn].v             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_2 = PyFloat_FromDouble((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).v); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (unlikely(__Pyx_SetItemInt(__pyx_v_val, __pyx_v_k, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1061, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "ttcrpy/rgrid.pyx":1062
 *                                 indices[k] = j
 *                                 val[k] = l_data[n][i][nn].v
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                 indptr[MM] = k
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1059
 *                     for j in range(NN):
 *                         for nn in range(l_data[n][i].size()):
 *                             if self._f2c_ind(l_data[n][i][nn].i) == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = l_data[n][i][nn].v
 */
            }
          }
        }
      }

      /* "ttcrpy/rgrid.pyx":1064
 *                                 k += 1
 * 
 *                 indptr[MM] = k             # <<<<<<<<<<<<<<
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_MM, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1065
 * 
 *                 indptr[MM] = k
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 * 
 *             if evID is None:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_val);
      __Pyx_GIVEREF(__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_val);
      __Pyx_INCREF(__pyx_v_indices);
      __Pyx_GIVEREF(__pyx_v_indices);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_indices);
      __Pyx_INCREF(__pyx_v_indptr);
      __Pyx_GIVEREF(__pyx_v_indptr);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_indptr);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_MM); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_NN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_49 = PyTuple_New(2); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_49, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_49, 1, __pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_49) < 0) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
      __pyx_t_49 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_20 = __Pyx_PyObject_Append(__pyx_v_L, __pyx_t_49); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
    }

    /* "ttcrpy/rgrid.pyx":1067
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *             if evID is None:             # <<<<<<<<<<<<<<
 *                 # we want a single matrix
 *                 tmp = sp.vstack(L)
 */
    __pyx_t_4 = (__pyx_v_evID == Py_None);
    __pyx_t_1 = (__pyx_t_4 != 0);
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":1069
 *             if evID is None:
 *                 # we want a single matrix
 *                 tmp = sp.vstack(L)             # <<<<<<<<<<<<<<
 *                 itmp = []
 *                 for n in range(nTx):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vstack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_49 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_L) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_L);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_49);
      __pyx_t_49 = 0;

      /* "ttcrpy/rgrid.pyx":1070
 *                 # we want a single matrix
 *                 tmp = sp.vstack(L)
 *                 itmp = []             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     for nt in range(vtt[n].size()):
 */
      __pyx_t_49 = PyList_New(0); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __pyx_v_itmp = ((PyObject*)__pyx_t_49);
      __pyx_t_49 = 0;

      /* "ttcrpy/rgrid.pyx":1071
 *                 tmp = sp.vstack(L)
 *                 itmp = []
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     for nt in range(vtt[n].size()):
 *                         itmp.append(iRx[n][nt])
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1071, __pyx_L1_error)
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_n = __pyx_t_23;

        /* "ttcrpy/rgrid.pyx":1072
 *                 itmp = []
 *                 for n in range(nTx):
 *                     for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                         itmp.append(iRx[n][nt])
 *                 L = tmp[itmp,:]
 */
        __pyx_t_28 = (__pyx_v_vtt[__pyx_v_n]).size();
        __pyx_t_29 = __pyx_t_28;
        for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
          __pyx_v_nt = __pyx_t_30;

          /* "ttcrpy/rgrid.pyx":1073
 *                 for n in range(nTx):
 *                     for nt in range(vtt[n].size()):
 *                         itmp.append(iRx[n][nt])             # <<<<<<<<<<<<<<
 *                 L = tmp[itmp,:]
 * 
 */
          __pyx_t_49 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1073, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_49);
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_49, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
          __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_itmp, __pyx_t_6); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1073, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }

      /* "ttcrpy/rgrid.pyx":1074
 *                     for nt in range(vtt[n].size()):
 *                         itmp.append(iRx[n][nt])
 *                 L = tmp[itmp,:]             # <<<<<<<<<<<<<<
 * 
 *         if compute_M:
 */
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_itmp);
      __Pyx_GIVEREF(__pyx_v_itmp);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_itmp);
      __Pyx_INCREF(__pyx_slice__3);
      __Pyx_GIVEREF(__pyx_slice__3);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_slice__3);
      __pyx_t_49 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_t_6); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1074, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_L, __pyx_t_49);
      __pyx_t_49 = 0;

      /* "ttcrpy/rgrid.pyx":1067
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *             if evID is None:             # <<<<<<<<<<<<<<
 *                 # we want a single matrix
 *                 tmp = sp.vstack(L)
 */
    }

    /* "ttcrpy/rgrid.pyx":1040
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L:             # <<<<<<<<<<<<<<
 * 
 *             # we build an array of matrices, for each event
 */
  }

  /* "ttcrpy/rgrid.pyx":1076
 *                 L = tmp[itmp,:]
 * 
 *         if compute_M:             # <<<<<<<<<<<<<<
 *             # we return array of matrices, one for each event
 *             M = []
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1076, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1078
 *         if compute_M:
 *             # we return array of matrices, one for each event
 *             M = []             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 nnz = 0
 */
    __pyx_t_49 = PyList_New(0); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_49);
    __pyx_v_M = ((PyObject*)__pyx_t_49);
    __pyx_t_49 = 0;

    /* "ttcrpy/rgrid.pyx":1079
 *             # we return array of matrices, one for each event
 *             M = []
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):
 */
    __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
      __pyx_v_n = __pyx_t_23;

      /* "ttcrpy/rgrid.pyx":1080
 *             M = []
 *             for n in range(nTx):
 *                 nnz = 0             # <<<<<<<<<<<<<<
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_nnz, __pyx_int_0);

      /* "ttcrpy/rgrid.pyx":1081
 *             for n in range(nTx):
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):             # <<<<<<<<<<<<<<
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 */
      __pyx_t_40 = (__pyx_v_m_data[__pyx_v_n]).size();
      __pyx_t_50 = __pyx_t_40;
      for (__pyx_t_51 = 0; __pyx_t_51 < __pyx_t_50; __pyx_t_51+=1) {
        __pyx_v_ni = __pyx_t_51;

        /* "ttcrpy/rgrid.pyx":1082
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()             # <<<<<<<<<<<<<<
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 */
        __pyx_t_49 = __Pyx_PyInt_FromSize_t(((__pyx_v_m_data[__pyx_v_n])[__pyx_v_ni]).size()); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_49);
        __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_nnz, __pyx_t_49); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
        __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_6);
        __pyx_t_6 = 0;
      }

      /* "ttcrpy/rgrid.pyx":1083
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_49 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_FromSize_t(((__pyx_v_vRx[__pyx_v_n]).size() + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_49, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indptr, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":1084
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 val = np.empty((nnz,))
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nnz);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_49, __pyx_n_s_np); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_49, __pyx_n_s_int64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indices, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "ttcrpy/rgrid.pyx":1085
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *                 k = 0
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nnz);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "ttcrpy/rgrid.pyx":1087
 *                 val = np.empty((nnz,))
 * 
 *                 k = 0             # <<<<<<<<<<<<<<
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()
 */
      __pyx_v_k = 0;

      /* "ttcrpy/rgrid.pyx":1088
 * 
 *                 k = 0
 *                 MM = vRx[n].size()             # <<<<<<<<<<<<<<
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):
 */
      __pyx_v_MM = (__pyx_v_vRx[__pyx_v_n]).size();

      /* "ttcrpy/rgrid.pyx":1089
 *                 k = 0
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()             # <<<<<<<<<<<<<<
 *                 for i in range(MM):
 *                     indptr[i] = k
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1089, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1089, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_26 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_26 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1089, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_NN = __pyx_t_26;

      /* "ttcrpy/rgrid.pyx":1090
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):             # <<<<<<<<<<<<<<
 *                     indptr[i] = k
 *                     for j in range(NN):
 */
      __pyx_t_26 = __pyx_v_MM;
      __pyx_t_41 = __pyx_t_26;
      for (__pyx_t_42 = 0; __pyx_t_42 < __pyx_t_41; __pyx_t_42+=1) {
        __pyx_v_i = __pyx_t_42;

        /* "ttcrpy/rgrid.pyx":1091
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):
 *                     indptr[i] = k             # <<<<<<<<<<<<<<
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1091, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_i, __pyx_t_7, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1091, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "ttcrpy/rgrid.pyx":1092
 *                 for i in range(MM):
 *                     indptr[i] = k
 *                     for j in range(NN):             # <<<<<<<<<<<<<<
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 */
        __pyx_t_43 = __pyx_v_NN;
        __pyx_t_44 = __pyx_t_43;
        for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_44; __pyx_t_45+=1) {
          __pyx_v_j = __pyx_t_45;

          /* "ttcrpy/rgrid.pyx":1093
 *                     indptr[i] = k
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j
 */
          __pyx_t_52 = ((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i]).size();
          __pyx_t_53 = __pyx_t_52;
          for (__pyx_t_48 = 0; __pyx_t_48 < __pyx_t_53; __pyx_t_48+=1) {
            __pyx_v_nn = __pyx_t_48;

            /* "ttcrpy/rgrid.pyx":1094
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 */
            __pyx_t_4 = (((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).i == __pyx_v_i) != 0);
            if (__pyx_t_4) {
            } else {
              __pyx_t_1 = __pyx_t_4;
              goto __pyx_L141_bool_binop_done;
            }
            __pyx_t_4 = (((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).j == __pyx_v_j) != 0);
            __pyx_t_1 = __pyx_t_4;
            __pyx_L141_bool_binop_done:;
            if (__pyx_t_1) {

              /* "ttcrpy/rgrid.pyx":1095
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j             # <<<<<<<<<<<<<<
 *                                 val[k] = m_data[n][i][nn].v
 *                                 k += 1
 */
              __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1095, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(__Pyx_SetItemInt(__pyx_v_indices, __pyx_v_k, __pyx_t_7, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1095, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "ttcrpy/rgrid.pyx":1096
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_7 = PyFloat_FromDouble((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).v); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1096, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(__Pyx_SetItemInt(__pyx_v_val, __pyx_v_k, __pyx_t_7, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1096, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "ttcrpy/rgrid.pyx":1097
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                 indptr[MM] = k
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1094
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 */
            }
          }
        }
      }

      /* "ttcrpy/rgrid.pyx":1099
 *                                 k += 1
 * 
 *                 indptr[MM] = k             # <<<<<<<<<<<<<<
 *                 M.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 */
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_MM, __pyx_t_7, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "ttcrpy/rgrid.pyx":1100
 * 
 *                 indptr[MM] = k
 *                 M.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 * 
 *         if compute_L==False and compute_M==False and return_rays==False:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_sp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_val);
      __Pyx_GIVEREF(__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_val);
      __Pyx_INCREF(__pyx_v_indices);
      __Pyx_GIVEREF(__pyx_v_indices);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_indices);
      __Pyx_INCREF(__pyx_v_indptr);
      __Pyx_GIVEREF(__pyx_v_indptr);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_indptr);
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_MM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_49 = __Pyx_PyInt_From_int(__pyx_v_NN); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_49);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_49);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_49);
      __pyx_t_2 = 0;
      __pyx_t_49 = 0;
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_M, __pyx_t_3); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "ttcrpy/rgrid.pyx":1076
 *                 L = tmp[itmp,:]
 * 
 *         if compute_M:             # <<<<<<<<<<<<<<
 *             # we return array of matrices, one for each event
 *             M = []
 */
  }

  /* "ttcrpy/rgrid.pyx":1102
 *                 M.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *         if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_M and return_rays:
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L144_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_M, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L144_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L144_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1103
 * 
 *         if compute_L==False and compute_M==False and return_rays==False:
 *             return tt             # <<<<<<<<<<<<<<
 *         elif compute_M and return_rays:
 *             return tt, rays, M
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_tt);
    __pyx_r = __pyx_v_tt;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1102
 *                 M.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *         if compute_L==False and compute_M==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_M and return_rays:
 */
  }

  /* "ttcrpy/rgrid.pyx":1104
 *         if compute_L==False and compute_M==False and return_rays==False:
 *             return tt
 *         elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, M
 *         elif compute_L and return_rays:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L147_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L147_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1105
 *             return tt
 *         elif compute_M and return_rays:
 *             return tt, rays, M             # <<<<<<<<<<<<<<
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 1105, __pyx_L1_error) }
    if (unlikely(!__pyx_v_M)) { __Pyx_RaiseUnboundLocalError("M"); __PYX_ERR(0, 1105, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __Pyx_INCREF(__pyx_v_M);
    __Pyx_GIVEREF(__pyx_v_M);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_M);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1104
 *         if compute_L==False and compute_M==False and return_rays==False:
 *             return tt
 *         elif compute_M and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, M
 *         elif compute_L and return_rays:
 */
  }

  /* "ttcrpy/rgrid.pyx":1106
 *         elif compute_M and return_rays:
 *             return tt, rays, M
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L149_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L149_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1107
 *             return tt, rays, M
 *         elif compute_L and return_rays:
 *             return tt, rays, L             # <<<<<<<<<<<<<<
 *         elif compute_L:
 *             return tt, L
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 1107, __pyx_L1_error) }
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 1107, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_L);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1106
 *         elif compute_M and return_rays:
 *             return tt, rays, M
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  }

  /* "ttcrpy/rgrid.pyx":1108
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         elif compute_M:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1108, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1109
 *             return tt, rays, L
 *         elif compute_L:
 *             return tt, L             # <<<<<<<<<<<<<<
 *         elif compute_M:
 *             return tt, M
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 1109, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_L);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1108
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         elif compute_M:
 */
  }

  /* "ttcrpy/rgrid.pyx":1110
 *         elif compute_L:
 *             return tt, L
 *         elif compute_M:             # <<<<<<<<<<<<<<
 *             return tt, M
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_M); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1110, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1111
 *             return tt, L
 *         elif compute_M:
 *             return tt, M             # <<<<<<<<<<<<<<
 *         else:
 *             return tt, rays
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_M)) { __Pyx_RaiseUnboundLocalError("M"); __PYX_ERR(0, 1111, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_M);
    __Pyx_GIVEREF(__pyx_v_M);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_M);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1110
 *         elif compute_L:
 *             return tt, L
 *         elif compute_M:             # <<<<<<<<<<<<<<
 *             return tt, M
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":1113
 *             return tt, M
 *         else:
 *             return tt, rays             # <<<<<<<<<<<<<<
 * 
 *     def to_vtk(self, fields, filename):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 1113, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":751
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, compute_M=False,
 *                  return_rays=False):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_49);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_evID);
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_t0);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_nTx);
  __Pyx_XDECREF(__pyx_v_Tx);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_iRx);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XDECREF(__pyx_v_ii);
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_rays);
  __Pyx_XDECREF(__pyx_v_L);
  __Pyx_XDECREF(__pyx_v_nnz);
  __Pyx_XDECREF(__pyx_v_indptr);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_itmp);
  __Pyx_XDECREF(__pyx_v_M);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1115
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_35to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_34to_vtk[] = "\n        to_vtk(fields, filename)\n\n        Save grid variables and/or raypaths to VTK format\n\n        Parameters\n        ----------\n        fields: dict\n            dict of variables to save to file. Variables should be np.ndarray of\n            size equal to either the number of nodes of the number of cells of\n            the grid, or a list of raypath coordinates.\n        filename: str\n            Name of file without extension for saving (extension vtr will be\n            added).  Raypaths are saved in separate files, and filename will\n            be appended by the dict key and have a vtp extension.\n\n        Notes\n        -----\n        VTK files can be visualized with Paraview (https://www.paraview.org)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_35to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_vtk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fields,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, 1); __PYX_ERR(0, 1115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "to_vtk") < 0)) __PYX_ERR(0, 1115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fields = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_34to_vtk(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_fields, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_34to_vtk(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_xCoords = NULL;
  PyObject *__pyx_v_yCoords = NULL;
  PyObject *__pyx_v_zCoords = NULL;
  PyObject *__pyx_v_rgrid = NULL;
  int __pyx_v_save_grid;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_scalar = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_vtk", 0);

  /* "ttcrpy/rgrid.pyx":1136
 *         VTK files can be visualized with Paraview (https://www.paraview.org)
 *         """
 *         xCoords = numpy_support.numpy_to_vtk(self.x)             # <<<<<<<<<<<<<<
 *         yCoords = numpy_support.numpy_to_vtk(self.y)
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_xCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1137
 *         """
 *         xCoords = numpy_support.numpy_to_vtk(self.x)
 *         yCoords = numpy_support.numpy_to_vtk(self.y)             # <<<<<<<<<<<<<<
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_yCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1138
 *         xCoords = numpy_support.numpy_to_vtk(self.x)
 *         yCoords = numpy_support.numpy_to_vtk(self.y)
 *         zCoords = numpy_support.numpy_to_vtk(self.z)             # <<<<<<<<<<<<<<
 * 
 *         rgrid = vtk.vtkRectilinearGrid()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_z); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_zCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1140
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 * 
 *         rgrid = vtk.vtkRectilinearGrid()             # <<<<<<<<<<<<<<
 *         rgrid.SetDimensions(self._x.size(), self._y.size(), self._z.size())
 *         rgrid.SetXCoordinates(xCoords)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkRectilinearGrid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rgrid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1141
 * 
 *         rgrid = vtk.vtkRectilinearGrid()
 *         rgrid.SetDimensions(self._x.size(), self._y.size(), self._z.size())             # <<<<<<<<<<<<<<
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetDimensions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1142
 *         rgrid = vtk.vtkRectilinearGrid()
 *         rgrid.SetDimensions(self._x.size(), self._y.size(), self._z.size())
 *         rgrid.SetXCoordinates(xCoords)             # <<<<<<<<<<<<<<
 *         rgrid.SetYCoordinates(yCoords)
 *         rgrid.SetZCoordinates(zCoords)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetXCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_v_xCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_xCoords);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1143
 *         rgrid.SetDimensions(self._x.size(), self._y.size(), self._z.size())
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)             # <<<<<<<<<<<<<<
 *         rgrid.SetZCoordinates(zCoords)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetYCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_v_yCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_yCoords);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1144
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)
 *         rgrid.SetZCoordinates(zCoords)             # <<<<<<<<<<<<<<
 * 
 *         save_grid = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetZCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_v_zCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_zCoords);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1146
 *         rgrid.SetZCoordinates(zCoords)
 * 
 *         save_grid = False             # <<<<<<<<<<<<<<
 *         for fn in fields:
 *             data = fields[fn]
 */
  __pyx_v_save_grid = 0;

  /* "ttcrpy/rgrid.pyx":1147
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
    __pyx_t_1 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1147, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1147, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1147, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_10(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1147, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":1148
 *         save_grid = False
 *         for fn in fields:
 *             data = fields[fn]             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(data, list):
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_fields, __pyx_v_fn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":1150
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
    __pyx_t_11 = PyList_Check(__pyx_v_data); 
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {

      /* "ttcrpy/rgrid.pyx":1151
 * 
 *             if isinstance(data, list):
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')             # <<<<<<<<<<<<<<
 *             else:
 *                 save_grid = True
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_save_raypaths); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = PyNumber_Add(__pyx_v_filename, __pyx_n_u__34); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_v_fn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_kp_u_vtp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_data, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_data, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_data);
        __Pyx_GIVEREF(__pyx_v_data);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_data);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1150
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
      goto __pyx_L5;
    }

    /* "ttcrpy/rgrid.pyx":1153
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 *                 save_grid = True             # <<<<<<<<<<<<<<
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 */
    /*else*/ {
      __pyx_v_save_grid = 1;

      /* "ttcrpy/rgrid.pyx":1154
 *             else:
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()             # <<<<<<<<<<<<<<
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_vtk); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_vtkDoubleArray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_scalar, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1155
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetName); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1155, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_fn);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1156
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfComponents); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_int_1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1157
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)             # <<<<<<<<<<<<<<
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfTuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1158
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1158, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1158, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_12) {

        /* "ttcrpy/rgrid.pyx":1159
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._y.size(), self._z.size())             # <<<<<<<<<<<<<<
 *                     if data.ndim == 3:
 *                         if data.shape != shape:
 */
        __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_y.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
        __pyx_t_8 = 0;
        __pyx_t_3 = 0;
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/rgrid.pyx":1160
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_12) {

          /* "ttcrpy/rgrid.pyx":1161
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1161, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_v_shape, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1161, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1161, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(__pyx_t_12)) {

            /* "ttcrpy/rgrid.pyx":1162
 *                     if data.ndim == 3:
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))             # <<<<<<<<<<<<<<
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_shape, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1162, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_fn);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1162, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1162, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 1162, __pyx_L1_error)

            /* "ttcrpy/rgrid.pyx":1161
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          }

          /* "ttcrpy/rgrid.pyx":1164
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')             # <<<<<<<<<<<<<<
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1164, __pyx_L1_error)
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "ttcrpy/rgrid.pyx":1160
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
          goto __pyx_L7;
        }

        /* "ttcrpy/rgrid.pyx":1165
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L9_bool_binop_done:;
        if (likely(__pyx_t_12)) {

          /* "ttcrpy/rgrid.pyx":1167
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_shape);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1167, __pyx_L1_error)
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":1165
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
          goto __pyx_L7;
        }

        /* "ttcrpy/rgrid.pyx":1169
 *                         tmp = data.reshape(shape).flatten(order='F')
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_ndim_1_d, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_fn, __pyx_t_3};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1169, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_fn, __pyx_t_3};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1169, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1169, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fn);
            __Pyx_GIVEREF(__pyx_v_fn);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_v_fn);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1169, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 1169, __pyx_L1_error)
        }
        __pyx_L7:;

        /* "ttcrpy/rgrid.pyx":1170
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_2 = __pyx_t_5; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1170, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_2))) {
              if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_5); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1170, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1170, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_5); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1170, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1170, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_14(__pyx_t_2);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1170, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":1171
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])             # <<<<<<<<<<<<<<
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1171, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1171, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_n, __pyx_t_3};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1171, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_n, __pyx_t_3};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1171, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1171, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_n);
            __Pyx_GIVEREF(__pyx_v_n);
            PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_n);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1171, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/rgrid.pyx":1170
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 */
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":1172
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":1158
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size(), self._y.size(), self._z.size())
 *                     if data.ndim == 3:
 */
        goto __pyx_L6;
      }

      /* "ttcrpy/rgrid.pyx":1173
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(__pyx_t_12)) {

        /* "ttcrpy/rgrid.pyx":1174
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)             # <<<<<<<<<<<<<<
 *                     if data.ndim == 3:
 *                         if data.shape != shape:
 */
        __pyx_t_5 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_y.size() - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
        __pyx_t_5 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":1175
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_12) {

          /* "ttcrpy/rgrid.pyx":1176
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_v_shape, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1176, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1176, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(__pyx_t_12)) {

            /* "ttcrpy/rgrid.pyx":1177
 *                     if data.ndim == 3:
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))             # <<<<<<<<<<<<<<
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_shape, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_fn);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1177, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 1177, __pyx_L1_error)

            /* "ttcrpy/rgrid.pyx":1176
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          }

          /* "ttcrpy/rgrid.pyx":1179
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')             # <<<<<<<<<<<<<<
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1179, __pyx_L1_error)
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1179, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "ttcrpy/rgrid.pyx":1175
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
          goto __pyx_L14;
        }

        /* "ttcrpy/rgrid.pyx":1180
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L16_bool_binop_done:;
        if (likely(__pyx_t_12)) {

          /* "ttcrpy/rgrid.pyx":1182
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1182, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_shape);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1182, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1182, __pyx_L1_error)
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1182, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "ttcrpy/rgrid.pyx":1180
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
          goto __pyx_L14;
        }

        /* "ttcrpy/rgrid.pyx":1184
 *                         tmp = data.reshape(shape).flatten(order='F')
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_ndim_1_d, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_fn, __pyx_t_4};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_fn, __pyx_t_4};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fn);
            __Pyx_GIVEREF(__pyx_v_fn);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_fn);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 1184, __pyx_L1_error)
        }
        __pyx_L14:;

        /* "ttcrpy/rgrid.pyx":1185
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
          __pyx_t_2 = __pyx_t_6; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1185, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_2))) {
              if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_6); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1185, __pyx_L1_error)
              #else
              __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1185, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              #endif
            } else {
              if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_6); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1185, __pyx_L1_error)
              #else
              __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1185, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              #endif
            }
          } else {
            __pyx_t_6 = __pyx_t_14(__pyx_t_2);
            if (unlikely(!__pyx_t_6)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1185, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "ttcrpy/rgrid.pyx":1186
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])             # <<<<<<<<<<<<<<
 *                     rgrid.GetCellData().AddArray(scalar)
 *                 else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1186, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_v_n); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1186, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_n, __pyx_t_4};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_n, __pyx_t_4};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1186, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_n);
            __Pyx_GIVEREF(__pyx_v_n);
            PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_n);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "ttcrpy/rgrid.pyx":1185
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim ({1:d})'.format(fn, data.ndim))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)
 */
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":1187
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":1173
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size()-1, self._y.size()-1, self._z.size()-1)
 *                     if data.ndim == 3:
 */
        goto __pyx_L6;
      }

      /* "ttcrpy/rgrid.pyx":1189
 *                     rgrid.GetCellData().AddArray(scalar)
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))             # <<<<<<<<<<<<<<
 * 
 *         if save_grid:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_fn);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 1189, __pyx_L1_error)
      }
      __pyx_L6:;
    }
    __pyx_L5:;

    /* "ttcrpy/rgrid.pyx":1147
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1191
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 */
  __pyx_t_12 = (__pyx_v_save_grid != 0);
  if (__pyx_t_12) {

    /* "ttcrpy/rgrid.pyx":1192
 * 
 *         if save_grid:
 *             writer = vtk.vtkXMLRectilinearGridWriter()             # <<<<<<<<<<<<<<
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkXMLRectilinearGridWriter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_writer = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1193
 *         if save_grid:
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')             # <<<<<<<<<<<<<<
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Add(__pyx_v_filename, __pyx_kp_u_vtr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1194
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)             # <<<<<<<<<<<<<<
 *             writer.SetDataModeToBinary()
 *             writer.Update()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_rgrid) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_rgrid);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1195
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *             writer.Update()
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1196
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()
 *             writer.Update()             # <<<<<<<<<<<<<<
 * 
 *     def _save_raypaths(self, rays, filename):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1191
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 */
  }

  /* "ttcrpy/rgrid.pyx":1115
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xCoords);
  __Pyx_XDECREF(__pyx_v_yCoords);
  __Pyx_XDECREF(__pyx_v_zCoords);
  __Pyx_XDECREF(__pyx_v_rgrid);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_scalar);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1198
 *             writer.Update()
 * 
 *     def _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_37_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_37_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rays = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_save_raypaths (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rays,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rays)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, 1); __PYX_ERR(0, 1198, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_save_raypaths") < 0)) __PYX_ERR(0, 1198, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rays = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1198, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_36_save_raypaths(((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_v_self), __pyx_v_rays, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_36_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_polydata = NULL;
  PyObject *__pyx_v_cellarray = NULL;
  PyObject *__pyx_v_pts = NULL;
  PyObject *__pyx_v_npts = NULL;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_save_raypaths", 0);

  /* "ttcrpy/rgrid.pyx":1199
 * 
 *     def _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()             # <<<<<<<<<<<<<<
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPolyData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_polydata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1200
 *     def _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()             # <<<<<<<<<<<<<<
 *         pts = vtk.vtkPoints()
 *         npts = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkCellArray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cellarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1201
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1202
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_npts = __pyx_int_0;

  /* "ttcrpy/rgrid.pyx":1203
 *         pts = vtk.vtkPoints()
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1203, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":1204
 *         npts = 0
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]             # <<<<<<<<<<<<<<
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_npts, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1205
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_npts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_npts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1206
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/rgrid.pyx":1207
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1207, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":1208
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1208, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1208, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1208, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1208, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1209
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])             # <<<<<<<<<<<<<<
 *                 npts += 1
 *         polydata.SetPoints(pts)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_0);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_1);
      __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_2);
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_v_npts, __pyx_t_11, __pyx_t_12, __pyx_t_13};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 4+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_v_npts, __pyx_t_11, __pyx_t_12, __pyx_t_13};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 4+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_14, __pyx_v_npts);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_14, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_14, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_14, __pyx_t_13);
        __pyx_t_11 = 0;
        __pyx_t_12 = 0;
        __pyx_t_13 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1210
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1             # <<<<<<<<<<<<<<
 *         polydata.SetPoints(pts)
 *         npts = 0
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1208
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1211
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 *         polydata.SetPoints(pts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_pts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_pts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1212
 *                 npts += 1
 *         polydata.SetPoints(pts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/rgrid.pyx":1213
 *         polydata.SetPoints(pts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":1214
 *         npts = 0
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()             # <<<<<<<<<<<<<<
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkPolyLine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":1215
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetNumberOfIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shape); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_9, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":1216
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1216, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1216, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1216, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1216, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1217
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)             # <<<<<<<<<<<<<<
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetId); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_p);
        __Pyx_GIVEREF(__pyx_v_p);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_14, __pyx_v_p);
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_14, __pyx_v_npts);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1218
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1             # <<<<<<<<<<<<<<
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":1216
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":1219
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 *             cellarray.InsertNextCell(line)             # <<<<<<<<<<<<<<
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cellarray, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_line);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1220
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)             # <<<<<<<<<<<<<<
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetLines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_cellarray) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1221
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()             # <<<<<<<<<<<<<<
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkXMLPolyDataWriter); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_writer = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1222
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1223
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)             # <<<<<<<<<<<<<<
 *         writer.SetDataModeToBinary()
 *         writer.Update()
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_polydata) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_polydata);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1224
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *         writer.Update()
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1225
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 *         writer.Update()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":1198
 *             writer.Update()
 * 
 *     def _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_polydata);
  __Pyx_XDECREF(__pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_v_pts);
  __Pyx_XDECREF(__pyx_v_npts);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1229
 * 
 *     @staticmethod
 *     def builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0,             # <<<<<<<<<<<<<<
 *                 eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5,
 *                 n_secondary=2, n_tertiary=2, radius_tertiary=1.0):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_39builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_38builder[] = "\n        builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0, eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5, n_secondary=2, n_tertiary=2, radius_tertiary=1.0)\n\n        Build instance of Grid3d from VTK file\n\n        Parameters\n        ----------\n        filename : str\n            Name of file holding a vtkRectilinearGrid.\n            The grid must have point or cell attribute named either\n            'Slowness', 'slowness', 'Velocity', 'velocity', or\n            'P-wave velocity'\n\n        Other parameters are defined in Constructor\n\n        Returns\n        -------\n        grid: :obj:`Grid3d`\n            grid instance\n        ";
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_6Grid3d_39builder = {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_39builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_38builder};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_39builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_n_threads = 0;
  PyObject *__pyx_v_method = 0;
  PyObject *__pyx_v_tt_from_rp = 0;
  PyObject *__pyx_v_interp_vel = 0;
  PyObject *__pyx_v_eps = 0;
  PyObject *__pyx_v_maxit = 0;
  PyObject *__pyx_v_weno = 0;
  PyObject *__pyx_v_nsnx = 0;
  PyObject *__pyx_v_nsny = 0;
  PyObject *__pyx_v_nsnz = 0;
  PyObject *__pyx_v_n_secondary = 0;
  PyObject *__pyx_v_n_tertiary = 0;
  PyObject *__pyx_v_radius_tertiary = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("builder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_n_threads,&__pyx_n_s_method,&__pyx_n_s_tt_from_rp,&__pyx_n_s_interp_vel,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_weno,&__pyx_n_s_nsnx,&__pyx_n_s_nsny,&__pyx_n_s_nsnz,&__pyx_n_s_n_secondary,&__pyx_n_s_n_tertiary,&__pyx_n_s_radius_tertiary,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)__pyx_int_1);
    values[2] = ((PyObject *)__pyx_n_u_FSM);
    values[3] = ((PyObject *)__pyx_int_1);
    values[4] = ((PyObject *)__pyx_int_0);
    values[5] = ((PyObject *)__pyx_float_1_eneg_15);
    values[6] = ((PyObject *)__pyx_int_20);
    values[7] = ((PyObject *)__pyx_int_1);
    values[8] = ((PyObject *)__pyx_int_5);
    values[9] = ((PyObject *)__pyx_int_5);
    values[10] = ((PyObject *)__pyx_int_5);
    values[11] = ((PyObject *)__pyx_int_2);
    values[12] = ((PyObject *)__pyx_int_2);
    values[13] = ((PyObject *)__pyx_float_1_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt_from_rp);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interp_vel);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_weno);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnx);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsny);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnz);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_tertiary);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius_tertiary);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "builder") < 0)) __PYX_ERR(0, 1229, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    __pyx_v_n_threads = values[1];
    __pyx_v_method = values[2];
    __pyx_v_tt_from_rp = values[3];
    __pyx_v_interp_vel = values[4];
    __pyx_v_eps = values[5];
    __pyx_v_maxit = values[6];
    __pyx_v_weno = values[7];
    __pyx_v_nsnx = values[8];
    __pyx_v_nsny = values[9];
    __pyx_v_nsnz = values[10];
    __pyx_v_n_secondary = values[11];
    __pyx_v_n_tertiary = values[12];
    __pyx_v_radius_tertiary = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("builder", 0, 1, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1229, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_38builder(__pyx_v_filename, __pyx_v_n_threads, __pyx_v_method, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_nsnx, __pyx_v_nsny, __pyx_v_nsnz, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_radius_tertiary);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_38builder(PyObject *__pyx_v_filename, PyObject *__pyx_v_n_threads, PyObject *__pyx_v_method, PyObject *__pyx_v_tt_from_rp, PyObject *__pyx_v_interp_vel, PyObject *__pyx_v_eps, PyObject *__pyx_v_maxit, PyObject *__pyx_v_weno, PyObject *__pyx_v_nsnx, PyObject *__pyx_v_nsny, PyObject *__pyx_v_nsnz, PyObject *__pyx_v_n_secondary, PyObject *__pyx_v_n_tertiary, PyObject *__pyx_v_radius_tertiary) {
  PyObject *__pyx_v_reader = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_z = NULL;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_name = NULL;
  long __pyx_v_cell_slowness;
  PyObject *__pyx_v_dim = NULL;
  PyObject *__pyx_v_slowness = NULL;
  struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_g = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("builder", 0);

  /* "ttcrpy/rgrid.pyx":1253
 *         """
 * 
 *         reader = vtk.vtkXMLRectilinearGridReader()             # <<<<<<<<<<<<<<
 *         reader.SetFileName(filename)
 *         reader.Update()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkXMLRectilinearGridReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_reader = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1254
 * 
 *         reader = vtk.vtkXMLRectilinearGridReader()
 *         reader.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         reader.Update()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1255
 *         reader = vtk.vtkXMLRectilinearGridReader()
 *         reader.SetFileName(filename)
 *         reader.Update()             # <<<<<<<<<<<<<<
 * 
 *         data = reader.GetOutput()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_Update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1257
 *         reader.Update()
 * 
 *         data = reader.GetOutput()             # <<<<<<<<<<<<<<
 *         x = numpy_support.vtk_to_numpy(data.GetXCoordinates())
 *         y = numpy_support.vtk_to_numpy(data.GetYCoordinates())
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_GetOutput); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1258
 * 
 *         data = reader.GetOutput()
 *         x = numpy_support.vtk_to_numpy(data.GetXCoordinates())             # <<<<<<<<<<<<<<
 *         y = numpy_support.vtk_to_numpy(data.GetYCoordinates())
 *         z = numpy_support.vtk_to_numpy(data.GetZCoordinates())
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetXCoordinates); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1259
 *         data = reader.GetOutput()
 *         x = numpy_support.vtk_to_numpy(data.GetXCoordinates())
 *         y = numpy_support.vtk_to_numpy(data.GetYCoordinates())             # <<<<<<<<<<<<<<
 *         z = numpy_support.vtk_to_numpy(data.GetZCoordinates())
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetYCoordinates); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_y = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1260
 *         x = numpy_support.vtk_to_numpy(data.GetXCoordinates())
 *         y = numpy_support.vtk_to_numpy(data.GetYCoordinates())
 *         z = numpy_support.vtk_to_numpy(data.GetZCoordinates())             # <<<<<<<<<<<<<<
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetZCoordinates); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_z = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1262
 *         z = numpy_support.vtk_to_numpy(data.GetZCoordinates())
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',             # <<<<<<<<<<<<<<
 *                  'P-wave velocity')
 *         for name in names:
 */
  __Pyx_INCREF(__pyx_tuple__35);
  __pyx_v_names = __pyx_tuple__35;

  /* "ttcrpy/rgrid.pyx":1264
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_t_1 = __pyx_v_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
  for (;;) {
    if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1264, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":1265
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {

      /* "ttcrpy/rgrid.pyx":1266
 *         for name in names:
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 dim = (x.size, y.size, z.size)
 */
      __pyx_v_cell_slowness = 0;

      /* "ttcrpy/rgrid.pyx":1267
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 dim = (x.size, y.size, z.size)
 *                 break
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":1268
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 dim = (x.size, y.size, z.size)             # <<<<<<<<<<<<<<
 *                 break
 *             if data.GetCellData().HasArray(name):
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_z, __pyx_n_s_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_4);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_v_dim = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;

      /* "ttcrpy/rgrid.pyx":1269
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 dim = (x.size, y.size, z.size)
 *                 break             # <<<<<<<<<<<<<<
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 */
      goto __pyx_L4_break;

      /* "ttcrpy/rgrid.pyx":1265
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    }

    /* "ttcrpy/rgrid.pyx":1270
 *                 dim = (x.size, y.size, z.size)
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_7) {

      /* "ttcrpy/rgrid.pyx":1271
 *                 break
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 dim = (x.size-1, y.size-1, z.size-1)
 */
      __pyx_v_cell_slowness = 1;

      /* "ttcrpy/rgrid.pyx":1272
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 dim = (x.size-1, y.size-1, z.size-1)
 *                 break
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "ttcrpy/rgrid.pyx":1273
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 dim = (x.size-1, y.size-1, z.size-1)             # <<<<<<<<<<<<<<
 *                 break
 *         else:
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_z, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_5);
      __pyx_t_4 = 0;
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      __pyx_v_dim = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;

      /* "ttcrpy/rgrid.pyx":1274
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 dim = (x.size-1, y.size-1, z.size-1)
 *                 break             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')
 */
      goto __pyx_L4_break;

      /* "ttcrpy/rgrid.pyx":1270
 *                 dim = (x.size, y.size, z.size)
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    }

    /* "ttcrpy/rgrid.pyx":1264
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  }
  /*else*/ {

    /* "ttcrpy/rgrid.pyx":1276
 *                 break
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')             # <<<<<<<<<<<<<<
 * 
 *         if 'lowness' in name:
 */
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 1276, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":1264
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_L4_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1278
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data.reshape(dim, order='F').flatten()
 *         else:
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1278, __pyx_L1_error)
  }
  __pyx_t_7 = (__Pyx_PyUnicode_ContainsTF(__pyx_n_u_lowness, __pyx_v_name, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __pyx_t_10 = (__pyx_t_7 != 0);
  if (__pyx_t_10) {

    /* "ttcrpy/rgrid.pyx":1279
 * 
 *         if 'lowness' in name:
 *             slowness = data.reshape(dim, order='F').flatten()             # <<<<<<<<<<<<<<
 *         else:
 *             slowness = 1.0 / data.reshape(dim, order='F').flatten()
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_dim);
    __Pyx_GIVEREF(__pyx_v_dim);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_dim);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_slowness = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1278
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data.reshape(dim, order='F').flatten()
 *         else:
 */
    goto __pyx_L8;
  }

  /* "ttcrpy/rgrid.pyx":1281
 *             slowness = data.reshape(dim, order='F').flatten()
 *         else:
 *             slowness = 1.0 / data.reshape(dim, order='F').flatten()             # <<<<<<<<<<<<<<
 * 
 *         g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_dim);
    __Pyx_GIVEREF(__pyx_v_dim);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_dim);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 1281, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_flatten); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_0, __pyx_t_1, 1.0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_slowness = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_L8:;

  /* "ttcrpy/rgrid.pyx":1283
 *             slowness = 1.0 / data.reshape(dim, order='F').flatten()
 * 
 *         g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,             # <<<<<<<<<<<<<<
 *                    interp_vel, eps, maxit, weno, nsnx, nsny, nsnz,
 *                    n_secondary, n_tertiary, radius_tertiary)
 */
  __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_cell_slowness); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "ttcrpy/rgrid.pyx":1285
 *         g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,
 *                    interp_vel, eps, maxit, weno, nsnx, nsny, nsnz,
 *                    n_secondary, n_tertiary, radius_tertiary)             # <<<<<<<<<<<<<<
 *         g.set_slowness(slowness)
 *         return g
 */
  __pyx_t_1 = PyTuple_New(17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_x);
  __Pyx_INCREF(__pyx_v_y);
  __Pyx_GIVEREF(__pyx_v_y);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_y);
  __Pyx_INCREF(__pyx_v_z);
  __Pyx_GIVEREF(__pyx_v_z);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_z);
  __Pyx_INCREF(__pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_v_n_threads);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_tt_from_rp);
  __Pyx_GIVEREF(__pyx_v_tt_from_rp);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_tt_from_rp);
  __Pyx_INCREF(__pyx_v_interp_vel);
  __Pyx_GIVEREF(__pyx_v_interp_vel);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_interp_vel);
  __Pyx_INCREF(__pyx_v_eps);
  __Pyx_GIVEREF(__pyx_v_eps);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_v_eps);
  __Pyx_INCREF(__pyx_v_maxit);
  __Pyx_GIVEREF(__pyx_v_maxit);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_v_maxit);
  __Pyx_INCREF(__pyx_v_weno);
  __Pyx_GIVEREF(__pyx_v_weno);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_v_weno);
  __Pyx_INCREF(__pyx_v_nsnx);
  __Pyx_GIVEREF(__pyx_v_nsnx);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_v_nsnx);
  __Pyx_INCREF(__pyx_v_nsny);
  __Pyx_GIVEREF(__pyx_v_nsny);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_v_nsny);
  __Pyx_INCREF(__pyx_v_nsnz);
  __Pyx_GIVEREF(__pyx_v_nsnz);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_v_nsnz);
  __Pyx_INCREF(__pyx_v_n_secondary);
  __Pyx_GIVEREF(__pyx_v_n_secondary);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_v_n_secondary);
  __Pyx_INCREF(__pyx_v_n_tertiary);
  __Pyx_GIVEREF(__pyx_v_n_tertiary);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_v_n_tertiary);
  __Pyx_INCREF(__pyx_v_radius_tertiary);
  __Pyx_GIVEREF(__pyx_v_radius_tertiary);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_v_radius_tertiary);
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":1283
 *             slowness = 1.0 / data.reshape(dim, order='F').flatten()
 * 
 *         g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,             # <<<<<<<<<<<<<<
 *                    interp_vel, eps, maxit, weno, nsnx, nsny, nsnz,
 *                    n_secondary, n_tertiary, radius_tertiary)
 */
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_g = ((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":1286
 *                    interp_vel, eps, maxit, weno, nsnx, nsny, nsnz,
 *                    n_secondary, n_tertiary, radius_tertiary)
 *         g.set_slowness(slowness)             # <<<<<<<<<<<<<<
 *         return g
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_g), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_slowness);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":1287
 *                    n_secondary, n_tertiary, radius_tertiary)
 *         g.set_slowness(slowness)
 *         return g             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_g));
  __pyx_r = ((PyObject *)__pyx_v_g);
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1229
 * 
 *     @staticmethod
 *     def builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0,             # <<<<<<<<<<<<<<
 *                 eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5,
 *                 n_secondary=2, n_tertiary=2, radius_tertiary=1.0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_reader);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_z);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_dim);
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1290
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays[] = "\n        data_kernel_straight_rays(Tx, Rx, grx, gry, grz, centers) -> L, (xc, yc, zc)\n\n        Raytracing with straight rays in 3D\n\n        Parameters\n        ----------\n        Tx : np.ndarray\n            source coordinates, nTx by 3\n                - 1st column contains X coordinates,\n                - 2nd contains Y coordinates\n                - 3rd contains Z coordinates\n        Rx : np.ndarray\n            receiver coordinates, nTx by 3\n                - 1st column contains X coordinates,\n                - 2nd contains Y coordinates\n                - 3rd contains Z coordinates\n        grx : np.ndarray\n            grid node coordinates along x\n        gry : np.ndarray\n            grid node coordinates along y\n        grz : np.ndarray\n            grid node coordinates along z\n        centers : bool\n            return coordinates of center of cells (False by default)\n\n        Returns\n        -------\n        L : scipy csr_matrix\n            data kernel matrix (tt = L*slowness)\n        (xc, yc, zc) : :obj:`tuple` of `np.ndarray`\n            vectors of coordinates of center of cells\n\n        Note\n        ----\n        Tx and Rx should contain the same number of rows, each row corresponding\n        to a source-receiver pair\n\n        ";
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays = {"data_kernel_straight_rays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_Tx = 0;
  PyArrayObject *__pyx_v_Rx = 0;
  PyArrayObject *__pyx_v_grx = 0;
  PyArrayObject *__pyx_v_gry = 0;
  PyArrayObject *__pyx_v_grz = 0;
  PyObject *__pyx_v_centers = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("data_kernel_straight_rays (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_Tx,&__pyx_n_s_Rx,&__pyx_n_s_grx,&__pyx_n_s_gry,&__pyx_n_s_grz,&__pyx_n_s_centers,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "ttcrpy/rgrid.pyx":1295
 *                                   np.ndarray[np.double_t, ndim=1] gry,
 *                                   np.ndarray[np.double_t, ndim=1] grz,
 *                                   centers=False):             # <<<<<<<<<<<<<<
 *         """
 *         data_kernel_straight_rays(Tx, Rx, grx, gry, grz, centers) -> L, (xc, yc, zc)
 */
    values[5] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Tx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Rx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 5, 6, 1); __PYX_ERR(0, 1290, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_grx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 5, 6, 2); __PYX_ERR(0, 1290, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gry)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 5, 6, 3); __PYX_ERR(0, 1290, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_grz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 5, 6, 4); __PYX_ERR(0, 1290, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_centers);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "data_kernel_straight_rays") < 0)) __PYX_ERR(0, 1290, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_Tx = ((PyArrayObject *)values[0]);
    __pyx_v_Rx = ((PyArrayObject *)values[1]);
    __pyx_v_grx = ((PyArrayObject *)values[2]);
    __pyx_v_gry = ((PyArrayObject *)values[3]);
    __pyx_v_grz = ((PyArrayObject *)values[4]);
    __pyx_v_centers = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 5, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1290, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.data_kernel_straight_rays", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_Tx), __pyx_ptype_5numpy_ndarray, 1, "Tx", 0))) __PYX_ERR(0, 1290, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_Rx), __pyx_ptype_5numpy_ndarray, 1, "Rx", 0))) __PYX_ERR(0, 1291, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_grx), __pyx_ptype_5numpy_ndarray, 1, "grx", 0))) __PYX_ERR(0, 1292, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gry), __pyx_ptype_5numpy_ndarray, 1, "gry", 0))) __PYX_ERR(0, 1293, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_grz), __pyx_ptype_5numpy_ndarray, 1, "grz", 0))) __PYX_ERR(0, 1294, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays(__pyx_v_Tx, __pyx_v_Rx, __pyx_v_grx, __pyx_v_gry, __pyx_v_grz, __pyx_v_centers);

  /* "ttcrpy/rgrid.pyx":1290
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays(PyArrayObject *__pyx_v_Tx, PyArrayObject *__pyx_v_Rx, PyArrayObject *__pyx_v_grx, PyArrayObject *__pyx_v_gry, PyArrayObject *__pyx_v_grz, PyObject *__pyx_v_centers) {
  size_t __pyx_v_nTx;
  size_t __pyx_v_n_grx;
  size_t __pyx_v_n_gry;
  size_t __pyx_v_n_grz;
  double __pyx_v_small;
  PyObject *__pyx_v_data_p = NULL;
  PyObject *__pyx_v_indices_p = NULL;
  PyObject *__pyx_v_indptr_p = NULL;
  size_t __pyx_v_k;
  size_t __pyx_v_ix;
  size_t __pyx_v_iy;
  size_t __pyx_v_iz;
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_z;
  double __pyx_v_x1;
  double __pyx_v_y1;
  double __pyx_v_z1;
  double __pyx_v_x2;
  double __pyx_v_y2;
  double __pyx_v_z2;
  double __pyx_v_dtmp;
  double __pyx_v_d;
  double __pyx_v_l;
  double __pyx_v_m;
  double __pyx_v_n;
  double __pyx_v_m_y;
  double __pyx_v_b_y;
  double __pyx_v_m_z;
  double __pyx_v_b_z;
  double __pyx_v_dlx;
  double __pyx_v_dly;
  double __pyx_v_dlz;
  double __pyx_v_dl;
  double __pyx_v_xe;
  double __pyx_v_ye;
  double __pyx_v_ze;
  int __pyx_v_sy;
  int __pyx_v_sz;
  bool __pyx_v_up_y;
  bool __pyx_v_up_z;
  Py_ssize_t __pyx_v_nt;
  size_t __pyx_v_iCell;
  PyObject *__pyx_v_zi = NULL;
  PyObject *__pyx_v_yi = NULL;
  PyObject *__pyx_v_L = NULL;
  PyObject *__pyx_v_xc = NULL;
  PyObject *__pyx_v_yc = NULL;
  PyObject *__pyx_v_zc = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_Rx;
  __Pyx_Buffer __pyx_pybuffer_Rx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_Tx;
  __Pyx_Buffer __pyx_pybuffer_Tx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_grx;
  __Pyx_Buffer __pyx_pybuffer_grx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gry;
  __Pyx_Buffer __pyx_pybuffer_gry;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_grz;
  __Pyx_Buffer __pyx_pybuffer_grz;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  double __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  int __pyx_t_15;
  __pyx_t_5numpy_double_t __pyx_t_16;
  double __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_kernel_straight_rays", 0);
  __pyx_pybuffer_Tx.pybuffer.buf = NULL;
  __pyx_pybuffer_Tx.refcount = 0;
  __pyx_pybuffernd_Tx.data = NULL;
  __pyx_pybuffernd_Tx.rcbuffer = &__pyx_pybuffer_Tx;
  __pyx_pybuffer_Rx.pybuffer.buf = NULL;
  __pyx_pybuffer_Rx.refcount = 0;
  __pyx_pybuffernd_Rx.data = NULL;
  __pyx_pybuffernd_Rx.rcbuffer = &__pyx_pybuffer_Rx;
  __pyx_pybuffer_grx.pybuffer.buf = NULL;
  __pyx_pybuffer_grx.refcount = 0;
  __pyx_pybuffernd_grx.data = NULL;
  __pyx_pybuffernd_grx.rcbuffer = &__pyx_pybuffer_grx;
  __pyx_pybuffer_gry.pybuffer.buf = NULL;
  __pyx_pybuffer_gry.refcount = 0;
  __pyx_pybuffernd_gry.data = NULL;
  __pyx_pybuffernd_gry.rcbuffer = &__pyx_pybuffer_gry;
  __pyx_pybuffer_grz.pybuffer.buf = NULL;
  __pyx_pybuffer_grz.refcount = 0;
  __pyx_pybuffernd_grz.data = NULL;
  __pyx_pybuffernd_grz.rcbuffer = &__pyx_pybuffer_grz;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer, (PyObject*)__pyx_v_Tx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  }
  __pyx_pybuffernd_Tx.diminfo[0].strides = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_Tx.diminfo[0].shape = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_Tx.diminfo[1].strides = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_Tx.diminfo[1].shape = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer, (PyObject*)__pyx_v_Rx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  }
  __pyx_pybuffernd_Rx.diminfo[0].strides = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_Rx.diminfo[0].shape = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_Rx.diminfo[1].strides = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_Rx.diminfo[1].shape = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_grx.rcbuffer->pybuffer, (PyObject*)__pyx_v_grx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  }
  __pyx_pybuffernd_grx.diminfo[0].strides = __pyx_pybuffernd_grx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_grx.diminfo[0].shape = __pyx_pybuffernd_grx.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gry.rcbuffer->pybuffer, (PyObject*)__pyx_v_gry, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  }
  __pyx_pybuffernd_gry.diminfo[0].strides = __pyx_pybuffernd_gry.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gry.diminfo[0].shape = __pyx_pybuffernd_gry.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_grz.rcbuffer->pybuffer, (PyObject*)__pyx_v_grz, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  }
  __pyx_pybuffernd_grz.diminfo[0].strides = __pyx_pybuffernd_grz.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_grz.diminfo[0].shape = __pyx_pybuffernd_grz.rcbuffer->pybuffer.shape[0];

  /* "ttcrpy/rgrid.pyx":1336
 *         """
 * 
 *         cdef size_t nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_gry = gry.shape[0]
 */
  __pyx_v_nTx = (__pyx_v_Tx->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":1337
 * 
 *         cdef size_t nTx = Tx.shape[0]
 *         cdef size_t n_grx = grx.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t n_gry = gry.shape[0]
 *         cdef size_t n_grz = grz.shape[0]
 */
  __pyx_v_n_grx = (__pyx_v_grx->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":1338
 *         cdef size_t nTx = Tx.shape[0]
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_gry = gry.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t n_grz = grz.shape[0]
 * 
 */
  __pyx_v_n_gry = (__pyx_v_gry->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":1339
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_gry = gry.shape[0]
 *         cdef size_t n_grz = grz.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         cdef double small = 1.e-10
 */
  __pyx_v_n_grz = (__pyx_v_grz->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":1341
 *         cdef size_t n_grz = grz.shape[0]
 * 
 *         cdef double small = 1.e-10             # <<<<<<<<<<<<<<
 * 
 *         data_p = []
 */
  __pyx_v_small = 1.e-10;

  /* "ttcrpy/rgrid.pyx":1343
 *         cdef double small = 1.e-10
 * 
 *         data_p = []             # <<<<<<<<<<<<<<
 *         indices_p = []
 *         indptr_p = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1344
 * 
 *         data_p = []
 *         indices_p = []             # <<<<<<<<<<<<<<
 *         indptr_p = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_indices_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1345
 *         data_p = []
 *         indices_p = []
 *         indptr_p = []             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t k = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_indptr_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1347
 *         indptr_p = []
 * 
 *         cdef size_t k = 0             # <<<<<<<<<<<<<<
 *         cdef size_t ix = 0
 *         cdef size_t iy = 0
 */
  __pyx_v_k = 0;

  /* "ttcrpy/rgrid.pyx":1348
 * 
 *         cdef size_t k = 0
 *         cdef size_t ix = 0             # <<<<<<<<<<<<<<
 *         cdef size_t iy = 0
 *         cdef size_t iz = 0
 */
  __pyx_v_ix = 0;

  /* "ttcrpy/rgrid.pyx":1349
 *         cdef size_t k = 0
 *         cdef size_t ix = 0
 *         cdef size_t iy = 0             # <<<<<<<<<<<<<<
 *         cdef size_t iz = 0
 * 
 */
  __pyx_v_iy = 0;

  /* "ttcrpy/rgrid.pyx":1350
 *         cdef size_t ix = 0
 *         cdef size_t iy = 0
 *         cdef size_t iz = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double x, y, z, x1, y1, z1, x2, y2, z2, dtmp, d, l, m, n
 */
  __pyx_v_iz = 0;

  /* "ttcrpy/rgrid.pyx":1358
 *         cdef Py_ssize_t nt
 * 
 *         for nt in range(nTx):             # <<<<<<<<<<<<<<
 *             indptr_p.append(k)
 * 
 */
  __pyx_t_2 = __pyx_v_nTx;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_nt = __pyx_t_4;

    /* "ttcrpy/rgrid.pyx":1359
 * 
 *         for nt in range(nTx):
 *             indptr_p.append(k)             # <<<<<<<<<<<<<<
 * 
 *             x1 = Tx[nt, 0]
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indptr_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1359, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1361
 *             indptr_p.append(k)
 * 
 *             x1 = Tx[nt, 0]             # <<<<<<<<<<<<<<
 *             y1 = Tx[nt, 1]
 *             z1 = Tx[nt, 2]
 */
    __pyx_t_6 = __pyx_v_nt;
    __pyx_t_7 = 0;
    __pyx_t_8 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Tx.diminfo[0].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Tx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Tx.diminfo[1].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Tx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1361, __pyx_L1_error)
    }
    __pyx_v_x1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Tx.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_Tx.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_Tx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1362
 * 
 *             x1 = Tx[nt, 0]
 *             y1 = Tx[nt, 1]             # <<<<<<<<<<<<<<
 *             z1 = Tx[nt, 2]
 *             x2 = Rx[nt, 0]
 */
    __pyx_t_7 = __pyx_v_nt;
    __pyx_t_6 = 1;
    __pyx_t_8 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Tx.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Tx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Tx.diminfo[1].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Tx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1362, __pyx_L1_error)
    }
    __pyx_v_y1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Tx.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_Tx.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_Tx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1363
 *             x1 = Tx[nt, 0]
 *             y1 = Tx[nt, 1]
 *             z1 = Tx[nt, 2]             # <<<<<<<<<<<<<<
 *             x2 = Rx[nt, 0]
 *             y2 = Rx[nt, 1]
 */
    __pyx_t_6 = __pyx_v_nt;
    __pyx_t_7 = 2;
    __pyx_t_8 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Tx.diminfo[0].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Tx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Tx.diminfo[1].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Tx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1363, __pyx_L1_error)
    }
    __pyx_v_z1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Tx.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_Tx.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_Tx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1364
 *             y1 = Tx[nt, 1]
 *             z1 = Tx[nt, 2]
 *             x2 = Rx[nt, 0]             # <<<<<<<<<<<<<<
 *             y2 = Rx[nt, 1]
 *             z2 = Rx[nt, 2]
 */
    __pyx_t_7 = __pyx_v_nt;
    __pyx_t_6 = 0;
    __pyx_t_8 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Rx.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Rx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Rx.diminfo[1].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Rx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1364, __pyx_L1_error)
    }
    __pyx_v_x2 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Rx.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_Rx.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_Rx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1365
 *             z1 = Tx[nt, 2]
 *             x2 = Rx[nt, 0]
 *             y2 = Rx[nt, 1]             # <<<<<<<<<<<<<<
 *             z2 = Rx[nt, 2]
 * 
 */
    __pyx_t_6 = __pyx_v_nt;
    __pyx_t_7 = 1;
    __pyx_t_8 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Rx.diminfo[0].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Rx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Rx.diminfo[1].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Rx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1365, __pyx_L1_error)
    }
    __pyx_v_y2 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Rx.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_Rx.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_Rx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1366
 *             x2 = Rx[nt, 0]
 *             y2 = Rx[nt, 1]
 *             z2 = Rx[nt, 2]             # <<<<<<<<<<<<<<
 * 
 *             if x1 > x2:  # on veut x croissant
 */
    __pyx_t_7 = __pyx_v_nt;
    __pyx_t_6 = 2;
    __pyx_t_8 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Rx.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Rx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Rx.diminfo[1].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Rx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 1366, __pyx_L1_error)
    }
    __pyx_v_z2 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Rx.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_Rx.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_Rx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":1368
 *             z2 = Rx[nt, 2]
 * 
 *             if x1 > x2:  # on veut x croissant             # <<<<<<<<<<<<<<
 *                 dtmp = x1
 *                 x1 = x2
 */
    __pyx_t_9 = ((__pyx_v_x1 > __pyx_v_x2) != 0);
    if (__pyx_t_9) {

      /* "ttcrpy/rgrid.pyx":1369
 * 
 *             if x1 > x2:  # on veut x croissant
 *                 dtmp = x1             # <<<<<<<<<<<<<<
 *                 x1 = x2
 *                 x2 = dtmp
 */
      __pyx_v_dtmp = __pyx_v_x1;

      /* "ttcrpy/rgrid.pyx":1370
 *             if x1 > x2:  # on veut x croissant
 *                 dtmp = x1
 *                 x1 = x2             # <<<<<<<<<<<<<<
 *                 x2 = dtmp
 *                 dtmp = y1
 */
      __pyx_v_x1 = __pyx_v_x2;

      /* "ttcrpy/rgrid.pyx":1371
 *                 dtmp = x1
 *                 x1 = x2
 *                 x2 = dtmp             # <<<<<<<<<<<<<<
 *                 dtmp = y1
 *                 y1 = y2
 */
      __pyx_v_x2 = __pyx_v_dtmp;

      /* "ttcrpy/rgrid.pyx":1372
 *                 x1 = x2
 *                 x2 = dtmp
 *                 dtmp = y1             # <<<<<<<<<<<<<<
 *                 y1 = y2
 *                 y2 = dtmp
 */
      __pyx_v_dtmp = __pyx_v_y1;

      /* "ttcrpy/rgrid.pyx":1373
 *                 x2 = dtmp
 *                 dtmp = y1
 *                 y1 = y2             # <<<<<<<<<<<<<<
 *                 y2 = dtmp
 *                 dtmp = z1
 */
      __pyx_v_y1 = __pyx_v_y2;

      /* "ttcrpy/rgrid.pyx":1374
 *                 dtmp = y1
 *                 y1 = y2
 *                 y2 = dtmp             # <<<<<<<<<<<<<<
 *                 dtmp = z1
 *                 z1 = z2
 */
      __pyx_v_y2 = __pyx_v_dtmp;

      /* "ttcrpy/rgrid.pyx":1375
 *                 y1 = y2
 *                 y2 = dtmp
 *                 dtmp = z1             # <<<<<<<<<<<<<<
 *                 z1 = z2
 *                 z2 = dtmp
 */
      __pyx_v_dtmp = __pyx_v_z1;

      /* "ttcrpy/rgrid.pyx":1376
 *                 y2 = dtmp
 *                 dtmp = z1
 *                 z1 = z2             # <<<<<<<<<<<<<<
 *                 z2 = dtmp
 * 
 */
      __pyx_v_z1 = __pyx_v_z2;

      /* "ttcrpy/rgrid.pyx":1377
 *                 dtmp = z1
 *                 z1 = z2
 *                 z2 = dtmp             # <<<<<<<<<<<<<<
 * 
 *             d = sqrt( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1) )
 */
      __pyx_v_z2 = __pyx_v_dtmp;

      /* "ttcrpy/rgrid.pyx":1368
 *             z2 = Rx[nt, 2]
 * 
 *             if x1 > x2:  # on veut x croissant             # <<<<<<<<<<<<<<
 *                 dtmp = x1
 *                 x1 = x2
 */
    }

    /* "ttcrpy/rgrid.pyx":1379
 *                 z2 = dtmp
 * 
 *             d = sqrt( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1) )             # <<<<<<<<<<<<<<
 *             # cosinus directeurs
 *             l = (x2-x1)/d
 */
    __pyx_v_d = sqrt(((((__pyx_v_x2 - __pyx_v_x1) * (__pyx_v_x2 - __pyx_v_x1)) + ((__pyx_v_y2 - __pyx_v_y1) * (__pyx_v_y2 - __pyx_v_y1))) + ((__pyx_v_z2 - __pyx_v_z1) * (__pyx_v_z2 - __pyx_v_z1))));

    /* "ttcrpy/rgrid.pyx":1381
 *             d = sqrt( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1) )
 *             # cosinus directeurs
 *             l = (x2-x1)/d             # <<<<<<<<<<<<<<
 *             m = (y2-y1)/d
 *             n = (z2-z1)/d
 */
    __pyx_t_10 = (__pyx_v_x2 - __pyx_v_x1);
    if (unlikely(__pyx_v_d == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 1381, __pyx_L1_error)
    }
    __pyx_v_l = (__pyx_t_10 / __pyx_v_d);

    /* "ttcrpy/rgrid.pyx":1382
 *             # cosinus directeurs
 *             l = (x2-x1)/d
 *             m = (y2-y1)/d             # <<<<<<<<<<<<<<
 *             n = (z2-z1)/d
 * 
 */
    __pyx_t_10 = (__pyx_v_y2 - __pyx_v_y1);
    if (unlikely(__pyx_v_d == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 1382, __pyx_L1_error)
    }
    __pyx_v_m = (__pyx_t_10 / __pyx_v_d);

    /* "ttcrpy/rgrid.pyx":1383
 *             l = (x2-x1)/d
 *             m = (y2-y1)/d
 *             n = (z2-z1)/d             # <<<<<<<<<<<<<<
 * 
 *             sy = (m > 0) - (m < 0)  # sign of m
 */
    __pyx_t_10 = (__pyx_v_z2 - __pyx_v_z1);
    if (unlikely(__pyx_v_d == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 1383, __pyx_L1_error)
    }
    __pyx_v_n = (__pyx_t_10 / __pyx_v_d);

    /* "ttcrpy/rgrid.pyx":1385
 *             n = (z2-z1)/d
 * 
 *             sy = (m > 0) - (m < 0)  # sign of m             # <<<<<<<<<<<<<<
 *             sz = (n > 0) - (n < 0)
 * 
 */
    __pyx_v_sy = ((__pyx_v_m > 0.0) - (__pyx_v_m < 0.0));

    /* "ttcrpy/rgrid.pyx":1386
 * 
 *             sy = (m > 0) - (m < 0)  # sign of m
 *             sz = (n > 0) - (n < 0)             # <<<<<<<<<<<<<<
 * 
 *             x = x1
 */
    __pyx_v_sz = ((__pyx_v_n > 0.0) - (__pyx_v_n < 0.0));

    /* "ttcrpy/rgrid.pyx":1388
 *             sz = (n > 0) - (n < 0)
 * 
 *             x = x1             # <<<<<<<<<<<<<<
 *             y = y1
 *             z = z1
 */
    __pyx_v_x = __pyx_v_x1;

    /* "ttcrpy/rgrid.pyx":1389
 * 
 *             x = x1
 *             y = y1             # <<<<<<<<<<<<<<
 *             z = z1
 * 
 */
    __pyx_v_y = __pyx_v_y1;

    /* "ttcrpy/rgrid.pyx":1390
 *             x = x1
 *             y = y1
 *             z = z1             # <<<<<<<<<<<<<<
 * 
 *             for ix in range(n_grx-1):
 */
    __pyx_v_z = __pyx_v_z1;

    /* "ttcrpy/rgrid.pyx":1392
 *             z = z1
 * 
 *             for ix in range(n_grx-1):             # <<<<<<<<<<<<<<
 *                 if x < grx[ix+1] and x >= grx[ix]:
 *                     break
 */
    __pyx_t_11 = (__pyx_v_n_grx - 1);
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_ix = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":1393
 * 
 *             for ix in range(n_grx-1):
 *                 if x < grx[ix+1] and x >= grx[ix]:             # <<<<<<<<<<<<<<
 *                     break
 *             for iy in range(n_gry-1):
 */
      __pyx_t_14 = (__pyx_v_ix + 1);
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1393, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_x < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grx.diminfo[0].strides))) != 0);
      if (__pyx_t_15) {
      } else {
        __pyx_t_9 = __pyx_t_15;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_14 = __pyx_v_ix;
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1393, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_x >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grx.diminfo[0].strides))) != 0);
      __pyx_t_9 = __pyx_t_15;
      __pyx_L9_bool_binop_done:;
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1394
 *             for ix in range(n_grx-1):
 *                 if x < grx[ix+1] and x >= grx[ix]:
 *                     break             # <<<<<<<<<<<<<<
 *             for iy in range(n_gry-1):
 *                 if y < gry[iy+1] and y >= gry[iy]:
 */
        goto __pyx_L7_break;

        /* "ttcrpy/rgrid.pyx":1393
 * 
 *             for ix in range(n_grx-1):
 *                 if x < grx[ix+1] and x >= grx[ix]:             # <<<<<<<<<<<<<<
 *                     break
 *             for iy in range(n_gry-1):
 */
      }
    }
    __pyx_L7_break:;

    /* "ttcrpy/rgrid.pyx":1395
 *                 if x < grx[ix+1] and x >= grx[ix]:
 *                     break
 *             for iy in range(n_gry-1):             # <<<<<<<<<<<<<<
 *                 if y < gry[iy+1] and y >= gry[iy]:
 *                     break
 */
    __pyx_t_11 = (__pyx_v_n_gry - 1);
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_iy = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":1396
 *                     break
 *             for iy in range(n_gry-1):
 *                 if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                     break
 *             for iz in range(n_grz-1):
 */
      __pyx_t_14 = (__pyx_v_iy + 1);
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1396, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_y < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
      if (__pyx_t_15) {
      } else {
        __pyx_t_9 = __pyx_t_15;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_14 = __pyx_v_iy;
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1396, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_y >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
      __pyx_t_9 = __pyx_t_15;
      __pyx_L14_bool_binop_done:;
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1397
 *             for iy in range(n_gry-1):
 *                 if y < gry[iy+1] and y >= gry[iy]:
 *                     break             # <<<<<<<<<<<<<<
 *             for iz in range(n_grz-1):
 *                 if z < grz[iz+1] and z >= grz[iz]:
 */
        goto __pyx_L12_break;

        /* "ttcrpy/rgrid.pyx":1396
 *                     break
 *             for iy in range(n_gry-1):
 *                 if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                     break
 *             for iz in range(n_grz-1):
 */
      }
    }
    __pyx_L12_break:;

    /* "ttcrpy/rgrid.pyx":1398
 *                 if y < gry[iy+1] and y >= gry[iy]:
 *                     break
 *             for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                 if z < grz[iz+1] and z >= grz[iz]:
 *                     break
 */
    __pyx_t_11 = (__pyx_v_n_grz - 1);
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_iz = __pyx_t_13;

      /* "ttcrpy/rgrid.pyx":1399
 *                     break
 *             for iz in range(n_grz-1):
 *                 if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      __pyx_t_14 = (__pyx_v_iz + 1);
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1399, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
      if (__pyx_t_15) {
      } else {
        __pyx_t_9 = __pyx_t_15;
        goto __pyx_L19_bool_binop_done;
      }
      __pyx_t_14 = __pyx_v_iz;
      __pyx_t_8 = -1;
      if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
      if (unlikely(__pyx_t_8 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_8);
        __PYX_ERR(0, 1399, __pyx_L1_error)
      }
      __pyx_t_15 = ((__pyx_v_z >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
      __pyx_t_9 = __pyx_t_15;
      __pyx_L19_bool_binop_done:;
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1400
 *             for iz in range(n_grz-1):
 *                 if z < grz[iz+1] and z >= grz[iz]:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             if abs(l) < small:
 */
        goto __pyx_L17_break;

        /* "ttcrpy/rgrid.pyx":1399
 *                     break
 *             for iz in range(n_grz-1):
 *                 if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      }
    }
    __pyx_L17_break:;

    /* "ttcrpy/rgrid.pyx":1402
 *                     break
 * 
 *             if abs(l) < small:             # <<<<<<<<<<<<<<
 *                 if abs(m) < small:
 *                     # X & Y constants
 */
    __pyx_t_9 = ((fabs(__pyx_v_l) < __pyx_v_small) != 0);
    if (__pyx_t_9) {

      /* "ttcrpy/rgrid.pyx":1403
 * 
 *             if abs(l) < small:
 *                 if abs(m) < small:             # <<<<<<<<<<<<<<
 *                     # X & Y constants
 *                     if z1 > z2:
 */
      __pyx_t_9 = ((fabs(__pyx_v_m) < __pyx_v_small) != 0);
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1405
 *                 if abs(m) < small:
 *                     # X & Y constants
 *                     if z1 > z2:             # <<<<<<<<<<<<<<
 *                         dtmp = z1
 *                         z1 = z2
 */
        __pyx_t_9 = ((__pyx_v_z1 > __pyx_v_z2) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":1406
 *                     # X & Y constants
 *                     if z1 > z2:
 *                         dtmp = z1             # <<<<<<<<<<<<<<
 *                         z1 = z2
 *                         z2 = dtmp
 */
          __pyx_v_dtmp = __pyx_v_z1;

          /* "ttcrpy/rgrid.pyx":1407
 *                     if z1 > z2:
 *                         dtmp = z1
 *                         z1 = z2             # <<<<<<<<<<<<<<
 *                         z2 = dtmp
 *                         z = z1
 */
          __pyx_v_z1 = __pyx_v_z2;

          /* "ttcrpy/rgrid.pyx":1408
 *                         dtmp = z1
 *                         z1 = z2
 *                         z2 = dtmp             # <<<<<<<<<<<<<<
 *                         z = z1
 *                         for iz in range(n_grz-1):
 */
          __pyx_v_z2 = __pyx_v_dtmp;

          /* "ttcrpy/rgrid.pyx":1409
 *                         z1 = z2
 *                         z2 = dtmp
 *                         z = z1             # <<<<<<<<<<<<<<
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:
 */
          __pyx_v_z = __pyx_v_z1;

          /* "ttcrpy/rgrid.pyx":1410
 *                         z2 = dtmp
 *                         z = z1
 *                         for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                             if z < grz[iz+1] and z >= grz[iz]:
 *                                 break
 */
          __pyx_t_11 = (__pyx_v_n_grz - 1);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
            __pyx_v_iz = __pyx_t_13;

            /* "ttcrpy/rgrid.pyx":1411
 *                         z = z1
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                                 break
 *                     while z < z2:
 */
            __pyx_t_14 = (__pyx_v_iz + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1411, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L27_bool_binop_done;
            }
            __pyx_t_14 = __pyx_v_iz;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1411, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_z >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L27_bool_binop_done:;
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1412
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:
 *                                 break             # <<<<<<<<<<<<<<
 *                     while z < z2:
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
              goto __pyx_L25_break;

              /* "ttcrpy/rgrid.pyx":1411
 *                         z = z1
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                                 break
 *                     while z < z2:
 */
            }
          }
          __pyx_L25_break:;

          /* "ttcrpy/rgrid.pyx":1405
 *                 if abs(m) < small:
 *                     # X & Y constants
 *                     if z1 > z2:             # <<<<<<<<<<<<<<
 *                         dtmp = z1
 *                         z1 = z2
 */
        }

        /* "ttcrpy/rgrid.pyx":1413
 *                             if z < grz[iz+1] and z >= grz[iz]:
 *                                 break
 *                     while z < z2:             # <<<<<<<<<<<<<<
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_z < __pyx_v_z2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1414
 *                                 break
 *                     while z < z2:
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                         dlz = ( grz[iz+1] if grz[iz+1] < z2 else z2 ) - z
 */
          __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

          /* "ttcrpy/rgrid.pyx":1416
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                         dlz = ( grz[iz+1] if grz[iz+1] < z2 else z2 ) - z             # <<<<<<<<<<<<<<
 * 
 *                         indices_p.append(iCell)
 */
          __pyx_t_11 = (__pyx_v_iz + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1416, __pyx_L1_error)
          }
          if ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)) < __pyx_v_z2) != 0)) {
            __pyx_t_12 = (__pyx_v_iz + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_12 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1416, __pyx_L1_error)
            }
            __pyx_t_16 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_grz.diminfo[0].strides));
          } else {
            __pyx_t_16 = __pyx_v_z2;
          }
          __pyx_v_dlz = (__pyx_t_16 - __pyx_v_z);

          /* "ttcrpy/rgrid.pyx":1418
 *                         dlz = ( grz[iz+1] if grz[iz+1] < z2 else z2 ) - z
 * 
 *                         indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                         data_p.append(dlz)
 *                         k += 1
 */
          __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1418, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1418, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1419
 * 
 *                         indices_p.append(iCell)
 *                         data_p.append(dlz)             # <<<<<<<<<<<<<<
 *                         k += 1
 * 
 */
          __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1420
 *                         indices_p.append(iCell)
 *                         data_p.append(dlz)
 *                         k += 1             # <<<<<<<<<<<<<<
 * 
 *                         iz += 1
 */
          __pyx_v_k = (__pyx_v_k + 1);

          /* "ttcrpy/rgrid.pyx":1422
 *                         k += 1
 * 
 *                         iz += 1             # <<<<<<<<<<<<<<
 *                         z = grz[iz]
 * 
 */
          __pyx_v_iz = (__pyx_v_iz + 1);

          /* "ttcrpy/rgrid.pyx":1423
 * 
 *                         iz += 1
 *                         z = grz[iz]             # <<<<<<<<<<<<<<
 * 
 *                 elif abs(n) < small:
 */
          __pyx_t_11 = __pyx_v_iz;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1423, __pyx_L1_error)
          }
          __pyx_v_z = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
        }

        /* "ttcrpy/rgrid.pyx":1403
 * 
 *             if abs(l) < small:
 *                 if abs(m) < small:             # <<<<<<<<<<<<<<
 *                     # X & Y constants
 *                     if z1 > z2:
 */
        goto __pyx_L22;
      }

      /* "ttcrpy/rgrid.pyx":1425
 *                         z = grz[iz]
 * 
 *                 elif abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # X & Z constants
 *                     if y1 > y2:
 */
      __pyx_t_9 = ((fabs(__pyx_v_n) < __pyx_v_small) != 0);
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1427
 *                 elif abs(n) < small:
 *                     # X & Z constants
 *                     if y1 > y2:             # <<<<<<<<<<<<<<
 *                         dtmp = y1
 *                         y1 = y2
 */
        __pyx_t_9 = ((__pyx_v_y1 > __pyx_v_y2) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":1428
 *                     # X & Z constants
 *                     if y1 > y2:
 *                         dtmp = y1             # <<<<<<<<<<<<<<
 *                         y1 = y2
 *                         y2 = dtmp
 */
          __pyx_v_dtmp = __pyx_v_y1;

          /* "ttcrpy/rgrid.pyx":1429
 *                     if y1 > y2:
 *                         dtmp = y1
 *                         y1 = y2             # <<<<<<<<<<<<<<
 *                         y2 = dtmp
 *                         y = y1
 */
          __pyx_v_y1 = __pyx_v_y2;

          /* "ttcrpy/rgrid.pyx":1430
 *                         dtmp = y1
 *                         y1 = y2
 *                         y2 = dtmp             # <<<<<<<<<<<<<<
 *                         y = y1
 *                         for iy in range(n_gry-1):
 */
          __pyx_v_y2 = __pyx_v_dtmp;

          /* "ttcrpy/rgrid.pyx":1431
 *                         y1 = y2
 *                         y2 = dtmp
 *                         y = y1             # <<<<<<<<<<<<<<
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:
 */
          __pyx_v_y = __pyx_v_y1;

          /* "ttcrpy/rgrid.pyx":1432
 *                         y2 = dtmp
 *                         y = y1
 *                         for iy in range(n_gry-1):             # <<<<<<<<<<<<<<
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break
 */
          __pyx_t_11 = (__pyx_v_n_gry - 1);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
            __pyx_v_iy = __pyx_t_13;

            /* "ttcrpy/rgrid.pyx":1433
 *                         y = y1
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                                 break
 *                     while y < y2:
 */
            __pyx_t_14 = (__pyx_v_iy + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1433, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_y < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L35_bool_binop_done;
            }
            __pyx_t_14 = __pyx_v_iy;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1433, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_y >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L35_bool_binop_done:;
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1434
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break             # <<<<<<<<<<<<<<
 *                     while y < y2:
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
              goto __pyx_L33_break;

              /* "ttcrpy/rgrid.pyx":1433
 *                         y = y1
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                                 break
 *                     while y < y2:
 */
            }
          }
          __pyx_L33_break:;

          /* "ttcrpy/rgrid.pyx":1427
 *                 elif abs(n) < small:
 *                     # X & Z constants
 *                     if y1 > y2:             # <<<<<<<<<<<<<<
 *                         dtmp = y1
 *                         y1 = y2
 */
        }

        /* "ttcrpy/rgrid.pyx":1435
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break
 *                     while y < y2:             # <<<<<<<<<<<<<<
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_y < __pyx_v_y2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1436
 *                                 break
 *                     while y < y2:
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                         dly = ( gry[iy+1] if gry[iy+1] < y2 else y2 ) - y
 */
          __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

          /* "ttcrpy/rgrid.pyx":1438
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                         dly = ( gry[iy+1] if gry[iy+1] < y2 else y2 ) - y             # <<<<<<<<<<<<<<
 * 
 *                         indices_p.append(iCell)
 */
          __pyx_t_11 = (__pyx_v_iy + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1438, __pyx_L1_error)
          }
          if ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides)) < __pyx_v_y2) != 0)) {
            __pyx_t_12 = (__pyx_v_iy + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_12 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1438, __pyx_L1_error)
            }
            __pyx_t_16 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_gry.diminfo[0].strides));
          } else {
            __pyx_t_16 = __pyx_v_y2;
          }
          __pyx_v_dly = (__pyx_t_16 - __pyx_v_y);

          /* "ttcrpy/rgrid.pyx":1440
 *                         dly = ( gry[iy+1] if gry[iy+1] < y2 else y2 ) - y
 * 
 *                         indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                         data_p.append(dly)
 *                         k += 1
 */
          __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1440, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1441
 * 
 *                         indices_p.append(iCell)
 *                         data_p.append(dly)             # <<<<<<<<<<<<<<
 *                         k += 1
 * 
 */
          __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dly); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1441, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1441, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1442
 *                         indices_p.append(iCell)
 *                         data_p.append(dly)
 *                         k += 1             # <<<<<<<<<<<<<<
 * 
 *                         iy += 1
 */
          __pyx_v_k = (__pyx_v_k + 1);

          /* "ttcrpy/rgrid.pyx":1444
 *                         k += 1
 * 
 *                         iy += 1             # <<<<<<<<<<<<<<
 *                         y = gry[iy]
 * 
 */
          __pyx_v_iy = (__pyx_v_iy + 1);

          /* "ttcrpy/rgrid.pyx":1445
 * 
 *                         iy += 1
 *                         y = gry[iy]             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
          __pyx_t_11 = __pyx_v_iy;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1445, __pyx_L1_error)
          }
          __pyx_v_y = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
        }

        /* "ttcrpy/rgrid.pyx":1425
 *                         z = grz[iz]
 * 
 *                 elif abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # X & Z constants
 *                     if y1 > y2:
 */
        goto __pyx_L22;
      }

      /* "ttcrpy/rgrid.pyx":1449
 *                 else:
 *                     # seul X constant
 *                     if y1 > y2:             # <<<<<<<<<<<<<<
 *                         dtmp = y1
 *                         y1 = y2
 */
      /*else*/ {
        __pyx_t_9 = ((__pyx_v_y1 > __pyx_v_y2) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":1450
 *                     # seul X constant
 *                     if y1 > y2:
 *                         dtmp = y1             # <<<<<<<<<<<<<<
 *                         y1 = y2
 *                         y2 = dtmp
 */
          __pyx_v_dtmp = __pyx_v_y1;

          /* "ttcrpy/rgrid.pyx":1451
 *                     if y1 > y2:
 *                         dtmp = y1
 *                         y1 = y2             # <<<<<<<<<<<<<<
 *                         y2 = dtmp
 *                         dtmp = z1
 */
          __pyx_v_y1 = __pyx_v_y2;

          /* "ttcrpy/rgrid.pyx":1452
 *                         dtmp = y1
 *                         y1 = y2
 *                         y2 = dtmp             # <<<<<<<<<<<<<<
 *                         dtmp = z1
 *                         z1 = z2
 */
          __pyx_v_y2 = __pyx_v_dtmp;

          /* "ttcrpy/rgrid.pyx":1453
 *                         y1 = y2
 *                         y2 = dtmp
 *                         dtmp = z1             # <<<<<<<<<<<<<<
 *                         z1 = z2
 *                         z2 = dtmp
 */
          __pyx_v_dtmp = __pyx_v_z1;

          /* "ttcrpy/rgrid.pyx":1454
 *                         y2 = dtmp
 *                         dtmp = z1
 *                         z1 = z2             # <<<<<<<<<<<<<<
 *                         z2 = dtmp
 *                         y = y1
 */
          __pyx_v_z1 = __pyx_v_z2;

          /* "ttcrpy/rgrid.pyx":1455
 *                         dtmp = z1
 *                         z1 = z2
 *                         z2 = dtmp             # <<<<<<<<<<<<<<
 *                         y = y1
 *                         for iy in range(n_gry-1):
 */
          __pyx_v_z2 = __pyx_v_dtmp;

          /* "ttcrpy/rgrid.pyx":1456
 *                         z1 = z2
 *                         z2 = dtmp
 *                         y = y1             # <<<<<<<<<<<<<<
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:
 */
          __pyx_v_y = __pyx_v_y1;

          /* "ttcrpy/rgrid.pyx":1457
 *                         z2 = dtmp
 *                         y = y1
 *                         for iy in range(n_gry-1):             # <<<<<<<<<<<<<<
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break
 */
          __pyx_t_11 = (__pyx_v_n_gry - 1);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
            __pyx_v_iy = __pyx_t_13;

            /* "ttcrpy/rgrid.pyx":1458
 *                         y = y1
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                                 break
 *                         z = z1
 */
            __pyx_t_14 = (__pyx_v_iy + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1458, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_y < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L43_bool_binop_done;
            }
            __pyx_t_14 = __pyx_v_iy;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1458, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_y >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_gry.diminfo[0].strides))) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L43_bool_binop_done:;
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1459
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break             # <<<<<<<<<<<<<<
 *                         z = z1
 *                         for iz in range(n_grz-1):
 */
              goto __pyx_L41_break;

              /* "ttcrpy/rgrid.pyx":1458
 *                         y = y1
 *                         for iy in range(n_gry-1):
 *                             if y < gry[iy+1] and y >= gry[iy]:             # <<<<<<<<<<<<<<
 *                                 break
 *                         z = z1
 */
            }
          }
          __pyx_L41_break:;

          /* "ttcrpy/rgrid.pyx":1460
 *                             if y < gry[iy+1] and y >= gry[iy]:
 *                                 break
 *                         z = z1             # <<<<<<<<<<<<<<
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:
 */
          __pyx_v_z = __pyx_v_z1;

          /* "ttcrpy/rgrid.pyx":1461
 *                                 break
 *                         z = z1
 *                         for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                             if z < grz[iz+1] and z >= grz[iz]:
 *                                 break
 */
          __pyx_t_11 = (__pyx_v_n_grz - 1);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
            __pyx_v_iz = __pyx_t_13;

            /* "ttcrpy/rgrid.pyx":1462
 *                         z = z1
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                                 break
 * 
 */
            __pyx_t_14 = (__pyx_v_iz + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1462, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L48_bool_binop_done;
            }
            __pyx_t_14 = __pyx_v_iz;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1462, __pyx_L1_error)
            }
            __pyx_t_15 = ((__pyx_v_z >= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L48_bool_binop_done:;
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1463
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:
 *                                 break             # <<<<<<<<<<<<<<
 * 
 *                     m_z = (z2-z1) / (y2-y1)
 */
              goto __pyx_L46_break;

              /* "ttcrpy/rgrid.pyx":1462
 *                         z = z1
 *                         for iz in range(n_grz-1):
 *                             if z < grz[iz+1] and z >= grz[iz]:             # <<<<<<<<<<<<<<
 *                                 break
 * 
 */
            }
          }
          __pyx_L46_break:;

          /* "ttcrpy/rgrid.pyx":1449
 *                 else:
 *                     # seul X constant
 *                     if y1 > y2:             # <<<<<<<<<<<<<<
 *                         dtmp = y1
 *                         y1 = y2
 */
        }

        /* "ttcrpy/rgrid.pyx":1465
 *                                 break
 * 
 *                     m_z = (z2-z1) / (y2-y1)             # <<<<<<<<<<<<<<
 *                     b_z = z2 - m_z*y2
 *                     up_z = m_z > 0
 */
        __pyx_t_10 = (__pyx_v_z2 - __pyx_v_z1);
        __pyx_t_17 = (__pyx_v_y2 - __pyx_v_y1);
        if (unlikely(__pyx_t_17 == 0)) {
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");
          __PYX_ERR(0, 1465, __pyx_L1_error)
        }
        __pyx_v_m_z = (__pyx_t_10 / __pyx_t_17);

        /* "ttcrpy/rgrid.pyx":1466
 * 
 *                     m_z = (z2-z1) / (y2-y1)
 *                     b_z = z2 - m_z*y2             # <<<<<<<<<<<<<<
 *                     up_z = m_z > 0
 * 
 */
        __pyx_v_b_z = (__pyx_v_z2 - (__pyx_v_m_z * __pyx_v_y2));

        /* "ttcrpy/rgrid.pyx":1467
 *                     m_z = (z2-z1) / (y2-y1)
 *                     b_z = z2 - m_z*y2
 *                     up_z = m_z > 0             # <<<<<<<<<<<<<<
 * 
 *                     while y < y2:
 */
        __pyx_v_up_z = (__pyx_v_m_z > 0.0);

        /* "ttcrpy/rgrid.pyx":1469
 *                     up_z = m_z > 0
 * 
 *                     while y < y2:             # <<<<<<<<<<<<<<
 * 
 *                         zi = m_z*gry[iy+1] + b_z
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_y < __pyx_v_y2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1471
 *                     while y < y2:
 * 
 *                         zi = m_z*gry[iy+1] + b_z             # <<<<<<<<<<<<<<
 *                         if up_z:
 *                             while z < zi and z < z2:
 */
          __pyx_t_11 = (__pyx_v_iy + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1471, __pyx_L1_error)
          }
          __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m_z * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides))) + __pyx_v_b_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1471, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_zi, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1472
 * 
 *                         zi = m_z*gry[iy+1] + b_z
 *                         if up_z:             # <<<<<<<<<<<<<<
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
          __pyx_t_9 = (__pyx_v_up_z != 0);
          if (__pyx_t_9) {

            /* "ttcrpy/rgrid.pyx":1473
 *                         zi = m_z*gry[iy+1] + b_z
 *                         if up_z:
 *                             while z < zi and z < z2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1473, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1473, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1473, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L55_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_z < __pyx_v_z2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L55_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1474
 *                         if up_z:
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ze = grz[iz+1] if grz[iz+1]<zi else zi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1476
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ze = grz[iz+1] if grz[iz+1]<zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze < z2 else z2
 *                                 ye = (ze-b_z) / m_z
 */
              __pyx_t_11 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1476, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1476, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1476, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = (__pyx_v_iz + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1476, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1477
 * 
 *                                 ze = grz[iz+1] if grz[iz+1]<zi else zi
 *                                 ze = ze if ze < z2 else z2             # <<<<<<<<<<<<<<
 *                                 ye = (ze-b_z) / m_z
 *                                 dly = ye - y
 */
              if (((__pyx_v_ze < __pyx_v_z2) != 0)) {
                __pyx_t_17 = __pyx_v_ze;
              } else {
                __pyx_t_17 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1478
 *                                 ze = grz[iz+1] if grz[iz+1]<zi else zi
 *                                 ze = ze if ze < z2 else z2
 *                                 ye = (ze-b_z) / m_z             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dlz = ze - z
 */
              __pyx_t_17 = (__pyx_v_ze - __pyx_v_b_z);
              if (unlikely(__pyx_v_m_z == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1478, __pyx_L1_error)
              }
              __pyx_v_ye = (__pyx_t_17 / __pyx_v_m_z);

              /* "ttcrpy/rgrid.pyx":1479
 *                                 ze = ze if ze < z2 else z2
 *                                 ye = (ze-b_z) / m_z
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1480
 *                                 ye = (ze-b_z) / m_z
 *                                 dly = ye - y
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 * 
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1481
 *                                 dly = ye - y
 *                                 dlz = ze - z
 *                                 dl = sqrt( dly*dly + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dly * __pyx_v_dly) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1483
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1483, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1484
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1484, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1485
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 y = ye
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1487
 *                                 k += 1
 * 
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 z = ze
 * 
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1488
 * 
 *                                 y = ye
 *                                 z = ze             # <<<<<<<<<<<<<<
 * 
 *                                 if abs(z-grz[iz+1]) < small:
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1490
 *                                 z = ze
 * 
 *                                 if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iz += 1
 * 
 */
              __pyx_t_11 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1490, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1491
 * 
 *                                 if abs(z-grz[iz+1]) < small:
 *                                     iz += 1             # <<<<<<<<<<<<<<
 * 
 *                         else: # down
 */
                __pyx_v_iz = (__pyx_v_iz + 1);

                /* "ttcrpy/rgrid.pyx":1490
 *                                 z = ze
 * 
 *                                 if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iz += 1
 * 
 */
              }
            }

            /* "ttcrpy/rgrid.pyx":1472
 * 
 *                         zi = m_z*gry[iy+1] + b_z
 *                         if up_z:             # <<<<<<<<<<<<<<
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
            goto __pyx_L52;
          }

          /* "ttcrpy/rgrid.pyx":1494
 * 
 *                         else: # down
 *                             while z > zi and z > z2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
          /*else*/ {
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1494, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1494, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L60_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_z > __pyx_v_z2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L60_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1495
 *                         else: # down
 *                             while z > zi and z > z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1497
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze > z2 else z2
 *                                 ye = (ze-b_z)/m_z
 */
              __pyx_t_11 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1497, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1497, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1497, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1497, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = __pyx_v_iz;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1497, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1497, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1498
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 *                                 ze = ze if ze > z2 else z2             # <<<<<<<<<<<<<<
 *                                 ye = (ze-b_z)/m_z
 *                                 dly = ye - y
 */
              if (((__pyx_v_ze > __pyx_v_z2) != 0)) {
                __pyx_t_17 = __pyx_v_ze;
              } else {
                __pyx_t_17 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1499
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 *                                 ze = ze if ze > z2 else z2
 *                                 ye = (ze-b_z)/m_z             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dlz = ze - z
 */
              __pyx_t_17 = (__pyx_v_ze - __pyx_v_b_z);
              if (unlikely(__pyx_v_m_z == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1499, __pyx_L1_error)
              }
              __pyx_v_ye = (__pyx_t_17 / __pyx_v_m_z);

              /* "ttcrpy/rgrid.pyx":1500
 *                                 ze = ze if ze > z2 else z2
 *                                 ye = (ze-b_z)/m_z
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1501
 *                                 ye = (ze-b_z)/m_z
 *                                 dly = ye - y
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 * 
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1502
 *                                 dly = ye - y
 *                                 dlz = ze - z
 *                                 dl = sqrt( dly*dly + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dly * __pyx_v_dly) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1504
 *                                 dl = sqrt( dly*dly + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1504, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1505
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1505, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1506
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 y = ye
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1508
 *                                 k += 1
 * 
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1509
 * 
 *                                 y = ye
 *                                 z = ze             # <<<<<<<<<<<<<<
 *                                 if abs(z-grz[iz]) < small:
 *                                     iz -= 1
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1510
 *                                 y = ye
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                     iz -= 1
 * 
 */
              __pyx_t_11 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1510, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1511
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:
 *                                     iz -= 1             # <<<<<<<<<<<<<<
 * 
 *                         iy += 1
 */
                __pyx_v_iz = (__pyx_v_iz - 1);

                /* "ttcrpy/rgrid.pyx":1510
 *                                 y = ye
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                     iz -= 1
 * 
 */
              }
            }
          }
          __pyx_L52:;

          /* "ttcrpy/rgrid.pyx":1513
 *                                     iz -= 1
 * 
 *                         iy += 1             # <<<<<<<<<<<<<<
 *                         y = gry[iy]
 *             else:
 */
          __pyx_v_iy = (__pyx_v_iy + 1);

          /* "ttcrpy/rgrid.pyx":1514
 * 
 *                         iy += 1
 *                         y = gry[iy]             # <<<<<<<<<<<<<<
 *             else:
 *                 if abs(m) < small and abs(n) < small:
 */
          __pyx_t_11 = __pyx_v_iy;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1514, __pyx_L1_error)
          }
          __pyx_v_y = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
        }
      }
      __pyx_L22:;

      /* "ttcrpy/rgrid.pyx":1402
 *                     break
 * 
 *             if abs(l) < small:             # <<<<<<<<<<<<<<
 *                 if abs(m) < small:
 *                     # X & Y constants
 */
      goto __pyx_L21;
    }

    /* "ttcrpy/rgrid.pyx":1516
 *                         y = gry[iy]
 *             else:
 *                 if abs(m) < small and abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # Y & Z constants
 *                     while x < x2:
 */
    /*else*/ {
      __pyx_t_15 = ((fabs(__pyx_v_m) < __pyx_v_small) != 0);
      if (__pyx_t_15) {
      } else {
        __pyx_t_9 = __pyx_t_15;
        goto __pyx_L64_bool_binop_done;
      }
      __pyx_t_15 = ((fabs(__pyx_v_n) < __pyx_v_small) != 0);
      __pyx_t_9 = __pyx_t_15;
      __pyx_L64_bool_binop_done:;
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1518
 *                 if abs(m) < small and abs(n) < small:
 *                     # Y & Z constants
 *                     while x < x2:             # <<<<<<<<<<<<<<
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_x < __pyx_v_x2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1519
 *                     # Y & Z constants
 *                     while x < x2:
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                         dlx = ( grx[ix+1] if grx[ix+1] < x2 else x2 ) - x
 */
          __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

          /* "ttcrpy/rgrid.pyx":1521
 *                         iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                         dlx = ( grx[ix+1] if grx[ix+1] < x2 else x2 ) - x             # <<<<<<<<<<<<<<
 * 
 *                         indices_p.append(iCell)
 */
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1521, __pyx_L1_error)
          }
          if ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides)) < __pyx_v_x2) != 0)) {
            __pyx_t_12 = (__pyx_v_ix + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_12 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 1521, __pyx_L1_error)
            }
            __pyx_t_16 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_grx.diminfo[0].strides));
          } else {
            __pyx_t_16 = __pyx_v_x2;
          }
          __pyx_v_dlx = (__pyx_t_16 - __pyx_v_x);

          /* "ttcrpy/rgrid.pyx":1523
 *                         dlx = ( grx[ix+1] if grx[ix+1] < x2 else x2 ) - x
 * 
 *                         indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                         data_p.append(dlx)
 *                         k += 1
 */
          __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1523, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1524
 * 
 *                         indices_p.append(iCell)
 *                         data_p.append(dlx)             # <<<<<<<<<<<<<<
 *                         k += 1
 * 
 */
          __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1524, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1525
 *                         indices_p.append(iCell)
 *                         data_p.append(dlx)
 *                         k += 1             # <<<<<<<<<<<<<<
 * 
 *                         ix += 1
 */
          __pyx_v_k = (__pyx_v_k + 1);

          /* "ttcrpy/rgrid.pyx":1527
 *                         k += 1
 * 
 *                         ix += 1             # <<<<<<<<<<<<<<
 *                         x = grx[ix]
 * 
 */
          __pyx_v_ix = (__pyx_v_ix + 1);

          /* "ttcrpy/rgrid.pyx":1528
 * 
 *                         ix += 1
 *                         x = grx[ix]             # <<<<<<<<<<<<<<
 * 
 *                 elif abs(m) < small:
 */
          __pyx_t_11 = __pyx_v_ix;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1528, __pyx_L1_error)
          }
          __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides));
        }

        /* "ttcrpy/rgrid.pyx":1516
 *                         y = gry[iy]
 *             else:
 *                 if abs(m) < small and abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # Y & Z constants
 *                     while x < x2:
 */
        goto __pyx_L63;
      }

      /* "ttcrpy/rgrid.pyx":1530
 *                         x = grx[ix]
 * 
 *                 elif abs(m) < small:             # <<<<<<<<<<<<<<
 *                     # seul Y constant
 *                     m_z = (z2-z1) / (x2-x1)
 */
      __pyx_t_9 = ((fabs(__pyx_v_m) < __pyx_v_small) != 0);
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1532
 *                 elif abs(m) < small:
 *                     # seul Y constant
 *                     m_z = (z2-z1) / (x2-x1)             # <<<<<<<<<<<<<<
 *                     b_z = z2 - m_z*x2
 *                     up_z = m_z>0
 */
        __pyx_t_17 = (__pyx_v_z2 - __pyx_v_z1);
        __pyx_t_10 = (__pyx_v_x2 - __pyx_v_x1);
        if (unlikely(__pyx_t_10 == 0)) {
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");
          __PYX_ERR(0, 1532, __pyx_L1_error)
        }
        __pyx_v_m_z = (__pyx_t_17 / __pyx_t_10);

        /* "ttcrpy/rgrid.pyx":1533
 *                     # seul Y constant
 *                     m_z = (z2-z1) / (x2-x1)
 *                     b_z = z2 - m_z*x2             # <<<<<<<<<<<<<<
 *                     up_z = m_z>0
 * 
 */
        __pyx_v_b_z = (__pyx_v_z2 - (__pyx_v_m_z * __pyx_v_x2));

        /* "ttcrpy/rgrid.pyx":1534
 *                     m_z = (z2-z1) / (x2-x1)
 *                     b_z = z2 - m_z*x2
 *                     up_z = m_z>0             # <<<<<<<<<<<<<<
 * 
 *                     while x < x2:
 */
        __pyx_v_up_z = (__pyx_v_m_z > 0.0);

        /* "ttcrpy/rgrid.pyx":1536
 *                     up_z = m_z>0
 * 
 *                     while x < x2:             # <<<<<<<<<<<<<<
 * 
 *                         zi = m_z*grx[ix+1] + b_z
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_x < __pyx_v_x2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1538
 *                     while x < x2:
 * 
 *                         zi = m_z*grx[ix+1] + b_z             # <<<<<<<<<<<<<<
 * 
 *                         if up_z:
 */
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1538, __pyx_L1_error)
          }
          __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m_z * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides))) + __pyx_v_b_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1538, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_zi, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1540
 *                         zi = m_z*grx[ix+1] + b_z
 * 
 *                         if up_z:             # <<<<<<<<<<<<<<
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
          __pyx_t_9 = (__pyx_v_up_z != 0);
          if (__pyx_t_9) {

            /* "ttcrpy/rgrid.pyx":1541
 * 
 *                         if up_z:
 *                             while z < zi and z < z2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1541, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1541, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L73_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_z < __pyx_v_z2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L73_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1542
 *                         if up_z:
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1544
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze < z2 else z2
 *                                 xe = (ze-b_z) / m_z
 */
              __pyx_t_11 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1544, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1544, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1544, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = (__pyx_v_iz + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1544, __pyx_L1_error)
                }
                __pyx_t_10 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1544, __pyx_L1_error)
                __pyx_t_10 = __pyx_t_17;
              }
              __pyx_v_ze = __pyx_t_10;

              /* "ttcrpy/rgrid.pyx":1545
 * 
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 *                                 ze = ze if ze < z2 else z2             # <<<<<<<<<<<<<<
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x
 */
              if (((__pyx_v_ze < __pyx_v_z2) != 0)) {
                __pyx_t_10 = __pyx_v_ze;
              } else {
                __pyx_t_10 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_10;

              /* "ttcrpy/rgrid.pyx":1546
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 *                                 ze = ze if ze < z2 else z2
 *                                 xe = (ze-b_z) / m_z             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dlz = ze - z
 */
              __pyx_t_10 = (__pyx_v_ze - __pyx_v_b_z);
              if (unlikely(__pyx_v_m_z == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1546, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_10 / __pyx_v_m_z);

              /* "ttcrpy/rgrid.pyx":1547
 *                                 ze = ze if ze < z2 else z2
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1548
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 * 
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1549
 *                                 dlx = xe - x
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1551
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1551, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1551, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1552
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1552, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1553
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1555
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 z = ze
 *                                 if abs(z-grz[iz+1]) < small:
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1556
 * 
 *                                 x = xe
 *                                 z = ze             # <<<<<<<<<<<<<<
 *                                 if abs(z-grz[iz+1]) < small:
 *                                     iz += 1
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1557
 *                                 x = xe
 *                                 z = ze
 *                                 if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iz += 1
 *                         else: # down
 */
              __pyx_t_11 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1557, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1558
 *                                 z = ze
 *                                 if abs(z-grz[iz+1]) < small:
 *                                     iz += 1             # <<<<<<<<<<<<<<
 *                         else: # down
 *                             while z > zi and z > z2:
 */
                __pyx_v_iz = (__pyx_v_iz + 1);

                /* "ttcrpy/rgrid.pyx":1557
 *                                 x = xe
 *                                 z = ze
 *                                 if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iz += 1
 *                         else: # down
 */
              }
            }

            /* "ttcrpy/rgrid.pyx":1540
 *                         zi = m_z*grx[ix+1] + b_z
 * 
 *                         if up_z:             # <<<<<<<<<<<<<<
 *                             while z < zi and z < z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
            goto __pyx_L70;
          }

          /* "ttcrpy/rgrid.pyx":1560
 *                                     iz += 1
 *                         else: # down
 *                             while z > zi and z > z2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
          /*else*/ {
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1560, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1560, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1560, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L78_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_z > __pyx_v_z2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L78_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1561
 *                         else: # down
 *                             while z > zi and z > z2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1563
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze > z2 else z2
 *                                 xe = (ze-b_z) / m_z
 */
              __pyx_t_11 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1563, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1563, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1563, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1563, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = __pyx_v_iz;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1563, __pyx_L1_error)
                }
                __pyx_t_10 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L1_error)
                __pyx_t_10 = __pyx_t_17;
              }
              __pyx_v_ze = __pyx_t_10;

              /* "ttcrpy/rgrid.pyx":1564
 * 
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 *                                 ze = ze if ze > z2 else z2             # <<<<<<<<<<<<<<
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x
 */
              if (((__pyx_v_ze > __pyx_v_z2) != 0)) {
                __pyx_t_10 = __pyx_v_ze;
              } else {
                __pyx_t_10 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_10;

              /* "ttcrpy/rgrid.pyx":1565
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 *                                 ze = ze if ze > z2 else z2
 *                                 xe = (ze-b_z) / m_z             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dlz = ze - z
 */
              __pyx_t_10 = (__pyx_v_ze - __pyx_v_b_z);
              if (unlikely(__pyx_v_m_z == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1565, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_10 / __pyx_v_m_z);

              /* "ttcrpy/rgrid.pyx":1566
 *                                 ze = ze if ze > z2 else z2
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1567
 *                                 xe = (ze-b_z) / m_z
 *                                 dlx = xe - x
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 * 
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1568
 *                                 dlx = xe - x
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1570
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1570, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1571
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1571, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1572
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1574
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1575
 * 
 *                                 x = xe
 *                                 z = ze             # <<<<<<<<<<<<<<
 *                                 if abs(z-grz[iz]) < small:
 *                                     iz -= 1
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1576
 *                                 x = xe
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                     iz -= 1
 * 
 */
              __pyx_t_11 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1576, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1577
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:
 *                                     iz -= 1             # <<<<<<<<<<<<<<
 * 
 *                         ix += 1
 */
                __pyx_v_iz = (__pyx_v_iz - 1);

                /* "ttcrpy/rgrid.pyx":1576
 *                                 x = xe
 *                                 z = ze
 *                                 if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                     iz -= 1
 * 
 */
              }
            }
          }
          __pyx_L70:;

          /* "ttcrpy/rgrid.pyx":1579
 *                                     iz -= 1
 * 
 *                         ix += 1             # <<<<<<<<<<<<<<
 *                         x = grx[ix]
 *                 elif abs(n) < small:
 */
          __pyx_v_ix = (__pyx_v_ix + 1);

          /* "ttcrpy/rgrid.pyx":1580
 * 
 *                         ix += 1
 *                         x = grx[ix]             # <<<<<<<<<<<<<<
 *                 elif abs(n) < small:
 *                     # seul Z constant
 */
          __pyx_t_11 = __pyx_v_ix;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1580, __pyx_L1_error)
          }
          __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides));
        }

        /* "ttcrpy/rgrid.pyx":1530
 *                         x = grx[ix]
 * 
 *                 elif abs(m) < small:             # <<<<<<<<<<<<<<
 *                     # seul Y constant
 *                     m_z = (z2-z1) / (x2-x1)
 */
        goto __pyx_L63;
      }

      /* "ttcrpy/rgrid.pyx":1581
 *                         ix += 1
 *                         x = grx[ix]
 *                 elif abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # seul Z constant
 *                     m_y = (y2-y1) / (x2-x1)
 */
      __pyx_t_9 = ((fabs(__pyx_v_n) < __pyx_v_small) != 0);
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":1583
 *                 elif abs(n) < small:
 *                     # seul Z constant
 *                     m_y = (y2-y1) / (x2-x1)             # <<<<<<<<<<<<<<
 *                     b_y = y2 - m_y*x2
 *                     up_y = m_y > 0
 */
        __pyx_t_10 = (__pyx_v_y2 - __pyx_v_y1);
        __pyx_t_17 = (__pyx_v_x2 - __pyx_v_x1);
        if (unlikely(__pyx_t_17 == 0)) {
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");
          __PYX_ERR(0, 1583, __pyx_L1_error)
        }
        __pyx_v_m_y = (__pyx_t_10 / __pyx_t_17);

        /* "ttcrpy/rgrid.pyx":1584
 *                     # seul Z constant
 *                     m_y = (y2-y1) / (x2-x1)
 *                     b_y = y2 - m_y*x2             # <<<<<<<<<<<<<<
 *                     up_y = m_y > 0
 * 
 */
        __pyx_v_b_y = (__pyx_v_y2 - (__pyx_v_m_y * __pyx_v_x2));

        /* "ttcrpy/rgrid.pyx":1585
 *                     m_y = (y2-y1) / (x2-x1)
 *                     b_y = y2 - m_y*x2
 *                     up_y = m_y > 0             # <<<<<<<<<<<<<<
 * 
 *                     while x < x2:
 */
        __pyx_v_up_y = (__pyx_v_m_y > 0.0);

        /* "ttcrpy/rgrid.pyx":1587
 *                     up_y = m_y > 0
 * 
 *                     while x < x2:             # <<<<<<<<<<<<<<
 * 
 *                         yi = m_y*grx[ix+1] + b_y
 */
        while (1) {
          __pyx_t_9 = ((__pyx_v_x < __pyx_v_x2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1589
 *                     while x < x2:
 * 
 *                         yi = m_y*grx[ix+1] + b_y             # <<<<<<<<<<<<<<
 * 
 *                         if up_y:
 */
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1589, __pyx_L1_error)
          }
          __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m_y * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides))) + __pyx_v_b_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1589, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_yi, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1591
 *                         yi = m_y*grx[ix+1] + b_y
 * 
 *                         if up_y:             # <<<<<<<<<<<<<<
 *                             while y < yi and y < y2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
          __pyx_t_9 = (__pyx_v_up_y != 0);
          if (__pyx_t_9) {

            /* "ttcrpy/rgrid.pyx":1592
 * 
 *                         if up_y:
 *                             while y < yi and y < y2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1592, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_yi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1592, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1592, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L86_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_y < __pyx_v_y2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L86_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1593
 *                         if up_y:
 *                             while y < yi and y < y2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1595
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi             # <<<<<<<<<<<<<<
 *                                 ye = ye if ye < y2 else y2
 *                                 xe = (ye-b_y) / m_y
 */
              __pyx_t_11 = (__pyx_v_iy + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1595, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1595, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_yi, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1595, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1595, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = (__pyx_v_iy + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1595, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_yi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1595, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1596
 * 
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 *                                 ye = ye if ye < y2 else y2             # <<<<<<<<<<<<<<
 *                                 xe = (ye-b_y) / m_y
 *                                 dlx = xe - x
 */
              if (((__pyx_v_ye < __pyx_v_y2) != 0)) {
                __pyx_t_17 = __pyx_v_ye;
              } else {
                __pyx_t_17 = __pyx_v_y2;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1597
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 *                                 ye = ye if ye < y2 else y2
 *                                 xe = (ye-b_y) / m_y             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dly = ye - y
 */
              __pyx_t_17 = (__pyx_v_ye - __pyx_v_b_y);
              if (unlikely(__pyx_v_m_y == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1597, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_17 / __pyx_v_m_y);

              /* "ttcrpy/rgrid.pyx":1598
 *                                 ye = ye if ye < y2 else y2
 *                                 xe = (ye-b_y) / m_y
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1599
 *                                 xe = (ye-b_y) / m_y
 *                                 dlx = xe - x
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 * 
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1600
 *                                 dlx = xe - x
 *                                 dly = ye - y
 *                                 dl = sqrt( dlx*dlx + dly*dly )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dly * __pyx_v_dly)));

              /* "ttcrpy/rgrid.pyx":1602
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1602, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1602, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1603
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1603, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1603, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1604
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1606
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 y = ye
 *                                 if abs(y-gry[iy+1]) < small:
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1607
 * 
 *                                 x = xe
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 if abs(y-gry[iy+1]) < small:
 *                                     iy += 1
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1608
 *                                 x = xe
 *                                 y = ye
 *                                 if abs(y-gry[iy+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iy += 1
 *                         else:  # down
 */
              __pyx_t_11 = (__pyx_v_iy + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1608, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_y - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1609
 *                                 y = ye
 *                                 if abs(y-gry[iy+1]) < small:
 *                                     iy += 1             # <<<<<<<<<<<<<<
 *                         else:  # down
 *                             while y > yi and y > y2:
 */
                __pyx_v_iy = (__pyx_v_iy + 1);

                /* "ttcrpy/rgrid.pyx":1608
 *                                 x = xe
 *                                 y = ye
 *                                 if abs(y-gry[iy+1]) < small:             # <<<<<<<<<<<<<<
 *                                     iy += 1
 *                         else:  # down
 */
              }
            }

            /* "ttcrpy/rgrid.pyx":1591
 *                         yi = m_y*grx[ix+1] + b_y
 * 
 *                         if up_y:             # <<<<<<<<<<<<<<
 *                             while y < yi and y < y2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 */
            goto __pyx_L83;
          }

          /* "ttcrpy/rgrid.pyx":1611
 *                                     iy += 1
 *                         else:  # down
 *                             while y > yi and y > y2:             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
          /*else*/ {
            while (1) {
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1611, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_yi, Py_GT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1611, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1611, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_15) {
              } else {
                __pyx_t_9 = __pyx_t_15;
                goto __pyx_L91_bool_binop_done;
              }
              __pyx_t_15 = ((__pyx_v_y > __pyx_v_y2) != 0);
              __pyx_t_9 = __pyx_t_15;
              __pyx_L91_bool_binop_done:;
              if (!__pyx_t_9) break;

              /* "ttcrpy/rgrid.pyx":1612
 *                         else:  # down
 *                             while y > yi and y > y2:
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1614
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 ye = gry[iy] if gry[iy] > yi else yi             # <<<<<<<<<<<<<<
 *                                 ye = ye if ye > y2 else y2
 *                                 xe = (ye-b_y)/m_y
 */
              __pyx_t_11 = __pyx_v_iy;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1614, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1614, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_yi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1614, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1614, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = __pyx_v_iy;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1614, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_yi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1615
 * 
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 *                                 ye = ye if ye > y2 else y2             # <<<<<<<<<<<<<<
 *                                 xe = (ye-b_y)/m_y
 *                                 dlx = xe - x
 */
              if (((__pyx_v_ye > __pyx_v_y2) != 0)) {
                __pyx_t_17 = __pyx_v_ye;
              } else {
                __pyx_t_17 = __pyx_v_y2;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1616
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 *                                 ye = ye if ye > y2 else y2
 *                                 xe = (ye-b_y)/m_y             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dly = ye - y
 */
              __pyx_t_17 = (__pyx_v_ye - __pyx_v_b_y);
              if (unlikely(__pyx_v_m_y == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1616, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_17 / __pyx_v_m_y);

              /* "ttcrpy/rgrid.pyx":1617
 *                                 ye = ye if ye > y2 else y2
 *                                 xe = (ye-b_y)/m_y
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1618
 *                                 xe = (ye-b_y)/m_y
 *                                 dlx = xe - x
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 * 
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1619
 *                                 dlx = xe - x
 *                                 dly = ye - y
 *                                 dl = sqrt( dlx*dlx + dly*dly )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dly * __pyx_v_dly)));

              /* "ttcrpy/rgrid.pyx":1621
 *                                 dl = sqrt( dlx*dlx + dly*dly )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1621, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1621, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1622
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1622, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1622, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1623
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1625
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 y = ye
 *                                 if abs(y-gry[iy]) < small:
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1626
 * 
 *                                 x = xe
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 if abs(y-gry[iy]) < small:
 *                                     iy -= 1
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1627
 *                                 x = xe
 *                                 y = ye
 *                                 if abs(y-gry[iy]) < small:             # <<<<<<<<<<<<<<
 *                                     iy -= 1
 * 
 */
              __pyx_t_11 = __pyx_v_iy;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1627, __pyx_L1_error)
              }
              __pyx_t_9 = ((fabs((__pyx_v_y - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides)))) < __pyx_v_small) != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1628
 *                                 y = ye
 *                                 if abs(y-gry[iy]) < small:
 *                                     iy -= 1             # <<<<<<<<<<<<<<
 * 
 *                         ix += 1
 */
                __pyx_v_iy = (__pyx_v_iy - 1);

                /* "ttcrpy/rgrid.pyx":1627
 *                                 x = xe
 *                                 y = ye
 *                                 if abs(y-gry[iy]) < small:             # <<<<<<<<<<<<<<
 *                                     iy -= 1
 * 
 */
              }
            }
          }
          __pyx_L83:;

          /* "ttcrpy/rgrid.pyx":1630
 *                                     iy -= 1
 * 
 *                         ix += 1             # <<<<<<<<<<<<<<
 *                         x = grx[ix]
 *                 else:
 */
          __pyx_v_ix = (__pyx_v_ix + 1);

          /* "ttcrpy/rgrid.pyx":1631
 * 
 *                         ix += 1
 *                         x = grx[ix]             # <<<<<<<<<<<<<<
 *                 else:
 *                     while x < x2:
 */
          __pyx_t_11 = __pyx_v_ix;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1631, __pyx_L1_error)
          }
          __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides));
        }

        /* "ttcrpy/rgrid.pyx":1581
 *                         ix += 1
 *                         x = grx[ix]
 *                 elif abs(n) < small:             # <<<<<<<<<<<<<<
 *                     # seul Z constant
 *                     m_y = (y2-y1) / (x2-x1)
 */
        goto __pyx_L63;
      }

      /* "ttcrpy/rgrid.pyx":1633
 *                         x = grx[ix]
 *                 else:
 *                     while x < x2:             # <<<<<<<<<<<<<<
 * 
 *                         m_y = (y2-y1) / (x2-x1)
 */
      /*else*/ {
        while (1) {
          __pyx_t_9 = ((__pyx_v_x < __pyx_v_x2) != 0);
          if (!__pyx_t_9) break;

          /* "ttcrpy/rgrid.pyx":1635
 *                     while x < x2:
 * 
 *                         m_y = (y2-y1) / (x2-x1)             # <<<<<<<<<<<<<<
 *                         b_y = y2 - m_y*x2
 *                         up_y = m_y > 0
 */
          __pyx_t_17 = (__pyx_v_y2 - __pyx_v_y1);
          __pyx_t_10 = (__pyx_v_x2 - __pyx_v_x1);
          if (unlikely(__pyx_t_10 == 0)) {
            PyErr_SetString(PyExc_ZeroDivisionError, "float division");
            __PYX_ERR(0, 1635, __pyx_L1_error)
          }
          __pyx_v_m_y = (__pyx_t_17 / __pyx_t_10);

          /* "ttcrpy/rgrid.pyx":1636
 * 
 *                         m_y = (y2-y1) / (x2-x1)
 *                         b_y = y2 - m_y*x2             # <<<<<<<<<<<<<<
 *                         up_y = m_y > 0
 * 
 */
          __pyx_v_b_y = (__pyx_v_y2 - (__pyx_v_m_y * __pyx_v_x2));

          /* "ttcrpy/rgrid.pyx":1637
 *                         m_y = (y2-y1) / (x2-x1)
 *                         b_y = y2 - m_y*x2
 *                         up_y = m_y > 0             # <<<<<<<<<<<<<<
 * 
 *                         m_z = (z2-z1) / (x2-x1)
 */
          __pyx_v_up_y = (__pyx_v_m_y > 0.0);

          /* "ttcrpy/rgrid.pyx":1639
 *                         up_y = m_y > 0
 * 
 *                         m_z = (z2-z1) / (x2-x1)             # <<<<<<<<<<<<<<
 *                         b_z = z2 - m_z*x2
 *                         up_z = m_z > 0
 */
          __pyx_t_10 = (__pyx_v_z2 - __pyx_v_z1);
          __pyx_t_17 = (__pyx_v_x2 - __pyx_v_x1);
          if (unlikely(__pyx_t_17 == 0)) {
            PyErr_SetString(PyExc_ZeroDivisionError, "float division");
            __PYX_ERR(0, 1639, __pyx_L1_error)
          }
          __pyx_v_m_z = (__pyx_t_10 / __pyx_t_17);

          /* "ttcrpy/rgrid.pyx":1640
 * 
 *                         m_z = (z2-z1) / (x2-x1)
 *                         b_z = z2 - m_z*x2             # <<<<<<<<<<<<<<
 *                         up_z = m_z > 0
 * 
 */
          __pyx_v_b_z = (__pyx_v_z2 - (__pyx_v_m_z * __pyx_v_x2));

          /* "ttcrpy/rgrid.pyx":1641
 *                         m_z = (z2-z1) / (x2-x1)
 *                         b_z = z2 - m_z*x2
 *                         up_z = m_z > 0             # <<<<<<<<<<<<<<
 * 
 *                         yi = m_y*grx[ix+1] + b_y
 */
          __pyx_v_up_z = (__pyx_v_m_z > 0.0);

          /* "ttcrpy/rgrid.pyx":1643
 *                         up_z = m_z > 0
 * 
 *                         yi = m_y*grx[ix+1] + b_y             # <<<<<<<<<<<<<<
 *                         zi = m_z*grx[ix+1] + b_z
 * 
 */
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1643, __pyx_L1_error)
          }
          __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m_y * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides))) + __pyx_v_b_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1643, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_yi, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1644
 * 
 *                         yi = m_y*grx[ix+1] + b_y
 *                         zi = m_z*grx[ix+1] + b_z             # <<<<<<<<<<<<<<
 * 
 *                         while (sy*(yi-y)) > 0 and (sy*(y2-y)) > 0 and \
 */
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1644, __pyx_L1_error)
          }
          __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m_z * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides))) + __pyx_v_b_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_zi, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":1646
 *                         zi = m_z*grx[ix+1] + b_z
 * 
 *                         while (sy*(yi-y)) > 0 and (sy*(y2-y)) > 0 and \             # <<<<<<<<<<<<<<
 * 							  (sz*(zi-z)) > 0 and (sz*(z2-z)) > 0:
 * 
 */
          while (1) {
            __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_sy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_18 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_19 = PyNumber_Subtract(__pyx_v_yi, __pyx_t_18); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_18 = PyNumber_Multiply(__pyx_t_1, __pyx_t_19); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __pyx_t_19 = PyObject_RichCompare(__pyx_t_18, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_19); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_19); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1646, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L98_bool_binop_done;
            }
            __pyx_t_15 = (((__pyx_v_sy * (__pyx_v_y2 - __pyx_v_y)) > 0.0) != 0);
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L98_bool_binop_done;
            }

            /* "ttcrpy/rgrid.pyx":1647
 * 
 *                         while (sy*(yi-y)) > 0 and (sy*(y2-y)) > 0 and \
 * 							  (sz*(zi-z)) > 0 and (sz*(z2-z)) > 0:             # <<<<<<<<<<<<<<
 * 
 *                             if up_y:
 */
            __pyx_t_19 = __Pyx_PyInt_From_int(__pyx_v_sz); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_18 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_1 = PyNumber_Subtract(__pyx_v_zi, __pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_18 = PyNumber_Multiply(__pyx_t_19, __pyx_t_1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1647, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L98_bool_binop_done;
            }
            __pyx_t_15 = (((__pyx_v_sz * (__pyx_v_z2 - __pyx_v_z)) > 0.0) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L98_bool_binop_done:;
            if (!__pyx_t_9) break;

            /* "ttcrpy/rgrid.pyx":1649
 * 							  (sz*(zi-z)) > 0 and (sz*(z2-z)) > 0:
 * 
 *                             if up_y:             # <<<<<<<<<<<<<<
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 *                                 ye = ye if ye<y2 else y2
 */
            __pyx_t_9 = (__pyx_v_up_y != 0);
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1650
 * 
 *                             if up_y:
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi             # <<<<<<<<<<<<<<
 *                                 ye = ye if ye<y2 else y2
 *                             else:
 */
              __pyx_t_11 = (__pyx_v_iy + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1650, __pyx_L1_error)
              }
              __pyx_t_1 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1650, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_yi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1650, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1650, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = (__pyx_v_iy + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1650, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_yi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1651
 *                             if up_y:
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 *                                 ye = ye if ye<y2 else y2             # <<<<<<<<<<<<<<
 *                             else:
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 */
              if (((__pyx_v_ye < __pyx_v_y2) != 0)) {
                __pyx_t_17 = __pyx_v_ye;
              } else {
                __pyx_t_17 = __pyx_v_y2;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1649
 * 							  (sz*(zi-z)) > 0 and (sz*(z2-z)) > 0:
 * 
 *                             if up_y:             # <<<<<<<<<<<<<<
 *                                 ye = gry[iy+1] if gry[iy+1] < yi else yi
 *                                 ye = ye if ye<y2 else y2
 */
              goto __pyx_L102;
            }

            /* "ttcrpy/rgrid.pyx":1653
 *                                 ye = ye if ye<y2 else y2
 *                             else:
 *                                 ye = gry[iy] if gry[iy] > yi else yi             # <<<<<<<<<<<<<<
 *                                 ye = ye if ye>y2 else y2
 *                             if up_z:
 */
            /*else*/ {
              __pyx_t_11 = __pyx_v_iy;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1653, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1653, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_yi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1653, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1653, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = __pyx_v_iy;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1653, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_yi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1653, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ye = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1654
 *                             else:
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 *                                 ye = ye if ye>y2 else y2             # <<<<<<<<<<<<<<
 *                             if up_z:
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 */
              if (((__pyx_v_ye > __pyx_v_y2) != 0)) {
                __pyx_t_17 = __pyx_v_ye;
              } else {
                __pyx_t_17 = __pyx_v_y2;
              }
              __pyx_v_ye = __pyx_t_17;
            }
            __pyx_L102:;

            /* "ttcrpy/rgrid.pyx":1655
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 *                                 ye = ye if ye>y2 else y2
 *                             if up_z:             # <<<<<<<<<<<<<<
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 *                                 ze = ze if ze < z2 else z2
 */
            __pyx_t_9 = (__pyx_v_up_z != 0);
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1656
 *                                 ye = ye if ye>y2 else y2
 *                             if up_z:
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze < z2 else z2
 *                             else:
 */
              __pyx_t_11 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1656, __pyx_L1_error)
              }
              __pyx_t_1 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1656, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1656, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = (__pyx_v_iz + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1656, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1656, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1657
 *                             if up_z:
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 *                                 ze = ze if ze < z2 else z2             # <<<<<<<<<<<<<<
 *                             else:
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 */
              if (((__pyx_v_ze < __pyx_v_z2) != 0)) {
                __pyx_t_17 = __pyx_v_ze;
              } else {
                __pyx_t_17 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1655
 *                                 ye = gry[iy] if gry[iy] > yi else yi
 *                                 ye = ye if ye>y2 else y2
 *                             if up_z:             # <<<<<<<<<<<<<<
 *                                 ze = grz[iz+1] if grz[iz+1] < zi else zi
 *                                 ze = ze if ze < z2 else z2
 */
              goto __pyx_L103;
            }

            /* "ttcrpy/rgrid.pyx":1659
 *                                 ze = ze if ze < z2 else z2
 *                             else:
 *                                 ze = grz[iz] if grz[iz] > zi else zi             # <<<<<<<<<<<<<<
 *                                 ze = ze if ze > z2 else z2
 * 
 */
            /*else*/ {
              __pyx_t_11 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 1659, __pyx_L1_error)
              }
              __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1659, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1659, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1659, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_9) {
                __pyx_t_11 = __pyx_v_iz;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1659, __pyx_L1_error)
                }
                __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
              } else {
                __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1659, __pyx_L1_error)
                __pyx_t_17 = __pyx_t_10;
              }
              __pyx_v_ze = __pyx_t_17;

              /* "ttcrpy/rgrid.pyx":1660
 *                             else:
 *                                 ze = grz[iz] if grz[iz] > zi else zi
 *                                 ze = ze if ze > z2 else z2             # <<<<<<<<<<<<<<
 * 
 *                             if (ze-b_z)/m_z < (ye-b_y)/m_y:  # we cross z before y
 */
              if (((__pyx_v_ze > __pyx_v_z2) != 0)) {
                __pyx_t_17 = __pyx_v_ze;
              } else {
                __pyx_t_17 = __pyx_v_z2;
              }
              __pyx_v_ze = __pyx_t_17;
            }
            __pyx_L103:;

            /* "ttcrpy/rgrid.pyx":1662
 *                                 ze = ze if ze > z2 else z2
 * 
 *                             if (ze-b_z)/m_z < (ye-b_y)/m_y:  # we cross z before y             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
            __pyx_t_17 = (__pyx_v_ze - __pyx_v_b_z);
            if (unlikely(__pyx_v_m_z == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 1662, __pyx_L1_error)
            }
            __pyx_t_10 = (__pyx_v_ye - __pyx_v_b_y);
            if (unlikely(__pyx_v_m_y == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 1662, __pyx_L1_error)
            }
            __pyx_t_9 = (((__pyx_t_17 / __pyx_v_m_z) < (__pyx_t_10 / __pyx_v_m_y)) != 0);
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":1663
 * 
 *                             if (ze-b_z)/m_z < (ye-b_y)/m_y:  # we cross z before y
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 xe = (ze-b_z) / m_z
 */
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1665
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 xe = (ze-b_z) / m_z             # <<<<<<<<<<<<<<
 *                                 ye = m_y*xe + b_y
 *                                 dlx = xe - x
 */
              __pyx_t_10 = (__pyx_v_ze - __pyx_v_b_z);
              if (unlikely(__pyx_v_m_z == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1665, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_10 / __pyx_v_m_z);

              /* "ttcrpy/rgrid.pyx":1666
 * 
 *                                 xe = (ze-b_z) / m_z
 *                                 ye = m_y*xe + b_y             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dly = ye - y
 */
              __pyx_v_ye = ((__pyx_v_m_y * __pyx_v_xe) + __pyx_v_b_y);

              /* "ttcrpy/rgrid.pyx":1667
 *                                 xe = (ze-b_z) / m_z
 *                                 ye = m_y*xe + b_y
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dlz = ze - z
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1668
 *                                 ye = m_y*xe + b_y
 *                                 dlx = xe - x
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1669
 *                                 dlx = xe - x
 *                                 dly = ye - y
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )
 * 
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1670
 *                                 dly = ye - y
 *                                 dlz = ze - z
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt((((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dly * __pyx_v_dly)) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1672
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1672, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1673
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1673, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1673, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1674
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1676
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 y = ye
 *                                 z = ze
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1677
 * 
 *                                 x = xe
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 z = ze
 *                                 if up_z:
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1678
 *                                 x = xe
 *                                 y = ye
 *                                 z = ze             # <<<<<<<<<<<<<<
 *                                 if up_z:
 *                                     if abs(z-grz[iz+1]) < small:
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1679
 *                                 y = ye
 *                                 z = ze
 *                                 if up_z:             # <<<<<<<<<<<<<<
 *                                     if abs(z-grz[iz+1]) < small:
 *                                         iz += 1
 */
              __pyx_t_9 = (__pyx_v_up_z != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1680
 *                                 z = ze
 *                                 if up_z:
 *                                     if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                         iz += 1
 *                                 else:
 */
                __pyx_t_11 = (__pyx_v_iz + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1680, __pyx_L1_error)
                }
                __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
                if (__pyx_t_9) {

                  /* "ttcrpy/rgrid.pyx":1681
 *                                 if up_z:
 *                                     if abs(z-grz[iz+1]) < small:
 *                                         iz += 1             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     if abs(z-grz[iz]) < small:
 */
                  __pyx_v_iz = (__pyx_v_iz + 1);

                  /* "ttcrpy/rgrid.pyx":1680
 *                                 z = ze
 *                                 if up_z:
 *                                     if abs(z-grz[iz+1]) < small:             # <<<<<<<<<<<<<<
 *                                         iz += 1
 *                                 else:
 */
                }

                /* "ttcrpy/rgrid.pyx":1679
 *                                 y = ye
 *                                 z = ze
 *                                 if up_z:             # <<<<<<<<<<<<<<
 *                                     if abs(z-grz[iz+1]) < small:
 *                                         iz += 1
 */
                goto __pyx_L105;
              }

              /* "ttcrpy/rgrid.pyx":1683
 *                                         iz += 1
 *                                 else:
 *                                     if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                         iz -= 1
 * 
 */
              /*else*/ {
                __pyx_t_11 = __pyx_v_iz;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1683, __pyx_L1_error)
                }
                __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
                if (__pyx_t_9) {

                  /* "ttcrpy/rgrid.pyx":1684
 *                                 else:
 *                                     if abs(z-grz[iz]) < small:
 *                                         iz -= 1             # <<<<<<<<<<<<<<
 * 
 *                             else:  # we cross y before z
 */
                  __pyx_v_iz = (__pyx_v_iz - 1);

                  /* "ttcrpy/rgrid.pyx":1683
 *                                         iz += 1
 *                                 else:
 *                                     if abs(z-grz[iz]) < small:             # <<<<<<<<<<<<<<
 *                                         iz -= 1
 * 
 */
                }
              }
              __pyx_L105:;

              /* "ttcrpy/rgrid.pyx":1662
 *                                 ze = ze if ze > z2 else z2
 * 
 *                             if (ze-b_z)/m_z < (ye-b_y)/m_y:  # we cross z before y             # <<<<<<<<<<<<<<
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 */
              goto __pyx_L104;
            }

            /* "ttcrpy/rgrid.pyx":1687
 * 
 *                             else:  # we cross y before z
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                                 xe = (ye-b_y)/m_y
 */
            /*else*/ {
              __pyx_v_iCell = ((((__pyx_v_ix * (__pyx_v_n_gry - 1)) + __pyx_v_iy) * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

              /* "ttcrpy/rgrid.pyx":1689
 *                                 iCell = (ix*(n_gry-1)+iy)*(n_grz-1) + iz
 * 
 *                                 xe = (ye-b_y)/m_y             # <<<<<<<<<<<<<<
 *                                 ze = m_z*xe + b_z
 *                                 dlx = xe - x
 */
              __pyx_t_10 = (__pyx_v_ye - __pyx_v_b_y);
              if (unlikely(__pyx_v_m_y == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1689, __pyx_L1_error)
              }
              __pyx_v_xe = (__pyx_t_10 / __pyx_v_m_y);

              /* "ttcrpy/rgrid.pyx":1690
 * 
 *                                 xe = (ye-b_y)/m_y
 *                                 ze = m_z*xe + b_z             # <<<<<<<<<<<<<<
 *                                 dlx = xe - x
 *                                 dly = ye - y
 */
              __pyx_v_ze = ((__pyx_v_m_z * __pyx_v_xe) + __pyx_v_b_z);

              /* "ttcrpy/rgrid.pyx":1691
 *                                 xe = (ye-b_y)/m_y
 *                                 ze = m_z*xe + b_z
 *                                 dlx = xe - x             # <<<<<<<<<<<<<<
 *                                 dly = ye - y
 *                                 dlz = ze - z
 */
              __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

              /* "ttcrpy/rgrid.pyx":1692
 *                                 ze = m_z*xe + b_z
 *                                 dlx = xe - x
 *                                 dly = ye - y             # <<<<<<<<<<<<<<
 *                                 dlz = ze - z
 * 
 */
              __pyx_v_dly = (__pyx_v_ye - __pyx_v_y);

              /* "ttcrpy/rgrid.pyx":1693
 *                                 dlx = xe - x
 *                                 dly = ye - y
 *                                 dlz = ze - z             # <<<<<<<<<<<<<<
 * 
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )
 */
              __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

              /* "ttcrpy/rgrid.pyx":1695
 *                                 dlz = ze - z
 * 
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )             # <<<<<<<<<<<<<<
 * 
 *                                 indices_p.append(iCell)
 */
              __pyx_v_dl = sqrt((((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dly * __pyx_v_dly)) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":1697
 *                                 dl = sqrt( dlx*dlx + dly*dly + dlz*dlz )
 * 
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1697, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1697, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1698
 * 
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1698, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1698, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":1699
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                                 x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":1701
 *                                 k += 1
 * 
 *                                 x = xe             # <<<<<<<<<<<<<<
 *                                 y = ye
 *                                 z = ze
 */
              __pyx_v_x = __pyx_v_xe;

              /* "ttcrpy/rgrid.pyx":1702
 * 
 *                                 x = xe
 *                                 y = ye             # <<<<<<<<<<<<<<
 *                                 z = ze
 *                                 if up_y:
 */
              __pyx_v_y = __pyx_v_ye;

              /* "ttcrpy/rgrid.pyx":1703
 *                                 x = xe
 *                                 y = ye
 *                                 z = ze             # <<<<<<<<<<<<<<
 *                                 if up_y:
 *                                     if abs(y-gry[iy+1]) < small:
 */
              __pyx_v_z = __pyx_v_ze;

              /* "ttcrpy/rgrid.pyx":1704
 *                                 y = ye
 *                                 z = ze
 *                                 if up_y:             # <<<<<<<<<<<<<<
 *                                     if abs(y-gry[iy+1]) < small:
 *                                         iy += 1
 */
              __pyx_t_9 = (__pyx_v_up_y != 0);
              if (__pyx_t_9) {

                /* "ttcrpy/rgrid.pyx":1705
 *                                 z = ze
 *                                 if up_y:
 *                                     if abs(y-gry[iy+1]) < small:             # <<<<<<<<<<<<<<
 *                                         iy += 1
 *                                 else:
 */
                __pyx_t_11 = (__pyx_v_iy + 1);
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1705, __pyx_L1_error)
                }
                __pyx_t_9 = ((fabs((__pyx_v_y - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides)))) < __pyx_v_small) != 0);
                if (__pyx_t_9) {

                  /* "ttcrpy/rgrid.pyx":1706
 *                                 if up_y:
 *                                     if abs(y-gry[iy+1]) < small:
 *                                         iy += 1             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     if abs(y-gry[iy]) < small:
 */
                  __pyx_v_iy = (__pyx_v_iy + 1);

                  /* "ttcrpy/rgrid.pyx":1705
 *                                 z = ze
 *                                 if up_y:
 *                                     if abs(y-gry[iy+1]) < small:             # <<<<<<<<<<<<<<
 *                                         iy += 1
 *                                 else:
 */
                }

                /* "ttcrpy/rgrid.pyx":1704
 *                                 y = ye
 *                                 z = ze
 *                                 if up_y:             # <<<<<<<<<<<<<<
 *                                     if abs(y-gry[iy+1]) < small:
 *                                         iy += 1
 */
                goto __pyx_L108;
              }

              /* "ttcrpy/rgrid.pyx":1708
 *                                         iy += 1
 *                                 else:
 *                                     if abs(y-gry[iy]) < small:             # <<<<<<<<<<<<<<
 *                                         iy -= 1
 * 
 */
              /*else*/ {
                __pyx_t_11 = __pyx_v_iy;
                __pyx_t_8 = -1;
                if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_gry.diminfo[0].shape)) __pyx_t_8 = 0;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  __PYX_ERR(0, 1708, __pyx_L1_error)
                }
                __pyx_t_9 = ((fabs((__pyx_v_y - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_gry.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_gry.diminfo[0].strides)))) < __pyx_v_small) != 0);
                if (__pyx_t_9) {

                  /* "ttcrpy/rgrid.pyx":1709
 *                                 else:
 *                                     if abs(y-gry[iy]) < small:
 *                                         iy -= 1             # <<<<<<<<<<<<<<
 * 
 *                         ix += 1
 */
                  __pyx_v_iy = (__pyx_v_iy - 1);

                  /* "ttcrpy/rgrid.pyx":1708
 *                                         iy += 1
 *                                 else:
 *                                     if abs(y-gry[iy]) < small:             # <<<<<<<<<<<<<<
 *                                         iy -= 1
 * 
 */
                }
              }
              __pyx_L108:;
            }
            __pyx_L104:;
          }

          /* "ttcrpy/rgrid.pyx":1711
 *                                         iy -= 1
 * 
 *                         ix += 1             # <<<<<<<<<<<<<<
 *                         x = grx[ix]
 * 
 */
          __pyx_v_ix = (__pyx_v_ix + 1);

          /* "ttcrpy/rgrid.pyx":1712
 * 
 *                         ix += 1
 *                         x = grx[ix]             # <<<<<<<<<<<<<<
 * 
 *         indptr_p.append(k)
 */
          __pyx_t_11 = __pyx_v_ix;
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 1712, __pyx_L1_error)
          }
          __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides));
        }
      }
      __pyx_L63:;
    }
    __pyx_L21:;
  }

  /* "ttcrpy/rgrid.pyx":1714
 *                         x = grx[ix]
 * 
 *         indptr_p.append(k)             # <<<<<<<<<<<<<<
 *         L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indptr_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1715
 * 
 *         indptr_p.append(k)
 *         L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data_p);
  __Pyx_GIVEREF(__pyx_v_data_p);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data_p);
  __Pyx_INCREF(__pyx_v_indices_p);
  __Pyx_GIVEREF(__pyx_v_indices_p);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_indices_p);
  __Pyx_INCREF(__pyx_v_indptr_p);
  __Pyx_GIVEREF(__pyx_v_indptr_p);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_indptr_p);
  __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1716
 *         indptr_p.append(k)
 *         L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))             # <<<<<<<<<<<<<<
 * 
 *         if centers:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = __Pyx_PyInt_FromSize_t(__pyx_v_nTx); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_PyInt_FromSize_t((((__pyx_v_n_grx - 1) * (__pyx_v_n_gry - 1)) * (__pyx_v_n_grz - 1))); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_21);
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_22) < 0) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "ttcrpy/rgrid.pyx":1715
 * 
 *         indptr_p.append(k)
 *         L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))
 * 
 */
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_19, __pyx_t_1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_L = __pyx_t_22;
  __pyx_t_22 = 0;

  /* "ttcrpy/rgrid.pyx":1718
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))
 * 
 *         if centers:             # <<<<<<<<<<<<<<
 *             xc = (grx[1:]+grx[:-1])/2
 *             yc = (gry[1:]+gry[:-1])/2
 */
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_centers); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1718, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "ttcrpy/rgrid.pyx":1719
 * 
 *         if centers:
 *             xc = (grx[1:]+grx[:-1])/2             # <<<<<<<<<<<<<<
 *             yc = (gry[1:]+gry[:-1])/2
 *             zc = (grz[1:]+grz[:-1])/2
 */
    __pyx_t_22 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_grx), __pyx_slice__37); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_grx), __pyx_slice__38); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_19 = PyNumber_Add(__pyx_t_22, __pyx_t_1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_19, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_v_xc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":1720
 *         if centers:
 *             xc = (grx[1:]+grx[:-1])/2
 *             yc = (gry[1:]+gry[:-1])/2             # <<<<<<<<<<<<<<
 *             zc = (grz[1:]+grz[:-1])/2
 *             return L, (xc, yc, zc)
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_gry), __pyx_slice__37); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_19 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_gry), __pyx_slice__38); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_22 = PyNumber_Add(__pyx_t_1, __pyx_t_19); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_19 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_22, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_v_yc = __pyx_t_19;
    __pyx_t_19 = 0;

    /* "ttcrpy/rgrid.pyx":1721
 *             xc = (grx[1:]+grx[:-1])/2
 *             yc = (gry[1:]+gry[:-1])/2
 *             zc = (grz[1:]+grz[:-1])/2             # <<<<<<<<<<<<<<
 *             return L, (xc, yc, zc)
 *         else:
 */
    __pyx_t_19 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_grz), __pyx_slice__37); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_22 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_grz), __pyx_slice__38); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_1 = PyNumber_Add(__pyx_t_19, __pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_zc = __pyx_t_22;
    __pyx_t_22 = 0;

    /* "ttcrpy/rgrid.pyx":1722
 *             yc = (gry[1:]+gry[:-1])/2
 *             zc = (grz[1:]+grz[:-1])/2
 *             return L, (xc, yc, zc)             # <<<<<<<<<<<<<<
 *         else:
 *             return L
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_22 = PyTuple_New(3); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_v_xc);
    __Pyx_GIVEREF(__pyx_v_xc);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_v_xc);
    __Pyx_INCREF(__pyx_v_yc);
    __Pyx_GIVEREF(__pyx_v_yc);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_v_yc);
    __Pyx_INCREF(__pyx_v_zc);
    __Pyx_GIVEREF(__pyx_v_zc);
    PyTuple_SET_ITEM(__pyx_t_22, 2, __pyx_v_zc);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_L);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_22);
    __pyx_t_22 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1718
 *                           shape=(nTx, (n_grx-1)*(n_gry-1)*(n_grz-1)))
 * 
 *         if centers:             # <<<<<<<<<<<<<<
 *             xc = (grx[1:]+grx[:-1])/2
 *             yc = (gry[1:]+gry[:-1])/2
 */
  }

  /* "ttcrpy/rgrid.pyx":1724
 *             return L, (xc, yc, zc)
 *         else:
 *             return L             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_L);
    __pyx_r = __pyx_v_L;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":1290
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gry.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grz.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid3d.data_kernel_straight_rays", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gry.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grz.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_data_p);
  __Pyx_XDECREF(__pyx_v_indices_p);
  __Pyx_XDECREF(__pyx_v_indptr_p);
  __Pyx_XDECREF(__pyx_v_zi);
  __Pyx_XDECREF(__pyx_v_yi);
  __Pyx_XDECREF(__pyx_v_L);
  __Pyx_XDECREF(__pyx_v_xc);
  __Pyx_XDECREF(__pyx_v_yc);
  __Pyx_XDECREF(__pyx_v_zc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1804
 *     cdef Grid2D[double,uint32_t,sxz[double]]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=1] x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.double_t, ndim=1] z, size_t n_threads=1,
 *                   bool cell_slowness=1, str method='SPM', str aniso='iso',
 */

/* Python wrapper */
static int __pyx_pw_6ttcrpy_5rgrid_6Grid2d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ttcrpy_5rgrid_6Grid2d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_z = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  PyObject *__pyx_v_aniso = 0;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  bool __pyx_v_weno;
  bool __pyx_v_rotated_template;
  uint32_t __pyx_v_nsnx;
  uint32_t __pyx_v_nsnz;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_z,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_aniso,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_weno,&__pyx_n_s_rotated_template,&__pyx_n_s_nsnx,&__pyx_n_s_nsnz,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    values[4] = ((PyObject*)__pyx_n_u_SPM);
    values[5] = ((PyObject*)__pyx_n_u_iso);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 12, 1); __PYX_ERR(0, 1804, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aniso);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_weno);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rotated_template);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnx);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsnz);
          if (value) { values[11] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1804, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_z = ((PyArrayObject *)values[1]);
    if (values[2]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1805, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[3]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1806, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[4]);
    __pyx_v_aniso = ((PyObject*)values[5]);
    if (values[6]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1807, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[7]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1807, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[8]) {
      __pyx_v_weno = __Pyx_PyObject_IsTrue(values[8]); if (unlikely((__pyx_v_weno == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1807, __pyx_L3_error)
    } else {
      __pyx_v_weno = ((bool)1);
    }
    if (values[9]) {
      __pyx_v_rotated_template = __Pyx_PyObject_IsTrue(values[9]); if (unlikely((__pyx_v_rotated_template == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1808, __pyx_L3_error)
    } else {
      __pyx_v_rotated_template = ((bool)0);
    }
    if (values[10]) {
      __pyx_v_nsnx = __Pyx_PyInt_As_uint32_t(values[10]); if (unlikely((__pyx_v_nsnx == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1808, __pyx_L3_error)
    } else {
      __pyx_v_nsnx = ((uint32_t)10);
    }
    if (values[11]) {
      __pyx_v_nsnz = __Pyx_PyInt_As_uint32_t(values[11]); if (unlikely((__pyx_v_nsnz == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1808, __pyx_L3_error)
    } else {
      __pyx_v_nsnz = ((uint32_t)10);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1804, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 1804, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 1805, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 1806, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_aniso), (&PyUnicode_Type), 1, "aniso", 1))) __PYX_ERR(0, 1806, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d___cinit__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_x, __pyx_v_z, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_aniso, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_rotated_template, __pyx_v_nsnx, __pyx_v_nsnz);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ttcrpy_5rgrid_6Grid2d___cinit__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_z, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, PyObject *__pyx_v_aniso, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_weno, bool __pyx_v_rotated_template, uint32_t __pyx_v_nsnx, uint32_t __pyx_v_nsnz) {
  uint32_t __pyx_v_nx;
  uint32_t __pyx_v_nz;
  double __pyx_v_xmin;
  double __pyx_v_zmin;
  PyObject *__pyx_v_val = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  double __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d>  *__pyx_t_12;
  ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_e>  *__pyx_t_13;
  ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_te>  *__pyx_t_14;
  ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_p>  *__pyx_t_15;
  ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_h>  *__pyx_t_16;
  ttcr::Grid2Drcfs<double,uint32_t,ttcr::sxz<double> >  *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  ttcr::Grid2Drnsp<double,uint32_t,ttcr::sxz<double> >  *__pyx_t_19;
  ttcr::Grid2Drnfs<double,uint32_t,ttcr::sxz<double> >  *__pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1804, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1804, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "ttcrpy/rgrid.pyx":1810
 *                   bool rotated_template=0, uint32_t nsnx=10, uint32_t nsnz=10):
 * 
 *         cdef uint32_t nx = x.size-1             # <<<<<<<<<<<<<<
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_x), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nx = __pyx_t_3;

  /* "ttcrpy/rgrid.pyx":1811
 * 
 *         cdef uint32_t nx = x.size-1
 *         cdef uint32_t nz = z.size-1             # <<<<<<<<<<<<<<
 *         self._dx = x[1] - x[0]
 *         self._dz = z[1] - z[0]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_z), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nz = __pyx_t_3;

  /* "ttcrpy/rgrid.pyx":1812
 *         cdef uint32_t nx = x.size-1
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]             # <<<<<<<<<<<<<<
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]
 */
  __pyx_t_4 = 1;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1812, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1812, __pyx_L1_error)
  }
  __pyx_v_self->_dx = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_x.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides)));

  /* "ttcrpy/rgrid.pyx":1813
 *         cdef uint32_t nz = z.size-1
 *         self._dx = x[1] - x[0]
 *         self._dz = z[1] - z[0]             # <<<<<<<<<<<<<<
 *         cdef double xmin = x[0]
 *         cdef double zmin = z[0]
 */
  __pyx_t_6 = 1;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1813, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1813, __pyx_L1_error)
  }
  __pyx_v_self->_dz = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_z.diminfo[0].strides)));

  /* "ttcrpy/rgrid.pyx":1814
 *         self._dx = x[1] - x[0]
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]             # <<<<<<<<<<<<<<
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_x.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_x.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1814, __pyx_L1_error)
  }
  __pyx_v_xmin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_x.diminfo[0].strides));

  /* "ttcrpy/rgrid.pyx":1815
 *         self._dz = z[1] - z[0]
 *         cdef double xmin = x[0]
 *         cdef double zmin = z[0]             # <<<<<<<<<<<<<<
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_pybuffernd_z.diminfo[0].shape;
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_z.diminfo[0].shape)) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(0, 1815, __pyx_L1_error)
  }
  __pyx_v_zmin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_z.diminfo[0].strides));

  /* "ttcrpy/rgrid.pyx":1816
 *         cdef double xmin = x[0]
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness             # <<<<<<<<<<<<<<
 *         self._n_threads = n_threads
 *         self.eps = eps
 */
  __pyx_v_self->cell_slowness = __pyx_v_cell_slowness;

  /* "ttcrpy/rgrid.pyx":1817
 *         cdef double zmin = z[0]
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads             # <<<<<<<<<<<<<<
 *         self.eps = eps
 *         self.maxit = maxit
 */
  __pyx_v_self->_n_threads = __pyx_v_n_threads;

  /* "ttcrpy/rgrid.pyx":1818
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 *         self.eps = eps             # <<<<<<<<<<<<<<
 *         self.maxit = maxit
 *         self.weno = weno
 */
  __pyx_v_self->eps = __pyx_v_eps;

  /* "ttcrpy/rgrid.pyx":1819
 *         self._n_threads = n_threads
 *         self.eps = eps
 *         self.maxit = maxit             # <<<<<<<<<<<<<<
 *         self.weno = weno
 *         self.rotated_template = rotated_template
 */
  __pyx_v_self->maxit = __pyx_v_maxit;

  /* "ttcrpy/rgrid.pyx":1820
 *         self.eps = eps
 *         self.maxit = maxit
 *         self.weno = weno             # <<<<<<<<<<<<<<
 *         self.rotated_template = rotated_template
 *         self.nsnx = nsnx
 */
  __pyx_v_self->weno = __pyx_v_weno;

  /* "ttcrpy/rgrid.pyx":1821
 *         self.maxit = maxit
 *         self.weno = weno
 *         self.rotated_template = rotated_template             # <<<<<<<<<<<<<<
 *         self.nsnx = nsnx
 *         self.nsnz = nsnz
 */
  __pyx_v_self->rotated_template = __pyx_v_rotated_template;

  /* "ttcrpy/rgrid.pyx":1822
 *         self.weno = weno
 *         self.rotated_template = rotated_template
 *         self.nsnx = nsnx             # <<<<<<<<<<<<<<
 *         self.nsnz = nsnz
 *         self.iso = b'i'
 */
  __pyx_v_self->nsnx = __pyx_v_nsnx;

  /* "ttcrpy/rgrid.pyx":1823
 *         self.rotated_template = rotated_template
 *         self.nsnx = nsnx
 *         self.nsnz = nsnz             # <<<<<<<<<<<<<<
 *         self.iso = b'i'
 * 
 */
  __pyx_v_self->nsnz = __pyx_v_nsnz;

  /* "ttcrpy/rgrid.pyx":1824
 *         self.nsnx = nsnx
 *         self.nsnz = nsnz
 *         self.iso = b'i'             # <<<<<<<<<<<<<<
 * 
 *         for val in x:
 */
  __pyx_v_self->iso = 'i';

  /* "ttcrpy/rgrid.pyx":1826
 *         self.iso = b'i'
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             self._x.push_back(val)
 *         for val in z:
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_x))) || PyTuple_CheckExact(((PyObject *)__pyx_v_x))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_x); __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1826, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1826, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1826, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1826, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":1827
 * 
 *         for val in x:
 *             self._x.push_back(val)             # <<<<<<<<<<<<<<
 *         for val in z:
 *             self._z.push_back(val)
 */
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1827, __pyx_L1_error)
    try {
      __pyx_v_self->_x.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1827, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":1826
 *         self.iso = b'i'
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             self._x.push_back(val)
 *         for val in z:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1828
 *         for val in x:
 *             self._x.push_back(val)
 *         for val in z:             # <<<<<<<<<<<<<<
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_z))) || PyTuple_CheckExact(((PyObject *)__pyx_v_z))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_z); __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1828, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1828, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1828, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1828, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":1829
 *             self._x.push_back(val)
 *         for val in z:
 *             self._z.push_back(val)             # <<<<<<<<<<<<<<
 *         self._x.shrink_to_fit()
 *         self._z.shrink_to_fit()
 */
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1829, __pyx_L1_error)
    try {
      __pyx_v_self->_z.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1829, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":1828
 *         for val in x:
 *             self._x.push_back(val)
 *         for val in z:             # <<<<<<<<<<<<<<
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1830
 *         for val in z:
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()             # <<<<<<<<<<<<<<
 *         self._z.shrink_to_fit()
 * 
 */
  __pyx_v_self->_x.shrink_to_fit();

  /* "ttcrpy/rgrid.pyx":1831
 *             self._z.push_back(val)
 *         self._x.shrink_to_fit()
 *         self._z.shrink_to_fit()             # <<<<<<<<<<<<<<
 * 
 *         if cell_slowness:
 */
  __pyx_v_self->_z.shrink_to_fit();

  /* "ttcrpy/rgrid.pyx":1833
 *         self._z.shrink_to_fit()
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'SPM':
 *                 self.method = b's'
 */
  __pyx_t_10 = (__pyx_v_cell_slowness != 0);
  if (__pyx_t_10) {

    /* "ttcrpy/rgrid.pyx":1834
 * 
 *         if cell_slowness:
 *             if method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 if aniso == 'iso':
 */
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1834, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_10 != 0);
    if (__pyx_t_11) {

      /* "ttcrpy/rgrid.pyx":1835
 *         if cell_slowness:
 *             if method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 if aniso == 'iso':
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d](
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/rgrid.pyx":1836
 *             if method == 'SPM':
 *                 self.method = b's'
 *                 if aniso == 'iso':             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d](
 *                                     nx, nz, self._dx, self._dz,
 */
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_aniso, __pyx_n_u_iso, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1836, __pyx_L1_error)
      __pyx_t_10 = (__pyx_t_11 != 0);
      if (__pyx_t_10) {

        /* "ttcrpy/rgrid.pyx":1837
 *                 self.method = b's'
 *                 if aniso == 'iso':
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d](             # <<<<<<<<<<<<<<
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 */
        try {
          __pyx_t_12 = new ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d> (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1837, __pyx_L1_error)
        }
        __pyx_v_self->grid = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":1836
 *             if method == 'SPM':
 *                 self.method = b's'
 *                 if aniso == 'iso':             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d](
 *                                     nx, nz, self._dx, self._dz,
 */
        goto __pyx_L9;
      }

      /* "ttcrpy/rgrid.pyx":1840
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'elliptical':             # <<<<<<<<<<<<<<
 *                     self.iso = b'e'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_e](
 */
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_aniso, __pyx_n_u_elliptical, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1840, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_10 != 0);
      if (__pyx_t_11) {

        /* "ttcrpy/rgrid.pyx":1841
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'elliptical':
 *                     self.iso = b'e'             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_e](
 *                                     nx, nz, self._dx, self._dz,
 */
        __pyx_v_self->iso = 'e';

        /* "ttcrpy/rgrid.pyx":1842
 *                 elif aniso == 'elliptical':
 *                     self.iso = b'e'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_e](             # <<<<<<<<<<<<<<
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 */
        try {
          __pyx_t_13 = new ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_e> (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1842, __pyx_L1_error)
        }
        __pyx_v_self->grid = __pyx_t_13;

        /* "ttcrpy/rgrid.pyx":1840
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'elliptical':             # <<<<<<<<<<<<<<
 *                     self.iso = b'e'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_e](
 */
        goto __pyx_L9;
      }

      /* "ttcrpy/rgrid.pyx":1845
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'tilted_elliptical':             # <<<<<<<<<<<<<<
 *                     self.iso = b't'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_te](
 */
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_aniso, __pyx_n_u_tilted_elliptical, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1845, __pyx_L1_error)
      __pyx_t_10 = (__pyx_t_11 != 0);
      if (__pyx_t_10) {

        /* "ttcrpy/rgrid.pyx":1846
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'tilted_elliptical':
 *                     self.iso = b't'             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_te](
 *                                     nx, nz, self._dx, self._dz,
 */
        __pyx_v_self->iso = 't';

        /* "ttcrpy/rgrid.pyx":1847
 *                 elif aniso == 'tilted_elliptical':
 *                     self.iso = b't'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_te](             # <<<<<<<<<<<<<<
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 */
        try {
          __pyx_t_14 = new ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_te> (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1847, __pyx_L1_error)
        }
        __pyx_v_self->grid = __pyx_t_14;

        /* "ttcrpy/rgrid.pyx":1845
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'tilted_elliptical':             # <<<<<<<<<<<<<<
 *                     self.iso = b't'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_te](
 */
        goto __pyx_L9;
      }

      /* "ttcrpy/rgrid.pyx":1850
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_psv':             # <<<<<<<<<<<<<<
 *                     self.iso = b'p'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_p](
 */
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_aniso, __pyx_n_u_vti_psv, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1850, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_10 != 0);
      if (__pyx_t_11) {

        /* "ttcrpy/rgrid.pyx":1851
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_psv':
 *                     self.iso = b'p'             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_p](
 *                                     nx, nz, self._dx, self._dz,
 */
        __pyx_v_self->iso = 'p';

        /* "ttcrpy/rgrid.pyx":1852
 *                 elif aniso == 'vti_psv':
 *                     self.iso = b'p'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_p](             # <<<<<<<<<<<<<<
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 */
        try {
          __pyx_t_15 = new ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_p> (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1852, __pyx_L1_error)
        }
        __pyx_v_self->grid = __pyx_t_15;

        /* "ttcrpy/rgrid.pyx":1850
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_psv':             # <<<<<<<<<<<<<<
 *                     self.iso = b'p'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_p](
 */
        goto __pyx_L9;
      }

      /* "ttcrpy/rgrid.pyx":1855
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_sh':             # <<<<<<<<<<<<<<
 *                     self.iso = b'h'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_h](
 */
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_aniso, __pyx_n_u_vti_sh, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1855, __pyx_L1_error)
      __pyx_t_10 = (__pyx_t_11 != 0);
      if (likely(__pyx_t_10)) {

        /* "ttcrpy/rgrid.pyx":1856
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_sh':
 *                     self.iso = b'h'             # <<<<<<<<<<<<<<
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_h](
 *                                     nx, nz, self._dx, self._dz,
 */
        __pyx_v_self->iso = 'h';

        /* "ttcrpy/rgrid.pyx":1857
 *                 elif aniso == 'vti_sh':
 *                     self.iso = b'h'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_h](             # <<<<<<<<<<<<<<
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 */
        try {
          __pyx_t_16 = new ttcr::Grid2Drcsp<double,uint32_t,ttcr::sxz<double> ,ttcr::cell2d_h> (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1857, __pyx_L1_error)
        }
        __pyx_v_self->grid = __pyx_t_16;

        /* "ttcrpy/rgrid.pyx":1855
 *                                     nx, nz, self._dx, self._dz,
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 elif aniso == 'vti_sh':             # <<<<<<<<<<<<<<
 *                     self.iso = b'h'
 *                     self.grid = new Grid2Drcsp[double,uint32_t,sxz[double],cell2d_h](
 */
        goto __pyx_L9;
      }

      /* "ttcrpy/rgrid.pyx":1861
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 else:
 *                     raise ValueError('Anisotropy model not implemented')             # <<<<<<<<<<<<<<
 *             elif method == 'FSM':
 *                 self.method = b'f'
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1861, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 1861, __pyx_L1_error)
      }
      __pyx_L9:;

      /* "ttcrpy/rgrid.pyx":1834
 * 
 *         if cell_slowness:
 *             if method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 if aniso == 'iso':
 */
      goto __pyx_L8;
    }

    /* "ttcrpy/rgrid.pyx":1862
 *                 else:
 *                     raise ValueError('Anisotropy model not implemented')
 *             elif method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drcfs[double,uint32_t,sxz[double]](nx, nz,
 */
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1862, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_10 != 0);
    if (likely(__pyx_t_11)) {

      /* "ttcrpy/rgrid.pyx":1863
 *                     raise ValueError('Anisotropy model not implemented')
 *             elif method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Drcfs[double,uint32_t,sxz[double]](nx, nz,
 *                                 self._dx, self._dz, xmin, zmin, eps,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/rgrid.pyx":1864
 *             elif method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drcfs[double,uint32_t,sxz[double]](nx, nz,             # <<<<<<<<<<<<<<
 *                                 self._dx, self._dz, xmin, zmin, eps,
 *                                 maxit, weno, rotated_template, n_threads)
 */
      try {
        __pyx_t_17 = new ttcr::Grid2Drcfs<double,uint32_t,ttcr::sxz<double> > (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_rotated_template, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1864, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_17;

      /* "ttcrpy/rgrid.pyx":1862
 *                 else:
 *                     raise ValueError('Anisotropy model not implemented')
 *             elif method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drcfs[double,uint32_t,sxz[double]](nx, nz,
 */
      goto __pyx_L8;
    }

    /* "ttcrpy/rgrid.pyx":1868
 *                                 maxit, weno, rotated_template, n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 *         else:
 *             if method == 'SPM':
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1868, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_18 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_18, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1868, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1868, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1868, __pyx_L1_error)
    }
    __pyx_L8:;

    /* "ttcrpy/rgrid.pyx":1833
 *         self._z.shrink_to_fit()
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'SPM':
 *                 self.method = b's'
 */
    goto __pyx_L7;
  }

  /* "ttcrpy/rgrid.pyx":1870
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Drnsp[double,uint32_t,sxz[double]](nx, nz,
 */
  /*else*/ {
    __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1870, __pyx_L1_error)
    __pyx_t_10 = (__pyx_t_11 != 0);
    if (__pyx_t_10) {

      /* "ttcrpy/rgrid.pyx":1871
 *         else:
 *             if method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Drnsp[double,uint32_t,sxz[double]](nx, nz,
 *                                 self._dx, self._dz, xmin, zmin,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/rgrid.pyx":1872
 *             if method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid2Drnsp[double,uint32_t,sxz[double]](nx, nz,             # <<<<<<<<<<<<<<
 *                                 self._dx, self._dz, xmin, zmin,
 *                                 nsnx, nsnz, n_threads)
 */
      try {
        __pyx_t_19 = new ttcr::Grid2Drnsp<double,uint32_t,ttcr::sxz<double> > (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_nsnx, __pyx_v_nsnz, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1872, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_19;

      /* "ttcrpy/rgrid.pyx":1870
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Drnsp[double,uint32_t,sxz[double]](nx, nz,
 */
      goto __pyx_L10;
    }

    /* "ttcrpy/rgrid.pyx":1875
 *                                 self._dx, self._dz, xmin, zmin,
 *                                 nsnx, nsnz, n_threads)
 *             elif method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drnfs[double,uint32_t,sxz[double]](nx, nz,
 */
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1875, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_10 != 0);
    if (__pyx_t_11) {

      /* "ttcrpy/rgrid.pyx":1876
 *                                 nsnx, nsnz, n_threads)
 *             elif method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Drnfs[double,uint32_t,sxz[double]](nx, nz,
 *                                 self._dx, self._dz, xmin, zmin, eps,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/rgrid.pyx":1877
 *             elif method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drnfs[double,uint32_t,sxz[double]](nx, nz,             # <<<<<<<<<<<<<<
 *                                 self._dx, self._dz, xmin, zmin, eps,
 *                                 maxit, weno, rotated_template, n_threads)
 */
      try {
        __pyx_t_20 = new ttcr::Grid2Drnfs<double,uint32_t,ttcr::sxz<double> > (__pyx_v_nx, __pyx_v_nz, __pyx_v_self->_dx, __pyx_v_self->_dz, __pyx_v_xmin, __pyx_v_zmin, __pyx_v_eps, __pyx_v_maxit, __pyx_v_weno, __pyx_v_rotated_template, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1877, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_20;

      /* "ttcrpy/rgrid.pyx":1875
 *                                 self._dx, self._dz, xmin, zmin,
 *                                 nsnx, nsnz, n_threads)
 *             elif method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Drnfs[double,uint32_t,sxz[double]](nx, nz,
 */
    }
    __pyx_L10:;
  }
  __pyx_L7:;

  /* "ttcrpy/rgrid.pyx":1804
 *     cdef Grid2D[double,uint32_t,sxz[double]]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=1] x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.double_t, ndim=1] z, size_t n_threads=1,
 *                   bool cell_slowness=1, str method='SPM', str aniso='iso',
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_18);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1881
 *                                 maxit, weno, rotated_template, n_threads)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

/* Python wrapper */
static void __pyx_pw_6ttcrpy_5rgrid_6Grid2d_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6ttcrpy_5rgrid_6Grid2d_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6ttcrpy_5rgrid_6Grid2d_2__dealloc__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6ttcrpy_5rgrid_6Grid2d_2__dealloc__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ttcrpy/rgrid.pyx":1882
 * 
 *     def __dealloc__(self):
 *         del self.grid             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  delete __pyx_v_self->grid;

  /* "ttcrpy/rgrid.pyx":1881
 *                                 maxit, weno, rotated_template, n_threads)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ttcrpy/rgrid.pyx":1884
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_4__reduce__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_4__reduce__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_aniso = NULL;
  PyObject *__pyx_v_constructor_params = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ttcrpy/rgrid.pyx":1885
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
  switch (__pyx_v_self->method) {
    case 'f':

    /* "ttcrpy/rgrid.pyx":1886
 *     def __reduce__(self):
 *         if self.method == b'f':
 *             method = 'FSM'             # <<<<<<<<<<<<<<
 *         elif self.method == b's':
 *             method = 'SPM'
 */
    __Pyx_INCREF(__pyx_n_u_FSM);
    __pyx_v_method = __pyx_n_u_FSM;

    /* "ttcrpy/rgrid.pyx":1885
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
    break;
    case 's':

    /* "ttcrpy/rgrid.pyx":1888
 *             method = 'FSM'
 *         elif self.method == b's':
 *             method = 'SPM'             # <<<<<<<<<<<<<<
 *         elif self.method == b'd':
 *             method = 'DSPM'
 */
    __Pyx_INCREF(__pyx_n_u_SPM);
    __pyx_v_method = __pyx_n_u_SPM;

    /* "ttcrpy/rgrid.pyx":1887
 *         if self.method == b'f':
 *             method = 'FSM'
 *         elif self.method == b's':             # <<<<<<<<<<<<<<
 *             method = 'SPM'
 *         elif self.method == b'd':
 */
    break;
    case 'd':

    /* "ttcrpy/rgrid.pyx":1890
 *             method = 'SPM'
 *         elif self.method == b'd':
 *             method = 'DSPM'             # <<<<<<<<<<<<<<
 * 
 *         if self.iso == b'i':
 */
    __Pyx_INCREF(__pyx_n_u_DSPM);
    __pyx_v_method = __pyx_n_u_DSPM;

    /* "ttcrpy/rgrid.pyx":1889
 *         elif self.method == b's':
 *             method = 'SPM'
 *         elif self.method == b'd':             # <<<<<<<<<<<<<<
 *             method = 'DSPM'
 * 
 */
    break;
    default: break;
  }

  /* "ttcrpy/rgrid.pyx":1892
 *             method = 'DSPM'
 * 
 *         if self.iso == b'i':             # <<<<<<<<<<<<<<
 *             aniso = 'iso'
 *         elif self.iso == b'e':
 */
  switch (__pyx_v_self->iso) {
    case 'i':

    /* "ttcrpy/rgrid.pyx":1893
 * 
 *         if self.iso == b'i':
 *             aniso = 'iso'             # <<<<<<<<<<<<<<
 *         elif self.iso == b'e':
 *             aniso = 'elliptical'
 */
    __Pyx_INCREF(__pyx_n_u_iso);
    __pyx_v_aniso = __pyx_n_u_iso;

    /* "ttcrpy/rgrid.pyx":1892
 *             method = 'DSPM'
 * 
 *         if self.iso == b'i':             # <<<<<<<<<<<<<<
 *             aniso = 'iso'
 *         elif self.iso == b'e':
 */
    break;
    case 'e':

    /* "ttcrpy/rgrid.pyx":1895
 *             aniso = 'iso'
 *         elif self.iso == b'e':
 *             aniso = 'elliptical'             # <<<<<<<<<<<<<<
 *         elif self.iso == b't':
 *             aniso = 'tilted_elliptical'
 */
    __Pyx_INCREF(__pyx_n_u_elliptical);
    __pyx_v_aniso = __pyx_n_u_elliptical;

    /* "ttcrpy/rgrid.pyx":1894
 *         if self.iso == b'i':
 *             aniso = 'iso'
 *         elif self.iso == b'e':             # <<<<<<<<<<<<<<
 *             aniso = 'elliptical'
 *         elif self.iso == b't':
 */
    break;
    case 't':

    /* "ttcrpy/rgrid.pyx":1897
 *             aniso = 'elliptical'
 *         elif self.iso == b't':
 *             aniso = 'tilted_elliptical'             # <<<<<<<<<<<<<<
 *         elif self.iso == b'p':
 *             aniso = 'vti_psv'
 */
    __Pyx_INCREF(__pyx_n_u_tilted_elliptical);
    __pyx_v_aniso = __pyx_n_u_tilted_elliptical;

    /* "ttcrpy/rgrid.pyx":1896
 *         elif self.iso == b'e':
 *             aniso = 'elliptical'
 *         elif self.iso == b't':             # <<<<<<<<<<<<<<
 *             aniso = 'tilted_elliptical'
 *         elif self.iso == b'p':
 */
    break;
    case 'p':

    /* "ttcrpy/rgrid.pyx":1899
 *             aniso = 'tilted_elliptical'
 *         elif self.iso == b'p':
 *             aniso = 'vti_psv'             # <<<<<<<<<<<<<<
 *         elif self.iso == b'h':
 *             aniso = 'vti_sh'
 */
    __Pyx_INCREF(__pyx_n_u_vti_psv);
    __pyx_v_aniso = __pyx_n_u_vti_psv;

    /* "ttcrpy/rgrid.pyx":1898
 *         elif self.iso == b't':
 *             aniso = 'tilted_elliptical'
 *         elif self.iso == b'p':             # <<<<<<<<<<<<<<
 *             aniso = 'vti_psv'
 *         elif self.iso == b'h':
 */
    break;
    case 'h':

    /* "ttcrpy/rgrid.pyx":1901
 *             aniso = 'vti_psv'
 *         elif self.iso == b'h':
 *             aniso = 'vti_sh'             # <<<<<<<<<<<<<<
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 */
    __Pyx_INCREF(__pyx_n_u_vti_sh);
    __pyx_v_aniso = __pyx_n_u_vti_sh;

    /* "ttcrpy/rgrid.pyx":1900
 *         elif self.iso == b'p':
 *             aniso = 'vti_psv'
 *         elif self.iso == b'h':             # <<<<<<<<<<<<<<
 *             aniso = 'vti_sh'
 * 
 */
    break;
    default: break;
  }

  /* "ttcrpy/rgrid.pyx":1903
 *             aniso = 'vti_sh'
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,             # <<<<<<<<<<<<<<
 *                               aniso, self.eps, self.maxit, self.weno,
 *                               self.rotated_template, self.nsnx, self.nsnz)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->cell_slowness); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_method)) { __Pyx_RaiseUnboundLocalError("method"); __PYX_ERR(0, 1903, __pyx_L1_error) }

  /* "ttcrpy/rgrid.pyx":1904
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 *                               aniso, self.eps, self.maxit, self.weno,             # <<<<<<<<<<<<<<
 *                               self.rotated_template, self.nsnx, self.nsnz)
 *         return (_rebuild2d, (self.x, self.z, constructor_params))
 */
  if (unlikely(!__pyx_v_aniso)) { __Pyx_RaiseUnboundLocalError("aniso"); __PYX_ERR(0, 1904, __pyx_L1_error) }
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->eps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->maxit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->weno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "ttcrpy/rgrid.pyx":1905
 *         constructor_params = (self.n_threads, self.cell_slowness, method,
 *                               aniso, self.eps, self.maxit, self.weno,
 *                               self.rotated_template, self.nsnx, self.nsnz)             # <<<<<<<<<<<<<<
 *         return (_rebuild2d, (self.x, self.z, constructor_params))
 * 
 */
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->rotated_template); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->nsnx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->nsnz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "ttcrpy/rgrid.pyx":1903
 *             aniso = 'vti_sh'
 * 
 *         constructor_params = (self.n_threads, self.cell_slowness, method,             # <<<<<<<<<<<<<<
 *                               aniso, self.eps, self.maxit, self.weno,
 *                               self.rotated_template, self.nsnx, self.nsnz)
 */
  __pyx_t_9 = PyTuple_New(10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_aniso);
  __Pyx_GIVEREF(__pyx_v_aniso);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_v_aniso);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_constructor_params = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "ttcrpy/rgrid.pyx":1906
 *                               aniso, self.eps, self.maxit, self.weno,
 *                               self.rotated_template, self.nsnx, self.nsnz)
 *         return (_rebuild2d, (self.x, self.z, constructor_params))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_rebuild2d); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_x); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_z); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
  __Pyx_INCREF(__pyx_v_constructor_params);
  __Pyx_GIVEREF(__pyx_v_constructor_params);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_constructor_params);
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_9 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1884
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_aniso);
  __Pyx_XDECREF(__pyx_v_constructor_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1909
 * 
 *     @property
 *     def x(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_1x_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_1x_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_1x___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_1x___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_v_tmp = NULL;
  int __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<double> ::size_type __pyx_t_5;
  std::vector<double> ::size_type __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1911
 *     def x(self):
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))             # <<<<<<<<<<<<<<
 *         cdef int n
 *         for n in range(self._x.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tmp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1913
 *         tmp = np.empty((self._x.size(),))
 *         cdef int n
 *         for n in range(self._x.size()):             # <<<<<<<<<<<<<<
 *             tmp[n] = self._x[n]
 *         return tmp
 */
  __pyx_t_5 = __pyx_v_self->_x.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/rgrid.pyx":1914
 *         cdef int n
 *         for n in range(self._x.size()):
 *             tmp[n] = self._x[n]             # <<<<<<<<<<<<<<
 *         return tmp
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_x[__pyx_v_n])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tmp, __pyx_v_n, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1914, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1915
 *         for n in range(self._x.size()):
 *             tmp[n] = self._x[n]
 *         return tmp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_r = __pyx_v_tmp;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1909
 * 
 *     @property
 *     def x(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along x"""
 *         tmp = np.empty((self._x.size(),))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.x.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1918
 * 
 *     @property
 *     def z(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_1z_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_1z_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_1z___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_1z___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_v_tmp = NULL;
  int __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<double> ::size_type __pyx_t_5;
  std::vector<double> ::size_type __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1920
 *     def z(self):
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))             # <<<<<<<<<<<<<<
 *         cdef int n
 *         for n in range(self._z.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tmp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1922
 *         tmp = np.empty((self._z.size(),))
 *         cdef int n
 *         for n in range(self._z.size()):             # <<<<<<<<<<<<<<
 *             tmp[n] = self._z[n]
 *         return tmp
 */
  __pyx_t_5 = __pyx_v_self->_z.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/rgrid.pyx":1923
 *         cdef int n
 *         for n in range(self._z.size()):
 *             tmp[n] = self._z[n]             # <<<<<<<<<<<<<<
 *         return tmp
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_z[__pyx_v_n])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tmp, __pyx_v_n, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1923, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1924
 *         for n in range(self._z.size()):
 *             tmp[n] = self._z[n]
 *         return tmp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_r = __pyx_v_tmp;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1918
 * 
 *     @property
 *     def z(self):             # <<<<<<<<<<<<<<
 *         """np.ndarray: node coordinates along z"""
 *         tmp = np.empty((self._z.size(),))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.z.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1927
 * 
 *     @property
 *     def dx(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dx
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dx_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dx___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dx___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1929
 *     def dx(self):
 *         """float: node separation along x"""
 *         return self._dx             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1927
 * 
 *     @property
 *     def dx(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dx
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.dx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1932
 * 
 *     @property
 *     def dz(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dz
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dz_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dz_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dz___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_2dz___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1934
 *     def dz(self):
 *         """float: node separation along x"""
 *         return self._dz             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_dz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1932
 * 
 *     @property
 *     def dz(self):             # <<<<<<<<<<<<<<
 *         """float: node separation along x"""
 *         return self._dz
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.dz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1937
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_5shape___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_5shape___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1939
 *     def shape(self):
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1, self._z.size()-1)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1940
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 *             return (self._x.size()-1, self._z.size()-1)             # <<<<<<<<<<<<<<
 *         else:
 *             return (self._x.size(), self._z.size())
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1939
 *     def shape(self):
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1, self._z.size()-1)
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":1942
 *             return (self._x.size()-1, self._z.size()-1)
 *         else:
 *             return (self._x.size(), self._z.size())             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":1937
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         """:obj:`list` of :obj:`int`: number of parameters along each dimension"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1945
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_9n_threads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_9n_threads_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_9n_threads___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1947
 *     def n_threads(self):
 *         """int: number of threads for raytracing"""
 *         return self._n_threads             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1945
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.n_threads.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1950
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_7nparams_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_7nparams_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_7nparams___get__(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_7nparams___get__(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/rgrid.pyx":1952
 *     def nparams(self):
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1) * (self._z.size()-1)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":1953
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 *             return (self._x.size()-1) * (self._z.size()-1)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._x.size() * self._z.size()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->_x.size() - 1) * (__pyx_v_self->_z.size() - 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":1952
 *     def nparams(self):
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return (self._x.size()-1) * (self._z.size()-1)
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":1955
 *             return (self._x.size()-1) * (self._z.size()-1)
 *         else:
 *             return self._x.size() * self._z.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_nodes(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() * __pyx_v_self->_z.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":1950
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for grid"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.nparams.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1957
 *             return self._x.size() * self._z.size()
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_7get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_6get_number_of_nodes[] = "\n        Returns\n        -------\n        int:\n            number of nodes in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_7get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_nodes (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_6get_number_of_nodes(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_6get_number_of_nodes(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_nodes", 0);

  /* "ttcrpy/rgrid.pyx":1964
 *             number of nodes in grid
 *         """
 *         return self._x.size() * self._z.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_cells(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() * __pyx_v_self->_z.size())); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1957
 *             return self._x.size() * self._z.size()
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.get_number_of_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1966
 *         return self._x.size() * self._z.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_9get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_8get_number_of_cells[] = "\n        Returns\n        -------\n        int:\n            number of cells in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_9get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_cells (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_8get_number_of_cells(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_8get_number_of_cells(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_cells", 0);

  /* "ttcrpy/rgrid.pyx":1973
 *             number of cells in grid
 *         """
 *         return (self._x.size()-1) * (self._z.size()-1)             # <<<<<<<<<<<<<<
 * 
 *     def get_grid_traveltimes(self, thread_no=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->_x.size() - 1) * (__pyx_v_self->_z.size() - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1966
 *         return self._x.size() * self._z.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.get_number_of_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":1975
 *         return (self._x.size()-1) * (self._z.size()-1)
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_11get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_10get_grid_traveltimes[] = "\n        get_grid_traveltimes(thread_no=0)\n\n        Obtain traveltimes computed at primary grid nodes\n\n        Parameters\n        ----------\n        thread_no : int\n            thread used to computed traveltimes (default is 0)\n\n        Returns\n        -------\n        tt: np ndarray, shape (nx, nz)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_11get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_thread_no = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_grid_traveltimes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_thread_no,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_grid_traveltimes") < 0)) __PYX_ERR(0, 1975, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_thread_no = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_grid_traveltimes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1975, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_10get_grid_traveltimes(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_thread_no);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_10get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_thread_no) {
  std::vector<double>  __pyx_v_tmp;
  int __pyx_v_n;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<double> ::size_type __pyx_t_7;
  std::vector<double> ::size_type __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_grid_traveltimes", 0);

  /* "ttcrpy/rgrid.pyx":1990
 *         tt: np ndarray, shape (nx, nz)
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_thread_no, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/rgrid.pyx":1991
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1991, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":1990
 *         tt: np ndarray, shape (nx, nz)
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  }

  /* "ttcrpy/rgrid.pyx":1994
 *         cdef vector[double] tmp
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)             # <<<<<<<<<<<<<<
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 */
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1994, __pyx_L1_error)
  try {
    __pyx_v_self->grid->getTT(__pyx_v_tmp, __pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1994, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":1995
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))             # <<<<<<<<<<<<<<
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_tmp.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_tt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":1996
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):             # <<<<<<<<<<<<<<
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._z.size())
 */
  __pyx_t_7 = __pyx_v_tmp.size();
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/rgrid.pyx":1997
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]             # <<<<<<<<<<<<<<
 *         shape = (self._x.size(), self._z.size())
 *         return tt.reshape(shape)
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_tmp[__pyx_v_n])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_n, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/rgrid.pyx":1998
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._z.size())             # <<<<<<<<<<<<<<
 *         return tt.reshape(shape)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_v_shape = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "ttcrpy/rgrid.pyx":1999
 *             tt[n] = tmp[n]
 *         shape = (self._x.size(), self._z.size())
 *         return tt.reshape(shape)             # <<<<<<<<<<<<<<
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tt, __pyx_n_s_reshape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_shape);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":1975
 *         return (self._x.size()-1) * (self._z.size()-1)
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2001
 *         return tt.reshape(shape)
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_13is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_12is_outside[] = "\n        is_outside(pts)\n\n        Check if points are outside grid\n\n        Parameters\n        ----------\n        pts : np ndarray, shape(npts, 3)\n            coordinates of points to check\n\n        Returns\n        -------\n        bool:\n            True if at least one point outside grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_13is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_outside (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pts), __pyx_ptype_5numpy_ndarray, 1, "pts", 0))) __PYX_ERR(0, 2001, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_12is_outside(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyArrayObject *)__pyx_v_pts));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_12is_outside(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyArrayObject *__pyx_v_pts) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pts;
  __Pyx_Buffer __pyx_pybuffer_pts;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_outside", 0);
  __pyx_pybuffer_pts.pybuffer.buf = NULL;
  __pyx_pybuffer_pts.refcount = 0;
  __pyx_pybuffernd_pts.data = NULL;
  __pyx_pybuffernd_pts.rcbuffer = &__pyx_pybuffer_pts;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pts.rcbuffer->pybuffer, (PyObject*)__pyx_v_pts, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2001, __pyx_L1_error)
  }
  __pyx_pybuffernd_pts.diminfo[0].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pts.diminfo[0].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pts.diminfo[1].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pts.diminfo[1].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[1];

  /* "ttcrpy/rgrid.pyx":2017
 *             True if at least one point outside grid
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or             # <<<<<<<<<<<<<<
 *                 np.min(pts[:,1]) < self._z.front() or np.max(pts[:,1]) > self._z.back() )
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_min); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_x.front()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 2017, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_x.back()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 2017, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ttcrpy/rgrid.pyx":2018
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or
 *                 np.min(pts[:,1]) < self._z.front() or np.max(pts[:,1]) > self._z.back() )             # <<<<<<<<<<<<<<
 * 
 *     def get_slowness(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_min); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_z.front()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 2018, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_pts), __pyx_tuple__5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_z.back()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":2001
 *         return tt.reshape(shape)
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.is_outside", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2020
 *                 np.min(pts[:,1]) < self._z.front() or np.max(pts[:,1]) > self._z.back() )
 * 
 *     def get_slowness(self):             # <<<<<<<<<<<<<<
 *         """Returns slowness of grid
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_15get_slowness(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_14get_slowness[] = "Returns slowness of grid\n\n        Returns\n        -------\n        slowness : np ndarray, shape (nx, nz)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_15get_slowness(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_14get_slowness(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_14get_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self) {
  int __pyx_v_i;
  std::vector<double>  __pyx_v_slown;
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  PyObject *__pyx_v_slowness = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  std::vector<double> ::size_type __pyx_t_6;
  std::vector<double> ::size_type __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slowness", 0);

  /* "ttcrpy/rgrid.pyx":2029
 *         cdef int i
 *         cdef vector[double] slown
 *         self.grid.getSlowness(slown)             # <<<<<<<<<<<<<<
 *         nx, nz = self.shape
 *         slowness = np.ndarray((nx*nz,))
 */
  try {
    __pyx_v_self->grid->getSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2029, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2030
 *         cdef vector[double] slown
 *         self.grid.getSlowness(slown)
 *         nx, nz = self.shape             # <<<<<<<<<<<<<<
 *         slowness = np.ndarray((nx*nz,))
 *         for i in range(slown.size()):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 2030, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 2030, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 2030, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_nx = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nz = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2031
 *         self.grid.getSlowness(slown)
 *         nx, nz = self.shape
 *         slowness = np.ndarray((nx*nz,))             # <<<<<<<<<<<<<<
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]
 */
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_slowness = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2032
 *         nx, nz = self.shape
 *         slowness = np.ndarray((nx*nz,))
 *         for i in range(slown.size()):             # <<<<<<<<<<<<<<
 *             slowness[i] = slown[i]
 *         return slowness.reshape((nx, nz))
 */
  __pyx_t_6 = __pyx_v_slown.size();
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ttcrpy/rgrid.pyx":2033
 *         slowness = np.ndarray((nx*nz,))
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]             # <<<<<<<<<<<<<<
 *         return slowness.reshape((nx, nz))
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_slown[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_slowness, __pyx_v_i, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2033, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/rgrid.pyx":2034
 *         for i in range(slown.size()):
 *             slowness[i] = slown[i]
 *         return slowness.reshape((nx, nz))             # <<<<<<<<<<<<<<
 * 
 *     def set_slowness(self, slowness):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nx);
  __Pyx_GIVEREF(__pyx_v_nx);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nx);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nz);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":2020
 *                 np.min(pts[:,1]) < self._z.front() or np.max(pts[:,1]) > self._z.back() )
 * 
 *     def get_slowness(self):             # <<<<<<<<<<<<<<
 *         """Returns slowness of grid
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.get_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2036
 *         return slowness.reshape((nx, nz))
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_17set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_16set_slowness[] = "\n        set_slowness(slowness)\n\n        Assign slowness to grid\n\n        Parameters\n        ----------\n        slowness : np ndarray, shape (nx, nz)\n            slowness may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_17set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_16set_slowness(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_slowness));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_16set_slowness(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_slowness", 0);

  /* "ttcrpy/rgrid.pyx":2047
 *             slowness may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2048
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2049
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2047
 *             slowness may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2051
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if slowness.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2052
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if slowness.size != nx*nz:
 *             raise ValueError('Slowness vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2053
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if slowness.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2054
 *             nz = self._z.size()
 *         if slowness.size != nx*nz:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2054, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2053
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if slowness.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2058
 *         cdef vector[double] data
 *         cdef int i
 *         if slowness.ndim == 2:             # <<<<<<<<<<<<<<
 *             if slowness.shape != (nx, nz):
 *                 raise ValueError('Slowness has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2059
 *         cdef int i
 *         if slowness.ndim == 2:
 *             if slowness.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2059, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2059, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2060
 *         if slowness.ndim == 2:
 *             if slowness.shape != (nx, nz):
 *                 raise ValueError('Slowness has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = slowness.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2060, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2060, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2059
 *         cdef int i
 *         if slowness.ndim == 2:
 *             if slowness.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2061
 *             if slowness.shape != (nx, nz):
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2062
 *                 raise ValueError('Slowness has wrong shape')
 *             tmp = slowness.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif slowness.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2063
 *             tmp = slowness.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif slowness.ndim == 1:
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2063, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2063, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2058
 *         cdef vector[double] data
 *         cdef int i
 *         if slowness.ndim == 2:             # <<<<<<<<<<<<<<
 *             if slowness.shape != (nx, nz):
 *                 raise ValueError('Slowness has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2064
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif slowness.ndim == 1:             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(slowness[i])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2065
 *                 data.push_back(tmp[i])
 *         elif slowness.ndim == 1:
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(slowness[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2066
 *         elif slowness.ndim == 1:
 *             for i in range(nx*nz):
 *                 data.push_back(slowness[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_slowness, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2066, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2066, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2066, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2064
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif slowness.ndim == 1:             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(slowness[i])
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2068
 *                 data.push_back(slowness[i])
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2068, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2068, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2069
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')
 *         self.grid.setSlowness(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_velocity(self, velocity):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2069, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2036
 *         return slowness.reshape((nx, nz))
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2071
 *         self.grid.setSlowness(data)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_19set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_18set_velocity[] = "\n        set_velocity(velocity)\n\n        Assign velocity to grid\n\n        Parameters\n        ----------\n        velocity : np ndarray, shape (nx, nz)\n            velocity may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_19set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_velocity (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_18set_velocity(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_velocity));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_18set_velocity(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_velocity", 0);

  /* "ttcrpy/rgrid.pyx":2082
 *             velocity may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2083
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2084
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2082
 *             velocity may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2086
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if velocity.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2087
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if velocity.size != nx*nz:
 *             raise ValueError('velocity vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2087, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2088
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if velocity.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2089
 *             nz = self._z.size()
 *         if velocity.size != nx*nz:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2089, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2089, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2088
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if velocity.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2093
 *         cdef vector[double] data
 *         cdef int i
 *         if velocity.ndim == 2:             # <<<<<<<<<<<<<<
 *             if velocity.shape != (nx, nz):
 *                 raise ValueError('velocity has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2094
 *         cdef int i
 *         if velocity.ndim == 2:
 *             if velocity.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2095
 *         if velocity.ndim == 2:
 *             if velocity.shape != (nx, nz):
 *                 raise ValueError('velocity has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = velocity.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2095, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2095, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2094
 *         cdef int i
 *         if velocity.ndim == 2:
 *             if velocity.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2096
 *             if velocity.shape != (nx, nz):
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(1./tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2097
 *                 raise ValueError('velocity has wrong shape')
 *             tmp = velocity.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2097, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2098
 *             tmp = velocity.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(1./tmp[i])             # <<<<<<<<<<<<<<
 *         elif velocity.ndim == 1:
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_2, 1., 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2098, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2093
 *         cdef vector[double] data
 *         cdef int i
 *         if velocity.ndim == 2:             # <<<<<<<<<<<<<<
 *             if velocity.shape != (nx, nz):
 *                 raise ValueError('velocity has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2099
 *             for i in range(nx*nz):
 *                 data.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(1./velocity[i])
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2100
 *                 data.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(1./velocity[i])
 *         else:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2101
 *         elif velocity.ndim == 1:
 *             for i in range(nx*nz):
 *                 data.push_back(1./velocity[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_velocity, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_2, 1., 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2101, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2101, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2099
 *             for i in range(nx*nz):
 *                 data.push_back(1./tmp[i])
 *         elif velocity.ndim == 1:             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(1./velocity[i])
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2103
 *                 data.push_back(1./velocity[i])
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2103, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2104
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')
 *         self.grid.setSlowness(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_xi(self, xi):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2104, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2071
 *         self.grid.setSlowness(data)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_velocity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2106
 *         self.grid.setSlowness(data)
 * 
 *     def set_xi(self, xi):             # <<<<<<<<<<<<<<
 *         """
 *         set_xi(xi)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_21set_xi(PyObject *__pyx_v_self, PyObject *__pyx_v_xi); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_20set_xi[] = "\n        set_xi(xi)\n\n        Assign elliptical anisotropy ratio to grid\n\n        Parameters\n        ----------\n        xi : np ndarray, shape (nx, nz)\n            xi may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_21set_xi(PyObject *__pyx_v_self, PyObject *__pyx_v_xi) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_xi (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_20set_xi(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_xi));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_20set_xi(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_xi) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_xi", 0);

  /* "ttcrpy/rgrid.pyx":2117
 *             xi may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2118
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2119
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2117
 *             xi may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2121
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if xi.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2122
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if xi.size != nx*nz:
 *             raise ValueError('xi vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2123
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if xi.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('xi vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2124
 *             nz = self._z.size()
 *         if xi.size != nx*nz:
 *             raise ValueError('xi vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2124, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2123
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if xi.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('xi vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2128
 *         cdef vector[double] data
 *         cdef int i
 *         if xi.ndim == 2:             # <<<<<<<<<<<<<<
 *             if xi.shape != (nx, nz):
 *                 raise ValueError('xi has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2129
 *         cdef int i
 *         if xi.ndim == 2:
 *             if xi.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('xi has wrong shape')
 *             tmp = xi.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2130
 *         if xi.ndim == 2:
 *             if xi.shape != (nx, nz):
 *                 raise ValueError('xi has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = xi.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2130, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2129
 *         cdef int i
 *         if xi.ndim == 2:
 *             if xi.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('xi has wrong shape')
 *             tmp = xi.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2131
 *             if xi.shape != (nx, nz):
 *                 raise ValueError('xi has wrong shape')
 *             tmp = xi.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2132
 *                 raise ValueError('xi has wrong shape')
 *             tmp = xi.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif xi.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2133
 *             tmp = xi.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif xi.ndim == 1:
 *             tmp = xi.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2133, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2133, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2128
 *         cdef vector[double] data
 *         cdef int i
 *         if xi.ndim == 2:             # <<<<<<<<<<<<<<
 *             if xi.shape != (nx, nz):
 *                 raise ValueError('xi has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2134
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif xi.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = xi.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2135
 *                 data.push_back(tmp[i])
 *         elif xi.ndim == 1:
 *             tmp = xi.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_xi, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2136
 *         elif xi.ndim == 1:
 *             tmp = xi.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2136, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2137
 *             tmp = xi.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('xi must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2137, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2137, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2134
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif xi.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = xi.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2139
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('xi must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setXi(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2139, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2140
 *         else:
 *             raise ValueError('xi must be 1D or 3D ndarray')
 *         self.grid.setXi(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_tilt_angle(self, theta):
 */
  try {
    __pyx_v_self->grid->setXi(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2140, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2106
 *         self.grid.setSlowness(data)
 * 
 *     def set_xi(self, xi):             # <<<<<<<<<<<<<<
 *         """
 *         set_xi(xi)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_xi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2142
 *         self.grid.setXi(data)
 * 
 *     def set_tilt_angle(self, theta):             # <<<<<<<<<<<<<<
 *         """
 *         set_tilt_angle(theta)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_23set_tilt_angle(PyObject *__pyx_v_self, PyObject *__pyx_v_theta); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_22set_tilt_angle[] = "\n        set_tilt_angle(theta)\n\n        Assign tilted elliptical anisotropy angle to grid\n\n        Parameters\n        ----------\n        theta : np ndarray, shape (nx, nz)\n            theta may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_23set_tilt_angle(PyObject *__pyx_v_self, PyObject *__pyx_v_theta) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_tilt_angle (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_22set_tilt_angle(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_theta));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_22set_tilt_angle(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_theta) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_tilt_angle", 0);

  /* "ttcrpy/rgrid.pyx":2153
 *             theta may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2154
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2155
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2153
 *             theta may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2157
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if theta.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2158
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if theta.size != nx*nz:
 *             raise ValueError('theta vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2159
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if theta.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('theta vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2160
 *             nz = self._z.size()
 *         if theta.size != nx*nz:
 *             raise ValueError('theta vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2160, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2159
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if theta.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('theta vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2164
 *         cdef vector[double] data
 *         cdef int i
 *         if theta.ndim == 2:             # <<<<<<<<<<<<<<
 *             if theta.shape != (nx, nz):
 *                 raise ValueError('theta has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2165
 *         cdef int i
 *         if theta.ndim == 2:
 *             if theta.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('theta has wrong shape')
 *             tmp = theta.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2166
 *         if theta.ndim == 2:
 *             if theta.shape != (nx, nz):
 *                 raise ValueError('theta has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = theta.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2166, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2165
 *         cdef int i
 *         if theta.ndim == 2:
 *             if theta.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('theta has wrong shape')
 *             tmp = theta.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2167
 *             if theta.shape != (nx, nz):
 *                 raise ValueError('theta has wrong shape')
 *             tmp = theta.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2168
 *                 raise ValueError('theta has wrong shape')
 *             tmp = theta.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif theta.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2168, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2169
 *             tmp = theta.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif theta.ndim == 1:
 *             tmp = theta.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2169, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2164
 *         cdef vector[double] data
 *         cdef int i
 *         if theta.ndim == 2:             # <<<<<<<<<<<<<<
 *             if theta.shape != (nx, nz):
 *                 raise ValueError('theta has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2170
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif theta.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = theta.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2171
 *                 data.push_back(tmp[i])
 *         elif theta.ndim == 1:
 *             tmp = theta.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_theta, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2172
 *         elif theta.ndim == 1:
 *             tmp = theta.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2172, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2173
 *             tmp = theta.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('theta must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2173, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2170
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif theta.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = theta.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2175
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('theta must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setTiltAngle(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2175, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2176
 *         else:
 *             raise ValueError('theta must be 1D or 3D ndarray')
 *         self.grid.setTiltAngle(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_Vp0(self, v):
 */
  try {
    __pyx_v_self->grid->setTiltAngle(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2176, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2142
 *         self.grid.setXi(data)
 * 
 *     def set_tilt_angle(self, theta):             # <<<<<<<<<<<<<<
 *         """
 *         set_tilt_angle(theta)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_tilt_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2178
 *         self.grid.setTiltAngle(data)
 * 
 *     def set_Vp0(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_Vp0(v)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_25set_Vp0(PyObject *__pyx_v_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_24set_Vp0[] = "\n        set_Vp0(v)\n\n        Assign vertical Vp to grid (VTI medium)\n\n        Parameters\n        ----------\n        v : np ndarray, shape (nx, nz)\n            v may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_25set_Vp0(PyObject *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_Vp0 (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_24set_Vp0(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_24set_Vp0(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_Vp0", 0);

  /* "ttcrpy/rgrid.pyx":2189
 *             v may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2190
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2191
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2189
 *             v may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2193
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2194
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2195
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2196
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2196, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2195
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2200
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2201
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2202
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2202, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2201
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2203
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2204
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2205
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2205, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2205, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2200
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2206
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2207
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2208
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2209
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2209, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2209, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2206
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2211
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setVp0(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2211, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2212
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 *         self.grid.setVp0(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_Vs0(self, v):
 */
  try {
    __pyx_v_self->grid->setVp0(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2212, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2178
 *         self.grid.setTiltAngle(data)
 * 
 *     def set_Vp0(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_Vp0(v)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_Vp0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2214
 *         self.grid.setVp0(data)
 * 
 *     def set_Vs0(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_Vs0(v)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_27set_Vs0(PyObject *__pyx_v_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_26set_Vs0[] = "\n        set_Vs0(v)\n\n        Assign vertical Vs to grid (VTI medium)\n\n        Parameters\n        ----------\n        v : np ndarray, shape (nx, nz)\n            v may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_27set_Vs0(PyObject *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_Vs0 (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_26set_Vs0(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_26set_Vs0(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_Vs0", 0);

  /* "ttcrpy/rgrid.pyx":2225
 *             v may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2226
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2227
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2225
 *             v may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2229
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2230
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2231
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2232
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2232, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2231
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2236
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2237
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2238
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2238, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2237
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2239
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2240
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2240, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2241
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2241, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2241, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2236
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2242
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2243
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2244
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2245
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2245, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2245, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2242
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2247
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setVs0(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2247, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2248
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 *         self.grid.setVs0(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_delta(self, v):
 */
  try {
    __pyx_v_self->grid->setVs0(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2248, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2214
 *         self.grid.setVp0(data)
 * 
 *     def set_Vs0(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_Vs0(v)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_Vs0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2250
 *         self.grid.setVs0(data)
 * 
 *     def set_delta(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_delta(d)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_29set_delta(PyObject *__pyx_v_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_28set_delta[] = "\n        set_delta(d)\n\n        Assign Thomsen delta parameter to grid (VTI medium, P-SV waves)\n\n        Parameters\n        ----------\n        d : np ndarray, shape (nx, nz)\n            d may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_29set_delta(PyObject *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_delta (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_28set_delta(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_28set_delta(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_delta", 0);

  /* "ttcrpy/rgrid.pyx":2261
 *             d may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2262
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2263
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2261
 *             d may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2265
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2266
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2267
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2268
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2268, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2267
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2272
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2273
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2273, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2273, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2274
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2274, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2273
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2275
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2276
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2277
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2277, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2277, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2272
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2278
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2279
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2280
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2280, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2281
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2281, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2281, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2278
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2283
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setDelta(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2283, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2284
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 *         self.grid.setDelta(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_epsilon(self, v):
 */
  try {
    __pyx_v_self->grid->setDelta(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2284, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2250
 *         self.grid.setVs0(data)
 * 
 *     def set_delta(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_delta(d)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2286
 *         self.grid.setDelta(data)
 * 
 *     def set_epsilon(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_epsilon(e)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_31set_epsilon(PyObject *__pyx_v_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_30set_epsilon[] = "\n        set_epsilon(e)\n\n        Assign Thomsen epsilon parameter to grid (VTI medium, P-SV waves)\n\n        Parameters\n        ----------\n        e : np ndarray, shape (nx, nz)\n            e may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_31set_epsilon(PyObject *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_epsilon (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_30set_epsilon(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_30set_epsilon(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_epsilon", 0);

  /* "ttcrpy/rgrid.pyx":2297
 *             e may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2298
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2299
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2297
 *             e may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2301
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2302
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2303
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2304
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2304, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2303
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2308
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2309
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2310
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2310, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2309
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2311
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2312
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2312, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2313
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2313, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2313, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2308
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2314
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2315
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2316
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2317
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2317, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2317, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2314
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2319
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setEpsilon(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2319, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2320
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 *         self.grid.setEpsilon(data)             # <<<<<<<<<<<<<<
 * 
 *     def set_gamma(self, v):
 */
  try {
    __pyx_v_self->grid->setEpsilon(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2320, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2286
 *         self.grid.setDelta(data)
 * 
 *     def set_epsilon(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_epsilon(e)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2322
 *         self.grid.setEpsilon(data)
 * 
 *     def set_gamma(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_gamma(g)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_33set_gamma(PyObject *__pyx_v_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_32set_gamma[] = "\n        set_gamma(g)\n\n        Assign Thomsen gamma parameter to grid (VTI medium, SH waves)\n\n        Parameters\n        ----------\n        g : np ndarray, shape (nx, nz)\n            g may also have been flattened (with default 'C' order)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_33set_gamma(PyObject *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_gamma (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_32set_gamma(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_32set_gamma(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  std::vector<double>  __pyx_v_data;
  int __pyx_v_i;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_gamma", 0);

  /* "ttcrpy/rgrid.pyx":2333
 *             g may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2334
 *         """
 *         if self.cell_slowness:
 *             nx = self._x.size()-1             # <<<<<<<<<<<<<<
 *             nz = self._z.size()-1
 *         else:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2335
 *         if self.cell_slowness:
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1             # <<<<<<<<<<<<<<
 *         else:
 *             nx = self._x.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2333
 *             g may also have been flattened (with default 'C' order)
 *         """
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             nx = self._x.size()-1
 *             nz = self._z.size()-1
 */
    goto __pyx_L3;
  }

  /* "ttcrpy/rgrid.pyx":2337
 *             nz = self._z.size()-1
 *         else:
 *             nx = self._x.size()             # <<<<<<<<<<<<<<
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2338
 *         else:
 *             nx = self._x.size()
 *             nz = self._z.size()             # <<<<<<<<<<<<<<
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nz = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "ttcrpy/rgrid.pyx":2339
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2340
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2340, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2339
 *             nx = self._x.size()
 *             nz = self._z.size()
 *         if v.size != nx*nz:             # <<<<<<<<<<<<<<
 *             raise ValueError('v vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2344
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2345
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nz);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2345, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2345, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2346
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2346, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2345
 *         cdef int i
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):             # <<<<<<<<<<<<<<
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 */
    }

    /* "ttcrpy/rgrid.pyx":2347
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2348
 *                 raise ValueError('v has wrong shape')
 *             tmp = v.flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2348, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2349
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2349, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2349, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2344
 *         cdef vector[double] data
 *         cdef int i
 *         if v.ndim == 2:             # <<<<<<<<<<<<<<
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2350
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2351
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()             # <<<<<<<<<<<<<<
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_nx);
    __Pyx_GIVEREF(__pyx_v_nx);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nx);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_nz);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2352
 *         elif v.ndim == 1:
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):             # <<<<<<<<<<<<<<
 *                 data.push_back(tmp[i])
 *         else:
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2352, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "ttcrpy/rgrid.pyx":2353
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_tmp, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2353, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      try {
        __pyx_v_data.push_back(__pyx_t_8);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2353, __pyx_L1_error)
      }
    }

    /* "ttcrpy/rgrid.pyx":2350
 *             for i in range(nx*nz):
 *                 data.push_back(tmp[i])
 *         elif v.ndim == 1:             # <<<<<<<<<<<<<<
 *             tmp = v.reshape((nx, nz)).flatten()
 *             for i in range(nx*nz):
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2355
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setGamma(data)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2355, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2356
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')
 *         self.grid.setGamma(data)             # <<<<<<<<<<<<<<
 * 
 *     def compute_K(self, order=1):
 */
  try {
    __pyx_v_self->grid->setGamma(__pyx_v_data);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2356, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2322
 *         self.grid.setEpsilon(data)
 * 
 *     def set_gamma(self, v):             # <<<<<<<<<<<<<<
 *         """
 *         set_gamma(g)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.set_gamma", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2358
 *         self.grid.setGamma(data)
 * 
 *     def compute_K(self, order=1):             # <<<<<<<<<<<<<<
 *         """
 *         Compute smoothing matrices
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_35compute_K(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_34compute_K[] = "\n        Compute smoothing matrices\n\n        Parameters\n        ----------\n        order : int\n            order of smoothing operator, accept 1 or 2 (1 by default)\n\n        Returns\n        -------\n        Kx, Kz : :obj:`tuple` of :obj:`csr_matrix`\n            matrices for derivatives along x & z\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_35compute_K(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_order = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_K (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_order,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_1);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_K") < 0)) __PYX_ERR(0, 2358, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_order = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_K", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2358, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.compute_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_34compute_K(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_order);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_34compute_K(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_order) {
  PyObject *__pyx_v_nx = NULL;
  PyObject *__pyx_v_nz = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_idz = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_jj = NULL;
  PyObject *__pyx_v_vd = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_Kx = NULL;
  PyObject *__pyx_v_Kz = NULL;
  PyObject *__pyx_v_idx2 = NULL;
  PyObject *__pyx_v_idz2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_K", 0);

  /* "ttcrpy/rgrid.pyx":2373
 *         """
 * 
 *         nx, nz = self.shape             # <<<<<<<<<<<<<<
 *         if order == 1:
 *             # forward operator is (u_{i+1} - u_i)/dx
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 2373, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 2373, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 2373, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_nx = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nz = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2374
 * 
 *         nx, nz = self.shape
 *         if order == 1:             # <<<<<<<<<<<<<<
 *             # forward operator is (u_{i+1} - u_i)/dx
 *             # centered operator is (u_{i+1} - u_{i-1})/(2dx)
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_order, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "ttcrpy/rgrid.pyx":2379
 *             # backward operator is (u_i - u_{i-1})/dx
 * 
 *             idx = 1 / self.dx             # <<<<<<<<<<<<<<
 *             idz = 1 / self.dz
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_idx = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2380
 * 
 *             idx = 1 / self.dx
 *             idz = 1 / self.dz             # <<<<<<<<<<<<<<
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((2, ), dtype=np.int32))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_idz = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2382
 *             idz = 1 / self.dz
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((2, ), dtype=np.int32))             # <<<<<<<<<<<<<<
 *             j = np.zeros((nz * nx * 2, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 2, ))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_kron); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ones); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__50, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2382, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2382, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_10, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_10, __pyx_t_9);
      __pyx_t_3 = 0;
      __pyx_t_9 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2383
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((2, ), dtype=np.int32))
 *             j = np.zeros((nz * nx * 2, ), dtype=np.int32)             # <<<<<<<<<<<<<<
 *             v = np.zeros((nz * nx * 2, ))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_v_nz, __pyx_v_nx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_int_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_int32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_j = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2384
 *             i = np.kron(np.arange(nx * nz), np.ones((2, ), dtype=np.int32))
 *             j = np.zeros((nz * nx * 2, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 2, ))             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz))).T
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_v_nz, __pyx_v_nx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_v = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2386
 *             v = np.zeros((nz * nx * 2, ))
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             j[:2 * nz] = jj
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vstack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_2, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_9 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_v_nz, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_jj = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2387
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             j[:2 * nz] = jj
 *             vd = idx * np.tile(np.array([-1, 1]), (nz, ))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2388
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz))).T
 *             jj = jj.flatten()
 *             j[:2 * nz] = jj             # <<<<<<<<<<<<<<
 *             vd = idx * np.tile(np.array([-1, 1]), (nz, ))
 *             v[:2 * nz] = vd
 */
    __pyx_t_1 = PyNumber_Multiply(__pyx_int_2, __pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_v_jj, 0, 0, NULL, &__pyx_t_1, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2388, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2389
 *             jj = jj.flatten()
 *             j[:2 * nz] = jj
 *             vd = idx * np.tile(np.array([-1, 1]), (nz, ))             # <<<<<<<<<<<<<<
 *             v[:2 * nz] = vd
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_tile); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_7, 0, __pyx_int_neg_1);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyList_SET_ITEM(__pyx_t_7, 1, __pyx_int_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_nz);
    __pyx_t_7 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_10, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_10, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Multiply(__pyx_v_idx, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_vd = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2390
 *             j[:2 * nz] = jj
 *             vd = idx * np.tile(np.array([-1, 1]), (nz, ))
 *             v[:2 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((-nz + np.arange(nz), nz + np.arange(nz))).T
 */
    __pyx_t_9 = PyNumber_Multiply(__pyx_int_2, __pyx_v_nz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, NULL, &__pyx_t_9, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2390, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2392
 *             v[:2 * nz] = vd
 * 
 *             jj = np.vstack((-nz + np.arange(nz), nz + np.arange(nz))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             for n in range(1, nx - 1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Negative(__pyx_v_nz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_3, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_v_nz, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_9 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_T); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2393
 * 
 *             jj = np.vstack((-nz + np.arange(nz), nz + np.arange(nz))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             for n in range(1, nx - 1):
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2394
 *             jj = np.vstack((-nz + np.arange(nz), nz + np.arange(nz))).T
 *             jj = jj.flatten()
 *             for n in range(1, nx - 1):             # <<<<<<<<<<<<<<
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = 0.5 * vd
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_9 = __pyx_t_2; __Pyx_INCREF(__pyx_t_9); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_12 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2394, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2394, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_9, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2394, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_9, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_12(__pyx_t_9);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2394, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2395
 *             jj = jj.flatten()
 *             for n in range(1, nx - 1):
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj             # <<<<<<<<<<<<<<
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = 0.5 * vd
 * 
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_v_jj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_2, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_4, 0, 0, &__pyx_t_3, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "ttcrpy/rgrid.pyx":2396
 *             for n in range(1, nx - 1):
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = 0.5 * vd             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((-nz + np.arange(nz), np.arange(nz))).T
 */
      __pyx_t_4 = PyNumber_Multiply(__pyx_float_0_5, __pyx_v_vd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_2, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_t_4, 0, 0, &__pyx_t_3, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2396, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "ttcrpy/rgrid.pyx":2394
 *             jj = np.vstack((-nz + np.arange(nz), nz + np.arange(nz))).T
 *             jj = jj.flatten()
 *             for n in range(1, nx - 1):             # <<<<<<<<<<<<<<
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = 0.5 * vd
 */
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2398
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = 0.5 * vd
 * 
 *             jj = np.vstack((-nz + np.arange(nz), np.arange(nz))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             j[(nx - 1) * 2 * nz:nx * 2 * nz] = (nx - 1) * nz + jj
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Negative(__pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_arange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_9 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_T); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2399
 * 
 *             jj = np.vstack((-nz + np.arange(nz), np.arange(nz))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             j[(nx - 1) * 2 * nz:nx * 2 * nz] = (nx - 1) * nz + jj
 *             v[(nx - 1) * 2 * nz:nx * 2 * nz] = vd
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2400
 *             jj = np.vstack((-nz + np.arange(nz), np.arange(nz))).T
 *             jj = jj.flatten()
 *             j[(nx - 1) * 2 * nz:nx * 2 * nz] = (nx - 1) * nz + jj             # <<<<<<<<<<<<<<
 *             v[(nx - 1) * 2 * nz:nx * 2 * nz] = vd
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_9, __pyx_v_jj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_9, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Multiply(__pyx_v_nx, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_2, 0, 0, &__pyx_t_9, &__pyx_t_3, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2401
 *             jj = jj.flatten()
 *             j[(nx - 1) * 2 * nz:nx * 2 * nz] = (nx - 1) * nz + jj
 *             v[(nx - 1) * 2 * nz:nx * 2 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             Kx = sp.csr_matrix((v, (i, j)))
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_int_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_int_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, &__pyx_t_2, &__pyx_t_9, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2403
 *             v[(nx - 1) * 2 * nz:nx * 2 * nz] = vd
 * 
 *             Kx = sp.csr_matrix((v, (i, j)))             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_j);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_v);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_9 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_Kx = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2405
 *             Kx = sp.csr_matrix((v, (i, j)))
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),             # <<<<<<<<<<<<<<
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T
 *             jj = jj.flatten()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vstack); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_hstack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_13, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2406
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             vd = idz * np.hstack((np.array([-1, 1]),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_hstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_arange); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_int_1, __pyx_v_nz};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2406, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_8);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_int_1, __pyx_v_nz};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2406, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_8);
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, __pyx_int_1);
      __Pyx_INCREF(__pyx_v_nz);
      __Pyx_GIVEREF(__pyx_v_nz);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_v_nz);
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_13);
    __pyx_t_8 = 0;
    __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2405
 *             Kx = sp.csr_matrix((v, (i, j)))
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),             # <<<<<<<<<<<<<<
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T
 *             jj = jj.flatten()
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2406
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             vd = idz * np.hstack((np.array([-1, 1]),
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_T); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2407
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 1))),
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             vd = idz * np.hstack((np.array([-1, 1]),
 *                                   np.tile(np.array([-0.5, 0.5]), (nz - 2,)),
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2408
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T
 *             jj = jj.flatten()
 *             vd = idz * np.hstack((np.array([-1, 1]),             # <<<<<<<<<<<<<<
 *                                   np.tile(np.array([-0.5, 0.5]), (nz - 2,)),
 *                                   np.array([-1, 1])))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_hstack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_int_neg_1);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyList_SET_ITEM(__pyx_t_1, 1, __pyx_int_1);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_9 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_14, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2409
 *             jj = jj.flatten()
 *             vd = idz * np.hstack((np.array([-1, 1]),
 *                                   np.tile(np.array([-0.5, 0.5]), (nz - 2,)),             # <<<<<<<<<<<<<<
 *                                   np.array([-1, 1])))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_tile); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyList_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_float_neg_0_5);
    __Pyx_GIVEREF(__pyx_float_neg_0_5);
    PyList_SET_ITEM(__pyx_t_13, 0, __pyx_float_neg_0_5);
    __Pyx_INCREF(__pyx_float_0_5);
    __Pyx_GIVEREF(__pyx_float_0_5);
    PyList_SET_ITEM(__pyx_t_13, 1, __pyx_float_0_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_13};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_13};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_10, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_10, __pyx_t_13);
      __pyx_t_1 = 0;
      __pyx_t_13 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2410
 *             vd = idz * np.hstack((np.array([-1, 1]),
 *                                   np.tile(np.array([-0.5, 0.5]), (nz - 2,)),
 *                                   np.array([-1, 1])))             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nx):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_int_neg_1);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyList_SET_ITEM(__pyx_t_4, 1, __pyx_int_1);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_14 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2408
 *                             np.hstack((np.arange(1, nz), nz - 1)))).T
 *             jj = jj.flatten()
 *             vd = idz * np.hstack((np.array([-1, 1]),             # <<<<<<<<<<<<<<
 *                                   np.tile(np.array([-0.5, 0.5]), (nz - 2,)),
 *                                   np.array([-1, 1])))
 */
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_t_14);
    __pyx_t_9 = 0;
    __pyx_t_3 = 0;
    __pyx_t_14 = 0;
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_idz, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_vd, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2412
 *                                   np.array([-1, 1])))
 * 
 *             for n in range(nx):             # <<<<<<<<<<<<<<
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = vd
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_nx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_7 = __pyx_t_2; __Pyx_INCREF(__pyx_t_7); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_12 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2412, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2412, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2412, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2412, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2412, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_12(__pyx_t_7);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2412, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2413
 * 
 *             for n in range(nx):
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj             # <<<<<<<<<<<<<<
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = vd
 * 
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_13 = PyNumber_Add(__pyx_t_2, __pyx_v_jj); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_int_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_13, 0, 0, &__pyx_t_14, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "ttcrpy/rgrid.pyx":2414
 *             for n in range(nx):
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))
 */
      __pyx_t_13 = PyNumber_Multiply(__pyx_v_n, __pyx_int_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_13, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = PyNumber_Multiply(__pyx_t_13, __pyx_int_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = PyNumber_Multiply(__pyx_t_14, __pyx_v_nz); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, &__pyx_t_2, &__pyx_t_13, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2414, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "ttcrpy/rgrid.pyx":2412
 *                                   np.array([-1, 1])))
 * 
 *             for n in range(nx):             # <<<<<<<<<<<<<<
 *                 j[n * 2 * nz:(n + 1) * 2 * nz] = n * nz + jj
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = vd
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2416
 *                 v[n * 2 * nz:(n + 1) * 2 * nz] = vd
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))             # <<<<<<<<<<<<<<
 *         elif order == 2:
 *             # forward operator is (u_i - 2u_{i+1} + u_{i+2})/dx^2
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_sp); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_j);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_v);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_13);
    __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_Kz = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2374
 * 
 *         nx, nz = self.shape
 *         if order == 1:             # <<<<<<<<<<<<<<
 *             # forward operator is (u_{i+1} - u_i)/dx
 *             # centered operator is (u_{i+1} - u_{i-1})/(2dx)
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2417
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))
 *         elif order == 2:             # <<<<<<<<<<<<<<
 *             # forward operator is (u_i - 2u_{i+1} + u_{i+2})/dx^2
 *             # centered operator is (u_{i-1} - 2u_i + u_{i+1})/dx^2
 */
  __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_v_order, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 2417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(__pyx_t_6)) {

    /* "ttcrpy/rgrid.pyx":2422
 *             # backward operator is (u_{i-2} - 2u_{i-1} + u_i)/dx^2
 * 
 *             idx2 = 1 / (self.dx * self.dx)             # <<<<<<<<<<<<<<
 *             idz2 = 1 / (self.dz * self.dz)
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = PyNumber_Multiply(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_int_1, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_v_idx2 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2423
 * 
 *             idx2 = 1 / (self.dx * self.dx)
 *             idz2 = 1 / (self.dz * self.dz)             # <<<<<<<<<<<<<<
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((3, ), dtype=np.int32))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_2, __pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyNumber_Divide(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_idz2 = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2425
 *             idz2 = 1 / (self.dz * self.dz)
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((3, ), dtype=np.int32))             # <<<<<<<<<<<<<<
 *             j = np.zeros((nz * nx * 3, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 3, ))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_kron); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyNumber_Multiply(__pyx_v_nx, __pyx_v_nz); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_int32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_tuple__51, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_4};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2425, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_4};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2425, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_10, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_10, __pyx_t_4);
      __pyx_t_7 = 0;
      __pyx_t_4 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_13, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2426
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((3, ), dtype=np.int32))
 *             j = np.zeros((nz * nx * 3, ), dtype=np.int32)             # <<<<<<<<<<<<<<
 *             v = np.zeros((nz * nx * 3, ))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyNumber_Multiply(__pyx_v_nz, __pyx_v_nx); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = PyNumber_Multiply(__pyx_t_14, __pyx_int_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __pyx_t_13 = 0;
    __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_13, __pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_v_j = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2427
 *             i = np.kron(np.arange(nx * nz), np.ones((3, ), dtype=np.int32))
 *             j = np.zeros((nz * nx * 3, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 3, ))             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_zeros); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyNumber_Multiply(__pyx_v_nz, __pyx_v_nx); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_14, __pyx_int_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_2, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_v_v = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2429
 *             v = np.zeros((nz * nx * 3, ))
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),             # <<<<<<<<<<<<<<
 *                             2 * nz + np.arange(nz))).T
 *             jj = jj.flatten()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_vstack); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_13 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_arange); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_v_nz, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2430
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),
 *                             2 * nz + np.arange(nz))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             j[:3 * nz] = jj
 */
    __pyx_t_4 = PyNumber_Multiply(__pyx_int_2, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_arange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_nz) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_nz);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2429
 *             v = np.zeros((nz * nx * 3, ))
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),             # <<<<<<<<<<<<<<
 *                             2 * nz + np.arange(nz))).T
 *             jj = jj.flatten()
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
    __pyx_t_13 = 0;
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
      }
    }
    __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2430
 * 
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),
 *                             2 * nz + np.arange(nz))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             j[:3 * nz] = jj
 */
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_jj = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2431
 *             jj = np.vstack((np.arange(nz), nz + np.arange(nz),
 *                             2 * nz + np.arange(nz))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             j[:3 * nz] = jj
 *             vd = idx2 * np.tile(np.array([1.0, -2.0, 1.0]), (nz, ))
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_14 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2432
 *                             2 * nz + np.arange(nz))).T
 *             jj = jj.flatten()
 *             j[:3 * nz] = jj             # <<<<<<<<<<<<<<
 *             vd = idx2 * np.tile(np.array([1.0, -2.0, 1.0]), (nz, ))
 *             v[:3 * nz] = vd
 */
    __pyx_t_14 = PyNumber_Multiply(__pyx_int_3, __pyx_v_nz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_v_jj, 0, 0, NULL, &__pyx_t_14, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2433
 *             jj = jj.flatten()
 *             j[:3 * nz] = jj
 *             vd = idx2 * np.tile(np.array([1.0, -2.0, 1.0]), (nz, ))             # <<<<<<<<<<<<<<
 *             v[:3 * nz] = vd
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_tile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_float_1_0);
    __Pyx_GIVEREF(__pyx_float_1_0);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_float_1_0);
    __Pyx_INCREF(__pyx_float_neg_2_0);
    __Pyx_GIVEREF(__pyx_float_neg_2_0);
    PyList_SET_ITEM(__pyx_t_1, 1, __pyx_float_neg_2_0);
    __Pyx_INCREF(__pyx_float_1_0);
    __Pyx_GIVEREF(__pyx_float_1_0);
    PyList_SET_ITEM(__pyx_t_1, 2, __pyx_float_1_0);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_7 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_13, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_nz);
    __Pyx_GIVEREF(__pyx_v_nz);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nz);
    __pyx_t_1 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_7, __pyx_t_3};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2433, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_7, __pyx_t_3};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2433, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_10, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_10, __pyx_t_3);
      __pyx_t_7 = 0;
      __pyx_t_3 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_13, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_idx2, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_v_vd = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2434
 *             j[:3 * nz] = jj
 *             vd = idx2 * np.tile(np.array([1.0, -2.0, 1.0]), (nz, ))
 *             v[:3 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             for n in range(1, nx - 1):
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_int_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, NULL, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2434, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2436
 *             v[:3 * nz] = vd
 * 
 *             for n in range(1, nx - 1):             # <<<<<<<<<<<<<<
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = (n - 1) * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_14 = __pyx_t_2; __Pyx_INCREF(__pyx_t_14); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_12 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2436, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_14))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_14)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2436, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_14, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2436, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_14, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_12(__pyx_t_14);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2436, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2437
 * 
 *             for n in range(1, nx - 1):
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = (n - 1) * nz + jj             # <<<<<<<<<<<<<<
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_13 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_t_13, __pyx_v_jj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = PyNumber_Multiply(__pyx_v_n, __pyx_int_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_13, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_13, __pyx_int_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_2, 0, 0, &__pyx_t_3, &__pyx_t_13, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2437, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2438
 *             for n in range(1, nx - 1):
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = (n - 1) * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             j[(nx - 1) * 3 * nz:nx * 3 * nz] = (nx - 3) * nz + jj
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_int_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_13 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_int_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, &__pyx_t_13, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2438, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2436
 *             v[:3 * nz] = vd
 * 
 *             for n in range(1, nx - 1):             # <<<<<<<<<<<<<<
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = (n - 1) * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 */
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2440
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 * 
 *             j[(nx - 1) * 3 * nz:nx * 3 * nz] = (nx - 3) * nz + jj             # <<<<<<<<<<<<<<
 *             v[(nx - 1) * 3 * nz:nx * 3 * nz] = vd
 * 
 */
    __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_14, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyNumber_Add(__pyx_t_2, __pyx_v_jj); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyNumber_Multiply(__pyx_t_2, __pyx_int_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_13, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyNumber_Multiply(__pyx_v_nx, __pyx_int_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_13, __pyx_v_nz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_14, 0, 0, &__pyx_t_2, &__pyx_t_3, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2440, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "ttcrpy/rgrid.pyx":2441
 * 
 *             j[(nx - 1) * 3 * nz:nx * 3 * nz] = (nx - 3) * nz + jj
 *             v[(nx - 1) * 3 * nz:nx * 3 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             Kx = sp.csr_matrix((v, (i, j)))
 */
    __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_nx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_3 = PyNumber_Multiply(__pyx_t_14, __pyx_int_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Multiply(__pyx_v_nx, __pyx_int_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_3, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, &__pyx_t_14, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2443
 *             v[(nx - 1) * 3 * nz:nx * 3 * nz] = vd
 * 
 *             Kx = sp.csr_matrix((v, (i, j)))             # <<<<<<<<<<<<<<
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_sp); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_j);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_v);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_14);
    __pyx_t_14 = 0;
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_Kx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2445
 *             Kx = sp.csr_matrix((v, (i, j)))
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),             # <<<<<<<<<<<<<<
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vstack); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_hstack); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_14 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
    __pyx_t_14 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "ttcrpy/rgrid.pyx":2446
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),             # <<<<<<<<<<<<<<
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T
 *             jj = jj.flatten()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_hstack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_arange); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_8 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_1, __pyx_t_14};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2446, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_1, __pyx_t_14};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2446, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else
    #endif
    {
      __pyx_t_15 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_10, __pyx_int_1);
      __Pyx_GIVEREF(__pyx_t_14);
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_10, __pyx_t_14);
      __pyx_t_14 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_9);
    __pyx_t_1 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ttcrpy/rgrid.pyx":2447
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             vd = vd * idz2 / idx2
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_hstack); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_int_2, __pyx_v_nz};
      __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_15);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_int_2, __pyx_v_nz};
      __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_15);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_10, __pyx_int_2);
      __Pyx_INCREF(__pyx_v_nz);
      __Pyx_GIVEREF(__pyx_v_nz);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_10, __pyx_v_nz);
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_8, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_nz, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_int_2);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_14);
    __pyx_t_15 = 0;
    __pyx_t_14 = 0;
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_4 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_14, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "ttcrpy/rgrid.pyx":2445
 *             Kx = sp.csr_matrix((v, (i, j)))
 * 
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),             # <<<<<<<<<<<<<<
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T
 */
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_7 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2447
 *             jj = np.vstack((np.hstack((0, np.arange(nz - 2), nz - 3)),
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T             # <<<<<<<<<<<<<<
 *             jj = jj.flatten()
 *             vd = vd * idz2 / idx2
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_13);
    __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2448
 *                             np.hstack((1, np.arange(1, nz - 1), nz - 2)),
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T
 *             jj = jj.flatten()             # <<<<<<<<<<<<<<
 *             vd = vd * idz2 / idx2
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_jj, __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_13 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_jj, __pyx_t_13);
    __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2449
 *                             np.hstack((2, np.arange(2, nz), nz - 1)))).T
 *             jj = jj.flatten()
 *             vd = vd * idz2 / idx2             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nx):
 */
    __pyx_t_13 = PyNumber_Multiply(__pyx_v_vd, __pyx_v_idz2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_13, __pyx_v_idx2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF_SET(__pyx_v_vd, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2451
 *             vd = vd * idz2 / idx2
 * 
 *             for n in range(nx):             # <<<<<<<<<<<<<<
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = n * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_nx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_13 = __pyx_t_2; __Pyx_INCREF(__pyx_t_13); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_13 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_12 = Py_TYPE(__pyx_t_13)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2451, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_13))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_13)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_13, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2451, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_13, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_13)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_13, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 2451, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_13, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_12(__pyx_t_13);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2451, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2452
 * 
 *             for n in range(nx):
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = n * nz + jj             # <<<<<<<<<<<<<<
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 * 
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_v_jj); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_int_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_v_nz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_2, __pyx_int_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_7, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_j, __pyx_t_9, 0, 0, &__pyx_t_4, &__pyx_t_2, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2452, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ttcrpy/rgrid.pyx":2453
 *             for n in range(nx):
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = n * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd             # <<<<<<<<<<<<<<
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))
 */
      __pyx_t_9 = PyNumber_Multiply(__pyx_v_n, __pyx_int_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_9, __pyx_v_nz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = PyNumber_Multiply(__pyx_t_9, __pyx_int_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Multiply(__pyx_t_4, __pyx_v_nz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_v, __pyx_v_vd, 0, 0, &__pyx_t_2, &__pyx_t_9, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 2453, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ttcrpy/rgrid.pyx":2451
 *             vd = vd * idz2 / idx2
 * 
 *             for n in range(nx):             # <<<<<<<<<<<<<<
 *                 j[n * 3 * nz:(n + 1) * 3 * nz] = n * nz + jj
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 */
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2455
 *                 v[n * 3 * nz:(n + 1) * 3 * nz] = vd
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_sp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_j);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_v);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_13 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_Kz = __pyx_t_13;
    __pyx_t_13 = 0;

    /* "ttcrpy/rgrid.pyx":2417
 * 
 *             Kz = sp.csr_matrix((v, (i, j)))
 *         elif order == 2:             # <<<<<<<<<<<<<<
 *             # forward operator is (u_i - 2u_{i+1} + u_{i+2})/dx^2
 *             # centered operator is (u_{i-1} - 2u_i + u_{i+1})/dx^2
 */
    goto __pyx_L5;
  }

  /* "ttcrpy/rgrid.pyx":2458
 * 
 *         else:
 *             raise ValueError('order value not valid (1 or 2 accepted)')             # <<<<<<<<<<<<<<
 * 
 *         return Kx, Kz
 */
  /*else*/ {
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_Raise(__pyx_t_13, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __PYX_ERR(0, 2458, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "ttcrpy/rgrid.pyx":2460
 *             raise ValueError('order value not valid (1 or 2 accepted)')
 * 
 *         return Kx, Kz             # <<<<<<<<<<<<<<
 * 
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_Kx);
  __Pyx_GIVEREF(__pyx_v_Kx);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_Kx);
  __Pyx_INCREF(__pyx_v_Kz);
  __Pyx_GIVEREF(__pyx_v_Kz);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_Kz);
  __pyx_r = __pyx_t_13;
  __pyx_t_13 = 0;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":2358
 *         self.grid.setGamma(data)
 * 
 *     def compute_K(self, order=1):             # <<<<<<<<<<<<<<
 *         """
 *         Compute smoothing matrices
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.compute_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nx);
  __Pyx_XDECREF(__pyx_v_nz);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_idz);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_jj);
  __Pyx_XDECREF(__pyx_v_vd);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_Kx);
  __Pyx_XDECREF(__pyx_v_Kz);
  __Pyx_XDECREF(__pyx_v_idx2);
  __Pyx_XDECREF(__pyx_v_idz2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2462
 *         return Kx, Kz
 * 
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,             # <<<<<<<<<<<<<<
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,
 *                  thread_no=None, aggregate_src=False, compute_L=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_37raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_36raytrace[] = "\n        raytrace(source, rcv, slowness=None, xi=None, theta=None, Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None, thread_no=None, aggregate_src=False, compute_L=False, return_rays=False) -> tt, rays, L\n\n        Perform raytracing\n\n        Parameters\n        ----------\n        source : 2D np.ndarray with 2 or 3 columns\n            see notes below\n        rcv : 2D np.ndarray with 2 columns\n            Columns correspond to x, y and z coordinates\n        slowness : np ndarray, shape (nx, nz) (None by default)\n            slowness at grid nodes or cells (depending on cell_slowness)\n            slowness may also have been flattened (with default 'C' order)\n            if None, slowness must have been assigned previously\n        xi : np ndarray, shape (nx, nz) (None by default)\n            xi at grid cells (only for SPM & cell_slowness=True)\n            xi may also have been flattened (with default 'C' order)\n            if None, xi must have been assigned previously\n        theta : np ndarray, shape (nx, nz) (None by default)\n            theta at grid cells (only for SPM & cell_slowness=True)\n            theta may also have been flattened (with default 'C' order)\n            if None, theta must have been assigned previously\n        Vp0 : np ndarray, shape (nx, nz) (None by default)\n            Vp0 at grid cells (only for SPM & cell_slowness=True)\n            Vp0 may also have been flattened (with default 'C' order)\n            if None, Vp0 must have been assigned previously\n        Vs0 : np ndarray, shape (nx, nz) (None by default)\n            Vs0 at grid cells (only for SPM & cell_slowness=True)\n            Vs0 may also have been flattened (with default 'C' order)\n            if None, Vs0 must have been assigned previously\n        delta : np ndarray, shape (nx, nz) (None by default)\n            delta at grid cells (only for SPM & cell_slowness=True)\n            delta may also have been flattened (with default 'C' order)\n     ""       if None, delta must have been assigned previously\n        epsilon : np ndarray, shape (nx, nz) (None by default)\n            epsilon at grid cells (only for SPM & cell_slowness=True)\n            epsilon may also have been flattened (with default 'C' order)\n            if None, epsilon must have been assigned previously\n        gamma : np ndarray, shape (nx, nz) (None by default)\n            gamma at grid cells (only for SPM & cell_slowness=True)\n            gamma may also have been flattened (with default 'C' order)\n            if None, gamma must have been assigned previously\n        thread_no : int (None by default)\n            Perform calculations in thread number \"thread_no\"\n            if None, attempt to run in parallel if warranted by number of\n            sources and value of n_threads in constructor\n        aggregate_src : bool (False by default)\n            if True, all source coordinates belong to a single event\n        compute_L : bool (False by default)\n            Compute matrices of partial derivative of travel time w/r to slowness\n        return_rays : bool (False by default)\n            Return raypaths\n\n        Returns\n        -------\n        tt : np.ndarray\n            travel times for the appropriate source-rcv  (see Notes below)\n        rays : :obj:`list` of :obj:`np.ndarray`\n            Coordinates of segments forming raypaths (if return_rays is True)\n        L : scipy csr_matrix\n            Matrix of partial derivative of travel time w/r to slowness\n\n        Notes\n        -----\n        If source has 2 columns:\n            - Columns correspond to x and z coordinates\n            - Origin time (t0) is 0 for all points\n        If source has 3 columns:\n            - 1st column corresponds to origin times\n            - 2nd & 3rd columns correspond to x and z coordinates\n\n        source and rcv can contain the same number of rows, each row\n        corresponding to a source-receiver pair, or the number of"" rows may\n        differ if aggregate_src is True or if all rows in source are identical.\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_37raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_rcv = 0;
  PyObject *__pyx_v_slowness = 0;
  PyObject *__pyx_v_xi = 0;
  PyObject *__pyx_v_theta = 0;
  PyObject *__pyx_v_Vp0 = 0;
  PyObject *__pyx_v_Vs0 = 0;
  PyObject *__pyx_v_delta = 0;
  PyObject *__pyx_v_epsilon = 0;
  PyObject *__pyx_v_gamma = 0;
  PyObject *__pyx_v_thread_no = 0;
  PyObject *__pyx_v_aggregate_src = 0;
  PyObject *__pyx_v_compute_L = 0;
  PyObject *__pyx_v_return_rays = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("raytrace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_rcv,&__pyx_n_s_slowness,&__pyx_n_s_xi,&__pyx_n_s_theta,&__pyx_n_s_Vp0,&__pyx_n_s_Vs0,&__pyx_n_s_delta,&__pyx_n_s_epsilon,&__pyx_n_s_gamma,&__pyx_n_s_thread_no,&__pyx_n_s_aggregate_src,&__pyx_n_s_compute_L,&__pyx_n_s_return_rays,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);

    /* "ttcrpy/rgrid.pyx":2463
 * 
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,             # <<<<<<<<<<<<<<
 *                  thread_no=None, aggregate_src=False, compute_L=False,
 *                  return_rays=False):
 */
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_None);
    values[8] = ((PyObject *)Py_None);
    values[9] = ((PyObject *)Py_None);

    /* "ttcrpy/rgrid.pyx":2464
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,
 *                  thread_no=None, aggregate_src=False, compute_L=False,             # <<<<<<<<<<<<<<
 *                  return_rays=False):
 *         """
 */
    values[10] = ((PyObject *)Py_None);
    values[11] = ((PyObject *)Py_False);
    values[12] = ((PyObject *)Py_False);

    /* "ttcrpy/rgrid.pyx":2465
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,
 *                  thread_no=None, aggregate_src=False, compute_L=False,
 *                  return_rays=False):             # <<<<<<<<<<<<<<
 *         """
 *         raytrace(source, rcv, slowness=None, xi=None, theta=None, Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None, thread_no=None, aggregate_src=False, compute_L=False, return_rays=False) -> tt, rays, L
 */
    values[13] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rcv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 14, 1); __PYX_ERR(0, 2462, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_xi);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Vp0);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Vs0);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_epsilon);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gamma);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aggregate_src);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_L);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_rays);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "raytrace") < 0)) __PYX_ERR(0, 2462, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_rcv = values[1];
    __pyx_v_slowness = values[2];
    __pyx_v_xi = values[3];
    __pyx_v_theta = values[4];
    __pyx_v_Vp0 = values[5];
    __pyx_v_Vs0 = values[6];
    __pyx_v_delta = values[7];
    __pyx_v_epsilon = values[8];
    __pyx_v_gamma = values[9];
    __pyx_v_thread_no = values[10];
    __pyx_v_aggregate_src = values[11];
    __pyx_v_compute_L = values[12];
    __pyx_v_return_rays = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_36raytrace(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_source, __pyx_v_rcv, __pyx_v_slowness, __pyx_v_xi, __pyx_v_theta, __pyx_v_Vp0, __pyx_v_Vs0, __pyx_v_delta, __pyx_v_epsilon, __pyx_v_gamma, __pyx_v_thread_no, __pyx_v_aggregate_src, __pyx_v_compute_L, __pyx_v_return_rays);

  /* "ttcrpy/rgrid.pyx":2462
 *         return Kx, Kz
 * 
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,             # <<<<<<<<<<<<<<
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,
 *                  thread_no=None, aggregate_src=False, compute_L=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_36raytrace(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_xi, PyObject *__pyx_v_theta, PyObject *__pyx_v_Vp0, PyObject *__pyx_v_Vs0, PyObject *__pyx_v_delta, PyObject *__pyx_v_epsilon, PyObject *__pyx_v_gamma, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_return_rays) {
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_Tx = NULL;
  PyObject *__pyx_v_t0 = NULL;
  PyObject *__pyx_v_nTx = NULL;
  PyObject *__pyx_v_tmp = NULL;
  std::vector<std::vector<ttcr::sxz<double> > >  __pyx_v_vTx;
  std::vector<std::vector<ttcr::sxz<double> > >  __pyx_v_vRx;
  std::vector<std::vector<double> >  __pyx_v_vt0;
  std::vector<std::vector<double> >  __pyx_v_vtt;
  std::vector<std::vector<std::vector<ttcr::sxz<double> > > >  __pyx_v_r_data;
  std::vector<std::vector<std::vector<ttcr::siv2<double> > > >  __pyx_v_l_data;
  size_t __pyx_v_thread_nb;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_n;
  int __pyx_v_nn;
  int __pyx_v_MM;
  int __pyx_v_NN;
  PyObject *__pyx_v_iRx = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_ind = NULL;
  PyObject *__pyx_v_tt = NULL;
  std::vector<double> ::size_type __pyx_v_nt;
  PyObject *__pyx_v_rays = NULL;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_v_n2;
  PyObject *__pyx_v_L = NULL;
  PyObject *__pyx_v_ncells = NULL;
  PyObject *__pyx_v_nnz = NULL;
  std::vector<std::vector<ttcr::siv2<double> > > ::size_type __pyx_v_ni;
  PyObject *__pyx_v_indptr = NULL;
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_itmp = NULL;
  CYTHON_UNUSED int __pyx_8genexpr2__pyx_v_n;
  CYTHON_UNUSED int __pyx_8genexpr3__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_7;
  std::vector<std::vector<double> > ::size_type __pyx_t_8;
  std::vector<std::vector<std::vector<ttcr::siv2<double> > > > ::size_type __pyx_t_9;
  std::vector<std::vector<std::vector<ttcr::sxz<double> > > > ::size_type __pyx_t_10;
  double __pyx_t_11;
  double __pyx_t_12;
  ttcr::sxz<double>  __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  std::vector<double> ::size_type __pyx_t_16;
  int __pyx_t_17;
  long __pyx_t_18;
  long __pyx_t_19;
  int __pyx_t_20;
  size_t __pyx_t_21;
  std::vector<double> ::size_type __pyx_t_22;
  std::vector<double> ::size_type __pyx_t_23;
  std::vector<double> ::size_type __pyx_t_24;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_25;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_26;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_27;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_28;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_29;
  int __pyx_t_30;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_31;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_32;
  std::vector<std::vector<ttcr::siv2<double> > > ::size_type __pyx_t_33;
  std::vector<std::vector<ttcr::siv2<double> > > ::size_type __pyx_t_34;
  std::vector<std::vector<ttcr::siv2<double> > > ::size_type __pyx_t_35;
  PyObject *__pyx_t_36 = NULL;
  int __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  int __pyx_t_41;
  std::vector<ttcr::siv2<double> > ::size_type __pyx_t_42;
  std::vector<ttcr::siv2<double> > ::size_type __pyx_t_43;
  int __pyx_t_44;
  PyObject *__pyx_t_45 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raytrace", 0);

  /* "ttcrpy/rgrid.pyx":2544
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2545
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if self.iso != b'i' and self.method != b's' and not self.cell_slowness:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2545, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2544
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2547
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.iso != b'i' and self.method != b's' and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')
 * 
 */
  __pyx_t_4 = ((__pyx_v_self->iso != 'i') != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_self->method != 's') != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_v_self->cell_slowness != 0)) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2548
 * 
 *         if self.iso != b'i' and self.method != b's' and not self.cell_slowness:
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and not self.cell_slowness:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2548, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2547
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.iso != b'i' and self.method != b's' and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2550
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2550, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_v_self->cell_slowness != 0)) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L11_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2551
 * 
 *         if compute_L and not self.cell_slowness:
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and (self.iso == b'p' or self.iso == b'h'):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2551, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2550
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2553
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 *         if compute_L and (self.iso == b'p' or self.iso == b'h'):             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L not implemented for VTI media')
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2553, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L14_bool_binop_done;
  }
  switch (__pyx_v_self->iso) {
    case 'p':
    case 'h':
    __pyx_t_4 = 1;
    break;
    default:
    __pyx_t_4 = 0;
    break;
  }
  __pyx_t_1 = __pyx_t_4;
  __pyx_L14_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2554
 * 
 *         if compute_L and (self.iso == b'p' or self.iso == b'h'):
 *             raise NotImplementedError('compute_L not implemented for VTI media')             # <<<<<<<<<<<<<<
 * 
 *         if source.shape[1] == 2:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2554, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2553
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')
 * 
 *         if compute_L and (self.iso == b'p' or self.iso == b'h'):             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L not implemented for VTI media')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2556
 *             raise NotImplementedError('compute_L not implemented for VTI media')
 * 
 *         if source.shape[1] == 2:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2557
 * 
 *         if source.shape[1] == 2:
 *             src = source             # <<<<<<<<<<<<<<
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 */
    __Pyx_INCREF(__pyx_v_source);
    __pyx_v_src = __pyx_v_source;

    /* "ttcrpy/rgrid.pyx":2558
 *         if source.shape[1] == 2:
 *             src = source
 *             Tx = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 2558, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_Tx = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "ttcrpy/rgrid.pyx":2559
 *             src = source
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))             # <<<<<<<<<<<<<<
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
    __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_t0 = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "ttcrpy/rgrid.pyx":2560
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2556
 *             raise NotImplementedError('compute_L not implemented for VTI media')
 * 
 *         if source.shape[1] == 2:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
    goto __pyx_L16;
  }

  /* "ttcrpy/rgrid.pyx":2561
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2562
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__56); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2563
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 2563, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_tmp = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2564
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]             # <<<<<<<<<<<<<<
 *             Tx = tmp[:,1:3]
 *             t0 = tmp[:,0]
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmp, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_nTx = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2565
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]             # <<<<<<<<<<<<<<
 *             t0 = tmp[:,0]
 *         else:
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__56); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_Tx = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2566
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]
 *             t0 = tmp[:,0]             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_t0 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2561
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 */
    goto __pyx_L16;
  }

  /* "ttcrpy/rgrid.pyx":2568
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 2568, __pyx_L1_error)
  }
  __pyx_L16:;

  /* "ttcrpy/rgrid.pyx":2570
 *             raise ValueError('source should be either nsrc x 2 or 3')
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L18_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L18_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2571
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 *             raise ValueError('src and rcv should be ndata x 2')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 2571, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2570
 *             raise ValueError('source should be either nsrc x 2 or 3')
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2573
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_src);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2574
 * 
 *         if self.is_outside(src):
 *             raise ValueError('Source point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(rcv):
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 2574, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2573
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2576
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_rcv) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_rcv);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/rgrid.pyx":2577
 * 
 *         if self.is_outside(rcv):
 *             raise ValueError('Receiver outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 2577, __pyx_L1_error)

    /* "ttcrpy/rgrid.pyx":2576
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2579
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 *         if xi is not None:
 */
  __pyx_t_1 = (__pyx_v_slowness != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2580
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 *         if xi is not None:
 *             self.set_xi(xi)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2579
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 *         if xi is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2581
 *         if slowness is not None:
 *             self.set_slowness(slowness)
 *         if xi is not None:             # <<<<<<<<<<<<<<
 *             self.set_xi(xi)
 *         if theta is not None:
 */
  __pyx_t_4 = (__pyx_v_xi != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2582
 *             self.set_slowness(slowness)
 *         if xi is not None:
 *             self.set_xi(xi)             # <<<<<<<<<<<<<<
 *         if theta is not None:
 *             self.set_tilt_angle(theta)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_xi); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_xi) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_xi);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2581
 *         if slowness is not None:
 *             self.set_slowness(slowness)
 *         if xi is not None:             # <<<<<<<<<<<<<<
 *             self.set_xi(xi)
 *         if theta is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2583
 *         if xi is not None:
 *             self.set_xi(xi)
 *         if theta is not None:             # <<<<<<<<<<<<<<
 *             self.set_tilt_angle(theta)
 *         if Vp0 is not None:
 */
  __pyx_t_1 = (__pyx_v_theta != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2584
 *             self.set_xi(xi)
 *         if theta is not None:
 *             self.set_tilt_angle(theta)             # <<<<<<<<<<<<<<
 *         if Vp0 is not None:
 *             self.set_Vp0(Vp0)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_tilt_angle); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_theta) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_theta);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2583
 *         if xi is not None:
 *             self.set_xi(xi)
 *         if theta is not None:             # <<<<<<<<<<<<<<
 *             self.set_tilt_angle(theta)
 *         if Vp0 is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2585
 *         if theta is not None:
 *             self.set_tilt_angle(theta)
 *         if Vp0 is not None:             # <<<<<<<<<<<<<<
 *             self.set_Vp0(Vp0)
 *         if Vs0 is not None:
 */
  __pyx_t_4 = (__pyx_v_Vp0 != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2586
 *             self.set_tilt_angle(theta)
 *         if Vp0 is not None:
 *             self.set_Vp0(Vp0)             # <<<<<<<<<<<<<<
 *         if Vs0 is not None:
 *             self.set_Vs0(Vs0)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_Vp0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_Vp0) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_Vp0);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2585
 *         if theta is not None:
 *             self.set_tilt_angle(theta)
 *         if Vp0 is not None:             # <<<<<<<<<<<<<<
 *             self.set_Vp0(Vp0)
 *         if Vs0 is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2587
 *         if Vp0 is not None:
 *             self.set_Vp0(Vp0)
 *         if Vs0 is not None:             # <<<<<<<<<<<<<<
 *             self.set_Vs0(Vs0)
 *         if delta is not None:
 */
  __pyx_t_1 = (__pyx_v_Vs0 != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2588
 *             self.set_Vp0(Vp0)
 *         if Vs0 is not None:
 *             self.set_Vs0(Vs0)             # <<<<<<<<<<<<<<
 *         if delta is not None:
 *             self.set_delta(delta)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_Vs0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_Vs0) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_Vs0);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2587
 *         if Vp0 is not None:
 *             self.set_Vp0(Vp0)
 *         if Vs0 is not None:             # <<<<<<<<<<<<<<
 *             self.set_Vs0(Vs0)
 *         if delta is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2589
 *         if Vs0 is not None:
 *             self.set_Vs0(Vs0)
 *         if delta is not None:             # <<<<<<<<<<<<<<
 *             self.set_delta(delta)
 *         if epsilon is not None:
 */
  __pyx_t_4 = (__pyx_v_delta != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2590
 *             self.set_Vs0(Vs0)
 *         if delta is not None:
 *             self.set_delta(delta)             # <<<<<<<<<<<<<<
 *         if epsilon is not None:
 *             self.set_epsilon(epsilon)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_delta); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_delta) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_delta);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2589
 *         if Vs0 is not None:
 *             self.set_Vs0(Vs0)
 *         if delta is not None:             # <<<<<<<<<<<<<<
 *             self.set_delta(delta)
 *         if epsilon is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2591
 *         if delta is not None:
 *             self.set_delta(delta)
 *         if epsilon is not None:             # <<<<<<<<<<<<<<
 *             self.set_epsilon(epsilon)
 *         if gamma is not None:
 */
  __pyx_t_1 = (__pyx_v_epsilon != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2592
 *             self.set_delta(delta)
 *         if epsilon is not None:
 *             self.set_epsilon(epsilon)             # <<<<<<<<<<<<<<
 *         if gamma is not None:
 *             self.set_gamma(gamma)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_epsilon); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_epsilon) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_epsilon);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2591
 *         if delta is not None:
 *             self.set_delta(delta)
 *         if epsilon is not None:             # <<<<<<<<<<<<<<
 *             self.set_epsilon(epsilon)
 *         if gamma is not None:
 */
  }

  /* "ttcrpy/rgrid.pyx":2593
 *         if epsilon is not None:
 *             self.set_epsilon(epsilon)
 *         if gamma is not None:             # <<<<<<<<<<<<<<
 *             self.set_gamma(gamma)
 * 
 */
  __pyx_t_4 = (__pyx_v_gamma != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2594
 *             self.set_epsilon(epsilon)
 *         if gamma is not None:
 *             self.set_gamma(gamma)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxz[double]]] vTx
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_gamma); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_gamma) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_gamma);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2593
 *         if epsilon is not None:
 *             self.set_epsilon(epsilon)
 *         if gamma is not None:             # <<<<<<<<<<<<<<
 *             self.set_gamma(gamma)
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2607
 *         cdef int i, j, k, n, nn, MM, NN
 * 
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2607, __pyx_L1_error)
  try {
    __pyx_v_vTx.resize(__pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2607, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2608
 * 
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2608, __pyx_L1_error)
  try {
    __pyx_v_vRx.resize(__pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2608, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2609
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)             # <<<<<<<<<<<<<<
 *         vtt.resize(nTx)
 *         if compute_L:
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2609, __pyx_L1_error)
  try {
    __pyx_v_vt0.resize(__pyx_t_8);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2609, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2610
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)             # <<<<<<<<<<<<<<
 *         if compute_L:
 *             l_data.resize(nTx)
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2610, __pyx_L1_error)
  try {
    __pyx_v_vtt.resize(__pyx_t_8);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2610, __pyx_L1_error)
  }

  /* "ttcrpy/rgrid.pyx":2611
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             l_data.resize(nTx)
 *         if return_rays:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2611, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2612
 *         vtt.resize(nTx)
 *         if compute_L:
 *             l_data.resize(nTx)             # <<<<<<<<<<<<<<
 *         if return_rays:
 *             r_data.resize(nTx)
 */
    __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2612, __pyx_L1_error)
    try {
      __pyx_v_l_data.resize(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2612, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2611
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             l_data.resize(nTx)
 *         if return_rays:
 */
  }

  /* "ttcrpy/rgrid.pyx":2613
 *         if compute_L:
 *             l_data.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2613, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2614
 *             l_data.resize(nTx)
 *         if return_rays:
 *             r_data.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iRx = []
 */
    __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2614, __pyx_L1_error)
    try {
      __pyx_v_r_data.resize(__pyx_t_10);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2614, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2613
 *         if compute_L:
 *             l_data.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  }

  /* "ttcrpy/rgrid.pyx":2616
 *             r_data.resize(nTx)
 * 
 *         iRx = []             # <<<<<<<<<<<<<<
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_iRx = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":2617
 * 
 *         iRx = []
 *         if nTx == 1:             # <<<<<<<<<<<<<<
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 */
  __pyx_t_5 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2618
 *         iRx = []
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))             # <<<<<<<<<<<<<<
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__30); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__31); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    try {
      __pyx_t_13 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_12);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2618, __pyx_L1_error)
    }
    try {
      (__pyx_v_vTx[0]).push_back(__pyx_t_13);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2618, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2619
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 */
    if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
      __pyx_t_5 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_5); __pyx_t_14 = 0;
      __pyx_t_15 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_15 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2619, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_15)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_3); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2619, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_3); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2619, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_15(__pyx_t_5);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2619, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/rgrid.pyx":2620
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2620, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2620, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_t_13 = ttcr::sxz<double> (__pyx_t_12, __pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2620, __pyx_L1_error)
      }
      try {
        (__pyx_v_vRx[0]).push_back(__pyx_t_13);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2620, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2619
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2621
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])             # <<<<<<<<<<<<<<
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_t0, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2621, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    try {
      (__pyx_v_vt0[0]).push_back(__pyx_t_11);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2621, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2622
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_16 = __Pyx_PyInt_As_size_t(__pyx_t_3); if (unlikely((__pyx_t_16 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2622, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    try {
      (__pyx_v_vtt[0]).resize(__pyx_t_16);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2622, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2623
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *         elif aggregate_src:
 *             for t in Tx:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_3); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2617
 * 
 *         iRx = []
 *         if nTx == 1:             # <<<<<<<<<<<<<<
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 */
    goto __pyx_L32;
  }

  /* "ttcrpy/rgrid.pyx":2624
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:             # <<<<<<<<<<<<<<
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2624, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2625
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 *             for t in Tx:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 */
    if (likely(PyList_CheckExact(__pyx_v_Tx)) || PyTuple_CheckExact(__pyx_v_Tx)) {
      __pyx_t_3 = __pyx_v_Tx; __Pyx_INCREF(__pyx_t_3); __pyx_t_14 = 0;
      __pyx_t_15 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_Tx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2625, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_15)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2625, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2625, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2625, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2625, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_15(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2625, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2626
 *         elif aggregate_src:
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))             # <<<<<<<<<<<<<<
 *             for t in t0:
 *                 vt0[0].push_back(t)
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2626, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2626, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_t_13 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2626, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[0]).push_back(__pyx_t_13);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2626, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2625
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 *             for t in Tx:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2627
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 */
    if (likely(PyList_CheckExact(__pyx_v_t0)) || PyTuple_CheckExact(__pyx_v_t0)) {
      __pyx_t_3 = __pyx_v_t0; __Pyx_INCREF(__pyx_t_3); __pyx_t_14 = 0;
      __pyx_t_15 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_t0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2627, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_15)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2627, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2627, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2627, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2627, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_15(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2627, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2628
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 *                 vt0[0].push_back(t)             # <<<<<<<<<<<<<<
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 */
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_v_t); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2628, __pyx_L1_error)
      try {
        (__pyx_v_vt0[0]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2628, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2627
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2629
 *             for t in t0:
 *                 vt0[0].push_back(t)
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 */
    if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
      __pyx_t_3 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_3); __pyx_t_14 = 0;
      __pyx_t_15 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2629, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_15)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2629, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2629, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_15(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2629, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2630
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2630, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2630, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_t_13 = ttcr::sxz<double> (__pyx_t_12, __pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2630, __pyx_L1_error)
      }
      try {
        (__pyx_v_vRx[0]).push_back(__pyx_t_13);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2630, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2629
 *             for t in t0:
 *                 vt0[0].push_back(t)
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2631
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *             nTx = 1
 *             iRx.append(np.arange(rcv.shape[0]))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_16 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_16 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2631, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      (__pyx_v_vtt[0]).resize(__pyx_t_16);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2631, __pyx_L1_error)
    }

    /* "ttcrpy/rgrid.pyx":2632
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1             # <<<<<<<<<<<<<<
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 */
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_DECREF_SET(__pyx_v_nTx, __pyx_int_1);

    /* "ttcrpy/rgrid.pyx":2633
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1
 *             iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *         else:
 *             if src.shape != rcv.shape:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_2); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2633, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2624
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:             # <<<<<<<<<<<<<<
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 */
    goto __pyx_L32;
  }

  /* "ttcrpy/rgrid.pyx":2635
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2635, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2635, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/rgrid.pyx":2636
 *         else:
 *             if src.shape != rcv.shape:
 *                 raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nTx):
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 2636, __pyx_L1_error)

      /* "ttcrpy/rgrid.pyx":2635
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
    }

    /* "ttcrpy/rgrid.pyx":2638
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])
 */
    __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2638, __pyx_L1_error)
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_n = __pyx_t_20;

      /* "ttcrpy/rgrid.pyx":2639
 * 
 *             for n in range(nTx):
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2             # <<<<<<<<<<<<<<
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_slice__3);
      __Pyx_GIVEREF(__pyx_slice__3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice__3);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_v_src, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_1) < 0) __PYX_ERR(0, 2639, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ind, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2640
 *             for n in range(nTx):
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])             # <<<<<<<<<<<<<<
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2640, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2640, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_ind) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_ind);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2640, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2640, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2640, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":2641
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))             # <<<<<<<<<<<<<<
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        __pyx_t_13 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2641, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_13);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2641, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2642
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])             # <<<<<<<<<<<<<<
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_t0, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2642, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2642, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2643
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())
 */
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_ind);
      __Pyx_GIVEREF(__pyx_v_ind);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_ind);
      __Pyx_INCREF(__pyx_slice__3);
      __Pyx_GIVEREF(__pyx_slice__3);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_slice__3);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_15 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2643, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2643, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2643, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_15(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 2643, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":2644
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *                 vtt[n].resize(vRx[n].size())
 * 
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2644, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2644, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2644, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2644, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_13 = ttcr::sxz<double> (__pyx_t_12, __pyx_t_11);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2644, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_13);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2644, __pyx_L1_error)
        }

        /* "ttcrpy/rgrid.pyx":2643
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/rgrid.pyx":2645
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 * 
 *         tt = np.zeros((rcv.shape[0],))
 */
      try {
        (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2645, __pyx_L1_error)
      }
    }
  }
  __pyx_L32:;

  /* "ttcrpy/rgrid.pyx":2647
 *                 vtt[n].resize(vRx[n].size())
 * 
 *         tt = np.zeros((rcv.shape[0],))             # <<<<<<<<<<<<<<
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tt = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/rgrid.pyx":2648
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 */
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_nTx, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L47_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_self->_n_threads == 1) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L47_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/rgrid.pyx":2649
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2649, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2649, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L50_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2649, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2649, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_L50_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":2650
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:
 */
      __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2650, __pyx_L1_error)
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_n = __pyx_t_20;

        /* "ttcrpy/rgrid.pyx":2651
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2651, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":2649
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
      goto __pyx_L49;
    }

    /* "ttcrpy/rgrid.pyx":2652
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2652, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L54_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2652, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L54_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":2653
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:
 */
      __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2653, __pyx_L1_error)
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_n = __pyx_t_20;

        /* "ttcrpy/rgrid.pyx":2654
 *             elif compute_L and return_rays:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2654, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":2652
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 */
      goto __pyx_L49;
    }

    /* "ttcrpy/rgrid.pyx":2655
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2655, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/rgrid.pyx":2656
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             else:
 */
      __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2656, __pyx_L1_error)
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_n = __pyx_t_20;

        /* "ttcrpy/rgrid.pyx":2657
 *             elif compute_L:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2657, __pyx_L1_error)
        }
      }

      /* "ttcrpy/rgrid.pyx":2655
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
      goto __pyx_L49;
    }

    /* "ttcrpy/rgrid.pyx":2659
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *             else:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 */
    /*else*/ {
      __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2659, __pyx_L1_error)
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_n = __pyx_t_20;

        /* "ttcrpy/rgrid.pyx":2660
 *             else:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)             # <<<<<<<<<<<<<<
 * 
 *         elif thread_no is not None:
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2660, __pyx_L1_error)
        }
      }
    }
    __pyx_L49:;

    /* "ttcrpy/rgrid.pyx":2648
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 */
    goto __pyx_L46;
  }

  /* "ttcrpy/rgrid.pyx":2662
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  __pyx_t_1 = (__pyx_v_thread_no != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2664
 *         elif thread_no is not None:
 *             # we should be here for just one event
 *             assert nTx == 1             # <<<<<<<<<<<<<<
 *             # normally we should not need to compute L
 *             assert compute_L is False
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 2664, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/rgrid.pyx":2666
 *             assert nTx == 1
 *             # normally we should not need to compute L
 *             assert compute_L is False             # <<<<<<<<<<<<<<
 *             thread_nb = thread_no
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_4 = (__pyx_v_compute_L == Py_False);
      if (unlikely(!(__pyx_t_4 != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 2666, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/rgrid.pyx":2667
 *             # normally we should not need to compute L
 *             assert compute_L is False
 *             thread_nb = thread_no             # <<<<<<<<<<<<<<
 * 
 *             if return_rays:
 */
    __pyx_t_21 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_21 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2667, __pyx_L1_error)
    __pyx_v_thread_nb = __pyx_t_21;

    /* "ttcrpy/rgrid.pyx":2669
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2669, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":2670
 * 
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), (__pyx_v_r_data[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2670, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2671
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 */
      __pyx_t_22 = (__pyx_v_vtt[0]).size();
      __pyx_t_23 = __pyx_t_22;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
        __pyx_v_nt = __pyx_t_24;

        /* "ttcrpy/rgrid.pyx":2672
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2672, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 2672, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/rgrid.pyx":2673
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_rays = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/rgrid.pyx":2674
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 *                 for n2 in range(vRx.size()):             # <<<<<<<<<<<<<<
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):
 */
      __pyx_t_25 = __pyx_v_vRx.size();
      __pyx_t_26 = __pyx_t_25;
      for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
        __pyx_v_n2 = __pyx_t_27;

        /* "ttcrpy/rgrid.pyx":2675
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[0])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_2);
        __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":2676
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z
 */
        __pyx_t_28 = ((__pyx_v_r_data[0])[__pyx_v_n2]).size();
        __pyx_t_29 = __pyx_t_28;
        for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_29; __pyx_t_20+=1) {
          __pyx_v_nn = __pyx_t_20;

          /* "ttcrpy/rgrid.pyx":2677
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
          __pyx_t_2 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_6, __pyx_t_3) < 0)) __PYX_ERR(0, 2677, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ttcrpy/rgrid.pyx":2678
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z             # <<<<<<<<<<<<<<
 *                     rays.append(r)
 *                 return tt, rays
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
          __pyx_t_6 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 2678, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "ttcrpy/rgrid.pyx":2679
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)             # <<<<<<<<<<<<<<
 *                 return tt, rays
 * 
 */
        __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_rays, __pyx_v_r); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2679, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2680
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)
 *                 return tt, rays             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_tt);
      __Pyx_GIVEREF(__pyx_v_tt);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
      __Pyx_INCREF(__pyx_v_rays);
      __Pyx_GIVEREF(__pyx_v_rays);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ttcrpy/rgrid.pyx":2669
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    }

    /* "ttcrpy/rgrid.pyx":2683
 * 
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2683, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2684
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 return tt
 */
      __pyx_t_22 = (__pyx_v_vtt[0]).size();
      __pyx_t_23 = __pyx_t_22;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
        __pyx_v_nt = __pyx_t_24;

        /* "ttcrpy/rgrid.pyx":2685
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 return tt
 * 
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 2685, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/rgrid.pyx":2686
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 return tt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_tt);
      __pyx_r = __pyx_v_tt;
      goto __pyx_L0;
    }

    /* "ttcrpy/rgrid.pyx":2662
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  }

  /* "ttcrpy/rgrid.pyx":2689
 * 
 *         else:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 */
  /*else*/ {
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L72_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_L72_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":2690
 *         else:
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2690, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2689
 * 
 *         else:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 */
      goto __pyx_L71;
    }

    /* "ttcrpy/rgrid.pyx":2691
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2691, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L74_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2691, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L74_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":2692
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data, __pyx_v_l_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2692, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2691
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 */
      goto __pyx_L71;
    }

    /* "ttcrpy/rgrid.pyx":2693
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2693, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/rgrid.pyx":2694
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_l_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2694, __pyx_L1_error)
      }

      /* "ttcrpy/rgrid.pyx":2693
 *             elif compute_L and return_rays:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             else:
 */
      goto __pyx_L71;
    }

    /* "ttcrpy/rgrid.pyx":2696
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 2696, __pyx_L1_error)
      }
    }
    __pyx_L71:;
  }
  __pyx_L46:;

  /* "ttcrpy/rgrid.pyx":2698
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 */
  __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2698, __pyx_L1_error)
  __pyx_t_19 = __pyx_t_18;
  for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
    __pyx_v_n = __pyx_t_20;

    /* "ttcrpy/rgrid.pyx":2699
 * 
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 */
    __pyx_t_22 = (__pyx_v_vtt[__pyx_v_n]).size();
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
      __pyx_v_nt = __pyx_t_24;

      /* "ttcrpy/rgrid.pyx":2700
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]             # <<<<<<<<<<<<<<
 * 
 *         if return_rays:
 */
      __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[__pyx_v_n])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tt, __pyx_t_6, __pyx_t_3) < 0)) __PYX_ERR(0, 2700, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "ttcrpy/rgrid.pyx":2702
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2702, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2703
 * 
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2703, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_8genexpr2__pyx_v_n = __pyx_t_20;
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2703, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2703, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    } /* exit inner scope */
    __pyx_v_rays = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2704
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 */
    __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2704, __pyx_L1_error)
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_n = __pyx_t_20;

      /* "ttcrpy/rgrid.pyx":2705
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_28 = (__pyx_v_vRx[__pyx_v_n]).size();
        __pyx_t_29 = __pyx_t_28;
        for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
          __pyx_8genexpr3__pyx_v_i = __pyx_t_30;
          __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2705, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_float_0_0);
          __Pyx_GIVEREF(__pyx_float_0_0);
          PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2705, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/rgrid.pyx":2706
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):             # <<<<<<<<<<<<<<
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):
 */
      __pyx_t_28 = (__pyx_v_vRx[__pyx_v_n]).size();
      __pyx_t_29 = __pyx_t_28;
      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_29; __pyx_t_25+=1) {
        __pyx_v_n2 = __pyx_t_25;

        /* "ttcrpy/rgrid.pyx":2707
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_v_r, __pyx_v_n2, __pyx_t_3, std::vector<std::vector<ttcr::sxz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 2707, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":2708
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 */
        __pyx_t_31 = ((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size();
        __pyx_t_32 = __pyx_t_31;
        for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_32; __pyx_t_30+=1) {
          __pyx_v_nn = __pyx_t_30;

          /* "ttcrpy/rgrid.pyx":2709
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, std::vector<std::vector<ttcr::sxz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
          __pyx_t_5 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 2709, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ttcrpy/rgrid.pyx":2710
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2710, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, std::vector<std::vector<ttcr::sxz<double> > > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2710, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2710, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2710, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
          __pyx_t_6 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_5, __pyx_t_3) < 0)) __PYX_ERR(0, 2710, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
      }

      /* "ttcrpy/rgrid.pyx":2711
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 */
      __pyx_t_22 = (__pyx_v_vtt[__pyx_v_n]).size();
      __pyx_t_23 = __pyx_t_22;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
        __pyx_v_nt = __pyx_t_24;

        /* "ttcrpy/rgrid.pyx":2712
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]             # <<<<<<<<<<<<<<
 * 
 *         if compute_L:
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_rays, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 2712, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }

    /* "ttcrpy/rgrid.pyx":2702
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  }

  /* "ttcrpy/rgrid.pyx":2714
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             # first build an array of matrices, for each event
 *             L = []
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2714, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2716
 *         if compute_L:
 *             # first build an array of matrices, for each event
 *             L = []             # <<<<<<<<<<<<<<
 *             ncells = self.get_number_of_cells()
 *             for n in range(nTx):
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_L = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2717
 *             # first build an array of matrices, for each event
 *             L = []
 *             ncells = self.get_number_of_cells()             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 nnz = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_ncells = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2718
 *             L = []
 *             ncells = self.get_number_of_cells()
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):
 */
    __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2718, __pyx_L1_error)
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_n = __pyx_t_20;

      /* "ttcrpy/rgrid.pyx":2719
 *             ncells = self.get_number_of_cells()
 *             for n in range(nTx):
 *                 nnz = 0             # <<<<<<<<<<<<<<
 *                 for ni in range(l_data[n].size()):
 *                     nnz += l_data[n][ni].size()
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_nnz, __pyx_int_0);

      /* "ttcrpy/rgrid.pyx":2720
 *             for n in range(nTx):
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):             # <<<<<<<<<<<<<<
 *                     nnz += l_data[n][ni].size()
 * 
 */
      __pyx_t_33 = (__pyx_v_l_data[__pyx_v_n]).size();
      __pyx_t_34 = __pyx_t_33;
      for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_34; __pyx_t_35+=1) {
        __pyx_v_ni = __pyx_t_35;

        /* "ttcrpy/rgrid.pyx":2721
 *                 nnz = 0
 *                 for ni in range(l_data[n].size()):
 *                     nnz += l_data[n][ni].size()             # <<<<<<<<<<<<<<
 * 
 *                 if self.iso == b'i':
 */
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(((__pyx_v_l_data[__pyx_v_n])[__pyx_v_ni]).size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_nnz, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_2);
        __pyx_t_2 = 0;
      }

      /* "ttcrpy/rgrid.pyx":2723
 *                     nnz += l_data[n][ni].size()
 * 
 *                 if self.iso == b'i':             # <<<<<<<<<<<<<<
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)
 */
      __pyx_t_4 = ((__pyx_v_self->iso == 'i') != 0);
      if (__pyx_t_4) {

        /* "ttcrpy/rgrid.pyx":2724
 * 
 *                 if self.iso == b'i':
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                     indices = np.empty((nnz,), dtype=np.int64)
 *                     val = np.empty((nnz,))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_vRx[__pyx_v_n]).size() + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_36 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int64); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_36) < 0) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __pyx_t_36 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2724, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_indptr, __pyx_t_36);
        __pyx_t_36 = 0;

        /* "ttcrpy/rgrid.pyx":2725
 *                 if self.iso == b'i':
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                     val = np.empty((nnz,))
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_36, __pyx_n_s_np); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_INCREF(__pyx_v_nnz);
        __Pyx_GIVEREF(__pyx_v_nnz);
        PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_v_nnz);
        __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_36);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_36);
        __pyx_t_36 = 0;
        __pyx_t_36 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_36); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2725, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __Pyx_XDECREF_SET(__pyx_v_indices, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":2726
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)
 *                     val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *                     k = 0
 */
        __Pyx_GetModuleGlobalName(__pyx_t_36, __pyx_n_s_np); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_INCREF(__pyx_v_nnz);
        __Pyx_GIVEREF(__pyx_v_nnz);
        PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_v_nnz);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_36) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_36);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":2728
 *                     val = np.empty((nnz,))
 * 
 *                     k = 0             # <<<<<<<<<<<<<<
 *                     MM = vRx[n].size()
 *                     NN = ncells
 */
        __pyx_v_k = 0;

        /* "ttcrpy/rgrid.pyx":2729
 * 
 *                     k = 0
 *                     MM = vRx[n].size()             # <<<<<<<<<<<<<<
 *                     NN = ncells
 *                     for i in range(MM):
 */
        __pyx_v_MM = (__pyx_v_vRx[__pyx_v_n]).size();

        /* "ttcrpy/rgrid.pyx":2730
 *                     k = 0
 *                     MM = vRx[n].size()
 *                     NN = ncells             # <<<<<<<<<<<<<<
 *                     for i in range(MM):
 *                         indptr[i] = k
 */
        __pyx_t_30 = __Pyx_PyInt_As_int(__pyx_v_ncells); if (unlikely((__pyx_t_30 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2730, __pyx_L1_error)
        __pyx_v_NN = __pyx_t_30;

        /* "ttcrpy/rgrid.pyx":2731
 *                     MM = vRx[n].size()
 *                     NN = ncells
 *                     for i in range(MM):             # <<<<<<<<<<<<<<
 *                         indptr[i] = k
 *                         for j in range(NN):
 */
        __pyx_t_30 = __pyx_v_MM;
        __pyx_t_37 = __pyx_t_30;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
          __pyx_v_i = __pyx_t_38;

          /* "ttcrpy/rgrid.pyx":2732
 *                     NN = ncells
 *                     for i in range(MM):
 *                         indptr[i] = k             # <<<<<<<<<<<<<<
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 */
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2732, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_i, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2732, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "ttcrpy/rgrid.pyx":2733
 *                     for i in range(MM):
 *                         indptr[i] = k
 *                         for j in range(NN):             # <<<<<<<<<<<<<<
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:
 */
          __pyx_t_39 = __pyx_v_NN;
          __pyx_t_40 = __pyx_t_39;
          for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_40; __pyx_t_41+=1) {
            __pyx_v_j = __pyx_t_41;

            /* "ttcrpy/rgrid.pyx":2734
 *                         indptr[i] = k
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j
 */
            __pyx_t_42 = ((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i]).size();
            __pyx_t_43 = __pyx_t_42;
            for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_43; __pyx_t_44+=1) {
              __pyx_v_nn = __pyx_t_44;

              /* "ttcrpy/rgrid.pyx":2735
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 */
              __pyx_t_4 = (((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).i == __pyx_v_j) != 0);
              if (__pyx_t_4) {

                /* "ttcrpy/rgrid.pyx":2736
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j             # <<<<<<<<<<<<<<
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1
 */
                __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2736, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_indices, __pyx_v_k, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2736, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "ttcrpy/rgrid.pyx":2737
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v             # <<<<<<<<<<<<<<
 *                                     k += 1
 * 
 */
                __pyx_t_6 = PyFloat_FromDouble((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).v); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2737, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_val, __pyx_v_k, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2737, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "ttcrpy/rgrid.pyx":2738
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1             # <<<<<<<<<<<<<<
 * 
 *                     indptr[MM] = k
 */
                __pyx_v_k = (__pyx_v_k + 1);

                /* "ttcrpy/rgrid.pyx":2735
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 */
              }
            }
          }
        }

        /* "ttcrpy/rgrid.pyx":2740
 *                                     k += 1
 * 
 *                     indptr[MM] = k             # <<<<<<<<<<<<<<
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 *                 else:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2740, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_MM, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2740, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":2741
 * 
 *                     indptr[MM] = k
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 *                 else:
 *                     nnz *= 2
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_v_val);
        __Pyx_GIVEREF(__pyx_v_val);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_val);
        __Pyx_INCREF(__pyx_v_indices);
        __Pyx_GIVEREF(__pyx_v_indices);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_indices);
        __Pyx_INCREF(__pyx_v_indptr);
        __Pyx_GIVEREF(__pyx_v_indptr);
        PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_indptr);
        __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_MM); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_NN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_3);
        __pyx_t_5 = 0;
        __pyx_t_3 = 0;
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_shape, __pyx_t_45) < 0) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __pyx_t_45 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_36, __pyx_t_6); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_17 = __Pyx_PyObject_Append(__pyx_v_L, __pyx_t_45); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

        /* "ttcrpy/rgrid.pyx":2723
 *                     nnz += l_data[n][ni].size()
 * 
 *                 if self.iso == b'i':             # <<<<<<<<<<<<<<
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)
 */
        goto __pyx_L98;
      }

      /* "ttcrpy/rgrid.pyx":2743
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 *                 else:
 *                     nnz *= 2             # <<<<<<<<<<<<<<
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)
 */
      /*else*/ {
        __pyx_t_45 = PyNumber_InPlaceMultiply(__pyx_v_nnz, __pyx_int_2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_45);
        __pyx_t_45 = 0;

        /* "ttcrpy/rgrid.pyx":2744
 *                 else:
 *                     nnz *= 2
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                     indices = np.empty((nnz,), dtype=np.int64)
 *                     val = np.empty((nnz,))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_45, __pyx_n_s_np); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_45, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __pyx_t_45 = __Pyx_PyInt_FromSize_t(((__pyx_v_vRx[__pyx_v_n]).size() + 1)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_GIVEREF(__pyx_t_45);
        PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_t_45);
        __pyx_t_45 = 0;
        __pyx_t_45 = PyTuple_New(1); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_GIVEREF(__pyx_t_36);
        PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_36);
        __pyx_t_36 = 0;
        __pyx_t_36 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_45, __pyx_t_36); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __Pyx_XDECREF_SET(__pyx_v_indptr, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":2745
 *                     nnz *= 2
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                     val = np.empty((nnz,))
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_36 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_nnz);
        __Pyx_GIVEREF(__pyx_v_nnz);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nnz);
        __pyx_t_45 = PyTuple_New(1); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_36, __pyx_t_45, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_indices, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":2746
 *                     indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                     indices = np.empty((nnz,), dtype=np.int64)
 *                     val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *                     k = 0
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_45 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_nnz);
        __Pyx_GIVEREF(__pyx_v_nnz);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nnz);
        __pyx_t_36 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_45))) {
          __pyx_t_36 = PyMethod_GET_SELF(__pyx_t_45);
          if (likely(__pyx_t_36)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_45);
            __Pyx_INCREF(__pyx_t_36);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_45, function);
          }
        }
        __pyx_t_2 = (__pyx_t_36) ? __Pyx_PyObject_Call2Args(__pyx_t_45, __pyx_t_36, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_45, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_36); __pyx_t_36 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":2748
 *                     val = np.empty((nnz,))
 * 
 *                     k = 0             # <<<<<<<<<<<<<<
 *                     MM = vRx[n].size()
 *                     NN = 2*ncells
 */
        __pyx_v_k = 0;

        /* "ttcrpy/rgrid.pyx":2749
 * 
 *                     k = 0
 *                     MM = vRx[n].size()             # <<<<<<<<<<<<<<
 *                     NN = 2*ncells
 *                     for i in range(MM):
 */
        __pyx_v_MM = (__pyx_v_vRx[__pyx_v_n]).size();

        /* "ttcrpy/rgrid.pyx":2750
 *                     k = 0
 *                     MM = vRx[n].size()
 *                     NN = 2*ncells             # <<<<<<<<<<<<<<
 *                     for i in range(MM):
 *                         indptr[i] = k
 */
        __pyx_t_2 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ncells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_30 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_30 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2750, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_NN = __pyx_t_30;

        /* "ttcrpy/rgrid.pyx":2751
 *                     MM = vRx[n].size()
 *                     NN = 2*ncells
 *                     for i in range(MM):             # <<<<<<<<<<<<<<
 *                         indptr[i] = k
 *                         for j in range(NN):
 */
        __pyx_t_30 = __pyx_v_MM;
        __pyx_t_37 = __pyx_t_30;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
          __pyx_v_i = __pyx_t_38;

          /* "ttcrpy/rgrid.pyx":2752
 *                     NN = 2*ncells
 *                     for i in range(MM):
 *                         indptr[i] = k             # <<<<<<<<<<<<<<
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2752, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2752, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":2753
 *                     for i in range(MM):
 *                         indptr[i] = k
 *                         for j in range(NN):             # <<<<<<<<<<<<<<
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:
 */
          __pyx_t_39 = __pyx_v_NN;
          __pyx_t_40 = __pyx_t_39;
          for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_40; __pyx_t_41+=1) {
            __pyx_v_j = __pyx_t_41;

            /* "ttcrpy/rgrid.pyx":2754
 *                         indptr[i] = k
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j
 */
            __pyx_t_42 = ((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i]).size();
            __pyx_t_43 = __pyx_t_42;
            for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_43; __pyx_t_44+=1) {
              __pyx_v_nn = __pyx_t_44;

              /* "ttcrpy/rgrid.pyx":2755
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 */
              __pyx_t_4 = (((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).i == __pyx_v_j) != 0);
              if (__pyx_t_4) {

                /* "ttcrpy/rgrid.pyx":2756
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j             # <<<<<<<<<<<<<<
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1
 */
                __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2756, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_indices, __pyx_v_k, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2756, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

                /* "ttcrpy/rgrid.pyx":2757
 *                                 if l_data[n][i][nn].i == j:
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v             # <<<<<<<<<<<<<<
 *                                     k += 1
 *                                 elif l_data[n][i][nn].i+ncells == j:
 */
                __pyx_t_2 = PyFloat_FromDouble((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).v); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2757, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_val, __pyx_v_k, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2757, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

                /* "ttcrpy/rgrid.pyx":2758
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1             # <<<<<<<<<<<<<<
 *                                 elif l_data[n][i][nn].i+ncells == j:
 *                                     indices[k] = j
 */
                __pyx_v_k = (__pyx_v_k + 1);

                /* "ttcrpy/rgrid.pyx":2755
 *                         for j in range(NN):
 *                             for nn in range(l_data[n][i].size()):
 *                                 if l_data[n][i][nn].i == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v
 */
                goto __pyx_L112;
              }

              /* "ttcrpy/rgrid.pyx":2759
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1
 *                                 elif l_data[n][i][nn].i+ncells == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v2
 */
              __pyx_t_2 = __Pyx_PyInt_FromSize_t((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2759, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_45 = PyNumber_Add(__pyx_t_2, __pyx_v_ncells); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2759, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_45);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2759, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = PyObject_RichCompare(__pyx_t_45, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2759, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2759, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (__pyx_t_4) {

                /* "ttcrpy/rgrid.pyx":2760
 *                                     k += 1
 *                                 elif l_data[n][i][nn].i+ncells == j:
 *                                     indices[k] = j             # <<<<<<<<<<<<<<
 *                                     val[k] = l_data[n][i][nn].v2
 *                                     k += 1
 */
                __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2760, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_3);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_indices, __pyx_v_k, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2760, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "ttcrpy/rgrid.pyx":2761
 *                                 elif l_data[n][i][nn].i+ncells == j:
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v2             # <<<<<<<<<<<<<<
 *                                     k += 1
 * 
 */
                __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_l_data[__pyx_v_n])[__pyx_v_i])[__pyx_v_nn]).v2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2761, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_3);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_val, __pyx_v_k, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2761, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "ttcrpy/rgrid.pyx":2762
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v2
 *                                     k += 1             # <<<<<<<<<<<<<<
 * 
 *                     indptr[MM] = k
 */
                __pyx_v_k = (__pyx_v_k + 1);

                /* "ttcrpy/rgrid.pyx":2759
 *                                     val[k] = l_data[n][i][nn].v
 *                                     k += 1
 *                                 elif l_data[n][i][nn].i+ncells == j:             # <<<<<<<<<<<<<<
 *                                     indices[k] = j
 *                                     val[k] = l_data[n][i][nn].v2
 */
              }
              __pyx_L112:;
            }
          }
        }

        /* "ttcrpy/rgrid.pyx":2764
 *                                     k += 1
 * 
 *                     indptr[MM] = k             # <<<<<<<<<<<<<<
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 *             # we want a single matrix
 */
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2764, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_MM, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2764, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/rgrid.pyx":2765
 * 
 *                     indptr[MM] = k
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 *             # we want a single matrix
 *             tmp = sp.vstack(L)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_val);
        __Pyx_GIVEREF(__pyx_v_val);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_val);
        __Pyx_INCREF(__pyx_v_indices);
        __Pyx_GIVEREF(__pyx_v_indices);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_indices);
        __Pyx_INCREF(__pyx_v_indptr);
        __Pyx_GIVEREF(__pyx_v_indptr);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_indptr);
        __pyx_t_45 = PyTuple_New(1); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_36 = __Pyx_PyInt_From_int(__pyx_v_MM); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_36);
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_NN); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_36);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_36);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
        __pyx_t_36 = 0;
        __pyx_t_6 = 0;
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_45, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_17 = __Pyx_PyObject_Append(__pyx_v_L, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2765, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_L98:;
    }

    /* "ttcrpy/rgrid.pyx":2767
 *                     L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 *             # we want a single matrix
 *             tmp = sp.vstack(L)             # <<<<<<<<<<<<<<
 *             itmp = []
 *             for n in range(nTx):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_45 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vstack); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_45))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_45);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_45);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_45, function);
      }
    }
    __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_45, __pyx_t_3, __pyx_v_L) : __Pyx_PyObject_CallOneArg(__pyx_t_45, __pyx_v_L);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2768
 *             # we want a single matrix
 *             tmp = sp.vstack(L)
 *             itmp = []             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 for nt in range(vtt[n].size()):
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2768, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_itmp = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2769
 *             tmp = sp.vstack(L)
 *             itmp = []
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[n].size()):
 *                     itmp.append(iRx[n][nt])
 */
    __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 2769, __pyx_L1_error)
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_n = __pyx_t_20;

      /* "ttcrpy/rgrid.pyx":2770
 *             itmp = []
 *             for n in range(nTx):
 *                 for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                     itmp.append(iRx[n][nt])
 *             L = tmp[itmp,:]
 */
      __pyx_t_22 = (__pyx_v_vtt[__pyx_v_n]).size();
      __pyx_t_23 = __pyx_t_22;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
        __pyx_v_nt = __pyx_t_24;

        /* "ttcrpy/rgrid.pyx":2771
 *             for n in range(nTx):
 *                 for nt in range(vtt[n].size()):
 *                     itmp.append(iRx[n][nt])             # <<<<<<<<<<<<<<
 *             L = tmp[itmp,:]
 * 
 */
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_45 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_nt, std::vector<double> ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_45);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_itmp, __pyx_t_45); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 2771, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
      }
    }

    /* "ttcrpy/rgrid.pyx":2772
 *                 for nt in range(vtt[n].size()):
 *                     itmp.append(iRx[n][nt])
 *             L = tmp[itmp,:]             # <<<<<<<<<<<<<<
 * 
 *         if compute_L==False and return_rays==False:
 */
    __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 2772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_45);
    __Pyx_INCREF(__pyx_v_itmp);
    __Pyx_GIVEREF(__pyx_v_itmp);
    PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_v_itmp);
    __Pyx_INCREF(__pyx_slice__3);
    __Pyx_GIVEREF(__pyx_slice__3);
    PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_slice__3);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_t_45); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
    __Pyx_DECREF_SET(__pyx_v_L, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/rgrid.pyx":2714
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L:             # <<<<<<<<<<<<<<
 *             # first build an array of matrices, for each event
 *             L = []
 */
  }

  /* "ttcrpy/rgrid.pyx":2774
 *             L = tmp[itmp,:]
 * 
 *         if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_L and return_rays:
 */
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2774, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L118_bool_binop_done;
  }
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2774, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __pyx_t_1;
  __pyx_L118_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2775
 * 
 *         if compute_L==False and return_rays==False:
 *             return tt             # <<<<<<<<<<<<<<
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_tt);
    __pyx_r = __pyx_v_tt;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":2774
 *             L = tmp[itmp,:]
 * 
 *         if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_L and return_rays:
 */
  }

  /* "ttcrpy/rgrid.pyx":2776
 *         if compute_L==False and return_rays==False:
 *             return tt
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2776, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L120_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2776, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_1;
  __pyx_L120_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2777
 *             return tt
 *         elif compute_L and return_rays:
 *             return tt, rays, L             # <<<<<<<<<<<<<<
 *         elif compute_L:
 *             return tt, L
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 2777, __pyx_L1_error) }
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 2777, __pyx_L1_error) }
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_rays);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_L);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":2776
 *         if compute_L==False and return_rays==False:
 *             return tt
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  }

  /* "ttcrpy/rgrid.pyx":2778
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2778, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/rgrid.pyx":2779
 *             return tt, rays, L
 *         elif compute_L:
 *             return tt, L             # <<<<<<<<<<<<<<
 *         else:
 *             return tt, rays
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 2779, __pyx_L1_error) }
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_L);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "ttcrpy/rgrid.pyx":2778
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         else:
 */
  }

  /* "ttcrpy/rgrid.pyx":2781
 *             return tt, L
 *         else:
 *             return tt, rays             # <<<<<<<<<<<<<<
 * 
 *     def to_vtk(self, fields, filename):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 2781, __pyx_L1_error) }
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_rays);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/rgrid.pyx":2462
 *         return Kx, Kz
 * 
 *     def raytrace(self, source, rcv, slowness=None, xi=None, theta=None,             # <<<<<<<<<<<<<<
 *                  Vp0=None, Vs0=None, delta=None, epsilon=None, gamma=None,
 *                  thread_no=None, aggregate_src=False, compute_L=False,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_Tx);
  __Pyx_XDECREF(__pyx_v_t0);
  __Pyx_XDECREF(__pyx_v_nTx);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_iRx);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_rays);
  __Pyx_XDECREF(__pyx_v_L);
  __Pyx_XDECREF(__pyx_v_ncells);
  __Pyx_XDECREF(__pyx_v_nnz);
  __Pyx_XDECREF(__pyx_v_indptr);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_itmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2783
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_39to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_38to_vtk[] = "\n        to_vtk(fields, filename)\n\n        Save grid variables and/or raypaths to VTK format\n\n        Parameters\n        ----------\n        fields: dict\n            dict of variables to save to file. Variables should be np.ndarray of\n            size equal to either the number of nodes of the number of cells of\n            the grid, or a list of raypath coordinates.\n        filename: str\n            Name of file without extension for saving (extension vtr will be\n            added).  Raypaths are saved in separate files, and filename will\n            be appended by the dict key and have a vtp extension.\n\n        Notes\n        -----\n        VTK files can be visualized with Paraview (https://www.paraview.org)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_39to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_vtk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fields,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, 1); __PYX_ERR(0, 2783, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "to_vtk") < 0)) __PYX_ERR(0, 2783, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fields = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2783, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_38to_vtk(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_fields, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_38to_vtk(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_xCoords = NULL;
  PyObject *__pyx_v_yCoords = NULL;
  PyObject *__pyx_v_zCoords = NULL;
  PyObject *__pyx_v_rgrid = NULL;
  int __pyx_v_save_grid;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_scalar = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_vtk", 0);

  /* "ttcrpy/rgrid.pyx":2804
 *         VTK files can be visualized with Paraview (https://www.paraview.org)
 *         """
 *         xCoords = numpy_support.numpy_to_vtk(self.x)             # <<<<<<<<<<<<<<
 *         yCoords = numpy_support.numpy_to_vtk(np.array([0.0]))
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_xCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2805
 *         """
 *         xCoords = numpy_support.numpy_to_vtk(self.x)
 *         yCoords = numpy_support.numpy_to_vtk(np.array([0.0]))             # <<<<<<<<<<<<<<
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_float_0_0);
  __Pyx_GIVEREF(__pyx_float_0_0);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_float_0_0);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_yCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2806
 *         xCoords = numpy_support.numpy_to_vtk(self.x)
 *         yCoords = numpy_support.numpy_to_vtk(np.array([0.0]))
 *         zCoords = numpy_support.numpy_to_vtk(self.z)             # <<<<<<<<<<<<<<
 * 
 *         rgrid = vtk.vtkRectilinearGrid()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_numpy_to_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_z); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_zCoords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2808
 *         zCoords = numpy_support.numpy_to_vtk(self.z)
 * 
 *         rgrid = vtk.vtkRectilinearGrid()             # <<<<<<<<<<<<<<
 *         rgrid.SetDimensions(self._x.size(), 1, self._z.size())
 *         rgrid.SetXCoordinates(xCoords)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkRectilinearGrid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rgrid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2809
 * 
 *         rgrid = vtk.vtkRectilinearGrid()
 *         rgrid.SetDimensions(self._x.size(), 1, self._z.size())             # <<<<<<<<<<<<<<
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetDimensions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_3, __pyx_int_1, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2809, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_3, __pyx_int_1, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2809, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2810
 *         rgrid = vtk.vtkRectilinearGrid()
 *         rgrid.SetDimensions(self._x.size(), 1, self._z.size())
 *         rgrid.SetXCoordinates(xCoords)             # <<<<<<<<<<<<<<
 *         rgrid.SetYCoordinates(yCoords)
 *         rgrid.SetZCoordinates(zCoords)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetXCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_xCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_xCoords);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2811
 *         rgrid.SetDimensions(self._x.size(), 1, self._z.size())
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)             # <<<<<<<<<<<<<<
 *         rgrid.SetZCoordinates(zCoords)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetYCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_yCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_yCoords);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2812
 *         rgrid.SetXCoordinates(xCoords)
 *         rgrid.SetYCoordinates(yCoords)
 *         rgrid.SetZCoordinates(zCoords)             # <<<<<<<<<<<<<<
 * 
 *         save_grid = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_SetZCoordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_zCoords) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_zCoords);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2814
 *         rgrid.SetZCoordinates(zCoords)
 * 
 *         save_grid = False             # <<<<<<<<<<<<<<
 *         for fn in fields:
 *             data = fields[fn]
 */
  __pyx_v_save_grid = 0;

  /* "ttcrpy/rgrid.pyx":2815
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
    __pyx_t_1 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2815, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2815, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2815, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2815, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2816
 *         save_grid = False
 *         for fn in fields:
 *             data = fields[fn]             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(data, list):
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_fields, __pyx_v_fn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/rgrid.pyx":2818
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
    __pyx_t_10 = PyList_Check(__pyx_v_data); 
    __pyx_t_11 = (__pyx_t_10 != 0);
    if (__pyx_t_11) {

      /* "ttcrpy/rgrid.pyx":2819
 * 
 *             if isinstance(data, list):
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')             # <<<<<<<<<<<<<<
 *             else:
 *                 save_grid = True
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_save_raypaths); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyNumber_Add(__pyx_v_filename, __pyx_n_u__34); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_v_fn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_kp_u_vtp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_data, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2819, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_data, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2819, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2819, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_data);
        __Pyx_GIVEREF(__pyx_v_data);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_v_data);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2819, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2818
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
      goto __pyx_L5;
    }

    /* "ttcrpy/rgrid.pyx":2821
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 *                 save_grid = True             # <<<<<<<<<<<<<<
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 */
    /*else*/ {
      __pyx_v_save_grid = 1;

      /* "ttcrpy/rgrid.pyx":2822
 *             else:
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()             # <<<<<<<<<<<<<<
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_vtk); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_vtkDoubleArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_scalar, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2823
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_fn);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2824
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfComponents); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_int_1);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2825
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)             # <<<<<<<<<<<<<<
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._z.size())
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfTuples); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/rgrid.pyx":2826
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_11) {

        /* "ttcrpy/rgrid.pyx":2827
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._z.size())             # <<<<<<<<<<<<<<
 *                     if data.ndim == 2:
 *                         if data.shape != shape:
 */
        __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_x.size()); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2827, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_z.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2827, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2827, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
        __pyx_t_6 = 0;
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/rgrid.pyx":2828
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_11) {

          /* "ttcrpy/rgrid.pyx":2829
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2829, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_shape, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2829, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2829, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(__pyx_t_11)) {

            /* "ttcrpy/rgrid.pyx":2830
 *                     if data.ndim == 2:
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))             # <<<<<<<<<<<<<<
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_shape, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2830, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_fn);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2830, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2830, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 2830, __pyx_L1_error)

            /* "ttcrpy/rgrid.pyx":2829
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          }

          /* "ttcrpy/rgrid.pyx":2832
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')             # <<<<<<<<<<<<<<
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_flatten); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2832, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2832, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2832, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "ttcrpy/rgrid.pyx":2828
 *                 if data.size == self.get_number_of_nodes():
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
          goto __pyx_L7;
        }

        /* "ttcrpy/rgrid.pyx":2833
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_10) {
        } else {
          __pyx_t_11 = __pyx_t_10;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_10) {
        } else {
          __pyx_t_11 = __pyx_t_10;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __pyx_t_10;
        __pyx_L9_bool_binop_done:;
        if (likely(__pyx_t_11)) {

          /* "ttcrpy/rgrid.pyx":2835
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2835, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_shape);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2835, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_flatten); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2835, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2835, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 2835, __pyx_L1_error)
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2835, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":2833
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
          goto __pyx_L7;
        }

        /* "ttcrpy/rgrid.pyx":2837
 *                         tmp = data.reshape(shape).flatten(order='F')
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_ndim, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_fn);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 2837, __pyx_L1_error)
        }
        __pyx_L7:;

        /* "ttcrpy/rgrid.pyx":2838
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_12 = 0;
          __pyx_t_13 = NULL;
        } else {
          __pyx_t_12 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2838, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_13)) {
            if (likely(PyList_CheckExact(__pyx_t_4))) {
              if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_2); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2838, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2838, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_2); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2838, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2838, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_13(__pyx_t_4);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 2838, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":2839
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])             # <<<<<<<<<<<<<<
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2839, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2839, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_n, __pyx_t_5};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_n, __pyx_t_5};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_14 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2839, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_n);
            __Pyx_GIVEREF(__pyx_v_n);
            PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_7, __pyx_v_n);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_7, __pyx_t_5);
            __pyx_t_5 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":2838
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 */
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "ttcrpy/rgrid.pyx":2840
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._z.size()-1)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_14 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "ttcrpy/rgrid.pyx":2826
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size(), self._z.size())
 *                     if data.ndim == 2:
 */
        goto __pyx_L6;
      }

      /* "ttcrpy/rgrid.pyx":2841
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_6 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(__pyx_t_11)) {

        /* "ttcrpy/rgrid.pyx":2842
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._z.size()-1)             # <<<<<<<<<<<<<<
 *                     if data.ndim == 2:
 *                         if data.shape != shape:
 */
        __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_x.size() - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyInt_FromSize_t((__pyx_v_self->_z.size() - 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
        __pyx_t_2 = 0;
        __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "ttcrpy/rgrid.pyx":2843
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2843, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2843, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2843, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_11) {

          /* "ttcrpy/rgrid.pyx":2844
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_v_shape, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2844, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2844, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(__pyx_t_11)) {

            /* "ttcrpy/rgrid.pyx":2845
 *                     if data.ndim == 2:
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))             # <<<<<<<<<<<<<<
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_shape, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2845, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_fn);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2845, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2845, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_Raise(__pyx_t_6, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __PYX_ERR(0, 2845, __pyx_L1_error)

            /* "ttcrpy/rgrid.pyx":2844
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:
 *                         if data.shape != shape:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 */
          }

          /* "ttcrpy/rgrid.pyx":2847
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')             # <<<<<<<<<<<<<<
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_flatten); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2847, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2847, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 2847, __pyx_L1_error)
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2847, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/rgrid.pyx":2843
 *                 elif data.size == self.get_number_of_cells():
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:             # <<<<<<<<<<<<<<
 *                         if data.shape != shape:
 *                             raise ValueError('Field {0:s} has incorrect shape'.format(fn))
 */
          goto __pyx_L14;
        }

        /* "ttcrpy/rgrid.pyx":2848
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_10) {
        } else {
          __pyx_t_11 = __pyx_t_10;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_10) {
        } else {
          __pyx_t_11 = __pyx_t_10;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2848, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_11 = __pyx_t_10;
        __pyx_L16_bool_binop_done:;
        if (likely(__pyx_t_11)) {

          /* "ttcrpy/rgrid.pyx":2850
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_reshape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2850, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_shape) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_shape);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2850, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2850, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2850, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_order, __pyx_n_u_F) < 0) __PYX_ERR(0, 2850, __pyx_L1_error)
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2850, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "ttcrpy/rgrid.pyx":2848
 *                         # VTK stores in 'F' order
 *                         tmp = data.flatten(order='F')
 *                     elif data.ndim == 1 or (data.ndim == 2 and data.shape[0] == data.size):             # <<<<<<<<<<<<<<
 *                         # 'C' order assumed, reshape back and flatten to 'F' order
 *                         tmp = data.reshape(shape).flatten(order='F')
 */
          goto __pyx_L14;
        }

        /* "ttcrpy/rgrid.pyx":2852
 *                         tmp = data.reshape(shape).flatten(order='F')
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 */
        /*else*/ {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_ndim, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2852, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_fn);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2852, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2852, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_6, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __PYX_ERR(0, 2852, __pyx_L1_error)
        }
        __pyx_L14:;

        /* "ttcrpy/rgrid.pyx":2853
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2853, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2853, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
          __pyx_t_6 = __pyx_t_4; __Pyx_INCREF(__pyx_t_6); __pyx_t_12 = 0;
          __pyx_t_13 = NULL;
        } else {
          __pyx_t_12 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_13 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2853, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        for (;;) {
          if (likely(!__pyx_t_13)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2853, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2853, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2853, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2853, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_13(__pyx_t_6);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 2853, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "ttcrpy/rgrid.pyx":2854
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])             # <<<<<<<<<<<<<<
 *                     rgrid.GetCellData().AddArray(scalar)
 *                 else:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2854, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_v_n); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2854, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_5 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_n, __pyx_t_14};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2854, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_n, __pyx_t_14};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2854, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2854, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_n);
            __Pyx_GIVEREF(__pyx_v_n);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_n);
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_14);
            __pyx_t_14 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2854, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "ttcrpy/rgrid.pyx":2853
 *                     else:
 *                         raise ValueError('Field {0:s} has incorrect ndim'.format(fn))
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":2855
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetCellData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rgrid, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2855, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2855, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2855, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2855, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "ttcrpy/rgrid.pyx":2841
 *                         scalar.SetTuple1(n, tmp[n])
 *                     rgrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     shape = (self._x.size()-1, self._z.size()-1)
 *                     if data.ndim == 2:
 */
        goto __pyx_L6;
      }

      /* "ttcrpy/rgrid.pyx":2857
 *                     rgrid.GetCellData().AddArray(scalar)
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))             # <<<<<<<<<<<<<<
 * 
 *         if save_grid:
 */
      /*else*/ {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2857, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_fn);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2857, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2857, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 2857, __pyx_L1_error)
      }
      __pyx_L6:;
    }
    __pyx_L5:;

    /* "ttcrpy/rgrid.pyx":2815
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2859
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 */
  __pyx_t_11 = (__pyx_v_save_grid != 0);
  if (__pyx_t_11) {

    /* "ttcrpy/rgrid.pyx":2860
 * 
 *         if save_grid:
 *             writer = vtk.vtkXMLRectilinearGridWriter()             # <<<<<<<<<<<<<<
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkXMLRectilinearGridWriter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_writer = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2861
 *         if save_grid:
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')             # <<<<<<<<<<<<<<
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyNumber_Add(__pyx_v_filename, __pyx_kp_u_vtr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2862
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)             # <<<<<<<<<<<<<<
 *             writer.SetDataModeToBinary()
 *             writer.Update()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_rgrid) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_rgrid);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2863
 *             writer.SetFileName(filename+'.vtr')
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *             writer.Update()
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2864
 *             writer.SetInputData(rgrid)
 *             writer.SetDataModeToBinary()
 *             writer.Update()             # <<<<<<<<<<<<<<
 * 
 *     def _save_raypaths(self, rays, filename):
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2859
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLRectilinearGridWriter()
 *             writer.SetFileName(filename+'.vtr')
 */
  }

  /* "ttcrpy/rgrid.pyx":2783
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xCoords);
  __Pyx_XDECREF(__pyx_v_yCoords);
  __Pyx_XDECREF(__pyx_v_zCoords);
  __Pyx_XDECREF(__pyx_v_rgrid);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_scalar);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2866
 *             writer.Update()
 * 
 *     def _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_41_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_41_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rays = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_save_raypaths (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rays,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rays)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, 1); __PYX_ERR(0, 2866, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_save_raypaths") < 0)) __PYX_ERR(0, 2866, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rays = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2866, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_40_save_raypaths(((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_v_self), __pyx_v_rays, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_40_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_polydata = NULL;
  PyObject *__pyx_v_cellarray = NULL;
  PyObject *__pyx_v_pts = NULL;
  PyObject *__pyx_v_npts = NULL;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_save_raypaths", 0);

  /* "ttcrpy/rgrid.pyx":2867
 * 
 *     def _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()             # <<<<<<<<<<<<<<
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPolyData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_polydata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2868
 *     def _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()             # <<<<<<<<<<<<<<
 *         pts = vtk.vtkPoints()
 *         npts = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkCellArray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cellarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2869
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2870
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_npts = __pyx_int_0;

  /* "ttcrpy/rgrid.pyx":2871
 *         pts = vtk.vtkPoints()
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2871, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":2872
 *         npts = 0
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]             # <<<<<<<<<<<<<<
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_npts, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":2873
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_npts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_npts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2874
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/rgrid.pyx":2875
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2875, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":2876
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2876, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2876, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2876, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2876, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2876, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2876, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2877
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])             # <<<<<<<<<<<<<<
 *                 npts += 1
 *         polydata.SetPoints(pts)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_0);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_1);
      __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_npts, __pyx_t_11, __pyx_float_0_0, __pyx_t_12};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_npts, __pyx_t_11, __pyx_float_0_0, __pyx_t_12};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_13, __pyx_v_npts);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_13, __pyx_t_11);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_13, __pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_13, __pyx_t_12);
        __pyx_t_11 = 0;
        __pyx_t_12 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2878
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])
 *                 npts += 1             # <<<<<<<<<<<<<<
 *         polydata.SetPoints(pts)
 *         npts = 0
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2876
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":2879
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 1])
 *                 npts += 1
 *         polydata.SetPoints(pts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_pts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_pts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2880
 *                 npts += 1
 *         polydata.SetPoints(pts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/rgrid.pyx":2881
 *         polydata.SetPoints(pts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2881, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/rgrid.pyx":2882
 *         npts = 0
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()             # <<<<<<<<<<<<<<
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkPolyLine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2883
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetNumberOfIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shape); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2884
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2884, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2884, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2884, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2884, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2885
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)             # <<<<<<<<<<<<<<
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_2 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetId); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_p);
        __Pyx_GIVEREF(__pyx_v_p);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_13, __pyx_v_p);
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_13, __pyx_v_npts);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2886
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1             # <<<<<<<<<<<<<<
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/rgrid.pyx":2884
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/rgrid.pyx":2887
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 *             cellarray.InsertNextCell(line)             # <<<<<<<<<<<<<<
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cellarray, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_line);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/rgrid.pyx":2888
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)             # <<<<<<<<<<<<<<
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetLines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_v_cellarray) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2889
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()             # <<<<<<<<<<<<<<
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkXMLPolyDataWriter); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_writer = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2890
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2891
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)             # <<<<<<<<<<<<<<
 *         writer.SetDataModeToBinary()
 *         writer.Update()
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_polydata) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_polydata);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2892
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *         writer.Update()
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2893
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 *         writer.Update()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/rgrid.pyx":2866
 *             writer.Update()
 * 
 *     def _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_polydata);
  __Pyx_XDECREF(__pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_v_pts);
  __Pyx_XDECREF(__pyx_v_npts);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":2897
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays[] = "\n        data_kernel_straight_rays(Tx, Rx, grx, grz, aniso=False) -> L\n\n        Raytracing with straight rays in 2D\n\n        Parameters\n        ----------\n        Tx : np.ndarray\n            source coordinates, nTx by 2\n                - 1st column contains X coordinates,\n                - 2nd contains Z coordinates\n        Rx : np.ndarray\n            receiver coordinates, nTx by 2\n                - 1st column contains X coordinates,\n                - 2nd contains Z coordinates\n        grx : np.ndarray\n            grid node coordinates along x\n        grz : np.ndarray\n            grid node coordinates along z\n        aniso : bool\n            compute L for elliptically anisotropic medium (True) or isotropic\n            medium (False)\n\n        Returns\n        -------\n        L : scipy csr_matrix\n            data kernel matrix (tt = L*slowness)\n\n        Note\n        ----\n        Tx and Rx should contain the same number of rows, each row corresponding\n        to a source-receiver pair\n        ";
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays = {"data_kernel_straight_rays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_Tx = 0;
  PyArrayObject *__pyx_v_Rx = 0;
  PyArrayObject *__pyx_v_grx = 0;
  PyArrayObject *__pyx_v_grz = 0;
  PyObject *__pyx_v_aniso = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("data_kernel_straight_rays (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_Tx,&__pyx_n_s_Rx,&__pyx_n_s_grx,&__pyx_n_s_grz,&__pyx_n_s_aniso,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "ttcrpy/rgrid.pyx":2901
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 *                                   np.ndarray[np.double_t, ndim=1] grz,
 *                                   aniso=False):             # <<<<<<<<<<<<<<
 *         """
 *         data_kernel_straight_rays(Tx, Rx, grx, grz, aniso=False) -> L
 */
    values[4] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Tx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Rx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 4, 5, 1); __PYX_ERR(0, 2897, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_grx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 4, 5, 2); __PYX_ERR(0, 2897, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_grz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 4, 5, 3); __PYX_ERR(0, 2897, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aniso);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "data_kernel_straight_rays") < 0)) __PYX_ERR(0, 2897, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_Tx = ((PyArrayObject *)values[0]);
    __pyx_v_Rx = ((PyArrayObject *)values[1]);
    __pyx_v_grx = ((PyArrayObject *)values[2]);
    __pyx_v_grz = ((PyArrayObject *)values[3]);
    __pyx_v_aniso = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("data_kernel_straight_rays", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2897, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.data_kernel_straight_rays", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_Tx), __pyx_ptype_5numpy_ndarray, 1, "Tx", 0))) __PYX_ERR(0, 2897, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_Rx), __pyx_ptype_5numpy_ndarray, 1, "Rx", 0))) __PYX_ERR(0, 2898, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_grx), __pyx_ptype_5numpy_ndarray, 1, "grx", 0))) __PYX_ERR(0, 2899, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_grz), __pyx_ptype_5numpy_ndarray, 1, "grz", 0))) __PYX_ERR(0, 2900, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays(__pyx_v_Tx, __pyx_v_Rx, __pyx_v_grx, __pyx_v_grz, __pyx_v_aniso);

  /* "ttcrpy/rgrid.pyx":2897
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays(PyArrayObject *__pyx_v_Tx, PyArrayObject *__pyx_v_Rx, PyArrayObject *__pyx_v_grx, PyArrayObject *__pyx_v_grz, PyObject *__pyx_v_aniso) {
  size_t __pyx_v_nTx;
  size_t __pyx_v_n_grx;
  size_t __pyx_v_n_grz;
  size_t __pyx_v_nCells;
  double __pyx_v_small;
  PyObject *__pyx_v_data_p = NULL;
  PyObject *__pyx_v_indices_p = NULL;
  PyObject *__pyx_v_indptr_p = NULL;
  size_t __pyx_v_k;
  size_t __pyx_v_ix;
  size_t __pyx_v_iz;
  double __pyx_v_x;
  double __pyx_v_z;
  double __pyx_v_xs;
  double __pyx_v_xr;
  double __pyx_v_zs;
  double __pyx_v_zr;
  double __pyx_v_dtmp;
  double __pyx_v_dlx;
  double __pyx_v_dlz;
  double __pyx_v_m;
  double __pyx_v_b;
  double __pyx_v_ze;
  double __pyx_v_xe;
  int64_t __pyx_v_iCell;
  bool __pyx_v_up;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_zi = NULL;
  double __pyx_v_dl;
  PyObject *__pyx_v_L = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_Rx;
  __Pyx_Buffer __pyx_pybuffer_Rx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_Tx;
  __Pyx_Buffer __pyx_pybuffer_Tx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_grx;
  __Pyx_Buffer __pyx_pybuffer_grx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_grz;
  __Pyx_Buffer __pyx_pybuffer_grz;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __pyx_t_5numpy_double_t __pyx_t_14;
  int __pyx_t_15;
  double __pyx_t_16;
  double __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_kernel_straight_rays", 0);
  __pyx_pybuffer_Tx.pybuffer.buf = NULL;
  __pyx_pybuffer_Tx.refcount = 0;
  __pyx_pybuffernd_Tx.data = NULL;
  __pyx_pybuffernd_Tx.rcbuffer = &__pyx_pybuffer_Tx;
  __pyx_pybuffer_Rx.pybuffer.buf = NULL;
  __pyx_pybuffer_Rx.refcount = 0;
  __pyx_pybuffernd_Rx.data = NULL;
  __pyx_pybuffernd_Rx.rcbuffer = &__pyx_pybuffer_Rx;
  __pyx_pybuffer_grx.pybuffer.buf = NULL;
  __pyx_pybuffer_grx.refcount = 0;
  __pyx_pybuffernd_grx.data = NULL;
  __pyx_pybuffernd_grx.rcbuffer = &__pyx_pybuffer_grx;
  __pyx_pybuffer_grz.pybuffer.buf = NULL;
  __pyx_pybuffer_grz.refcount = 0;
  __pyx_pybuffernd_grz.data = NULL;
  __pyx_pybuffernd_grz.rcbuffer = &__pyx_pybuffer_grz;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer, (PyObject*)__pyx_v_Tx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  }
  __pyx_pybuffernd_Tx.diminfo[0].strides = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_Tx.diminfo[0].shape = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_Tx.diminfo[1].strides = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_Tx.diminfo[1].shape = __pyx_pybuffernd_Tx.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer, (PyObject*)__pyx_v_Rx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  }
  __pyx_pybuffernd_Rx.diminfo[0].strides = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_Rx.diminfo[0].shape = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_Rx.diminfo[1].strides = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_Rx.diminfo[1].shape = __pyx_pybuffernd_Rx.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_grx.rcbuffer->pybuffer, (PyObject*)__pyx_v_grx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  }
  __pyx_pybuffernd_grx.diminfo[0].strides = __pyx_pybuffernd_grx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_grx.diminfo[0].shape = __pyx_pybuffernd_grx.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_grz.rcbuffer->pybuffer, (PyObject*)__pyx_v_grz, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  }
  __pyx_pybuffernd_grz.diminfo[0].strides = __pyx_pybuffernd_grz.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_grz.diminfo[0].shape = __pyx_pybuffernd_grz.rcbuffer->pybuffer.shape[0];

  /* "ttcrpy/rgrid.pyx":2936
 *         """
 * 
 *         cdef size_t nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_grz = grz.shape[0]
 */
  __pyx_v_nTx = (__pyx_v_Tx->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":2937
 * 
 *         cdef size_t nTx = Tx.shape[0]
 *         cdef size_t n_grx = grx.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t n_grz = grz.shape[0]
 *         cdef size_t nCells = (n_grx-1)*(n_grz-1)
 */
  __pyx_v_n_grx = (__pyx_v_grx->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":2938
 *         cdef size_t nTx = Tx.shape[0]
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_grz = grz.shape[0]             # <<<<<<<<<<<<<<
 *         cdef size_t nCells = (n_grx-1)*(n_grz-1)
 * 
 */
  __pyx_v_n_grz = (__pyx_v_grz->dimensions[0]);

  /* "ttcrpy/rgrid.pyx":2939
 *         cdef size_t n_grx = grx.shape[0]
 *         cdef size_t n_grz = grz.shape[0]
 *         cdef size_t nCells = (n_grx-1)*(n_grz-1)             # <<<<<<<<<<<<<<
 * 
 *         cdef double small = 1.e-10
 */
  __pyx_v_nCells = ((__pyx_v_n_grx - 1) * (__pyx_v_n_grz - 1));

  /* "ttcrpy/rgrid.pyx":2941
 *         cdef size_t nCells = (n_grx-1)*(n_grz-1)
 * 
 *         cdef double small = 1.e-10             # <<<<<<<<<<<<<<
 * 
 *         data_p = []
 */
  __pyx_v_small = 1.e-10;

  /* "ttcrpy/rgrid.pyx":2943
 *         cdef double small = 1.e-10
 * 
 *         data_p = []             # <<<<<<<<<<<<<<
 *         indices_p = []
 *         indptr_p = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2944
 * 
 *         data_p = []
 *         indices_p = []             # <<<<<<<<<<<<<<
 *         indptr_p = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_indices_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2945
 *         data_p = []
 *         indices_p = []
 *         indptr_p = []             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t k = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_indptr_p = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":2947
 *         indptr_p = []
 * 
 *         cdef size_t k = 0             # <<<<<<<<<<<<<<
 *         cdef size_t ix = 0
 *         cdef size_t iz = 0
 */
  __pyx_v_k = 0;

  /* "ttcrpy/rgrid.pyx":2948
 * 
 *         cdef size_t k = 0
 *         cdef size_t ix = 0             # <<<<<<<<<<<<<<
 *         cdef size_t iz = 0
 * 
 */
  __pyx_v_ix = 0;

  /* "ttcrpy/rgrid.pyx":2949
 *         cdef size_t k = 0
 *         cdef size_t ix = 0
 *         cdef size_t iz = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef double x, z, xs, xr, zs, zr, dtmp, dlx, dlz, m, b, ze, xe
 */
  __pyx_v_iz = 0;

  /* "ttcrpy/rgrid.pyx":2956
 *         cdef Py_ssize_t n
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             indptr_p.append(k)
 * 
 */
  __pyx_t_2 = __pyx_v_nTx;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_n = __pyx_t_4;

    /* "ttcrpy/rgrid.pyx":2957
 * 
 *         for n in range(nTx):
 *             indptr_p.append(k)             # <<<<<<<<<<<<<<
 * 
 *             xs = Tx[n, 0]
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indptr_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 2957, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":2959
 *             indptr_p.append(k)
 * 
 *             xs = Tx[n, 0]             # <<<<<<<<<<<<<<
 *             zs = Tx[n, 1]
 *             xr = Rx[n, 0]
 */
    __pyx_t_6 = __pyx_v_n;
    __pyx_t_7 = 0;
    __pyx_t_8 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Tx.diminfo[0].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Tx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Tx.diminfo[1].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Tx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 2959, __pyx_L1_error)
    }
    __pyx_v_xs = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Tx.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_Tx.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_Tx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":2960
 * 
 *             xs = Tx[n, 0]
 *             zs = Tx[n, 1]             # <<<<<<<<<<<<<<
 *             xr = Rx[n, 0]
 *             zr = Rx[n, 1]
 */
    __pyx_t_7 = __pyx_v_n;
    __pyx_t_6 = 1;
    __pyx_t_8 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Tx.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Tx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Tx.diminfo[1].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Tx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 2960, __pyx_L1_error)
    }
    __pyx_v_zs = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Tx.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_Tx.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_Tx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":2961
 *             xs = Tx[n, 0]
 *             zs = Tx[n, 1]
 *             xr = Rx[n, 0]             # <<<<<<<<<<<<<<
 *             zr = Rx[n, 1]
 * 
 */
    __pyx_t_6 = __pyx_v_n;
    __pyx_t_7 = 0;
    __pyx_t_8 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Rx.diminfo[0].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Rx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Rx.diminfo[1].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Rx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 2961, __pyx_L1_error)
    }
    __pyx_v_xr = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Rx.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_Rx.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_Rx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":2962
 *             zs = Tx[n, 1]
 *             xr = Rx[n, 0]
 *             zr = Rx[n, 1]             # <<<<<<<<<<<<<<
 * 
 *             if xs > xr:  # on va de s  r, on veut x croissant
 */
    __pyx_t_7 = __pyx_v_n;
    __pyx_t_6 = 1;
    __pyx_t_8 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_Rx.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_Rx.diminfo[0].shape)) __pyx_t_8 = 0;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_pybuffernd_Rx.diminfo[1].shape;
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_8 = 1;
    } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_Rx.diminfo[1].shape)) __pyx_t_8 = 1;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(0, 2962, __pyx_L1_error)
    }
    __pyx_v_zr = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_Rx.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_Rx.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_Rx.diminfo[1].strides));

    /* "ttcrpy/rgrid.pyx":2964
 *             zr = Rx[n, 1]
 * 
 *             if xs > xr:  # on va de s  r, on veut x croissant             # <<<<<<<<<<<<<<
 *                 dtmp = xs
 *                 xs = xr
 */
    __pyx_t_9 = ((__pyx_v_xs > __pyx_v_xr) != 0);
    if (__pyx_t_9) {

      /* "ttcrpy/rgrid.pyx":2965
 * 
 *             if xs > xr:  # on va de s  r, on veut x croissant
 *                 dtmp = xs             # <<<<<<<<<<<<<<
 *                 xs = xr
 *                 xr = dtmp
 */
      __pyx_v_dtmp = __pyx_v_xs;

      /* "ttcrpy/rgrid.pyx":2966
 *             if xs > xr:  # on va de s  r, on veut x croissant
 *                 dtmp = xs
 *                 xs = xr             # <<<<<<<<<<<<<<
 *                 xr = dtmp
 *                 dtmp = zs
 */
      __pyx_v_xs = __pyx_v_xr;

      /* "ttcrpy/rgrid.pyx":2967
 *                 dtmp = xs
 *                 xs = xr
 *                 xr = dtmp             # <<<<<<<<<<<<<<
 *                 dtmp = zs
 *                 zs = zr
 */
      __pyx_v_xr = __pyx_v_dtmp;

      /* "ttcrpy/rgrid.pyx":2968
 *                 xs = xr
 *                 xr = dtmp
 *                 dtmp = zs             # <<<<<<<<<<<<<<
 *                 zs = zr
 *                 zr = dtmp
 */
      __pyx_v_dtmp = __pyx_v_zs;

      /* "ttcrpy/rgrid.pyx":2969
 *                 xr = dtmp
 *                 dtmp = zs
 *                 zs = zr             # <<<<<<<<<<<<<<
 *                 zr = dtmp
 * 
 */
      __pyx_v_zs = __pyx_v_zr;

      /* "ttcrpy/rgrid.pyx":2970
 *                 dtmp = zs
 *                 zs = zr
 *                 zr = dtmp             # <<<<<<<<<<<<<<
 * 
 *             # point de dpart
 */
      __pyx_v_zr = __pyx_v_dtmp;

      /* "ttcrpy/rgrid.pyx":2964
 *             zr = Rx[n, 1]
 * 
 *             if xs > xr:  # on va de s  r, on veut x croissant             # <<<<<<<<<<<<<<
 *                 dtmp = xs
 *                 xs = xr
 */
    }

    /* "ttcrpy/rgrid.pyx":2973
 * 
 *             # point de dpart
 *             x = xs             # <<<<<<<<<<<<<<
 *             z = zs
 * 
 */
    __pyx_v_x = __pyx_v_xs;

    /* "ttcrpy/rgrid.pyx":2974
 *             # point de dpart
 *             x = xs
 *             z = zs             # <<<<<<<<<<<<<<
 * 
 *             if abs(zs-zr)<small: # rai horizontal
 */
    __pyx_v_z = __pyx_v_zs;

    /* "ttcrpy/rgrid.pyx":2976
 *             z = zs
 * 
 *             if abs(zs-zr)<small: # rai horizontal             # <<<<<<<<<<<<<<
 * 
 *                 for ix in range(n_grx-1):
 */
    __pyx_t_9 = ((fabs((__pyx_v_zs - __pyx_v_zr)) < __pyx_v_small) != 0);
    if (__pyx_t_9) {

      /* "ttcrpy/rgrid.pyx":2978
 *             if abs(zs-zr)<small: # rai horizontal
 * 
 *                 for ix in range(n_grx-1):             # <<<<<<<<<<<<<<
 *                     if x < grx[ix+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grx - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_ix = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":2979
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        __pyx_t_13 = (__pyx_v_ix + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 2979, __pyx_L1_error)
        }
        __pyx_t_9 = ((__pyx_v_x < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grx.diminfo[0].strides))) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":2980
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:
 *                         break             # <<<<<<<<<<<<<<
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 */
          goto __pyx_L8_break;

          /* "ttcrpy/rgrid.pyx":2979
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        }
      }
      __pyx_L8_break:;

      /* "ttcrpy/rgrid.pyx":2981
 *                     if x < grx[ix+1]:
 *                         break
 *                 for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                     if z < grz[iz+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grz - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_iz = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":2982
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        __pyx_t_13 = (__pyx_v_iz + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 2982, __pyx_L1_error)
        }
        __pyx_t_9 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":2983
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                 while x < xr:
 */
          goto __pyx_L11_break;

          /* "ttcrpy/rgrid.pyx":2982
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        }
      }
      __pyx_L11_break:;

      /* "ttcrpy/rgrid.pyx":2985
 *                         break
 * 
 *                 while x < xr:             # <<<<<<<<<<<<<<
 *                     iCell = ix*(n_grz-1) + iz
 * 
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_x < __pyx_v_xr) != 0);
        if (!__pyx_t_9) break;

        /* "ttcrpy/rgrid.pyx":2986
 * 
 *                 while x < xr:
 *                     iCell = ix*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                     dlx = ( grx[ix+1] if grx[ix+1]<xr else xr ) - x
 */
        __pyx_v_iCell = ((__pyx_v_ix * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

        /* "ttcrpy/rgrid.pyx":2988
 *                     iCell = ix*(n_grz-1) + iz
 * 
 *                     dlx = ( grx[ix+1] if grx[ix+1]<xr else xr ) - x             # <<<<<<<<<<<<<<
 * 
 *                     indices_p.append(iCell)
 */
        __pyx_t_10 = (__pyx_v_ix + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 2988, __pyx_L1_error)
        }
        if ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grx.diminfo[0].strides)) < __pyx_v_xr) != 0)) {
          __pyx_t_11 = (__pyx_v_ix + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 2988, __pyx_L1_error)
          }
          __pyx_t_14 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grx.diminfo[0].strides));
        } else {
          __pyx_t_14 = __pyx_v_xr;
        }
        __pyx_v_dlx = (__pyx_t_14 - __pyx_v_x);

        /* "ttcrpy/rgrid.pyx":2990
 *                     dlx = ( grx[ix+1] if grx[ix+1]<xr else xr ) - x
 * 
 *                     indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                     data_p.append(dlx)
 *                     k += 1
 */
        __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2990, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 2990, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":2991
 * 
 *                     indices_p.append(iCell)
 *                     data_p.append(dlx)             # <<<<<<<<<<<<<<
 *                     k += 1
 * 
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 2991, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":2992
 *                     indices_p.append(iCell)
 *                     data_p.append(dlx)
 *                     k += 1             # <<<<<<<<<<<<<<
 * 
 *                     ix += 1
 */
        __pyx_v_k = (__pyx_v_k + 1);

        /* "ttcrpy/rgrid.pyx":2994
 *                     k += 1
 * 
 *                     ix += 1             # <<<<<<<<<<<<<<
 *                     x = grx[ix]
 * 
 */
        __pyx_v_ix = (__pyx_v_ix + 1);

        /* "ttcrpy/rgrid.pyx":2995
 * 
 *                     ix += 1
 *                     x = grx[ix]             # <<<<<<<<<<<<<<
 * 
 *             elif abs(xs-xr) < small:   # rai vertical
 */
        __pyx_t_10 = __pyx_v_ix;
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 2995, __pyx_L1_error)
        }
        __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grx.diminfo[0].strides));
      }

      /* "ttcrpy/rgrid.pyx":2976
 *             z = zs
 * 
 *             if abs(zs-zr)<small: # rai horizontal             # <<<<<<<<<<<<<<
 * 
 *                 for ix in range(n_grx-1):
 */
      goto __pyx_L6;
    }

    /* "ttcrpy/rgrid.pyx":2997
 *                     x = grx[ix]
 * 
 *             elif abs(xs-xr) < small:   # rai vertical             # <<<<<<<<<<<<<<
 *                 if zs > zr:
 *                     dtmp = zs
 */
    __pyx_t_9 = ((fabs((__pyx_v_xs - __pyx_v_xr)) < __pyx_v_small) != 0);
    if (__pyx_t_9) {

      /* "ttcrpy/rgrid.pyx":2998
 * 
 *             elif abs(xs-xr) < small:   # rai vertical
 *                 if zs > zr:             # <<<<<<<<<<<<<<
 *                     dtmp = zs
 *                     zs = zr
 */
      __pyx_t_9 = ((__pyx_v_zs > __pyx_v_zr) != 0);
      if (__pyx_t_9) {

        /* "ttcrpy/rgrid.pyx":2999
 *             elif abs(xs-xr) < small:   # rai vertical
 *                 if zs > zr:
 *                     dtmp = zs             # <<<<<<<<<<<<<<
 *                     zs = zr
 *                     zr = dtmp
 */
        __pyx_v_dtmp = __pyx_v_zs;

        /* "ttcrpy/rgrid.pyx":3000
 *                 if zs > zr:
 *                     dtmp = zs
 *                     zs = zr             # <<<<<<<<<<<<<<
 *                     zr = dtmp
 *                 z = zs
 */
        __pyx_v_zs = __pyx_v_zr;

        /* "ttcrpy/rgrid.pyx":3001
 *                     dtmp = zs
 *                     zs = zr
 *                     zr = dtmp             # <<<<<<<<<<<<<<
 *                 z = zs
 * 
 */
        __pyx_v_zr = __pyx_v_dtmp;

        /* "ttcrpy/rgrid.pyx":2998
 * 
 *             elif abs(xs-xr) < small:   # rai vertical
 *                 if zs > zr:             # <<<<<<<<<<<<<<
 *                     dtmp = zs
 *                     zs = zr
 */
      }

      /* "ttcrpy/rgrid.pyx":3002
 *                     zs = zr
 *                     zr = dtmp
 *                 z = zs             # <<<<<<<<<<<<<<
 * 
 *                 for ix in range(n_grx-1):
 */
      __pyx_v_z = __pyx_v_zs;

      /* "ttcrpy/rgrid.pyx":3004
 *                 z = zs
 * 
 *                 for ix in range(n_grx-1):             # <<<<<<<<<<<<<<
 *                     if x < grx[ix+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grx - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_ix = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":3005
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        __pyx_t_13 = (__pyx_v_ix + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3005, __pyx_L1_error)
        }
        __pyx_t_9 = ((__pyx_v_x < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grx.diminfo[0].strides))) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":3006
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:
 *                         break             # <<<<<<<<<<<<<<
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 */
          goto __pyx_L17_break;

          /* "ttcrpy/rgrid.pyx":3005
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        }
      }
      __pyx_L17_break:;

      /* "ttcrpy/rgrid.pyx":3007
 *                     if x < grx[ix+1]:
 *                         break
 *                 for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                     if z < grz[iz+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grz - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_iz = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":3008
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        __pyx_t_13 = (__pyx_v_iz + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3008, __pyx_L1_error)
        }
        __pyx_t_9 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
        if (__pyx_t_9) {

          /* "ttcrpy/rgrid.pyx":3009
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                 while z < zr:
 */
          goto __pyx_L20_break;

          /* "ttcrpy/rgrid.pyx":3008
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        }
      }
      __pyx_L20_break:;

      /* "ttcrpy/rgrid.pyx":3011
 *                         break
 * 
 *                 while z < zr:             # <<<<<<<<<<<<<<
 *                     iCell = ix*(n_grz-1) + iz
 * 
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_z < __pyx_v_zr) != 0);
        if (!__pyx_t_9) break;

        /* "ttcrpy/rgrid.pyx":3012
 * 
 *                 while z < zr:
 *                     iCell = ix*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                     dlz = ( grz[iz+1] if grz[iz+1]<zr else zr ) - z
 */
        __pyx_v_iCell = ((__pyx_v_ix * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

        /* "ttcrpy/rgrid.pyx":3014
 *                     iCell = ix*(n_grz-1) + iz
 * 
 *                     dlz = ( grz[iz+1] if grz[iz+1]<zr else zr ) - z             # <<<<<<<<<<<<<<
 * 
 *                     if not aniso:
 */
        __pyx_t_10 = (__pyx_v_iz + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3014, __pyx_L1_error)
        }
        if ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides)) < __pyx_v_zr) != 0)) {
          __pyx_t_11 = (__pyx_v_iz + 1);
          __pyx_t_8 = -1;
          if (unlikely(__pyx_t_11 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            __PYX_ERR(0, 3014, __pyx_L1_error)
          }
          __pyx_t_14 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_grz.diminfo[0].strides));
        } else {
          __pyx_t_14 = __pyx_v_zr;
        }
        __pyx_v_dlz = (__pyx_t_14 - __pyx_v_z);

        /* "ttcrpy/rgrid.pyx":3016
 *                     dlz = ( grz[iz+1] if grz[iz+1]<zr else zr ) - z
 * 
 *                     if not aniso:             # <<<<<<<<<<<<<<
 *                         indices_p.append(iCell)
 *                     else:
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_aniso); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3016, __pyx_L1_error)
        __pyx_t_15 = ((!__pyx_t_9) != 0);
        if (__pyx_t_15) {

          /* "ttcrpy/rgrid.pyx":3017
 * 
 *                     if not aniso:
 *                         indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                     else:
 *                         indices_p.append(iCell + nCells)
 */
          __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3017, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3017, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ttcrpy/rgrid.pyx":3016
 *                     dlz = ( grz[iz+1] if grz[iz+1]<zr else zr ) - z
 * 
 *                     if not aniso:             # <<<<<<<<<<<<<<
 *                         indices_p.append(iCell)
 *                     else:
 */
          goto __pyx_L24;
        }

        /* "ttcrpy/rgrid.pyx":3019
 *                         indices_p.append(iCell)
 *                     else:
 *                         indices_p.append(iCell + nCells)             # <<<<<<<<<<<<<<
 *                     data_p.append(dlz)
 *                     k += 1
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_iCell + __pyx_v_nCells)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3019, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3019, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __pyx_L24:;

        /* "ttcrpy/rgrid.pyx":3020
 *                     else:
 *                         indices_p.append(iCell + nCells)
 *                     data_p.append(dlz)             # <<<<<<<<<<<<<<
 *                     k += 1
 * 
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3020, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3020, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":3021
 *                         indices_p.append(iCell + nCells)
 *                     data_p.append(dlz)
 *                     k += 1             # <<<<<<<<<<<<<<
 * 
 *                     iz += 1
 */
        __pyx_v_k = (__pyx_v_k + 1);

        /* "ttcrpy/rgrid.pyx":3023
 *                     k += 1
 * 
 *                     iz += 1             # <<<<<<<<<<<<<<
 *                     z = grz[iz]
 *             else:   # rai oblique
 */
        __pyx_v_iz = (__pyx_v_iz + 1);

        /* "ttcrpy/rgrid.pyx":3024
 * 
 *                     iz += 1
 *                     z = grz[iz]             # <<<<<<<<<<<<<<
 *             else:   # rai oblique
 *                 m = (zr-zs)/(xr-xs)
 */
        __pyx_t_10 = __pyx_v_iz;
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3024, __pyx_L1_error)
        }
        __pyx_v_z = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides));
      }

      /* "ttcrpy/rgrid.pyx":2997
 *                     x = grx[ix]
 * 
 *             elif abs(xs-xr) < small:   # rai vertical             # <<<<<<<<<<<<<<
 *                 if zs > zr:
 *                     dtmp = zs
 */
      goto __pyx_L6;
    }

    /* "ttcrpy/rgrid.pyx":3026
 *                     z = grz[iz]
 *             else:   # rai oblique
 *                 m = (zr-zs)/(xr-xs)             # <<<<<<<<<<<<<<
 *                 b = zr - m*xr
 *                 up = m>0
 */
    /*else*/ {
      __pyx_t_16 = (__pyx_v_zr - __pyx_v_zs);
      __pyx_t_17 = (__pyx_v_xr - __pyx_v_xs);
      if (unlikely(__pyx_t_17 == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 3026, __pyx_L1_error)
      }
      __pyx_v_m = (__pyx_t_16 / __pyx_t_17);

      /* "ttcrpy/rgrid.pyx":3027
 *             else:   # rai oblique
 *                 m = (zr-zs)/(xr-xs)
 *                 b = zr - m*xr             # <<<<<<<<<<<<<<
 *                 up = m>0
 * 
 */
      __pyx_v_b = (__pyx_v_zr - (__pyx_v_m * __pyx_v_xr));

      /* "ttcrpy/rgrid.pyx":3028
 *                 m = (zr-zs)/(xr-xs)
 *                 b = zr - m*xr
 *                 up = m>0             # <<<<<<<<<<<<<<
 * 
 *                 for ix in range(n_grx-1):
 */
      __pyx_v_up = (__pyx_v_m > 0.0);

      /* "ttcrpy/rgrid.pyx":3030
 *                 up = m>0
 * 
 *                 for ix in range(n_grx-1):             # <<<<<<<<<<<<<<
 *                     if x < grx[ix+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grx - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_ix = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":3031
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        __pyx_t_13 = (__pyx_v_ix + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3031, __pyx_L1_error)
        }
        __pyx_t_15 = ((__pyx_v_x < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grx.diminfo[0].strides))) != 0);
        if (__pyx_t_15) {

          /* "ttcrpy/rgrid.pyx":3032
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:
 *                         break             # <<<<<<<<<<<<<<
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 */
          goto __pyx_L26_break;

          /* "ttcrpy/rgrid.pyx":3031
 * 
 *                 for ix in range(n_grx-1):
 *                     if x < grx[ix+1]:             # <<<<<<<<<<<<<<
 *                         break
 *                 for iz in range(n_grz-1):
 */
        }
      }
      __pyx_L26_break:;

      /* "ttcrpy/rgrid.pyx":3033
 *                     if x < grx[ix+1]:
 *                         break
 *                 for iz in range(n_grz-1):             # <<<<<<<<<<<<<<
 *                     if z < grz[iz+1]:
 *                         break
 */
      __pyx_t_10 = (__pyx_v_n_grz - 1);
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_iz = __pyx_t_12;

        /* "ttcrpy/rgrid.pyx":3034
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        __pyx_t_13 = (__pyx_v_iz + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3034, __pyx_L1_error)
        }
        __pyx_t_15 = ((__pyx_v_z < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_grz.diminfo[0].strides))) != 0);
        if (__pyx_t_15) {

          /* "ttcrpy/rgrid.pyx":3035
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                 while x < xr:
 */
          goto __pyx_L29_break;

          /* "ttcrpy/rgrid.pyx":3034
 *                         break
 *                 for iz in range(n_grz-1):
 *                     if z < grz[iz+1]:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        }
      }
      __pyx_L29_break:;

      /* "ttcrpy/rgrid.pyx":3037
 *                         break
 * 
 *                 while x < xr:             # <<<<<<<<<<<<<<
 * 
 *                     zi = m*grx[ix+1] + b
 */
      while (1) {
        __pyx_t_15 = ((__pyx_v_x < __pyx_v_xr) != 0);
        if (!__pyx_t_15) break;

        /* "ttcrpy/rgrid.pyx":3039
 *                 while x < xr:
 * 
 *                     zi = m*grx[ix+1] + b             # <<<<<<<<<<<<<<
 * 
 *                     if up:
 */
        __pyx_t_10 = (__pyx_v_ix + 1);
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3039, __pyx_L1_error)
        }
        __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grx.diminfo[0].strides))) + __pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_zi, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "ttcrpy/rgrid.pyx":3041
 *                     zi = m*grx[ix+1] + b
 * 
 *                     if up:             # <<<<<<<<<<<<<<
 *                         while z < zi and z < zr:
 *                             iCell = ix*(n_grz-1) + iz
 */
        __pyx_t_15 = (__pyx_v_up != 0);
        if (__pyx_t_15) {

          /* "ttcrpy/rgrid.pyx":3042
 * 
 *                     if up:
 *                         while z < zi and z < zr:             # <<<<<<<<<<<<<<
 *                             iCell = ix*(n_grz-1) + iz
 * 
 */
          while (1) {
            __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3042, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3042, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3042, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (__pyx_t_9) {
            } else {
              __pyx_t_15 = __pyx_t_9;
              goto __pyx_L36_bool_binop_done;
            }
            __pyx_t_9 = ((__pyx_v_z < __pyx_v_zr) != 0);
            __pyx_t_15 = __pyx_t_9;
            __pyx_L36_bool_binop_done:;
            if (!__pyx_t_15) break;

            /* "ttcrpy/rgrid.pyx":3043
 *                     if up:
 *                         while z < zi and z < zr:
 *                             iCell = ix*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                             ze = grz[iz+1] if grz[iz+1]<zi else zi
 */
            __pyx_v_iCell = ((__pyx_v_ix * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

            /* "ttcrpy/rgrid.pyx":3045
 *                             iCell = ix*(n_grz-1) + iz
 * 
 *                             ze = grz[iz+1] if grz[iz+1]<zi else zi             # <<<<<<<<<<<<<<
 *                             ze = ze if ze<zr else zr
 *                             xe = (ze-b)/m
 */
            __pyx_t_10 = (__pyx_v_iz + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 3045, __pyx_L1_error)
            }
            __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3045, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3045, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 3045, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_15) {
              __pyx_t_10 = (__pyx_v_iz + 1);
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 3045, __pyx_L1_error)
              }
              __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides));
            } else {
              __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 3045, __pyx_L1_error)
              __pyx_t_17 = __pyx_t_16;
            }
            __pyx_v_ze = __pyx_t_17;

            /* "ttcrpy/rgrid.pyx":3046
 * 
 *                             ze = grz[iz+1] if grz[iz+1]<zi else zi
 *                             ze = ze if ze<zr else zr             # <<<<<<<<<<<<<<
 *                             xe = (ze-b)/m
 *                             dlx = xe - x
 */
            if (((__pyx_v_ze < __pyx_v_zr) != 0)) {
              __pyx_t_17 = __pyx_v_ze;
            } else {
              __pyx_t_17 = __pyx_v_zr;
            }
            __pyx_v_ze = __pyx_t_17;

            /* "ttcrpy/rgrid.pyx":3047
 *                             ze = grz[iz+1] if grz[iz+1]<zi else zi
 *                             ze = ze if ze<zr else zr
 *                             xe = (ze-b)/m             # <<<<<<<<<<<<<<
 *                             dlx = xe - x
 *                             dlz = ze - z
 */
            __pyx_t_17 = (__pyx_v_ze - __pyx_v_b);
            if (unlikely(__pyx_v_m == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 3047, __pyx_L1_error)
            }
            __pyx_v_xe = (__pyx_t_17 / __pyx_v_m);

            /* "ttcrpy/rgrid.pyx":3048
 *                             ze = ze if ze<zr else zr
 *                             xe = (ze-b)/m
 *                             dlx = xe - x             # <<<<<<<<<<<<<<
 *                             dlz = ze - z
 * 
 */
            __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

            /* "ttcrpy/rgrid.pyx":3049
 *                             xe = (ze-b)/m
 *                             dlx = xe - x
 *                             dlz = ze - z             # <<<<<<<<<<<<<<
 * 
 *                             if not aniso:
 */
            __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

            /* "ttcrpy/rgrid.pyx":3051
 *                             dlz = ze - z
 * 
 *                             if not aniso:             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 */
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_aniso); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 3051, __pyx_L1_error)
            __pyx_t_9 = ((!__pyx_t_15) != 0);
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":3052
 * 
 *                             if not aniso:
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )             # <<<<<<<<<<<<<<
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":3053
 *                             if not aniso:
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3053, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3053, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3054
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 *                             else:
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3054, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3054, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3055
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 *                             else:
 *                                 indices_p.append(iCell)
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":3051
 *                             dlz = ze - z
 * 
 *                             if not aniso:             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 */
              goto __pyx_L38;
            }

            /* "ttcrpy/rgrid.pyx":3057
 *                                 k += 1
 *                             else:
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dlx)
 *                                 k += 1
 */
            /*else*/ {
              __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3057, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3057, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3058
 *                             else:
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dlx)             # <<<<<<<<<<<<<<
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3058, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3058, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3059
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dlx)
 *                                 k += 1             # <<<<<<<<<<<<<<
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":3060
 *                                 data_p.append(dlx)
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dlz)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_iCell + __pyx_v_nCells)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3060, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3060, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3061
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3061, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3061, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3062
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                             x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);
            }
            __pyx_L38:;

            /* "ttcrpy/rgrid.pyx":3064
 *                                 k += 1
 * 
 *                             x = xe             # <<<<<<<<<<<<<<
 *                             z = ze
 *                             if abs(z-grz[iz+1])<small:
 */
            __pyx_v_x = __pyx_v_xe;

            /* "ttcrpy/rgrid.pyx":3065
 * 
 *                             x = xe
 *                             z = ze             # <<<<<<<<<<<<<<
 *                             if abs(z-grz[iz+1])<small:
 *                                 iz += 1
 */
            __pyx_v_z = __pyx_v_ze;

            /* "ttcrpy/rgrid.pyx":3066
 *                             x = xe
 *                             z = ze
 *                             if abs(z-grz[iz+1])<small:             # <<<<<<<<<<<<<<
 *                                 iz += 1
 *                     else:
 */
            __pyx_t_10 = (__pyx_v_iz + 1);
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 3066, __pyx_L1_error)
            }
            __pyx_t_9 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
            if (__pyx_t_9) {

              /* "ttcrpy/rgrid.pyx":3067
 *                             z = ze
 *                             if abs(z-grz[iz+1])<small:
 *                                 iz += 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         while z > zi and z > zr:
 */
              __pyx_v_iz = (__pyx_v_iz + 1);

              /* "ttcrpy/rgrid.pyx":3066
 *                             x = xe
 *                             z = ze
 *                             if abs(z-grz[iz+1])<small:             # <<<<<<<<<<<<<<
 *                                 iz += 1
 *                     else:
 */
            }
          }

          /* "ttcrpy/rgrid.pyx":3041
 *                     zi = m*grx[ix+1] + b
 * 
 *                     if up:             # <<<<<<<<<<<<<<
 *                         while z < zi and z < zr:
 *                             iCell = ix*(n_grz-1) + iz
 */
          goto __pyx_L33;
        }

        /* "ttcrpy/rgrid.pyx":3069
 *                                 iz += 1
 *                     else:
 *                         while z > zi and z > zr:             # <<<<<<<<<<<<<<
 *                             iCell = ix*(n_grz-1) + iz
 * 
 */
        /*else*/ {
          while (1) {
            __pyx_t_1 = PyFloat_FromDouble(__pyx_v_z); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3069, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_18 = PyObject_RichCompare(__pyx_t_1, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3069, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 3069, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (__pyx_t_15) {
            } else {
              __pyx_t_9 = __pyx_t_15;
              goto __pyx_L42_bool_binop_done;
            }
            __pyx_t_15 = ((__pyx_v_z > __pyx_v_zr) != 0);
            __pyx_t_9 = __pyx_t_15;
            __pyx_L42_bool_binop_done:;
            if (!__pyx_t_9) break;

            /* "ttcrpy/rgrid.pyx":3070
 *                     else:
 *                         while z > zi and z > zr:
 *                             iCell = ix*(n_grz-1) + iz             # <<<<<<<<<<<<<<
 * 
 *                             ze = grz[iz] if grz[iz]>zi else zi
 */
            __pyx_v_iCell = ((__pyx_v_ix * (__pyx_v_n_grz - 1)) + __pyx_v_iz);

            /* "ttcrpy/rgrid.pyx":3072
 *                             iCell = ix*(n_grz-1) + iz
 * 
 *                             ze = grz[iz] if grz[iz]>zi else zi             # <<<<<<<<<<<<<<
 *                             ze = ze if ze>zr else zr
 *                             xe = (ze-b)/m
 */
            __pyx_t_10 = __pyx_v_iz;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 3072, __pyx_L1_error)
            }
            __pyx_t_18 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides))); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3072, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_1 = PyObject_RichCompare(__pyx_t_18, __pyx_v_zi, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3072, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3072, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_9) {
              __pyx_t_10 = __pyx_v_iz;
              __pyx_t_8 = -1;
              if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                __PYX_ERR(0, 3072, __pyx_L1_error)
              }
              __pyx_t_17 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides));
            } else {
              __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_v_zi); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 3072, __pyx_L1_error)
              __pyx_t_17 = __pyx_t_16;
            }
            __pyx_v_ze = __pyx_t_17;

            /* "ttcrpy/rgrid.pyx":3073
 * 
 *                             ze = grz[iz] if grz[iz]>zi else zi
 *                             ze = ze if ze>zr else zr             # <<<<<<<<<<<<<<
 *                             xe = (ze-b)/m
 *                             dlx = xe - x
 */
            if (((__pyx_v_ze > __pyx_v_zr) != 0)) {
              __pyx_t_17 = __pyx_v_ze;
            } else {
              __pyx_t_17 = __pyx_v_zr;
            }
            __pyx_v_ze = __pyx_t_17;

            /* "ttcrpy/rgrid.pyx":3074
 *                             ze = grz[iz] if grz[iz]>zi else zi
 *                             ze = ze if ze>zr else zr
 *                             xe = (ze-b)/m             # <<<<<<<<<<<<<<
 *                             dlx = xe - x
 *                             dlz = ze - z
 */
            __pyx_t_17 = (__pyx_v_ze - __pyx_v_b);
            if (unlikely(__pyx_v_m == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 3074, __pyx_L1_error)
            }
            __pyx_v_xe = (__pyx_t_17 / __pyx_v_m);

            /* "ttcrpy/rgrid.pyx":3075
 *                             ze = ze if ze>zr else zr
 *                             xe = (ze-b)/m
 *                             dlx = xe - x             # <<<<<<<<<<<<<<
 *                             dlz = ze - z
 * 
 */
            __pyx_v_dlx = (__pyx_v_xe - __pyx_v_x);

            /* "ttcrpy/rgrid.pyx":3076
 *                             xe = (ze-b)/m
 *                             dlx = xe - x
 *                             dlz = ze - z             # <<<<<<<<<<<<<<
 * 
 *                             if not aniso:
 */
            __pyx_v_dlz = (__pyx_v_ze - __pyx_v_z);

            /* "ttcrpy/rgrid.pyx":3078
 *                             dlz = ze - z
 * 
 *                             if not aniso:             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 */
            __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_aniso); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3078, __pyx_L1_error)
            __pyx_t_15 = ((!__pyx_t_9) != 0);
            if (__pyx_t_15) {

              /* "ttcrpy/rgrid.pyx":3079
 * 
 *                             if not aniso:
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )             # <<<<<<<<<<<<<<
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 */
              __pyx_v_dl = sqrt(((__pyx_v_dlx * __pyx_v_dlx) + (__pyx_v_dlz * __pyx_v_dlz)));

              /* "ttcrpy/rgrid.pyx":3080
 *                             if not aniso:
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dl)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3080, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3080, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3081
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)             # <<<<<<<<<<<<<<
 *                                 k += 1
 *                             else:
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3081, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3082
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dl)
 *                                 k += 1             # <<<<<<<<<<<<<<
 *                             else:
 *                                 indices_p.append(iCell)
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":3078
 *                             dlz = ze - z
 * 
 *                             if not aniso:             # <<<<<<<<<<<<<<
 *                                 dl = sqrt( dlx*dlx + dlz*dlz )
 *                                 indices_p.append(iCell)
 */
              goto __pyx_L44;
            }

            /* "ttcrpy/rgrid.pyx":3084
 *                                 k += 1
 *                             else:
 *                                 indices_p.append(iCell)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dlx)
 *                                 k += 1
 */
            /*else*/ {
              __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_iCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3084, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3084, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3085
 *                             else:
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dlx)             # <<<<<<<<<<<<<<
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3085, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3085, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3086
 *                                 indices_p.append(iCell)
 *                                 data_p.append(dlx)
 *                                 k += 1             # <<<<<<<<<<<<<<
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)
 */
              __pyx_v_k = (__pyx_v_k + 1);

              /* "ttcrpy/rgrid.pyx":3087
 *                                 data_p.append(dlx)
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)             # <<<<<<<<<<<<<<
 *                                 data_p.append(dlz)
 *                                 k += 1
 */
              __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_iCell + __pyx_v_nCells)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3087, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indices_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3087, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3088
 *                                 k += 1
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dlz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3088, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_data_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3088, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "ttcrpy/rgrid.pyx":3089
 *                                 indices_p.append(iCell+nCells)
 *                                 data_p.append(dlz)
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                             x = xe
 */
              __pyx_v_k = (__pyx_v_k + 1);
            }
            __pyx_L44:;

            /* "ttcrpy/rgrid.pyx":3091
 *                                 k += 1
 * 
 *                             x = xe             # <<<<<<<<<<<<<<
 *                             z = ze
 *                             if abs(z-grz[iz])<small:
 */
            __pyx_v_x = __pyx_v_xe;

            /* "ttcrpy/rgrid.pyx":3092
 * 
 *                             x = xe
 *                             z = ze             # <<<<<<<<<<<<<<
 *                             if abs(z-grz[iz])<small:
 *                                 iz -= 1
 */
            __pyx_v_z = __pyx_v_ze;

            /* "ttcrpy/rgrid.pyx":3093
 *                             x = xe
 *                             z = ze
 *                             if abs(z-grz[iz])<small:             # <<<<<<<<<<<<<<
 *                                 iz -= 1
 * 
 */
            __pyx_t_10 = __pyx_v_iz;
            __pyx_t_8 = -1;
            if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grz.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              __PYX_ERR(0, 3093, __pyx_L1_error)
            }
            __pyx_t_15 = ((fabs((__pyx_v_z - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grz.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grz.diminfo[0].strides)))) < __pyx_v_small) != 0);
            if (__pyx_t_15) {

              /* "ttcrpy/rgrid.pyx":3094
 *                             z = ze
 *                             if abs(z-grz[iz])<small:
 *                                 iz -= 1             # <<<<<<<<<<<<<<
 * 
 *                     ix += 1
 */
              __pyx_v_iz = (__pyx_v_iz - 1);

              /* "ttcrpy/rgrid.pyx":3093
 *                             x = xe
 *                             z = ze
 *                             if abs(z-grz[iz])<small:             # <<<<<<<<<<<<<<
 *                                 iz -= 1
 * 
 */
            }
          }
        }
        __pyx_L33:;

        /* "ttcrpy/rgrid.pyx":3096
 *                                 iz -= 1
 * 
 *                     ix += 1             # <<<<<<<<<<<<<<
 *                     x = grx[ix]
 * 
 */
        __pyx_v_ix = (__pyx_v_ix + 1);

        /* "ttcrpy/rgrid.pyx":3097
 * 
 *                     ix += 1
 *                     x = grx[ix]             # <<<<<<<<<<<<<<
 * 
 *         indptr_p.append(k)
 */
        __pyx_t_10 = __pyx_v_ix;
        __pyx_t_8 = -1;
        if (unlikely(__pyx_t_10 >= (size_t)__pyx_pybuffernd_grx.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          __PYX_ERR(0, 3097, __pyx_L1_error)
        }
        __pyx_v_x = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_grx.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_grx.diminfo[0].strides));
      }
    }
    __pyx_L6:;
  }

  /* "ttcrpy/rgrid.pyx":3099
 *                     x = grx[ix]
 * 
 *         indptr_p.append(k)             # <<<<<<<<<<<<<<
 *         if not aniso:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_indptr_p, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":3100
 * 
 *         indptr_p.append(k)
 *         if not aniso:             # <<<<<<<<<<<<<<
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                               shape=(nTx, nCells))
 */
  __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_aniso); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __pyx_t_9 = ((!__pyx_t_15) != 0);
  if (__pyx_t_9) {

    /* "ttcrpy/rgrid.pyx":3101
 *         indptr_p.append(k)
 *         if not aniso:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                               shape=(nTx, nCells))
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_data_p);
    __Pyx_GIVEREF(__pyx_v_data_p);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data_p);
    __Pyx_INCREF(__pyx_v_indices_p);
    __Pyx_GIVEREF(__pyx_v_indices_p);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_indices_p);
    __Pyx_INCREF(__pyx_v_indptr_p);
    __Pyx_GIVEREF(__pyx_v_indptr_p);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_indptr_p);
    __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "ttcrpy/rgrid.pyx":3102
 *         if not aniso:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                               shape=(nTx, nCells))             # <<<<<<<<<<<<<<
 *         else:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_20 = __Pyx_PyInt_FromSize_t(__pyx_v_nTx); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_21 = __Pyx_PyInt_FromSize_t(__pyx_v_nCells); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_21);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_21);
    __pyx_t_20 = 0;
    __pyx_t_21 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_22) < 0) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "ttcrpy/rgrid.pyx":3101
 *         indptr_p.append(k)
 *         if not aniso:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                               shape=(nTx, nCells))
 *         else:
 */
    __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_19, __pyx_t_1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_L = __pyx_t_22;
    __pyx_t_22 = 0;

    /* "ttcrpy/rgrid.pyx":3100
 * 
 *         indptr_p.append(k)
 *         if not aniso:             # <<<<<<<<<<<<<<
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                               shape=(nTx, nCells))
 */
    goto __pyx_L46;
  }

  /* "ttcrpy/rgrid.pyx":3104
 *                               shape=(nTx, nCells))
 *         else:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                               shape=(nTx, 2*nCells))
 *         return L
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_sp); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = PyTuple_New(3); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_v_data_p);
    __Pyx_GIVEREF(__pyx_v_data_p);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_v_data_p);
    __Pyx_INCREF(__pyx_v_indices_p);
    __Pyx_GIVEREF(__pyx_v_indices_p);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_v_indices_p);
    __Pyx_INCREF(__pyx_v_indptr_p);
    __Pyx_GIVEREF(__pyx_v_indptr_p);
    PyTuple_SET_ITEM(__pyx_t_22, 2, __pyx_v_indptr_p);
    __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_22);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_22);
    __pyx_t_22 = 0;

    /* "ttcrpy/rgrid.pyx":3105
 *         else:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                               shape=(nTx, 2*nCells))             # <<<<<<<<<<<<<<
 *         return L
 * 
 */
    __pyx_t_22 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_18 = __Pyx_PyInt_FromSize_t(__pyx_v_nTx); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_21 = __Pyx_PyInt_FromSize_t((2 * __pyx_v_nCells)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_21);
    PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_21);
    __pyx_t_18 = 0;
    __pyx_t_21 = 0;
    if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_shape, __pyx_t_20) < 0) __PYX_ERR(0, 3105, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "ttcrpy/rgrid.pyx":3104
 *                               shape=(nTx, nCells))
 *         else:
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),             # <<<<<<<<<<<<<<
 *                               shape=(nTx, 2*nCells))
 *         return L
 */
    __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_19, __pyx_t_22); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_v_L = __pyx_t_20;
    __pyx_t_20 = 0;
  }
  __pyx_L46:;

  /* "ttcrpy/rgrid.pyx":3106
 *             L = sp.csr_matrix((data_p, indices_p, indptr_p),
 *                               shape=(nTx, 2*nCells))
 *         return L             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_L);
  __pyx_r = __pyx_v_L;
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":2897
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grz.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.rgrid.Grid2d.data_kernel_straight_rays", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Rx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_grz.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_data_p);
  __Pyx_XDECREF(__pyx_v_indices_p);
  __Pyx_XDECREF(__pyx_v_indptr_p);
  __Pyx_XDECREF(__pyx_v_zi);
  __Pyx_XDECREF(__pyx_v_L);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":3109
 * 
 * 
 * def _rebuild3d(x, y, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,
 *      weno, nsnx, nsny, nsnz, n_secondary,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_3_rebuild3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_3_rebuild3d = {"_rebuild3d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_3_rebuild3d, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_3_rebuild3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  PyObject *__pyx_v_z = 0;
  PyObject *__pyx_v_constructor_params = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rebuild3d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_constructor_params,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rebuild3d", 1, 4, 4, 1); __PYX_ERR(0, 3109, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rebuild3d", 1, 4, 4, 2); __PYX_ERR(0, 3109, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_constructor_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rebuild3d", 1, 4, 4, 3); __PYX_ERR(0, 3109, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_rebuild3d") < 0)) __PYX_ERR(0, 3109, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
    __pyx_v_z = values[2];
    __pyx_v_constructor_params = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rebuild3d", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3109, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid._rebuild3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_2_rebuild3d(__pyx_self, __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_constructor_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_2_rebuild3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y, PyObject *__pyx_v_z, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_v_n_threads = NULL;
  PyObject *__pyx_v_cell_slowness = NULL;
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_tt_from_rp = NULL;
  PyObject *__pyx_v_interp_vel = NULL;
  PyObject *__pyx_v_eps = NULL;
  PyObject *__pyx_v_maxit = NULL;
  PyObject *__pyx_v_weno = NULL;
  PyObject *__pyx_v_nsnx = NULL;
  PyObject *__pyx_v_nsny = NULL;
  PyObject *__pyx_v_nsnz = NULL;
  PyObject *__pyx_v_n_secondary = NULL;
  PyObject *__pyx_v_n_tertiary = NULL;
  PyObject *__pyx_v_radius_tertiary = NULL;
  struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *__pyx_v_g = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rebuild3d", 0);

  /* "ttcrpy/rgrid.pyx":3112
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,
 *      weno, nsnx, nsny, nsnz, n_secondary,
 *      n_tertiary, radius_tertiary) = constructor_params             # <<<<<<<<<<<<<<
 *     g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,
 *                interp_vel, eps, maxit, weno, nsnx, nsny, nsnz, n_secondary,
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_constructor_params))) || (PyList_CheckExact(__pyx_v_constructor_params))) {
    PyObject* sequence = __pyx_v_constructor_params;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 14)) {
      if (size > 14) __Pyx_RaiseTooManyValuesError(14);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 3110, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 9); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 10); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 11); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 12); 
      __pyx_t_14 = PyTuple_GET_ITEM(sequence, 13); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 9); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 10); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 11); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 12); 
      __pyx_t_14 = PyList_GET_ITEM(sequence, 13); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_14);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[14] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10,&__pyx_t_11,&__pyx_t_12,&__pyx_t_13,&__pyx_t_14};
      for (i=0; i < 14; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 3110, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[14] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10,&__pyx_t_11,&__pyx_t_12,&__pyx_t_13,&__pyx_t_14};
    __pyx_t_15 = PyObject_GetIter(__pyx_v_constructor_params); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
    for (index=0; index < 14; index++) {
      PyObject* item = __pyx_t_16(__pyx_t_15); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 14) < 0) __PYX_ERR(0, 3110, __pyx_L1_error)
    __pyx_t_16 = NULL;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_16 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 3110, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "ttcrpy/rgrid.pyx":3110
 * 
 * def _rebuild3d(x, y, z, constructor_params):
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,             # <<<<<<<<<<<<<<
 *      weno, nsnx, nsny, nsnz, n_secondary,
 *      n_tertiary, radius_tertiary) = constructor_params
 */
  __pyx_v_n_threads = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_cell_slowness = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_method = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_tt_from_rp = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_interp_vel = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_eps = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_maxit = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_weno = __pyx_t_8;
  __pyx_t_8 = 0;
  __pyx_v_nsnx = __pyx_t_9;
  __pyx_t_9 = 0;
  __pyx_v_nsny = __pyx_t_10;
  __pyx_t_10 = 0;
  __pyx_v_nsnz = __pyx_t_11;
  __pyx_t_11 = 0;
  __pyx_v_n_secondary = __pyx_t_12;
  __pyx_t_12 = 0;
  __pyx_v_n_tertiary = __pyx_t_13;
  __pyx_t_13 = 0;
  __pyx_v_radius_tertiary = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "ttcrpy/rgrid.pyx":3113
 *      weno, nsnx, nsny, nsnz, n_secondary,
 *      n_tertiary, radius_tertiary) = constructor_params
 *     g = Grid3d(x, y, z, n_threads, cell_slowness, method, tt_from_rp,             # <<<<<<<<<<<<<<
 *                interp_vel, eps, maxit, weno, nsnx, nsny, nsnz, n_secondary,
 *                n_tertiary, radius_tertiary)
 */
  __pyx_t_14 = PyTuple_New(17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_x);
  __Pyx_INCREF(__pyx_v_y);
  __Pyx_GIVEREF(__pyx_v_y);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_y);
  __Pyx_INCREF(__pyx_v_z);
  __Pyx_GIVEREF(__pyx_v_z);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_v_z);
  __Pyx_INCREF(__pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_v_n_threads);
  PyTuple_SET_ITEM(__pyx_t_14, 3, __pyx_v_n_threads);
  __Pyx_INCREF(__pyx_v_cell_slowness);
  __Pyx_GIVEREF(__pyx_v_cell_slowness);
  PyTuple_SET_ITEM(__pyx_t_14, 4, __pyx_v_cell_slowness);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_14, 5, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_tt_from_rp);
  __Pyx_GIVEREF(__pyx_v_tt_from_rp);
  PyTuple_SET_ITEM(__pyx_t_14, 6, __pyx_v_tt_from_rp);
  __Pyx_INCREF(__pyx_v_interp_vel);
  __Pyx_GIVEREF(__pyx_v_interp_vel);
  PyTuple_SET_ITEM(__pyx_t_14, 7, __pyx_v_interp_vel);
  __Pyx_INCREF(__pyx_v_eps);
  __Pyx_GIVEREF(__pyx_v_eps);
  PyTuple_SET_ITEM(__pyx_t_14, 8, __pyx_v_eps);
  __Pyx_INCREF(__pyx_v_maxit);
  __Pyx_GIVEREF(__pyx_v_maxit);
  PyTuple_SET_ITEM(__pyx_t_14, 9, __pyx_v_maxit);
  __Pyx_INCREF(__pyx_v_weno);
  __Pyx_GIVEREF(__pyx_v_weno);
  PyTuple_SET_ITEM(__pyx_t_14, 10, __pyx_v_weno);
  __Pyx_INCREF(__pyx_v_nsnx);
  __Pyx_GIVEREF(__pyx_v_nsnx);
  PyTuple_SET_ITEM(__pyx_t_14, 11, __pyx_v_nsnx);
  __Pyx_INCREF(__pyx_v_nsny);
  __Pyx_GIVEREF(__pyx_v_nsny);
  PyTuple_SET_ITEM(__pyx_t_14, 12, __pyx_v_nsny);
  __Pyx_INCREF(__pyx_v_nsnz);
  __Pyx_GIVEREF(__pyx_v_nsnz);
  PyTuple_SET_ITEM(__pyx_t_14, 13, __pyx_v_nsnz);
  __Pyx_INCREF(__pyx_v_n_secondary);
  __Pyx_GIVEREF(__pyx_v_n_secondary);
  PyTuple_SET_ITEM(__pyx_t_14, 14, __pyx_v_n_secondary);
  __Pyx_INCREF(__pyx_v_n_tertiary);
  __Pyx_GIVEREF(__pyx_v_n_tertiary);
  PyTuple_SET_ITEM(__pyx_t_14, 15, __pyx_v_n_tertiary);
  __Pyx_INCREF(__pyx_v_radius_tertiary);
  __Pyx_GIVEREF(__pyx_v_radius_tertiary);
  PyTuple_SET_ITEM(__pyx_t_14, 16, __pyx_v_radius_tertiary);
  __pyx_t_13 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d), __pyx_t_14, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_g = ((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)__pyx_t_13);
  __pyx_t_13 = 0;

  /* "ttcrpy/rgrid.pyx":3116
 *                interp_vel, eps, maxit, weno, nsnx, nsny, nsnz, n_secondary,
 *                n_tertiary, radius_tertiary)
 *     return g             # <<<<<<<<<<<<<<
 * 
 * def _rebuild2d(x, z, constructor_params):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_g));
  __pyx_r = ((PyObject *)__pyx_v_g);
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":3109
 * 
 * 
 * def _rebuild3d(x, y, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,
 *      weno, nsnx, nsny, nsnz, n_secondary,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("ttcrpy.rgrid._rebuild3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_threads);
  __Pyx_XDECREF(__pyx_v_cell_slowness);
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_tt_from_rp);
  __Pyx_XDECREF(__pyx_v_interp_vel);
  __Pyx_XDECREF(__pyx_v_eps);
  __Pyx_XDECREF(__pyx_v_maxit);
  __Pyx_XDECREF(__pyx_v_weno);
  __Pyx_XDECREF(__pyx_v_nsnx);
  __Pyx_XDECREF(__pyx_v_nsny);
  __Pyx_XDECREF(__pyx_v_nsnz);
  __Pyx_XDECREF(__pyx_v_n_secondary);
  __Pyx_XDECREF(__pyx_v_n_tertiary);
  __Pyx_XDECREF(__pyx_v_radius_tertiary);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/rgrid.pyx":3118
 *     return g
 * 
 * def _rebuild2d(x, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5rgrid_5_rebuild2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6ttcrpy_5rgrid_5_rebuild2d = {"_rebuild2d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_5_rebuild2d, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6ttcrpy_5rgrid_5_rebuild2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_z = 0;
  PyObject *__pyx_v_constructor_params = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rebuild2d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_z,&__pyx_n_s_constructor_params,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rebuild2d", 1, 3, 3, 1); __PYX_ERR(0, 3118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_constructor_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rebuild2d", 1, 3, 3, 2); __PYX_ERR(0, 3118, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_rebuild2d") < 0)) __PYX_ERR(0, 3118, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_x = values[0];
    __pyx_v_z = values[1];
    __pyx_v_constructor_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rebuild2d", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3118, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.rgrid._rebuild2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5rgrid_4_rebuild2d(__pyx_self, __pyx_v_x, __pyx_v_z, __pyx_v_constructor_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5rgrid_4_rebuild2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_z, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_v_n_threads = NULL;
  PyObject *__pyx_v_cell_slowness = NULL;
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_aniso = NULL;
  PyObject *__pyx_v_eps = NULL;
  PyObject *__pyx_v_maxit = NULL;
  PyObject *__pyx_v_weno = NULL;
  PyObject *__pyx_v_rotated_template = NULL;
  PyObject *__pyx_v_nsnx = NULL;
  PyObject *__pyx_v_nsnz = NULL;
  struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *__pyx_v_g = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rebuild2d", 0);

  /* "ttcrpy/rgrid.pyx":3120
 * def _rebuild2d(x, z, constructor_params):
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params             # <<<<<<<<<<<<<<
 *     g = Grid2d(x, z, n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *                rotated_template, nsnx, nsnz)
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_constructor_params))) || (PyList_CheckExact(__pyx_v_constructor_params))) {
    PyObject* sequence = __pyx_v_constructor_params;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 10)) {
      if (size > 10) __Pyx_RaiseTooManyValuesError(10);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 3119, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 9); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 9); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_10);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[10] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10};
      for (i=0; i < 10; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 3119, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[10] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10};
    __pyx_t_11 = PyObject_GetIter(__pyx_v_constructor_params); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
    for (index=0; index < 10; index++) {
      PyObject* item = __pyx_t_12(__pyx_t_11); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 10) < 0) __PYX_ERR(0, 3119, __pyx_L1_error)
    __pyx_t_12 = NULL;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 3119, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "ttcrpy/rgrid.pyx":3119
 * 
 * def _rebuild2d(x, z, constructor_params):
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,             # <<<<<<<<<<<<<<
 *      rotated_template, nsnx, nsnz) = constructor_params
 *     g = Grid2d(x, z, n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 */
  __pyx_v_n_threads = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_cell_slowness = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_method = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_aniso = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_eps = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_maxit = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_weno = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_rotated_template = __pyx_t_8;
  __pyx_t_8 = 0;
  __pyx_v_nsnx = __pyx_t_9;
  __pyx_t_9 = 0;
  __pyx_v_nsnz = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "ttcrpy/rgrid.pyx":3121
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params
 *     g = Grid2d(x, z, n_threads, cell_slowness, method, aniso, eps, maxit, weno,             # <<<<<<<<<<<<<<
 *                rotated_template, nsnx, nsnz)
 *     return g
 */
  __pyx_t_10 = PyTuple_New(12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_x);
  __Pyx_INCREF(__pyx_v_z);
  __Pyx_GIVEREF(__pyx_v_z);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_z);
  __Pyx_INCREF(__pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_v_n_threads);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_n_threads);
  __Pyx_INCREF(__pyx_v_cell_slowness);
  __Pyx_GIVEREF(__pyx_v_cell_slowness);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_v_cell_slowness);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_aniso);
  __Pyx_GIVEREF(__pyx_v_aniso);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_v_aniso);
  __Pyx_INCREF(__pyx_v_eps);
  __Pyx_GIVEREF(__pyx_v_eps);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_v_eps);
  __Pyx_INCREF(__pyx_v_maxit);
  __Pyx_GIVEREF(__pyx_v_maxit);
  PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_v_maxit);
  __Pyx_INCREF(__pyx_v_weno);
  __Pyx_GIVEREF(__pyx_v_weno);
  PyTuple_SET_ITEM(__pyx_t_10, 8, __pyx_v_weno);
  __Pyx_INCREF(__pyx_v_rotated_template);
  __Pyx_GIVEREF(__pyx_v_rotated_template);
  PyTuple_SET_ITEM(__pyx_t_10, 9, __pyx_v_rotated_template);
  __Pyx_INCREF(__pyx_v_nsnx);
  __Pyx_GIVEREF(__pyx_v_nsnx);
  PyTuple_SET_ITEM(__pyx_t_10, 10, __pyx_v_nsnx);
  __Pyx_INCREF(__pyx_v_nsnz);
  __Pyx_GIVEREF(__pyx_v_nsnz);
  PyTuple_SET_ITEM(__pyx_t_10, 11, __pyx_v_nsnz);
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid2d), __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_g = ((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "ttcrpy/rgrid.pyx":3123
 *     g = Grid2d(x, z, n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *                rotated_template, nsnx, nsnz)
 *     return g             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_g));
  __pyx_r = ((PyObject *)__pyx_v_g);
  goto __pyx_L0;

  /* "ttcrpy/rgrid.pyx":3118
 *     return g
 * 
 * def _rebuild2d(x, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ttcrpy.rgrid._rebuild2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_threads);
  __Pyx_XDECREF(__pyx_v_cell_slowness);
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_aniso);
  __Pyx_XDECREF(__pyx_v_eps);
  __Pyx_XDECREF(__pyx_v_maxit);
  __Pyx_XDECREF(__pyx_v_weno);
  __Pyx_XDECREF(__pyx_v_rotated_template);
  __Pyx_XDECREF(__pyx_v_nsnx);
  __Pyx_XDECREF(__pyx_v_nsnz);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":869
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":870
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":873
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":875
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":876
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":882
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 882, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":883
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 883, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 884, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 884, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":888
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 888, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":889
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 889, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 890, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 890, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":894
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 894, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":895
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 895, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":896
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 896, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 896, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6ttcrpy_5rgrid_Grid3d __pyx_vtable_6ttcrpy_5rgrid_Grid3d;

static PyObject *__pyx_tp_new_6ttcrpy_5rgrid_Grid3d(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)o);
  p->__pyx_vtab = __pyx_vtabptr_6ttcrpy_5rgrid_Grid3d;
  new((void*)&(p->_x)) std::vector<double> ();
  new((void*)&(p->_y)) std::vector<double> ();
  new((void*)&(p->_z)) std::vector<double> ();
  if (unlikely(__pyx_pw_6ttcrpy_5rgrid_6Grid3d_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6ttcrpy_5rgrid_Grid3d(PyObject *o) {
  struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *p = (struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6ttcrpy_5rgrid_6Grid3d_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->_x);
  __Pyx_call_destructor(p->_y);
  __Pyx_call_destructor(p->_z);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_x(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_1x_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_y(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_1y_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_z(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_1z_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dx_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dy_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dz(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_2dz_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_5shape_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_n_threads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_9n_threads_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid3d_nparams(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid3d_7nparams_1__get__(o);
}

static PyMethodDef __pyx_methods_6ttcrpy_5rgrid_Grid3d[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_5__reduce__, METH_NOARGS, 0},
  {"set_use_thread_pool", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_7set_use_thread_pool, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_6set_use_thread_pool},
  {"get_number_of_nodes", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_9get_number_of_nodes, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_8get_number_of_nodes},
  {"get_number_of_cells", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_11get_number_of_cells, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_10get_number_of_cells},
  {"get_grid_traveltimes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_13get_grid_traveltimes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_12get_grid_traveltimes},
  {"ind", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_15ind, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_14ind},
  {"indc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_17indc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_16indc},
  {"is_outside", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_19is_outside, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_18is_outside},
  {"get_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_21get_slowness, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_20get_slowness},
  {"set_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_23set_slowness, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_22set_slowness},
  {"set_velocity", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_25set_velocity, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_24set_velocity},
  {"compute_D", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_27compute_D, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_26compute_D},
  {"compute_K", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_29compute_K, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_28compute_K},
  {"get_s0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_31get_s0, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_30get_s0},
  {"raytrace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_33raytrace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_32raytrace},
  {"to_vtk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_35to_vtk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_34to_vtk},
  {"_save_raypaths", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_37_save_raypaths, METH_VARARGS|METH_KEYWORDS, 0},
  {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_39builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_38builder},
  {"data_kernel_straight_rays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid3d_40data_kernel_straight_rays},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6ttcrpy_5rgrid_Grid3d[] = {
  {(char *)"x", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_x, 0, (char *)"np.ndarray: node coordinates along x", 0},
  {(char *)"y", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_y, 0, (char *)"np.ndarray: node coordinates along y", 0},
  {(char *)"z", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_z, 0, (char *)"np.ndarray: node coordinates along z", 0},
  {(char *)"dx", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dx, 0, (char *)"float: node separation along x", 0},
  {(char *)"dy", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dy, 0, (char *)"float: node separation along y", 0},
  {(char *)"dz", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_dz, 0, (char *)"float: node separation along z", 0},
  {(char *)"shape", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_shape, 0, (char *)":obj:`list` of :obj:`int`: number of parameters along each dimension", 0},
  {(char *)"n_threads", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_n_threads, 0, (char *)"int: number of threads for raytracing", 0},
  {(char *)"nparams", __pyx_getprop_6ttcrpy_5rgrid_6Grid3d_nparams, 0, (char *)"int: total number of parameters for grid", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ttcrpy_5rgrid_Grid3d = {
  PyVarObject_HEAD_INIT(0, 0)
  "ttcrpy.rgrid.Grid3d", /*tp_name*/
  sizeof(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ttcrpy_5rgrid_Grid3d, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    class to perform raytracing with 3D rectilinear grids\n\n    Attributes\n    ----------\n    x: np.ndarray\n        node coordinates along x\n    y: np.ndarray\n        node coordinates along y\n    z: np.ndarray\n        node coordinates along z\n    dx: float\n        node separation along x\n    dy: float\n        node separation along y\n    dz: float\n        node separation along z\n    shape: (int, int, int)\n        number of parameters along each dimension\n    nparams: int\n        total number of parameters for grid\n    n_threads: int\n        number of threads for raytracing\n\n    Constructor:\n\n    Grid3d(x, y, z, n_threads=1, cell_slowness=1, method='FSM', tt_from_rp=1, interp_vel=0, eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5, n_secondary=2, n_tertiary=2, radius_tertiary=1.0) -> Grid3d\n\n        Parameters\n        ----------\n        x : np.ndarray\n            node coordinates along x\n        y : np.ndarray\n            node coordinates along y\n        z : np.ndarray\n            node coordinates along z\n        n_threads : int\n            number of threads for raytracing (default is 1)\n        cell_slowness : bool\n            slowness defined for cells (True) or nodes (False) (default is 1)\n        method : string\n            raytracing method (default is FSM)\n                - 'FSM' : fast marching method\n                - 'SPM' : shortest path method\n                - 'DSPM' : dynamic shortest path\n        tt_from_rp : bool\n            compute traveltimes from raypaths (SPM od DSPM) (default is 1)\n        interp_vel : bool\n            interpolate velocity instead of slowness at nodes (for\n            cell_slowness == False or FSM) (defauls is False)\n        eps : double\n            convergence criterion (FSM) (default is 1e-15)\n        maxit : int\n            max number of sweeping iterations (FSM) (default is 20)\n        weno : bool\n            use 3rd order weighted essentially non-oscillatory operator (""FSM)\n            (default is True)\n        nsnx : int\n            number of secondary nodes in x (SPM) (default is 5)\n        nsny : int\n            number of secondary nodes in y (SPM) (default is 5)\n        nsnz : int\n            number of secondary nodes in z (SPM) (default is 5)\n        n_secondary : int\n            number of secondary nodes (DSPM) (default is 2)\n        n_tertiary : int\n            number of tertiary nodes (DSPM) (default is 2)\n        radius_tertiary : double\n            radius of sphere around source that includes tertiary nodes (DSPM)\n            (default is 1)\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ttcrpy_5rgrid_Grid3d, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6ttcrpy_5rgrid_Grid3d, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ttcrpy_5rgrid_Grid3d, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6ttcrpy_5rgrid_Grid2d(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)o);
  new((void*)&(p->_x)) std::vector<double> ();
  new((void*)&(p->_z)) std::vector<double> ();
  if (unlikely(__pyx_pw_6ttcrpy_5rgrid_6Grid2d_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6ttcrpy_5rgrid_Grid2d(PyObject *o) {
  struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *p = (struct __pyx_obj_6ttcrpy_5rgrid_Grid2d *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6ttcrpy_5rgrid_6Grid2d_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->_x);
  __Pyx_call_destructor(p->_z);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_x(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_1x_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_z(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_1z_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_dx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dx_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_dz(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_2dz_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_5shape_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_n_threads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_9n_threads_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5rgrid_6Grid2d_nparams(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5rgrid_6Grid2d_7nparams_1__get__(o);
}

static PyMethodDef __pyx_methods_6ttcrpy_5rgrid_Grid2d[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_5__reduce__, METH_NOARGS, 0},
  {"get_number_of_nodes", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_7get_number_of_nodes, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_6get_number_of_nodes},
  {"get_number_of_cells", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_9get_number_of_cells, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_8get_number_of_cells},
  {"get_grid_traveltimes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_11get_grid_traveltimes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_10get_grid_traveltimes},
  {"is_outside", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_13is_outside, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_12is_outside},
  {"get_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_15get_slowness, METH_NOARGS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_14get_slowness},
  {"set_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_17set_slowness, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_16set_slowness},
  {"set_velocity", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_19set_velocity, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_18set_velocity},
  {"set_xi", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_21set_xi, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_20set_xi},
  {"set_tilt_angle", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_23set_tilt_angle, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_22set_tilt_angle},
  {"set_Vp0", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_25set_Vp0, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_24set_Vp0},
  {"set_Vs0", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_27set_Vs0, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_26set_Vs0},
  {"set_delta", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_29set_delta, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_28set_delta},
  {"set_epsilon", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_31set_epsilon, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_30set_epsilon},
  {"set_gamma", (PyCFunction)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_33set_gamma, METH_O, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_32set_gamma},
  {"compute_K", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_35compute_K, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_34compute_K},
  {"raytrace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_37raytrace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_36raytrace},
  {"to_vtk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_39to_vtk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_38to_vtk},
  {"_save_raypaths", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_41_save_raypaths, METH_VARARGS|METH_KEYWORDS, 0},
  {"data_kernel_straight_rays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5rgrid_6Grid2d_42data_kernel_straight_rays},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6ttcrpy_5rgrid_Grid2d[] = {
  {(char *)"x", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_x, 0, (char *)"np.ndarray: node coordinates along x", 0},
  {(char *)"z", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_z, 0, (char *)"np.ndarray: node coordinates along z", 0},
  {(char *)"dx", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_dx, 0, (char *)"float: node separation along x", 0},
  {(char *)"dz", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_dz, 0, (char *)"float: node separation along x", 0},
  {(char *)"shape", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_shape, 0, (char *)":obj:`list` of :obj:`int`: number of parameters along each dimension", 0},
  {(char *)"n_threads", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_n_threads, 0, (char *)"int: number of threads for raytracing", 0},
  {(char *)"nparams", __pyx_getprop_6ttcrpy_5rgrid_6Grid2d_nparams, 0, (char *)"int: total number of parameters for grid", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ttcrpy_5rgrid_Grid2d = {
  PyVarObject_HEAD_INIT(0, 0)
  "ttcrpy.rgrid.Grid2d", /*tp_name*/
  sizeof(struct __pyx_obj_6ttcrpy_5rgrid_Grid2d), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ttcrpy_5rgrid_Grid2d, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    class to perform raytracing with 2D rectilinear grids\n\n    Attributes\n    ----------\n    x: np.ndarray\n        node coordinates along x\n    z: np.ndarray\n        node coordinates along z\n    dx: float\n        node separation along x\n    dz: float\n        node separation along z\n    shape: (int, int)\n        number of parameters along each dimension\n    nparams: int\n        total number of parameters for grid\n    n_threads: int\n        number of threads for raytracing\n\n    Constructor:\n\n    Grid2d(x, z, n_threads=1, cell_slowness=1, method='SPM', aniso='iso', eps=1.e-15, maxit=20, weno=1, rotated_template=0, nsnx=10, nsnz=10) -> Grid2d\n\n    Parameters\n    ----------\n    x : np.ndarray\n        node coordinates along x\n    z : np.ndarray\n        node coordinates along z\n    n_threads : int\n        number of threads for raytracing (default is 1)\n    cell_slowness : bool\n        slowness defined for cells (True) or nodes (False) (default is 1)\n    method : string\n        raytracing method (default is SPM)\n            - 'FSM' : fast marching method\n            - 'SPM' : shortest path method\n    aniso : string\n        type of anisotropy (implemented only for the SPM method)\n            - 'iso' : isotropic medium\n            - 'elliptical' : elliptical anisotropy\n            - 'tilted_elliptical' : tilted elliptical anisotropy\n            - 'vti_psv' : vertical transverse isotropy, P and SV waves\n            - 'vti_sh' : vertical transverse isotropy, SH waves\n    eps : double\n        convergence criterion (FSM) (default is 1e-15)\n    maxit : int\n        max number of sweeping iterations (FSM) (default is 20)\n    weno : bool\n        use 3rd order weighted essentially non-oscillatory operator (FSM)\n        (default is True)\n    rotated_template : bool\n        use rotated templates (FSM)\n    nsnx : int\n        number of secondary nodes in x (SPM) (default is 10)\n    nsnz : int\n        number of secondary nodes in z ""(SPM) (default is 10)\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ttcrpy_5rgrid_Grid2d, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6ttcrpy_5rgrid_Grid2d, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ttcrpy_5rgrid_Grid2d, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_rgrid(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_rgrid},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "rgrid",
    __pyx_k_Raytracing_on_rectilinear_grids, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AddArray, __pyx_k_AddArray, sizeof(__pyx_k_AddArray), 0, 0, 1, 1},
  {&__pyx_kp_u_Anisotropic_raytracing_implement, __pyx_k_Anisotropic_raytracing_implement, sizeof(__pyx_k_Anisotropic_raytracing_implement), 0, 1, 0, 0},
  {&__pyx_kp_u_Anisotropy_model_not_implemented, __pyx_k_Anisotropy_model_not_implemented, sizeof(__pyx_k_Anisotropy_model_not_implemented), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_aggregate_source_with_DSP, __pyx_k_Cannot_aggregate_source_with_DSP, sizeof(__pyx_k_Cannot_aggregate_source_with_DSP), 0, 1, 0, 0},
  {&__pyx_n_u_DSPM, __pyx_k_DSPM, sizeof(__pyx_k_DSPM), 0, 1, 0, 1},
  {&__pyx_n_u_F, __pyx_k_F, sizeof(__pyx_k_F), 0, 1, 0, 1},
  {&__pyx_n_u_FSM, __pyx_k_FSM, sizeof(__pyx_k_FSM), 0, 1, 0, 1},
  {&__pyx_kp_u_FSM_Grid_cells_must_be_cubic, __pyx_k_FSM_Grid_cells_must_be_cubic, sizeof(__pyx_k_FSM_Grid_cells_must_be_cubic), 0, 1, 0, 0},
  {&__pyx_kp_u_Field_0_s_has_incorrect_ndim, __pyx_k_Field_0_s_has_incorrect_ndim, sizeof(__pyx_k_Field_0_s_has_incorrect_ndim), 0, 1, 0, 0},
  {&__pyx_kp_u_Field_0_s_has_incorrect_ndim_1_d, __pyx_k_Field_0_s_has_incorrect_ndim_1_d, sizeof(__pyx_k_Field_0_s_has_incorrect_ndim_1_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Field_0_s_has_incorrect_shape, __pyx_k_Field_0_s_has_incorrect_shape, sizeof(__pyx_k_Field_0_s_has_incorrect_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_k_Field_0_s_has_incorrect_size, sizeof(__pyx_k_Field_0_s_has_incorrect_size), 0, 1, 0, 0},
  {&__pyx_kp_u_File_should_contain_slowness_or, __pyx_k_File_should_contain_slowness_or, sizeof(__pyx_k_File_should_contain_slowness_or), 0, 1, 0, 0},
  {&__pyx_n_s_GetArray, __pyx_k_GetArray, sizeof(__pyx_k_GetArray), 0, 0, 1, 1},
  {&__pyx_n_s_GetCellData, __pyx_k_GetCellData, sizeof(__pyx_k_GetCellData), 0, 0, 1, 1},
  {&__pyx_n_s_GetOutput, __pyx_k_GetOutput, sizeof(__pyx_k_GetOutput), 0, 0, 1, 1},
  {&__pyx_n_s_GetPointData, __pyx_k_GetPointData, sizeof(__pyx_k_GetPointData), 0, 0, 1, 1},
  {&__pyx_n_s_GetPointIds, __pyx_k_GetPointIds, sizeof(__pyx_k_GetPointIds), 0, 0, 1, 1},
  {&__pyx_n_s_GetXCoordinates, __pyx_k_GetXCoordinates, sizeof(__pyx_k_GetXCoordinates), 0, 0, 1, 1},
  {&__pyx_n_s_GetYCoordinates, __pyx_k_GetYCoordinates, sizeof(__pyx_k_GetYCoordinates), 0, 0, 1, 1},
  {&__pyx_n_s_GetZCoordinates, __pyx_k_GetZCoordinates, sizeof(__pyx_k_GetZCoordinates), 0, 0, 1, 1},
  {&__pyx_n_s_Grid2d, __pyx_k_Grid2d, sizeof(__pyx_k_Grid2d), 0, 0, 1, 1},
  {&__pyx_n_s_Grid3d, __pyx_k_Grid3d, sizeof(__pyx_k_Grid3d), 0, 0, 1, 1},
  {&__pyx_n_s_HasArray, __pyx_k_HasArray, sizeof(__pyx_k_HasArray), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_InsertNextCell, __pyx_k_InsertNextCell, sizeof(__pyx_k_InsertNextCell), 0, 0, 1, 1},
  {&__pyx_n_s_InsertPoint, __pyx_k_InsertPoint, sizeof(__pyx_k_InsertPoint), 0, 0, 1, 1},
  {&__pyx_n_s_L, __pyx_k_L, sizeof(__pyx_k_L), 0, 0, 1, 1},
  {&__pyx_kp_u_Method_0_s_undefined, __pyx_k_Method_0_s_undefined, sizeof(__pyx_k_Method_0_s_undefined), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_kp_u_P_wave_velocity, __pyx_k_P_wave_velocity, sizeof(__pyx_k_P_wave_velocity), 0, 1, 0, 0},
  {&__pyx_kp_u_Receiver_outside_grid, __pyx_k_Receiver_outside_grid, sizeof(__pyx_k_Receiver_outside_grid), 0, 1, 0, 0},
  {&__pyx_n_s_Rx, __pyx_k_Rx, sizeof(__pyx_k_Rx), 0, 0, 1, 1},
  {&__pyx_n_u_SPM, __pyx_k_SPM, sizeof(__pyx_k_SPM), 0, 1, 0, 1},
  {&__pyx_n_s_SetDataModeToBinary, __pyx_k_SetDataModeToBinary, sizeof(__pyx_k_SetDataModeToBinary), 0, 0, 1, 1},
  {&__pyx_n_s_SetDimensions, __pyx_k_SetDimensions, sizeof(__pyx_k_SetDimensions), 0, 0, 1, 1},
  {&__pyx_n_s_SetFileName, __pyx_k_SetFileName, sizeof(__pyx_k_SetFileName), 0, 0, 1, 1},
  {&__pyx_n_s_SetId, __pyx_k_SetId, sizeof(__pyx_k_SetId), 0, 0, 1, 1},
  {&__pyx_n_s_SetInputData, __pyx_k_SetInputData, sizeof(__pyx_k_SetInputData), 0, 0, 1, 1},
  {&__pyx_n_s_SetLines, __pyx_k_SetLines, sizeof(__pyx_k_SetLines), 0, 0, 1, 1},
  {&__pyx_n_s_SetName, __pyx_k_SetName, sizeof(__pyx_k_SetName), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfComponents, __pyx_k_SetNumberOfComponents, sizeof(__pyx_k_SetNumberOfComponents), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfIds, __pyx_k_SetNumberOfIds, sizeof(__pyx_k_SetNumberOfIds), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfPoints, __pyx_k_SetNumberOfPoints, sizeof(__pyx_k_SetNumberOfPoints), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfTuples, __pyx_k_SetNumberOfTuples, sizeof(__pyx_k_SetNumberOfTuples), 0, 0, 1, 1},
  {&__pyx_n_s_SetPoints, __pyx_k_SetPoints, sizeof(__pyx_k_SetPoints), 0, 0, 1, 1},
  {&__pyx_n_s_SetTuple1, __pyx_k_SetTuple1, sizeof(__pyx_k_SetTuple1), 0, 0, 1, 1},
  {&__pyx_n_s_SetXCoordinates, __pyx_k_SetXCoordinates, sizeof(__pyx_k_SetXCoordinates), 0, 0, 1, 1},
  {&__pyx_n_s_SetYCoordinates, __pyx_k_SetYCoordinates, sizeof(__pyx_k_SetYCoordinates), 0, 0, 1, 1},
  {&__pyx_n_s_SetZCoordinates, __pyx_k_SetZCoordinates, sizeof(__pyx_k_SetZCoordinates), 0, 0, 1, 1},
  {&__pyx_n_u_Slowness, __pyx_k_Slowness, sizeof(__pyx_k_Slowness), 0, 1, 0, 1},
  {&__pyx_kp_u_Slowness_has_wrong_shape, __pyx_k_Slowness_has_wrong_shape, sizeof(__pyx_k_Slowness_has_wrong_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_Slowness_must_be_1D_or_3D_ndarra, __pyx_k_Slowness_must_be_1D_or_3D_ndarra, sizeof(__pyx_k_Slowness_must_be_1D_or_3D_ndarra), 0, 1, 0, 0},
  {&__pyx_kp_u_Slowness_vector_has_wrong_size, __pyx_k_Slowness_vector_has_wrong_size, sizeof(__pyx_k_Slowness_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_kp_u_Source_point_outside_grid, __pyx_k_Source_point_outside_grid, sizeof(__pyx_k_Source_point_outside_grid), 0, 1, 0, 0},
  {&__pyx_n_s_T, __pyx_k_T, sizeof(__pyx_k_T), 0, 0, 1, 1},
  {&__pyx_kp_u_Thread_number_is_larger_than_num, __pyx_k_Thread_number_is_larger_than_num, sizeof(__pyx_k_Thread_number_is_larger_than_num), 0, 1, 0, 0},
  {&__pyx_n_s_Tx, __pyx_k_Tx, sizeof(__pyx_k_Tx), 0, 0, 1, 1},
  {&__pyx_n_s_Update, __pyx_k_Update, sizeof(__pyx_k_Update), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_u_Velocity, __pyx_k_Velocity, sizeof(__pyx_k_Velocity), 0, 1, 0, 1},
  {&__pyx_kp_u_Velocity_data_point_outside_grid, __pyx_k_Velocity_data_point_outside_grid, sizeof(__pyx_k_Velocity_data_point_outside_grid), 0, 1, 0, 0},
  {&__pyx_n_s_Vp0, __pyx_k_Vp0, sizeof(__pyx_k_Vp0), 0, 0, 1, 1},
  {&__pyx_n_s_Vs0, __pyx_k_Vs0, sizeof(__pyx_k_Vs0), 0, 0, 1, 1},
  {&__pyx_n_u__34, __pyx_k__34, sizeof(__pyx_k__34), 0, 1, 0, 1},
  {&__pyx_n_s__61, __pyx_k__61, sizeof(__pyx_k__61), 0, 0, 1, 1},
  {&__pyx_n_s_abs, __pyx_k_abs, sizeof(__pyx_k_abs), 0, 0, 1, 1},
  {&__pyx_n_s_aggregate_src, __pyx_k_aggregate_src, sizeof(__pyx_k_aggregate_src), 0, 0, 1, 1},
  {&__pyx_n_s_aniso, __pyx_k_aniso, sizeof(__pyx_k_aniso), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_b_y, __pyx_k_b_y, sizeof(__pyx_k_b_y), 0, 0, 1, 1},
  {&__pyx_n_s_b_z, __pyx_k_b_z, sizeof(__pyx_k_b_z), 0, 0, 1, 1},
  {&__pyx_n_s_builder, __pyx_k_builder, sizeof(__pyx_k_builder), 0, 0, 1, 1},
  {&__pyx_n_s_cell_slowness, __pyx_k_cell_slowness, sizeof(__pyx_k_cell_slowness), 0, 0, 1, 1},
  {&__pyx_n_s_centers, __pyx_k_centers, sizeof(__pyx_k_centers), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_compute_L, __pyx_k_compute_L, sizeof(__pyx_k_compute_L), 0, 0, 1, 1},
  {&__pyx_kp_u_compute_L_and_compute_M_are_mutu, __pyx_k_compute_L_and_compute_M_are_mutu, sizeof(__pyx_k_compute_L_and_compute_M_are_mutu), 0, 1, 0, 0},
  {&__pyx_kp_u_compute_L_defined_only_for_grids, __pyx_k_compute_L_defined_only_for_grids, sizeof(__pyx_k_compute_L_defined_only_for_grids), 0, 1, 0, 0},
  {&__pyx_kp_u_compute_L_not_implemented_for_VT, __pyx_k_compute_L_not_implemented_for_VT, sizeof(__pyx_k_compute_L_not_implemented_for_VT), 0, 1, 0, 0},
  {&__pyx_n_s_compute_M, __pyx_k_compute_M, sizeof(__pyx_k_compute_M), 0, 0, 1, 1},
  {&__pyx_kp_u_compute_M_not_defined_for_grids, __pyx_k_compute_M_not_defined_for_grids, sizeof(__pyx_k_compute_M_not_defined_for_grids), 0, 1, 0, 0},
  {&__pyx_n_s_constructor_params, __pyx_k_constructor_params, sizeof(__pyx_k_constructor_params), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_csr_matrix, __pyx_k_csr_matrix, sizeof(__pyx_k_csr_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_kernel_straight_rays, __pyx_k_data_kernel_straight_rays, sizeof(__pyx_k_data_kernel_straight_rays), 0, 0, 1, 1},
  {&__pyx_n_s_data_p, __pyx_k_data_p, sizeof(__pyx_k_data_p), 0, 0, 1, 1},
  {&__pyx_n_s_delta, __pyx_k_delta, sizeof(__pyx_k_delta), 0, 0, 1, 1},
  {&__pyx_n_s_dim, __pyx_k_dim, sizeof(__pyx_k_dim), 0, 0, 1, 1},
  {&__pyx_n_s_dl, __pyx_k_dl, sizeof(__pyx_k_dl), 0, 0, 1, 1},
  {&__pyx_n_s_dlx, __pyx_k_dlx, sizeof(__pyx_k_dlx), 0, 0, 1, 1},
  {&__pyx_n_s_dly, __pyx_k_dly, sizeof(__pyx_k_dly), 0, 0, 1, 1},
  {&__pyx_n_s_dlz, __pyx_k_dlz, sizeof(__pyx_k_dlz), 0, 0, 1, 1},
  {&__pyx_n_s_dtmp, __pyx_k_dtmp, sizeof(__pyx_k_dtmp), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dx, __pyx_k_dx, sizeof(__pyx_k_dx), 0, 0, 1, 1},
  {&__pyx_n_s_dz, __pyx_k_dz, sizeof(__pyx_k_dz), 0, 0, 1, 1},
  {&__pyx_n_u_elliptical, __pyx_k_elliptical, sizeof(__pyx_k_elliptical), 0, 1, 0, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_eps, __pyx_k_eps, sizeof(__pyx_k_eps), 0, 0, 1, 1},
  {&__pyx_n_s_epsilon, __pyx_k_epsilon, sizeof(__pyx_k_epsilon), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_flatten, __pyx_k_flatten, sizeof(__pyx_k_flatten), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_gamma, __pyx_k_gamma, sizeof(__pyx_k_gamma), 0, 0, 1, 1},
  {&__pyx_n_s_get_number_of_cells, __pyx_k_get_number_of_cells, sizeof(__pyx_k_get_number_of_cells), 0, 0, 1, 1},
  {&__pyx_n_s_get_number_of_nodes, __pyx_k_get_number_of_nodes, sizeof(__pyx_k_get_number_of_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_grx, __pyx_k_grx, sizeof(__pyx_k_grx), 0, 0, 1, 1},
  {&__pyx_n_s_gry, __pyx_k_gry, sizeof(__pyx_k_gry), 0, 0, 1, 1},
  {&__pyx_n_s_grz, __pyx_k_grz, sizeof(__pyx_k_grz), 0, 0, 1, 1},
  {&__pyx_n_s_hstack, __pyx_k_hstack, sizeof(__pyx_k_hstack), 0, 0, 1, 1},
  {&__pyx_n_s_hypo, __pyx_k_hypo, sizeof(__pyx_k_hypo), 0, 0, 1, 1},
  {&__pyx_kp_u_hypo_should_be_npts_x_5, __pyx_k_hypo_should_be_npts_x_5, sizeof(__pyx_k_hypo_should_be_npts_x_5), 0, 1, 0, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_iCell, __pyx_k_iCell, sizeof(__pyx_k_iCell), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_ind, __pyx_k_ind, sizeof(__pyx_k_ind), 0, 0, 1, 1},
  {&__pyx_n_s_indc, __pyx_k_indc, sizeof(__pyx_k_indc), 0, 0, 1, 1},
  {&__pyx_n_s_indices_p, __pyx_k_indices_p, sizeof(__pyx_k_indices_p), 0, 0, 1, 1},
  {&__pyx_n_s_indptr_p, __pyx_k_indptr_p, sizeof(__pyx_k_indptr_p), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_interp_vel, __pyx_k_interp_vel, sizeof(__pyx_k_interp_vel), 0, 0, 1, 1},
  {&__pyx_n_s_is_outside, __pyx_k_is_outside, sizeof(__pyx_k_is_outside), 0, 0, 1, 1},
  {&__pyx_n_u_iso, __pyx_k_iso, sizeof(__pyx_k_iso), 0, 1, 0, 1},
  {&__pyx_n_s_ix, __pyx_k_ix, sizeof(__pyx_k_ix), 0, 0, 1, 1},
  {&__pyx_n_s_iy, __pyx_k_iy, sizeof(__pyx_k_iy), 0, 0, 1, 1},
  {&__pyx_n_s_iz, __pyx_k_iz, sizeof(__pyx_k_iz), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_kron, __pyx_k_kron, sizeof(__pyx_k_kron), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_u_lowness, __pyx_k_lowness, sizeof(__pyx_k_lowness), 0, 1, 0, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_m_y, __pyx_k_m_y, sizeof(__pyx_k_m_y), 0, 0, 1, 1},
  {&__pyx_n_s_m_z, __pyx_k_m_z, sizeof(__pyx_k_m_z), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_maxit, __pyx_k_maxit, sizeof(__pyx_k_maxit), 0, 0, 1, 1},
  {&__pyx_n_s_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nCells, __pyx_k_nCells, sizeof(__pyx_k_nCells), 0, 0, 1, 1},
  {&__pyx_n_s_nTx, __pyx_k_nTx, sizeof(__pyx_k_nTx), 0, 0, 1, 1},
  {&__pyx_n_s_n_grx, __pyx_k_n_grx, sizeof(__pyx_k_n_grx), 0, 0, 1, 1},
  {&__pyx_n_s_n_gry, __pyx_k_n_gry, sizeof(__pyx_k_n_gry), 0, 0, 1, 1},
  {&__pyx_n_s_n_grz, __pyx_k_n_grz, sizeof(__pyx_k_n_grz), 0, 0, 1, 1},
  {&__pyx_n_s_n_secondary, __pyx_k_n_secondary, sizeof(__pyx_k_n_secondary), 0, 0, 1, 1},
  {&__pyx_n_s_n_tertiary, __pyx_k_n_tertiary, sizeof(__pyx_k_n_tertiary), 0, 0, 1, 1},
  {&__pyx_n_s_n_threads, __pyx_k_n_threads, sizeof(__pyx_k_n_threads), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_nonzero, __pyx_k_nonzero, sizeof(__pyx_k_nonzero), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nsnx, __pyx_k_nsnx, sizeof(__pyx_k_nsnx), 0, 0, 1, 1},
  {&__pyx_n_s_nsny, __pyx_k_nsny, sizeof(__pyx_k_nsny), 0, 0, 1, 1},
  {&__pyx_n_s_nsnz, __pyx_k_nsnz, sizeof(__pyx_k_nsnz), 0, 0, 1, 1},
  {&__pyx_n_s_nt, __pyx_k_nt, sizeof(__pyx_k_nt), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_numpy_support, __pyx_k_numpy_support, sizeof(__pyx_k_numpy_support), 0, 0, 1, 1},
  {&__pyx_n_s_numpy_to_vtk, __pyx_k_numpy_to_vtk, sizeof(__pyx_k_numpy_to_vtk), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_kp_u_order_value_not_valid_1_or_2_acc, __pyx_k_order_value_not_valid_1_or_2_acc, sizeof(__pyx_k_order_value_not_valid_1_or_2_acc), 0, 1, 0, 0},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_radius_tertiary, __pyx_k_radius_tertiary, sizeof(__pyx_k_radius_tertiary), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rays, __pyx_k_rays, sizeof(__pyx_k_rays), 0, 0, 1, 1},
  {&__pyx_n_s_rcv, __pyx_k_rcv, sizeof(__pyx_k_rcv), 0, 0, 1, 1},
  {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
  {&__pyx_n_s_rebuild2d, __pyx_k_rebuild2d, sizeof(__pyx_k_rebuild2d), 0, 0, 1, 1},
  {&__pyx_n_s_rebuild3d, __pyx_k_rebuild3d, sizeof(__pyx_k_rebuild3d), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_return_rays, __pyx_k_return_rays, sizeof(__pyx_k_return_rays), 0, 0, 1, 1},
  {&__pyx_n_s_rotated_template, __pyx_k_rotated_template, sizeof(__pyx_k_rotated_template), 0, 0, 1, 1},
  {&__pyx_n_s_save_raypaths, __pyx_k_save_raypaths, sizeof(__pyx_k_save_raypaths), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_sparse, __pyx_k_scipy_sparse, sizeof(__pyx_k_scipy_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_set_Vp0, __pyx_k_set_Vp0, sizeof(__pyx_k_set_Vp0), 0, 0, 1, 1},
  {&__pyx_n_s_set_Vs0, __pyx_k_set_Vs0, sizeof(__pyx_k_set_Vs0), 0, 0, 1, 1},
  {&__pyx_n_s_set_delta, __pyx_k_set_delta, sizeof(__pyx_k_set_delta), 0, 0, 1, 1},
  {&__pyx_n_s_set_epsilon, __pyx_k_set_epsilon, sizeof(__pyx_k_set_epsilon), 0, 0, 1, 1},
  {&__pyx_n_s_set_gamma, __pyx_k_set_gamma, sizeof(__pyx_k_set_gamma), 0, 0, 1, 1},
  {&__pyx_n_s_set_slowness, __pyx_k_set_slowness, sizeof(__pyx_k_set_slowness), 0, 0, 1, 1},
  {&__pyx_n_s_set_tilt_angle, __pyx_k_set_tilt_angle, sizeof(__pyx_k_set_tilt_angle), 0, 0, 1, 1},
  {&__pyx_n_s_set_verbose, __pyx_k_set_verbose, sizeof(__pyx_k_set_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_set_xi, __pyx_k_set_xi, sizeof(__pyx_k_set_xi), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_slowness, __pyx_k_slowness, sizeof(__pyx_k_slowness), 0, 0, 1, 1},
  {&__pyx_n_u_slowness, __pyx_k_slowness, sizeof(__pyx_k_slowness), 0, 1, 0, 1},
  {&__pyx_n_s_small, __pyx_k_small, sizeof(__pyx_k_small), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_kp_u_source_and_rcv_should_be_2D_arra, __pyx_k_source_and_rcv_should_be_2D_arra, sizeof(__pyx_k_source_and_rcv_should_be_2D_arra), 0, 1, 0, 0},
  {&__pyx_kp_u_source_should_be_either_nsrc_x_2, __pyx_k_source_should_be_either_nsrc_x_2, sizeof(__pyx_k_source_should_be_either_nsrc_x_2), 0, 1, 0, 0},
  {&__pyx_kp_u_source_should_be_either_nsrc_x_3, __pyx_k_source_should_be_either_nsrc_x_3, sizeof(__pyx_k_source_should_be_either_nsrc_x_3), 0, 1, 0, 0},
  {&__pyx_n_s_sp, __pyx_k_sp, sizeof(__pyx_k_sp), 0, 0, 1, 1},
  {&__pyx_kp_u_src_and_rcv_should_be_ndata_x_2, __pyx_k_src_and_rcv_should_be_ndata_x_2, sizeof(__pyx_k_src_and_rcv_should_be_ndata_x_2), 0, 1, 0, 0},
  {&__pyx_kp_u_src_and_rcv_should_be_ndata_x_3, __pyx_k_src_and_rcv_should_be_ndata_x_3, sizeof(__pyx_k_src_and_rcv_should_be_ndata_x_3), 0, 1, 0, 0},
  {&__pyx_kp_u_src_and_rcv_should_be_of_equal_s, __pyx_k_src_and_rcv_should_be_of_equal_s, sizeof(__pyx_k_src_and_rcv_should_be_of_equal_s), 0, 1, 0, 0},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_sy, __pyx_k_sy, sizeof(__pyx_k_sy), 0, 0, 1, 1},
  {&__pyx_n_s_sz, __pyx_k_sz, sizeof(__pyx_k_sz), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theta, __pyx_k_theta, sizeof(__pyx_k_theta), 0, 0, 1, 1},
  {&__pyx_kp_u_theta_has_wrong_shape, __pyx_k_theta_has_wrong_shape, sizeof(__pyx_k_theta_has_wrong_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_theta_must_be_1D_or_3D_ndarray, __pyx_k_theta_must_be_1D_or_3D_ndarray, sizeof(__pyx_k_theta_must_be_1D_or_3D_ndarray), 0, 1, 0, 0},
  {&__pyx_kp_u_theta_vector_has_wrong_size, __pyx_k_theta_vector_has_wrong_size, sizeof(__pyx_k_theta_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_n_s_thread_no, __pyx_k_thread_no, sizeof(__pyx_k_thread_no), 0, 0, 1, 1},
  {&__pyx_n_s_tile, __pyx_k_tile, sizeof(__pyx_k_tile), 0, 0, 1, 1},
  {&__pyx_n_u_tilted_elliptical, __pyx_k_tilted_elliptical, sizeof(__pyx_k_tilted_elliptical), 0, 1, 0, 1},
  {&__pyx_n_s_tt_from_rp, __pyx_k_tt_from_rp, sizeof(__pyx_k_tt_from_rp), 0, 0, 1, 1},
  {&__pyx_n_s_ttcrpy_rgrid, __pyx_k_ttcrpy_rgrid, sizeof(__pyx_k_ttcrpy_rgrid), 0, 0, 1, 1},
  {&__pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_k_ttcrpy_rgrid_pyx, sizeof(__pyx_k_ttcrpy_rgrid_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_unique, __pyx_k_unique, sizeof(__pyx_k_unique), 0, 0, 1, 1},
  {&__pyx_n_s_up, __pyx_k_up, sizeof(__pyx_k_up), 0, 0, 1, 1},
  {&__pyx_n_s_up_y, __pyx_k_up_y, sizeof(__pyx_k_up_y), 0, 0, 1, 1},
  {&__pyx_n_s_up_z, __pyx_k_up_z, sizeof(__pyx_k_up_z), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_kp_u_v_has_wrong_shape, __pyx_k_v_has_wrong_shape, sizeof(__pyx_k_v_has_wrong_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_v_must_be_1D_or_3D_ndarray, __pyx_k_v_must_be_1D_or_3D_ndarray, sizeof(__pyx_k_v_must_be_1D_or_3D_ndarray), 0, 1, 0, 0},
  {&__pyx_kp_u_v_vector_has_wrong_size, __pyx_k_v_vector_has_wrong_size, sizeof(__pyx_k_v_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_n_u_velocity, __pyx_k_velocity, sizeof(__pyx_k_velocity), 0, 1, 0, 1},
  {&__pyx_kp_u_velocity_has_wrong_shape, __pyx_k_velocity_has_wrong_shape, sizeof(__pyx_k_velocity_has_wrong_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_velocity_must_be_1D_or_3D_ndarra, __pyx_k_velocity_must_be_1D_or_3D_ndarra, sizeof(__pyx_k_velocity_must_be_1D_or_3D_ndarra), 0, 1, 0, 0},
  {&__pyx_kp_u_velocity_vector_has_wrong_size, __pyx_k_velocity_vector_has_wrong_size, sizeof(__pyx_k_velocity_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_n_s_vstack, __pyx_k_vstack, sizeof(__pyx_k_vstack), 0, 0, 1, 1},
  {&__pyx_n_u_vti_psv, __pyx_k_vti_psv, sizeof(__pyx_k_vti_psv), 0, 1, 0, 1},
  {&__pyx_n_u_vti_sh, __pyx_k_vti_sh, sizeof(__pyx_k_vti_sh), 0, 1, 0, 1},
  {&__pyx_n_s_vtk, __pyx_k_vtk, sizeof(__pyx_k_vtk), 0, 0, 1, 1},
  {&__pyx_n_s_vtkCellArray, __pyx_k_vtkCellArray, sizeof(__pyx_k_vtkCellArray), 0, 0, 1, 1},
  {&__pyx_n_s_vtkDoubleArray, __pyx_k_vtkDoubleArray, sizeof(__pyx_k_vtkDoubleArray), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPoints, __pyx_k_vtkPoints, sizeof(__pyx_k_vtkPoints), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPolyData, __pyx_k_vtkPolyData, sizeof(__pyx_k_vtkPolyData), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPolyLine, __pyx_k_vtkPolyLine, sizeof(__pyx_k_vtkPolyLine), 0, 0, 1, 1},
  {&__pyx_n_s_vtkRectilinearGrid, __pyx_k_vtkRectilinearGrid, sizeof(__pyx_k_vtkRectilinearGrid), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLPolyDataWriter, __pyx_k_vtkXMLPolyDataWriter, sizeof(__pyx_k_vtkXMLPolyDataWriter), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLRectilinearGridReader, __pyx_k_vtkXMLRectilinearGridReader, sizeof(__pyx_k_vtkXMLRectilinearGridReader), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLRectilinearGridWriter, __pyx_k_vtkXMLRectilinearGridWriter, sizeof(__pyx_k_vtkXMLRectilinearGridWriter), 0, 0, 1, 1},
  {&__pyx_n_s_vtk_to_numpy, __pyx_k_vtk_to_numpy, sizeof(__pyx_k_vtk_to_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_vtk_util, __pyx_k_vtk_util, sizeof(__pyx_k_vtk_util), 0, 0, 1, 1},
  {&__pyx_kp_u_vtp, __pyx_k_vtp, sizeof(__pyx_k_vtp), 0, 1, 0, 0},
  {&__pyx_kp_u_vtr, __pyx_k_vtr, sizeof(__pyx_k_vtr), 0, 1, 0, 0},
  {&__pyx_n_s_weno, __pyx_k_weno, sizeof(__pyx_k_weno), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_x1, __pyx_k_x1, sizeof(__pyx_k_x1), 0, 0, 1, 1},
  {&__pyx_n_s_x2, __pyx_k_x2, sizeof(__pyx_k_x2), 0, 0, 1, 1},
  {&__pyx_n_s_xc, __pyx_k_xc, sizeof(__pyx_k_xc), 0, 0, 1, 1},
  {&__pyx_n_s_xe, __pyx_k_xe, sizeof(__pyx_k_xe), 0, 0, 1, 1},
  {&__pyx_n_s_xi, __pyx_k_xi, sizeof(__pyx_k_xi), 0, 0, 1, 1},
  {&__pyx_kp_u_xi_has_wrong_shape, __pyx_k_xi_has_wrong_shape, sizeof(__pyx_k_xi_has_wrong_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_xi_must_be_1D_or_3D_ndarray, __pyx_k_xi_must_be_1D_or_3D_ndarray, sizeof(__pyx_k_xi_must_be_1D_or_3D_ndarray), 0, 1, 0, 0},
  {&__pyx_kp_u_xi_vector_has_wrong_size, __pyx_k_xi_vector_has_wrong_size, sizeof(__pyx_k_xi_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_n_s_xr, __pyx_k_xr, sizeof(__pyx_k_xr), 0, 0, 1, 1},
  {&__pyx_n_s_xs, __pyx_k_xs, sizeof(__pyx_k_xs), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_y1, __pyx_k_y1, sizeof(__pyx_k_y1), 0, 0, 1, 1},
  {&__pyx_n_s_y2, __pyx_k_y2, sizeof(__pyx_k_y2), 0, 0, 1, 1},
  {&__pyx_n_s_yc, __pyx_k_yc, sizeof(__pyx_k_yc), 0, 0, 1, 1},
  {&__pyx_n_s_ye, __pyx_k_ye, sizeof(__pyx_k_ye), 0, 0, 1, 1},
  {&__pyx_n_s_yi, __pyx_k_yi, sizeof(__pyx_k_yi), 0, 0, 1, 1},
  {&__pyx_n_s_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 0, 1, 1},
  {&__pyx_n_s_z1, __pyx_k_z1, sizeof(__pyx_k_z1), 0, 0, 1, 1},
  {&__pyx_n_s_z2, __pyx_k_z2, sizeof(__pyx_k_z2), 0, 0, 1, 1},
  {&__pyx_n_s_zc, __pyx_k_zc, sizeof(__pyx_k_zc), 0, 0, 1, 1},
  {&__pyx_n_s_ze, __pyx_k_ze, sizeof(__pyx_k_ze), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {&__pyx_n_s_zi, __pyx_k_zi, sizeof(__pyx_k_zi), 0, 0, 1, 1},
  {&__pyx_n_s_zr, __pyx_k_zr, sizeof(__pyx_k_zr), 0, 0, 1, 1},
  {&__pyx_n_s_zs, __pyx_k_zs, sizeof(__pyx_k_zs), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 1228, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 179, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 833, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 884, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ttcrpy/rgrid.pyx":179
 *         if method == 'FSM':
 *             if np.abs(self._dx - self._dy)>0.000001 or np.abs(self._dx - self._dz)>0.000001:
 *                 raise ValueError('FSM: Grid cells must be cubic')             # <<<<<<<<<<<<<<
 * 
 *         for val in x:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_FSM_Grid_cells_must_be_cubic); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "ttcrpy/rgrid.pyx":364
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_Thread_number_is_larger_than_num); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "ttcrpy/rgrid.pyx":440
 *             True if at least one point outside grid
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or             # <<<<<<<<<<<<<<
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 */
  __pyx_slice__3 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__3);
  __Pyx_GIVEREF(__pyx_slice__3);
  __pyx_tuple__4 = PyTuple_Pack(2, __pyx_slice__3, __pyx_int_0); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "ttcrpy/rgrid.pyx":441
 *         """
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or             # <<<<<<<<<<<<<<
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_slice__3, __pyx_int_1); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "ttcrpy/rgrid.pyx":442
 *         return ( np.min(pts[:,0]) < self._x.front() or np.max(pts[:,0]) > self._x.back() or
 *                 np.min(pts[:,1]) < self._y.front() or np.max(pts[:,1]) > self._y.back() or
 *                 np.min(pts[:,2]) < self._z.front() or np.max(pts[:,2]) > self._z.back() )             # <<<<<<<<<<<<<<
 * 
 *     def get_slowness(self):
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_slice__3, __pyx_int_2); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "ttcrpy/rgrid.pyx":480
 *             nz = self._z.size()
 *         if slowness.size != nx*ny*nz:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Slowness_vector_has_wrong_size); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "ttcrpy/rgrid.pyx":486
 *         if slowness.ndim == 3:
 *             if slowness.shape != (nx, ny, nz):
 *                 raise ValueError('Slowness has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = slowness.flatten('F')
 *             for i in range(nx*ny*nz):
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Slowness_has_wrong_shape); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "ttcrpy/rgrid.pyx":496
 *                 slown.push_back(tmp[i])
 *         else:
 *             raise ValueError('Slowness must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_Slowness_must_be_1D_or_3D_ndarra); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "ttcrpy/rgrid.pyx":519
 *             nz = self._z.size()
 *         if velocity.size != nx*ny*nz:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_velocity_vector_has_wrong_size); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "ttcrpy/rgrid.pyx":525
 *         if velocity.ndim == 3:
 *             if velocity.shape != (nx, ny, nz):
 *                 raise ValueError('velocity has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = velocity.flatten('F')
 *             for i in range(nx*ny*nz):
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_velocity_has_wrong_shape); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "ttcrpy/rgrid.pyx":535
 *                 slown.push_back(1./tmp[i])
 *         else:
 *             raise ValueError('velocity must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_velocity_must_be_1D_or_3D_ndarra); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "ttcrpy/rgrid.pyx":556
 *         """
 *         if self.is_outside(coord):
 *             raise ValueError('Velocity data point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         # D is npts x nparams
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Velocity_data_point_outside_grid); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "ttcrpy/rgrid.pyx":578
 *         else:
 *             # for each point in coord, we have 8 values in D
 *             ivec = np.kron(np.arange(coord.shape[0], dtype=np.int64),np.ones(8, dtype=np.int64))             # <<<<<<<<<<<<<<
 *             jvec = np.zeros(ivec.shape, dtype=np.int64)
 *             vec = np.zeros(ivec.shape)
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_int_8); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "ttcrpy/rgrid.pyx":619
 *         nx, ny, nz = self.shape
 *         # Kx
 *         iK = np.kron(np.arange(nx*ny*nz, dtype=np.int64), np.ones(3,dtype=np.int64))             # <<<<<<<<<<<<<<
 *         val = np.tile(np.array([1., -2., 1.]), nx*ny*nz) / (self._dx*self._dx)
 *         # i=0 -> forward op
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "ttcrpy/rgrid.pyx":710
 * 
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')             # <<<<<<<<<<<<<<
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_hypo_should_be_npts_x_5); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "ttcrpy/rgrid.pyx":711
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]             # <<<<<<<<<<<<<<
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 */
  __pyx_slice__17 = PySlice_New(__pyx_int_2, __pyx_int_5, Py_None); if (unlikely(!__pyx_slice__17)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__17);
  __Pyx_GIVEREF(__pyx_slice__17);
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_slice__3, __pyx_slice__17); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "ttcrpy/rgrid.pyx":824
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if self.method == b'd' and aggregate_src:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_source_and_rcv_should_be_2D_arra); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "ttcrpy/rgrid.pyx":827
 * 
 *         if self.method == b'd' and aggregate_src:
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and compute_M:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_Cannot_aggregate_source_with_DSP); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "ttcrpy/rgrid.pyx":830
 * 
 *         if compute_L and compute_M:
 *             raise ValueError('compute_L and compute_M are mutually exclusive')             # <<<<<<<<<<<<<<
 * 
 *         if self.cell_slowness and compute_M:
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_u_compute_L_and_compute_M_are_mutu); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "ttcrpy/rgrid.pyx":833
 * 
 *         if self.cell_slowness and compute_M:
 *             raise NotImplementedError('compute_M not defined for grids with slowness defined for cells')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and not self.cell_slowness:
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u_compute_M_not_defined_for_grids); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "ttcrpy/rgrid.pyx":836
 * 
 *         if compute_L and not self.cell_slowness:
 *             raise NotImplementedError('compute_L defined only for grids with slowness defined for cells')             # <<<<<<<<<<<<<<
 * 
 *         evID = None
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_compute_L_defined_only_for_grids); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "ttcrpy/rgrid.pyx":851
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
  __pyx_slice__24 = PySlice_New(__pyx_int_1, __pyx_int_4, Py_None); if (unlikely(!__pyx_slice__24)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__24);
  __Pyx_GIVEREF(__pyx_slice__24);
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_slice__3, __pyx_slice__24); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "ttcrpy/rgrid.pyx":857
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_source_should_be_either_nsrc_x_3); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "ttcrpy/rgrid.pyx":860
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 *             raise ValueError('src and rcv should be ndata x 3')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_ndata_x_3); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "ttcrpy/rgrid.pyx":863
 * 
 *         if self.is_outside(src):
 *             raise ValueError('Source point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(rcv):
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_Source_point_outside_grid); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "ttcrpy/rgrid.pyx":866
 * 
 *         if self.is_outside(rcv):
 *             raise ValueError('Receiver outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_Receiver_outside_grid); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "ttcrpy/rgrid.pyx":897
 *         if evID is None:
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_1); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_2); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "ttcrpy/rgrid.pyx":915
 *             else:
 *                 if src.shape != rcv.shape:
 *                     raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *                 for n in range(nTx):
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_of_equal_s); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "ttcrpy/rgrid.pyx":1262
 *         z = numpy_support.vtk_to_numpy(data.GetZCoordinates())
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',             # <<<<<<<<<<<<<<
 *                  'P-wave velocity')
 *         for name in names:
 */
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_u_Slowness, __pyx_n_u_slowness, __pyx_n_u_Velocity, __pyx_n_u_velocity, __pyx_kp_u_P_wave_velocity); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "ttcrpy/rgrid.pyx":1276
 *                 break
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')             # <<<<<<<<<<<<<<
 * 
 *         if 'lowness' in name:
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_u_File_should_contain_slowness_or); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "ttcrpy/rgrid.pyx":1719
 * 
 *         if centers:
 *             xc = (grx[1:]+grx[:-1])/2             # <<<<<<<<<<<<<<
 *             yc = (gry[1:]+gry[:-1])/2
 *             zc = (grz[1:]+grz[:-1])/2
 */
  __pyx_slice__37 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__37)) __PYX_ERR(0, 1719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__37);
  __Pyx_GIVEREF(__pyx_slice__37);
  __pyx_slice__38 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__38)) __PYX_ERR(0, 1719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__38);
  __Pyx_GIVEREF(__pyx_slice__38);

  /* "ttcrpy/rgrid.pyx":1861
 *                                     xmin, zmin, nsnx, nsnz, n_threads)
 *                 else:
 *                     raise ValueError('Anisotropy model not implemented')             # <<<<<<<<<<<<<<
 *             elif method == 'FSM':
 *                 self.method = b'f'
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_u_Anisotropy_model_not_implemented); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "ttcrpy/rgrid.pyx":2124
 *             nz = self._z.size()
 *         if xi.size != nx*nz:
 *             raise ValueError('xi vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_u_xi_vector_has_wrong_size); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 2124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "ttcrpy/rgrid.pyx":2130
 *         if xi.ndim == 2:
 *             if xi.shape != (nx, nz):
 *                 raise ValueError('xi has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = xi.flatten()
 *             for i in range(nx*nz):
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_u_xi_has_wrong_shape); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 2130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "ttcrpy/rgrid.pyx":2139
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('xi must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setXi(data)
 * 
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_u_xi_must_be_1D_or_3D_ndarray); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 2139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "ttcrpy/rgrid.pyx":2160
 *             nz = self._z.size()
 *         if theta.size != nx*nz:
 *             raise ValueError('theta vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_u_theta_vector_has_wrong_size); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 2160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "ttcrpy/rgrid.pyx":2166
 *         if theta.ndim == 2:
 *             if theta.shape != (nx, nz):
 *                 raise ValueError('theta has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = theta.flatten()
 *             for i in range(nx*nz):
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_u_theta_has_wrong_shape); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 2166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "ttcrpy/rgrid.pyx":2175
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('theta must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setTiltAngle(data)
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_u_theta_must_be_1D_or_3D_ndarray); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 2175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "ttcrpy/rgrid.pyx":2196
 *             nz = self._z.size()
 *         if v.size != nx*nz:
 *             raise ValueError('v vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] data
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_u_v_vector_has_wrong_size); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 2196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "ttcrpy/rgrid.pyx":2202
 *         if v.ndim == 2:
 *             if v.shape != (nx, nz):
 *                 raise ValueError('v has wrong shape')             # <<<<<<<<<<<<<<
 *             tmp = v.flatten()
 *             for i in range(nx*nz):
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_u_v_has_wrong_shape); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 2202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "ttcrpy/rgrid.pyx":2211
 *                 data.push_back(tmp[i])
 *         else:
 *             raise ValueError('v must be 1D or 3D ndarray')             # <<<<<<<<<<<<<<
 *         self.grid.setVp0(data)
 * 
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_u_v_must_be_1D_or_3D_ndarray); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 2211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "ttcrpy/rgrid.pyx":2382
 *             idz = 1 / self.dz
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((2, ), dtype=np.int32))             # <<<<<<<<<<<<<<
 *             j = np.zeros((nz * nx * 2, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 2, ))
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_tuple__49); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "ttcrpy/rgrid.pyx":2425
 *             idz2 = 1 / (self.dz * self.dz)
 * 
 *             i = np.kron(np.arange(nx * nz), np.ones((3, ), dtype=np.int32))             # <<<<<<<<<<<<<<
 *             j = np.zeros((nz * nx * 3, ), dtype=np.int32)
 *             v = np.zeros((nz * nx * 3, ))
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_tuple__15); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 2425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "ttcrpy/rgrid.pyx":2458
 * 
 *         else:
 *             raise ValueError('order value not valid (1 or 2 accepted)')             # <<<<<<<<<<<<<<
 * 
 *         return Kx, Kz
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_u_order_value_not_valid_1_or_2_acc); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 2458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "ttcrpy/rgrid.pyx":2548
 * 
 *         if self.iso != b'i' and self.method != b's' and not self.cell_slowness:
 *             raise NotImplementedError('Anisotropic raytracing implemented only for SPM')             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and not self.cell_slowness:
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_u_Anisotropic_raytracing_implement); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "ttcrpy/rgrid.pyx":2554
 * 
 *         if compute_L and (self.iso == b'p' or self.iso == b'h'):
 *             raise NotImplementedError('compute_L not implemented for VTI media')             # <<<<<<<<<<<<<<
 * 
 *         if source.shape[1] == 2:
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_u_compute_L_not_implemented_for_VT); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 2554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "ttcrpy/rgrid.pyx":2562
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
  __pyx_slice__55 = PySlice_New(__pyx_int_1, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice__55)) __PYX_ERR(0, 2562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__55);
  __Pyx_GIVEREF(__pyx_slice__55);
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_slice__3, __pyx_slice__55); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 2562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "ttcrpy/rgrid.pyx":2568
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_u_source_should_be_either_nsrc_x_2); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 2568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "ttcrpy/rgrid.pyx":2571
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 *             raise ValueError('src and rcv should be ndata x 2')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_ndata_x_2); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 2571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "ttcrpy/rgrid.pyx":38
 *     void setVerbose(int)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """
 *     Set verbosity level for C++ code
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_n_s_v); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_set_verbose, 38, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 38, __pyx_L1_error)

  /* "ttcrpy/rgrid.pyx":1229
 * 
 *     @staticmethod
 *     def builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0,             # <<<<<<<<<<<<<<
 *                 eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5,
 *                 n_secondary=2, n_tertiary=2, radius_tertiary=1.0):
 */
  __pyx_tuple__64 = PyTuple_Pack(25, __pyx_n_s_filename, __pyx_n_s_n_threads, __pyx_n_s_method, __pyx_n_s_tt_from_rp, __pyx_n_s_interp_vel, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_weno, __pyx_n_s_nsnx, __pyx_n_s_nsny, __pyx_n_s_nsnz, __pyx_n_s_n_secondary, __pyx_n_s_n_tertiary, __pyx_n_s_radius_tertiary, __pyx_n_s_reader, __pyx_n_s_data, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_z, __pyx_n_s_names, __pyx_n_s_name_2, __pyx_n_s_cell_slowness, __pyx_n_s_dim, __pyx_n_s_slowness, __pyx_n_s_g); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(14, 0, 25, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_builder, 1229, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 1229, __pyx_L1_error)

  /* "ttcrpy/rgrid.pyx":1290
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */
  __pyx_tuple__66 = PyTuple_Pack(55, __pyx_n_s_Tx, __pyx_n_s_Rx, __pyx_n_s_grx, __pyx_n_s_gry, __pyx_n_s_grz, __pyx_n_s_centers, __pyx_n_s_nTx, __pyx_n_s_n_grx, __pyx_n_s_n_gry, __pyx_n_s_n_grz, __pyx_n_s_small, __pyx_n_s_data_p, __pyx_n_s_indices_p, __pyx_n_s_indptr_p, __pyx_n_s_k, __pyx_n_s_ix, __pyx_n_s_iy, __pyx_n_s_iz, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_z, __pyx_n_s_x1, __pyx_n_s_y1, __pyx_n_s_z1, __pyx_n_s_x2, __pyx_n_s_y2, __pyx_n_s_z2, __pyx_n_s_dtmp, __pyx_n_s_d, __pyx_n_s_l, __pyx_n_s_m, __pyx_n_s_n, __pyx_n_s_m_y, __pyx_n_s_b_y, __pyx_n_s_m_z, __pyx_n_s_b_z, __pyx_n_s_dlx, __pyx_n_s_dly, __pyx_n_s_dlz, __pyx_n_s_dl, __pyx_n_s_xe, __pyx_n_s_ye, __pyx_n_s_ze, __pyx_n_s_sy, __pyx_n_s_sz, __pyx_n_s_up_y, __pyx_n_s_up_z, __pyx_n_s_nt, __pyx_n_s_iCell, __pyx_n_s_zi, __pyx_n_s_yi, __pyx_n_s_L, __pyx_n_s_xc, __pyx_n_s_yc, __pyx_n_s_zc); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(6, 0, 55, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_data_kernel_straight_rays, 1290, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 1290, __pyx_L1_error)

  /* "ttcrpy/rgrid.pyx":2897
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */
  __pyx_tuple__68 = PyTuple_Pack(35, __pyx_n_s_Tx, __pyx_n_s_Rx, __pyx_n_s_grx, __pyx_n_s_grz, __pyx_n_s_aniso, __pyx_n_s_nTx, __pyx_n_s_n_grx, __pyx_n_s_n_grz, __pyx_n_s_nCells, __pyx_n_s_small, __pyx_n_s_data_p, __pyx_n_s_indices_p, __pyx_n_s_indptr_p, __pyx_n_s_k, __pyx_n_s_ix, __pyx_n_s_iz, __pyx_n_s_x, __pyx_n_s_z, __pyx_n_s_xs, __pyx_n_s_xr, __pyx_n_s_zs, __pyx_n_s_zr, __pyx_n_s_dtmp, __pyx_n_s_dlx, __pyx_n_s_dlz, __pyx_n_s_m, __pyx_n_s_b, __pyx_n_s_ze, __pyx_n_s_xe, __pyx_n_s_iCell, __pyx_n_s_up, __pyx_n_s_n, __pyx_n_s_zi, __pyx_n_s_dl, __pyx_n_s_L); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 2897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(5, 0, 35, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_data_kernel_straight_rays, 2897, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 2897, __pyx_L1_error)

  /* "ttcrpy/rgrid.pyx":3109
 * 
 * 
 * def _rebuild3d(x, y, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,
 *      weno, nsnx, nsny, nsnz, n_secondary,
 */
  __pyx_tuple__70 = PyTuple_Pack(19, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_z, __pyx_n_s_constructor_params, __pyx_n_s_n_threads, __pyx_n_s_cell_slowness, __pyx_n_s_method, __pyx_n_s_tt_from_rp, __pyx_n_s_interp_vel, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_weno, __pyx_n_s_nsnx, __pyx_n_s_nsny, __pyx_n_s_nsnz, __pyx_n_s_n_secondary, __pyx_n_s_n_tertiary, __pyx_n_s_radius_tertiary, __pyx_n_s_g); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 3109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(4, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_rebuild3d, 3109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 3109, __pyx_L1_error)

  /* "ttcrpy/rgrid.pyx":3118
 *     return g
 * 
 * def _rebuild2d(x, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params
 */
  __pyx_tuple__72 = PyTuple_Pack(14, __pyx_n_s_x, __pyx_n_s_z, __pyx_n_s_constructor_params, __pyx_n_s_n_threads, __pyx_n_s_cell_slowness, __pyx_n_s_method, __pyx_n_s_aniso, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_weno, __pyx_n_s_rotated_template, __pyx_n_s_nsnx, __pyx_n_s_nsnz, __pyx_n_s_g); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 3118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(3, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_rgrid_pyx, __pyx_n_s_rebuild2d, 3118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 3118, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_eneg_6 = PyFloat_FromDouble(1.e-6); if (unlikely(!__pyx_float_1_eneg_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_eneg_15 = PyFloat_FromDouble(1.e-15); if (unlikely(!__pyx_float_1_eneg_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_000001 = PyFloat_FromDouble(0.000001); if (unlikely(!__pyx_float_0_000001)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_2_ = PyFloat_FromDouble(-2.); if (unlikely(!__pyx_float_neg_2_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_0_5 = PyFloat_FromDouble(-0.5); if (unlikely(!__pyx_float_neg_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_2_0 = PyFloat_FromDouble(-2.0); if (unlikely(!__pyx_float_neg_2_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6ttcrpy_5rgrid_Grid3d = &__pyx_vtable_6ttcrpy_5rgrid_Grid3d;
  __pyx_vtable_6ttcrpy_5rgrid_Grid3d._f2c_ind = (uint32_t (*)(struct __pyx_obj_6ttcrpy_5rgrid_Grid3d *, uint32_t))__pyx_f_6ttcrpy_5rgrid_6Grid3d__f2c_ind;
  if (PyType_Ready(&__pyx_type_6ttcrpy_5rgrid_Grid3d) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6ttcrpy_5rgrid_Grid3d.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ttcrpy_5rgrid_Grid3d.tp_dictoffset && __pyx_type_6ttcrpy_5rgrid_Grid3d.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ttcrpy_5rgrid_Grid3d.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ttcrpy_5rgrid_Grid3d.tp_dict, __pyx_vtabptr_6ttcrpy_5rgrid_Grid3d) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Grid3d, (PyObject *)&__pyx_type_6ttcrpy_5rgrid_Grid3d) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_ptype_6ttcrpy_5rgrid_Grid3d = &__pyx_type_6ttcrpy_5rgrid_Grid3d;
  if (PyType_Ready(&__pyx_type_6ttcrpy_5rgrid_Grid2d) < 0) __PYX_ERR(0, 1727, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6ttcrpy_5rgrid_Grid2d.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ttcrpy_5rgrid_Grid2d.tp_dictoffset && __pyx_type_6ttcrpy_5rgrid_Grid2d.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ttcrpy_5rgrid_Grid2d.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Grid2d, (PyObject *)&__pyx_type_6ttcrpy_5rgrid_Grid2d) < 0) __PYX_ERR(0, 1727, __pyx_L1_error)
  __pyx_ptype_6ttcrpy_5rgrid_Grid2d = &__pyx_type_6ttcrpy_5rgrid_Grid2d;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initrgrid(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initrgrid(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_rgrid(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_rgrid(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_rgrid(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'rgrid' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_rgrid(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("rgrid", __pyx_methods, __pyx_k_Raytracing_on_rectilinear_grids, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ttcrpy__rgrid) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ttcrpy.rgrid")) {
      if (unlikely(PyDict_SetItemString(modules, "ttcrpy.rgrid", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ttcrpy/rgrid.pyx":24
 * # distutils: language = c++
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * import scipy.sparse as sp
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":26
 * import numpy as np
 * cimport numpy as np
 * import scipy.sparse as sp             # <<<<<<<<<<<<<<
 * 
 * import vtk
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__61);
  __Pyx_GIVEREF(__pyx_n_s__61);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__61);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_sparse, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sp, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":28
 * import scipy.sparse as sp
 * 
 * import vtk             # <<<<<<<<<<<<<<
 * from vtk.util import numpy_support
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_vtk, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtk, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":29
 * 
 * import vtk
 * from vtk.util import numpy_support             # <<<<<<<<<<<<<<
 * 
 * from ttcrpy.rgrid cimport Grid3D, Grid3Drcfs, Grid3Drcsp, Grid3Drcdsp, \
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_numpy_support);
  __Pyx_GIVEREF(__pyx_n_s_numpy_support);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_numpy_support);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_vtk_util, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy_support, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":38
 *     void setVerbose(int)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """
 *     Set verbosity level for C++ code
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_1set_verbose, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_set_verbose, __pyx_t_1) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/rgrid.pyx":1229
 * 
 *     @staticmethod
 *     def builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0,             # <<<<<<<<<<<<<<
 *                 eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5,
 *                 n_secondary=2, n_tertiary=2, radius_tertiary=1.0):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_6Grid3d_39builder, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d->tp_dict, __pyx_n_s_builder, __pyx_t_1) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid3d);

  /* "ttcrpy/rgrid.pyx":1228
 * 
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def builder(filename, n_threads=1, method='FSM', tt_from_rp=1, interp_vel=0,
 *                 eps=1.e-15, maxit=20, weno=1, nsnx=5, nsny=5, nsnz=5,
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d, __pyx_n_s_builder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d->tp_dict, __pyx_n_s_builder, __pyx_t_2) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid3d);

  /* "ttcrpy/rgrid.pyx":1290
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_6Grid3d_41data_kernel_straight_rays, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d->tp_dict, __pyx_n_s_data_kernel_straight_rays, __pyx_t_2) < 0) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid3d);

  /* "ttcrpy/rgrid.pyx":1289
 *         return g
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d, __pyx_n_s_data_kernel_straight_rays); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid3d->tp_dict, __pyx_n_s_data_kernel_straight_rays, __pyx_t_1) < 0) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid3d);

  /* "ttcrpy/rgrid.pyx":2897
 * 
 *     @staticmethod
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,             # <<<<<<<<<<<<<<
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 *                                   np.ndarray[np.double_t, ndim=1] grx,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_6Grid2d_43data_kernel_straight_rays, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid2d->tp_dict, __pyx_n_s_data_kernel_straight_rays, __pyx_t_1) < 0) __PYX_ERR(0, 2897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid2d);

  /* "ttcrpy/rgrid.pyx":2896
 * 
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def data_kernel_straight_rays(np.ndarray[np.double_t, ndim=2] Tx,
 *                                   np.ndarray[np.double_t, ndim=2] Rx,
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid2d, __pyx_n_s_data_kernel_straight_rays); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5rgrid_Grid2d->tp_dict, __pyx_n_s_data_kernel_straight_rays, __pyx_t_2) < 0) __PYX_ERR(0, 2897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5rgrid_Grid2d);

  /* "ttcrpy/rgrid.pyx":3109
 * 
 * 
 * def _rebuild3d(x, y, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, tt_from_rp, interp_vel, eps, maxit,
 *      weno, nsnx, nsny, nsnz, n_secondary,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_3_rebuild3d, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rebuild3d, __pyx_t_2) < 0) __PYX_ERR(0, 3109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":3118
 *     return g
 * 
 * def _rebuild2d(x, z, constructor_params):             # <<<<<<<<<<<<<<
 *     (n_threads, cell_slowness, method, aniso, eps, maxit, weno,
 *      rotated_template, nsnx, nsnz) = constructor_params
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5rgrid_5_rebuild2d, NULL, __pyx_n_s_ttcrpy_rgrid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rebuild2d, __pyx_t_2) < 0) __PYX_ERR(0, 3118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/rgrid.pyx":1
 * # -*- coding: utf-8 -*-             # <<<<<<<<<<<<<<
 * """
 * Raytracing on rectilinear grids
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ttcrpy.rgrid", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ttcrpy.rgrid");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* BytesEquals */
  static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
  static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyIntCompare */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(operand) if (unlikely(zerodivision_check && ((operand) == 0))) {\
    PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");\
    return NULL;\
}
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(0) break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        }
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyFloatBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: b = 0.0; break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
    }
        
        PyFPE_START_PROTECT("add", return NULL)
        result = a + b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyFloatBinop */
      #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: b = 0.0; break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyIntCompare */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyObjectGetMethod */
        static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
        static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
        static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_sst_abs(size) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            if (unlikely(zerodivision_check && b == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");
                return NULL;
            }
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* SliceObject */
        static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* GetTopmostException */
        #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
        #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
        #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
        static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
        #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
        static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
        static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetNameInClass */
        static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


        /* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
    const int64_t neg_one = (int64_t) ((int64_t) 0 - (int64_t) 1), const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* Declarations */
        #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
        #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
        #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
        #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
