/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include/numpy/ufuncobject.h",
            "ttcr/Cell.h",
            "ttcr/Grid2D.h",
            "ttcr/Grid2Duc.h",
            "ttcr/Grid2Ducfs.h",
            "ttcr/Grid2Ducsp.h",
            "ttcr/Grid2Dun.h",
            "ttcr/Grid2Dunfs.h",
            "ttcr/Grid2Dunsp.h",
            "ttcr/Grid3D.h",
            "ttcr/Grid3Duc.h",
            "ttcr/Grid3Ducdsp.h",
            "ttcr/Grid3Ducfs.h",
            "ttcr/Grid3Ducsp.h",
            "ttcr/Grid3Dun.h",
            "ttcr/Grid3Dundsp.h",
            "ttcr/Grid3Dunfs.h",
            "ttcr/Grid3Dunsp.h",
            "ttcr/Node2Dc.h",
            "ttcr/Node2Dcsp.h",
            "ttcr/Node2Dn.h",
            "ttcr/Node2Dnsp.h",
            "ttcr/Node3Dc.h",
            "ttcr/Node3Dcsp.h",
            "ttcr/Node3Dn.h",
            "ttcr/Node3Dnsp.h",
            "ttcr/ttcr_t.h",
            "ttcrpy/utils_cython.h",
            "ttcrpy/verbose.h"
        ],
        "extra_compile_args": [
            "-std=c++11",
            "-stdlib=libc++",
            "-O3"
        ],
        "include_dirs": [
            "ttcrpy",
            "ttcr",
            "boost_1_72_0",
            "eigen-3.3.7",
            "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/core/include"
        ],
        "language": "c++",
        "name": "ttcrpy.tmesh",
        "sources": [
            "ttcrpy/tmesh.pyx",
            "ttcrpy/verbose.cpp"
        ]
    },
    "module_name": "ttcrpy.tmesh"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ttcrpy__tmesh
#define __PYX_HAVE_API__ttcrpy__tmesh
/* Early includes */
#include <string.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <string>
#include <vector>
#include <stdint.h>
#include <math.h>
#include "ttcr_t.h"
#include "Node3Dn.h"
#include "Node3Dnsp.h"
#include "Node3Dc.h"
#include "Node3Dcsp.h"
#include "Cell.h"
#include "Node2Dn.h"
#include "Node2Dc.h"
#include "Node2Dnsp.h"
#include "Node2Dcsp.h"
#include "Grid3D.h"
#include "Grid3Duc.h"
#include "Grid3Dun.h"
#include "Grid3Ducfs.h"
#include "Grid3Ducsp.h"
#include "Grid3Ducdsp.h"
#include "Grid3Dunfs.h"
#include "Grid3Dunsp.h"
#include "Grid3Dundsp.h"
#include "Grid2D.h"
#include "Grid2Duc.h"
#include "Grid2Dun.h"
#include "Grid2Ducsp.h"
#include "Grid2Ducfs.h"
#include "Grid2Dunsp.h"
#include "Grid2Dunfs.h"
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "verbose.h"
#include "utils_cython.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "ttcrpy/tmesh.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d;
struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "ttcrpy/tmesh.pyx":52
 * 
 * 
 * cdef class Mesh3d:             # <<<<<<<<<<<<<<
 *     """class to perform raytracing with tetrahedral meshes
 * 
 */
struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d {
  PyObject_HEAD
  bool cell_slowness;
  size_t _n_threads;
  char method;
  bool tt_from_rp;
  bool interp_vel;
  double eps;
  int maxit;
  int gradient_method;
  double min_dist;
  uint32_t n_secondary;
  uint32_t n_tertiary;
  double radius_tertiary;
  std::vector<ttcr::sxyz<double> >  no;
  std::vector<ttcr::tetrahedronElem<uint32_t> >  tet;
  ttcr::Grid3D<double,uint32_t>  *grid;
};


/* "ttcrpy/tmesh.pyx":1026
 * 
 * 
 * cdef class Mesh2d:             # <<<<<<<<<<<<<<
 *     """class to perform raytracing with triangular meshes
 * 
 */
struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d {
  PyObject_HEAD
  bool cell_slowness;
  bool process_obtuse;
  size_t _n_threads;
  double eps;
  int maxit;
  char method;
  uint32_t n_secondary;
  std::vector<ttcr::sxz<double> >  no;
  std::vector<ttcr::triangleElem<uint32_t> >  tri;
  ttcr::Grid2D<double,uint32_t,ttcr::sxz<double> >  *grid;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyUnicodeContains.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_ContainsTF(PyObject* substring, PyObject* text, int eq) {
    int result = PyUnicode_Contains(text, substring);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libcpp.string' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libcpp' */

/* Module declarations from 'ttcrpy.common' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'ttcrpy.tmesh' */
static PyTypeObject *__pyx_ptype_6ttcrpy_5tmesh_Mesh3d = 0;
static PyTypeObject *__pyx_ptype_6ttcrpy_5tmesh_Mesh2d = 0;
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_double_t = { "double_t", NULL, sizeof(__pyx_t_5numpy_double_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t = { "int64_t", NULL, sizeof(__pyx_t_5numpy_int64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int64_t), 0 };
#define __Pyx_MODULE_NAME "ttcrpy.tmesh"
extern int __pyx_module_is_main_ttcrpy__tmesh;
int __pyx_module_is_main_ttcrpy__tmesh = 0;

/* Implementation of 'ttcrpy.tmesh' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_nn[] = "nn";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_sp[] = "sp";
static const char __pyx_k_FSM[] = "FSM";
static const char __pyx_k_SPM[] = "SPM";
static const char __pyx_k__25[] = "_";
static const char __pyx_k__36[] = "*";
static const char __pyx_k_eps[] = "eps";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_nod[] = "nod";
static const char __pyx_k_rcv[] = "rcv";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_tet[] = "tet";
static const char __pyx_k_tri[] = "tri";
static const char __pyx_k_vtk[] = "vtk";
static const char __pyx_k_vtp[] = ".vtp";
static const char __pyx_k_vtu[] = ".vtu";
static const char __pyx_k_DSPM[] = "DSPM";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_hypo[] = "hypo";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_rays[] = "rays";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_GetId[] = "GetId";
static const char __pyx_k_SetId[] = "SetId";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_maxit[] = "maxit";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_tetra[] = "tetra";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_Mesh2d[] = "Mesh2d";
static const char __pyx_k_Mesh3d[] = "Mesh3d";
static const char __pyx_k_Update[] = "Update";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_unique[] = "unique";
static const char __pyx_k_GetCell[] = "GetCell";
static const char __pyx_k_GetData[] = "GetData";
static const char __pyx_k_SetName[] = "SetName";
static const char __pyx_k_builder[] = "builder";
static const char __pyx_k_lowness[] = "lowness";
static const char __pyx_k_nonzero[] = "nonzero";
static const char __pyx_k_nparams[] = "nparams";
static const char __pyx_k_squared[] = "squared";
static const char __pyx_k_AddArray[] = "AddArray";
static const char __pyx_k_GetArray[] = "GetArray";
static const char __pyx_k_HasArray[] = "HasArray";
static const char __pyx_k_SetLines[] = "SetLines";
static const char __pyx_k_Slowness[] = "Slowness";
static const char __pyx_k_Velocity[] = "Velocity";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_min_dist[] = "min_dist";
static const char __pyx_k_s0inside[] = "s0inside";
static const char __pyx_k_slowness[] = "slowness";
static const char __pyx_k_velocity[] = "velocity";
static const char __pyx_k_vtkTetra[] = "vtkTetra";
static const char __pyx_k_vtk_util[] = "vtk.util";
static const char __pyx_k_GetOutput[] = "GetOutput";
static const char __pyx_k_GetPoints[] = "GetPoints";
static const char __pyx_k_SetPoints[] = "SetPoints";
static const char __pyx_k_SetTuple1[] = "SetTuple1";
static const char __pyx_k_VTK_TETRA[] = "VTK_TETRA";
static const char __pyx_k_compute_L[] = "compute_L";
static const char __pyx_k_n_threads[] = "n_threads";
static const char __pyx_k_rebuild2d[] = "_rebuild2d";
static const char __pyx_k_rebuild3d[] = "_rebuild3d";
static const char __pyx_k_thread_no[] = "thread_no";
static const char __pyx_k_triangles[] = "triangles";
static const char __pyx_k_vtkPoints[] = "vtkPoints";
static const char __pyx_k_weighting[] = "weighting";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_csr_matrix[] = "csr_matrix";
static const char __pyx_k_interp_vel[] = "interp_vel";
static const char __pyx_k_is_outside[] = "is_outside";
static const char __pyx_k_n_tertiary[] = "n_tertiary";
static const char __pyx_k_tt_from_rp[] = "tt_from_rp";
static const char __pyx_k_GetCellData[] = "GetCellData";
static const char __pyx_k_GetCellType[] = "GetCellType";
static const char __pyx_k_GetPointIds[] = "GetPointIds";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_InsertPoint[] = "InsertPoint";
static const char __pyx_k_SetFileName[] = "SetFileName";
static const char __pyx_k_n_secondary[] = "n_secondary";
static const char __pyx_k_return_rays[] = "return_rays";
static const char __pyx_k_set_verbose[] = "set_verbose";
static const char __pyx_k_vtkPolyData[] = "vtkPolyData";
static const char __pyx_k_vtkPolyLine[] = "vtkPolyLine";
static const char __pyx_k_vtkTriangle[] = "vtkTriangle";
static const char __pyx_k_C_CONTIGUOUS[] = "C_CONTIGUOUS";
static const char __pyx_k_GetPointData[] = "GetPointData";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_SetInputData[] = "SetInputData";
static const char __pyx_k_VTK_TRIANGLE[] = "VTK_TRIANGLE";
static const char __pyx_k_scipy_sparse[] = "scipy.sparse";
static const char __pyx_k_set_slowness[] = "set_slowness";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_taylor_order[] = "taylor_order";
static const char __pyx_k_ttcrpy_tmesh[] = "ttcrpy.tmesh";
static const char __pyx_k_vtkCellArray[] = "vtkCellArray";
static const char __pyx_k_vtk_to_numpy[] = "vtk_to_numpy";
static const char __pyx_k_aggregate_src[] = "aggregate_src";
static const char __pyx_k_cell_slowness[] = "cell_slowness";
static const char __pyx_k_numpy_support[] = "numpy_support";
static const char __pyx_k_save_raypaths[] = "_save_raypaths";
static const char __pyx_k_InsertNextCell[] = "InsertNextCell";
static const char __pyx_k_SetNumberOfIds[] = "SetNumberOfIds";
static const char __pyx_k_process_obtuse[] = "process_obtuse";
static const char __pyx_k_vtkDoubleArray[] = "vtkDoubleArray";
static const char __pyx_k_P_wave_velocity[] = "P-wave velocity";
static const char __pyx_k_gradient_method[] = "gradient_method";
static const char __pyx_k_radius_tertiary[] = "radius_tertiary";
static const char __pyx_k_GetNumberOfCells[] = "GetNumberOfCells";
static const char __pyx_k_ttcrpy_tmesh_pyx[] = "ttcrpy/tmesh.pyx";
static const char __pyx_k_SetNumberOfPoints[] = "SetNumberOfPoints";
static const char __pyx_k_SetNumberOfTuples[] = "SetNumberOfTuples";
static const char __pyx_k_ascontiguousarray[] = "ascontiguousarray";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_constructor_params[] = "constructor_params";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_SetDataModeToBinary[] = "SetDataModeToBinary";
static const char __pyx_k_get_number_of_cells[] = "get_number_of_cells";
static const char __pyx_k_get_number_of_nodes[] = "get_number_of_nodes";
static const char __pyx_k_vtkUnstructuredGrid[] = "vtkUnstructuredGrid";
static const char __pyx_k_Method_0_s_undefined[] = "Method {0:s} undefined";
static const char __pyx_k_vtkXMLPolyDataWriter[] = "vtkXMLPolyDataWriter";
static const char __pyx_k_Receiver_outside_grid[] = "Receiver outside grid";
static const char __pyx_k_SetNumberOfComponents[] = "SetNumberOfComponents";
static const char __pyx_k_hypo_should_be_npts_x_5[] = "hypo should be npts x 5";
static const char __pyx_k_Source_point_outside_grid[] = "Source point outside grid";
static const char __pyx_k_Field_0_s_has_incorrect_size[] = "Field {0:s} has incorrect size";
static const char __pyx_k_vtkXMLUnstructuredGridReader[] = "vtkXMLUnstructuredGridReader";
static const char __pyx_k_vtkXMLUnstructuredGridWriter[] = "vtkXMLUnstructuredGridWriter";
static const char __pyx_k_Slowness_vector_has_wrong_size[] = "Slowness vector has wrong size";
static const char __pyx_k_velocity_vector_has_wrong_size[] = "velocity vector has wrong size";
static const char __pyx_k_0_s_should_only_contain_tetrahe[] = "{0:s} should only contain tetrahedra";
static const char __pyx_k_0_s_should_only_contain_triangl[] = "{0:s} should only contain triangles";
static const char __pyx_k_File_should_contain_slowness_or[] = "File should contain slowness or velocity data";
static const char __pyx_k_Raytracing_on_unstructured_tria[] = "\nRaytracing on unstructured triangular and tetrahedral meshes\n\nThis module contains two classes to perform traveltime computation and\nraytracing on unstructured meshes:\n    - `Mesh2d` for 2D media\n    - `Mesh3d` for 3D media\n\n    Two general algorithms are implemented\n        - the Shortest-Path Method\n        - the Fast-Sweeping Method\n    and `Mesh3d` also support the Dynamic Shortest-Path Method\n\n    Slowness model can be defined in two ways:\n        1) slowness constant within the voxels of the mesh (the default)\n        2) slowness defined at nodes of the mesh\n\n    This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )\n";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_src_and_rcv_should_be_ndata_x_2[] = "src and rcv should be ndata x 2";
static const char __pyx_k_src_and_rcv_should_be_ndata_x_3[] = "src and rcv should be ndata x 3";
static const char __pyx_k_Cannot_aggregate_source_with_DSP[] = "Cannot aggregate source with DSPM raytracing";
static const char __pyx_k_Thread_number_is_larger_than_num[] = "Thread number is larger than number of threads";
static const char __pyx_k_Velocity_data_point_outside_grid[] = "Velocity data point outside grid";
static const char __pyx_k_compute_L_not_implemented_for_me[] = "compute_L not implemented for mesh with slowness defined in cells";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_source_and_rcv_should_be_2D_arra[] = "source and rcv should be 2D arrays";
static const char __pyx_k_source_should_be_either_nsrc_x_2[] = "source should be either nsrc x 2 or 3";
static const char __pyx_k_source_should_be_either_nsrc_x_3[] = "source should be either nsrc x 3, 4 or 5";
static const char __pyx_k_src_and_rcv_should_be_of_equal_s[] = "src and rcv should be of equal size";
static PyObject *__pyx_kp_u_0_s_should_only_contain_tetrahe;
static PyObject *__pyx_kp_u_0_s_should_only_contain_triangl;
static PyObject *__pyx_n_s_AddArray;
static PyObject *__pyx_n_u_C_CONTIGUOUS;
static PyObject *__pyx_kp_u_Cannot_aggregate_source_with_DSP;
static PyObject *__pyx_n_u_DSPM;
static PyObject *__pyx_n_u_FSM;
static PyObject *__pyx_kp_u_Field_0_s_has_incorrect_size;
static PyObject *__pyx_kp_u_File_should_contain_slowness_or;
static PyObject *__pyx_n_s_GetArray;
static PyObject *__pyx_n_s_GetCell;
static PyObject *__pyx_n_s_GetCellData;
static PyObject *__pyx_n_s_GetCellType;
static PyObject *__pyx_n_s_GetData;
static PyObject *__pyx_n_s_GetId;
static PyObject *__pyx_n_s_GetNumberOfCells;
static PyObject *__pyx_n_s_GetOutput;
static PyObject *__pyx_n_s_GetPointData;
static PyObject *__pyx_n_s_GetPointIds;
static PyObject *__pyx_n_s_GetPoints;
static PyObject *__pyx_n_s_HasArray;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_InsertNextCell;
static PyObject *__pyx_n_s_InsertPoint;
static PyObject *__pyx_n_s_Mesh2d;
static PyObject *__pyx_n_s_Mesh3d;
static PyObject *__pyx_kp_u_Method_0_s_undefined;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_kp_u_P_wave_velocity;
static PyObject *__pyx_kp_u_Receiver_outside_grid;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_u_SPM;
static PyObject *__pyx_n_s_SetDataModeToBinary;
static PyObject *__pyx_n_s_SetFileName;
static PyObject *__pyx_n_s_SetId;
static PyObject *__pyx_n_s_SetInputData;
static PyObject *__pyx_n_s_SetLines;
static PyObject *__pyx_n_s_SetName;
static PyObject *__pyx_n_s_SetNumberOfComponents;
static PyObject *__pyx_n_s_SetNumberOfIds;
static PyObject *__pyx_n_s_SetNumberOfPoints;
static PyObject *__pyx_n_s_SetNumberOfTuples;
static PyObject *__pyx_n_s_SetPoints;
static PyObject *__pyx_n_s_SetTuple1;
static PyObject *__pyx_n_u_Slowness;
static PyObject *__pyx_kp_u_Slowness_vector_has_wrong_size;
static PyObject *__pyx_kp_u_Source_point_outside_grid;
static PyObject *__pyx_kp_u_Thread_number_is_larger_than_num;
static PyObject *__pyx_n_s_Update;
static PyObject *__pyx_n_s_VTK_TETRA;
static PyObject *__pyx_n_s_VTK_TRIANGLE;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_u_Velocity;
static PyObject *__pyx_kp_u_Velocity_data_point_outside_grid;
static PyObject *__pyx_n_u__25;
static PyObject *__pyx_n_s__36;
static PyObject *__pyx_n_s_aggregate_src;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_ascontiguousarray;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_builder;
static PyObject *__pyx_n_s_cell_slowness;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_compute_L;
static PyObject *__pyx_kp_u_compute_L_not_implemented_for_me;
static PyObject *__pyx_n_s_constructor_params;
static PyObject *__pyx_n_s_csr_matrix;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_eps;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_get_number_of_cells;
static PyObject *__pyx_n_s_get_number_of_nodes;
static PyObject *__pyx_n_s_gradient_method;
static PyObject *__pyx_n_s_hypo;
static PyObject *__pyx_kp_u_hypo_should_be_npts_x_5;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_interp_vel;
static PyObject *__pyx_n_s_is_outside;
static PyObject *__pyx_n_u_lowness;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_maxit;
static PyObject *__pyx_n_s_method;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_min_dist;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_n_secondary;
static PyObject *__pyx_n_s_n_tertiary;
static PyObject *__pyx_n_s_n_threads;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_nn;
static PyObject *__pyx_n_s_nod;
static PyObject *__pyx_n_s_nodes;
static PyObject *__pyx_n_s_nonzero;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nparams;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_numpy_support;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_n_s_process_obtuse;
static PyObject *__pyx_n_s_radius_tertiary;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rays;
static PyObject *__pyx_n_s_rcv;
static PyObject *__pyx_n_s_reader;
static PyObject *__pyx_n_s_rebuild2d;
static PyObject *__pyx_n_s_rebuild3d;
static PyObject *__pyx_n_s_return_rays;
static PyObject *__pyx_n_s_s0inside;
static PyObject *__pyx_n_s_save_raypaths;
static PyObject *__pyx_n_s_scipy_sparse;
static PyObject *__pyx_n_s_set_slowness;
static PyObject *__pyx_n_s_set_verbose;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_slowness;
static PyObject *__pyx_n_u_slowness;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_kp_u_source_and_rcv_should_be_2D_arra;
static PyObject *__pyx_kp_u_source_should_be_either_nsrc_x_2;
static PyObject *__pyx_kp_u_source_should_be_either_nsrc_x_3;
static PyObject *__pyx_n_s_sp;
static PyObject *__pyx_n_s_squared;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_ndata_x_2;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_ndata_x_3;
static PyObject *__pyx_kp_u_src_and_rcv_should_be_of_equal_s;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_taylor_order;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tet;
static PyObject *__pyx_n_s_tetra;
static PyObject *__pyx_n_s_thread_no;
static PyObject *__pyx_n_s_tri;
static PyObject *__pyx_n_s_triangles;
static PyObject *__pyx_n_s_tt_from_rp;
static PyObject *__pyx_n_s_ttcrpy_tmesh;
static PyObject *__pyx_kp_s_ttcrpy_tmesh_pyx;
static PyObject *__pyx_n_s_unique;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_u_velocity;
static PyObject *__pyx_kp_u_velocity_vector_has_wrong_size;
static PyObject *__pyx_n_s_vtk;
static PyObject *__pyx_n_s_vtkCellArray;
static PyObject *__pyx_n_s_vtkDoubleArray;
static PyObject *__pyx_n_s_vtkPoints;
static PyObject *__pyx_n_s_vtkPolyData;
static PyObject *__pyx_n_s_vtkPolyLine;
static PyObject *__pyx_n_s_vtkTetra;
static PyObject *__pyx_n_s_vtkTriangle;
static PyObject *__pyx_n_s_vtkUnstructuredGrid;
static PyObject *__pyx_n_s_vtkXMLPolyDataWriter;
static PyObject *__pyx_n_s_vtkXMLUnstructuredGridReader;
static PyObject *__pyx_n_s_vtkXMLUnstructuredGridWriter;
static PyObject *__pyx_n_s_vtk_to_numpy;
static PyObject *__pyx_n_s_vtk_util;
static PyObject *__pyx_kp_u_vtp;
static PyObject *__pyx_kp_u_vtu;
static PyObject *__pyx_n_s_weighting;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_6ttcrpy_5tmesh_set_verbose(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static int __pyx_pf_6ttcrpy_5tmesh_6Mesh3d___cinit__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyArrayObject *__pyx_v_nodes, PyArrayObject *__pyx_v_tetra, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, int __pyx_v_gradient_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, double __pyx_v_min_dist, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary); /* proto */
static void __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_2__dealloc__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_4__reduce__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_7nparams___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_6set_use_thread_pool(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_use_thread_pool); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_8is_outside(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyArrayObject *__pyx_v_pts); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_10get_number_of_nodes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_12get_number_of_cells(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_14get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_thread_no); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_16set_slowness(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_18set_velocity(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_velocity); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_20compute_D(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_coord); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_22compute_K(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_order, PyObject *__pyx_v_taylor_order, PyObject *__pyx_v_weighting, PyObject *__pyx_v_squared, PyObject *__pyx_v_s0inside); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_24get_s0(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_hypo, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_26raytrace(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_return_rays); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_28to_vtk(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_30_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_32builder(PyObject *__pyx_v_filename, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, int __pyx_v_gradient_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, double __pyx_v_min_dist, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary); /* proto */
static int __pyx_pf_6ttcrpy_5tmesh_6Mesh2d___cinit__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyArrayObject *__pyx_v_nodes, PyArrayObject *__pyx_v_triangles, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_process_obtuse, uint32_t __pyx_v_n_secondary); /* proto */
static void __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_2__dealloc__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_4__reduce__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_7nparams___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_6get_number_of_nodes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_8get_number_of_cells(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_10get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_thread_no); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_12set_slowness(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_slowness); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_14set_velocity(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_velocity); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_16raytrace(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_return_rays); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_18to_vtk(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_20_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_22builder(PyObject *__pyx_v_filename, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_process_obtuse, uint32_t __pyx_v_n_secondary); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_2_rebuild3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_constructor_params); /* proto */
static PyObject *__pyx_pf_6ttcrpy_5tmesh_4_rebuild2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_constructor_params); /* proto */
static PyObject *__pyx_tp_new_6ttcrpy_5tmesh_Mesh3d(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ttcrpy_5tmesh_Mesh2d(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_1_;
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_slice_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__10;
static PyObject *__pyx_slice__14;
static PyObject *__pyx_slice__29;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
/* Late includes */

/* "ttcrpy/tmesh.pyx":41
 *     int build_matrix_siv[T](size_t, size_t, vector[vector[siv[T]]]&, object)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """Set verbosity level for C++ code
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_1set_verbose(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_set_verbose[] = "Set verbosity level for C++ code\n\n    Parameters\n    ----------\n    v: int\n        verbosity level\n    ";
static PyMethodDef __pyx_mdef_6ttcrpy_5tmesh_1set_verbose = {"set_verbose", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_1set_verbose, METH_O, __pyx_doc_6ttcrpy_5tmesh_set_verbose};
static PyObject *__pyx_pw_6ttcrpy_5tmesh_1set_verbose(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_verbose (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_set_verbose(__pyx_self, ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_set_verbose(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_verbose", 0);

  /* "ttcrpy/tmesh.pyx":49
 *         verbosity level
 *     """
 *     setVerbose(v)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_v); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
  ttcr::setVerbose(__pyx_t_1);

  /* "ttcrpy/tmesh.pyx":41
 *     int build_matrix_siv[T](size_t, size_t, vector[vector[siv[T]]]&, object)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """Set verbosity level for C++ code
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.set_verbose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":122
 *     cdef Grid3D[double, uint32_t]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=2] tetra,
 *                   size_t n_threads=1, bool cell_slowness=1,
 */

/* Python wrapper */
static int __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_nodes = 0;
  PyArrayObject *__pyx_v_tetra = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  int __pyx_v_gradient_method;
  bool __pyx_v_tt_from_rp;
  bool __pyx_v_interp_vel;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  double __pyx_v_min_dist;
  uint32_t __pyx_v_n_secondary;
  uint32_t __pyx_v_n_tertiary;
  double __pyx_v_radius_tertiary;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,&__pyx_n_s_tetra,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_gradient_method,&__pyx_n_s_tt_from_rp,&__pyx_n_s_interp_vel,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_min_dist,&__pyx_n_s_n_secondary,&__pyx_n_s_n_tertiary,&__pyx_n_s_radius_tertiary,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[4] = ((PyObject*)__pyx_n_u_FSM);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tetra)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 14, 1); __PYX_ERR(0, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gradient_method);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt_from_rp);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interp_vel);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_dist);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_tertiary);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius_tertiary);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = ((PyArrayObject *)values[0]);
    __pyx_v_tetra = ((PyArrayObject *)values[1]);
    if (values[2]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[3]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[4]);
    if (values[5]) {
      __pyx_v_gradient_method = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_gradient_method == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 125, __pyx_L3_error)
    } else {
      __pyx_v_gradient_method = ((int)1);
    }
    if (values[6]) {
      __pyx_v_tt_from_rp = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_tt_from_rp == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)
    } else {
      __pyx_v_tt_from_rp = ((bool)1);
    }
    if (values[7]) {
      __pyx_v_interp_vel = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_interp_vel == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)
    } else {
      __pyx_v_interp_vel = ((bool)0);
    }
    if (values[8]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[9]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[10]) {
      __pyx_v_min_dist = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_min_dist == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)
    } else {
      __pyx_v_min_dist = ((double)1.e-5);
    }
    if (values[11]) {
      __pyx_v_n_secondary = __Pyx_PyInt_As_uint32_t(values[11]); if (unlikely((__pyx_v_n_secondary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    } else {
      __pyx_v_n_secondary = ((uint32_t)2);
    }
    if (values[12]) {
      __pyx_v_n_tertiary = __Pyx_PyInt_As_uint32_t(values[12]); if (unlikely((__pyx_v_n_tertiary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    } else {
      __pyx_v_n_tertiary = ((uint32_t)2);
    }
    if (values[13]) {
      __pyx_v_radius_tertiary = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_radius_tertiary == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
    } else {
      __pyx_v_radius_tertiary = ((double)1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nodes), __pyx_ptype_5numpy_ndarray, 1, "nodes", 0))) __PYX_ERR(0, 122, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tetra), __pyx_ptype_5numpy_ndarray, 1, "tetra", 0))) __PYX_ERR(0, 123, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 125, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d___cinit__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_nodes, __pyx_v_tetra, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_gradient_method, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_eps, __pyx_v_maxit, __pyx_v_min_dist, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_radius_tertiary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ttcrpy_5tmesh_6Mesh3d___cinit__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyArrayObject *__pyx_v_nodes, PyArrayObject *__pyx_v_tetra, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, int __pyx_v_gradient_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, double __pyx_v_min_dist, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary) {
  double __pyx_v_source_radius;
  std::vector<ttcr::sxyz<double> >  __pyx_v_pts_ref;
  int __pyx_v_n;
  PyObject *__pyx_v_xmin = NULL;
  PyObject *__pyx_v_xmax = NULL;
  PyObject *__pyx_v_ymin = NULL;
  PyObject *__pyx_v_ymax = NULL;
  PyObject *__pyx_v_zmin = NULL;
  PyObject *__pyx_v_zmax = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_nodes;
  __Pyx_Buffer __pyx_pybuffer_nodes;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tetra;
  __Pyx_Buffer __pyx_pybuffer_tetra;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  ttcr::sxyz<double>  __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  ttcr::tetrahedronElem<uint32_t>  __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  double __pyx_t_26;
  double __pyx_t_27;
  double __pyx_t_28;
  ttcr::Grid3Ducfs<double,uint32_t>  *__pyx_t_29;
  ttcr::Grid3Ducsp<double,uint32_t>  *__pyx_t_30;
  ttcr::Grid3Ducdsp<double,uint32_t>  *__pyx_t_31;
  ttcr::Grid3Dunfs<double,uint32_t>  *__pyx_t_32;
  ttcr::Grid3Dunsp<double,uint32_t>  *__pyx_t_33;
  ttcr::Grid3Dundsp<double,uint32_t>  *__pyx_t_34;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_nodes);
  __Pyx_INCREF((PyObject *)__pyx_v_tetra);
  __pyx_pybuffer_nodes.pybuffer.buf = NULL;
  __pyx_pybuffer_nodes.refcount = 0;
  __pyx_pybuffernd_nodes.data = NULL;
  __pyx_pybuffernd_nodes.rcbuffer = &__pyx_pybuffer_nodes;
  __pyx_pybuffer_tetra.pybuffer.buf = NULL;
  __pyx_pybuffer_tetra.refcount = 0;
  __pyx_pybuffernd_tetra.data = NULL;
  __pyx_pybuffernd_tetra.rcbuffer = &__pyx_pybuffer_tetra;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_v_nodes, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 122, __pyx_L1_error)
  }
  __pyx_pybuffernd_nodes.diminfo[0].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nodes.diminfo[0].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_nodes.diminfo[1].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_nodes.diminfo[1].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer, (PyObject*)__pyx_v_tetra, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 122, __pyx_L1_error)
  }
  __pyx_pybuffernd_tetra.diminfo[0].strides = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tetra.diminfo[0].shape = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tetra.diminfo[1].strides = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tetra.diminfo[1].shape = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.shape[1];

  /* "ttcrpy/tmesh.pyx":131
 *                   double radius_tertiary=1.0):
 * 
 *         self.cell_slowness = cell_slowness             # <<<<<<<<<<<<<<
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp
 */
  __pyx_v_self->cell_slowness = __pyx_v_cell_slowness;

  /* "ttcrpy/tmesh.pyx":132
 * 
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads             # <<<<<<<<<<<<<<
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel
 */
  __pyx_v_self->_n_threads = __pyx_v_n_threads;

  /* "ttcrpy/tmesh.pyx":133
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp             # <<<<<<<<<<<<<<
 *         self.interp_vel = interp_vel
 *         self.eps = eps
 */
  __pyx_v_self->tt_from_rp = __pyx_v_tt_from_rp;

  /* "ttcrpy/tmesh.pyx":134
 *         self._n_threads = n_threads
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel             # <<<<<<<<<<<<<<
 *         self.eps = eps
 *         self.maxit = maxit
 */
  __pyx_v_self->interp_vel = __pyx_v_interp_vel;

  /* "ttcrpy/tmesh.pyx":135
 *         self.tt_from_rp = tt_from_rp
 *         self.interp_vel = interp_vel
 *         self.eps = eps             # <<<<<<<<<<<<<<
 *         self.maxit = maxit
 *         self.gradient_method = gradient_method
 */
  __pyx_v_self->eps = __pyx_v_eps;

  /* "ttcrpy/tmesh.pyx":136
 *         self.interp_vel = interp_vel
 *         self.eps = eps
 *         self.maxit = maxit             # <<<<<<<<<<<<<<
 *         self.gradient_method = gradient_method
 *         self.min_dist = min_dist
 */
  __pyx_v_self->maxit = __pyx_v_maxit;

  /* "ttcrpy/tmesh.pyx":137
 *         self.eps = eps
 *         self.maxit = maxit
 *         self.gradient_method = gradient_method             # <<<<<<<<<<<<<<
 *         self.min_dist = min_dist
 *         self.n_secondary = n_secondary
 */
  __pyx_v_self->gradient_method = __pyx_v_gradient_method;

  /* "ttcrpy/tmesh.pyx":138
 *         self.maxit = maxit
 *         self.gradient_method = gradient_method
 *         self.min_dist = min_dist             # <<<<<<<<<<<<<<
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary
 */
  __pyx_v_self->min_dist = __pyx_v_min_dist;

  /* "ttcrpy/tmesh.pyx":139
 *         self.gradient_method = gradient_method
 *         self.min_dist = min_dist
 *         self.n_secondary = n_secondary             # <<<<<<<<<<<<<<
 *         self.n_tertiary = n_tertiary
 *         self.radius_tertiary = radius_tertiary
 */
  __pyx_v_self->n_secondary = __pyx_v_n_secondary;

  /* "ttcrpy/tmesh.pyx":140
 *         self.min_dist = min_dist
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary             # <<<<<<<<<<<<<<
 *         self.radius_tertiary = radius_tertiary
 * 
 */
  __pyx_v_self->n_tertiary = __pyx_v_n_tertiary;

  /* "ttcrpy/tmesh.pyx":141
 *         self.n_secondary = n_secondary
 *         self.n_tertiary = n_tertiary
 *         self.radius_tertiary = radius_tertiary             # <<<<<<<<<<<<<<
 * 
 *         cdef double source_radius = 0.0
 */
  __pyx_v_self->radius_tertiary = __pyx_v_radius_tertiary;

  /* "ttcrpy/tmesh.pyx":143
 *         self.radius_tertiary = radius_tertiary
 * 
 *         cdef double source_radius = 0.0             # <<<<<<<<<<<<<<
 *         cdef vector[sxyz[double]] pts_ref
 *         cdef int n
 */
  __pyx_v_source_radius = 0.0;

  /* "ttcrpy/tmesh.pyx":147
 *         cdef int n
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             nodes = np.ascontiguousarray(nodes)
 *         if not tetra.flags['C_CONTIGUOUS']:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nodes), __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":148
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)             # <<<<<<<<<<<<<<
 *         if not tetra.flags['C_CONTIGUOUS']:
 *             tetra = np.ascontiguousarray(tetra)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, ((PyObject *)__pyx_v_nodes)) : __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_nodes));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_t_6 = ((PyArrayObject *)__pyx_t_2);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_v_nodes, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        }
        __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;
      }
      __pyx_pybuffernd_nodes.diminfo[0].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nodes.diminfo[0].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_nodes.diminfo[1].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_nodes.diminfo[1].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[1];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 148, __pyx_L1_error)
    }
    __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_nodes, ((PyArrayObject *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":147
 *         cdef int n
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             nodes = np.ascontiguousarray(nodes)
 *         if not tetra.flags['C_CONTIGUOUS']:
 */
  }

  /* "ttcrpy/tmesh.pyx":149
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)
 *         if not tetra.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             tetra = np.ascontiguousarray(tetra)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_tetra), __pyx_n_s_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  if (__pyx_t_3) {

    /* "ttcrpy/tmesh.pyx":150
 *             nodes = np.ascontiguousarray(nodes)
 *         if not tetra.flags['C_CONTIGUOUS']:
 *             tetra = np.ascontiguousarray(tetra)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nodes.shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, ((PyObject *)__pyx_v_tetra)) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_v_tetra));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 150, __pyx_L1_error)
    __pyx_t_11 = ((PyArrayObject *)__pyx_t_5);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer, (PyObject*)__pyx_v_tetra, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        }
        __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;
      }
      __pyx_pybuffernd_tetra.diminfo[0].strides = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tetra.diminfo[0].shape = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tetra.diminfo[1].strides = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tetra.diminfo[1].shape = __pyx_pybuffernd_tetra.rcbuffer->pybuffer.shape[1];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 150, __pyx_L1_error)
    }
    __pyx_t_11 = 0;
    __Pyx_DECREF_SET(__pyx_v_tetra, ((PyArrayObject *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":149
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)
 *         if not tetra.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             tetra = np.ascontiguousarray(tetra)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":152
 *             tetra = np.ascontiguousarray(tetra)
 * 
 *         for n in range(nodes.shape[0]):             # <<<<<<<<<<<<<<
 *             self.no.push_back(sxyz[double](nodes[n, 0],
 *                                            nodes[n, 1],
 */
  __pyx_t_12 = (__pyx_v_nodes->dimensions[0]);
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":153
 * 
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxyz[double](nodes[n, 0],             # <<<<<<<<<<<<<<
 *                                            nodes[n, 1],
 *                                            nodes[n, 2]))
 */
    __pyx_t_14 = __pyx_v_n;
    __pyx_t_15 = 0;
    __pyx_t_16 = -1;
    if (__pyx_t_14 < 0) {
      __pyx_t_14 += __pyx_pybuffernd_nodes.diminfo[0].shape;
      if (unlikely(__pyx_t_14 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_14 >= __pyx_pybuffernd_nodes.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_15 < 0) {
      __pyx_t_15 += __pyx_pybuffernd_nodes.diminfo[1].shape;
      if (unlikely(__pyx_t_15 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_15 >= __pyx_pybuffernd_nodes.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 153, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":154
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxyz[double](nodes[n, 0],
 *                                            nodes[n, 1],             # <<<<<<<<<<<<<<
 *                                            nodes[n, 2]))
 *         for n in range(tetra.shape[0]):
 */
    __pyx_t_17 = __pyx_v_n;
    __pyx_t_18 = 1;
    __pyx_t_16 = -1;
    if (__pyx_t_17 < 0) {
      __pyx_t_17 += __pyx_pybuffernd_nodes.diminfo[0].shape;
      if (unlikely(__pyx_t_17 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_nodes.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_18 < 0) {
      __pyx_t_18 += __pyx_pybuffernd_nodes.diminfo[1].shape;
      if (unlikely(__pyx_t_18 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_nodes.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 154, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":155
 *             self.no.push_back(sxyz[double](nodes[n, 0],
 *                                            nodes[n, 1],
 *                                            nodes[n, 2]))             # <<<<<<<<<<<<<<
 *         for n in range(tetra.shape[0]):
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],
 */
    __pyx_t_19 = __pyx_v_n;
    __pyx_t_20 = 2;
    __pyx_t_16 = -1;
    if (__pyx_t_19 < 0) {
      __pyx_t_19 += __pyx_pybuffernd_nodes.diminfo[0].shape;
      if (unlikely(__pyx_t_19 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_19 >= __pyx_pybuffernd_nodes.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_20 < 0) {
      __pyx_t_20 += __pyx_pybuffernd_nodes.diminfo[1].shape;
      if (unlikely(__pyx_t_20 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_20 >= __pyx_pybuffernd_nodes.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 155, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":153
 * 
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxyz[double](nodes[n, 0],             # <<<<<<<<<<<<<<
 *                                            nodes[n, 1],
 *                                            nodes[n, 2]))
 */
    try {
      __pyx_t_21 = ttcr::sxyz<double> ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_nodes.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_nodes.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_nodes.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_nodes.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_nodes.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_nodes.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_nodes.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_nodes.diminfo[0].strides, __pyx_t_20, __pyx_pybuffernd_nodes.diminfo[1].strides)));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 153, __pyx_L1_error)
    }
    try {
      __pyx_v_self->no.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 153, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":156
 *                                            nodes[n, 1],
 *                                            nodes[n, 2]))
 *         for n in range(tetra.shape[0]):             # <<<<<<<<<<<<<<
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],
 *                                                          tetra[n, 1],
 */
  __pyx_t_12 = (__pyx_v_tetra->dimensions[0]);
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":157
 *                                            nodes[n, 2]))
 *         for n in range(tetra.shape[0]):
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],             # <<<<<<<<<<<<<<
 *                                                          tetra[n, 1],
 *                                                          tetra[n, 2],
 */
    __pyx_t_20 = __pyx_v_n;
    __pyx_t_19 = 0;
    __pyx_t_16 = -1;
    if (__pyx_t_20 < 0) {
      __pyx_t_20 += __pyx_pybuffernd_tetra.diminfo[0].shape;
      if (unlikely(__pyx_t_20 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_20 >= __pyx_pybuffernd_tetra.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_19 < 0) {
      __pyx_t_19 += __pyx_pybuffernd_tetra.diminfo[1].shape;
      if (unlikely(__pyx_t_19 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_19 >= __pyx_pybuffernd_tetra.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 157, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":158
 *         for n in range(tetra.shape[0]):
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],
 *                                                          tetra[n, 1],             # <<<<<<<<<<<<<<
 *                                                          tetra[n, 2],
 *                                                          tetra[n, 3]))
 */
    __pyx_t_18 = __pyx_v_n;
    __pyx_t_17 = 1;
    __pyx_t_16 = -1;
    if (__pyx_t_18 < 0) {
      __pyx_t_18 += __pyx_pybuffernd_tetra.diminfo[0].shape;
      if (unlikely(__pyx_t_18 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_tetra.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_17 < 0) {
      __pyx_t_17 += __pyx_pybuffernd_tetra.diminfo[1].shape;
      if (unlikely(__pyx_t_17 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_tetra.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 158, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":159
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],
 *                                                          tetra[n, 1],
 *                                                          tetra[n, 2],             # <<<<<<<<<<<<<<
 *                                                          tetra[n, 3]))
 * 
 */
    __pyx_t_15 = __pyx_v_n;
    __pyx_t_14 = 2;
    __pyx_t_16 = -1;
    if (__pyx_t_15 < 0) {
      __pyx_t_15 += __pyx_pybuffernd_tetra.diminfo[0].shape;
      if (unlikely(__pyx_t_15 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_15 >= __pyx_pybuffernd_tetra.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_14 < 0) {
      __pyx_t_14 += __pyx_pybuffernd_tetra.diminfo[1].shape;
      if (unlikely(__pyx_t_14 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_14 >= __pyx_pybuffernd_tetra.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 159, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":160
 *                                                          tetra[n, 1],
 *                                                          tetra[n, 2],
 *                                                          tetra[n, 3]))             # <<<<<<<<<<<<<<
 * 
 *         if method == 'FSM':
 */
    __pyx_t_22 = __pyx_v_n;
    __pyx_t_23 = 3;
    __pyx_t_16 = -1;
    if (__pyx_t_22 < 0) {
      __pyx_t_22 += __pyx_pybuffernd_tetra.diminfo[0].shape;
      if (unlikely(__pyx_t_22 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_22 >= __pyx_pybuffernd_tetra.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_23 < 0) {
      __pyx_t_23 += __pyx_pybuffernd_tetra.diminfo[1].shape;
      if (unlikely(__pyx_t_23 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_23 >= __pyx_pybuffernd_tetra.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 160, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":157
 *                                            nodes[n, 2]))
 *         for n in range(tetra.shape[0]):
 *             self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],             # <<<<<<<<<<<<<<
 *                                                          tetra[n, 1],
 *                                                          tetra[n, 2],
 */
    try {
      __pyx_t_24 = ttcr::tetrahedronElem<uint32_t> ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_tetra.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_tetra.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_tetra.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_tetra.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_tetra.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_tetra.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_tetra.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_tetra.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_tetra.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_tetra.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_tetra.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_tetra.diminfo[1].strides)));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 157, __pyx_L1_error)
    }
    try {
      __pyx_v_self->tet.push_back(__pyx_t_24);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 157, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":162
 *                                                          tetra[n, 3]))
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 */
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 162, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":163
 * 
 *         if method == 'FSM':
 *             xmin = np.min(nodes[:, 0])             # <<<<<<<<<<<<<<
 *             xmax = np.max(nodes[:, 0])
 *             ymin = np.min(nodes[:, 1])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_25, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_xmin = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":164
 *         if method == 'FSM':
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])             # <<<<<<<<<<<<<<
 *             ymin = np.min(nodes[:, 1])
 *             ymax = np.max(nodes[:, 1])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_25, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_xmax = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":165
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 *             ymin = np.min(nodes[:, 1])             # <<<<<<<<<<<<<<
 *             ymax = np.max(nodes[:, 1])
 *             zmin = np.min(nodes[:, 2])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_25, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_ymin = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":166
 *             xmax = np.max(nodes[:, 0])
 *             ymin = np.min(nodes[:, 1])
 *             ymax = np.max(nodes[:, 1])             # <<<<<<<<<<<<<<
 *             zmin = np.min(nodes[:, 2])
 *             zmax = np.max(nodes[:, 2])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_25, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ymax = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":167
 *             ymin = np.min(nodes[:, 1])
 *             ymax = np.max(nodes[:, 1])
 *             zmin = np.min(nodes[:, 2])             # <<<<<<<<<<<<<<
 *             zmax = np.max(nodes[:, 2])
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_25, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_zmin = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":168
 *             ymax = np.max(nodes[:, 1])
 *             zmin = np.min(nodes[:, 2])
 *             zmax = np.max(nodes[:, 2])             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmax))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_25, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_zmax = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":169
 *             zmin = np.min(nodes[:, 2])
 *             zmax = np.max(nodes[:, 2])
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmin))
 */
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 169, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymin); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 169, __pyx_L1_error)
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 169, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_26, __pyx_t_27, __pyx_t_28);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 169, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 169, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":170
 *             zmax = np.max(nodes[:, 2])
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmax))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmax))
 */
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymin); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L1_error)
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_28, __pyx_t_27, __pyx_t_26);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 170, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 170, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":171
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmin))
 */
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymax); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_26, __pyx_t_27, __pyx_t_28);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 171, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":172
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmax))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmax))
 */
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymax); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_28, __pyx_t_27, __pyx_t_26);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 172, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 172, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":173
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmin))
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmin))
 */
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymin); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_26, __pyx_t_27, __pyx_t_28);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 173, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 173, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":174
 *             pts_ref.push_back(sxyz[double](xmin, ymax, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmax))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmin))
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmax))
 */
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymin); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L1_error)
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_28, __pyx_t_27, __pyx_t_26);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 174, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 174, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":175
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmin))
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmax))
 * 
 */
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymax); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_26, __pyx_t_27, __pyx_t_28);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 175, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 175, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":176
 *             pts_ref.push_back(sxyz[double](xmax, ymin, zmax))
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmin))
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmax))             # <<<<<<<<<<<<<<
 * 
 *         if cell_slowness:
 */
    __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)
    __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_ymax); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)
    __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)
    try {
      __pyx_t_21 = ttcr::sxyz<double> (__pyx_t_28, __pyx_t_27, __pyx_t_26);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 176, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_21);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 176, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":162
 *                                                          tetra[n, 3]))
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 */
  }

  /* "ttcrpy/tmesh.pyx":178
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmax))
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
  __pyx_t_4 = (__pyx_v_cell_slowness != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":179
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Ducfs[double,uint32_t](self.no, self.tet,
 */
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 179, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "ttcrpy/tmesh.pyx":180
 *         if cell_slowness:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Ducfs[double,uint32_t](self.no, self.tet,
 *                                                             eps, maxit,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/tmesh.pyx":181
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid3Ducfs[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                             eps, maxit,
 *                                                             pts_ref, 2,
 */
      try {
        __pyx_t_29 = new ttcr::Grid3Ducfs<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_eps, __pyx_v_maxit, __pyx_v_pts_ref, 2, __pyx_v_gradient_method, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 181, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_29;

      /* "ttcrpy/tmesh.pyx":179
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Ducfs[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":187
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Ducsp[double,uint32_t](self.no, self.tet,
 */
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 187, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":188
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Ducsp[double,uint32_t](self.no, self.tet,
 *                                                             n_secondary,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/tmesh.pyx":189
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid3Ducsp[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                             n_secondary,
 *                                                             tt_from_rp,
 */
      try {
        __pyx_t_30 = new ttcr::Grid3Ducsp<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_n_secondary, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 189, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_30;

      /* "ttcrpy/tmesh.pyx":187
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Ducsp[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":193
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Ducdsp[double,uint32_t](self.no, self.tet,
 */
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_DSPM, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (likely(__pyx_t_3)) {

      /* "ttcrpy/tmesh.pyx":194
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':
 *                 self.method = b'd'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Ducdsp[double,uint32_t](self.no, self.tet,
 *                                                              n_secondary,
 */
      __pyx_v_self->method = 'd';

      /* "ttcrpy/tmesh.pyx":195
 *             elif method == 'DSPM':
 *                 self.method = b'd'
 *                 self.grid = new Grid3Ducdsp[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                              n_secondary,
 *                                                              n_tertiary,
 */
      try {
        __pyx_t_31 = new ttcr::Grid3Ducdsp<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_source_radius, __pyx_v_gradient_method, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_radius_tertiary, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 195, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_31;

      /* "ttcrpy/tmesh.pyx":193
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Ducdsp[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":206
 * 
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 *         else:
 *             if method == 'FSM':
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 206, __pyx_L1_error)
    }
    __pyx_L11:;

    /* "ttcrpy/tmesh.pyx":178
 *             pts_ref.push_back(sxyz[double](xmax, ymax, zmax))
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
    goto __pyx_L10;
  }

  /* "ttcrpy/tmesh.pyx":208
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Dunfs[double,uint32_t](self.no, self.tet,
 */
  /*else*/ {
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 208, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":209
 *         else:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Dunfs[double,uint32_t](self.no, self.tet,
 *                                                             eps, maxit,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/tmesh.pyx":210
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid3Dunfs[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                             eps, maxit,
 *                                                             pts_ref, 2,
 */
      try {
        __pyx_t_32 = new ttcr::Grid3Dunfs<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_eps, __pyx_v_maxit, __pyx_v_pts_ref, 2, __pyx_v_gradient_method, __pyx_v_interp_vel, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 210, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_32;

      /* "ttcrpy/tmesh.pyx":208
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid3Dunfs[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L12;
    }

    /* "ttcrpy/tmesh.pyx":217
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Dunsp[double,uint32_t](self.no, self.tet,
 */
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 217, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "ttcrpy/tmesh.pyx":218
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Dunsp[double,uint32_t](self.no, self.tet,
 *                                                             n_secondary,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/tmesh.pyx":219
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid3Dunsp[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                             n_secondary,
 *                                                             interp_vel,
 */
      try {
        __pyx_t_33 = new ttcr::Grid3Dunsp<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_n_secondary, __pyx_v_interp_vel, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 219, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_33;

      /* "ttcrpy/tmesh.pyx":217
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid3Dunsp[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L12;
    }

    /* "ttcrpy/tmesh.pyx":224
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Dundsp[double,uint32_t](self.no, self.tet,
 */
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_DSPM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 224, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (likely(__pyx_t_4)) {

      /* "ttcrpy/tmesh.pyx":225
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':
 *                 self.method = b'd'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid3Dundsp[double,uint32_t](self.no, self.tet,
 *                                                              n_secondary,
 */
      __pyx_v_self->method = 'd';

      /* "ttcrpy/tmesh.pyx":226
 *             elif method == 'DSPM':
 *                 self.method = b'd'
 *                 self.grid = new Grid3Dundsp[double,uint32_t](self.no, self.tet,             # <<<<<<<<<<<<<<
 *                                                              n_secondary,
 *                                                              n_tertiary,
 */
      try {
        __pyx_t_34 = new ttcr::Grid3Dundsp<double,uint32_t> (__pyx_v_self->no, __pyx_v_self->tet, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_source_radius, __pyx_v_interp_vel, __pyx_v_gradient_method, __pyx_v_tt_from_rp, __pyx_v_min_dist, __pyx_v_radius_tertiary, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 226, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_34;

      /* "ttcrpy/tmesh.pyx":224
 *                                                             tt_from_rp,
 *                                                             min_dist, n_threads)
 *             elif method == 'DSPM':             # <<<<<<<<<<<<<<
 *                 self.method = b'd'
 *                 self.grid = new Grid3Dundsp[double,uint32_t](self.no, self.tet,
 */
      goto __pyx_L12;
    }

    /* "ttcrpy/tmesh.pyx":237
 *                                                              n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 237, __pyx_L1_error)
    }
    __pyx_L12:;
  }
  __pyx_L10:;

  /* "ttcrpy/tmesh.pyx":122
 *     cdef Grid3D[double, uint32_t]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=2] tetra,
 *                   size_t n_threads=1, bool cell_slowness=1,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_25);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tetra.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_xmin);
  __Pyx_XDECREF(__pyx_v_xmax);
  __Pyx_XDECREF(__pyx_v_ymin);
  __Pyx_XDECREF(__pyx_v_ymax);
  __Pyx_XDECREF(__pyx_v_zmin);
  __Pyx_XDECREF(__pyx_v_zmax);
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_tetra);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":239
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

/* Python wrapper */
static void __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_2__dealloc__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_2__dealloc__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ttcrpy/tmesh.pyx":240
 * 
 *     def __dealloc__(self):
 *         del self.grid             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  delete __pyx_v_self->grid;

  /* "ttcrpy/tmesh.pyx":239
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ttcrpy/tmesh.pyx":242
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_4__reduce__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_4__reduce__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_tetra = NULL;
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_constructor_params = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ttcrpy/tmesh.pyx":243
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
  switch (__pyx_v_self->method) {
    case 'f':

    /* "ttcrpy/tmesh.pyx":244
 *     def __reduce__(self):
 *         if self.method == b'f':
 *             method = 'FSM'             # <<<<<<<<<<<<<<
 *         elif self.method == b's':
 *             method = 'SPM'
 */
    __Pyx_INCREF(__pyx_n_u_FSM);
    __pyx_v_method = __pyx_n_u_FSM;

    /* "ttcrpy/tmesh.pyx":243
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
    break;
    case 's':

    /* "ttcrpy/tmesh.pyx":246
 *             method = 'FSM'
 *         elif self.method == b's':
 *             method = 'SPM'             # <<<<<<<<<<<<<<
 *         elif self.method == b'd':
 *             method = 'DSPM'
 */
    __Pyx_INCREF(__pyx_n_u_SPM);
    __pyx_v_method = __pyx_n_u_SPM;

    /* "ttcrpy/tmesh.pyx":245
 *         if self.method == b'f':
 *             method = 'FSM'
 *         elif self.method == b's':             # <<<<<<<<<<<<<<
 *             method = 'SPM'
 *         elif self.method == b'd':
 */
    break;
    case 'd':

    /* "ttcrpy/tmesh.pyx":248
 *             method = 'SPM'
 *         elif self.method == b'd':
 *             method = 'DSPM'             # <<<<<<<<<<<<<<
 * 
 *         nodes = np.ndarray((self.no.size(), 3))
 */
    __Pyx_INCREF(__pyx_n_u_DSPM);
    __pyx_v_method = __pyx_n_u_DSPM;

    /* "ttcrpy/tmesh.pyx":247
 *         elif self.method == b's':
 *             method = 'SPM'
 *         elif self.method == b'd':             # <<<<<<<<<<<<<<
 *             method = 'DSPM'
 * 
 */
    break;
    default: break;
  }

  /* "ttcrpy/tmesh.pyx":250
 *             method = 'DSPM'
 * 
 *         nodes = np.ndarray((self.no.size(), 3))             # <<<<<<<<<<<<<<
 *         tetra = np.ndarray((self.tet.size(), 4), dtype=int)
 *         cdef int n
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":251
 * 
 *         nodes = np.ndarray((self.no.size(), 3))
 *         tetra = np.ndarray((self.tet.size(), 4), dtype=int)             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef int nn
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tet.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_4);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tetra = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":254
 *         cdef int n
 *         cdef int nn
 *         for n in range(nodes.shape[0]):             # <<<<<<<<<<<<<<
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].y
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_nodes, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":255
 *         cdef int nn
 *         for n in range(nodes.shape[0]):
 *             nodes[n, 0] = self.no[n].x             # <<<<<<<<<<<<<<
 *             nodes[n, 1] = self.no[n].y
 *             nodes[n, 2] = self.no[n].z
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_0);
    __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_nodes, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":256
 *         for n in range(nodes.shape[0]):
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].y             # <<<<<<<<<<<<<<
 *             nodes[n, 2] = self.no[n].z
 *         for n in range(tetra.shape[0]):
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
    __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_nodes, __pyx_t_3, __pyx_t_2) < 0)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":257
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].y
 *             nodes[n, 2] = self.no[n].z             # <<<<<<<<<<<<<<
 *         for n in range(tetra.shape[0]):
 *             for nn in range(4):
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).z); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_2);
    __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_nodes, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":258
 *             nodes[n, 1] = self.no[n].y
 *             nodes[n, 2] = self.no[n].z
 *         for n in range(tetra.shape[0]):             # <<<<<<<<<<<<<<
 *             for nn in range(4):
 *                 tetra[n, nn] = self.tet[n].i[nn]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tetra, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":259
 *             nodes[n, 2] = self.no[n].z
 *         for n in range(tetra.shape[0]):
 *             for nn in range(4):             # <<<<<<<<<<<<<<
 *                 tetra[n, nn] = self.tet[n].i[nn]
 * 
 */
    for (__pyx_t_7 = 0; __pyx_t_7 < 4; __pyx_t_7+=1) {
      __pyx_v_nn = __pyx_t_7;

      /* "ttcrpy/tmesh.pyx":260
 *         for n in range(tetra.shape[0]):
 *             for nn in range(4):
 *                 tetra[n, nn] = self.tet[n].i[nn]             # <<<<<<<<<<<<<<
 * 
 *         constructor_params = (nodes, tetra, method, self.cell_slowness,
 */
      __pyx_t_1 = __Pyx_PyInt_From_uint32_t(((__pyx_v_self->tet[__pyx_v_n]).i[__pyx_v_nn])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tetra, __pyx_t_8, __pyx_t_1) < 0)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":262
 *                 tetra[n, nn] = self.tet[n].i[nn]
 * 
 *         constructor_params = (nodes, tetra, method, self.cell_slowness,             # <<<<<<<<<<<<<<
 *                               self._n_threads, self.tt_from_rp, self.interp_vel,
 *                               self.eps, self.maxit, self.gradient_method,
 */
  if (unlikely(!__pyx_v_method)) { __Pyx_RaiseUnboundLocalError("method"); __PYX_ERR(0, 262, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->cell_slowness); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "ttcrpy/tmesh.pyx":263
 * 
 *         constructor_params = (nodes, tetra, method, self.cell_slowness,
 *                               self._n_threads, self.tt_from_rp, self.interp_vel,             # <<<<<<<<<<<<<<
 *                               self.eps, self.maxit, self.gradient_method,
 *                               self.min_dist, self.n_secondary, self.n_tertiary,
 */
  __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->tt_from_rp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->interp_vel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "ttcrpy/tmesh.pyx":264
 *         constructor_params = (nodes, tetra, method, self.cell_slowness,
 *                               self._n_threads, self.tt_from_rp, self.interp_vel,
 *                               self.eps, self.maxit, self.gradient_method,             # <<<<<<<<<<<<<<
 *                               self.min_dist, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->eps); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->maxit); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->gradient_method); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "ttcrpy/tmesh.pyx":265
 *                               self._n_threads, self.tt_from_rp, self.interp_vel,
 *                               self.eps, self.maxit, self.gradient_method,
 *                               self.min_dist, self.n_secondary, self.n_tertiary,             # <<<<<<<<<<<<<<
 *                               self.radius_tertiary)
 *         return (_rebuild3d, constructor_params)
 */
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_self->min_dist); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->n_secondary); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->n_tertiary); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "ttcrpy/tmesh.pyx":266
 *                               self.eps, self.maxit, self.gradient_method,
 *                               self.min_dist, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)             # <<<<<<<<<<<<<<
 *         return (_rebuild3d, constructor_params)
 * 
 */
  __pyx_t_15 = PyFloat_FromDouble(__pyx_v_self->radius_tertiary); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "ttcrpy/tmesh.pyx":262
 *                 tetra[n, nn] = self.tet[n].i[nn]
 * 
 *         constructor_params = (nodes, tetra, method, self.cell_slowness,             # <<<<<<<<<<<<<<
 *                               self._n_threads, self.tt_from_rp, self.interp_vel,
 *                               self.eps, self.maxit, self.gradient_method,
 */
  __pyx_t_16 = PyTuple_New(14); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_nodes);
  __Pyx_INCREF(__pyx_v_tetra);
  __Pyx_GIVEREF(__pyx_v_tetra);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_tetra);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_v_method);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_16, 4, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_16, 5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_16, 6, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_16, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_16, 8, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_16, 9, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_16, 10, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_16, 11, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_16, 12, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_16, 13, __pyx_t_15);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_v_constructor_params = ((PyObject*)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "ttcrpy/tmesh.pyx":267
 *                               self.min_dist, self.n_secondary, self.n_tertiary,
 *                               self.radius_tertiary)
 *         return (_rebuild3d, constructor_params)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_rebuild3d); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16);
  __Pyx_INCREF(__pyx_v_constructor_params);
  __Pyx_GIVEREF(__pyx_v_constructor_params);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_constructor_params);
  __pyx_t_16 = 0;
  __pyx_r = __pyx_t_15;
  __pyx_t_15 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":242
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_tetra);
  __Pyx_XDECREF(__pyx_v_constructor_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":270
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9n_threads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9n_threads_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_9n_threads___get__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/tmesh.pyx":272
 *     def n_threads(self):
 *         """int: number of threads for raytracing"""
 *         return self._n_threads             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":270
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.n_threads.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":275
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7nparams_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7nparams_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_7nparams___get__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_7nparams___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/tmesh.pyx":277
 *     def nparams(self):
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return self.tet.size()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":278
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 *             return self.tet.size()             # <<<<<<<<<<<<<<
 *         else:
 *             return self.no.size()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tet.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":277
 *     def nparams(self):
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return self.tet.size()
 *         else:
 */
  }

  /* "ttcrpy/tmesh.pyx":280
 *             return self.tet.size()
 *         else:
 *             return self.no.size()             # <<<<<<<<<<<<<<
 * 
 *     def set_use_thread_pool(self, use_thread_pool):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/tmesh.pyx":275
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.nparams.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":282
 *             return self.no.size()
 * 
 *     def set_use_thread_pool(self, use_thread_pool):             # <<<<<<<<<<<<<<
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7set_use_thread_pool(PyObject *__pyx_v_self, PyObject *__pyx_v_use_thread_pool); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_6set_use_thread_pool[] = "bool: use thread pool instead of parallel loop";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7set_use_thread_pool(PyObject *__pyx_v_self, PyObject *__pyx_v_use_thread_pool) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_use_thread_pool (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_6set_use_thread_pool(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), ((PyObject *)__pyx_v_use_thread_pool));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_6set_use_thread_pool(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_use_thread_pool) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_use_thread_pool", 0);

  /* "ttcrpy/tmesh.pyx":284
 *     def set_use_thread_pool(self, use_thread_pool):
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)             # <<<<<<<<<<<<<<
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_use_thread_pool); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 284, __pyx_L1_error)
  __pyx_v_self->grid->setUsePool(__pyx_t_1);

  /* "ttcrpy/tmesh.pyx":282
 *             return self.no.size()
 * 
 *     def set_use_thread_pool(self, use_thread_pool):             # <<<<<<<<<<<<<<
 *         """bool: use thread pool instead of parallel loop"""
 *         self.grid.setUsePool(use_thread_pool)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.set_use_thread_pool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":286
 *         self.grid.setUsePool(use_thread_pool)
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_8is_outside[] = "\n        is_outside(pts)\n\n        Check if points are outside grid\n\n        Parameters\n        ----------\n        pts : np ndarray, shape(npts, 3)\n            coordinates of points to check\n\n        Returns\n        -------\n        bool:\n            True if at least one point outside grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9is_outside(PyObject *__pyx_v_self, PyObject *__pyx_v_pts) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_outside (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pts), __pyx_ptype_5numpy_ndarray, 1, "pts", 0))) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_8is_outside(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), ((PyArrayObject *)__pyx_v_pts));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_8is_outside(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyArrayObject *__pyx_v_pts) {
  std::vector<ttcr::sxyz<double> >  __pyx_v_vpts;
  npy_intp __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pts;
  __Pyx_Buffer __pyx_pybuffer_pts;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  ttcr::sxyz<double>  __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_outside", 0);
  __pyx_pybuffer_pts.pybuffer.buf = NULL;
  __pyx_pybuffer_pts.refcount = 0;
  __pyx_pybuffernd_pts.data = NULL;
  __pyx_pybuffernd_pts.rcbuffer = &__pyx_pybuffer_pts;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pts.rcbuffer->pybuffer, (PyObject*)__pyx_v_pts, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 286, __pyx_L1_error)
  }
  __pyx_pybuffernd_pts.diminfo[0].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pts.diminfo[0].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pts.diminfo[1].strides = __pyx_pybuffernd_pts.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pts.diminfo[1].shape = __pyx_pybuffernd_pts.rcbuffer->pybuffer.shape[1];

  /* "ttcrpy/tmesh.pyx":303
 *         """
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(pts.shape[0]):             # <<<<<<<<<<<<<<
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
 *         try:
 */
  __pyx_t_1 = (__pyx_v_pts->dimensions[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_n = __pyx_t_3;

    /* "ttcrpy/tmesh.pyx":304
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(pts.shape[0]):
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))             # <<<<<<<<<<<<<<
 *         try:
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 */
    __pyx_t_4 = __pyx_v_n;
    __pyx_t_5 = 0;
    __pyx_t_6 = -1;
    if (__pyx_t_4 < 0) {
      __pyx_t_4 += __pyx_pybuffernd_pts.diminfo[0].shape;
      if (unlikely(__pyx_t_4 < 0)) __pyx_t_6 = 0;
    } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_pts.diminfo[0].shape)) __pyx_t_6 = 0;
    if (__pyx_t_5 < 0) {
      __pyx_t_5 += __pyx_pybuffernd_pts.diminfo[1].shape;
      if (unlikely(__pyx_t_5 < 0)) __pyx_t_6 = 1;
    } else if (unlikely(__pyx_t_5 >= __pyx_pybuffernd_pts.diminfo[1].shape)) __pyx_t_6 = 1;
    if (unlikely(__pyx_t_6 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_6);
      __PYX_ERR(0, 304, __pyx_L1_error)
    }
    __pyx_t_7 = __pyx_v_n;
    __pyx_t_8 = 1;
    __pyx_t_6 = -1;
    if (__pyx_t_7 < 0) {
      __pyx_t_7 += __pyx_pybuffernd_pts.diminfo[0].shape;
      if (unlikely(__pyx_t_7 < 0)) __pyx_t_6 = 0;
    } else if (unlikely(__pyx_t_7 >= __pyx_pybuffernd_pts.diminfo[0].shape)) __pyx_t_6 = 0;
    if (__pyx_t_8 < 0) {
      __pyx_t_8 += __pyx_pybuffernd_pts.diminfo[1].shape;
      if (unlikely(__pyx_t_8 < 0)) __pyx_t_6 = 1;
    } else if (unlikely(__pyx_t_8 >= __pyx_pybuffernd_pts.diminfo[1].shape)) __pyx_t_6 = 1;
    if (unlikely(__pyx_t_6 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_6);
      __PYX_ERR(0, 304, __pyx_L1_error)
    }
    __pyx_t_9 = __pyx_v_n;
    __pyx_t_10 = 2;
    __pyx_t_6 = -1;
    if (__pyx_t_9 < 0) {
      __pyx_t_9 += __pyx_pybuffernd_pts.diminfo[0].shape;
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_6 = 0;
    } else if (unlikely(__pyx_t_9 >= __pyx_pybuffernd_pts.diminfo[0].shape)) __pyx_t_6 = 0;
    if (__pyx_t_10 < 0) {
      __pyx_t_10 += __pyx_pybuffernd_pts.diminfo[1].shape;
      if (unlikely(__pyx_t_10 < 0)) __pyx_t_6 = 1;
    } else if (unlikely(__pyx_t_10 >= __pyx_pybuffernd_pts.diminfo[1].shape)) __pyx_t_6 = 1;
    if (unlikely(__pyx_t_6 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_6);
      __PYX_ERR(0, 304, __pyx_L1_error)
    }
    try {
      __pyx_t_11 = ttcr::sxyz<double> ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_pts.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_pts.diminfo[0].strides, __pyx_t_5, __pyx_pybuffernd_pts.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_pts.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_pts.diminfo[0].strides, __pyx_t_8, __pyx_pybuffernd_pts.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_pts.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_pts.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_pts.diminfo[1].strides)));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 304, __pyx_L1_error)
    }
    try {
      __pyx_v_vpts.push_back(__pyx_t_11);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 304, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":305
 *         for n in range(pts.shape[0]):
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
 *         try:             # <<<<<<<<<<<<<<
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 *         except RuntimeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
    __Pyx_XGOTREF(__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_13);
    __Pyx_XGOTREF(__pyx_t_14);
    /*try:*/ {

      /* "ttcrpy/tmesh.pyx":306
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
 *         try:
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside             # <<<<<<<<<<<<<<
 *         except RuntimeError:
 *             return True
 */
      try {
        __pyx_v_self->grid->checkPts(__pyx_v_vpts);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 306, __pyx_L5_error)
      }

      /* "ttcrpy/tmesh.pyx":305
 *         for n in range(pts.shape[0]):
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
 *         try:             # <<<<<<<<<<<<<<
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 *         except RuntimeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;

    /* "ttcrpy/tmesh.pyx":307
 *         try:
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 *         except RuntimeError:             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_RuntimeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.is_outside", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0) __PYX_ERR(0, 307, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_GOTREF(__pyx_t_17);

      /* "ttcrpy/tmesh.pyx":308
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 *         except RuntimeError:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L8_except_return;
    }
    goto __pyx_L7_except_error;
    __pyx_L7_except_error:;

    /* "ttcrpy/tmesh.pyx":305
 *         for n in range(pts.shape[0]):
 *             vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
 *         try:             # <<<<<<<<<<<<<<
 *             self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
 *         except RuntimeError:
 */
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_XGIVEREF(__pyx_t_14);
    __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_XGIVEREF(__pyx_t_14);
    __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
    goto __pyx_L0;
    __pyx_L10_try_end:;
  }

  /* "ttcrpy/tmesh.pyx":310
 *             return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_nodes(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":286
 *         self.grid.setUsePool(use_thread_pool)
 * 
 *     def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):             # <<<<<<<<<<<<<<
 *         """
 *         is_outside(pts)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.is_outside", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pts.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":312
 *         return False
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_11get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_10get_number_of_nodes[] = "\n        Returns\n        -------\n        int:\n            number of nodes in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_11get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_nodes (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_10get_number_of_nodes(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_10get_number_of_nodes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_nodes", 0);

  /* "ttcrpy/tmesh.pyx":319
 *             number of nodes in grid
 *         """
 *         return self.no.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_cells(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":312
 *         return False
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_number_of_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":321
 *         return self.no.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_13get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_12get_number_of_cells[] = "\n        Returns\n        -------\n        int:\n            number of cells in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_13get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_cells (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_12get_number_of_cells(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_12get_number_of_cells(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_cells", 0);

  /* "ttcrpy/tmesh.pyx":328
 *             number of cells in grid
 *         """
 *         return self.tet.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_grid_traveltimes(self, thread_no=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tet.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":321
 *         return self.no.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_number_of_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":330
 *         return self.tet.size()
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_15get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_14get_grid_traveltimes[] = "\n        get_grid_traveltimes(thread_no=0)\n\n        Obtain traveltimes computed at primary grid nodes\n\n        Parameters\n        ----------\n        thread_no : int\n            thread used to computed traveltimes (default is 0)\n\n        Returns\n        -------\n        tt: np ndarray, shape (nnodes,)\n            traveltimes\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_15get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_thread_no = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_grid_traveltimes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_thread_no,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_grid_traveltimes") < 0)) __PYX_ERR(0, 330, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_thread_no = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_grid_traveltimes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 330, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_14get_grid_traveltimes(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_thread_no);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_14get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_thread_no) {
  std::vector<double>  __pyx_v_tmp;
  int __pyx_v_n;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<double> ::size_type __pyx_t_7;
  std::vector<double> ::size_type __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_grid_traveltimes", 0);

  /* "ttcrpy/tmesh.pyx":346
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_thread_no, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/tmesh.pyx":347
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 347, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":346
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  }

  /* "ttcrpy/tmesh.pyx":350
 *         cdef vector[double] tmp
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)             # <<<<<<<<<<<<<<
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 */
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L1_error)
  try {
    __pyx_v_self->grid->getTT(__pyx_v_tmp, __pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 350, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":351
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))             # <<<<<<<<<<<<<<
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_tmp.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_tt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":352
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):             # <<<<<<<<<<<<<<
 *             tt[n] = tmp[n]
 *         return tt
 */
  __pyx_t_7 = __pyx_v_tmp.size();
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":353
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]             # <<<<<<<<<<<<<<
 *         return tt
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_tmp[__pyx_v_n])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_n, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":354
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 *         return tt             # <<<<<<<<<<<<<<
 * 
 *     def set_slowness(self, slowness):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tt);
  __pyx_r = __pyx_v_tt;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":330
 *         return self.tet.size()
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":356
 *         return tt
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_17set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_16set_slowness[] = "\n        set_slowness(slowness)\n\n        Assign slowness to grid\n\n        Parameters\n        ----------\n        slowness : np ndarray, shape (nparams, )\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_17set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_16set_slowness(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), ((PyObject *)__pyx_v_slowness));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_16set_slowness(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_slowness) {
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_slowness", 0);
  __Pyx_INCREF(__pyx_v_slowness);

  /* "ttcrpy/tmesh.pyx":366
 *         slowness : np ndarray, shape (nparams, )
 *         """
 *         if slowness.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":367
 *         """
 *         if slowness.size != self.nparams:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 367, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":366
 *         slowness : np ndarray, shape (nparams, )
 *         """
 *         if slowness.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":369
 *             raise ValueError('Slowness vector has wrong size')
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             slowness = np.ascontiguousarray(slowness)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "ttcrpy/tmesh.pyx":370
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:
 *             slowness = np.ascontiguousarray(slowness)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_slowness, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":369
 *             raise ValueError('Slowness vector has wrong size')
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             slowness = np.ascontiguousarray(slowness)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":374
 *         cdef vector[double] slown
 *         cdef int i
 *         for i in range(slowness.size):             # <<<<<<<<<<<<<<
 *             slown.push_back(slowness[i])
 *         self.grid.setSlowness(slown)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ttcrpy/tmesh.pyx":375
 *         cdef int i
 *         for i in range(slowness.size):
 *             slown.push_back(slowness[i])             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_slowness, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      __pyx_v_slown.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 375, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":376
 *         for i in range(slowness.size):
 *             slown.push_back(slowness[i])
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def set_velocity(self, velocity):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 376, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":356
 *         return tt
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.set_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":378
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_19set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_18set_velocity[] = "\n        set_velocity(velocity)\n\n        Assign velocity to grid\n\n        Parameters\n        ----------\n        velocity : np ndarray, shape (nparams, )\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_19set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_velocity (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_18set_velocity(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), ((PyObject *)__pyx_v_velocity));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_18set_velocity(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_velocity) {
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_velocity", 0);
  __Pyx_INCREF(__pyx_v_velocity);

  /* "ttcrpy/tmesh.pyx":388
 *         velocity : np ndarray, shape (nparams, )
 *         """
 *         if velocity.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":389
 *         """
 *         if velocity.size != self.nparams:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 389, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":388
 *         velocity : np ndarray, shape (nparams, )
 *         """
 *         if velocity.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":391
 *             raise ValueError('velocity vector has wrong size')
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             velocity = np.ascontiguousarray(velocity)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "ttcrpy/tmesh.pyx":392
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:
 *             velocity = np.ascontiguousarray(velocity)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_velocity) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_velocity);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_velocity, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":391
 *             raise ValueError('velocity vector has wrong size')
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             velocity = np.ascontiguousarray(velocity)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":396
 *         cdef vector[double] slown
 *         cdef int i
 *         for i in range(velocity.size):             # <<<<<<<<<<<<<<
 *             slown.push_back(1./velocity[i])
 *         self.grid.setSlowness(slown)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ttcrpy/tmesh.pyx":397
 *         cdef int i
 *         for i in range(velocity.size):
 *             slown.push_back(1./velocity[i])             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_velocity, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_2, 1., 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    try {
      __pyx_v_slown.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 397, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":398
 *         for i in range(velocity.size):
 *             slown.push_back(1./velocity[i])
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def compute_D(self, coord):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 398, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":378
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.set_velocity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_velocity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":400
 *         self.grid.setSlowness(slown)
 * 
 *     def compute_D(self, coord):             # <<<<<<<<<<<<<<
 *         """
 *         compute_D(coord)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_21compute_D(PyObject *__pyx_v_self, PyObject *__pyx_v_coord); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_20compute_D[] = "\n        compute_D(coord)\n\n        Return matrix of interpolation weights for velocity data points\n        constraint\n\n        Parameters\n        ----------\n        coord : np.ndarray, shape (npts, 3)\n            coordinates of data points\n\n        Returns\n        -------\n        D : scipy csr_matrix, shape (npts, nparams)\n            Matrix of interpolation weights\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_21compute_D(PyObject *__pyx_v_self, PyObject *__pyx_v_coord) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_D (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_20compute_D(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), ((PyObject *)__pyx_v_coord));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_20compute_D(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_coord) {
  std::vector<ttcr::sxyz<double> >  __pyx_v_vpts;
  PyObject *__pyx_v_n = NULL;
  std::vector<std::vector<ttcr::sijv<double> > >  __pyx_v_d_data;
  PyObject *__pyx_v_nnz = NULL;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_v_ni;
  PyObject *__pyx_v_MM = NULL;
  PyObject *__pyx_v_NN = NULL;
  PyObject *__pyx_v_indptr = NULL;
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_v_nn;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  double __pyx_t_7;
  double __pyx_t_8;
  double __pyx_t_9;
  ttcr::sxyz<double>  __pyx_t_10;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_11;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_12;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_18;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_19;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_20;
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_D", 0);

  /* "ttcrpy/tmesh.pyx":417
 *             Matrix of interpolation weights
 *         """
 *         if self.is_outside(coord):             # <<<<<<<<<<<<<<
 *             raise ValueError('Velocity data point outside grid')
 *         cdef vector[sxyz[double]] vpts
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_coord) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_coord);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":418
 *         """
 *         if self.is_outside(coord):
 *             raise ValueError('Velocity data point outside grid')             # <<<<<<<<<<<<<<
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(coord.shape[0]):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 418, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":417
 *             Matrix of interpolation weights
 *         """
 *         if self.is_outside(coord):             # <<<<<<<<<<<<<<
 *             raise ValueError('Velocity data point outside grid')
 *         cdef vector[sxyz[double]] vpts
 */
  }

  /* "ttcrpy/tmesh.pyx":420
 *             raise ValueError('Velocity data point outside grid')
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(coord.shape[0]):             # <<<<<<<<<<<<<<
 *             vpts.push_back(sxyz[double](coord[n, 0], coord[n, 1], coord[n, 2]))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coord, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 420, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 420, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 420, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 420, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":421
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(coord.shape[0]):
 *             vpts.push_back(sxyz[double](coord[n, 0], coord[n, 1], coord[n, 2]))             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sijv[double]]] d_data
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_n);
    __Pyx_GIVEREF(__pyx_v_n);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_n);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_0);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_n);
    __Pyx_GIVEREF(__pyx_v_n);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_n);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_n);
    __Pyx_GIVEREF(__pyx_v_n);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_n);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_2);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_coord, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    try {
      __pyx_t_10 = ttcr::sxyz<double> (__pyx_t_7, __pyx_t_8, __pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 421, __pyx_L1_error)
    }
    try {
      __pyx_v_vpts.push_back(__pyx_t_10);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 421, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":420
 *             raise ValueError('Velocity data point outside grid')
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(coord.shape[0]):             # <<<<<<<<<<<<<<
 *             vpts.push_back(sxyz[double](coord[n, 0], coord[n, 1], coord[n, 2]))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":425
 *         cdef vector[vector[sijv[double]]] d_data
 * 
 *         self.grid.computeD(vpts, d_data)             # <<<<<<<<<<<<<<
 * 
 *         nnz = 0
 */
  __pyx_v_self->grid->computeD(__pyx_v_vpts, __pyx_v_d_data);

  /* "ttcrpy/tmesh.pyx":427
 *         self.grid.computeD(vpts, d_data)
 * 
 *         nnz = 0             # <<<<<<<<<<<<<<
 *         for ni in range(d_data.size()):
 *             nnz += d_data[ni].size()
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_nnz = __pyx_int_0;

  /* "ttcrpy/tmesh.pyx":428
 * 
 *         nnz = 0
 *         for ni in range(d_data.size()):             # <<<<<<<<<<<<<<
 *             nnz += d_data[ni].size()
 * 
 */
  __pyx_t_11 = __pyx_v_d_data.size();
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_ni = __pyx_t_13;

    /* "ttcrpy/tmesh.pyx":429
 *         nnz = 0
 *         for ni in range(d_data.size()):
 *             nnz += d_data[ni].size()             # <<<<<<<<<<<<<<
 * 
 *         MM = vpts.size()
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_d_data[__pyx_v_ni]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_nnz, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":431
 *             nnz += d_data[ni].size()
 * 
 *         MM = vpts.size()             # <<<<<<<<<<<<<<
 *         NN = self.nparams
 *         indptr = np.empty((MM+1,), dtype=np.int64)
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_vpts.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_MM = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":432
 * 
 *         MM = vpts.size()
 *         NN = self.nparams             # <<<<<<<<<<<<<<
 *         indptr = np.empty((MM+1,), dtype=np.int64)
 *         indices = np.empty((nnz,), dtype=np.int64)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_NN = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":433
 *         MM = vpts.size()
 *         NN = self.nparams
 *         indptr = np.empty((MM+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *         indices = np.empty((nnz,), dtype=np.int64)
 *         val = np.empty((nnz,))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_MM, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_int64); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_15) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_indptr = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "ttcrpy/tmesh.pyx":434
 *         NN = self.nparams
 *         indptr = np.empty((MM+1,), dtype=np.int64)
 *         indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *         val = np.empty((nnz,))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_v_nnz);
  __Pyx_GIVEREF(__pyx_v_nnz);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_nnz);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_dtype, __pyx_t_14) < 0) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_v_indices = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "ttcrpy/tmesh.pyx":435
 *         indptr = np.empty((MM+1,), dtype=np.int64)
 *         indices = np.empty((nnz,), dtype=np.int64)
 *         val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *         k = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_v_nnz);
  __Pyx_GIVEREF(__pyx_v_nnz);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_nnz);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_14 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_15);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_val = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "ttcrpy/tmesh.pyx":437
 *         val = np.empty((nnz,))
 * 
 *         k = 0             # <<<<<<<<<<<<<<
 *         for i in range(MM):
 *             indptr[i] = k
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_k = __pyx_int_0;

  /* "ttcrpy/tmesh.pyx":438
 * 
 *         k = 0
 *         for i in range(MM):             # <<<<<<<<<<<<<<
 *             indptr[i] = k
 *             for j in range(NN):
 */
  __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_MM); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
    __pyx_t_3 = __pyx_t_14; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 438, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_14); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 438, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_14); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 438, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      }
    } else {
      __pyx_t_14 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_14)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 438, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_14);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "ttcrpy/tmesh.pyx":439
 *         k = 0
 *         for i in range(MM):
 *             indptr[i] = k             # <<<<<<<<<<<<<<
 *             for j in range(NN):
 *                 for nn in range(d_data[i].size()):
 */
    if (unlikely(PyObject_SetItem(__pyx_v_indptr, __pyx_v_i, __pyx_v_k) < 0)) __PYX_ERR(0, 439, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":440
 *         for i in range(MM):
 *             indptr[i] = k
 *             for j in range(NN):             # <<<<<<<<<<<<<<
 *                 for nn in range(d_data[i].size()):
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:
 */
    __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_NN); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
      __pyx_t_15 = __pyx_t_14; __Pyx_INCREF(__pyx_t_15); __pyx_t_16 = 0;
      __pyx_t_17 = NULL;
    } else {
      __pyx_t_16 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_17 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 440, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    for (;;) {
      if (likely(!__pyx_t_17)) {
        if (likely(PyList_CheckExact(__pyx_t_15))) {
          if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_16); __Pyx_INCREF(__pyx_t_14); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_15, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        } else {
          if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_16); __Pyx_INCREF(__pyx_t_14); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_15, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        }
      } else {
        __pyx_t_14 = __pyx_t_17(__pyx_t_15);
        if (unlikely(!__pyx_t_14)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 440, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_14);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "ttcrpy/tmesh.pyx":441
 *             indptr[i] = k
 *             for j in range(NN):
 *                 for nn in range(d_data[i].size()):             # <<<<<<<<<<<<<<
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:
 *                         indices[k] = j
 */
      __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_v_i); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
      __pyx_t_19 = (__pyx_v_d_data[__pyx_t_18]).size();
      __pyx_t_20 = __pyx_t_19;
      for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
        __pyx_v_nn = __pyx_t_21;

        /* "ttcrpy/tmesh.pyx":442
 *             for j in range(NN):
 *                 for nn in range(d_data[i].size()):
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:             # <<<<<<<<<<<<<<
 *                         indices[k] = j
 *                         val[k] = d_data[i][nn].v
 */
        __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_v_i); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 442, __pyx_L1_error)
        __pyx_t_14 = __Pyx_PyInt_FromSize_t(((__pyx_v_d_data[__pyx_t_18])[__pyx_v_nn]).i); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_1 = PyObject_RichCompare(__pyx_t_14, __pyx_v_i, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_22) {
        } else {
          __pyx_t_4 = __pyx_t_22;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_v_i); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 442, __pyx_L1_error)
        __pyx_t_1 = __Pyx_PyInt_FromSize_t(((__pyx_v_d_data[__pyx_t_18])[__pyx_v_nn]).j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = PyObject_RichCompare(__pyx_t_1, __pyx_v_j, Py_EQ); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_4 = __pyx_t_22;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_4) {

          /* "ttcrpy/tmesh.pyx":443
 *                 for nn in range(d_data[i].size()):
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:
 *                         indices[k] = j             # <<<<<<<<<<<<<<
 *                         val[k] = d_data[i][nn].v
 *                         k += 1
 */
          if (unlikely(PyObject_SetItem(__pyx_v_indices, __pyx_v_k, __pyx_v_j) < 0)) __PYX_ERR(0, 443, __pyx_L1_error)

          /* "ttcrpy/tmesh.pyx":444
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:
 *                         indices[k] = j
 *                         val[k] = d_data[i][nn].v             # <<<<<<<<<<<<<<
 *                         k += 1
 * 
 */
          __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_v_i); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L1_error)
          __pyx_t_14 = PyFloat_FromDouble(((__pyx_v_d_data[__pyx_t_18])[__pyx_v_nn]).v); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (unlikely(PyObject_SetItem(__pyx_v_val, __pyx_v_k, __pyx_t_14) < 0)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "ttcrpy/tmesh.pyx":445
 *                         indices[k] = j
 *                         val[k] = d_data[i][nn].v
 *                         k += 1             # <<<<<<<<<<<<<<
 * 
 *         indptr[MM] = k
 */
          __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_v_k, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 445, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF_SET(__pyx_v_k, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "ttcrpy/tmesh.pyx":442
 *             for j in range(NN):
 *                 for nn in range(d_data[i].size()):
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:             # <<<<<<<<<<<<<<
 *                         indices[k] = j
 *                         val[k] = d_data[i][nn].v
 */
        }
      }

      /* "ttcrpy/tmesh.pyx":440
 *         for i in range(MM):
 *             indptr[i] = k
 *             for j in range(NN):             # <<<<<<<<<<<<<<
 *                 for nn in range(d_data[i].size()):
 *                     if d_data[i][nn].i == i and d_data[i][nn].j == j:
 */
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "ttcrpy/tmesh.pyx":438
 * 
 *         k = 0
 *         for i in range(MM):             # <<<<<<<<<<<<<<
 *             indptr[i] = k
 *             for j in range(NN):
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":447
 *                         k += 1
 * 
 *         indptr[MM] = k             # <<<<<<<<<<<<<<
 *         return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))
 * 
 */
  if (unlikely(PyObject_SetItem(__pyx_v_indptr, __pyx_v_MM, __pyx_v_k) < 0)) __PYX_ERR(0, 447, __pyx_L1_error)

  /* "ttcrpy/tmesh.pyx":448
 * 
 *         indptr[MM] = k
 *         return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))             # <<<<<<<<<<<<<<
 * 
 *     def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_val);
  __Pyx_INCREF(__pyx_v_indices);
  __Pyx_GIVEREF(__pyx_v_indices);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_indices);
  __Pyx_INCREF(__pyx_v_indptr);
  __Pyx_GIVEREF(__pyx_v_indptr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_indptr);
  __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_MM);
  __Pyx_GIVEREF(__pyx_v_MM);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_MM);
  __Pyx_INCREF(__pyx_v_NN);
  __Pyx_GIVEREF(__pyx_v_NN);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_NN);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shape, __pyx_t_1) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":400
 *         self.grid.setSlowness(slown)
 * 
 *     def compute_D(self, coord):             # <<<<<<<<<<<<<<
 *         """
 *         compute_D(coord)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.compute_D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_nnz);
  __Pyx_XDECREF(__pyx_v_MM);
  __Pyx_XDECREF(__pyx_v_NN);
  __Pyx_XDECREF(__pyx_v_indptr);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":450
 *         return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))
 * 
 *     def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,             # <<<<<<<<<<<<<<
 *                   s0inside=False):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_23compute_K(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_22compute_K[] = "\n        Compute smoothing matrices (spatial derivative)\n\n        Parameters\n        ----------\n        order : int\n            order of derivative (1 or 2, 2 by default)\n        taylor_order : int\n            order of taylors series expansion (1 or 2, 2 by default)\n        weighting : bool\n            apply inverse distance weighting (True by default)\n        squared : bool\n            Second derivative evaluated by taking the square of first\n            derivative.  Applied only if order == 2 (True by default)\n        s0inside : bool\n            (experimental) ignore slowness value at local node (value is a\n            filtered estimate) (False by default)\n\n        Returns\n        -------\n        Kx, Ky, Kz : :obj:`tuple` of :obj:`csr_matrix`\n            matrices for derivatives along x, y, & z\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_23compute_K(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_order = 0;
  PyObject *__pyx_v_taylor_order = 0;
  PyObject *__pyx_v_weighting = 0;
  PyObject *__pyx_v_squared = 0;
  PyObject *__pyx_v_s0inside = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_K (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_order,&__pyx_n_s_taylor_order,&__pyx_n_s_weighting,&__pyx_n_s_squared,&__pyx_n_s_s0inside,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_2);
    values[1] = ((PyObject *)__pyx_int_2);
    values[2] = ((PyObject *)Py_True);
    values[3] = ((PyObject *)Py_True);

    /* "ttcrpy/tmesh.pyx":451
 * 
 *     def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,
 *                   s0inside=False):             # <<<<<<<<<<<<<<
 *         """
 *         Compute smoothing matrices (spatial derivative)
 */
    values[4] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_taylor_order);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_weighting);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_squared);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s0inside);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_K") < 0)) __PYX_ERR(0, 450, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_order = values[0];
    __pyx_v_taylor_order = values[1];
    __pyx_v_weighting = values[2];
    __pyx_v_squared = values[3];
    __pyx_v_s0inside = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_K", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 450, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.compute_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_22compute_K(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_order, __pyx_v_taylor_order, __pyx_v_weighting, __pyx_v_squared, __pyx_v_s0inside);

  /* "ttcrpy/tmesh.pyx":450
 *         return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))
 * 
 *     def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,             # <<<<<<<<<<<<<<
 *                   s0inside=False):
 *         """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_22compute_K(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_order, PyObject *__pyx_v_taylor_order, PyObject *__pyx_v_weighting, PyObject *__pyx_v_squared, PyObject *__pyx_v_s0inside) {
  int __pyx_v_o;
  int __pyx_v_to;
  bool __pyx_v_w;
  bool __pyx_v_s;
  std::vector<std::vector<std::vector<ttcr::siv<double> > > >  __pyx_v_k_data;
  PyObject *__pyx_v_K = NULL;
  size_t __pyx_v_MM;
  size_t __pyx_v_NN;
  long __pyx_v_nk;
  PyObject *__pyx_v_m_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  bool __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  size_t __pyx_t_6;
  long __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_K", 0);

  /* "ttcrpy/tmesh.pyx":476
 *         """
 * 
 *         cdef int o = order             # <<<<<<<<<<<<<<
 *         cdef int to = taylor_order
 *         cdef bool w = weighting
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_order); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_v_o = __pyx_t_1;

  /* "ttcrpy/tmesh.pyx":477
 * 
 *         cdef int o = order
 *         cdef int to = taylor_order             # <<<<<<<<<<<<<<
 *         cdef bool w = weighting
 *         cdef bool s = s0inside
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_taylor_order); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_v_to = __pyx_t_1;

  /* "ttcrpy/tmesh.pyx":478
 *         cdef int o = order
 *         cdef int to = taylor_order
 *         cdef bool w = weighting             # <<<<<<<<<<<<<<
 *         cdef bool s = s0inside
 *         cdef vector[vector[vector[siv[double]]]] k_data
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_weighting); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L1_error)
  __pyx_v_w = __pyx_t_2;

  /* "ttcrpy/tmesh.pyx":479
 *         cdef int to = taylor_order
 *         cdef bool w = weighting
 *         cdef bool s = s0inside             # <<<<<<<<<<<<<<
 *         cdef vector[vector[vector[siv[double]]]] k_data
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_s0inside); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L1_error)
  __pyx_v_s = __pyx_t_2;

  /* "ttcrpy/tmesh.pyx":482
 *         cdef vector[vector[vector[siv[double]]]] k_data
 * 
 *         if order == 2 and squared:             # <<<<<<<<<<<<<<
 *             o = 1
 *         self.grid.computeK(k_data, o, to, w, s)
 */
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_v_order, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_squared); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ttcrpy/tmesh.pyx":483
 * 
 *         if order == 2 and squared:
 *             o = 1             # <<<<<<<<<<<<<<
 *         self.grid.computeK(k_data, o, to, w, s)
 * 
 */
    __pyx_v_o = 1;

    /* "ttcrpy/tmesh.pyx":482
 *         cdef vector[vector[vector[siv[double]]]] k_data
 * 
 *         if order == 2 and squared:             # <<<<<<<<<<<<<<
 *             o = 1
 *         self.grid.computeK(k_data, o, to, w, s)
 */
  }

  /* "ttcrpy/tmesh.pyx":484
 *         if order == 2 and squared:
 *             o = 1
 *         self.grid.computeK(k_data, o, to, w, s)             # <<<<<<<<<<<<<<
 * 
 *         K = []
 */
  try {
    __pyx_v_self->grid->computeK(__pyx_v_k_data, __pyx_v_o, __pyx_v_to, __pyx_v_w, __pyx_v_s);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 484, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":486
 *         self.grid.computeK(k_data, o, to, w, s)
 * 
 *         K = []             # <<<<<<<<<<<<<<
 *         cdef size_t MM = self.nparams
 *         cdef size_t NN = self.nparams
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_K = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "ttcrpy/tmesh.pyx":487
 * 
 *         K = []
 *         cdef size_t MM = self.nparams             # <<<<<<<<<<<<<<
 *         cdef size_t NN = self.nparams
 *         for nk in range(3):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_MM = __pyx_t_6;

  /* "ttcrpy/tmesh.pyx":488
 *         K = []
 *         cdef size_t MM = self.nparams
 *         cdef size_t NN = self.nparams             # <<<<<<<<<<<<<<
 *         for nk in range(3):
 *             m_tuple = ([0.0], [0.0], [0.0])
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_NN = __pyx_t_6;

  /* "ttcrpy/tmesh.pyx":489
 *         cdef size_t MM = self.nparams
 *         cdef size_t NN = self.nparams
 *         for nk in range(3):             # <<<<<<<<<<<<<<
 *             m_tuple = ([0.0], [0.0], [0.0])
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)
 */
  for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
    __pyx_v_nk = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":490
 *         cdef size_t NN = self.nparams
 *         for nk in range(3):
 *             m_tuple = ([0.0], [0.0], [0.0])             # <<<<<<<<<<<<<<
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
 */
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_float_0_0);
    __Pyx_GIVEREF(__pyx_float_0_0);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_float_0_0);
    __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_float_0_0);
    __Pyx_GIVEREF(__pyx_float_0_0);
    PyList_SET_ITEM(__pyx_t_8, 0, __pyx_float_0_0);
    __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_float_0_0);
    __Pyx_GIVEREF(__pyx_float_0_0);
    PyList_SET_ITEM(__pyx_t_9, 0, __pyx_float_0_0);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
    __pyx_t_4 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_m_tuple, ((PyObject*)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "ttcrpy/tmesh.pyx":491
 *         for nk in range(3):
 *             m_tuple = ([0.0], [0.0], [0.0])
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)             # <<<<<<<<<<<<<<
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
 *             if order == 2 and squared:
 */
    (void)(build_matrix_siv<double>(__pyx_v_MM, __pyx_v_NN, (__pyx_v_k_data[__pyx_v_nk]), __pyx_v_m_tuple));

    /* "ttcrpy/tmesh.pyx":492
 *             m_tuple = ([0.0], [0.0], [0.0])
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 *             if order == 2 and squared:
 *                 K[-1] = K[-1] * K[-1]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_sp); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_m_tuple);
    __Pyx_GIVEREF(__pyx_v_m_tuple);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_m_tuple);
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_MM); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = __Pyx_PyInt_FromSize_t(__pyx_v_NN); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_11);
    __pyx_t_4 = 0;
    __pyx_t_11 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_shape, __pyx_t_12) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_K, __pyx_t_12); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "ttcrpy/tmesh.pyx":493
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
 *             if order == 2 and squared:             # <<<<<<<<<<<<<<
 *                 K[-1] = K[-1] * K[-1]
 * 
 */
    __pyx_t_12 = __Pyx_PyInt_EqObjC(__pyx_v_order, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_squared); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_5;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_3) {

      /* "ttcrpy/tmesh.pyx":494
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
 *             if order == 2 and squared:
 *                 K[-1] = K[-1] * K[-1]             # <<<<<<<<<<<<<<
 * 
 *         return tuple(K)
 */
      __pyx_t_12 = __Pyx_GetItemInt_List(__pyx_v_K, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_K, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = PyNumber_Multiply(__pyx_t_12, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__Pyx_SetItemInt(__pyx_v_K, -1L, __pyx_t_10, long, 1, __Pyx_PyInt_From_long, 1, 1, 1) < 0)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "ttcrpy/tmesh.pyx":493
 *             build_matrix_siv(MM, NN, k_data[nk], m_tuple)
 *             K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
 *             if order == 2 and squared:             # <<<<<<<<<<<<<<
 *                 K[-1] = K[-1] * K[-1]
 * 
 */
    }
  }

  /* "ttcrpy/tmesh.pyx":496
 *                 K[-1] = K[-1] * K[-1]
 * 
 *         return tuple(K)             # <<<<<<<<<<<<<<
 * 
 *     def get_s0(self, hypo, slowness=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = PyList_AsTuple(__pyx_v_K); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":450
 *         return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))
 * 
 *     def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,             # <<<<<<<<<<<<<<
 *                   s0inside=False):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.compute_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_K);
  __Pyx_XDECREF(__pyx_v_m_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":498
 *         return tuple(K)
 * 
 *     def get_s0(self, hypo, slowness=None):             # <<<<<<<<<<<<<<
 *         """
 *         get_s0(hypo, slowness=None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_25get_s0(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_24get_s0[] = "\n        get_s0(hypo, slowness=None)\n\n        Return slowness at source points\n\n        Parameters\n        ----------\n        hypo : np.ndarray with 5 columns\n            hypo holds source information, i.e.\n                - 1st column is event ID number\n                - 2nd column is origin time\n                - 3rd column is source easting\n                - 4th column is source northing\n                - 5th column is source elevation\n        slowness : np ndarray, shape (nparams, ) (optional)\n            slowness at grid nodes or cells (depending on cell_slowness)\n\n        Returns\n        -------\n        s0 : np.ndarray\n            slowness at source points\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_25get_s0(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hypo = 0;
  PyObject *__pyx_v_slowness = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_s0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hypo,&__pyx_n_s_slowness,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hypo)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_s0") < 0)) __PYX_ERR(0, 498, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_hypo = values[0];
    __pyx_v_slowness = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_s0", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 498, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_s0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_24get_s0(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_hypo, __pyx_v_slowness);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_24get_s0(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_hypo, PyObject *__pyx_v_slowness) {
  Py_ssize_t __pyx_v_n;
  Py_ssize_t __pyx_v_nn;
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_evID = NULL;
  PyObject *__pyx_v_eid = NULL;
  Py_ssize_t __pyx_v_nTx;
  PyObject *__pyx_v_i0 = NULL;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vTx;
  PyObject *__pyx_v_iTx = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_ii = NULL;
  PyObject *__pyx_v_s0 = NULL;
  double __pyx_v_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  long __pyx_t_11;
  long __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  double __pyx_t_14;
  double __pyx_t_15;
  double __pyx_t_16;
  ttcr::sxyz<double>  __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_t_19;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_20;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_s0", 0);

  /* "ttcrpy/tmesh.pyx":524
 *         cdef Py_ssize_t n, nn
 * 
 *         if hypo.shape[1] != 5:             # <<<<<<<<<<<<<<
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_hypo, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/tmesh.pyx":525
 * 
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')             # <<<<<<<<<<<<<<
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 525, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":524
 *         cdef Py_ssize_t n, nn
 * 
 *         if hypo.shape[1] != 5:             # <<<<<<<<<<<<<<
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 */
  }

  /* "ttcrpy/tmesh.pyx":526
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]             # <<<<<<<<<<<<<<
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_hypo, __pyx_tuple__11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_src = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":527
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]             # <<<<<<<<<<<<<<
 *         eid = np.sort(np.unique(evID))
 *         nTx = len(eid)
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_hypo, __pyx_tuple__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_evID = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":528
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))             # <<<<<<<<<<<<<<
 *         nTx = len(eid)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sort); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_unique); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_v_evID) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_evID);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_eid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":529
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 *         nTx = len(eid)             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_eid); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 529, __pyx_L1_error)
  __pyx_v_nTx = __pyx_t_7;

  /* "ttcrpy/tmesh.pyx":531
 *         nTx = len(eid)
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  __pyx_t_3 = (__pyx_v_slowness != Py_None);
  __pyx_t_8 = (__pyx_t_3 != 0);
  if (__pyx_t_8) {

    /* "ttcrpy/tmesh.pyx":532
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":531
 *         nTx = len(eid)
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":534
 *             self.set_slowness(slowness)
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_nTx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_i0 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/tmesh.pyx":535
 * 
 *         i0 = np.empty((nTx,), dtype=np.int64)
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/tmesh.pyx":536
 *         i0 = np.empty((nTx,), dtype=np.int64)
 *         for n in range(nTx):
 *             for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_5); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/tmesh.pyx":537
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0[n] = nn
 *                     break
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_evID, __pyx_v_nn, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_8) {

        /* "ttcrpy/tmesh.pyx":538
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_nn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_i0, __pyx_v_n, __pyx_t_1, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/tmesh.pyx":539
 *                 if eid[n] == evID[nn]:
 *                     i0[n] = nn
 *                     break             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 */
        goto __pyx_L8_break;

        /* "ttcrpy/tmesh.pyx":537
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0[n] = nn
 *                     break
 */
      }
    }
    __pyx_L8_break:;
  }

  /* "ttcrpy/tmesh.pyx":542
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iTx = []
 */
  try {
    __pyx_v_vTx.resize(__pyx_v_nTx);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 542, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":544
 *         vTx.resize(nTx)
 * 
 *         iTx = []             # <<<<<<<<<<<<<<
 *         i0 = 0
 *         for n in range(nTx):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_iTx = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":545
 * 
 *         iTx = []
 *         i0 = 0             # <<<<<<<<<<<<<<
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_i0, __pyx_int_0);

  /* "ttcrpy/tmesh.pyx":546
 *         iTx = []
 *         i0 = 0
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/tmesh.pyx":547
 *         i0 = 0
 *         for n in range(nTx):
 *             for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/tmesh.pyx":548
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0 = nn
 *                     break
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_evID, __pyx_v_nn, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 548, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 548, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 548, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_8) {

        /* "ttcrpy/tmesh.pyx":549
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn             # <<<<<<<<<<<<<<
 *                     break
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 */
        __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_nn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_i0, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/tmesh.pyx":550
 *                 if eid[n] == evID[nn]:
 *                     i0 = nn
 *                     break             # <<<<<<<<<<<<<<
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 */
        goto __pyx_L13_break;

        /* "ttcrpy/tmesh.pyx":548
 *         for n in range(nTx):
 *             for nn in range(evID.size):
 *                 if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                     i0 = nn
 *                     break
 */
      }
    }
    __pyx_L13_break:;

    /* "ttcrpy/tmesh.pyx":551
 *                     i0 = nn
 *                     break
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))             # <<<<<<<<<<<<<<
 * 
 *         for i in eid:
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i0);
    __Pyx_GIVEREF(__pyx_v_i0);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      __pyx_t_17 = ttcr::sxyz<double> (__pyx_t_14, __pyx_t_15, __pyx_t_16);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 551, __pyx_L1_error)
    }
    try {
      (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_17);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 551, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":553
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 *         for i in eid:             # <<<<<<<<<<<<<<
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])
 */
  if (likely(PyList_CheckExact(__pyx_v_eid)) || PyTuple_CheckExact(__pyx_v_eid)) {
    __pyx_t_2 = __pyx_v_eid; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_18 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_eid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 553, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_18)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_18(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 553, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":554
 * 
 *         for i in eid:
 *             ii = evID == i             # <<<<<<<<<<<<<<
 *             iTx.append(np.nonzero(ii)[0])
 * 
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_evID, __pyx_v_i, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_ii, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":555
 *         for i in eid:
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])             # <<<<<<<<<<<<<<
 * 
 *         s0 = np.zeros((src.shape[0],))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_ii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_ii);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iTx, __pyx_t_4); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ttcrpy/tmesh.pyx":553
 *             vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 * 
 *         for i in eid:             # <<<<<<<<<<<<<<
 *             ii = evID == i
 *             iTx.append(np.nonzero(ii)[0])
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":557
 *             iTx.append(np.nonzero(ii)[0])
 * 
 *         s0 = np.zeros((src.shape[0],))             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_s0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":559
 *         s0 = np.zeros((src.shape[0],))
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             s = 0.0
 *             for nn in range(vTx[n].size()):
 */
  __pyx_t_7 = __pyx_v_nTx;
  __pyx_t_9 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_n = __pyx_t_10;

    /* "ttcrpy/tmesh.pyx":560
 * 
 *         for n in range(nTx):
 *             s = 0.0             # <<<<<<<<<<<<<<
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 */
    __pyx_v_s = 0.0;

    /* "ttcrpy/tmesh.pyx":561
 *         for n in range(nTx):
 *             s = 0.0
 *             for nn in range(vTx[n].size()):             # <<<<<<<<<<<<<<
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()
 */
    __pyx_t_20 = (__pyx_v_vTx[__pyx_v_n]).size();
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_21; __pyx_t_13+=1) {
      __pyx_v_nn = __pyx_t_13;

      /* "ttcrpy/tmesh.pyx":562
 *             s = 0.0
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])             # <<<<<<<<<<<<<<
 *             s0[iTx[n]] = s/vTx[n].size()
 *         return s0
 */
      try {
        __pyx_t_16 = __pyx_v_self->grid->computeSlowness(((__pyx_v_vTx[__pyx_v_n])[__pyx_v_nn]));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 562, __pyx_L1_error)
      }
      __pyx_v_s = (__pyx_v_s + __pyx_t_16);
    }

    /* "ttcrpy/tmesh.pyx":563
 *             for nn in range(vTx[n].size()):
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()             # <<<<<<<<<<<<<<
 *         return s0
 * 
 */
    __pyx_t_20 = (__pyx_v_vTx[__pyx_v_n]).size();
    if (unlikely(__pyx_t_20 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 563, __pyx_L1_error)
    }
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_s / ((double)__pyx_t_20))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iTx, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(PyObject_SetItem(__pyx_v_s0, __pyx_t_5, __pyx_t_2) < 0)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":564
 *                 s += self.grid.computeSlowness(vTx[n][nn])
 *             s0[iTx[n]] = s/vTx[n].size()
 *         return s0             # <<<<<<<<<<<<<<
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_s0);
  __pyx_r = __pyx_v_s0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":498
 *         return tuple(K)
 * 
 *     def get_s0(self, hypo, slowness=None):             # <<<<<<<<<<<<<<
 *         """
 *         get_s0(hypo, slowness=None)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.get_s0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_evID);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_i0);
  __Pyx_XDECREF(__pyx_v_iTx);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_ii);
  __Pyx_XDECREF(__pyx_v_s0);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":566
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, return_rays=False):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_27raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_26raytrace[] = "\n        raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, compute_L=False, return_rays=False) -> tt, rays, L\n\n        Perform raytracing\n\n        Parameters\n        ----------\n        source : 2D np.ndarray with 3, 4 or 5 columns\n            see notes below\n        rcv : 2D np.ndarray with 3 columns\n            Columns correspond to x, y and z coordinates\n        slowness : np ndarray, (None by default)\n            slowness at grid nodes or cells (depending on cell_slowness)\n            if None, slowness must have been assigned previously\n        thread_no : int (None by default)\n            Perform calculations in thread number \"thread_no\"\n            if None, attempt to run in parallel if warranted by number of\n            sources and value of n_threads in constructor\n        aggregate_src : bool (False by default)\n            if True, all source coordinates belong to a single event\n        compute_L : bool (False by default)\n            Compute matrices of partial derivative of travel time w/r to slowness\n        return_rays : bool (False by default)\n            Return raypaths\n\n        Returns\n        -------\n        tt : np.ndarray\n            travel times for the appropriate source-rcv  (see Notes below)\n        rays : :obj:`list` of :obj:`np.ndarray`\n            Coordinates of segments forming raypaths (if return_rays is True)\n        L :  :obj:`list` of :obj:`csr_matrix`  or  scipy csr_matrix\n            Matrix of partial derivative of travel time w/r to slowness.\n            if input argument source has 5 columns or if slowness is defined at\n            nodes, L is a list of matrices and the number of matrices is equal\n            to the number of sources otherwise, L is a single csr_matrix\n\n        Notes\n        -----\n        If source has 3 columns:\n            - Columns correspond to x, y and z coordinates\n            - Origin time (t0) is 0 for all points\n        If source has 4 column""s:\n            - 1st column corresponds to origin times\n            - 2nd, 3rd & 4th columns correspond to x, y and z coordinates\n        If source has 5 columns:\n            - 1st column corresponds to event ID\n            - 2nd column corresponds to origin times\n            - 3rd, 4th & 5th columns correspond to x, y and z coordinates\n\n        For the latter case (5 columns), source and rcv should contain the same\n        number of rows, each row corresponding to a source-receiver pair.\n        For the 2 other cases, source and rcv can contain the same number of\n        rows, each row corresponding to a source-receiver pair, or the number\n        of rows may differ if aggregate_src is True or if all rows in source\n        are identical.\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_27raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_rcv = 0;
  PyObject *__pyx_v_slowness = 0;
  PyObject *__pyx_v_thread_no = 0;
  PyObject *__pyx_v_aggregate_src = 0;
  PyObject *__pyx_v_compute_L = 0;
  PyObject *__pyx_v_return_rays = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("raytrace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_rcv,&__pyx_n_s_slowness,&__pyx_n_s_thread_no,&__pyx_n_s_aggregate_src,&__pyx_n_s_compute_L,&__pyx_n_s_return_rays,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);

    /* "ttcrpy/tmesh.pyx":567
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 *                  aggregate_src=False, compute_L=False, return_rays=False):             # <<<<<<<<<<<<<<
 *         """
 *         raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, compute_L=False, return_rays=False) -> tt, rays, L
 */
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)Py_False);
    values[6] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rcv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 7, 1); __PYX_ERR(0, 566, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aggregate_src);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_L);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_rays);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "raytrace") < 0)) __PYX_ERR(0, 566, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_rcv = values[1];
    __pyx_v_slowness = values[2];
    __pyx_v_thread_no = values[3];
    __pyx_v_aggregate_src = values[4];
    __pyx_v_compute_L = values[5];
    __pyx_v_return_rays = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 566, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_26raytrace(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_source, __pyx_v_rcv, __pyx_v_slowness, __pyx_v_thread_no, __pyx_v_aggregate_src, __pyx_v_compute_L, __pyx_v_return_rays);

  /* "ttcrpy/tmesh.pyx":566
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, return_rays=False):
 *         """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_26raytrace(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_compute_L, PyObject *__pyx_v_return_rays) {
  PyObject *__pyx_v_evID = NULL;
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_t0 = NULL;
  PyObject *__pyx_v_eid = NULL;
  PyObject *__pyx_v_nTx = NULL;
  PyObject *__pyx_v_Tx = NULL;
  PyObject *__pyx_v_tmp = NULL;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vTx;
  std::vector<std::vector<ttcr::sxyz<double> > >  __pyx_v_vRx;
  std::vector<std::vector<double> >  __pyx_v_vt0;
  std::vector<std::vector<double> >  __pyx_v_vtt;
  std::vector<std::vector<std::vector<ttcr::sxyz<double> > > >  __pyx_v_r_data;
  std::vector<std::vector<std::vector<ttcr::siv<double> > > >  __pyx_v_l_data;
  std::vector<std::vector<std::vector<ttcr::sijv<double> > > >  __pyx_v_m_data;
  size_t __pyx_v_thread_nb;
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_v_n2;
  int __pyx_v_nt;
  PyObject *__pyx_v_iRx = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_ind = NULL;
  PyObject *__pyx_v_i0 = NULL;
  PyObject *__pyx_v_nn = NULL;
  PyObject *__pyx_v_ii = NULL;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_v_rays = NULL;
  PyObject *__pyx_v_L = NULL;
  PyObject *__pyx_v_nnz = NULL;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_v_ni;
  PyObject *__pyx_v_indptr = NULL;
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_k = NULL;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_v_MM;
  PyObject *__pyx_v_NN = NULL;
  PyObject *__pyx_v_j = NULL;
  CYTHON_UNUSED int __pyx_7genexpr__pyx_v_n;
  CYTHON_UNUSED int __pyx_8genexpr1__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_9;
  std::vector<std::vector<double> > ::size_type __pyx_t_10;
  std::vector<std::vector<std::vector<ttcr::sijv<double> > > > ::size_type __pyx_t_11;
  std::vector<std::vector<std::vector<ttcr::sxyz<double> > > > ::size_type __pyx_t_12;
  double __pyx_t_13;
  double __pyx_t_14;
  double __pyx_t_15;
  ttcr::sxyz<double>  __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  std::vector<double> ::size_type __pyx_t_18;
  int __pyx_t_19;
  long __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  size_t __pyx_t_23;
  std::vector<double> ::size_type __pyx_t_24;
  std::vector<double> ::size_type __pyx_t_25;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_26;
  std::vector<std::vector<ttcr::sxyz<double> > > ::size_type __pyx_t_27;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_28;
  int __pyx_t_29;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_30;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_31;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_32;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_33;
  std::vector<std::vector<ttcr::sijv<double> > > ::size_type __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  PyObject *(*__pyx_t_36)(PyObject *);
  std::vector<ttcr::sijv<double> > ::size_type __pyx_t_37;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raytrace", 0);

  /* "ttcrpy/tmesh.pyx":628
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":629
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if self.method == b'd' and aggregate_src:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 629, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":628
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":631
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.method == b'd' and aggregate_src:             # <<<<<<<<<<<<<<
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 */
  __pyx_t_4 = ((__pyx_v_self->method == 'd') != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 631, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":632
 * 
 *         if self.method == b'd' and aggregate_src:
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')             # <<<<<<<<<<<<<<
 * 
 *         evID = None
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 632, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":631
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if self.method == b'd' and aggregate_src:             # <<<<<<<<<<<<<<
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":634
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')
 * 
 *         evID = None             # <<<<<<<<<<<<<<
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_evID = Py_None;

  /* "ttcrpy/tmesh.pyx":635
 * 
 *         evID = None
 *         if source.shape[1] == 5:             # <<<<<<<<<<<<<<
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":636
 *         evID = None
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]             # <<<<<<<<<<<<<<
 *             t0 = source[:,1]
 *             evID = source[:,0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":637
 *         if source.shape[1] == 5:
 *             src = source[:,2:5]
 *             t0 = source[:,1]             # <<<<<<<<<<<<<<
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_t0 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":638
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 *             evID = source[:,0]             # <<<<<<<<<<<<<<
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_evID, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":639
 *             t0 = source[:,1]
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))             # <<<<<<<<<<<<<<
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sort); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_evID) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_evID);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_eid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":640
 *             evID = source[:,0]
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 3:
 *             src = source
 */
    __pyx_t_8 = PyObject_Length(__pyx_v_eid); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 640, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":635
 * 
 *         evID = None
 *         if source.shape[1] == 5:             # <<<<<<<<<<<<<<
 *             src = source[:,2:5]
 *             t0 = source[:,1]
 */
    goto __pyx_L9;
  }

  /* "ttcrpy/tmesh.pyx":641
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":642
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:
 *             src = source             # <<<<<<<<<<<<<<
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 */
    __Pyx_INCREF(__pyx_v_source);
    __pyx_v_src = __pyx_v_source;

    /* "ttcrpy/tmesh.pyx":643
 *         elif source.shape[1] == 3:
 *             src = source
 *             Tx = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_Tx = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/tmesh.pyx":644
 *             src = source
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))             # <<<<<<<<<<<<<<
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
    __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_t0 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "ttcrpy/tmesh.pyx":645
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":641
 *             eid = np.sort(np.unique(evID))
 *             nTx = len(eid)
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
    goto __pyx_L9;
  }

  /* "ttcrpy/tmesh.pyx":646
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:             # <<<<<<<<<<<<<<
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":647
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":648
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tmp = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":649
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]             # <<<<<<<<<<<<<<
 *             Tx = tmp[:,1:4]
 *             t0 = tmp[:,0]
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmp, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_nTx = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":650
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]             # <<<<<<<<<<<<<<
 *             t0 = tmp[:,0]
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_Tx = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":651
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:4]
 *             t0 = tmp[:,0]             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_t0 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":646
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:             # <<<<<<<<<<<<<<
 *             src = source[:,1:4]
 *             tmp = np.unique(source, axis=0)
 */
    goto __pyx_L9;
  }

  /* "ttcrpy/tmesh.pyx":653
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 653, __pyx_L1_error)
  }
  __pyx_L9:;

  /* "ttcrpy/tmesh.pyx":655
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L11_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":656
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 *             raise ValueError('src and rcv should be ndata x 3')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 656, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":655
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":658
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_src);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":659
 * 
 *         if self.is_outside(src):
 *             raise ValueError('Source point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(rcv):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 659, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":658
 *             raise ValueError('src and rcv should be ndata x 3')
 * 
 *         if self.is_outside(src):             # <<<<<<<<<<<<<<
 *             raise ValueError('Source point outside grid')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":661
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_outside); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_rcv) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_rcv);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":662
 * 
 *         if self.is_outside(rcv):
 *             raise ValueError('Receiver outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 662, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":661
 *             raise ValueError('Source point outside grid')
 * 
 *         if self.is_outside(rcv):             # <<<<<<<<<<<<<<
 *             raise ValueError('Receiver outside grid')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":664
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  __pyx_t_1 = (__pyx_v_slowness != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":665
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxyz[double]]] vTx
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":664
 *             raise ValueError('Receiver outside grid')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":679
 *         cdef int i, n, n2, nt
 * 
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 */
  __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L1_error)
  try {
    __pyx_v_vTx.resize(__pyx_t_9);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 679, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":680
 * 
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 */
  __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L1_error)
  try {
    __pyx_v_vRx.resize(__pyx_t_9);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 680, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":681
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)             # <<<<<<<<<<<<<<
 *         vtt.resize(nTx)
 *         if compute_L and self.cell_slowness:
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 681, __pyx_L1_error)
  try {
    __pyx_v_vt0.resize(__pyx_t_10);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 681, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":682
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)             # <<<<<<<<<<<<<<
 *         if compute_L and self.cell_slowness:
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L1_error)
  try {
    __pyx_v_vtt.resize(__pyx_t_10);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 682, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":683
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L and self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 *         elif compute_L and not self.cell_slowness:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L17_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  __pyx_t_4 = __pyx_t_1;
  __pyx_L17_bool_binop_done:;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":684
 *         vtt.resize(nTx)
 *         if compute_L and self.cell_slowness:
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')             # <<<<<<<<<<<<<<
 *         elif compute_L and not self.cell_slowness:
 *             m_data.resize(nTx)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 684, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":683
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if compute_L and self.cell_slowness:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 *         elif compute_L and not self.cell_slowness:
 */
  }

  /* "ttcrpy/tmesh.pyx":685
 *         if compute_L and self.cell_slowness:
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 *         elif compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             m_data.resize(nTx)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 685, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_1 = ((!(__pyx_v_self->cell_slowness != 0)) != 0);
  __pyx_t_4 = __pyx_t_1;
  __pyx_L19_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":686
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 *         elif compute_L and not self.cell_slowness:
 *             m_data.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         if return_rays:
 */
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 686, __pyx_L1_error)
    try {
      __pyx_v_m_data.resize(__pyx_t_11);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 686, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":685
 *         if compute_L and self.cell_slowness:
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
 *         elif compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             m_data.resize(nTx)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":688
 *             m_data.resize(nTx)
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 688, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":689
 * 
 *         if return_rays:
 *             r_data.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iRx = []
 */
    __pyx_t_12 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_12 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 689, __pyx_L1_error)
    try {
      __pyx_v_r_data.resize(__pyx_t_12);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 689, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":688
 *             m_data.resize(nTx)
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":691
 *             r_data.resize(nTx)
 * 
 *         iRx = []             # <<<<<<<<<<<<<<
 *         if evID is None:
 *             if nTx == 1:
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_iRx = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":692
 * 
 *         iRx = []
 *         if evID is None:             # <<<<<<<<<<<<<<
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 */
  __pyx_t_4 = (__pyx_v_evID == Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":693
 *         iRx = []
 *         if evID is None:
 *             if nTx == 1:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 */
    __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":694
 *         if evID is None:
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__21); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__22); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__23); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_13, __pyx_t_14, __pyx_t_15);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 694, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[0]).push_back(__pyx_t_16);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 694, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":695
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 */
      if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
        __pyx_t_3 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 695, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_17 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 695, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 695, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 695, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_17(__pyx_t_3);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 695, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "ttcrpy/tmesh.pyx":696
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        try {
          __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_15, __pyx_t_14, __pyx_t_13);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 696, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[0]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 696, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":695
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":697
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])             # <<<<<<<<<<<<<<
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_t0, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        (__pyx_v_vt0[0]).push_back(__pyx_t_13);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 697, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":698
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_t_5); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        (__pyx_v_vtt[0]).resize(__pyx_t_18);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 698, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":699
 *                 vt0[0].push_back(t0[0])
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *             elif aggregate_src:
 *                 for t in Tx:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_5); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":693
 *         iRx = []
 *         if evID is None:
 *             if nTx == 1:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 *                 for r in rcv:
 */
      goto __pyx_L23;
    }

    /* "ttcrpy/tmesh.pyx":700
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:             # <<<<<<<<<<<<<<
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 700, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":701
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 *                 for t in Tx:             # <<<<<<<<<<<<<<
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 */
      if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 701, __pyx_L1_error) }
      if (likely(PyList_CheckExact(__pyx_v_Tx)) || PyTuple_CheckExact(__pyx_v_Tx)) {
        __pyx_t_5 = __pyx_v_Tx; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_Tx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_17 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 701, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 701, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 701, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_17(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 701, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":702
 *             elif aggregate_src:
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))             # <<<<<<<<<<<<<<
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_13, __pyx_t_14, __pyx_t_15);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 702, __pyx_L1_error)
        }
        try {
          (__pyx_v_vTx[0]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 702, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":701
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:
 *                 for t in Tx:             # <<<<<<<<<<<<<<
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":703
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:             # <<<<<<<<<<<<<<
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 */
      if (likely(PyList_CheckExact(__pyx_v_t0)) || PyTuple_CheckExact(__pyx_v_t0)) {
        __pyx_t_5 = __pyx_v_t0; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_t0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_17 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 703, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_17(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 703, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":704
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:
 *                     vt0[0].push_back(t)             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_v_t); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L1_error)
        try {
          (__pyx_v_vt0[0]).push_back(__pyx_t_15);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 704, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":703
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 *                 for t in t0:             # <<<<<<<<<<<<<<
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":705
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 */
      if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
        __pyx_t_5 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_17 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 705, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_17(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 705, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":706
 *                     vt0[0].push_back(t)
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_15, __pyx_t_14, __pyx_t_13);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 706, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[0]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 706, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":705
 *                 for t in t0:
 *                     vt0[0].push_back(t)
 *                 for r in rcv:             # <<<<<<<<<<<<<<
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":707
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *                 nTx = 1
 *                 iRx.append(np.arange(rcv.shape[0]))
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_18 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_18 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        (__pyx_v_vtt[0]).resize(__pyx_t_18);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 707, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":708
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1             # <<<<<<<<<<<<<<
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 */
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_DECREF_SET(__pyx_v_nTx, __pyx_int_1);

      /* "ttcrpy/tmesh.pyx":709
 *                 vtt[0].resize(rcv.shape[0])
 *                 nTx = 1
 *                 iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *             else:
 *                 if src.shape != rcv.shape:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_2); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":700
 *                 vtt[0].resize(rcv.shape[0])
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             elif aggregate_src:             # <<<<<<<<<<<<<<
 *                 for t in Tx:
 *                     vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
 */
      goto __pyx_L23;
    }

    /* "ttcrpy/tmesh.pyx":711
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 *                 if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_t_1)) {

        /* "ttcrpy/tmesh.pyx":712
 *             else:
 *                 if src.shape != rcv.shape:
 *                     raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *                 for n in range(nTx):
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 712, __pyx_L1_error)

        /* "ttcrpy/tmesh.pyx":711
 *                 iRx.append(np.arange(rcv.shape[0]))
 *             else:
 *                 if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 */
      }

      /* "ttcrpy/tmesh.pyx":714
 *                     raise ValueError('src and rcv should be of equal size')
 * 
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])
 */
      __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_v_n = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":715
 * 
 *                 for n in range(nTx):
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3             # <<<<<<<<<<<<<<
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sum); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 715, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_slice_);
        __Pyx_GIVEREF(__pyx_slice_);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice_);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_v_src, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_1) < 0) __PYX_ERR(0, 715, __pyx_L1_error)
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ind, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":716
 *                 for n in range(nTx):
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])             # <<<<<<<<<<<<<<
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_ind) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ind);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_3); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 716, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":717
 *                     ind = np.sum(Tx[n,:] == src, axis=1) == 3
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))             # <<<<<<<<<<<<<<
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:
 */
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 717, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 717, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_1);
        __Pyx_GIVEREF(__pyx_int_1);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_Tx)) { __Pyx_RaiseUnboundLocalError("Tx"); __PYX_ERR(0, 717, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_13, __pyx_t_14, __pyx_t_15);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 717, __pyx_L1_error)
        }
        try {
          (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 717, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":718
 *                     iRx.append(np.nonzero(ind)[0])
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])             # <<<<<<<<<<<<<<
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_t0, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 718, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_15);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 718, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":719
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())
 */
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_ind);
        __Pyx_GIVEREF(__pyx_v_ind);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_ind);
        __Pyx_INCREF(__pyx_slice_);
        __Pyx_GIVEREF(__pyx_slice_);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_slice_);
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 719, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_17 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 719, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 719, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 719, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_17(__pyx_t_3);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 719, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/tmesh.pyx":720
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 */
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          try {
            __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_15, __pyx_t_14, __pyx_t_13);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 720, __pyx_L1_error)
          }
          try {
            (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_16);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 720, __pyx_L1_error)
          }

          /* "ttcrpy/tmesh.pyx":719
 *                     vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
 *                     vt0[n].push_back(t0[n])
 *                     for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":721
 *                     for r in rcv[ind,:]:
 *                         vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                     vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 *         else:
 *             if src.shape != rcv.shape:
 */
        try {
          (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 721, __pyx_L1_error)
        }
      }
    }
    __pyx_L23:;

    /* "ttcrpy/tmesh.pyx":692
 * 
 *         iRx = []
 *         if evID is None:             # <<<<<<<<<<<<<<
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
 */
    goto __pyx_L22;
  }

  /* "ttcrpy/tmesh.pyx":723
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 723, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 723, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 723, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "ttcrpy/tmesh.pyx":724
 *         else:
 *             if src.shape != rcv.shape:
 *                 raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *             i0 = 0
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 724, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 724, __pyx_L1_error)

      /* "ttcrpy/tmesh.pyx":723
 *                     vtt[n].resize(vRx[n].size())
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
    }

    /* "ttcrpy/tmesh.pyx":726
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 *             i0 = 0             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_i0 = __pyx_int_0;

    /* "ttcrpy/tmesh.pyx":727
 * 
 *             i0 = 0
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:
 */
    __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L1_error)
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
      __pyx_v_n = __pyx_t_22;

      /* "ttcrpy/tmesh.pyx":728
 *             i0 = 0
 *             for n in range(nTx):
 *                 for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_evID, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_17 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 728, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 728, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 728, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_17(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 728, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_nn, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":729
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                         i0 = nn
 *                         break
 */
        if (unlikely(!__pyx_v_eid)) { __Pyx_RaiseUnboundLocalError("eid"); __PYX_ERR(0, 729, __pyx_L1_error) }
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_eid, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_evID, __pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 729, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 729, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 729, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_1) {

          /* "ttcrpy/tmesh.pyx":730
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn             # <<<<<<<<<<<<<<
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 */
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_DECREF_SET(__pyx_v_i0, __pyx_v_nn);

          /* "ttcrpy/tmesh.pyx":731
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn
 *                         break             # <<<<<<<<<<<<<<
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 *                 vt0[n].push_back(t0[i0])
 */
          goto __pyx_L41_break;

          /* "ttcrpy/tmesh.pyx":729
 *             for n in range(nTx):
 *                 for nn in range(evID.size):
 *                     if eid[n] == evID[nn]:             # <<<<<<<<<<<<<<
 *                         i0 = nn
 *                         break
 */
        }

        /* "ttcrpy/tmesh.pyx":728
 *             i0 = 0
 *             for n in range(nTx):
 *                 for nn in range(evID.size):             # <<<<<<<<<<<<<<
 *                     if eid[n] == evID[nn]:
 *                         i0 = nn
 */
      }
      __pyx_L41_break:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":732
 *                         i0 = nn
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))             # <<<<<<<<<<<<<<
 *                 vt0[n].push_back(t0[i0])
 * 
 */
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_i0);
      __Pyx_GIVEREF(__pyx_v_i0);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_i0);
      __Pyx_GIVEREF(__pyx_v_i0);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_i0);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_1);
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_i0);
      __Pyx_GIVEREF(__pyx_v_i0);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i0);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      try {
        __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_13, __pyx_t_14, __pyx_t_15);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 732, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_16);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 732, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":733
 *                         break
 *                 vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
 *                 vt0[n].push_back(t0[i0])             # <<<<<<<<<<<<<<
 * 
 *             for i in eid:
 */
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_t0, __pyx_v_i0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      try {
        (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_15);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 733, __pyx_L1_error)
      }
    }

    /* "ttcrpy/tmesh.pyx":735
 *                 vt0[n].push_back(t0[i0])
 * 
 *             for i in eid:             # <<<<<<<<<<<<<<
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])
 */
    if (unlikely(!__pyx_v_eid)) { __Pyx_RaiseUnboundLocalError("eid"); __PYX_ERR(0, 735, __pyx_L1_error) }
    if (likely(PyList_CheckExact(__pyx_v_eid)) || PyTuple_CheckExact(__pyx_v_eid)) {
      __pyx_t_7 = __pyx_v_eid; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
      __pyx_t_17 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_eid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_17 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 735, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_17)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 735, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 735, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_17(__pyx_t_7);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 735, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_t_22 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_22 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_i = __pyx_t_22;

      /* "ttcrpy/tmesh.pyx":736
 * 
 *             for i in eid:
 *                 ii = evID == i             # <<<<<<<<<<<<<<
 *                 iRx.append(np.nonzero(ii)[0])
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 736, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_evID, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 736, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ii, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":737
 *             for i in eid:
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nTx):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_v_ii) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ii);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_2); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":735
 *                 vt0[n].push_back(t0[i0])
 * 
 *             for i in eid:             # <<<<<<<<<<<<<<
 *                 ii = evID == i
 *                 iRx.append(np.nonzero(ii)[0])
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "ttcrpy/tmesh.pyx":739
 *                 iRx.append(np.nonzero(ii)[0])
 * 
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 */
    __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L1_error)
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
      __pyx_v_n = __pyx_t_22;

      /* "ttcrpy/tmesh.pyx":740
 * 
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())
 */
      __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
      __Pyx_INCREF(__pyx_slice_);
      __Pyx_GIVEREF(__pyx_slice_);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice_);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
        __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
        __pyx_t_17 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 740, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_17 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 740, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      for (;;) {
        if (likely(!__pyx_t_17)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 740, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 740, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 740, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 740, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_17(__pyx_t_2);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 740, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "ttcrpy/tmesh.pyx":741
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))             # <<<<<<<<<<<<<<
 *                 vtt[n].resize(vRx[n].size())
 * 
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        try {
          __pyx_t_16 = ttcr::sxyz<double> (__pyx_t_15, __pyx_t_14, __pyx_t_13);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 741, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_16);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 741, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":740
 * 
 *             for n in range(nTx):
 *                 for r in rcv[iRx[n],:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":742
 *                 for r in rcv[iRx[n],:]:
 *                     vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
 *                 vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 * 
 *         tt = np.zeros((rcv.shape[0],))
 */
      try {
        (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 742, __pyx_L1_error)
      }
    }
  }
  __pyx_L22:;

  /* "ttcrpy/tmesh.pyx":744
 *                 vtt[n].resize(vRx[n].size())
 * 
 *         tt = np.zeros((rcv.shape[0],))             # <<<<<<<<<<<<<<
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":745
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_nTx, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L50_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_self->_n_threads == 1) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L50_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":746
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L53_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_L53_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":747
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:
 */
      __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L1_error)
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_v_n = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":748
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 748, __pyx_L1_error)
        }
      }

      /* "ttcrpy/tmesh.pyx":746
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
      goto __pyx_L52;
    }

    /* "ttcrpy/tmesh.pyx":749
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     # TODO: implement this in C++ codebase!
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 749, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L57_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 749, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L57_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":750
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     # TODO: implement this in C++ codebase!
 *                     for n in range(nTx):
 */
      __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
      if (__pyx_t_1) {

        /* "ttcrpy/tmesh.pyx":752
 *                 if self.cell_slowness:
 *                     # TODO: implement this in C++ codebase!
 *                     for n in range(nTx):             # <<<<<<<<<<<<<<
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *                 else:
 */
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 752, __pyx_L1_error)
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_n = __pyx_t_22;

          /* "ttcrpy/tmesh.pyx":753
 *                     # TODO: implement this in C++ codebase!
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *                 else:
 *                     for n in range(nTx):
 */
          try {
            __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 753, __pyx_L1_error)
          }
        }

        /* "ttcrpy/tmesh.pyx":750
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     # TODO: implement this in C++ codebase!
 *                     for n in range(nTx):
 */
        goto __pyx_L59;
      }

      /* "ttcrpy/tmesh.pyx":755
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
 *                 else:
 *                     for n in range(nTx):             # <<<<<<<<<<<<<<
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L:
 */
      /*else*/ {
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L1_error)
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_n = __pyx_t_22;

          /* "ttcrpy/tmesh.pyx":756
 *                 else:
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 if self.cell_slowness:
 */
          try {
            __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), (__pyx_v_m_data[__pyx_v_n]), 0);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 756, __pyx_L1_error)
          }
        }
      }
      __pyx_L59:;

      /* "ttcrpy/tmesh.pyx":749
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     # TODO: implement this in C++ codebase!
 */
      goto __pyx_L52;
    }

    /* "ttcrpy/tmesh.pyx":757
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     for n in range(nTx):
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 757, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":758
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
      __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
      if (__pyx_t_1) {

        /* "ttcrpy/tmesh.pyx":759
 *             elif compute_L:
 *                 if self.cell_slowness:
 *                     for n in range(nTx):             # <<<<<<<<<<<<<<
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *                 else:
 */
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 759, __pyx_L1_error)
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_n = __pyx_t_22;

          /* "ttcrpy/tmesh.pyx":760
 *                 if self.cell_slowness:
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)             # <<<<<<<<<<<<<<
 *                 else:
 *                     for n in range(nTx):
 */
          try {
            __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_l_data[__pyx_v_n]), 0);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 760, __pyx_L1_error)
          }
        }

        /* "ttcrpy/tmesh.pyx":758
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 */
        goto __pyx_L64;
      }

      /* "ttcrpy/tmesh.pyx":762
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
 *                 else:
 *                     for n in range(nTx):             # <<<<<<<<<<<<<<
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 *             else:
 */
      /*else*/ {
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L1_error)
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_n = __pyx_t_22;

          /* "ttcrpy/tmesh.pyx":763
 *                 else:
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 for n in range(nTx):
 */
          try {
            __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_m_data[__pyx_v_n]), 0);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 763, __pyx_L1_error)
          }
        }
      }
      __pyx_L64:;

      /* "ttcrpy/tmesh.pyx":757
 *                     for n in range(nTx):
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     for n in range(nTx):
 */
      goto __pyx_L52;
    }

    /* "ttcrpy/tmesh.pyx":765
 *                         self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
 *             else:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 */
    /*else*/ {
      __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L1_error)
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_v_n = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":766
 *             else:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)             # <<<<<<<<<<<<<<
 * 
 *         elif thread_no is not None:
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 766, __pyx_L1_error)
        }
      }
    }
    __pyx_L52:;

    /* "ttcrpy/tmesh.pyx":745
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if compute_L==False and return_rays==False:
 *                 for n in range(nTx):
 */
    goto __pyx_L49;
  }

  /* "ttcrpy/tmesh.pyx":768
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  __pyx_t_1 = (__pyx_v_thread_no != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":770
 *         elif thread_no is not None:
 *             # we should be here for just one event
 *             assert nTx == 1             # <<<<<<<<<<<<<<
 *             # normally we should not need to compute M or L
 *             assert compute_L is False
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 770, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 770, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 770, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/tmesh.pyx":772
 *             assert nTx == 1
 *             # normally we should not need to compute M or L
 *             assert compute_L is False             # <<<<<<<<<<<<<<
 *             thread_nb = thread_no
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_4 = (__pyx_v_compute_L == Py_False);
      if (unlikely(!(__pyx_t_4 != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 772, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/tmesh.pyx":773
 *             # normally we should not need to compute M or L
 *             assert compute_L is False
 *             thread_nb = thread_no             # <<<<<<<<<<<<<<
 * 
 *             if return_rays:
 */
    __pyx_t_23 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_23 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 773, __pyx_L1_error)
    __pyx_v_thread_nb = __pyx_t_23;

    /* "ttcrpy/tmesh.pyx":775
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 775, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":776
 * 
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), (__pyx_v_r_data[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 776, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":777
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 */
      __pyx_t_24 = (__pyx_v_vtt[0]).size();
      __pyx_t_25 = __pyx_t_24;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_25; __pyx_t_22+=1) {
        __pyx_v_nt = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":778
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 778, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/tmesh.pyx":779
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_rays = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":780
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 *                 for n2 in range(vRx.size()):             # <<<<<<<<<<<<<<
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):
 */
      __pyx_t_26 = __pyx_v_vRx.size();
      __pyx_t_27 = __pyx_t_26;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_27; __pyx_t_22+=1) {
        __pyx_v_n2 = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":781
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[0])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
        __Pyx_INCREF(__pyx_int_3);
        __Pyx_GIVEREF(__pyx_int_3);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_3);
        __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":782
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y
 */
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[0])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 782, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 782, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
          __pyx_t_3 = __pyx_t_7; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_17 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 782, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_7 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 782, __pyx_L1_error)
              #else
              __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 782, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              #endif
            } else {
              if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 782, __pyx_L1_error)
              #else
              __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 782, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              #endif
            }
          } else {
            __pyx_t_7 = __pyx_t_17(__pyx_t_3);
            if (unlikely(!__pyx_t_7)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 782, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_XDECREF_SET(__pyx_v_nn, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "ttcrpy/tmesh.pyx":783
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L1_error)
          __pyx_t_7 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_t_28]).x); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_5, __pyx_t_7) < 0)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "ttcrpy/tmesh.pyx":784
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y             # <<<<<<<<<<<<<<
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 784, __pyx_L1_error)
          __pyx_t_7 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_t_28]).y); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 784, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 784, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_5, __pyx_t_7) < 0)) __PYX_ERR(0, 784, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "ttcrpy/tmesh.pyx":785
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z             # <<<<<<<<<<<<<<
 *                     rays.append(r)
 *                 return tt, rays
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 785, __pyx_L1_error)
          __pyx_t_7 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_t_28]).z); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_2);
          __Pyx_GIVEREF(__pyx_int_2);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_5, __pyx_t_7) < 0)) __PYX_ERR(0, 785, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "ttcrpy/tmesh.pyx":782
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 3))
 *                     for nn in range(r_data[0][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].y
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":786
 *                         r[nn, 1] = r_data[0][n2][nn].y
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)             # <<<<<<<<<<<<<<
 *                 return tt, rays
 * 
 */
        __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_rays, __pyx_v_r); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 786, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":787
 *                         r[nn, 2] = r_data[0][n2][nn].z
 *                     rays.append(r)
 *                 return tt, rays             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_tt);
      __Pyx_GIVEREF(__pyx_v_tt);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
      __Pyx_INCREF(__pyx_v_rays);
      __Pyx_GIVEREF(__pyx_v_rays);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ttcrpy/tmesh.pyx":775
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    }

    /* "ttcrpy/tmesh.pyx":790
 * 
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 790, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":791
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 return tt
 */
      __pyx_t_24 = (__pyx_v_vtt[0]).size();
      __pyx_t_25 = __pyx_t_24;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_25; __pyx_t_22+=1) {
        __pyx_v_nt = __pyx_t_22;

        /* "ttcrpy/tmesh.pyx":792
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 return tt
 * 
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 792, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 792, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/tmesh.pyx":793
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 return tt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_tt);
      __pyx_r = __pyx_v_tt;
      goto __pyx_L0;
    }

    /* "ttcrpy/tmesh.pyx":768
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  }

  /* "ttcrpy/tmesh.pyx":796
 * 
 *         else:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 */
  /*else*/ {
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L81_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_L81_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":797
 *         else:
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)             # <<<<<<<<<<<<<<
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 797, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":796
 * 
 *         else:
 *             if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 */
      goto __pyx_L80;
    }

    /* "ttcrpy/tmesh.pyx":798
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L83_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L83_bool_binop_done:;
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":799
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *                 else:
 */
      __pyx_t_4 = (__pyx_v_self->cell_slowness != 0);
      if (__pyx_t_4) {

        /* "ttcrpy/tmesh.pyx":800
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 */
        try {
          __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data, __pyx_v_l_data);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 800, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":799
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *                 else:
 */
        goto __pyx_L85;
      }

      /* "ttcrpy/tmesh.pyx":802
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)             # <<<<<<<<<<<<<<
 *             elif compute_L:
 *                 if self.cell_slowness:
 */
      /*else*/ {
        try {
          __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data, __pyx_v_m_data);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 802, __pyx_L1_error)
        }
      }
      __pyx_L85:;

      /* "ttcrpy/tmesh.pyx":798
 *             if compute_L==False and return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
 */
      goto __pyx_L80;
    }

    /* "ttcrpy/tmesh.pyx":803
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 803, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":804
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *                 else:
 */
      __pyx_t_4 = (__pyx_v_self->cell_slowness != 0);
      if (__pyx_t_4) {

        /* "ttcrpy/tmesh.pyx":805
 *             elif compute_L:
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 */
        try {
          __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_l_data);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 805, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":804
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L:
 *                 if self.cell_slowness:             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *                 else:
 */
        goto __pyx_L86;
      }

      /* "ttcrpy/tmesh.pyx":807
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 */
      /*else*/ {
        try {
          __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_m_data);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 807, __pyx_L1_error)
        }
      }
      __pyx_L86:;

      /* "ttcrpy/tmesh.pyx":803
 *                 else:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
 *             elif compute_L:             # <<<<<<<<<<<<<<
 *                 if self.cell_slowness:
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
 */
      goto __pyx_L80;
    }

    /* "ttcrpy/tmesh.pyx":809
 *                     self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 809, __pyx_L1_error)
      }
    }
    __pyx_L80:;
  }
  __pyx_L49:;

  /* "ttcrpy/tmesh.pyx":811
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 */
  __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 811, __pyx_L1_error)
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
    __pyx_v_n = __pyx_t_22;

    /* "ttcrpy/tmesh.pyx":812
 * 
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 */
    __pyx_t_24 = (__pyx_v_vtt[__pyx_v_n]).size();
    __pyx_t_25 = __pyx_t_24;
    for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_25; __pyx_t_29+=1) {
      __pyx_v_nt = __pyx_t_29;

      /* "ttcrpy/tmesh.pyx":813
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]             # <<<<<<<<<<<<<<
 * 
 *         if return_rays:
 */
      __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[__pyx_v_n])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tt, __pyx_t_5, __pyx_t_3) < 0)) __PYX_ERR(0, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":815
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 815, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":816
 * 
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_t_7); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_7genexpr__pyx_v_n = __pyx_t_22;
        __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyList_SET_ITEM(__pyx_t_7, 0, __pyx_float_0_0);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    } /* exit inner scope */
    __pyx_v_rays = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":817
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 */
    __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 817, __pyx_L1_error)
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
      __pyx_v_n = __pyx_t_22;

      /* "ttcrpy/tmesh.pyx":818
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_30 = (__pyx_v_vRx[__pyx_v_n]).size();
        __pyx_t_31 = __pyx_t_30;
        for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_31; __pyx_t_29+=1) {
          __pyx_8genexpr1__pyx_v_i = __pyx_t_29;
          __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 818, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_float_0_0);
          __Pyx_GIVEREF(__pyx_float_0_0);
          PyList_SET_ITEM(__pyx_t_7, 0, __pyx_float_0_0);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 818, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":819
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):             # <<<<<<<<<<<<<<
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):
 */
      __pyx_t_30 = (__pyx_v_vRx[__pyx_v_n]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_31; __pyx_t_29+=1) {
        __pyx_v_n2 = __pyx_t_29;

        /* "ttcrpy/tmesh.pyx":820
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
        __Pyx_INCREF(__pyx_int_3);
        __Pyx_GIVEREF(__pyx_int_3);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
        __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_v_r, __pyx_v_n2, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 820, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":821
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 */
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_3 = __pyx_t_5; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_17 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 821, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 821, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 821, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_17(__pyx_t_3);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 821, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_nn, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "ttcrpy/tmesh.pyx":822
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_t_28]).x); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_0);
          if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_7, __pyx_t_5) < 0)) __PYX_ERR(0, 822, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/tmesh.pyx":823
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_t_28]).y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
          if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_t_2, __pyx_t_5) < 0)) __PYX_ERR(0, 823, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/tmesh.pyx":824
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]
 */
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L1_error)
          __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_t_28]).z); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 824, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_nn);
          __Pyx_GIVEREF(__pyx_v_nn);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nn);
          __Pyx_INCREF(__pyx_int_2);
          __Pyx_GIVEREF(__pyx_int_2);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_2);
          if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_7, __pyx_t_5) < 0)) __PYX_ERR(0, 824, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "ttcrpy/tmesh.pyx":821
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 3))
 *                     for nn in range(r_data[n][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/tmesh.pyx":825
 *                         r[n2][nn, 1] = r_data[n][n2][nn].y
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 */
      __pyx_t_24 = (__pyx_v_vtt[__pyx_v_n]).size();
      __pyx_t_25 = __pyx_t_24;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_25; __pyx_t_29+=1) {
        __pyx_v_nt = __pyx_t_29;

        /* "ttcrpy/tmesh.pyx":826
 *                         r[n2][nn, 2] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]             # <<<<<<<<<<<<<<
 * 
 *         if compute_L and not self.cell_slowness:
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_rays, __pyx_t_7, __pyx_t_3) < 0)) __PYX_ERR(0, 826, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }

    /* "ttcrpy/tmesh.pyx":815
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  }

  /* "ttcrpy/tmesh.pyx":828
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             # we return array of matrices, one for each event
 *             L = []
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L105_bool_binop_done;
  }
  __pyx_t_1 = ((!(__pyx_v_self->cell_slowness != 0)) != 0);
  __pyx_t_4 = __pyx_t_1;
  __pyx_L105_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":830
 *         if compute_L and not self.cell_slowness:
 *             # we return array of matrices, one for each event
 *             L = []             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 nnz = 0
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_L = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":831
 *             # we return array of matrices, one for each event
 *             L = []
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):
 */
    __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 831, __pyx_L1_error)
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
      __pyx_v_n = __pyx_t_22;

      /* "ttcrpy/tmesh.pyx":832
 *             L = []
 *             for n in range(nTx):
 *                 nnz = 0             # <<<<<<<<<<<<<<
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_nnz, __pyx_int_0);

      /* "ttcrpy/tmesh.pyx":833
 *             for n in range(nTx):
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):             # <<<<<<<<<<<<<<
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 */
      __pyx_t_32 = (__pyx_v_m_data[__pyx_v_n]).size();
      __pyx_t_33 = __pyx_t_32;
      for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
        __pyx_v_ni = __pyx_t_34;

        /* "ttcrpy/tmesh.pyx":834
 *                 nnz = 0
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()             # <<<<<<<<<<<<<<
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 */
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(((__pyx_v_m_data[__pyx_v_n])[__pyx_v_ni]).size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_nnz, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 834, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_nnz, __pyx_t_7);
        __pyx_t_7 = 0;
      }

      /* "ttcrpy/tmesh.pyx":835
 *                 for ni in range(m_data[n].size()):
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(((__pyx_v_vRx[__pyx_v_n]).size() + 1)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indptr, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "ttcrpy/tmesh.pyx":836
 *                     nnz += m_data[n][ni].size()
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)             # <<<<<<<<<<<<<<
 *                 val = np.empty((nnz,))
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_nnz);
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_indices, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":837
 *                 indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
 *                 indices = np.empty((nnz,), dtype=np.int64)
 *                 val = np.empty((nnz,))             # <<<<<<<<<<<<<<
 * 
 *                 k = 0
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_nnz);
      __Pyx_GIVEREF(__pyx_v_nnz);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_nnz);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":839
 *                 val = np.empty((nnz,))
 * 
 *                 k = 0             # <<<<<<<<<<<<<<
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_int_0);

      /* "ttcrpy/tmesh.pyx":840
 * 
 *                 k = 0
 *                 MM = vRx[n].size()             # <<<<<<<<<<<<<<
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):
 */
      __pyx_v_MM = (__pyx_v_vRx[__pyx_v_n]).size();

      /* "ttcrpy/tmesh.pyx":841
 *                 k = 0
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()             # <<<<<<<<<<<<<<
 *                 for i in range(MM):
 *                     indptr[i] = k
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_NN, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":842
 *                 MM = vRx[n].size()
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):             # <<<<<<<<<<<<<<
 *                     indptr[i] = k
 *                     for j in range(NN):
 */
      __pyx_t_30 = __pyx_v_MM;
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_31; __pyx_t_29+=1) {
        __pyx_v_i = __pyx_t_29;

        /* "ttcrpy/tmesh.pyx":843
 *                 NN = self.get_number_of_nodes()
 *                 for i in range(MM):
 *                     indptr[i] = k             # <<<<<<<<<<<<<<
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 */
        if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_i, __pyx_v_k, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 843, __pyx_L1_error)

        /* "ttcrpy/tmesh.pyx":844
 *                 for i in range(MM):
 *                     indptr[i] = k
 *                     for j in range(NN):             # <<<<<<<<<<<<<<
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 */
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_NN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_7 = __pyx_t_2; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_17 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 844, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_7))) {
              if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 844, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 844, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_17(__pyx_t_7);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 844, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "ttcrpy/tmesh.pyx":845
 *                     indptr[i] = k
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j
 */
          __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 845, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
            __pyx_t_2 = __pyx_t_6; __Pyx_INCREF(__pyx_t_2); __pyx_t_35 = 0;
            __pyx_t_36 = NULL;
          } else {
            __pyx_t_35 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_36 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 845, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          for (;;) {
            if (likely(!__pyx_t_36)) {
              if (likely(PyList_CheckExact(__pyx_t_2))) {
                if (__pyx_t_35 >= PyList_GET_SIZE(__pyx_t_2)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_35); __Pyx_INCREF(__pyx_t_6); __pyx_t_35++; if (unlikely(0 < 0)) __PYX_ERR(0, 845, __pyx_L1_error)
                #else
                __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_35); __pyx_t_35++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 845, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              } else {
                if (__pyx_t_35 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_35); __Pyx_INCREF(__pyx_t_6); __pyx_t_35++; if (unlikely(0 < 0)) __PYX_ERR(0, 845, __pyx_L1_error)
                #else
                __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_35); __pyx_t_35++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 845, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              }
            } else {
              __pyx_t_6 = __pyx_t_36(__pyx_t_2);
              if (unlikely(!__pyx_t_6)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 845, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_6);
            }
            __Pyx_XDECREF_SET(__pyx_v_nn, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "ttcrpy/tmesh.pyx":846
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 */
            __pyx_t_37 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_37 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
            __pyx_t_1 = (((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_t_37]).i == __pyx_v_i) != 0);
            if (__pyx_t_1) {
            } else {
              __pyx_t_4 = __pyx_t_1;
              goto __pyx_L118_bool_binop_done;
            }
            __pyx_t_37 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_37 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
            __pyx_t_6 = __Pyx_PyInt_FromSize_t((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_t_37]).j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 846, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_v_j, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 846, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 846, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_4 = __pyx_t_1;
            __pyx_L118_bool_binop_done:;
            if (__pyx_t_4) {

              /* "ttcrpy/tmesh.pyx":847
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j             # <<<<<<<<<<<<<<
 *                                 val[k] = m_data[n][i][nn].v
 *                                 k += 1
 */
              if (unlikely(PyObject_SetItem(__pyx_v_indices, __pyx_v_k, __pyx_v_j) < 0)) __PYX_ERR(0, 847, __pyx_L1_error)

              /* "ttcrpy/tmesh.pyx":848
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v             # <<<<<<<<<<<<<<
 *                                 k += 1
 * 
 */
              __pyx_t_37 = __Pyx_PyInt_As_size_t(__pyx_v_nn); if (unlikely((__pyx_t_37 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 848, __pyx_L1_error)
              __pyx_t_5 = PyFloat_FromDouble((((__pyx_v_m_data[__pyx_v_n])[__pyx_v_i])[__pyx_t_37]).v); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 848, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (unlikely(PyObject_SetItem(__pyx_v_val, __pyx_v_k, __pyx_t_5) < 0)) __PYX_ERR(0, 848, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "ttcrpy/tmesh.pyx":849
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 *                                 k += 1             # <<<<<<<<<<<<<<
 * 
 *                 indptr[MM] = k
 */
              __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_k, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 849, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF_SET(__pyx_v_k, __pyx_t_5);
              __pyx_t_5 = 0;

              /* "ttcrpy/tmesh.pyx":846
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:             # <<<<<<<<<<<<<<
 *                                 indices[k] = j
 *                                 val[k] = m_data[n][i][nn].v
 */
            }

            /* "ttcrpy/tmesh.pyx":845
 *                     indptr[i] = k
 *                     for j in range(NN):
 *                         for nn in range(m_data[n][i].size()):             # <<<<<<<<<<<<<<
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 *                                 indices[k] = j
 */
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "ttcrpy/tmesh.pyx":844
 *                 for i in range(MM):
 *                     indptr[i] = k
 *                     for j in range(NN):             # <<<<<<<<<<<<<<
 *                         for nn in range(m_data[n][i].size()):
 *                             if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
 */
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }

      /* "ttcrpy/tmesh.pyx":851
 *                                 k += 1
 * 
 *                 indptr[MM] = k             # <<<<<<<<<<<<<<
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 */
      if (unlikely(__Pyx_SetItemInt(__pyx_v_indptr, __pyx_v_MM, __pyx_v_k, std::vector<ttcr::sxyz<double> > ::size_type, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1) < 0)) __PYX_ERR(0, 851, __pyx_L1_error)

      /* "ttcrpy/tmesh.pyx":852
 * 
 *                 indptr[MM] = k
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )             # <<<<<<<<<<<<<<
 * 
 *         if compute_L==False and return_rays==False:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_sp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_val);
      __Pyx_GIVEREF(__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_val);
      __Pyx_INCREF(__pyx_v_indices);
      __Pyx_GIVEREF(__pyx_v_indices);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_indices);
      __Pyx_INCREF(__pyx_v_indptr);
      __Pyx_GIVEREF(__pyx_v_indptr);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_indptr);
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_MM); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_NN);
      __Pyx_GIVEREF(__pyx_v_NN);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_NN);
      __pyx_t_6 = 0;
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_L, __pyx_t_3); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 852, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "ttcrpy/tmesh.pyx":828
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if compute_L and not self.cell_slowness:             # <<<<<<<<<<<<<<
 *             # we return array of matrices, one for each event
 *             L = []
 */
  }

  /* "ttcrpy/tmesh.pyx":854
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *         if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_L and return_rays:
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_compute_L, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L121_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_t_1;
  __pyx_L121_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":855
 * 
 *         if compute_L==False and return_rays==False:
 *             return tt             # <<<<<<<<<<<<<<
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_tt);
    __pyx_r = __pyx_v_tt;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":854
 *                 L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )
 * 
 *         if compute_L==False and return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         elif compute_L and return_rays:
 */
  }

  /* "ttcrpy/tmesh.pyx":856
 *         if compute_L==False and return_rays==False:
 *             return tt
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 856, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L123_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_1;
  __pyx_L123_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":857
 *             return tt
 *         elif compute_L and return_rays:
 *             return tt, rays, L             # <<<<<<<<<<<<<<
 *         elif compute_L:
 *             return tt, L
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 857, __pyx_L1_error) }
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 857, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_L);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":856
 *         if compute_L==False and return_rays==False:
 *             return tt
 *         elif compute_L and return_rays:             # <<<<<<<<<<<<<<
 *             return tt, rays, L
 *         elif compute_L:
 */
  }

  /* "ttcrpy/tmesh.pyx":858
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_L); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 858, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":859
 *             return tt, rays, L
 *         elif compute_L:
 *             return tt, L             # <<<<<<<<<<<<<<
 *         else:
 *             return tt, rays
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_L)) { __Pyx_RaiseUnboundLocalError("L"); __PYX_ERR(0, 859, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_L);
    __Pyx_GIVEREF(__pyx_v_L);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_L);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":858
 *         elif compute_L and return_rays:
 *             return tt, rays, L
 *         elif compute_L:             # <<<<<<<<<<<<<<
 *             return tt, L
 *         else:
 */
  }

  /* "ttcrpy/tmesh.pyx":861
 *             return tt, L
 *         else:
 *             return tt, rays             # <<<<<<<<<<<<<<
 * 
 *     def to_vtk(self, fields, filename):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 861, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/tmesh.pyx":566
 *         return s0
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, compute_L=False, return_rays=False):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_evID);
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_t0);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_nTx);
  __Pyx_XDECREF(__pyx_v_Tx);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_iRx);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XDECREF(__pyx_v_i0);
  __Pyx_XDECREF(__pyx_v_nn);
  __Pyx_XDECREF(__pyx_v_ii);
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_rays);
  __Pyx_XDECREF(__pyx_v_L);
  __Pyx_XDECREF(__pyx_v_nnz);
  __Pyx_XDECREF(__pyx_v_indptr);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_NN);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":863
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_29to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_28to_vtk[] = "\n        to_vtk(fields, filename)\n\n        Save mesh variables and/or raypaths to VTK format\n\n        Parameters\n        ----------\n        fields: dict\n            dict of variables to save to file. Variables should be np.ndarray of\n            size equal to either the number of nodes of the number of cells of\n            the mesh, or a list of raypath coordinates.\n        filename: str\n            Name of file without extension for saving (extension vtu will be\n            added).  Raypaths are saved in separate files, and filename will\n            be appended by the dict key and have a vtp extension.\n\n        Notes\n        -----\n        VTK files can be visualized with Paraview (https://www.paraview.org)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_29to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_vtk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fields,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, 1); __PYX_ERR(0, 863, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "to_vtk") < 0)) __PYX_ERR(0, 863, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fields = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 863, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_28to_vtk(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_fields, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_28to_vtk(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename) {
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_ugrid = NULL;
  PyObject *__pyx_v_tPts = NULL;
  PyObject *__pyx_v_tet = NULL;
  int __pyx_v_save_grid;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_scalar = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_5;
  std::vector<ttcr::sxyz<double> > ::size_type __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  std::vector<ttcr::tetrahedronElem<uint32_t> > ::size_type __pyx_t_13;
  std::vector<ttcr::tetrahedronElem<uint32_t> > ::size_type __pyx_t_14;
  int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  int __pyx_t_18;
  int __pyx_t_19;
  long __pyx_t_20;
  long __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_vtk", 0);

  /* "ttcrpy/tmesh.pyx":885
 *         """
 *         cdef int n, nn
 *         ugrid = vtk.vtkUnstructuredGrid()             # <<<<<<<<<<<<<<
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkUnstructuredGrid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ugrid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":886
 *         cdef int n, nn
 *         ugrid = vtk.vtkUnstructuredGrid()
 *         tPts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tPts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":887
 *         ugrid = vtk.vtkUnstructuredGrid()
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())             # <<<<<<<<<<<<<<
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tPts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":888
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):             # <<<<<<<<<<<<<<
 *             tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)
 *         ugrid.SetPoints(tPts)
 */
  __pyx_t_5 = __pyx_v_self->no.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":889
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)             # <<<<<<<<<<<<<<
 *         ugrid.SetPoints(tPts)
 *         tet = vtk.vtkTetra()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tPts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).y); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).z); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_11 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_3, __pyx_t_4, __pyx_t_8, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_3, __pyx_t_4, __pyx_t_8, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_12 = PyTuple_New(4+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_11, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_11, __pyx_t_9);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/tmesh.pyx":890
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)
 *         ugrid.SetPoints(tPts)             # <<<<<<<<<<<<<<
 *         tet = vtk.vtkTetra()
 *         for n in range(self.tet.size()):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_v_tPts) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_tPts);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":891
 *             tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)
 *         ugrid.SetPoints(tPts)
 *         tet = vtk.vtkTetra()             # <<<<<<<<<<<<<<
 *         for n in range(self.tet.size()):
 *             for nn in range(4):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkTetra); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_12);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_12, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_v_tet = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":892
 *         ugrid.SetPoints(tPts)
 *         tet = vtk.vtkTetra()
 *         for n in range(self.tet.size()):             # <<<<<<<<<<<<<<
 *             for nn in range(4):
 *                 tet.GetPointIds().SetId(nn, self.tet[n].i[nn])
 */
  __pyx_t_13 = __pyx_v_self->tet.size();
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_14; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":893
 *         tet = vtk.vtkTetra()
 *         for n in range(self.tet.size()):
 *             for nn in range(4):             # <<<<<<<<<<<<<<
 *                 tet.GetPointIds().SetId(nn, self.tet[n].i[nn])
 *             ugrid.InsertNextCell(tet.GetCellType(), tet.GetPointIds())
 */
    for (__pyx_t_11 = 0; __pyx_t_11 < 4; __pyx_t_11+=1) {
      __pyx_v_nn = __pyx_t_11;

      /* "ttcrpy/tmesh.pyx":894
 *         for n in range(self.tet.size()):
 *             for nn in range(4):
 *                 tet.GetPointIds().SetId(nn, self.tet[n].i[nn])             # <<<<<<<<<<<<<<
 *             ugrid.InsertNextCell(tet.GetCellType(), tet.GetPointIds())
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tet, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_12 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_SetId); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_9 = __Pyx_PyInt_From_uint32_t(((__pyx_v_self->tet[__pyx_v_n]).i[__pyx_v_nn])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = NULL;
      __pyx_t_15 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_15 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_12, __pyx_t_9};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_12, __pyx_t_9};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_15, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_15, __pyx_t_9);
        __pyx_t_12 = 0;
        __pyx_t_9 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "ttcrpy/tmesh.pyx":895
 *             for nn in range(4):
 *                 tet.GetPointIds().SetId(nn, self.tet[n].i[nn])
 *             ugrid.InsertNextCell(tet.GetCellType(), tet.GetPointIds())             # <<<<<<<<<<<<<<
 * 
 *         save_grid = False
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_tet, __pyx_n_s_GetCellType); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_4 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_tet, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_9 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    __pyx_t_11 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_11 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_4, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_4, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_12) {
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_12); __pyx_t_12 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_11, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_11, __pyx_t_9);
      __pyx_t_4 = 0;
      __pyx_t_9 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/tmesh.pyx":897
 *             ugrid.InsertNextCell(tet.GetCellType(), tet.GetPointIds())
 * 
 *         save_grid = False             # <<<<<<<<<<<<<<
 *         for fn in fields:
 *             data = fields[fn]
 */
  __pyx_v_save_grid = 0;

  /* "ttcrpy/tmesh.pyx":898
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
    __pyx_t_1 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_1); __pyx_t_16 = 0;
    __pyx_t_17 = NULL;
  } else {
    __pyx_t_16 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_17 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 898, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_17)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_16); __Pyx_INCREF(__pyx_t_2); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 898, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_16); __Pyx_INCREF(__pyx_t_2); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 898, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_17(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 898, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":899
 *         save_grid = False
 *         for fn in fields:
 *             data = fields[fn]             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(data, list):
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_fields, __pyx_v_fn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 899, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":901
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
    __pyx_t_18 = PyList_Check(__pyx_v_data); 
    __pyx_t_19 = (__pyx_t_18 != 0);
    if (__pyx_t_19) {

      /* "ttcrpy/tmesh.pyx":902
 * 
 *             if isinstance(data, list):
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')             # <<<<<<<<<<<<<<
 *             else:
 *                 save_grid = True
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_save_raypaths); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PyNumber_Add(__pyx_v_filename, __pyx_n_u__25); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = PyNumber_Add(__pyx_t_9, __pyx_v_fn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Add(__pyx_t_4, __pyx_kp_u_vtp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_data, __pyx_t_9};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_data, __pyx_t_9};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 902, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_data);
        __Pyx_GIVEREF(__pyx_v_data);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_7, __pyx_v_data);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_7, __pyx_t_9);
        __pyx_t_9 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":901
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":904
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 *                 save_grid = True             # <<<<<<<<<<<<<<
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 */
    /*else*/ {
      __pyx_v_save_grid = 1;

      /* "ttcrpy/tmesh.pyx":905
 *             else:
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()             # <<<<<<<<<<<<<<
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_vtk); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_vtkDoubleArray); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF_SET(__pyx_v_scalar, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":906
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetName); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_8, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_fn);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":907
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfComponents); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_8, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_int_1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":908
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)             # <<<<<<<<<<<<<<
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfTuples); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":909
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_12 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_12, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_19 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_19 < 0)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_19) {

        /* "ttcrpy/tmesh.pyx":910
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 910, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_t_8); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_21; __pyx_t_7+=1) {
          __pyx_v_n = __pyx_t_7;

          /* "ttcrpy/tmesh.pyx":911
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])             # <<<<<<<<<<<<<<
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 911, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_data, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 911, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_4 = NULL;
          __pyx_t_11 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_11 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_9};
            __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_9};
            __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_11, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_9);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_11, __pyx_t_9);
            __pyx_t_2 = 0;
            __pyx_t_9 = 0;
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_3, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }

        /* "ttcrpy/tmesh.pyx":912
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_12 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 912, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_12, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 912, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "ttcrpy/tmesh.pyx":909
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
        goto __pyx_L12;
      }

      /* "ttcrpy/tmesh.pyx":913
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyObject_RichCompare(__pyx_t_8, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_19 < 0)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (likely(__pyx_t_19)) {

        /* "ttcrpy/tmesh.pyx":914
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetCellData().AddArray(scalar)
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 914, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_t_12); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 914, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_21; __pyx_t_7+=1) {
          __pyx_v_n = __pyx_t_7;

          /* "ttcrpy/tmesh.pyx":915
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])             # <<<<<<<<<<<<<<
 *                     ugrid.GetCellData().AddArray(scalar)
 *                 else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 915, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_data, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 915, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_2 = NULL;
          __pyx_t_11 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_11 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_8, __pyx_t_9};
            __pyx_t_12 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 915, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_8, __pyx_t_9};
            __pyx_t_12 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 915, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 915, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_11, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_11, __pyx_t_9);
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 915, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "ttcrpy/tmesh.pyx":916
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetCellData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

        /* "ttcrpy/tmesh.pyx":913
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
        goto __pyx_L12;
      }

      /* "ttcrpy/tmesh.pyx":918
 *                     ugrid.GetCellData().AddArray(scalar)
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))             # <<<<<<<<<<<<<<
 * 
 *         if save_grid:
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_fn);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 918, __pyx_L1_error)
      }
      __pyx_L12:;
    }
    __pyx_L11:;

    /* "ttcrpy/tmesh.pyx":898
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":920
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 */
  __pyx_t_19 = (__pyx_v_save_grid != 0);
  if (__pyx_t_19) {

    /* "ttcrpy/tmesh.pyx":921
 * 
 *         if save_grid:
 *             writer = vtk.vtkXMLUnstructuredGridWriter()             # <<<<<<<<<<<<<<
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_vtk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vtkXMLUnstructuredGridWriter); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_v_writer = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":922
 *         if save_grid:
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')             # <<<<<<<<<<<<<<
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = PyNumber_Add(__pyx_v_filename, __pyx_kp_u_vtu); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":923
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)             # <<<<<<<<<<<<<<
 *             writer.SetDataModeToBinary()
 *             writer.Update()
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_4, __pyx_v_ugrid) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_ugrid);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":924
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *             writer.Update()
 * 
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":925
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()
 *             writer.Update()             # <<<<<<<<<<<<<<
 * 
 *     def  _save_raypaths(self, rays, filename):
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":920
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 */
  }

  /* "ttcrpy/tmesh.pyx":863
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ugrid);
  __Pyx_XDECREF(__pyx_v_tPts);
  __Pyx_XDECREF(__pyx_v_tet);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_scalar);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":927
 *             writer.Update()
 * 
 *     def  _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_31_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_31_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rays = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_save_raypaths (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rays,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rays)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, 1); __PYX_ERR(0, 927, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_save_raypaths") < 0)) __PYX_ERR(0, 927, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rays = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 927, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_30_save_raypaths(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_v_self), __pyx_v_rays, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_30_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_polydata = NULL;
  PyObject *__pyx_v_cellarray = NULL;
  PyObject *__pyx_v_pts = NULL;
  PyObject *__pyx_v_npts = NULL;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_save_raypaths", 0);

  /* "ttcrpy/tmesh.pyx":928
 * 
 *     def  _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()             # <<<<<<<<<<<<<<
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPolyData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_polydata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":929
 *     def  _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()             # <<<<<<<<<<<<<<
 *         pts = vtk.vtkPoints()
 *         npts = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkCellArray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cellarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":930
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":931
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_npts = __pyx_int_0;

  /* "ttcrpy/tmesh.pyx":932
 *         pts = vtk.vtkPoints()
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 932, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":933
 *         npts = 0
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]             # <<<<<<<<<<<<<<
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_npts, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":934
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_npts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_npts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":935
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/tmesh.pyx":936
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 936, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":937
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 937, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 937, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 937, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 937, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":938
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])             # <<<<<<<<<<<<<<
 *                 npts += 1
 *         polydata.SetPoints(pts)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_0);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_1);
      __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_2);
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_v_npts, __pyx_t_11, __pyx_t_12, __pyx_t_13};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 4+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_v_npts, __pyx_t_11, __pyx_t_12, __pyx_t_13};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 4+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_14, __pyx_v_npts);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_14, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_14, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_14, __pyx_t_13);
        __pyx_t_11 = 0;
        __pyx_t_12 = 0;
        __pyx_t_13 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":939
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1             # <<<<<<<<<<<<<<
 *         polydata.SetPoints(pts)
 *         npts = 0
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":937
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":940
 *                 pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
 *                 npts += 1
 *         polydata.SetPoints(pts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_pts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_pts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":941
 *                 npts += 1
 *         polydata.SetPoints(pts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/tmesh.pyx":942
 *         polydata.SetPoints(pts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":943
 *         npts = 0
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()             # <<<<<<<<<<<<<<
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkPolyLine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":944
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetNumberOfIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shape); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_9, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":945
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 945, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 945, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 945, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 945, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":946
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)             # <<<<<<<<<<<<<<
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetId); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_p);
        __Pyx_GIVEREF(__pyx_v_p);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_14, __pyx_v_p);
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_14, __pyx_v_npts);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":947
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1             # <<<<<<<<<<<<<<
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":945
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":948
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 *             cellarray.InsertNextCell(line)             # <<<<<<<<<<<<<<
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cellarray, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_line);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":949
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)             # <<<<<<<<<<<<<<
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetLines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_cellarray) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":950
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()             # <<<<<<<<<<<<<<
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkXMLPolyDataWriter); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_writer = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":951
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":952
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)             # <<<<<<<<<<<<<<
 *         writer.SetDataModeToBinary()
 *         writer.Update()
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_polydata) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_polydata);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":953
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *         writer.Update()
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":954
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 *         writer.Update()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":927
 *             writer.Update()
 * 
 *     def  _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_polydata);
  __Pyx_XDECREF(__pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_v_pts);
  __Pyx_XDECREF(__pyx_v_npts);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":958
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM', int gradient_method=1,
 *                 bool tt_from_rp=1, bool interp_vel=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_33builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_32builder[] = "\n        builder(filename, n_threads, cell_slowness, method, gradient_method, tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary, n_tertiary, radius_tertiary)\n\n        Build instance of Mesh3d from VTK file\n\n        Parameters\n        ----------\n        filename : str\n            Name of file holding a vtkUnstructuredGrid.\n            The grid must have point or cell attribute named either\n            'Slowness', 'slowness', 'Velocity', 'velocity', or\n            'P-wave velocity'.  All cells must be of type vtkTetra\n\n        Other parameters are defined in Constructor\n\n        Returns\n        -------\n        mesh: :obj:`Mesh3d`\n            mesh instance\n        ";
static PyMethodDef __pyx_mdef_6ttcrpy_5tmesh_6Mesh3d_33builder = {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_33builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_32builder};
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_33builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  int __pyx_v_gradient_method;
  bool __pyx_v_tt_from_rp;
  bool __pyx_v_interp_vel;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  double __pyx_v_min_dist;
  uint32_t __pyx_v_n_secondary;
  uint32_t __pyx_v_n_tertiary;
  double __pyx_v_radius_tertiary;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("builder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_gradient_method,&__pyx_n_s_tt_from_rp,&__pyx_n_s_interp_vel,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_min_dist,&__pyx_n_s_n_secondary,&__pyx_n_s_n_tertiary,&__pyx_n_s_radius_tertiary,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[3] = ((PyObject*)__pyx_n_u_FSM);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gradient_method);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt_from_rp);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interp_vel);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_dist);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_tertiary);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius_tertiary);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "builder") < 0)) __PYX_ERR(0, 958, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    if (values[1]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 958, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[2]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 958, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[3]);
    if (values[4]) {
      __pyx_v_gradient_method = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_gradient_method == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 959, __pyx_L3_error)
    } else {
      __pyx_v_gradient_method = ((int)1);
    }
    if (values[5]) {
      __pyx_v_tt_from_rp = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_tt_from_rp == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    } else {
      __pyx_v_tt_from_rp = ((bool)1);
    }
    if (values[6]) {
      __pyx_v_interp_vel = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_interp_vel == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    } else {
      __pyx_v_interp_vel = ((bool)0);
    }
    if (values[7]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 961, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[8]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 961, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[9]) {
      __pyx_v_min_dist = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_min_dist == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 961, __pyx_L3_error)
    } else {
      __pyx_v_min_dist = ((double)1.e-5);
    }
    if (values[10]) {
      __pyx_v_n_secondary = __Pyx_PyInt_As_uint32_t(values[10]); if (unlikely((__pyx_v_n_secondary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    } else {
      __pyx_v_n_secondary = ((uint32_t)2);
    }
    if (values[11]) {
      __pyx_v_n_tertiary = __Pyx_PyInt_As_uint32_t(values[11]); if (unlikely((__pyx_v_n_tertiary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    } else {
      __pyx_v_n_tertiary = ((uint32_t)2);
    }
    if (values[12]) {
      __pyx_v_radius_tertiary = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_radius_tertiary == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L3_error)
    } else {
      __pyx_v_radius_tertiary = ((double)1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("builder", 0, 1, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 958, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 959, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh3d_32builder(__pyx_v_filename, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_gradient_method, __pyx_v_tt_from_rp, __pyx_v_interp_vel, __pyx_v_eps, __pyx_v_maxit, __pyx_v_min_dist, __pyx_v_n_secondary, __pyx_v_n_tertiary, __pyx_v_radius_tertiary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh3d_32builder(PyObject *__pyx_v_filename, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, int __pyx_v_gradient_method, bool __pyx_v_tt_from_rp, bool __pyx_v_interp_vel, double __pyx_v_eps, int __pyx_v_maxit, double __pyx_v_min_dist, uint32_t __pyx_v_n_secondary, uint32_t __pyx_v_n_tertiary, double __pyx_v_radius_tertiary) {
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_reader = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_tet = NULL;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_slowness = NULL;
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_m = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  long __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("builder", 0);

  /* "ttcrpy/tmesh.pyx":986
 * 
 *         cdef int n, nn
 *         reader = vtk.vtkXMLUnstructuredGridReader()             # <<<<<<<<<<<<<<
 *         reader.SetFileName(filename)
 *         reader.Update()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkXMLUnstructuredGridReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_reader = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":987
 *         cdef int n, nn
 *         reader = vtk.vtkXMLUnstructuredGridReader()
 *         reader.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         reader.Update()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":988
 *         reader = vtk.vtkXMLUnstructuredGridReader()
 *         reader.SetFileName(filename)
 *         reader.Update()             # <<<<<<<<<<<<<<
 * 
 *         data = reader.GetOutput()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_Update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":990
 *         reader.Update()
 * 
 *         data = reader.GetOutput()             # <<<<<<<<<<<<<<
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         for n in range(data.GetNumberOfCells()):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_GetOutput); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":991
 * 
 *         data = reader.GetOutput()
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())             # <<<<<<<<<<<<<<
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TETRA:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPoints); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_GetData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":992
 *         data = reader.GetOutput()
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         for n in range(data.GetNumberOfCells()):             # <<<<<<<<<<<<<<
 *             if data.GetCellType(n) != vtk.VTK_TETRA:
 *                 raise ValueError('{0:s} should only contain tetrahedra')
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":993
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TETRA:             # <<<<<<<<<<<<<<
 *                 raise ValueError('{0:s} should only contain tetrahedra')
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_VTK_TETRA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_10)) {

      /* "ttcrpy/tmesh.pyx":994
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TETRA:
 *                 raise ValueError('{0:s} should only contain tetrahedra')             # <<<<<<<<<<<<<<
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 *         for n in range(data.GetNumberOfCells()):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 994, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 994, __pyx_L1_error)

      /* "ttcrpy/tmesh.pyx":993
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TETRA:             # <<<<<<<<<<<<<<
 *                 raise ValueError('{0:s} should only contain tetrahedra')
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 */
    }
  }

  /* "ttcrpy/tmesh.pyx":995
 *             if data.GetCellType(n) != vtk.VTK_TETRA:
 *                 raise ValueError('{0:s} should only contain tetrahedra')
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))             # <<<<<<<<<<<<<<
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(4):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_4);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tet = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":996
 *                 raise ValueError('{0:s} should only contain tetrahedra')
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 *         for n in range(data.GetNumberOfCells()):             # <<<<<<<<<<<<<<
 *             for nn in range(4):
 *                 tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":997
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(4):             # <<<<<<<<<<<<<<
 *                 tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 * 
 */
    for (__pyx_t_11 = 0; __pyx_t_11 < 4; __pyx_t_11+=1) {
      __pyx_v_nn = __pyx_t_11;

      /* "ttcrpy/tmesh.pyx":998
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(4):
 *                 tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)             # <<<<<<<<<<<<<<
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCell); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_GetId); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
      __pyx_t_5 = 0;
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tet, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 998, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":1000
 *                 tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',             # <<<<<<<<<<<<<<
 *                  'P-wave velocity')
 *         for name in names:
 */
  __Pyx_INCREF(__pyx_tuple__27);
  __pyx_v_names = __pyx_tuple__27;

  /* "ttcrpy/tmesh.pyx":1002
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_t_2 = __pyx_v_names; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
  for (;;) {
    if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1002, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":1003
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "ttcrpy/tmesh.pyx":1004
 *         for name in names:
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 */
      __pyx_v_cell_slowness = 0;

      /* "ttcrpy/tmesh.pyx":1005
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 break
 *             if data.GetCellData().HasArray(name):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1006
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break             # <<<<<<<<<<<<<<
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 */
      goto __pyx_L11_break;

      /* "ttcrpy/tmesh.pyx":1003
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    }

    /* "ttcrpy/tmesh.pyx":1007
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "ttcrpy/tmesh.pyx":1008
 *                 break
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 break
 */
      __pyx_v_cell_slowness = 1;

      /* "ttcrpy/tmesh.pyx":1009
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 break
 *         else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_6 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1010
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 break             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')
 */
      goto __pyx_L11_break;

      /* "ttcrpy/tmesh.pyx":1007
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    }

    /* "ttcrpy/tmesh.pyx":1002
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  }
  /*else*/ {

    /* "ttcrpy/tmesh.pyx":1012
 *                 break
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')             # <<<<<<<<<<<<<<
 * 
 *         if 'lowness' in name:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1012, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1002
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_L11_break:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":1014
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data
 *         else:
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1014, __pyx_L1_error)
  }
  __pyx_t_10 = (__Pyx_PyUnicode_ContainsTF(__pyx_n_u_lowness, __pyx_v_name, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __pyx_t_14 = (__pyx_t_10 != 0);
  if (__pyx_t_14) {

    /* "ttcrpy/tmesh.pyx":1015
 * 
 *         if 'lowness' in name:
 *             slowness = data             # <<<<<<<<<<<<<<
 *         else:
 *             slowness = 1.0 / data
 */
    __Pyx_INCREF(__pyx_v_data);
    __pyx_v_slowness = __pyx_v_data;

    /* "ttcrpy/tmesh.pyx":1014
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data
 *         else:
 */
    goto __pyx_L15;
  }

  /* "ttcrpy/tmesh.pyx":1017
 *             slowness = data
 *         else:
 *             slowness = 1.0 / data             # <<<<<<<<<<<<<<
 * 
 *         m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_0, __pyx_v_data, 1.0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_slowness = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L15:;

  /* "ttcrpy/tmesh.pyx":1019
 *             slowness = 1.0 / data
 * 
 *         m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,             # <<<<<<<<<<<<<<
 *                    tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                    n_tertiary, radius_tertiary)
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_n_threads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_cell_slowness); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_gradient_method); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ttcrpy/tmesh.pyx":1020
 * 
 *         m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,
 *                    tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,             # <<<<<<<<<<<<<<
 *                    n_tertiary, radius_tertiary)
 *         m.set_slowness(slowness)
 */
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_tt_from_rp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_interp_vel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_eps); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_maxit); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_15 = PyFloat_FromDouble(__pyx_v_min_dist); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyInt_From_uint32_t(__pyx_v_n_secondary); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);

  /* "ttcrpy/tmesh.pyx":1021
 *         m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,
 *                    tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                    n_tertiary, radius_tertiary)             # <<<<<<<<<<<<<<
 *         m.set_slowness(slowness)
 *         return m
 */
  __pyx_t_17 = __Pyx_PyInt_From_uint32_t(__pyx_v_n_tertiary); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyFloat_FromDouble(__pyx_v_radius_tertiary); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);

  /* "ttcrpy/tmesh.pyx":1019
 *             slowness = 1.0 / data
 * 
 *         m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,             # <<<<<<<<<<<<<<
 *                    tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                    n_tertiary, radius_tertiary)
 */
  __pyx_t_19 = PyTuple_New(14); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_v_nodes);
  __Pyx_INCREF(__pyx_v_tet);
  __Pyx_GIVEREF(__pyx_v_tet);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_v_tet);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_19, 3, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_19, 4, __pyx_v_method);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_19, 5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_19, 6, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_19, 7, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_19, 8, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_19, 9, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_19, 10, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_19, 11, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_19, 12, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_19, 13, __pyx_t_18);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_13 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh3d), __pyx_t_19, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_v_m = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_t_18);
  __pyx_t_18 = 0;

  /* "ttcrpy/tmesh.pyx":1022
 *                    tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                    n_tertiary, radius_tertiary)
 *         m.set_slowness(slowness)             # <<<<<<<<<<<<<<
 *         return m
 * 
 */
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_m), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_17 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_19))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_19);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_19, function);
    }
  }
  __pyx_t_18 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_17, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_v_slowness);
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "ttcrpy/tmesh.pyx":1023
 *                    n_tertiary, radius_tertiary)
 *         m.set_slowness(slowness)
 *         return m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_m));
  __pyx_r = ((PyObject *)__pyx_v_m);
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":958
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM', int gradient_method=1,
 *                 bool tt_from_rp=1, bool interp_vel=0,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh3d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_reader);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_tet);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XDECREF((PyObject *)__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1076
 *     cdef Grid2D[double, uint32_t,sxz[double]]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=2] triangles,
 *                   size_t n_threads=1, bool cell_slowness=1,
 */

/* Python wrapper */
static int __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_nodes = 0;
  PyArrayObject *__pyx_v_triangles = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  bool __pyx_v_process_obtuse;
  uint32_t __pyx_v_n_secondary;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,&__pyx_n_s_triangles,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_process_obtuse,&__pyx_n_s_n_secondary,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[4] = ((PyObject*)__pyx_n_u_FSM);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_triangles)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 9, 1); __PYX_ERR(0, 1076, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_process_obtuse);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1076, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = ((PyArrayObject *)values[0]);
    __pyx_v_triangles = ((PyArrayObject *)values[1]);
    if (values[2]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[3]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[4]);
    if (values[5]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[6]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[7]) {
      __pyx_v_process_obtuse = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_process_obtuse == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    } else {
      __pyx_v_process_obtuse = ((bool)1);
    }
    if (values[8]) {
      __pyx_v_n_secondary = __Pyx_PyInt_As_uint32_t(values[8]); if (unlikely((__pyx_v_n_secondary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    } else {
      __pyx_v_n_secondary = ((uint32_t)5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1076, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nodes), __pyx_ptype_5numpy_ndarray, 1, "nodes", 0))) __PYX_ERR(0, 1076, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_triangles), __pyx_ptype_5numpy_ndarray, 1, "triangles", 0))) __PYX_ERR(0, 1077, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 1079, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d___cinit__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), __pyx_v_nodes, __pyx_v_triangles, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_eps, __pyx_v_maxit, __pyx_v_process_obtuse, __pyx_v_n_secondary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ttcrpy_5tmesh_6Mesh2d___cinit__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyArrayObject *__pyx_v_nodes, PyArrayObject *__pyx_v_triangles, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_process_obtuse, uint32_t __pyx_v_n_secondary) {
  std::vector<ttcr::sxz<double> >  __pyx_v_pts_ref;
  int __pyx_v_n;
  PyObject *__pyx_v_xmin = NULL;
  PyObject *__pyx_v_xmax = NULL;
  PyObject *__pyx_v_zmin = NULL;
  PyObject *__pyx_v_zmax = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_nodes;
  __Pyx_Buffer __pyx_pybuffer_nodes;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_triangles;
  __Pyx_Buffer __pyx_pybuffer_triangles;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  ttcr::sxz<double>  __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  ttcr::triangleElem<uint32_t>  __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  double __pyx_t_24;
  double __pyx_t_25;
  ttcr::Grid2Ducfs<double,uint32_t,ttcr::Node2Dc<double,uint32_t> ,ttcr::sxz<double> >  *__pyx_t_26;
  ttcr::Grid2Ducsp<double,uint32_t,ttcr::Node2Dcsp<double,uint32_t> ,ttcr::sxz<double> >  *__pyx_t_27;
  ttcr::Grid2Dunfs<double,uint32_t,ttcr::Node2Dn<double,uint32_t> ,ttcr::sxz<double> >  *__pyx_t_28;
  ttcr::Grid2Dunsp<double,uint32_t,ttcr::Node2Dnsp<double,uint32_t> ,ttcr::sxz<double> >  *__pyx_t_29;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_nodes);
  __Pyx_INCREF((PyObject *)__pyx_v_triangles);
  __pyx_pybuffer_nodes.pybuffer.buf = NULL;
  __pyx_pybuffer_nodes.refcount = 0;
  __pyx_pybuffernd_nodes.data = NULL;
  __pyx_pybuffernd_nodes.rcbuffer = &__pyx_pybuffer_nodes;
  __pyx_pybuffer_triangles.pybuffer.buf = NULL;
  __pyx_pybuffer_triangles.refcount = 0;
  __pyx_pybuffernd_triangles.data = NULL;
  __pyx_pybuffernd_triangles.rcbuffer = &__pyx_pybuffer_triangles;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_v_nodes, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  }
  __pyx_pybuffernd_nodes.diminfo[0].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nodes.diminfo[0].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_nodes.diminfo[1].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_nodes.diminfo[1].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer, (PyObject*)__pyx_v_triangles, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  }
  __pyx_pybuffernd_triangles.diminfo[0].strides = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_triangles.diminfo[0].shape = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_triangles.diminfo[1].strides = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_triangles.diminfo[1].shape = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.shape[1];

  /* "ttcrpy/tmesh.pyx":1082
 *                   bool process_obtuse=1, uint32_t n_secondary=5):
 * 
 *         self.cell_slowness = cell_slowness             # <<<<<<<<<<<<<<
 *         self._n_threads = n_threads
 *         self.eps = eps
 */
  __pyx_v_self->cell_slowness = __pyx_v_cell_slowness;

  /* "ttcrpy/tmesh.pyx":1083
 * 
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads             # <<<<<<<<<<<<<<
 *         self.eps = eps
 *         self.maxit = maxit
 */
  __pyx_v_self->_n_threads = __pyx_v_n_threads;

  /* "ttcrpy/tmesh.pyx":1084
 *         self.cell_slowness = cell_slowness
 *         self._n_threads = n_threads
 *         self.eps = eps             # <<<<<<<<<<<<<<
 *         self.maxit = maxit
 *         self.process_obtuse = process_obtuse
 */
  __pyx_v_self->eps = __pyx_v_eps;

  /* "ttcrpy/tmesh.pyx":1085
 *         self._n_threads = n_threads
 *         self.eps = eps
 *         self.maxit = maxit             # <<<<<<<<<<<<<<
 *         self.process_obtuse = process_obtuse
 *         self.n_secondary = n_secondary
 */
  __pyx_v_self->maxit = __pyx_v_maxit;

  /* "ttcrpy/tmesh.pyx":1086
 *         self.eps = eps
 *         self.maxit = maxit
 *         self.process_obtuse = process_obtuse             # <<<<<<<<<<<<<<
 *         self.n_secondary = n_secondary
 * 
 */
  __pyx_v_self->process_obtuse = __pyx_v_process_obtuse;

  /* "ttcrpy/tmesh.pyx":1087
 *         self.maxit = maxit
 *         self.process_obtuse = process_obtuse
 *         self.n_secondary = n_secondary             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[sxz[double]] pts_ref
 */
  __pyx_v_self->n_secondary = __pyx_v_n_secondary;

  /* "ttcrpy/tmesh.pyx":1092
 *         cdef int n
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             nodes = np.ascontiguousarray(nodes)
 *         if not triangles.flags['C_CONTIGUOUS']:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nodes), __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1093
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)             # <<<<<<<<<<<<<<
 *         if not triangles.flags['C_CONTIGUOUS']:
 *             triangles = np.ascontiguousarray(triangles)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, ((PyObject *)__pyx_v_nodes)) : __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_nodes));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1093, __pyx_L1_error)
    __pyx_t_6 = ((PyArrayObject *)__pyx_t_2);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer, (PyObject*)__pyx_v_nodes, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        }
        __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;
      }
      __pyx_pybuffernd_nodes.diminfo[0].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nodes.diminfo[0].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_nodes.diminfo[1].strides = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_nodes.diminfo[1].shape = __pyx_pybuffernd_nodes.rcbuffer->pybuffer.shape[1];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1093, __pyx_L1_error)
    }
    __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_nodes, ((PyArrayObject *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1092
 *         cdef int n
 * 
 *         if not nodes.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             nodes = np.ascontiguousarray(nodes)
 *         if not triangles.flags['C_CONTIGUOUS']:
 */
  }

  /* "ttcrpy/tmesh.pyx":1094
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)
 *         if not triangles.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             triangles = np.ascontiguousarray(triangles)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_triangles), __pyx_n_s_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  if (__pyx_t_3) {

    /* "ttcrpy/tmesh.pyx":1095
 *             nodes = np.ascontiguousarray(nodes)
 *         if not triangles.flags['C_CONTIGUOUS']:
 *             triangles = np.ascontiguousarray(triangles)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nodes.shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, ((PyObject *)__pyx_v_triangles)) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_v_triangles));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1095, __pyx_L1_error)
    __pyx_t_11 = ((PyArrayObject *)__pyx_t_5);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer, (PyObject*)__pyx_v_triangles, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        }
        __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;
      }
      __pyx_pybuffernd_triangles.diminfo[0].strides = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_triangles.diminfo[0].shape = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_triangles.diminfo[1].strides = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_triangles.diminfo[1].shape = __pyx_pybuffernd_triangles.rcbuffer->pybuffer.shape[1];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1095, __pyx_L1_error)
    }
    __pyx_t_11 = 0;
    __Pyx_DECREF_SET(__pyx_v_triangles, ((PyArrayObject *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1094
 *         if not nodes.flags['C_CONTIGUOUS']:
 *             nodes = np.ascontiguousarray(nodes)
 *         if not triangles.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             triangles = np.ascontiguousarray(triangles)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1097
 *             triangles = np.ascontiguousarray(triangles)
 * 
 *         for n in range(nodes.shape[0]):             # <<<<<<<<<<<<<<
 *             self.no.push_back(sxz[double](nodes[n, 0],
 *                                           nodes[n, 1]))
 */
  __pyx_t_12 = (__pyx_v_nodes->dimensions[0]);
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":1098
 * 
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxz[double](nodes[n, 0],             # <<<<<<<<<<<<<<
 *                                           nodes[n, 1]))
 *         for n in range(triangles.shape[0]):
 */
    __pyx_t_14 = __pyx_v_n;
    __pyx_t_15 = 0;
    __pyx_t_16 = -1;
    if (__pyx_t_14 < 0) {
      __pyx_t_14 += __pyx_pybuffernd_nodes.diminfo[0].shape;
      if (unlikely(__pyx_t_14 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_14 >= __pyx_pybuffernd_nodes.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_15 < 0) {
      __pyx_t_15 += __pyx_pybuffernd_nodes.diminfo[1].shape;
      if (unlikely(__pyx_t_15 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_15 >= __pyx_pybuffernd_nodes.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 1098, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1099
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxz[double](nodes[n, 0],
 *                                           nodes[n, 1]))             # <<<<<<<<<<<<<<
 *         for n in range(triangles.shape[0]):
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],
 */
    __pyx_t_17 = __pyx_v_n;
    __pyx_t_18 = 1;
    __pyx_t_16 = -1;
    if (__pyx_t_17 < 0) {
      __pyx_t_17 += __pyx_pybuffernd_nodes.diminfo[0].shape;
      if (unlikely(__pyx_t_17 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_nodes.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_18 < 0) {
      __pyx_t_18 += __pyx_pybuffernd_nodes.diminfo[1].shape;
      if (unlikely(__pyx_t_18 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_nodes.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 1099, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1098
 * 
 *         for n in range(nodes.shape[0]):
 *             self.no.push_back(sxz[double](nodes[n, 0],             # <<<<<<<<<<<<<<
 *                                           nodes[n, 1]))
 *         for n in range(triangles.shape[0]):
 */
    try {
      __pyx_t_19 = ttcr::sxz<double> ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_nodes.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_nodes.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_nodes.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_double_t *, __pyx_pybuffernd_nodes.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_nodes.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_nodes.diminfo[1].strides)));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1098, __pyx_L1_error)
    }
    try {
      __pyx_v_self->no.push_back(__pyx_t_19);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1098, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":1100
 *             self.no.push_back(sxz[double](nodes[n, 0],
 *                                           nodes[n, 1]))
 *         for n in range(triangles.shape[0]):             # <<<<<<<<<<<<<<
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],
 *                                                       triangles[n, 1],
 */
  __pyx_t_12 = (__pyx_v_triangles->dimensions[0]);
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":1101
 *                                           nodes[n, 1]))
 *         for n in range(triangles.shape[0]):
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],             # <<<<<<<<<<<<<<
 *                                                       triangles[n, 1],
 *                                                       triangles[n, 2]))
 */
    __pyx_t_18 = __pyx_v_n;
    __pyx_t_17 = 0;
    __pyx_t_16 = -1;
    if (__pyx_t_18 < 0) {
      __pyx_t_18 += __pyx_pybuffernd_triangles.diminfo[0].shape;
      if (unlikely(__pyx_t_18 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_triangles.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_17 < 0) {
      __pyx_t_17 += __pyx_pybuffernd_triangles.diminfo[1].shape;
      if (unlikely(__pyx_t_17 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_triangles.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 1101, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1102
 *         for n in range(triangles.shape[0]):
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],
 *                                                       triangles[n, 1],             # <<<<<<<<<<<<<<
 *                                                       triangles[n, 2]))
 * 
 */
    __pyx_t_15 = __pyx_v_n;
    __pyx_t_14 = 1;
    __pyx_t_16 = -1;
    if (__pyx_t_15 < 0) {
      __pyx_t_15 += __pyx_pybuffernd_triangles.diminfo[0].shape;
      if (unlikely(__pyx_t_15 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_15 >= __pyx_pybuffernd_triangles.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_14 < 0) {
      __pyx_t_14 += __pyx_pybuffernd_triangles.diminfo[1].shape;
      if (unlikely(__pyx_t_14 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_14 >= __pyx_pybuffernd_triangles.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 1102, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1103
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],
 *                                                       triangles[n, 1],
 *                                                       triangles[n, 2]))             # <<<<<<<<<<<<<<
 * 
 *         if method == 'FSM':
 */
    __pyx_t_20 = __pyx_v_n;
    __pyx_t_21 = 2;
    __pyx_t_16 = -1;
    if (__pyx_t_20 < 0) {
      __pyx_t_20 += __pyx_pybuffernd_triangles.diminfo[0].shape;
      if (unlikely(__pyx_t_20 < 0)) __pyx_t_16 = 0;
    } else if (unlikely(__pyx_t_20 >= __pyx_pybuffernd_triangles.diminfo[0].shape)) __pyx_t_16 = 0;
    if (__pyx_t_21 < 0) {
      __pyx_t_21 += __pyx_pybuffernd_triangles.diminfo[1].shape;
      if (unlikely(__pyx_t_21 < 0)) __pyx_t_16 = 1;
    } else if (unlikely(__pyx_t_21 >= __pyx_pybuffernd_triangles.diminfo[1].shape)) __pyx_t_16 = 1;
    if (unlikely(__pyx_t_16 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_16);
      __PYX_ERR(0, 1103, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1101
 *                                           nodes[n, 1]))
 *         for n in range(triangles.shape[0]):
 *             self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],             # <<<<<<<<<<<<<<
 *                                                       triangles[n, 1],
 *                                                       triangles[n, 2]))
 */
    try {
      __pyx_t_22 = ttcr::triangleElem<uint32_t> ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_triangles.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_triangles.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_triangles.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_triangles.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_triangles.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_triangles.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_triangles.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_triangles.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_triangles.diminfo[1].strides)));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1101, __pyx_L1_error)
    }
    try {
      __pyx_v_self->tri.push_back(__pyx_t_22);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1101, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":1105
 *                                                       triangles[n, 2]))
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 */
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1106
 * 
 *         if method == 'FSM':
 *             xmin = np.min(nodes[:, 0])             # <<<<<<<<<<<<<<
 *             xmax = np.max(nodes[:, 0])
 *             zmin = np.min(nodes[:, 1])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_23 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_23)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_23, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_xmin = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1107
 *         if method == 'FSM':
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])             # <<<<<<<<<<<<<<
 *             zmin = np.min(nodes[:, 1])
 *             zmax = np.max(nodes[:, 1])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_23 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_23)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_23, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_xmax = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1108
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 *             zmin = np.min(nodes[:, 1])             # <<<<<<<<<<<<<<
 *             zmax = np.max(nodes[:, 1])
 *             pts_ref.push_back(sxz[double](xmin, zmin))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_23 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_23)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_23, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_zmin = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1109
 *             xmax = np.max(nodes[:, 0])
 *             zmin = np.min(nodes[:, 1])
 *             zmax = np.max(nodes[:, 1])             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxz[double](xmin, zmin))
 *             pts_ref.push_back(sxz[double](xmin, zmax))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_nodes), __pyx_tuple__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_23 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_23)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_23, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_zmax = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1110
 *             zmin = np.min(nodes[:, 1])
 *             zmax = np.max(nodes[:, 1])
 *             pts_ref.push_back(sxz[double](xmin, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxz[double](xmin, zmax))
 *             pts_ref.push_back(sxz[double](xmax, zmin))
 */
    __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L1_error)
    __pyx_t_25 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_25 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L1_error)
    try {
      __pyx_t_19 = ttcr::sxz<double> (__pyx_t_24, __pyx_t_25);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1110, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_19);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1110, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1111
 *             zmax = np.max(nodes[:, 1])
 *             pts_ref.push_back(sxz[double](xmin, zmin))
 *             pts_ref.push_back(sxz[double](xmin, zmax))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxz[double](xmax, zmin))
 *             pts_ref.push_back(sxz[double](xmax, zmax))
 */
    __pyx_t_25 = __pyx_PyFloat_AsDouble(__pyx_v_xmin); if (unlikely((__pyx_t_25 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1111, __pyx_L1_error)
    __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1111, __pyx_L1_error)
    try {
      __pyx_t_19 = ttcr::sxz<double> (__pyx_t_25, __pyx_t_24);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1111, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_19);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1111, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1112
 *             pts_ref.push_back(sxz[double](xmin, zmin))
 *             pts_ref.push_back(sxz[double](xmin, zmax))
 *             pts_ref.push_back(sxz[double](xmax, zmin))             # <<<<<<<<<<<<<<
 *             pts_ref.push_back(sxz[double](xmax, zmax))
 * 
 */
    __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1112, __pyx_L1_error)
    __pyx_t_25 = __pyx_PyFloat_AsDouble(__pyx_v_zmin); if (unlikely((__pyx_t_25 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1112, __pyx_L1_error)
    try {
      __pyx_t_19 = ttcr::sxz<double> (__pyx_t_24, __pyx_t_25);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1112, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_19);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1112, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1113
 *             pts_ref.push_back(sxz[double](xmin, zmax))
 *             pts_ref.push_back(sxz[double](xmax, zmin))
 *             pts_ref.push_back(sxz[double](xmax, zmax))             # <<<<<<<<<<<<<<
 * 
 *         if cell_slowness:
 */
    __pyx_t_25 = __pyx_PyFloat_AsDouble(__pyx_v_xmax); if (unlikely((__pyx_t_25 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1113, __pyx_L1_error)
    __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_v_zmax); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1113, __pyx_L1_error)
    try {
      __pyx_t_19 = ttcr::sxz<double> (__pyx_t_25, __pyx_t_24);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1113, __pyx_L1_error)
    }
    try {
      __pyx_v_pts_ref.push_back(__pyx_t_19);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1113, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1105
 *                                                       triangles[n, 2]))
 * 
 *         if method == 'FSM':             # <<<<<<<<<<<<<<
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 */
  }

  /* "ttcrpy/tmesh.pyx":1115
 *             pts_ref.push_back(sxz[double](xmax, zmax))
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
  __pyx_t_4 = (__pyx_v_cell_slowness != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1116
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Ducfs[double,uint32_t,Node2Dc[double,uint32_t],sxz[double]](self.no,
 */
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1116, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "ttcrpy/tmesh.pyx":1117
 *         if cell_slowness:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Ducfs[double,uint32_t,Node2Dc[double,uint32_t],sxz[double]](self.no,
 *                                                                                                  self.tri,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/tmesh.pyx":1118
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid2Ducfs[double,uint32_t,Node2Dc[double,uint32_t],sxz[double]](self.no,             # <<<<<<<<<<<<<<
 *                                                                                                  self.tri,
 *                                                                                                  eps,
 */
      try {
        __pyx_t_26 = new ttcr::Grid2Ducfs<double,uint32_t,ttcr::Node2Dc<double,uint32_t> ,ttcr::sxz<double> > (__pyx_v_self->no, __pyx_v_self->tri, __pyx_v_eps, __pyx_v_maxit, __pyx_v_pts_ref, 2, __pyx_v_n_threads, __pyx_v_process_obtuse);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1118, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_26;

      /* "ttcrpy/tmesh.pyx":1116
 * 
 *         if cell_slowness:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Ducfs[double,uint32_t,Node2Dc[double,uint32_t],sxz[double]](self.no,
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":1125
 *                                                                                                  n_threads,
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Ducsp[double,uint32_t,Node2Dcsp[double,uint32_t],sxz[double]](self.no,
 */
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1125, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (likely(__pyx_t_4)) {

      /* "ttcrpy/tmesh.pyx":1126
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Ducsp[double,uint32_t,Node2Dcsp[double,uint32_t],sxz[double]](self.no,
 *                                                                                                    self.tri,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/tmesh.pyx":1127
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid2Ducsp[double,uint32_t,Node2Dcsp[double,uint32_t],sxz[double]](self.no,             # <<<<<<<<<<<<<<
 *                                                                                                    self.tri,
 *                                                                                                    n_secondary,
 */
      try {
        __pyx_t_27 = new ttcr::Grid2Ducsp<double,uint32_t,ttcr::Node2Dcsp<double,uint32_t> ,ttcr::sxz<double> > (__pyx_v_self->no, __pyx_v_self->tri, __pyx_v_n_secondary, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1127, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_27;

      /* "ttcrpy/tmesh.pyx":1125
 *                                                                                                  n_threads,
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Ducsp[double,uint32_t,Node2Dcsp[double,uint32_t],sxz[double]](self.no,
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":1132
 *                                                                                                    n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 *         else:
 *             if method == 'FSM':
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 1132, __pyx_L1_error)
    }
    __pyx_L11:;

    /* "ttcrpy/tmesh.pyx":1115
 *             pts_ref.push_back(sxz[double](xmax, zmax))
 * 
 *         if cell_slowness:             # <<<<<<<<<<<<<<
 *             if method == 'FSM':
 *                 self.method = b'f'
 */
    goto __pyx_L10;
  }

  /* "ttcrpy/tmesh.pyx":1134
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Dunfs[double,uint32_t,Node2Dn[double,uint32_t],sxz[double]](self.no,
 */
  /*else*/ {
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_FSM, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "ttcrpy/tmesh.pyx":1135
 *         else:
 *             if method == 'FSM':
 *                 self.method = b'f'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Dunfs[double,uint32_t,Node2Dn[double,uint32_t],sxz[double]](self.no,
 *                                                                                                  self.tri,
 */
      __pyx_v_self->method = 'f';

      /* "ttcrpy/tmesh.pyx":1136
 *             if method == 'FSM':
 *                 self.method = b'f'
 *                 self.grid = new Grid2Dunfs[double,uint32_t,Node2Dn[double,uint32_t],sxz[double]](self.no,             # <<<<<<<<<<<<<<
 *                                                                                                  self.tri,
 *                                                                                                  eps,
 */
      try {
        __pyx_t_28 = new ttcr::Grid2Dunfs<double,uint32_t,ttcr::Node2Dn<double,uint32_t> ,ttcr::sxz<double> > (__pyx_v_self->no, __pyx_v_self->tri, __pyx_v_eps, __pyx_v_maxit, __pyx_v_pts_ref, 2, __pyx_v_n_threads, __pyx_v_process_obtuse);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1136, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_28;

      /* "ttcrpy/tmesh.pyx":1134
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 *         else:
 *             if method == 'FSM':             # <<<<<<<<<<<<<<
 *                 self.method = b'f'
 *                 self.grid = new Grid2Dunfs[double,uint32_t,Node2Dn[double,uint32_t],sxz[double]](self.no,
 */
      goto __pyx_L12;
    }

    /* "ttcrpy/tmesh.pyx":1143
 *                                                                                                  n_threads,
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Dunsp[double,uint32_t,Node2Dnsp[double,uint32_t],sxz[double]](self.no,
 */
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_SPM, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1143, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (likely(__pyx_t_4)) {

      /* "ttcrpy/tmesh.pyx":1144
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':
 *                 self.method = b's'             # <<<<<<<<<<<<<<
 *                 self.grid = new Grid2Dunsp[double,uint32_t,Node2Dnsp[double,uint32_t],sxz[double]](self.no,
 *                                                                                                    self.tri,
 */
      __pyx_v_self->method = 's';

      /* "ttcrpy/tmesh.pyx":1145
 *             elif method == 'SPM':
 *                 self.method = b's'
 *                 self.grid = new Grid2Dunsp[double,uint32_t,Node2Dnsp[double,uint32_t],sxz[double]](self.no,             # <<<<<<<<<<<<<<
 *                                                                                                    self.tri,
 *                                                                                                    n_secondary,
 */
      try {
        __pyx_t_29 = new ttcr::Grid2Dunsp<double,uint32_t,ttcr::Node2Dnsp<double,uint32_t> ,ttcr::sxz<double> > (__pyx_v_self->no, __pyx_v_self->tri, __pyx_v_n_secondary, __pyx_v_n_threads);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1145, __pyx_L1_error)
      }
      __pyx_v_self->grid = __pyx_t_29;

      /* "ttcrpy/tmesh.pyx":1143
 *                                                                                                  n_threads,
 *                                                                                                  process_obtuse)
 *             elif method == 'SPM':             # <<<<<<<<<<<<<<
 *                 self.method = b's'
 *                 self.grid = new Grid2Dunsp[double,uint32_t,Node2Dnsp[double,uint32_t],sxz[double]](self.no,
 */
      goto __pyx_L12;
    }

    /* "ttcrpy/tmesh.pyx":1150
 *                                                                                                    n_threads)
 *             else:
 *                 raise ValueError('Method {0:s} undefined'.format(method))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Method_0_s_undefined, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_method) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_method);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1150, __pyx_L1_error)
    }
    __pyx_L12:;
  }
  __pyx_L10:;

  /* "ttcrpy/tmesh.pyx":1076
 *     cdef Grid2D[double, uint32_t,sxz[double]]* grid
 * 
 *     def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=2] triangles,
 *                   size_t n_threads=1, bool cell_slowness=1,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_23);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nodes.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_triangles.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_xmin);
  __Pyx_XDECREF(__pyx_v_xmax);
  __Pyx_XDECREF(__pyx_v_zmin);
  __Pyx_XDECREF(__pyx_v_zmax);
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_triangles);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1152
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

/* Python wrapper */
static void __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_2__dealloc__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_2__dealloc__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ttcrpy/tmesh.pyx":1153
 * 
 *     def __dealloc__(self):
 *         del self.grid             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  delete __pyx_v_self->grid;

  /* "ttcrpy/tmesh.pyx":1152
 *                 raise ValueError('Method {0:s} undefined'.format(method))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.grid
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ttcrpy/tmesh.pyx":1155
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_4__reduce__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_4__reduce__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_triangles = NULL;
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_constructor_params = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ttcrpy/tmesh.pyx":1156
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
  switch (__pyx_v_self->method) {
    case 'f':

    /* "ttcrpy/tmesh.pyx":1157
 *     def __reduce__(self):
 *         if self.method == b'f':
 *             method = 'FSM'             # <<<<<<<<<<<<<<
 *         elif self.method == b's':
 *             method = 'SPM'
 */
    __Pyx_INCREF(__pyx_n_u_FSM);
    __pyx_v_method = __pyx_n_u_FSM;

    /* "ttcrpy/tmesh.pyx":1156
 * 
 *     def __reduce__(self):
 *         if self.method == b'f':             # <<<<<<<<<<<<<<
 *             method = 'FSM'
 *         elif self.method == b's':
 */
    break;
    case 's':

    /* "ttcrpy/tmesh.pyx":1159
 *             method = 'FSM'
 *         elif self.method == b's':
 *             method = 'SPM'             # <<<<<<<<<<<<<<
 * 
 *         nodes = np.ndarray((self.no.size(), 2))
 */
    __Pyx_INCREF(__pyx_n_u_SPM);
    __pyx_v_method = __pyx_n_u_SPM;

    /* "ttcrpy/tmesh.pyx":1158
 *         if self.method == b'f':
 *             method = 'FSM'
 *         elif self.method == b's':             # <<<<<<<<<<<<<<
 *             method = 'SPM'
 * 
 */
    break;
    default: break;
  }

  /* "ttcrpy/tmesh.pyx":1161
 *             method = 'SPM'
 * 
 *         nodes = np.ndarray((self.no.size(), 2))             # <<<<<<<<<<<<<<
 *         triangles = np.ndarray((self.tri.size(), 3), dtype=int)
 *         cdef int n
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1162
 * 
 *         nodes = np.ndarray((self.no.size(), 2))
 *         triangles = np.ndarray((self.tri.size(), 3), dtype=int)             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef int nn
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tri.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1162, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_triangles = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1165
 *         cdef int n
 *         cdef int nn
 *         for n in range(nodes.shape[0]):             # <<<<<<<<<<<<<<
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].z
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_nodes, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":1166
 *         cdef int nn
 *         for n in range(nodes.shape[0]):
 *             nodes[n, 0] = self.no[n].x             # <<<<<<<<<<<<<<
 *             nodes[n, 1] = self.no[n].z
 *         for n in range(triangles.shape[0]):
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_0);
    __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_nodes, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 1166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1167
 *         for n in range(nodes.shape[0]):
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].z             # <<<<<<<<<<<<<<
 *         for n in range(triangles.shape[0]):
 *             for nn in range(3):
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).z); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
    __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_nodes, __pyx_t_3, __pyx_t_2) < 0)) __PYX_ERR(0, 1167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1168
 *             nodes[n, 0] = self.no[n].x
 *             nodes[n, 1] = self.no[n].z
 *         for n in range(triangles.shape[0]):             # <<<<<<<<<<<<<<
 *             for nn in range(3):
 *                 triangles[n, nn] = self.tri[n].i[nn]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_triangles, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":1169
 *             nodes[n, 1] = self.no[n].z
 *         for n in range(triangles.shape[0]):
 *             for nn in range(3):             # <<<<<<<<<<<<<<
 *                 triangles[n, nn] = self.tri[n].i[nn]
 * 
 */
    for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
      __pyx_v_nn = __pyx_t_7;

      /* "ttcrpy/tmesh.pyx":1170
 *         for n in range(triangles.shape[0]):
 *             for nn in range(3):
 *                 triangles[n, nn] = self.tri[n].i[nn]             # <<<<<<<<<<<<<<
 * 
 *         constructor_params = (nodes, triangles,
 */
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t(((__pyx_v_self->tri[__pyx_v_n]).i[__pyx_v_nn])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_triangles, __pyx_t_8, __pyx_t_3) < 0)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":1173
 * 
 *         constructor_params = (nodes, triangles,
 *                               method, self.cell_slowness,             # <<<<<<<<<<<<<<
 *                               self._n_threads,
 *                               self.eps, self.maxit, self.process_obtuse,
 */
  if (unlikely(!__pyx_v_method)) { __Pyx_RaiseUnboundLocalError("method"); __PYX_ERR(0, 1173, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->cell_slowness); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ttcrpy/tmesh.pyx":1174
 *         constructor_params = (nodes, triangles,
 *                               method, self.cell_slowness,
 *                               self._n_threads,             # <<<<<<<<<<<<<<
 *                               self.eps, self.maxit, self.process_obtuse,
 *                               self.n_secondary)
 */
  __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "ttcrpy/tmesh.pyx":1175
 *                               method, self.cell_slowness,
 *                               self._n_threads,
 *                               self.eps, self.maxit, self.process_obtuse,             # <<<<<<<<<<<<<<
 *                               self.n_secondary)
 *         return (_rebuild2d, constructor_params)
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->eps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->maxit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyBool_FromLong(__pyx_v_self->process_obtuse); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "ttcrpy/tmesh.pyx":1176
 *                               self._n_threads,
 *                               self.eps, self.maxit, self.process_obtuse,
 *                               self.n_secondary)             # <<<<<<<<<<<<<<
 *         return (_rebuild2d, constructor_params)
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->n_secondary); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "ttcrpy/tmesh.pyx":1172
 *                 triangles[n, nn] = self.tri[n].i[nn]
 * 
 *         constructor_params = (nodes, triangles,             # <<<<<<<<<<<<<<
 *                               method, self.cell_slowness,
 *                               self._n_threads,
 */
  __pyx_t_11 = PyTuple_New(9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_nodes);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_triangles);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_method);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_11, 4, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 5, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 6, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_11, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_11, 8, __pyx_t_10);
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_v_constructor_params = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;

  /* "ttcrpy/tmesh.pyx":1177
 *                               self.eps, self.maxit, self.process_obtuse,
 *                               self.n_secondary)
 *         return (_rebuild2d, constructor_params)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_rebuild2d); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11);
  __Pyx_INCREF(__pyx_v_constructor_params);
  __Pyx_GIVEREF(__pyx_v_constructor_params);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_constructor_params);
  __pyx_t_11 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1155
 *         del self.grid
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         if self.method == b'f':
 *             method = 'FSM'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_triangles);
  __Pyx_XDECREF(__pyx_v_constructor_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1180
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9n_threads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9n_threads_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_9n_threads___get__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_9n_threads___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/tmesh.pyx":1182
 *     def n_threads(self):
 *         """int: number of threads for raytracing"""
 *         return self._n_threads             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1180
 * 
 *     @property
 *     def n_threads(self):             # <<<<<<<<<<<<<<
 *         """int: number of threads for raytracing"""
 *         return self._n_threads
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.n_threads.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1185
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7nparams_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7nparams_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_7nparams___get__(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_7nparams___get__(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ttcrpy/tmesh.pyx":1187
 *     def nparams(self):
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return self.tri.size()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->cell_slowness != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":1188
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 *             return self.tri.size()             # <<<<<<<<<<<<<<
 *         else:
 *             return self.no.size()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tri.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":1187
 *     def nparams(self):
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:             # <<<<<<<<<<<<<<
 *             return self.tri.size()
 *         else:
 */
  }

  /* "ttcrpy/tmesh.pyx":1190
 *             return self.tri.size()
 *         else:
 *             return self.no.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_nodes(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/tmesh.pyx":1185
 * 
 *     @property
 *     def nparams(self):             # <<<<<<<<<<<<<<
 *         """int: total number of parameters for mesh"""
 *         if self.cell_slowness:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.nparams.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1192
 *             return self.no.size()
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_6get_number_of_nodes[] = "\n        Returns\n        -------\n        int:\n            number of nodes in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7get_number_of_nodes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_nodes (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_6get_number_of_nodes(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_6get_number_of_nodes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_nodes", 0);

  /* "ttcrpy/tmesh.pyx":1199
 *             number of nodes in grid
 *         """
 *         return self.no.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_number_of_cells(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1192
 *             return self.no.size()
 * 
 *     def get_number_of_nodes(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.get_number_of_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1201
 *         return self.no.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_8get_number_of_cells[] = "\n        Returns\n        -------\n        int:\n            number of cells in grid\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9get_number_of_cells(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_number_of_cells (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_8get_number_of_cells(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_8get_number_of_cells(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_number_of_cells", 0);

  /* "ttcrpy/tmesh.pyx":1208
 *             number of cells in grid
 *         """
 *         return self.tri.size()             # <<<<<<<<<<<<<<
 * 
 *     def get_grid_traveltimes(self, thread_no=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->tri.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1201
 *         return self.no.size()
 * 
 *     def get_number_of_cells(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.get_number_of_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1210
 *         return self.tri.size()
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_11get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_10get_grid_traveltimes[] = "\n        get_grid_traveltimes(thread_no=0)\n\n        Obtain traveltimes computed at primary grid nodes\n\n        Parameters\n        ----------\n        thread_no : int\n            thread used to computed traveltimes (default is 0)\n\n        Returns\n        -------\n        tt: np ndarray, shape (nnodes,)\n            traveltimes\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_11get_grid_traveltimes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_thread_no = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_grid_traveltimes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_thread_no,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_grid_traveltimes") < 0)) __PYX_ERR(0, 1210, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_thread_no = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_grid_traveltimes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1210, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_10get_grid_traveltimes(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), __pyx_v_thread_no);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_10get_grid_traveltimes(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_thread_no) {
  std::vector<double>  __pyx_v_tmp;
  int __pyx_v_n;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<double> ::size_type __pyx_t_7;
  std::vector<double> ::size_type __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_grid_traveltimes", 0);

  /* "ttcrpy/tmesh.pyx":1226
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_thread_no, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "ttcrpy/tmesh.pyx":1227
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1227, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":1226
 *             traveltimes
 *         """
 *         if thread_no >= self._n_threads:             # <<<<<<<<<<<<<<
 *             raise ValueError('Thread number is larger than number of threads')
 *         cdef vector[double] tmp
 */
  }

  /* "ttcrpy/tmesh.pyx":1230
 *         cdef vector[double] tmp
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)             # <<<<<<<<<<<<<<
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 */
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L1_error)
  try {
    __pyx_v_self->grid->getTT(__pyx_v_tmp, __pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1230, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1231
 *         cdef int n
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))             # <<<<<<<<<<<<<<
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_tmp.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_tt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":1232
 *         self.grid.getTT(tmp, thread_no)
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):             # <<<<<<<<<<<<<<
 *             tt[n] = tmp[n]
 *         return tt
 */
  __pyx_t_7 = __pyx_v_tmp.size();
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":1233
 *         tt = np.empty((tmp.size(),))
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]             # <<<<<<<<<<<<<<
 *         return tt
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_v_tmp[__pyx_v_n])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_n, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1234
 *         for n in range(tmp.size()):
 *             tt[n] = tmp[n]
 *         return tt             # <<<<<<<<<<<<<<
 * 
 *     def set_slowness(self, slowness):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tt);
  __pyx_r = __pyx_v_tt;
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1210
 *         return self.tri.size()
 * 
 *     def get_grid_traveltimes(self, thread_no=0):             # <<<<<<<<<<<<<<
 *         """
 *         get_grid_traveltimes(thread_no=0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.get_grid_traveltimes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1236
 *         return tt
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_13set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_12set_slowness[] = "\n        set_slowness(slowness)\n\n        Assign slowness to grid\n\n        Parameters\n        ----------\n        slowness : np ndarray, shape (nparams, )\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_13set_slowness(PyObject *__pyx_v_self, PyObject *__pyx_v_slowness) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_slowness (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_12set_slowness(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), ((PyObject *)__pyx_v_slowness));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_12set_slowness(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_slowness) {
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_slowness", 0);
  __Pyx_INCREF(__pyx_v_slowness);

  /* "ttcrpy/tmesh.pyx":1246
 *         slowness : np ndarray, shape (nparams, )
 *         """
 *         if slowness.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":1247
 *         """
 *         if slowness.size != self.nparams:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1247, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":1246
 *         slowness : np ndarray, shape (nparams, )
 *         """
 *         if slowness.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('Slowness vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1249
 *             raise ValueError('Slowness vector has wrong size')
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             slowness = np.ascontiguousarray(slowness)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "ttcrpy/tmesh.pyx":1250
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:
 *             slowness = np.ascontiguousarray(slowness)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_slowness, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1249
 *             raise ValueError('Slowness vector has wrong size')
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             slowness = np.ascontiguousarray(slowness)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1254
 *         cdef vector[double] slown
 *         cdef int i
 *         for i in range(slowness.size):             # <<<<<<<<<<<<<<
 *             slown.push_back(slowness[i])
 *         self.grid.setSlowness(slown)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_slowness, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ttcrpy/tmesh.pyx":1255
 *         cdef int i
 *         for i in range(slowness.size):
 *             slown.push_back(slowness[i])             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_slowness, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      __pyx_v_slown.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1255, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":1256
 *         for i in range(slowness.size):
 *             slown.push_back(slowness[i])
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def set_velocity(self, velocity):
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1256, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1236
 *         return tt
 * 
 *     def set_slowness(self, slowness):             # <<<<<<<<<<<<<<
 *         """
 *         set_slowness(slowness)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.set_slowness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1258
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_15set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_14set_velocity[] = "\n        set_velocity(velocity)\n\n        Assign velocity to grid\n\n        Parameters\n        ----------\n        velocity : np ndarray, shape (nparams, )\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_15set_velocity(PyObject *__pyx_v_self, PyObject *__pyx_v_velocity) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_velocity (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_14set_velocity(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), ((PyObject *)__pyx_v_velocity));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_14set_velocity(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_velocity) {
  std::vector<double>  __pyx_v_slown;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_velocity", 0);
  __Pyx_INCREF(__pyx_v_velocity);

  /* "ttcrpy/tmesh.pyx":1268
 *         velocity : np ndarray, shape (nparams, )
 *         """
 *         if velocity.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nparams); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "ttcrpy/tmesh.pyx":1269
 *         """
 *         if velocity.size != self.nparams:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1269, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":1268
 *         velocity : np ndarray, shape (nparams, )
 *         """
 *         if velocity.size != self.nparams:             # <<<<<<<<<<<<<<
 *             raise ValueError('velocity vector has wrong size')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1271
 *             raise ValueError('velocity vector has wrong size')
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             velocity = np.ascontiguousarray(velocity)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_C_CONTIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "ttcrpy/tmesh.pyx":1272
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:
 *             velocity = np.ascontiguousarray(velocity)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[double] slown
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_velocity) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_velocity);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_velocity, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1271
 *             raise ValueError('velocity vector has wrong size')
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:             # <<<<<<<<<<<<<<
 *             velocity = np.ascontiguousarray(velocity)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1276
 *         cdef vector[double] slown
 *         cdef int i
 *         for i in range(velocity.size):             # <<<<<<<<<<<<<<
 *             slown.push_back(1./velocity[i])
 *         self.grid.setSlowness(slown)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_velocity, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ttcrpy/tmesh.pyx":1277
 *         cdef int i
 *         for i in range(velocity.size):
 *             slown.push_back(1./velocity[i])             # <<<<<<<<<<<<<<
 *         self.grid.setSlowness(slown)
 * 
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_velocity, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_2, 1., 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    try {
      __pyx_v_slown.push_back(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1277, __pyx_L1_error)
    }
  }

  /* "ttcrpy/tmesh.pyx":1278
 *         for i in range(velocity.size):
 *             slown.push_back(1./velocity[i])
 *         self.grid.setSlowness(slown)             # <<<<<<<<<<<<<<
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 */
  try {
    __pyx_v_self->grid->setSlowness(__pyx_v_slown);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1278, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1258
 *         self.grid.setSlowness(slown)
 * 
 *     def set_velocity(self, velocity):             # <<<<<<<<<<<<<<
 *         """
 *         set_velocity(velocity)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.set_velocity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_velocity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1280
 *         self.grid.setSlowness(slown)
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, return_rays=False):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_17raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_16raytrace[] = "\n        raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, return_rays=False) -> tt, rays\n\n        Perform raytracing\n\n        Parameters\n        ----------\n        source : 2D np.ndarray with 2 or 3 columns\n            see notes below\n        rcv : 2D np.ndarray with 2 columns\n            Columns correspond to x and z coordinates\n        slowness : np ndarray, (None by default)\n            slowness at grid nodes or cells (depending on cell_slowness)\n            if None, slowness must have been assigned previously\n        thread_no : int (None by default)\n            Perform calculations in thread number \"thread_no\"\n            if None, attempt to run in parallel if warranted by number of\n            sources and value of n_threads in constructor\n        aggregate_src : bool (False by default)\n            if True, all source coordinates belong to a single event\n        return_rays : bool (False by default)\n            Return raypaths\n\n        Returns\n        -------\n        tt : np.ndarray\n            travel times for the appropriate source-rcv  (see Notes below)\n        rays : :obj:`list` of :obj:`np.ndarray`\n            Coordinates of segments forming raypaths (if return_rays is True)\n\n        Notes\n        -----\n        If source has 2 columns:\n            - Columns correspond to x and z coordinates\n            - Origin time (t0) is 0 for all points\n        If source has 3 columns:\n            - 1st column corresponds to origin times\n            - 2nd & 3rd columns correspond to x and z coordinates\n\n        source and rcv can contain the same number of rows, each row\n        corresponding to a source-receiver pair, or the number of rows may\n        differ if aggregate_src is True or if all rows in source are identical.\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_17raytrace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_rcv = 0;
  PyObject *__pyx_v_slowness = 0;
  PyObject *__pyx_v_thread_no = 0;
  PyObject *__pyx_v_aggregate_src = 0;
  PyObject *__pyx_v_return_rays = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("raytrace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_rcv,&__pyx_n_s_slowness,&__pyx_n_s_thread_no,&__pyx_n_s_aggregate_src,&__pyx_n_s_return_rays,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);

    /* "ttcrpy/tmesh.pyx":1281
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,
 *                  aggregate_src=False, return_rays=False):             # <<<<<<<<<<<<<<
 *         """
 *         raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, return_rays=False) -> tt, rays
 */
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rcv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 6, 1); __PYX_ERR(0, 1280, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thread_no);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aggregate_src);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_rays);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "raytrace") < 0)) __PYX_ERR(0, 1280, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_rcv = values[1];
    __pyx_v_slowness = values[2];
    __pyx_v_thread_no = values[3];
    __pyx_v_aggregate_src = values[4];
    __pyx_v_return_rays = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("raytrace", 0, 2, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1280, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_16raytrace(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), __pyx_v_source, __pyx_v_rcv, __pyx_v_slowness, __pyx_v_thread_no, __pyx_v_aggregate_src, __pyx_v_return_rays);

  /* "ttcrpy/tmesh.pyx":1280
 *         self.grid.setSlowness(slown)
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, return_rays=False):
 *         """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_16raytrace(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_rcv, PyObject *__pyx_v_slowness, PyObject *__pyx_v_thread_no, PyObject *__pyx_v_aggregate_src, PyObject *__pyx_v_return_rays) {
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_Tx = NULL;
  PyObject *__pyx_v_t0 = NULL;
  PyObject *__pyx_v_nTx = NULL;
  PyObject *__pyx_v_tmp = NULL;
  std::vector<std::vector<ttcr::sxz<double> > >  __pyx_v_vTx;
  std::vector<std::vector<ttcr::sxz<double> > >  __pyx_v_vRx;
  std::vector<std::vector<double> >  __pyx_v_vt0;
  std::vector<std::vector<double> >  __pyx_v_vtt;
  std::vector<std::vector<std::vector<ttcr::sxz<double> > > >  __pyx_v_r_data;
  size_t __pyx_v_thread_nb;
  int __pyx_v_n;
  int __pyx_v_n2;
  int __pyx_v_nt;
  PyObject *__pyx_v_iRx = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_ind = NULL;
  PyObject *__pyx_v_tt = NULL;
  PyObject *__pyx_v_rays = NULL;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_v_nn;
  CYTHON_UNUSED int __pyx_8genexpr2__pyx_v_n;
  CYTHON_UNUSED int __pyx_8genexpr3__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_7;
  std::vector<std::vector<double> > ::size_type __pyx_t_8;
  std::vector<std::vector<std::vector<ttcr::sxz<double> > > > ::size_type __pyx_t_9;
  double __pyx_t_10;
  double __pyx_t_11;
  ttcr::sxz<double>  __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  std::vector<double> ::size_type __pyx_t_15;
  int __pyx_t_16;
  long __pyx_t_17;
  long __pyx_t_18;
  int __pyx_t_19;
  size_t __pyx_t_20;
  std::vector<double> ::size_type __pyx_t_21;
  std::vector<double> ::size_type __pyx_t_22;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_23;
  std::vector<std::vector<ttcr::sxz<double> > > ::size_type __pyx_t_24;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_25;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_26;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_27;
  int __pyx_t_28;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_29;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_30;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raytrace", 0);

  /* "ttcrpy/tmesh.pyx":1328
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":1329
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if source.shape[1] == 2:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1329, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":1328
 *         # check input data consistency
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1331
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if source.shape[1] == 2:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":1332
 * 
 *         if source.shape[1] == 2:
 *             src = source             # <<<<<<<<<<<<<<
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 */
    __Pyx_INCREF(__pyx_v_source);
    __pyx_v_src = __pyx_v_source;

    /* "ttcrpy/tmesh.pyx":1333
 *         if source.shape[1] == 2:
 *             src = source
 *             Tx = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 1333, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_Tx = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "ttcrpy/tmesh.pyx":1334
 *             src = source
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))             # <<<<<<<<<<<<<<
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
    __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_t0 = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "ttcrpy/tmesh.pyx":1335
 *             Tx = np.unique(source, axis=0)
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]             # <<<<<<<<<<<<<<
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_Tx, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_nTx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1331
 *             raise ValueError('source and rcv should be 2D arrays')
 * 
 *         if source.shape[1] == 2:             # <<<<<<<<<<<<<<
 *             src = source
 *             Tx = np.unique(source, axis=0)
 */
    goto __pyx_L6;
  }

  /* "ttcrpy/tmesh.pyx":1336
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":1337
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_source, __pyx_tuple__30); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_src = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1338
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)             # <<<<<<<<<<<<<<
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unique); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_source);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 1338, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_tmp = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1339
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]             # <<<<<<<<<<<<<<
 *             Tx = tmp[:,1:3]
 *             t0 = tmp[:,0]
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmp, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_nTx = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1340
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]             # <<<<<<<<<<<<<<
 *             t0 = tmp[:,0]
 *         else:
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__30); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_Tx = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1341
 *             nTx = tmp.shape[0]
 *             Tx = tmp[:,1:3]
 *             t0 = tmp[:,0]             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_tmp, __pyx_tuple__2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_t0 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1336
 *             t0 = np.zeros((Tx.shape[0], 1))
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:             # <<<<<<<<<<<<<<
 *             src = source[:,1:3]
 *             tmp = np.unique(source, axis=0)
 */
    goto __pyx_L6;
  }

  /* "ttcrpy/tmesh.pyx":1343
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1343, __pyx_L1_error)
  }
  __pyx_L6:;

  /* "ttcrpy/tmesh.pyx":1345
 *             raise ValueError('source should be either nsrc x 2 or 3')
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L8_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "ttcrpy/tmesh.pyx":1346
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 *             raise ValueError('src and rcv should be ndata x 2')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1346, __pyx_L1_error)

    /* "ttcrpy/tmesh.pyx":1345
 *             raise ValueError('source should be either nsrc x 2 or 3')
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1348
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  __pyx_t_1 = (__pyx_v_slowness != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1349
 * 
 *         if slowness is not None:
 *             self.set_slowness(slowness)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[vector[sxz[double]]] vTx
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_slowness);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1348
 *             raise ValueError('src and rcv should be ndata x 2')
 * 
 *         if slowness is not None:             # <<<<<<<<<<<<<<
 *             self.set_slowness(slowness)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1361
 *         cdef int i, n, n2, nt
 * 
 *         vTx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L1_error)
  try {
    __pyx_v_vTx.resize(__pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1361, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1362
 * 
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)             # <<<<<<<<<<<<<<
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L1_error)
  try {
    __pyx_v_vRx.resize(__pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1362, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1363
 *         vTx.resize(nTx)
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)             # <<<<<<<<<<<<<<
 *         vtt.resize(nTx)
 *         if return_rays:
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L1_error)
  try {
    __pyx_v_vt0.resize(__pyx_t_8);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1363, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1364
 *         vRx.resize(nTx)
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)             # <<<<<<<<<<<<<<
 *         if return_rays:
 *             r_data.resize(nTx)
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  try {
    __pyx_v_vtt.resize(__pyx_t_8);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1364, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1365
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1365, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1366
 *         vtt.resize(nTx)
 *         if return_rays:
 *             r_data.resize(nTx)             # <<<<<<<<<<<<<<
 * 
 *         iRx = []
 */
    __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_nTx); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L1_error)
    try {
      __pyx_v_r_data.resize(__pyx_t_9);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1366, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1365
 *         vt0.resize(nTx)
 *         vtt.resize(nTx)
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             r_data.resize(nTx)
 * 
 */
  }

  /* "ttcrpy/tmesh.pyx":1368
 *             r_data.resize(nTx)
 * 
 *         iRx = []             # <<<<<<<<<<<<<<
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_iRx = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ttcrpy/tmesh.pyx":1369
 * 
 *         iRx = []
 *         if nTx == 1:             # <<<<<<<<<<<<<<
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 */
  __pyx_t_5 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1370
 *         iRx = []
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))             # <<<<<<<<<<<<<<
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__21); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_tuple__22); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    try {
      __pyx_t_12 = ttcr::sxz<double> (__pyx_t_10, __pyx_t_11);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1370, __pyx_L1_error)
    }
    try {
      (__pyx_v_vTx[0]).push_back(__pyx_t_12);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1370, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1371
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 */
    if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
      __pyx_t_5 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_5); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1371, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1371, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1371, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1371, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1371, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_14(__pyx_t_5);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1371, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":1372
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1372, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1372, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      try {
        __pyx_t_12 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_10);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1372, __pyx_L1_error)
      }
      try {
        (__pyx_v_vRx[0]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1372, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1371
 *         if nTx == 1:
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ttcrpy/tmesh.pyx":1373
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])             # <<<<<<<<<<<<<<
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_t0, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    try {
      (__pyx_v_vt0[0]).push_back(__pyx_t_10);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1373, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1374
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_15 = __Pyx_PyInt_As_size_t(__pyx_t_3); if (unlikely((__pyx_t_15 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    try {
      (__pyx_v_vtt[0]).resize(__pyx_t_15);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1374, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1375
 *             vt0[0].push_back(t0[0])
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *         elif aggregate_src:
 *             for t in Tx:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_3); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1369
 * 
 *         iRx = []
 *         if nTx == 1:             # <<<<<<<<<<<<<<
 *             vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
 *             for r in rcv:
 */
    goto __pyx_L12;
  }

  /* "ttcrpy/tmesh.pyx":1376
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:             # <<<<<<<<<<<<<<
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_aggregate_src); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1376, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1377
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 *             for t in Tx:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 */
    if (likely(PyList_CheckExact(__pyx_v_Tx)) || PyTuple_CheckExact(__pyx_v_Tx)) {
      __pyx_t_3 = __pyx_v_Tx; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_Tx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1377, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1377, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1377, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_14(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1377, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":1378
 *         elif aggregate_src:
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))             # <<<<<<<<<<<<<<
 *             for t in t0:
 *                 vt0[0].push_back(t)
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1378, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_t, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1378, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_t_12 = ttcr::sxz<double> (__pyx_t_10, __pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1378, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[0]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1378, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1377
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:
 *             for t in Tx:             # <<<<<<<<<<<<<<
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1379
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 */
    if (likely(PyList_CheckExact(__pyx_v_t0)) || PyTuple_CheckExact(__pyx_v_t0)) {
      __pyx_t_3 = __pyx_v_t0; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_t0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1379, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1379, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1379, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_14(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1379, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":1380
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:
 *                 vt0[0].push_back(t)             # <<<<<<<<<<<<<<
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 */
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_v_t); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L1_error)
      try {
        (__pyx_v_vt0[0]).push_back(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1380, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1379
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 *             for t in t0:             # <<<<<<<<<<<<<<
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1381
 *             for t in t0:
 *                 vt0[0].push_back(t)
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 */
    if (likely(PyList_CheckExact(__pyx_v_rcv)) || PyTuple_CheckExact(__pyx_v_rcv)) {
      __pyx_t_3 = __pyx_v_rcv; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_rcv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1381, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1381, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1381, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_14(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1381, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":1382
 *                 vt0[0].push_back(t)
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      try {
        __pyx_t_12 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_10);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1382, __pyx_L1_error)
      }
      try {
        (__pyx_v_vRx[0]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1382, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1381
 *             for t in t0:
 *                 vt0[0].push_back(t)
 *             for r in rcv:             # <<<<<<<<<<<<<<
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1383
 *             for r in rcv:
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])             # <<<<<<<<<<<<<<
 *             nTx = 1
 *             iRx.append(np.arange(rcv.shape[0]))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_15 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_15 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      (__pyx_v_vtt[0]).resize(__pyx_t_15);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1383, __pyx_L1_error)
    }

    /* "ttcrpy/tmesh.pyx":1384
 *                 vRx[0].push_back(sxz[double](r[0], r[1]))
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1             # <<<<<<<<<<<<<<
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 */
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_DECREF_SET(__pyx_v_nTx, __pyx_int_1);

    /* "ttcrpy/tmesh.pyx":1385
 *             vtt[0].resize(rcv.shape[0])
 *             nTx = 1
 *             iRx.append(np.arange(rcv.shape[0]))             # <<<<<<<<<<<<<<
 *         else:
 *             if src.shape != rcv.shape:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_arange); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_2); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1376
 *             vtt[0].resize(rcv.shape[0])
 *             iRx.append(np.arange(rcv.shape[0]))
 *         elif aggregate_src:             # <<<<<<<<<<<<<<
 *             for t in Tx:
 *                 vTx[0].push_back(sxz[double](t[0], t[1]))
 */
    goto __pyx_L12;
  }

  /* "ttcrpy/tmesh.pyx":1387
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_4)) {

      /* "ttcrpy/tmesh.pyx":1388
 *         else:
 *             if src.shape != rcv.shape:
 *                 raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *             for n in range(nTx):
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1388, __pyx_L1_error)

      /* "ttcrpy/tmesh.pyx":1387
 *             iRx.append(np.arange(rcv.shape[0]))
 *         else:
 *             if src.shape != rcv.shape:             # <<<<<<<<<<<<<<
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 */
    }

    /* "ttcrpy/tmesh.pyx":1390
 *                 raise ValueError('src and rcv should be of equal size')
 * 
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])
 */
    __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1390, __pyx_L1_error)
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_n = __pyx_t_19;

      /* "ttcrpy/tmesh.pyx":1391
 * 
 *             for n in range(nTx):
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2             # <<<<<<<<<<<<<<
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_slice_);
      __Pyx_GIVEREF(__pyx_slice_);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice_);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_v_src, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_1) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ind, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "ttcrpy/tmesh.pyx":1392
 *             for n in range(nTx):
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])             # <<<<<<<<<<<<<<
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_nonzero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_ind) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_ind);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_iRx, __pyx_t_5); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":1393
 *                 ind = np.sum(Tx[n,:] == src, axis=1) == 2
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))             # <<<<<<<<<<<<<<
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_Tx, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        __pyx_t_12 = ttcr::sxz<double> (__pyx_t_10, __pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1393, __pyx_L1_error)
      }
      try {
        (__pyx_v_vTx[__pyx_v_n]).push_back(__pyx_t_12);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1393, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1394
 *                 iRx.append(np.nonzero(ind)[0])
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])             # <<<<<<<<<<<<<<
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_t0, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      try {
        (__pyx_v_vt0[__pyx_v_n]).push_back(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1394, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1395
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())
 */
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_ind);
      __Pyx_GIVEREF(__pyx_v_ind);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_ind);
      __Pyx_INCREF(__pyx_slice_);
      __Pyx_GIVEREF(__pyx_slice_);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_slice_);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_rcv, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_13 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_13 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_14 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1395, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1395, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 1395, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_14(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1395, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "ttcrpy/tmesh.pyx":1396
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))             # <<<<<<<<<<<<<<
 *                 vtt[n].resize(vRx[n].size())
 * 
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        try {
          __pyx_t_12 = ttcr::sxz<double> (__pyx_t_11, __pyx_t_10);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1396, __pyx_L1_error)
        }
        try {
          (__pyx_v_vRx[__pyx_v_n]).push_back(__pyx_t_12);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1396, __pyx_L1_error)
        }

        /* "ttcrpy/tmesh.pyx":1395
 *                 vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
 *                 vt0[n].push_back(t0[n])
 *                 for r in rcv[ind,:]:             # <<<<<<<<<<<<<<
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ttcrpy/tmesh.pyx":1397
 *                 for r in rcv[ind,:]:
 *                     vRx[n].push_back(sxz[double](r[0], r[1]))
 *                 vtt[n].resize(vRx[n].size())             # <<<<<<<<<<<<<<
 * 
 *         tt = np.zeros((rcv.shape[0],))
 */
      try {
        (__pyx_v_vtt[__pyx_v_n]).resize((__pyx_v_vRx[__pyx_v_n]).size());
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1397, __pyx_L1_error)
      }
    }
  }
  __pyx_L12:;

  /* "ttcrpy/tmesh.pyx":1399
 *                 vtt[n].resize(vRx[n].size())
 * 
 *         tt = np.zeros((rcv.shape[0],))             # <<<<<<<<<<<<<<
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if return_rays==False:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_tt = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "ttcrpy/tmesh.pyx":1400
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if return_rays==False:
 *                 for n in range(nTx):
 */
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_n_threads); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_nTx, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L27_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_self->_n_threads == 1) != 0);
  __pyx_t_4 = __pyx_t_1;
  __pyx_L27_bool_binop_done:;
  if (__pyx_t_4) {

    /* "ttcrpy/tmesh.pyx":1401
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "ttcrpy/tmesh.pyx":1402
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if return_rays==False:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             else:
 */
      __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1402, __pyx_L1_error)
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_n = __pyx_t_19;

        /* "ttcrpy/tmesh.pyx":1403
 *             if return_rays==False:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 for n in range(nTx):
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1403, __pyx_L1_error)
        }
      }

      /* "ttcrpy/tmesh.pyx":1401
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:
 *             if return_rays==False:             # <<<<<<<<<<<<<<
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 */
      goto __pyx_L29;
    }

    /* "ttcrpy/tmesh.pyx":1405
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
 *             else:
 *                 for n in range(nTx):             # <<<<<<<<<<<<<<
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 */
    /*else*/ {
      __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1405, __pyx_L1_error)
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_n = __pyx_t_19;

        /* "ttcrpy/tmesh.pyx":1406
 *             else:
 *                 for n in range(nTx):
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)             # <<<<<<<<<<<<<<
 * 
 *         elif thread_no is not None:
 */
        try {
          __pyx_v_self->grid->raytrace((__pyx_v_vTx[__pyx_v_n]), (__pyx_v_vt0[__pyx_v_n]), (__pyx_v_vRx[__pyx_v_n]), (__pyx_v_vtt[__pyx_v_n]), (__pyx_v_r_data[__pyx_v_n]), 0);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1406, __pyx_L1_error)
        }
      }
    }
    __pyx_L29:;

    /* "ttcrpy/tmesh.pyx":1400
 * 
 *         tt = np.zeros((rcv.shape[0],))
 *         if nTx < self._n_threads or self._n_threads == 1:             # <<<<<<<<<<<<<<
 *             if return_rays==False:
 *                 for n in range(nTx):
 */
    goto __pyx_L26;
  }

  /* "ttcrpy/tmesh.pyx":1408
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  __pyx_t_4 = (__pyx_v_thread_no != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":1410
 *         elif thread_no is not None:
 *             # we should be here for just one event
 *             assert nTx == 1             # <<<<<<<<<<<<<<
 *             thread_nb = thread_no
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_nTx, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1410, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 1410, __pyx_L1_error)
      }
    }
    #endif

    /* "ttcrpy/tmesh.pyx":1411
 *             # we should be here for just one event
 *             assert nTx == 1
 *             thread_nb = thread_no             # <<<<<<<<<<<<<<
 * 
 *             if return_rays:
 */
    __pyx_t_20 = __Pyx_PyInt_As_size_t(__pyx_v_thread_no); if (unlikely((__pyx_t_20 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L1_error)
    __pyx_v_thread_nb = __pyx_t_20;

    /* "ttcrpy/tmesh.pyx":1413
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1413, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":1414
 * 
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), (__pyx_v_r_data[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1414, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1415
 *             if return_rays:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 */
      __pyx_t_21 = (__pyx_v_vtt[0]).size();
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_22; __pyx_t_19+=1) {
        __pyx_v_nt = __pyx_t_19;

        /* "ttcrpy/tmesh.pyx":1416
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1416, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1416, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/tmesh.pyx":1417
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_rays = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":1418
 *                     tt[nt] = vtt[0][nt]
 *                 rays = []
 *                 for n2 in range(vRx.size()):             # <<<<<<<<<<<<<<
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):
 */
      __pyx_t_23 = __pyx_v_vRx.size();
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_24; __pyx_t_19+=1) {
        __pyx_v_n2 = __pyx_t_19;

        /* "ttcrpy/tmesh.pyx":1419
 *                 rays = []
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[0])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_2);
        __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":1420
 *                 for n2 in range(vRx.size()):
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z
 */
        __pyx_t_25 = ((__pyx_v_r_data[0])[__pyx_v_n2]).size();
        __pyx_t_26 = __pyx_t_25;
        for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
          __pyx_v_nn = __pyx_t_27;

          /* "ttcrpy/tmesh.pyx":1421
 *                     r = np.empty((r_data[0][n2].size(), 2))
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1421, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1421, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1421, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
          __pyx_t_2 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_6, __pyx_t_3) < 0)) __PYX_ERR(0, 1421, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ttcrpy/tmesh.pyx":1422
 *                     for nn in range(r_data[0][n2].size()):
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z             # <<<<<<<<<<<<<<
 *                     rays.append(r)
 *                 return tt, rays
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[0])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1422, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_nn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1422, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1422, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
          __pyx_t_6 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_r, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 1422, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "ttcrpy/tmesh.pyx":1423
 *                         r[nn, 0] = r_data[0][n2][nn].x
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)             # <<<<<<<<<<<<<<
 *                 return tt, rays
 * 
 */
        __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_rays, __pyx_v_r); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1423, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1424
 *                         r[nn, 1] = r_data[0][n2][nn].z
 *                     rays.append(r)
 *                 return tt, rays             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1424, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_tt);
      __Pyx_GIVEREF(__pyx_v_tt);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
      __Pyx_INCREF(__pyx_v_rays);
      __Pyx_GIVEREF(__pyx_v_rays);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ttcrpy/tmesh.pyx":1413
 *             thread_nb = thread_no
 * 
 *             if return_rays:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 */
    }

    /* "ttcrpy/tmesh.pyx":1427
 * 
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace((__pyx_v_vTx[0]), (__pyx_v_vt0[0]), (__pyx_v_vRx[0]), (__pyx_v_vtt[0]), __pyx_v_thread_nb);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1427, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1428
 *             else:
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):             # <<<<<<<<<<<<<<
 *                     tt[nt] = vtt[0][nt]
 *                 return tt
 */
      __pyx_t_21 = (__pyx_v_vtt[0]).size();
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_22; __pyx_t_19+=1) {
        __pyx_v_nt = __pyx_t_19;

        /* "ttcrpy/tmesh.pyx":1429
 *                 self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]             # <<<<<<<<<<<<<<
 *                 return tt
 * 
 */
        __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[0])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_tt, __pyx_v_nt, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1429, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "ttcrpy/tmesh.pyx":1430
 *                 for nt in range(vtt[0].size()):
 *                     tt[nt] = vtt[0][nt]
 *                 return tt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_tt);
      __pyx_r = __pyx_v_tt;
      goto __pyx_L0;
    }

    /* "ttcrpy/tmesh.pyx":1408
 *                     self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)
 * 
 *         elif thread_no is not None:             # <<<<<<<<<<<<<<
 *             # we should be here for just one event
 *             assert nTx == 1
 */
  }

  /* "ttcrpy/tmesh.pyx":1433
 * 
 *         else:
 *             if return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             else:
 */
  /*else*/ {
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "ttcrpy/tmesh.pyx":1434
 *         else:
 *             if return_rays==False:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 */
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1434, __pyx_L1_error)
      }

      /* "ttcrpy/tmesh.pyx":1433
 * 
 *         else:
 *             if return_rays==False:             # <<<<<<<<<<<<<<
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             else:
 */
      goto __pyx_L43;
    }

    /* "ttcrpy/tmesh.pyx":1436
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt)
 *             else:
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)             # <<<<<<<<<<<<<<
 * 
 *         for n in range(nTx):
 */
    /*else*/ {
      try {
        __pyx_v_self->grid->raytrace(__pyx_v_vTx, __pyx_v_vt0, __pyx_v_vRx, __pyx_v_vtt, __pyx_v_r_data);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1436, __pyx_L1_error)
      }
    }
    __pyx_L43:;
  }
  __pyx_L26:;

  /* "ttcrpy/tmesh.pyx":1438
 *                 self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)
 * 
 *         for n in range(nTx):             # <<<<<<<<<<<<<<
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 */
  __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1438, __pyx_L1_error)
  __pyx_t_18 = __pyx_t_17;
  for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
    __pyx_v_n = __pyx_t_19;

    /* "ttcrpy/tmesh.pyx":1439
 * 
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 */
    __pyx_t_21 = (__pyx_v_vtt[__pyx_v_n]).size();
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_22; __pyx_t_28+=1) {
      __pyx_v_nt = __pyx_t_28;

      /* "ttcrpy/tmesh.pyx":1440
 *         for n in range(nTx):
 *             for nt in range(vtt[n].size()):
 *                 tt[iRx[n][nt]] = vtt[n][nt]             # <<<<<<<<<<<<<<
 * 
 *         if return_rays:
 */
      __pyx_t_3 = PyFloat_FromDouble(((__pyx_v_vtt[__pyx_v_n])[__pyx_v_nt])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tt, __pyx_t_6, __pyx_t_3) < 0)) __PYX_ERR(0, 1440, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":1442
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_rays); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1442, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":1443
 * 
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]             # <<<<<<<<<<<<<<
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rcv, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_8genexpr2__pyx_v_n = __pyx_t_19;
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1443, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 1443, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    } /* exit inner scope */
    __pyx_v_rays = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1444
 *         if return_rays:
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):             # <<<<<<<<<<<<<<
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 */
    __pyx_t_17 = __Pyx_PyInt_As_long(__pyx_v_nTx); if (unlikely((__pyx_t_17 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1444, __pyx_L1_error)
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_n = __pyx_t_19;

      /* "ttcrpy/tmesh.pyx":1445
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]             # <<<<<<<<<<<<<<
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1445, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_25 = (__pyx_v_vRx[__pyx_v_n]).size();
        __pyx_t_26 = __pyx_t_25;
        for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_26; __pyx_t_28+=1) {
          __pyx_8genexpr3__pyx_v_i = __pyx_t_28;
          __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1445, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_float_0_0);
          __Pyx_GIVEREF(__pyx_float_0_0);
          PyList_SET_ITEM(__pyx_t_2, 0, __pyx_float_0_0);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 1445, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "ttcrpy/tmesh.pyx":1446
 *             for n in range(nTx):
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):             # <<<<<<<<<<<<<<
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):
 */
      __pyx_t_25 = (__pyx_v_vRx[__pyx_v_n]).size();
      __pyx_t_26 = __pyx_t_25;
      for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_26; __pyx_t_28+=1) {
        __pyx_v_n2 = __pyx_t_28;

        /* "ttcrpy/tmesh.pyx":1447
 *                 r = [ [0.0] for i in range(vRx[n].size())]
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))             # <<<<<<<<<<<<<<
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_v_r, __pyx_v_n2, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1447, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "ttcrpy/tmesh.pyx":1448
 *                 for n2 in range(vRx[n].size()):
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 */
        __pyx_t_27 = ((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2]).size();
        __pyx_t_29 = __pyx_t_27;
        for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
          __pyx_v_nn = __pyx_t_30;

          /* "ttcrpy/tmesh.pyx":1449
 *                     r[n2] = np.empty((r_data[n][n2].size(), 2))
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x             # <<<<<<<<<<<<<<
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1449, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1449, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_nn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1449, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1449, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
          __pyx_t_5 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 1449, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ttcrpy/tmesh.pyx":1450
 *                     for nn in range(r_data[n][n2].size()):
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z             # <<<<<<<<<<<<<<
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]
 */
          __pyx_t_3 = PyFloat_FromDouble((((__pyx_v_r_data[__pyx_v_n])[__pyx_v_n2])[__pyx_v_nn]).z); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1450, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_n2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1450, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_nn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1450, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1450, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
          __pyx_t_6 = 0;
          if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_5, __pyx_t_3) < 0)) __PYX_ERR(0, 1450, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
      }

      /* "ttcrpy/tmesh.pyx":1451
 *                         r[n2][nn, 0] = r_data[n][n2][nn].x
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):             # <<<<<<<<<<<<<<
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 */
      __pyx_t_21 = (__pyx_v_vtt[__pyx_v_n]).size();
      __pyx_t_22 = __pyx_t_21;
      for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_22; __pyx_t_28+=1) {
        __pyx_v_nt = __pyx_t_28;

        /* "ttcrpy/tmesh.pyx":1452
 *                         r[n2][nn, 1] = r_data[n][n2][nn].z
 *                 for nt in range(vtt[n].size()):
 *                     rays[iRx[n][nt]] = r[nt]             # <<<<<<<<<<<<<<
 * 
 *         if return_rays==False:
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_r, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_iRx, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_nt, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_rays, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }

    /* "ttcrpy/tmesh.pyx":1442
 *                 tt[iRx[n][nt]] = vtt[n][nt]
 * 
 *         if return_rays:             # <<<<<<<<<<<<<<
 *             rays = [ [0.0] for n in range(rcv.shape[0])]
 *             for n in range(nTx):
 */
  }

  /* "ttcrpy/tmesh.pyx":1454
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         else:
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_return_rays, Py_False, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1454, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1454, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "ttcrpy/tmesh.pyx":1455
 * 
 *         if return_rays==False:
 *             return tt             # <<<<<<<<<<<<<<
 *         else:
 *             return tt, rays
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_tt);
    __pyx_r = __pyx_v_tt;
    goto __pyx_L0;

    /* "ttcrpy/tmesh.pyx":1454
 *                     rays[iRx[n][nt]] = r[nt]
 * 
 *         if return_rays==False:             # <<<<<<<<<<<<<<
 *             return tt
 *         else:
 */
  }

  /* "ttcrpy/tmesh.pyx":1457
 *             return tt
 *         else:
 *             return tt, rays             # <<<<<<<<<<<<<<
 * 
 *     def to_vtk(self, fields, filename):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_rays)) { __Pyx_RaiseUnboundLocalError("rays"); __PYX_ERR(0, 1457, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_tt);
    __Pyx_GIVEREF(__pyx_v_tt);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tt);
    __Pyx_INCREF(__pyx_v_rays);
    __Pyx_GIVEREF(__pyx_v_rays);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rays);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "ttcrpy/tmesh.pyx":1280
 *         self.grid.setSlowness(slown)
 * 
 *     def raytrace(self, source, rcv, slowness=None, thread_no=None,             # <<<<<<<<<<<<<<
 *                  aggregate_src=False, return_rays=False):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.raytrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_Tx);
  __Pyx_XDECREF(__pyx_v_t0);
  __Pyx_XDECREF(__pyx_v_nTx);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_iRx);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XDECREF(__pyx_v_tt);
  __Pyx_XDECREF(__pyx_v_rays);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1459
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_19to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_18to_vtk[] = "\n        to_vtk(fields, filename)\n\n        Save mesh variables and/or raypaths to VTK format\n\n        Parameters\n        ----------\n        fields: dict\n            dict of variables to save to file. Variables should be np.ndarray of\n            size equal to either the number of nodes of the number of cells of\n            the mesh, or a list of raypath coordinates.\n        filename: str\n            Name of file without extension for saving (extension vtu will be\n            added).  Raypaths are saved in separate files, and filename will\n            be appended by the dict key and have a vtp extension.\n\n        Notes\n        -----\n        VTK files can be visualized with Paraview (https://www.paraview.org)\n        ";
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_19to_vtk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_vtk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fields,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, 1); __PYX_ERR(0, 1459, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "to_vtk") < 0)) __PYX_ERR(0, 1459, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fields = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_vtk", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1459, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_18to_vtk(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), __pyx_v_fields, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_18to_vtk(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_fields, PyObject *__pyx_v_filename) {
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_ugrid = NULL;
  PyObject *__pyx_v_tPts = NULL;
  PyObject *__pyx_v_tri = NULL;
  int __pyx_v_save_grid;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_scalar = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_5;
  std::vector<ttcr::sxz<double> > ::size_type __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  std::vector<ttcr::triangleElem<uint32_t> > ::size_type __pyx_t_12;
  std::vector<ttcr::triangleElem<uint32_t> > ::size_type __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  int __pyx_t_18;
  long __pyx_t_19;
  long __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_vtk", 0);

  /* "ttcrpy/tmesh.pyx":1481
 *         """
 *         cdef int n, nn
 *         ugrid = vtk.vtkUnstructuredGrid()             # <<<<<<<<<<<<<<
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkUnstructuredGrid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ugrid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1482
 *         cdef int n, nn
 *         ugrid = vtk.vtkUnstructuredGrid()
 *         tPts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tPts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1483
 *         ugrid = vtk.vtkUnstructuredGrid()
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())             # <<<<<<<<<<<<<<
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tPts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->no.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1484
 *         tPts = vtk.vtkPoints()
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):             # <<<<<<<<<<<<<<
 *             tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)
 *         ugrid.SetPoints(tPts)
 */
  __pyx_t_5 = __pyx_v_self->no.size();
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":1485
 *         tPts.SetNumberOfPoints(self.no.size())
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)             # <<<<<<<<<<<<<<
 *         ugrid.SetPoints(tPts)
 *         tri = vtk.vtkTriangle
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tPts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyFloat_FromDouble((__pyx_v_self->no[__pyx_v_n]).z); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_float_0_0, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_float_0_0, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_4);
      __Pyx_INCREF(__pyx_float_0_0);
      __Pyx_GIVEREF(__pyx_float_0_0);
      PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_float_0_0);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_8);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1486
 *         for n in range(self.no.size()):
 *             tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)
 *         ugrid.SetPoints(tPts)             # <<<<<<<<<<<<<<
 *         tri = vtk.vtkTriangle
 *         for n in range(self.tri.size()):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_11, __pyx_v_tPts) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_tPts);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1487
 *             tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)
 *         ugrid.SetPoints(tPts)
 *         tri = vtk.vtkTriangle             # <<<<<<<<<<<<<<
 *         for n in range(self.tri.size()):
 *             for nn in range(3):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkTriangle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_tri = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":1488
 *         ugrid.SetPoints(tPts)
 *         tri = vtk.vtkTriangle
 *         for n in range(self.tri.size()):             # <<<<<<<<<<<<<<
 *             for nn in range(3):
 *                 tri.GetPointIds().SetId(nn, self.tet[n].i[nn])
 */
  __pyx_t_12 = __pyx_v_self->tri.size();
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_n = __pyx_t_7;

    /* "ttcrpy/tmesh.pyx":1489
 *         tri = vtk.vtkTriangle
 *         for n in range(self.tri.size()):
 *             for nn in range(3):             # <<<<<<<<<<<<<<
 *                 tri.GetPointIds().SetId(nn, self.tet[n].i[nn])
 *             ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())
 */
    for (__pyx_t_10 = 0; __pyx_t_10 < 3; __pyx_t_10+=1) {
      __pyx_v_nn = __pyx_t_10;

      /* "ttcrpy/tmesh.pyx":1490
 *         for n in range(self.tri.size()):
 *             for nn in range(3):
 *                 tri.GetPointIds().SetId(nn, self.tet[n].i[nn])             # <<<<<<<<<<<<<<
 *             ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())
 * 
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_tri, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SetId); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tet); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_8, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_8, __pyx_v_nn, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_11)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1490, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1490, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_14, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_14, __pyx_t_4);
        __pyx_t_1 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1490, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "ttcrpy/tmesh.pyx":1491
 *             for nn in range(3):
 *                 tri.GetPointIds().SetId(nn, self.tet[n].i[nn])
 *             ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())             # <<<<<<<<<<<<<<
 * 
 *         save_grid = False
 */
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_tri, __pyx_n_s_GetCellType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tri, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_3, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1491, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_3, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1491, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_10, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_10, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1493
 *             ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())
 * 
 *         save_grid = False             # <<<<<<<<<<<<<<
 *         for fn in fields:
 *             data = fields[fn]
 */
  __pyx_v_save_grid = 0;

  /* "ttcrpy/tmesh.pyx":1494
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
    __pyx_t_2 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
    __pyx_t_16 = NULL;
  } else {
    __pyx_t_15 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_16 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1494, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_16)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_11 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_11); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1494, __pyx_L1_error)
        #else
        __pyx_t_11 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
      } else {
        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_11); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1494, __pyx_L1_error)
        #else
        __pyx_t_11 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
      }
    } else {
      __pyx_t_11 = __pyx_t_16(__pyx_t_2);
      if (unlikely(!__pyx_t_11)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1494, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_11);
    }
    __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "ttcrpy/tmesh.pyx":1495
 *         save_grid = False
 *         for fn in fields:
 *             data = fields[fn]             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(data, list):
 */
    __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_fields, __pyx_v_fn); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "ttcrpy/tmesh.pyx":1497
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
    __pyx_t_17 = PyList_Check(__pyx_v_data); 
    __pyx_t_18 = (__pyx_t_17 != 0);
    if (__pyx_t_18) {

      /* "ttcrpy/tmesh.pyx":1498
 * 
 *             if isinstance(data, list):
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')             # <<<<<<<<<<<<<<
 *             else:
 *                 save_grid = True
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_save_raypaths); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = PyNumber_Add(__pyx_v_filename, __pyx_n_u__25); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_fn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_u_vtp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_data, __pyx_t_4};
        __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1498, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_data, __pyx_t_4};
        __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1498, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_data);
        __Pyx_GIVEREF(__pyx_v_data);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_7, __pyx_v_data);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ttcrpy/tmesh.pyx":1497
 *             data = fields[fn]
 * 
 *             if isinstance(data, list):             # <<<<<<<<<<<<<<
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 */
      goto __pyx_L11;
    }

    /* "ttcrpy/tmesh.pyx":1500
 *                 self._save_raypaths(data, filename+'_'+fn+'.vtp')
 *             else:
 *                 save_grid = True             # <<<<<<<<<<<<<<
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 */
    /*else*/ {
      __pyx_v_save_grid = 1;

      /* "ttcrpy/tmesh.pyx":1501
 *             else:
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()             # <<<<<<<<<<<<<<
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_vtk); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_vtkDoubleArray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_11 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_scalar, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "ttcrpy/tmesh.pyx":1502
 *                 save_grid = True
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_11 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_fn);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ttcrpy/tmesh.pyx":1503
 *                 scalar = vtk.vtkDoubleArray()
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)             # <<<<<<<<<<<<<<
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfComponents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_11 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ttcrpy/tmesh.pyx":1504
 *                 scalar.SetName(fn)
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)             # <<<<<<<<<<<<<<
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetNumberOfTuples); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_11 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ttcrpy/tmesh.pyx":1505
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_nodes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_11, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1505, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 1505, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_18) {

        /* "ttcrpy/tmesh.pyx":1506
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1506, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_19 = __Pyx_PyInt_As_long(__pyx_t_8); if (unlikely((__pyx_t_19 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_20 = __pyx_t_19;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_20; __pyx_t_7+=1) {
          __pyx_v_n = __pyx_t_7;

          /* "ttcrpy/tmesh.pyx":1507
 *                 if data.size == self.get_number_of_nodes():
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])             # <<<<<<<<<<<<<<
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1507, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_data, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1507, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_11, __pyx_t_4};
            __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1507, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_11, __pyx_t_4};
            __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1507, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_9 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1507, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_11);
            PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_t_4);
            __pyx_t_11 = 0;
            __pyx_t_4 = 0;
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1507, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }

        /* "ttcrpy/tmesh.pyx":1508
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1508, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1508, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_8 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1508, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "ttcrpy/tmesh.pyx":1505
 *                 scalar.SetNumberOfComponents(1)
 *                 scalar.SetNumberOfTuples(data.size)
 *                 if data.size == self.get_number_of_nodes():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
        goto __pyx_L12;
      }

      /* "ttcrpy/tmesh.pyx":1509
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_number_of_cells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_9 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_8, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(__pyx_t_18)) {

        /* "ttcrpy/tmesh.pyx":1510
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):             # <<<<<<<<<<<<<<
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetCellData().AddArray(scalar)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_19 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_19 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1510, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_20 = __pyx_t_19;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_20; __pyx_t_7+=1) {
          __pyx_v_n = __pyx_t_7;

          /* "ttcrpy/tmesh.pyx":1511
 *                 elif data.size == self.get_number_of_cells():
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])             # <<<<<<<<<<<<<<
 *                     ugrid.GetCellData().AddArray(scalar)
 *                 else:
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_scalar, __pyx_n_s_SetTuple1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1511, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1511, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_data, __pyx_v_n, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1511, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_8, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_8, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_11) {
              __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_11); __pyx_t_11 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_10, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_10, __pyx_t_4);
            __pyx_t_8 = 0;
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }

        /* "ttcrpy/tmesh.pyx":1512
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetCellData().AddArray(scalar)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ugrid, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_9 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_AddArray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_scalar) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_scalar);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ttcrpy/tmesh.pyx":1509
 *                         scalar.SetTuple1(n, data[n])
 *                     ugrid.GetPointData().AddArray(scalar)
 *                 elif data.size == self.get_number_of_cells():             # <<<<<<<<<<<<<<
 *                     for n in range(data.size):
 *                         scalar.SetTuple1(n, data[n])
 */
        goto __pyx_L12;
      }

      /* "ttcrpy/tmesh.pyx":1514
 *                     ugrid.GetCellData().AddArray(scalar)
 *                 else:
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))             # <<<<<<<<<<<<<<
 * 
 *         if save_grid:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_fn) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_fn);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1514, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 1514, __pyx_L1_error)
      }
      __pyx_L12:;
    }
    __pyx_L11:;

    /* "ttcrpy/tmesh.pyx":1494
 * 
 *         save_grid = False
 *         for fn in fields:             # <<<<<<<<<<<<<<
 *             data = fields[fn]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":1516
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 */
  __pyx_t_18 = (__pyx_v_save_grid != 0);
  if (__pyx_t_18) {

    /* "ttcrpy/tmesh.pyx":1517
 * 
 *         if save_grid:
 *             writer = vtk.vtkXMLUnstructuredGridWriter()             # <<<<<<<<<<<<<<
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkXMLUnstructuredGridWriter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_writer = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1518
 *         if save_grid:
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')             # <<<<<<<<<<<<<<
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_v_filename, __pyx_kp_u_vtu); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1519
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)             # <<<<<<<<<<<<<<
 *             writer.SetDataModeToBinary()
 *             writer.Update()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_ugrid) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_ugrid);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1520
 *             writer.SetFileName(filename+'.vtu')
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *             writer.Update()
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1521
 *             writer.SetInputData(ugrid)
 *             writer.SetDataModeToBinary()
 *             writer.Update()             # <<<<<<<<<<<<<<
 * 
 *     def  _save_raypaths(self, rays, filename):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ttcrpy/tmesh.pyx":1516
 *                     raise ValueError('Field {0:s} has incorrect size'.format(fn))
 * 
 *         if save_grid:             # <<<<<<<<<<<<<<
 *             writer = vtk.vtkXMLUnstructuredGridWriter()
 *             writer.SetFileName(filename+'.vtu')
 */
  }

  /* "ttcrpy/tmesh.pyx":1459
 *             return tt, rays
 * 
 *     def to_vtk(self, fields, filename):             # <<<<<<<<<<<<<<
 *         """
 *         to_vtk(fields, filename)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.to_vtk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ugrid);
  __Pyx_XDECREF(__pyx_v_tPts);
  __Pyx_XDECREF(__pyx_v_tri);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_scalar);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1523
 *             writer.Update()
 * 
 *     def  _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_21_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_21_save_raypaths(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rays = 0;
  PyObject *__pyx_v_filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_save_raypaths (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rays,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rays)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, 1); __PYX_ERR(0, 1523, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_save_raypaths") < 0)) __PYX_ERR(0, 1523, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rays = values[0];
    __pyx_v_filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_save_raypaths", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1523, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_20_save_raypaths(((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_v_self), __pyx_v_rays, __pyx_v_filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_20_save_raypaths(CYTHON_UNUSED struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_self, PyObject *__pyx_v_rays, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_polydata = NULL;
  PyObject *__pyx_v_cellarray = NULL;
  PyObject *__pyx_v_pts = NULL;
  PyObject *__pyx_v_npts = NULL;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_writer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_save_raypaths", 0);

  /* "ttcrpy/tmesh.pyx":1524
 * 
 *     def  _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()             # <<<<<<<<<<<<<<
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPolyData); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_polydata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1525
 *     def  _save_raypaths(self, rays, filename):
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()             # <<<<<<<<<<<<<<
 *         pts = vtk.vtkPoints()
 *         npts = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtkCellArray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cellarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1526
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkPoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pts = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1527
 *         cellarray = vtk.vtkCellArray()
 *         pts = vtk.vtkPoints()
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_npts = __pyx_int_0;

  /* "ttcrpy/tmesh.pyx":1528
 *         pts = vtk.vtkPoints()
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1528, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":1529
 *         npts = 0
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]             # <<<<<<<<<<<<<<
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_npts, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1530
 *         for n in range(len(rays)):
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_SetNumberOfPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_npts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_npts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1531
 *             npts += rays[n].shape[0]
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/tmesh.pyx":1532
 *         pts.SetNumberOfPoints(npts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1532, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":1533
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1533, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1533, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1533, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1533, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1534
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])             # <<<<<<<<<<<<<<
 *                 npts += 1
 *         polydata.SetPoints(pts)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pts, __pyx_n_s_InsertPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_0);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_p);
      __Pyx_GIVEREF(__pyx_v_p);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_p);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_2);
      __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_npts, __pyx_t_11, __pyx_float_0_0, __pyx_t_12};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1534, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_npts, __pyx_t_11, __pyx_float_0_0, __pyx_t_12};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1534, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1534, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_13, __pyx_v_npts);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_13, __pyx_t_11);
        __Pyx_INCREF(__pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_float_0_0);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_13, __pyx_float_0_0);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_13, __pyx_t_12);
        __pyx_t_11 = 0;
        __pyx_t_12 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1534, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1535
 *             for p in range(rays[n].shape[0]):
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
 *                 npts += 1             # <<<<<<<<<<<<<<
 *         polydata.SetPoints(pts)
 *         npts = 0
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1533
 *         npts = 0
 *         for n in range(len(rays)):
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1536
 *                 pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
 *                 npts += 1
 *         polydata.SetPoints(pts)             # <<<<<<<<<<<<<<
 *         npts = 0
 *         for n in range(len(rays)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetPoints); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_pts) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_pts);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1537
 *                 npts += 1
 *         polydata.SetPoints(pts)
 *         npts = 0             # <<<<<<<<<<<<<<
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_DECREF_SET(__pyx_v_npts, __pyx_int_0);

  /* "ttcrpy/tmesh.pyx":1538
 *         polydata.SetPoints(pts)
 *         npts = 0
 *         for n in range(len(rays)):             # <<<<<<<<<<<<<<
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_rays); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1538, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_n = __pyx_t_6;

    /* "ttcrpy/tmesh.pyx":1539
 *         npts = 0
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()             # <<<<<<<<<<<<<<
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkPolyLine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1540
 *         for n in range(len(rays)):
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])             # <<<<<<<<<<<<<<
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetNumberOfIds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shape); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1541
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_rays, __pyx_v_n, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1541, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1541, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1541, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1541, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1542
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)             # <<<<<<<<<<<<<<
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_2 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SetId); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_p, __pyx_v_npts};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1542, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_p);
        __Pyx_GIVEREF(__pyx_v_p);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_13, __pyx_v_p);
        __Pyx_INCREF(__pyx_v_npts);
        __Pyx_GIVEREF(__pyx_v_npts);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_13, __pyx_v_npts);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1543
 *             for p in range(rays[n].shape[0]):
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1             # <<<<<<<<<<<<<<
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_npts, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1543, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_npts, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1541
 *             line = vtk.vtkPolyLine()
 *             line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
 *             for p in range(rays[n].shape[0]):             # <<<<<<<<<<<<<<
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ttcrpy/tmesh.pyx":1544
 *                 line.GetPointIds().SetId(p, npts)
 *                 npts += 1
 *             cellarray.InsertNextCell(line)             # <<<<<<<<<<<<<<
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cellarray, __pyx_n_s_InsertNextCell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_line);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "ttcrpy/tmesh.pyx":1545
 *                 npts += 1
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)             # <<<<<<<<<<<<<<
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_polydata, __pyx_n_s_SetLines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_10, __pyx_v_cellarray) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1546
 *             cellarray.InsertNextCell(line)
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()             # <<<<<<<<<<<<<<
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_vtk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vtkXMLPolyDataWriter); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_writer = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1547
 *         polydata.SetLines(cellarray)
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1548
 *         writer = vtk.vtkXMLPolyDataWriter()
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)             # <<<<<<<<<<<<<<
 *         writer.SetDataModeToBinary()
 *         writer.Update()
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetInputData); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_polydata) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_polydata);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1549
 *         writer.SetFileName(filename)
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()             # <<<<<<<<<<<<<<
 *         writer.Update()
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_SetDataModeToBinary); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1550
 *         writer.SetInputData(polydata)
 *         writer.SetDataModeToBinary()
 *         writer.Update()             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_Update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1523
 *             writer.Update()
 * 
 *     def  _save_raypaths(self, rays, filename):             # <<<<<<<<<<<<<<
 *         polydata = vtk.vtkPolyData()
 *         cellarray = vtk.vtkCellArray()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d._save_raypaths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_polydata);
  __Pyx_XDECREF(__pyx_v_cellarray);
  __Pyx_XDECREF(__pyx_v_pts);
  __Pyx_XDECREF(__pyx_v_npts);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1553
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM',double eps=1.e-15, int maxit=20,
 *                 bool process_obtuse=1,uint32_t n_secondary=5):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_23builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_22builder[] = "\n        builder(filename, n_threads, cell_slowness, method, eps, maxit, process_obtuse, n_secondary)\n\n        Build instance of Mesh2d from VTK file\n\n        Parameters\n        ----------\n        filename : str\n            Name of file holding a vtkUnstructuredGrid.\n            The grid must have point or cell attribute named either\n            'Slowness', 'slowness', 'Velocity', 'velocity', or\n            'P-wave velocity'.  All cells must be of type vtkTriangle\n\n        Other parameters are defined in Constructor\n\n        Returns\n        -------\n        mesh: :obj:`Mesh2d`\n            mesh instance\n        ";
static PyMethodDef __pyx_mdef_6ttcrpy_5tmesh_6Mesh2d_23builder = {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_23builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_22builder};
static PyObject *__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_23builder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  size_t __pyx_v_n_threads;
  bool __pyx_v_cell_slowness;
  PyObject *__pyx_v_method = 0;
  double __pyx_v_eps;
  int __pyx_v_maxit;
  bool __pyx_v_process_obtuse;
  uint32_t __pyx_v_n_secondary;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("builder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_n_threads,&__pyx_n_s_cell_slowness,&__pyx_n_s_method,&__pyx_n_s_eps,&__pyx_n_s_maxit,&__pyx_n_s_process_obtuse,&__pyx_n_s_n_secondary,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[3] = ((PyObject*)__pyx_n_u_FSM);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_threads);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cell_slowness);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxit);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_process_obtuse);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_secondary);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "builder") < 0)) __PYX_ERR(0, 1553, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    if (values[1]) {
      __pyx_v_n_threads = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_n_threads == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    } else {
      __pyx_v_n_threads = ((size_t)1);
    }
    if (values[2]) {
      __pyx_v_cell_slowness = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_cell_slowness == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    } else {
      __pyx_v_cell_slowness = ((bool)1);
    }
    __pyx_v_method = ((PyObject*)values[3]);
    if (values[4]) {
      __pyx_v_eps = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    } else {
      __pyx_v_eps = ((double)1.e-15);
    }
    if (values[5]) {
      __pyx_v_maxit = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_maxit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    } else {
      __pyx_v_maxit = ((int)20);
    }
    if (values[6]) {
      __pyx_v_process_obtuse = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_process_obtuse == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1555, __pyx_L3_error)
    } else {
      __pyx_v_process_obtuse = ((bool)1);
    }
    if (values[7]) {
      __pyx_v_n_secondary = __Pyx_PyInt_As_uint32_t(values[7]); if (unlikely((__pyx_v_n_secondary == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1555, __pyx_L3_error)
    } else {
      __pyx_v_n_secondary = ((uint32_t)5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("builder", 0, 1, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1553, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_method), (&PyUnicode_Type), 1, "method", 1))) __PYX_ERR(0, 1554, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_6Mesh2d_22builder(__pyx_v_filename, __pyx_v_n_threads, __pyx_v_cell_slowness, __pyx_v_method, __pyx_v_eps, __pyx_v_maxit, __pyx_v_process_obtuse, __pyx_v_n_secondary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_6Mesh2d_22builder(PyObject *__pyx_v_filename, size_t __pyx_v_n_threads, bool __pyx_v_cell_slowness, PyObject *__pyx_v_method, double __pyx_v_eps, int __pyx_v_maxit, bool __pyx_v_process_obtuse, uint32_t __pyx_v_n_secondary) {
  int __pyx_v_n;
  int __pyx_v_nn;
  PyObject *__pyx_v_reader = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_nod = NULL;
  PyObject *__pyx_v_tri = NULL;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_slowness = NULL;
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_m = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  long __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("builder", 0);

  /* "ttcrpy/tmesh.pyx":1578
 * 
 *         cdef int n, nn
 *         reader = vtk.vtkXMLUnstructuredGridReader()             # <<<<<<<<<<<<<<
 *         reader.SetFileName(filename)
 *         reader.Update()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_vtk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_vtkXMLUnstructuredGridReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_reader = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1579
 *         cdef int n, nn
 *         reader = vtk.vtkXMLUnstructuredGridReader()
 *         reader.SetFileName(filename)             # <<<<<<<<<<<<<<
 *         reader.Update()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_SetFileName); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1580
 *         reader = vtk.vtkXMLUnstructuredGridReader()
 *         reader.SetFileName(filename)
 *         reader.Update()             # <<<<<<<<<<<<<<
 * 
 *         data = reader.GetOutput()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_Update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1582
 *         reader.Update()
 * 
 *         data = reader.GetOutput()             # <<<<<<<<<<<<<<
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader, __pyx_n_s_GetOutput); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1583
 * 
 *         data = reader.GetOutput()
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())             # <<<<<<<<<<<<<<
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])
 *         for n in range(data.GetNumberOfCells()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPoints); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_GetData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1584
 *         data = reader.GetOutput()
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])             # <<<<<<<<<<<<<<
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_int_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_slice_);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_nodes, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_nod = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1585
 *         nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])
 *         for n in range(data.GetNumberOfCells()):             # <<<<<<<<<<<<<<
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:
 *                 raise ValueError('{0:s} should only contain triangles')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":1586
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:             # <<<<<<<<<<<<<<
 *                 raise ValueError('{0:s} should only contain triangles')
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_vtk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_VTK_TRIANGLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_10)) {

      /* "ttcrpy/tmesh.pyx":1587
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:
 *                 raise ValueError('{0:s} should only contain triangles')             # <<<<<<<<<<<<<<
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 *         for n in range(data.GetNumberOfCells()):
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1587, __pyx_L1_error)

      /* "ttcrpy/tmesh.pyx":1586
 *         nod = np.ascontiguousarray(nodes[:,[0, 2]])
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:             # <<<<<<<<<<<<<<
 *                 raise ValueError('{0:s} should only contain triangles')
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 */
    }
  }

  /* "ttcrpy/tmesh.pyx":1588
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:
 *                 raise ValueError('{0:s} should only contain triangles')
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))             # <<<<<<<<<<<<<<
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(3):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_ndarray), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tri = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1589
 *                 raise ValueError('{0:s} should only contain triangles')
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 *         for n in range(data.GetNumberOfCells()):             # <<<<<<<<<<<<<<
 *             for nn in range(3):
 *                 tri[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetNumberOfCells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_n = __pyx_t_9;

    /* "ttcrpy/tmesh.pyx":1590
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(3):             # <<<<<<<<<<<<<<
 *                 tri[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 * 
 */
    for (__pyx_t_11 = 0; __pyx_t_11 < 3; __pyx_t_11+=1) {
      __pyx_v_nn = __pyx_t_11;

      /* "ttcrpy/tmesh.pyx":1591
 *         for n in range(data.GetNumberOfCells()):
 *             for nn in range(3):
 *                 tri[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)             # <<<<<<<<<<<<<<
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCell); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_GetPointIds); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_GetId); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_nn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_5 = 0;
      __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tri, __pyx_t_1, __pyx_t_3) < 0)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "ttcrpy/tmesh.pyx":1593
 *                 tri[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',             # <<<<<<<<<<<<<<
 *                  'P-wave velocity')
 *         for name in names:
 */
  __Pyx_INCREF(__pyx_tuple__27);
  __pyx_v_names = __pyx_tuple__27;

  /* "ttcrpy/tmesh.pyx":1595
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_t_3 = __pyx_v_names; __Pyx_INCREF(__pyx_t_3); __pyx_t_12 = 0;
  for (;;) {
    if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1595, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ttcrpy/tmesh.pyx":1596
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "ttcrpy/tmesh.pyx":1597
 *         for name in names:
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 */
      __pyx_v_cell_slowness = 0;

      /* "ttcrpy/tmesh.pyx":1598
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 break
 *             if data.GetCellData().HasArray(name):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetPointData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1599
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break             # <<<<<<<<<<<<<<
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 */
      goto __pyx_L11_break;

      /* "ttcrpy/tmesh.pyx":1596
 *                  'P-wave velocity')
 *         for name in names:
 *             if data.GetPointData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 0
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 */
    }

    /* "ttcrpy/tmesh.pyx":1600
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_HasArray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "ttcrpy/tmesh.pyx":1601
 *                 break
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1             # <<<<<<<<<<<<<<
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 break
 */
      __pyx_v_cell_slowness = 1;

      /* "ttcrpy/tmesh.pyx":1602
 *             if data.GetCellData().HasArray(name):
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))             # <<<<<<<<<<<<<<
 *                 break
 *         else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_vtk_to_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_GetCellData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_6 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_GetArray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ttcrpy/tmesh.pyx":1603
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 *                 break             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')
 */
      goto __pyx_L11_break;

      /* "ttcrpy/tmesh.pyx":1600
 *                 data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
 *                 break
 *             if data.GetCellData().HasArray(name):             # <<<<<<<<<<<<<<
 *                 cell_slowness = 1
 *                 data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
 */
    }

    /* "ttcrpy/tmesh.pyx":1595
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  }
  /*else*/ {

    /* "ttcrpy/tmesh.pyx":1605
 *                 break
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')             # <<<<<<<<<<<<<<
 * 
 *         if 'lowness' in name:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1605, __pyx_L1_error)
  }

  /* "ttcrpy/tmesh.pyx":1595
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',
 *                  'P-wave velocity')
 *         for name in names:             # <<<<<<<<<<<<<<
 *             if data.GetPointData().HasArray(name):
 *                 cell_slowness = 0
 */
  __pyx_L11_break:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ttcrpy/tmesh.pyx":1607
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data
 *         else:
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1607, __pyx_L1_error)
  }
  __pyx_t_10 = (__Pyx_PyUnicode_ContainsTF(__pyx_n_u_lowness, __pyx_v_name, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __pyx_t_14 = (__pyx_t_10 != 0);
  if (__pyx_t_14) {

    /* "ttcrpy/tmesh.pyx":1608
 * 
 *         if 'lowness' in name:
 *             slowness = data             # <<<<<<<<<<<<<<
 *         else:
 *             slowness = 1.0 / data
 */
    __Pyx_INCREF(__pyx_v_data);
    __pyx_v_slowness = __pyx_v_data;

    /* "ttcrpy/tmesh.pyx":1607
 *             raise ValueError('File should contain slowness or velocity data')
 * 
 *         if 'lowness' in name:             # <<<<<<<<<<<<<<
 *             slowness = data
 *         else:
 */
    goto __pyx_L15;
  }

  /* "ttcrpy/tmesh.pyx":1610
 *             slowness = data
 *         else:
 *             slowness = 1.0 / data             # <<<<<<<<<<<<<<
 * 
 *         m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_0, __pyx_v_data, 1.0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_slowness = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L15:;

  /* "ttcrpy/tmesh.pyx":1612
 *             slowness = 1.0 / data
 * 
 *         m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,             # <<<<<<<<<<<<<<
 *                    process_obtuse, n_secondary)
 *         m.set_slowness(slowness)
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n_threads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_cell_slowness); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_eps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_maxit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "ttcrpy/tmesh.pyx":1613
 * 
 *         m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,
 *                    process_obtuse, n_secondary)             # <<<<<<<<<<<<<<
 *         m.set_slowness(slowness)
 *         return m
 */
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_process_obtuse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_uint32_t(__pyx_v_n_secondary); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "ttcrpy/tmesh.pyx":1612
 *             slowness = 1.0 / data
 * 
 *         m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,             # <<<<<<<<<<<<<<
 *                    process_obtuse, n_secondary)
 *         m.set_slowness(slowness)
 */
  __pyx_t_13 = PyTuple_New(9); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_nod);
  __Pyx_GIVEREF(__pyx_v_nod);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_nod);
  __Pyx_INCREF(__pyx_v_tri);
  __Pyx_GIVEREF(__pyx_v_tri);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_tri);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_13, 3, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_13, 4, __pyx_v_method);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_13, 5, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_13, 6, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_13, 7, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_13, 8, __pyx_t_6);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh2d), __pyx_t_13, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_m = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "ttcrpy/tmesh.pyx":1614
 *         m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,
 *                    process_obtuse, n_secondary)
 *         m.set_slowness(slowness)             # <<<<<<<<<<<<<<
 *         return m
 * 
 */
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_m), __pyx_n_s_set_slowness); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_13, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_4, __pyx_v_slowness) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_slowness);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ttcrpy/tmesh.pyx":1615
 *                    process_obtuse, n_secondary)
 *         m.set_slowness(slowness)
 *         return m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_m));
  __pyx_r = ((PyObject *)__pyx_v_m);
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1553
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM',double eps=1.e-15, int maxit=20,
 *                 bool process_obtuse=1,uint32_t n_secondary=5):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("ttcrpy.tmesh.Mesh2d.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_reader);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_nod);
  __Pyx_XDECREF(__pyx_v_tri);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_slowness);
  __Pyx_XDECREF((PyObject *)__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1618
 * 
 * 
 * def _rebuild3d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_3_rebuild3d(PyObject *__pyx_self, PyObject *__pyx_v_constructor_params); /*proto*/
static PyMethodDef __pyx_mdef_6ttcrpy_5tmesh_3_rebuild3d = {"_rebuild3d", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_3_rebuild3d, METH_O, 0};
static PyObject *__pyx_pw_6ttcrpy_5tmesh_3_rebuild3d(PyObject *__pyx_self, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rebuild3d (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_2_rebuild3d(__pyx_self, ((PyObject *)__pyx_v_constructor_params));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_2_rebuild3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_tetra = NULL;
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_cell_slowness = NULL;
  PyObject *__pyx_v_n_threads = NULL;
  PyObject *__pyx_v_tt_from_rp = NULL;
  PyObject *__pyx_v_interp_vel = NULL;
  PyObject *__pyx_v_eps = NULL;
  PyObject *__pyx_v_maxit = NULL;
  PyObject *__pyx_v_gradient_method = NULL;
  PyObject *__pyx_v_min_dist = NULL;
  PyObject *__pyx_v_n_secondary = NULL;
  PyObject *__pyx_v_n_tertiary = NULL;
  PyObject *__pyx_v_radius_tertiary = NULL;
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *__pyx_v_g = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rebuild3d", 0);

  /* "ttcrpy/tmesh.pyx":1621
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 *      radius_tertiary) = constructor_params             # <<<<<<<<<<<<<<
 * 
 *     g = Mesh3d(nodes, tetra, n_threads, cell_slowness, method, gradient_method,
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_constructor_params))) || (PyList_CheckExact(__pyx_v_constructor_params))) {
    PyObject* sequence = __pyx_v_constructor_params;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 14)) {
      if (size > 14) __Pyx_RaiseTooManyValuesError(14);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1619, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 9); 
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 10); 
      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 11); 
      __pyx_t_13 = PyTuple_GET_ITEM(sequence, 12); 
      __pyx_t_14 = PyTuple_GET_ITEM(sequence, 13); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 8); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 9); 
      __pyx_t_11 = PyList_GET_ITEM(sequence, 10); 
      __pyx_t_12 = PyList_GET_ITEM(sequence, 11); 
      __pyx_t_13 = PyList_GET_ITEM(sequence, 12); 
      __pyx_t_14 = PyList_GET_ITEM(sequence, 13); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_14);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[14] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10,&__pyx_t_11,&__pyx_t_12,&__pyx_t_13,&__pyx_t_14};
      for (i=0; i < 14; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 1619, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[14] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10,&__pyx_t_11,&__pyx_t_12,&__pyx_t_13,&__pyx_t_14};
    __pyx_t_15 = PyObject_GetIter(__pyx_v_constructor_params); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
    for (index=0; index < 14; index++) {
      PyObject* item = __pyx_t_16(__pyx_t_15); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 14) < 0) __PYX_ERR(0, 1619, __pyx_L1_error)
    __pyx_t_16 = NULL;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_16 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1619, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "ttcrpy/tmesh.pyx":1619
 * 
 * def _rebuild3d(constructor_params):
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,             # <<<<<<<<<<<<<<
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 *      radius_tertiary) = constructor_params
 */
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_tetra = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_method = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_cell_slowness = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_n_threads = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_tt_from_rp = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_interp_vel = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_eps = __pyx_t_8;
  __pyx_t_8 = 0;
  __pyx_v_maxit = __pyx_t_9;
  __pyx_t_9 = 0;
  __pyx_v_gradient_method = __pyx_t_10;
  __pyx_t_10 = 0;
  __pyx_v_min_dist = __pyx_t_11;
  __pyx_t_11 = 0;
  __pyx_v_n_secondary = __pyx_t_12;
  __pyx_t_12 = 0;
  __pyx_v_n_tertiary = __pyx_t_13;
  __pyx_t_13 = 0;
  __pyx_v_radius_tertiary = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "ttcrpy/tmesh.pyx":1623
 *      radius_tertiary) = constructor_params
 * 
 *     g = Mesh3d(nodes, tetra, n_threads, cell_slowness, method, gradient_method,             # <<<<<<<<<<<<<<
 *                tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                n_tertiary, radius_tertiary)
 */
  __pyx_t_14 = PyTuple_New(14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_nodes);
  __Pyx_INCREF(__pyx_v_tetra);
  __Pyx_GIVEREF(__pyx_v_tetra);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_tetra);
  __Pyx_INCREF(__pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_v_n_threads);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_v_n_threads);
  __Pyx_INCREF(__pyx_v_cell_slowness);
  __Pyx_GIVEREF(__pyx_v_cell_slowness);
  PyTuple_SET_ITEM(__pyx_t_14, 3, __pyx_v_cell_slowness);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_14, 4, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_gradient_method);
  __Pyx_GIVEREF(__pyx_v_gradient_method);
  PyTuple_SET_ITEM(__pyx_t_14, 5, __pyx_v_gradient_method);
  __Pyx_INCREF(__pyx_v_tt_from_rp);
  __Pyx_GIVEREF(__pyx_v_tt_from_rp);
  PyTuple_SET_ITEM(__pyx_t_14, 6, __pyx_v_tt_from_rp);
  __Pyx_INCREF(__pyx_v_interp_vel);
  __Pyx_GIVEREF(__pyx_v_interp_vel);
  PyTuple_SET_ITEM(__pyx_t_14, 7, __pyx_v_interp_vel);
  __Pyx_INCREF(__pyx_v_eps);
  __Pyx_GIVEREF(__pyx_v_eps);
  PyTuple_SET_ITEM(__pyx_t_14, 8, __pyx_v_eps);
  __Pyx_INCREF(__pyx_v_maxit);
  __Pyx_GIVEREF(__pyx_v_maxit);
  PyTuple_SET_ITEM(__pyx_t_14, 9, __pyx_v_maxit);
  __Pyx_INCREF(__pyx_v_min_dist);
  __Pyx_GIVEREF(__pyx_v_min_dist);
  PyTuple_SET_ITEM(__pyx_t_14, 10, __pyx_v_min_dist);
  __Pyx_INCREF(__pyx_v_n_secondary);
  __Pyx_GIVEREF(__pyx_v_n_secondary);
  PyTuple_SET_ITEM(__pyx_t_14, 11, __pyx_v_n_secondary);
  __Pyx_INCREF(__pyx_v_n_tertiary);
  __Pyx_GIVEREF(__pyx_v_n_tertiary);
  PyTuple_SET_ITEM(__pyx_t_14, 12, __pyx_v_n_tertiary);
  __Pyx_INCREF(__pyx_v_radius_tertiary);
  __Pyx_GIVEREF(__pyx_v_radius_tertiary);
  PyTuple_SET_ITEM(__pyx_t_14, 13, __pyx_v_radius_tertiary);
  __pyx_t_13 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh3d), __pyx_t_14, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_g = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)__pyx_t_13);
  __pyx_t_13 = 0;

  /* "ttcrpy/tmesh.pyx":1626
 *                tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
 *                n_tertiary, radius_tertiary)
 *     return g             # <<<<<<<<<<<<<<
 * 
 * def _rebuild2d(constructor_params):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_g));
  __pyx_r = ((PyObject *)__pyx_v_g);
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1618
 * 
 * 
 * def _rebuild3d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("ttcrpy.tmesh._rebuild3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_tetra);
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_cell_slowness);
  __Pyx_XDECREF(__pyx_v_n_threads);
  __Pyx_XDECREF(__pyx_v_tt_from_rp);
  __Pyx_XDECREF(__pyx_v_interp_vel);
  __Pyx_XDECREF(__pyx_v_eps);
  __Pyx_XDECREF(__pyx_v_maxit);
  __Pyx_XDECREF(__pyx_v_gradient_method);
  __Pyx_XDECREF(__pyx_v_min_dist);
  __Pyx_XDECREF(__pyx_v_n_secondary);
  __Pyx_XDECREF(__pyx_v_n_tertiary);
  __Pyx_XDECREF(__pyx_v_radius_tertiary);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ttcrpy/tmesh.pyx":1628
 *     return g
 * 
 * def _rebuild2d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
 *      process_obtuse, n_secondary) = constructor_params
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ttcrpy_5tmesh_5_rebuild2d(PyObject *__pyx_self, PyObject *__pyx_v_constructor_params); /*proto*/
static PyMethodDef __pyx_mdef_6ttcrpy_5tmesh_5_rebuild2d = {"_rebuild2d", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_5_rebuild2d, METH_O, 0};
static PyObject *__pyx_pw_6ttcrpy_5tmesh_5_rebuild2d(PyObject *__pyx_self, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rebuild2d (wrapper)", 0);
  __pyx_r = __pyx_pf_6ttcrpy_5tmesh_4_rebuild2d(__pyx_self, ((PyObject *)__pyx_v_constructor_params));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ttcrpy_5tmesh_4_rebuild2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_constructor_params) {
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_triangles = NULL;
  PyObject *__pyx_v_method = NULL;
  PyObject *__pyx_v_cell_slowness = NULL;
  PyObject *__pyx_v_n_threads = NULL;
  PyObject *__pyx_v_eps = NULL;
  PyObject *__pyx_v_maxit = NULL;
  PyObject *__pyx_v_process_obtuse = NULL;
  PyObject *__pyx_v_n_secondary = NULL;
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *__pyx_v_g = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rebuild2d", 0);

  /* "ttcrpy/tmesh.pyx":1630
 * def _rebuild2d(constructor_params):
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
 *      process_obtuse, n_secondary) = constructor_params             # <<<<<<<<<<<<<<
 * 
 *     g = Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit,
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_constructor_params))) || (PyList_CheckExact(__pyx_v_constructor_params))) {
    PyObject* sequence = __pyx_v_constructor_params;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1629, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 1629, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9};
    __pyx_t_10 = PyObject_GetIter(__pyx_v_constructor_params); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_11(__pyx_t_10); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 9) < 0) __PYX_ERR(0, 1629, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1629, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "ttcrpy/tmesh.pyx":1629
 * 
 * def _rebuild2d(constructor_params):
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,             # <<<<<<<<<<<<<<
 *      process_obtuse, n_secondary) = constructor_params
 * 
 */
  __pyx_v_nodes = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_triangles = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_method = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_cell_slowness = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_n_threads = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_eps = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_maxit = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_process_obtuse = __pyx_t_8;
  __pyx_t_8 = 0;
  __pyx_v_n_secondary = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "ttcrpy/tmesh.pyx":1632
 *      process_obtuse, n_secondary) = constructor_params
 * 
 *     g = Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit,             # <<<<<<<<<<<<<<
 *         process_obtuse, n_secondary)
 *     return g
 */
  __pyx_t_9 = PyTuple_New(9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_nodes);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_triangles);
  __Pyx_INCREF(__pyx_v_n_threads);
  __Pyx_GIVEREF(__pyx_v_n_threads);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_n_threads);
  __Pyx_INCREF(__pyx_v_cell_slowness);
  __Pyx_GIVEREF(__pyx_v_cell_slowness);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_v_cell_slowness);
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_eps);
  __Pyx_GIVEREF(__pyx_v_eps);
  PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_v_eps);
  __Pyx_INCREF(__pyx_v_maxit);
  __Pyx_GIVEREF(__pyx_v_maxit);
  PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_v_maxit);
  __Pyx_INCREF(__pyx_v_process_obtuse);
  __Pyx_GIVEREF(__pyx_v_process_obtuse);
  PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_v_process_obtuse);
  __Pyx_INCREF(__pyx_v_n_secondary);
  __Pyx_GIVEREF(__pyx_v_n_secondary);
  PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_v_n_secondary);
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh2d), __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_g = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "ttcrpy/tmesh.pyx":1634
 *     g = Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit,
 *         process_obtuse, n_secondary)
 *     return g             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_g));
  __pyx_r = ((PyObject *)__pyx_v_g);
  goto __pyx_L0;

  /* "ttcrpy/tmesh.pyx":1628
 *     return g
 * 
 * def _rebuild2d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
 *      process_obtuse, n_secondary) = constructor_params
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ttcrpy.tmesh._rebuild2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_triangles);
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_cell_slowness);
  __Pyx_XDECREF(__pyx_v_n_threads);
  __Pyx_XDECREF(__pyx_v_eps);
  __Pyx_XDECREF(__pyx_v_maxit);
  __Pyx_XDECREF(__pyx_v_process_obtuse);
  __Pyx_XDECREF(__pyx_v_n_secondary);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":869
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":870
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":873
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":875
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":876
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":882
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 882, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":883
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 883, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 884, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 884, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":888
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 888, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":889
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 889, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 890, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 890, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":894
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 894, __pyx_L3_error)

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":895
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 895, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":896
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 896, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 896, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6ttcrpy_5tmesh_Mesh3d(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)o);
  new((void*)&(p->no)) std::vector<ttcr::sxyz<double> > ();
  new((void*)&(p->tet)) std::vector<ttcr::tetrahedronElem<uint32_t> > ();
  if (unlikely(__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6ttcrpy_5tmesh_Mesh3d(PyObject *o) {
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *p = (struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->no);
  __Pyx_call_destructor(p->tet);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5tmesh_6Mesh3d_n_threads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9n_threads_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5tmesh_6Mesh3d_nparams(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7nparams_1__get__(o);
}

static PyMethodDef __pyx_methods_6ttcrpy_5tmesh_Mesh3d[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_5__reduce__, METH_NOARGS, 0},
  {"set_use_thread_pool", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_7set_use_thread_pool, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_6set_use_thread_pool},
  {"is_outside", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_9is_outside, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_8is_outside},
  {"get_number_of_nodes", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_11get_number_of_nodes, METH_NOARGS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_10get_number_of_nodes},
  {"get_number_of_cells", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_13get_number_of_cells, METH_NOARGS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_12get_number_of_cells},
  {"get_grid_traveltimes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_15get_grid_traveltimes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_14get_grid_traveltimes},
  {"set_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_17set_slowness, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_16set_slowness},
  {"set_velocity", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_19set_velocity, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_18set_velocity},
  {"compute_D", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_21compute_D, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_20compute_D},
  {"compute_K", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_23compute_K, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_22compute_K},
  {"get_s0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_25get_s0, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_24get_s0},
  {"raytrace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_27raytrace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_26raytrace},
  {"to_vtk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_29to_vtk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_28to_vtk},
  {"_save_raypaths", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_31_save_raypaths, METH_VARARGS|METH_KEYWORDS, 0},
  {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh3d_33builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh3d_32builder},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6ttcrpy_5tmesh_Mesh3d[] = {
  {(char *)"n_threads", __pyx_getprop_6ttcrpy_5tmesh_6Mesh3d_n_threads, 0, (char *)"int: number of threads for raytracing", 0},
  {(char *)"nparams", __pyx_getprop_6ttcrpy_5tmesh_6Mesh3d_nparams, 0, (char *)"int: total number of parameters for mesh", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ttcrpy_5tmesh_Mesh3d = {
  PyVarObject_HEAD_INIT(0, 0)
  "ttcrpy.tmesh.Mesh3d", /*tp_name*/
  sizeof(struct __pyx_obj_6ttcrpy_5tmesh_Mesh3d), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ttcrpy_5tmesh_Mesh3d, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "class to perform raytracing with tetrahedral meshes\n\n    Attributes\n    ----------\n    nparams: int\n        total number of parameters for grid\n    n_threads: int\n        number of threads for raytracing\n\n    Constructor:\n\n    Mesh3d(nodes, tetra, n_threads, cell_slowness, method, gradient_method, tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary, n_tertiary, radius_tertiary) -> Mesh3d\n\n        Parameters\n        ----------\n        nodes : np.ndarray, shape (nnodes, 3)\n            node coordinates\n        tetra : np.ndarray of int, shape (ntetra, 4)\n            indices of nodes forming the tetrahedra\n        n_threads : int\n            number of threads for raytracing (default is 1)\n        cell_slowness : bool\n            slowness defined for cells (True) or nodes (False) (default is 1)\n        method : string\n            raytracing method (default is FSM)\n                - 'FSM' : fast marching method\n                - 'SPM' : shortest path method\n                - 'DSPM' : dynamic shortest path\n        gradient_method : int\n            method to compute traveltime gradient (default is 1)\n                - 0 : least-squares first-order\n                - 1 : least-squares second-order\n                - 2 : Averaging-Based method\n        tt_from_rp : bool\n            compute traveltimes from raypaths (SPM od DSPM) (default is 1)\n        interp_vel : bool\n            interpolate velocity instead of slowness at nodes (for\n            cell_slowness == False or FSM) (defauls is False)\n        eps : double\n            convergence criterion (FSM) (default is 1e-15)\n        maxit : int\n            max number of sweeping iterations (FSM) (default is 20)\n        min_dist : double\n            tolerance for backward raytracing (default is 1e-5)\n        n_secondary : int\n            number of secondary nodes (SPM & DSPM) (default is 2)\n        n_tertiary : int\n            number of tertiary nodes (DSPM) (default is 2)\n    ""    radius_tertiary : double\n            radius of sphere around source that includes tertiary nodes (DSPM)\n            (default is 1)\n\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ttcrpy_5tmesh_Mesh3d, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6ttcrpy_5tmesh_Mesh3d, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ttcrpy_5tmesh_Mesh3d, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6ttcrpy_5tmesh_Mesh2d(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)o);
  new((void*)&(p->no)) std::vector<ttcr::sxz<double> > ();
  new((void*)&(p->tri)) std::vector<ttcr::triangleElem<uint32_t> > ();
  if (unlikely(__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6ttcrpy_5tmesh_Mesh2d(PyObject *o) {
  struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *p = (struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->no);
  __Pyx_call_destructor(p->tri);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5tmesh_6Mesh2d_n_threads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9n_threads_1__get__(o);
}

static PyObject *__pyx_getprop_6ttcrpy_5tmesh_6Mesh2d_nparams(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7nparams_1__get__(o);
}

static PyMethodDef __pyx_methods_6ttcrpy_5tmesh_Mesh2d[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_5__reduce__, METH_NOARGS, 0},
  {"get_number_of_nodes", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_7get_number_of_nodes, METH_NOARGS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_6get_number_of_nodes},
  {"get_number_of_cells", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_9get_number_of_cells, METH_NOARGS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_8get_number_of_cells},
  {"get_grid_traveltimes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_11get_grid_traveltimes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_10get_grid_traveltimes},
  {"set_slowness", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_13set_slowness, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_12set_slowness},
  {"set_velocity", (PyCFunction)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_15set_velocity, METH_O, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_14set_velocity},
  {"raytrace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_17raytrace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_16raytrace},
  {"to_vtk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_19to_vtk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_18to_vtk},
  {"_save_raypaths", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_21_save_raypaths, METH_VARARGS|METH_KEYWORDS, 0},
  {"builder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ttcrpy_5tmesh_6Mesh2d_23builder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ttcrpy_5tmesh_6Mesh2d_22builder},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6ttcrpy_5tmesh_Mesh2d[] = {
  {(char *)"n_threads", __pyx_getprop_6ttcrpy_5tmesh_6Mesh2d_n_threads, 0, (char *)"int: number of threads for raytracing", 0},
  {(char *)"nparams", __pyx_getprop_6ttcrpy_5tmesh_6Mesh2d_nparams, 0, (char *)"int: total number of parameters for mesh", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ttcrpy_5tmesh_Mesh2d = {
  PyVarObject_HEAD_INIT(0, 0)
  "ttcrpy.tmesh.Mesh2d", /*tp_name*/
  sizeof(struct __pyx_obj_6ttcrpy_5tmesh_Mesh2d), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ttcrpy_5tmesh_Mesh2d, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "class to perform raytracing with triangular meshes\n\n    Attributes\n    ----------\n    nparams: int\n        total number of parameters for grid\n    n_threads: int\n        number of threads for raytracing\n\n    Constructor:\n\n    Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit, process_obtuse, n_secondary) -> Mesh2d\n\n        Parameters\n        ----------\n        nodes : np.ndarray, shape (nnodes, 2)\n            node coordinates\n        triangles : np.ndarray of int, shape (ntriangles, 3)\n            indices of nodes forming the triangles\n        n_threads : int\n            number of threads for raytracing (default is 1)\n        cell_slowness : bool\n            slowness defined for cells (True) or nodes (False) (default is 1)\n        method : string\n            raytracing method (default is FSM)\n                - 'FSM' : fast marching method\n                - 'SPM' : shortest path method\n                - 'DSPM' : dynamic shortest path\n        eps : double\n            convergence criterion (FSM) (default is 1e-15)\n        maxit : int\n            max number of sweeping iterations (FSM) (default is 20)\n        process_obtuse : bool\n            use method of Qian et al (2007) to improve accuracy for triangles\n            with obtuse angle\n        n_secondary : int\n            number of secondary nodes (SPM) (default is 5)\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ttcrpy_5tmesh_Mesh2d, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6ttcrpy_5tmesh_Mesh2d, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ttcrpy_5tmesh_Mesh2d, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_tmesh(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_tmesh},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "tmesh",
    __pyx_k_Raytracing_on_unstructured_tria, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_0_s_should_only_contain_tetrahe, __pyx_k_0_s_should_only_contain_tetrahe, sizeof(__pyx_k_0_s_should_only_contain_tetrahe), 0, 1, 0, 0},
  {&__pyx_kp_u_0_s_should_only_contain_triangl, __pyx_k_0_s_should_only_contain_triangl, sizeof(__pyx_k_0_s_should_only_contain_triangl), 0, 1, 0, 0},
  {&__pyx_n_s_AddArray, __pyx_k_AddArray, sizeof(__pyx_k_AddArray), 0, 0, 1, 1},
  {&__pyx_n_u_C_CONTIGUOUS, __pyx_k_C_CONTIGUOUS, sizeof(__pyx_k_C_CONTIGUOUS), 0, 1, 0, 1},
  {&__pyx_kp_u_Cannot_aggregate_source_with_DSP, __pyx_k_Cannot_aggregate_source_with_DSP, sizeof(__pyx_k_Cannot_aggregate_source_with_DSP), 0, 1, 0, 0},
  {&__pyx_n_u_DSPM, __pyx_k_DSPM, sizeof(__pyx_k_DSPM), 0, 1, 0, 1},
  {&__pyx_n_u_FSM, __pyx_k_FSM, sizeof(__pyx_k_FSM), 0, 1, 0, 1},
  {&__pyx_kp_u_Field_0_s_has_incorrect_size, __pyx_k_Field_0_s_has_incorrect_size, sizeof(__pyx_k_Field_0_s_has_incorrect_size), 0, 1, 0, 0},
  {&__pyx_kp_u_File_should_contain_slowness_or, __pyx_k_File_should_contain_slowness_or, sizeof(__pyx_k_File_should_contain_slowness_or), 0, 1, 0, 0},
  {&__pyx_n_s_GetArray, __pyx_k_GetArray, sizeof(__pyx_k_GetArray), 0, 0, 1, 1},
  {&__pyx_n_s_GetCell, __pyx_k_GetCell, sizeof(__pyx_k_GetCell), 0, 0, 1, 1},
  {&__pyx_n_s_GetCellData, __pyx_k_GetCellData, sizeof(__pyx_k_GetCellData), 0, 0, 1, 1},
  {&__pyx_n_s_GetCellType, __pyx_k_GetCellType, sizeof(__pyx_k_GetCellType), 0, 0, 1, 1},
  {&__pyx_n_s_GetData, __pyx_k_GetData, sizeof(__pyx_k_GetData), 0, 0, 1, 1},
  {&__pyx_n_s_GetId, __pyx_k_GetId, sizeof(__pyx_k_GetId), 0, 0, 1, 1},
  {&__pyx_n_s_GetNumberOfCells, __pyx_k_GetNumberOfCells, sizeof(__pyx_k_GetNumberOfCells), 0, 0, 1, 1},
  {&__pyx_n_s_GetOutput, __pyx_k_GetOutput, sizeof(__pyx_k_GetOutput), 0, 0, 1, 1},
  {&__pyx_n_s_GetPointData, __pyx_k_GetPointData, sizeof(__pyx_k_GetPointData), 0, 0, 1, 1},
  {&__pyx_n_s_GetPointIds, __pyx_k_GetPointIds, sizeof(__pyx_k_GetPointIds), 0, 0, 1, 1},
  {&__pyx_n_s_GetPoints, __pyx_k_GetPoints, sizeof(__pyx_k_GetPoints), 0, 0, 1, 1},
  {&__pyx_n_s_HasArray, __pyx_k_HasArray, sizeof(__pyx_k_HasArray), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_InsertNextCell, __pyx_k_InsertNextCell, sizeof(__pyx_k_InsertNextCell), 0, 0, 1, 1},
  {&__pyx_n_s_InsertPoint, __pyx_k_InsertPoint, sizeof(__pyx_k_InsertPoint), 0, 0, 1, 1},
  {&__pyx_n_s_Mesh2d, __pyx_k_Mesh2d, sizeof(__pyx_k_Mesh2d), 0, 0, 1, 1},
  {&__pyx_n_s_Mesh3d, __pyx_k_Mesh3d, sizeof(__pyx_k_Mesh3d), 0, 0, 1, 1},
  {&__pyx_kp_u_Method_0_s_undefined, __pyx_k_Method_0_s_undefined, sizeof(__pyx_k_Method_0_s_undefined), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_kp_u_P_wave_velocity, __pyx_k_P_wave_velocity, sizeof(__pyx_k_P_wave_velocity), 0, 1, 0, 0},
  {&__pyx_kp_u_Receiver_outside_grid, __pyx_k_Receiver_outside_grid, sizeof(__pyx_k_Receiver_outside_grid), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_u_SPM, __pyx_k_SPM, sizeof(__pyx_k_SPM), 0, 1, 0, 1},
  {&__pyx_n_s_SetDataModeToBinary, __pyx_k_SetDataModeToBinary, sizeof(__pyx_k_SetDataModeToBinary), 0, 0, 1, 1},
  {&__pyx_n_s_SetFileName, __pyx_k_SetFileName, sizeof(__pyx_k_SetFileName), 0, 0, 1, 1},
  {&__pyx_n_s_SetId, __pyx_k_SetId, sizeof(__pyx_k_SetId), 0, 0, 1, 1},
  {&__pyx_n_s_SetInputData, __pyx_k_SetInputData, sizeof(__pyx_k_SetInputData), 0, 0, 1, 1},
  {&__pyx_n_s_SetLines, __pyx_k_SetLines, sizeof(__pyx_k_SetLines), 0, 0, 1, 1},
  {&__pyx_n_s_SetName, __pyx_k_SetName, sizeof(__pyx_k_SetName), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfComponents, __pyx_k_SetNumberOfComponents, sizeof(__pyx_k_SetNumberOfComponents), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfIds, __pyx_k_SetNumberOfIds, sizeof(__pyx_k_SetNumberOfIds), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfPoints, __pyx_k_SetNumberOfPoints, sizeof(__pyx_k_SetNumberOfPoints), 0, 0, 1, 1},
  {&__pyx_n_s_SetNumberOfTuples, __pyx_k_SetNumberOfTuples, sizeof(__pyx_k_SetNumberOfTuples), 0, 0, 1, 1},
  {&__pyx_n_s_SetPoints, __pyx_k_SetPoints, sizeof(__pyx_k_SetPoints), 0, 0, 1, 1},
  {&__pyx_n_s_SetTuple1, __pyx_k_SetTuple1, sizeof(__pyx_k_SetTuple1), 0, 0, 1, 1},
  {&__pyx_n_u_Slowness, __pyx_k_Slowness, sizeof(__pyx_k_Slowness), 0, 1, 0, 1},
  {&__pyx_kp_u_Slowness_vector_has_wrong_size, __pyx_k_Slowness_vector_has_wrong_size, sizeof(__pyx_k_Slowness_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_kp_u_Source_point_outside_grid, __pyx_k_Source_point_outside_grid, sizeof(__pyx_k_Source_point_outside_grid), 0, 1, 0, 0},
  {&__pyx_kp_u_Thread_number_is_larger_than_num, __pyx_k_Thread_number_is_larger_than_num, sizeof(__pyx_k_Thread_number_is_larger_than_num), 0, 1, 0, 0},
  {&__pyx_n_s_Update, __pyx_k_Update, sizeof(__pyx_k_Update), 0, 0, 1, 1},
  {&__pyx_n_s_VTK_TETRA, __pyx_k_VTK_TETRA, sizeof(__pyx_k_VTK_TETRA), 0, 0, 1, 1},
  {&__pyx_n_s_VTK_TRIANGLE, __pyx_k_VTK_TRIANGLE, sizeof(__pyx_k_VTK_TRIANGLE), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_u_Velocity, __pyx_k_Velocity, sizeof(__pyx_k_Velocity), 0, 1, 0, 1},
  {&__pyx_kp_u_Velocity_data_point_outside_grid, __pyx_k_Velocity_data_point_outside_grid, sizeof(__pyx_k_Velocity_data_point_outside_grid), 0, 1, 0, 0},
  {&__pyx_n_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 1},
  {&__pyx_n_s__36, __pyx_k__36, sizeof(__pyx_k__36), 0, 0, 1, 1},
  {&__pyx_n_s_aggregate_src, __pyx_k_aggregate_src, sizeof(__pyx_k_aggregate_src), 0, 0, 1, 1},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_ascontiguousarray, __pyx_k_ascontiguousarray, sizeof(__pyx_k_ascontiguousarray), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_builder, __pyx_k_builder, sizeof(__pyx_k_builder), 0, 0, 1, 1},
  {&__pyx_n_s_cell_slowness, __pyx_k_cell_slowness, sizeof(__pyx_k_cell_slowness), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_compute_L, __pyx_k_compute_L, sizeof(__pyx_k_compute_L), 0, 0, 1, 1},
  {&__pyx_kp_u_compute_L_not_implemented_for_me, __pyx_k_compute_L_not_implemented_for_me, sizeof(__pyx_k_compute_L_not_implemented_for_me), 0, 1, 0, 0},
  {&__pyx_n_s_constructor_params, __pyx_k_constructor_params, sizeof(__pyx_k_constructor_params), 0, 0, 1, 1},
  {&__pyx_n_s_csr_matrix, __pyx_k_csr_matrix, sizeof(__pyx_k_csr_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_eps, __pyx_k_eps, sizeof(__pyx_k_eps), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_get_number_of_cells, __pyx_k_get_number_of_cells, sizeof(__pyx_k_get_number_of_cells), 0, 0, 1, 1},
  {&__pyx_n_s_get_number_of_nodes, __pyx_k_get_number_of_nodes, sizeof(__pyx_k_get_number_of_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_gradient_method, __pyx_k_gradient_method, sizeof(__pyx_k_gradient_method), 0, 0, 1, 1},
  {&__pyx_n_s_hypo, __pyx_k_hypo, sizeof(__pyx_k_hypo), 0, 0, 1, 1},
  {&__pyx_kp_u_hypo_should_be_npts_x_5, __pyx_k_hypo_should_be_npts_x_5, sizeof(__pyx_k_hypo_should_be_npts_x_5), 0, 1, 0, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_interp_vel, __pyx_k_interp_vel, sizeof(__pyx_k_interp_vel), 0, 0, 1, 1},
  {&__pyx_n_s_is_outside, __pyx_k_is_outside, sizeof(__pyx_k_is_outside), 0, 0, 1, 1},
  {&__pyx_n_u_lowness, __pyx_k_lowness, sizeof(__pyx_k_lowness), 0, 1, 0, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_maxit, __pyx_k_maxit, sizeof(__pyx_k_maxit), 0, 0, 1, 1},
  {&__pyx_n_s_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_min_dist, __pyx_k_min_dist, sizeof(__pyx_k_min_dist), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_n_secondary, __pyx_k_n_secondary, sizeof(__pyx_k_n_secondary), 0, 0, 1, 1},
  {&__pyx_n_s_n_tertiary, __pyx_k_n_tertiary, sizeof(__pyx_k_n_tertiary), 0, 0, 1, 1},
  {&__pyx_n_s_n_threads, __pyx_k_n_threads, sizeof(__pyx_k_n_threads), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_nn, __pyx_k_nn, sizeof(__pyx_k_nn), 0, 0, 1, 1},
  {&__pyx_n_s_nod, __pyx_k_nod, sizeof(__pyx_k_nod), 0, 0, 1, 1},
  {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_nonzero, __pyx_k_nonzero, sizeof(__pyx_k_nonzero), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nparams, __pyx_k_nparams, sizeof(__pyx_k_nparams), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_numpy_support, __pyx_k_numpy_support, sizeof(__pyx_k_numpy_support), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_n_s_process_obtuse, __pyx_k_process_obtuse, sizeof(__pyx_k_process_obtuse), 0, 0, 1, 1},
  {&__pyx_n_s_radius_tertiary, __pyx_k_radius_tertiary, sizeof(__pyx_k_radius_tertiary), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rays, __pyx_k_rays, sizeof(__pyx_k_rays), 0, 0, 1, 1},
  {&__pyx_n_s_rcv, __pyx_k_rcv, sizeof(__pyx_k_rcv), 0, 0, 1, 1},
  {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
  {&__pyx_n_s_rebuild2d, __pyx_k_rebuild2d, sizeof(__pyx_k_rebuild2d), 0, 0, 1, 1},
  {&__pyx_n_s_rebuild3d, __pyx_k_rebuild3d, sizeof(__pyx_k_rebuild3d), 0, 0, 1, 1},
  {&__pyx_n_s_return_rays, __pyx_k_return_rays, sizeof(__pyx_k_return_rays), 0, 0, 1, 1},
  {&__pyx_n_s_s0inside, __pyx_k_s0inside, sizeof(__pyx_k_s0inside), 0, 0, 1, 1},
  {&__pyx_n_s_save_raypaths, __pyx_k_save_raypaths, sizeof(__pyx_k_save_raypaths), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_sparse, __pyx_k_scipy_sparse, sizeof(__pyx_k_scipy_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_set_slowness, __pyx_k_set_slowness, sizeof(__pyx_k_set_slowness), 0, 0, 1, 1},
  {&__pyx_n_s_set_verbose, __pyx_k_set_verbose, sizeof(__pyx_k_set_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_slowness, __pyx_k_slowness, sizeof(__pyx_k_slowness), 0, 0, 1, 1},
  {&__pyx_n_u_slowness, __pyx_k_slowness, sizeof(__pyx_k_slowness), 0, 1, 0, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_kp_u_source_and_rcv_should_be_2D_arra, __pyx_k_source_and_rcv_should_be_2D_arra, sizeof(__pyx_k_source_and_rcv_should_be_2D_arra), 0, 1, 0, 0},
  {&__pyx_kp_u_source_should_be_either_nsrc_x_2, __pyx_k_source_should_be_either_nsrc_x_2, sizeof(__pyx_k_source_should_be_either_nsrc_x_2), 0, 1, 0, 0},
  {&__pyx_kp_u_source_should_be_either_nsrc_x_3, __pyx_k_source_should_be_either_nsrc_x_3, sizeof(__pyx_k_source_should_be_either_nsrc_x_3), 0, 1, 0, 0},
  {&__pyx_n_s_sp, __pyx_k_sp, sizeof(__pyx_k_sp), 0, 0, 1, 1},
  {&__pyx_n_s_squared, __pyx_k_squared, sizeof(__pyx_k_squared), 0, 0, 1, 1},
  {&__pyx_kp_u_src_and_rcv_should_be_ndata_x_2, __pyx_k_src_and_rcv_should_be_ndata_x_2, sizeof(__pyx_k_src_and_rcv_should_be_ndata_x_2), 0, 1, 0, 0},
  {&__pyx_kp_u_src_and_rcv_should_be_ndata_x_3, __pyx_k_src_and_rcv_should_be_ndata_x_3, sizeof(__pyx_k_src_and_rcv_should_be_ndata_x_3), 0, 1, 0, 0},
  {&__pyx_kp_u_src_and_rcv_should_be_of_equal_s, __pyx_k_src_and_rcv_should_be_of_equal_s, sizeof(__pyx_k_src_and_rcv_should_be_of_equal_s), 0, 1, 0, 0},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_taylor_order, __pyx_k_taylor_order, sizeof(__pyx_k_taylor_order), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tet, __pyx_k_tet, sizeof(__pyx_k_tet), 0, 0, 1, 1},
  {&__pyx_n_s_tetra, __pyx_k_tetra, sizeof(__pyx_k_tetra), 0, 0, 1, 1},
  {&__pyx_n_s_thread_no, __pyx_k_thread_no, sizeof(__pyx_k_thread_no), 0, 0, 1, 1},
  {&__pyx_n_s_tri, __pyx_k_tri, sizeof(__pyx_k_tri), 0, 0, 1, 1},
  {&__pyx_n_s_triangles, __pyx_k_triangles, sizeof(__pyx_k_triangles), 0, 0, 1, 1},
  {&__pyx_n_s_tt_from_rp, __pyx_k_tt_from_rp, sizeof(__pyx_k_tt_from_rp), 0, 0, 1, 1},
  {&__pyx_n_s_ttcrpy_tmesh, __pyx_k_ttcrpy_tmesh, sizeof(__pyx_k_ttcrpy_tmesh), 0, 0, 1, 1},
  {&__pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_k_ttcrpy_tmesh_pyx, sizeof(__pyx_k_ttcrpy_tmesh_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_unique, __pyx_k_unique, sizeof(__pyx_k_unique), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_u_velocity, __pyx_k_velocity, sizeof(__pyx_k_velocity), 0, 1, 0, 1},
  {&__pyx_kp_u_velocity_vector_has_wrong_size, __pyx_k_velocity_vector_has_wrong_size, sizeof(__pyx_k_velocity_vector_has_wrong_size), 0, 1, 0, 0},
  {&__pyx_n_s_vtk, __pyx_k_vtk, sizeof(__pyx_k_vtk), 0, 0, 1, 1},
  {&__pyx_n_s_vtkCellArray, __pyx_k_vtkCellArray, sizeof(__pyx_k_vtkCellArray), 0, 0, 1, 1},
  {&__pyx_n_s_vtkDoubleArray, __pyx_k_vtkDoubleArray, sizeof(__pyx_k_vtkDoubleArray), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPoints, __pyx_k_vtkPoints, sizeof(__pyx_k_vtkPoints), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPolyData, __pyx_k_vtkPolyData, sizeof(__pyx_k_vtkPolyData), 0, 0, 1, 1},
  {&__pyx_n_s_vtkPolyLine, __pyx_k_vtkPolyLine, sizeof(__pyx_k_vtkPolyLine), 0, 0, 1, 1},
  {&__pyx_n_s_vtkTetra, __pyx_k_vtkTetra, sizeof(__pyx_k_vtkTetra), 0, 0, 1, 1},
  {&__pyx_n_s_vtkTriangle, __pyx_k_vtkTriangle, sizeof(__pyx_k_vtkTriangle), 0, 0, 1, 1},
  {&__pyx_n_s_vtkUnstructuredGrid, __pyx_k_vtkUnstructuredGrid, sizeof(__pyx_k_vtkUnstructuredGrid), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLPolyDataWriter, __pyx_k_vtkXMLPolyDataWriter, sizeof(__pyx_k_vtkXMLPolyDataWriter), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLUnstructuredGridReader, __pyx_k_vtkXMLUnstructuredGridReader, sizeof(__pyx_k_vtkXMLUnstructuredGridReader), 0, 0, 1, 1},
  {&__pyx_n_s_vtkXMLUnstructuredGridWriter, __pyx_k_vtkXMLUnstructuredGridWriter, sizeof(__pyx_k_vtkXMLUnstructuredGridWriter), 0, 0, 1, 1},
  {&__pyx_n_s_vtk_to_numpy, __pyx_k_vtk_to_numpy, sizeof(__pyx_k_vtk_to_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_vtk_util, __pyx_k_vtk_util, sizeof(__pyx_k_vtk_util), 0, 0, 1, 1},
  {&__pyx_kp_u_vtp, __pyx_k_vtp, sizeof(__pyx_k_vtp), 0, 1, 0, 0},
  {&__pyx_kp_u_vtu, __pyx_k_vtu, sizeof(__pyx_k_vtu), 0, 1, 0, 0},
  {&__pyx_n_s_weighting, __pyx_k_weighting, sizeof(__pyx_k_weighting), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 957, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 206, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 307, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 684, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 884, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ttcrpy/tmesh.pyx":163
 * 
 *         if method == 'FSM':
 *             xmin = np.min(nodes[:, 0])             # <<<<<<<<<<<<<<
 *             xmax = np.max(nodes[:, 0])
 *             ymin = np.min(nodes[:, 1])
 */
  __pyx_slice_ = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_slice_, __pyx_int_0); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "ttcrpy/tmesh.pyx":165
 *             xmin = np.min(nodes[:, 0])
 *             xmax = np.max(nodes[:, 0])
 *             ymin = np.min(nodes[:, 1])             # <<<<<<<<<<<<<<
 *             ymax = np.max(nodes[:, 1])
 *             zmin = np.min(nodes[:, 2])
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_slice_, __pyx_int_1); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "ttcrpy/tmesh.pyx":167
 *             ymin = np.min(nodes[:, 1])
 *             ymax = np.max(nodes[:, 1])
 *             zmin = np.min(nodes[:, 2])             # <<<<<<<<<<<<<<
 *             zmax = np.max(nodes[:, 2])
 *             pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
 */
  __pyx_tuple__4 = PyTuple_Pack(2, __pyx_slice_, __pyx_int_2); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "ttcrpy/tmesh.pyx":347
 *         """
 *         if thread_no >= self._n_threads:
 *             raise ValueError('Thread number is larger than number of threads')             # <<<<<<<<<<<<<<
 *         cdef vector[double] tmp
 *         cdef int n
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_Thread_number_is_larger_than_num); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "ttcrpy/tmesh.pyx":367
 *         """
 *         if slowness.size != self.nparams:
 *             raise ValueError('Slowness vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not slowness.flags['C_CONTIGUOUS']:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_Slowness_vector_has_wrong_size); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "ttcrpy/tmesh.pyx":389
 *         """
 *         if velocity.size != self.nparams:
 *             raise ValueError('velocity vector has wrong size')             # <<<<<<<<<<<<<<
 * 
 *         if not velocity.flags['C_CONTIGUOUS']:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_velocity_vector_has_wrong_size); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "ttcrpy/tmesh.pyx":418
 *         """
 *         if self.is_outside(coord):
 *             raise ValueError('Velocity data point outside grid')             # <<<<<<<<<<<<<<
 *         cdef vector[sxyz[double]] vpts
 *         for n in range(coord.shape[0]):
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Velocity_data_point_outside_grid); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "ttcrpy/tmesh.pyx":525
 * 
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')             # <<<<<<<<<<<<<<
 *         src = hypo[:,2:5]
 *         evID = hypo[:,0]
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_hypo_should_be_npts_x_5); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "ttcrpy/tmesh.pyx":526
 *         if hypo.shape[1] != 5:
 *             raise ValueError('hypo should be npts x 5')
 *         src = hypo[:,2:5]             # <<<<<<<<<<<<<<
 *         evID = hypo[:,0]
 *         eid = np.sort(np.unique(evID))
 */
  __pyx_slice__10 = PySlice_New(__pyx_int_2, __pyx_int_5, Py_None); if (unlikely(!__pyx_slice__10)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__10);
  __Pyx_GIVEREF(__pyx_slice__10);
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_slice_, __pyx_slice__10); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "ttcrpy/tmesh.pyx":629
 * 
 *         if source.ndim != 2 or rcv.ndim != 2:
 *             raise ValueError('source and rcv should be 2D arrays')             # <<<<<<<<<<<<<<
 * 
 *         if self.method == b'd' and aggregate_src:
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_source_and_rcv_should_be_2D_arra); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "ttcrpy/tmesh.pyx":632
 * 
 *         if self.method == b'd' and aggregate_src:
 *             raise ValueError('Cannot aggregate source with DSPM raytracing')             # <<<<<<<<<<<<<<
 * 
 *         evID = None
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Cannot_aggregate_source_with_DSP); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "ttcrpy/tmesh.pyx":647
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 4:
 *             src = source[:,1:4]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
  __pyx_slice__14 = PySlice_New(__pyx_int_1, __pyx_int_4, Py_None); if (unlikely(!__pyx_slice__14)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__14);
  __Pyx_GIVEREF(__pyx_slice__14);
  __pyx_tuple__15 = PyTuple_Pack(2, __pyx_slice_, __pyx_slice__14); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "ttcrpy/tmesh.pyx":653
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 3, 4 or 5')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_source_should_be_either_nsrc_x_3); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "ttcrpy/tmesh.pyx":656
 * 
 *         if src.shape[1] != 3 or rcv.shape[1] != 3:
 *             raise ValueError('src and rcv should be ndata x 3')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(src):
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_ndata_x_3); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "ttcrpy/tmesh.pyx":659
 * 
 *         if self.is_outside(src):
 *             raise ValueError('Source point outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if self.is_outside(rcv):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_Source_point_outside_grid); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "ttcrpy/tmesh.pyx":662
 * 
 *         if self.is_outside(rcv):
 *             raise ValueError('Receiver outside grid')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_Receiver_outside_grid); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "ttcrpy/tmesh.pyx":684
 *         vtt.resize(nTx)
 *         if compute_L and self.cell_slowness:
 *             raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')             # <<<<<<<<<<<<<<
 *         elif compute_L and not self.cell_slowness:
 *             m_data.resize(nTx)
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_compute_L_not_implemented_for_me); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "ttcrpy/tmesh.pyx":694
 *         if evID is None:
 *             if nTx == 1:
 *                 vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))             # <<<<<<<<<<<<<<
 *                 for r in rcv:
 *                     vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
 */
  __pyx_tuple__21 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_1); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_2); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "ttcrpy/tmesh.pyx":712
 *             else:
 *                 if src.shape != rcv.shape:
 *                     raise ValueError('src and rcv should be of equal size')             # <<<<<<<<<<<<<<
 * 
 *                 for n in range(nTx):
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_of_equal_s); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "ttcrpy/tmesh.pyx":994
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TETRA:
 *                 raise ValueError('{0:s} should only contain tetrahedra')             # <<<<<<<<<<<<<<
 *         tet = np.ndarray((data.GetNumberOfCells(), 4))
 *         for n in range(data.GetNumberOfCells()):
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_0_s_should_only_contain_tetrahe); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "ttcrpy/tmesh.pyx":1000
 *                 tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)
 * 
 *         names = ('Slowness', 'slowness', 'Velocity', 'velocity',             # <<<<<<<<<<<<<<
 *                  'P-wave velocity')
 *         for name in names:
 */
  __pyx_tuple__27 = PyTuple_Pack(5, __pyx_n_u_Slowness, __pyx_n_u_slowness, __pyx_n_u_Velocity, __pyx_n_u_velocity, __pyx_kp_u_P_wave_velocity); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "ttcrpy/tmesh.pyx":1012
 *                 break
 *         else:
 *             raise ValueError('File should contain slowness or velocity data')             # <<<<<<<<<<<<<<
 * 
 *         if 'lowness' in name:
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_File_should_contain_slowness_or); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "ttcrpy/tmesh.pyx":1337
 *             nTx = Tx.shape[0]
 *         elif source.shape[1] == 3:
 *             src = source[:,1:3]             # <<<<<<<<<<<<<<
 *             tmp = np.unique(source, axis=0)
 *             nTx = tmp.shape[0]
 */
  __pyx_slice__29 = PySlice_New(__pyx_int_1, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice__29)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__29);
  __Pyx_GIVEREF(__pyx_slice__29);
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_slice_, __pyx_slice__29); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "ttcrpy/tmesh.pyx":1343
 *             t0 = tmp[:,0]
 *         else:
 *             raise ValueError('source should be either nsrc x 2 or 3')             # <<<<<<<<<<<<<<
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_u_source_should_be_either_nsrc_x_2); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "ttcrpy/tmesh.pyx":1346
 * 
 *         if src.shape[1] != 2 or rcv.shape[1] != 2:
 *             raise ValueError('src and rcv should be ndata x 2')             # <<<<<<<<<<<<<<
 * 
 *         if slowness is not None:
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_src_and_rcv_should_be_ndata_x_2); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "ttcrpy/tmesh.pyx":1587
 *         for n in range(data.GetNumberOfCells()):
 *             if data.GetCellType(n) != vtk.VTK_TRIANGLE:
 *                 raise ValueError('{0:s} should only contain triangles')             # <<<<<<<<<<<<<<
 *         tri = np.ndarray((data.GetNumberOfCells(), 3))
 *         for n in range(data.GetNumberOfCells()):
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_u_0_s_should_only_contain_triangl); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 1587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "ttcrpy/tmesh.pyx":41
 *     int build_matrix_siv[T](size_t, size_t, vector[vector[siv[T]]]&, object)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """Set verbosity level for C++ code
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_n_s_v); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_n_s_set_verbose, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 41, __pyx_L1_error)

  /* "ttcrpy/tmesh.pyx":958
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM', int gradient_method=1,
 *                 bool tt_from_rp=1, bool interp_vel=0,
 */
  __pyx_tuple__39 = PyTuple_Pack(23, __pyx_n_s_filename, __pyx_n_s_n_threads, __pyx_n_s_cell_slowness, __pyx_n_s_method, __pyx_n_s_gradient_method, __pyx_n_s_tt_from_rp, __pyx_n_s_interp_vel, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_min_dist, __pyx_n_s_n_secondary, __pyx_n_s_n_tertiary, __pyx_n_s_radius_tertiary, __pyx_n_s_n, __pyx_n_s_nn, __pyx_n_s_reader, __pyx_n_s_data, __pyx_n_s_nodes, __pyx_n_s_tet, __pyx_n_s_names, __pyx_n_s_name_2, __pyx_n_s_slowness, __pyx_n_s_m); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(13, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_n_s_builder, 958, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 958, __pyx_L1_error)

  /* "ttcrpy/tmesh.pyx":1553
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM',double eps=1.e-15, int maxit=20,
 *                 bool process_obtuse=1,uint32_t n_secondary=5):
 */
  __pyx_tuple__41 = PyTuple_Pack(19, __pyx_n_s_filename, __pyx_n_s_n_threads, __pyx_n_s_cell_slowness, __pyx_n_s_method, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_process_obtuse, __pyx_n_s_n_secondary, __pyx_n_s_n, __pyx_n_s_nn, __pyx_n_s_reader, __pyx_n_s_data, __pyx_n_s_nodes, __pyx_n_s_nod, __pyx_n_s_tri, __pyx_n_s_names, __pyx_n_s_name_2, __pyx_n_s_slowness, __pyx_n_s_m); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(8, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_n_s_builder, 1553, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 1553, __pyx_L1_error)

  /* "ttcrpy/tmesh.pyx":1618
 * 
 * 
 * def _rebuild3d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 */
  __pyx_tuple__43 = PyTuple_Pack(16, __pyx_n_s_constructor_params, __pyx_n_s_nodes, __pyx_n_s_tetra, __pyx_n_s_method, __pyx_n_s_cell_slowness, __pyx_n_s_n_threads, __pyx_n_s_tt_from_rp, __pyx_n_s_interp_vel, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_gradient_method, __pyx_n_s_min_dist, __pyx_n_s_n_secondary, __pyx_n_s_n_tertiary, __pyx_n_s_radius_tertiary, __pyx_n_s_g); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_n_s_rebuild3d, 1618, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 1618, __pyx_L1_error)

  /* "ttcrpy/tmesh.pyx":1628
 *     return g
 * 
 * def _rebuild2d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
 *      process_obtuse, n_secondary) = constructor_params
 */
  __pyx_tuple__45 = PyTuple_Pack(11, __pyx_n_s_constructor_params, __pyx_n_s_nodes, __pyx_n_s_triangles, __pyx_n_s_method, __pyx_n_s_cell_slowness, __pyx_n_s_n_threads, __pyx_n_s_eps, __pyx_n_s_maxit, __pyx_n_s_process_obtuse, __pyx_n_s_n_secondary, __pyx_n_s_g); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ttcrpy_tmesh_pyx, __pyx_n_s_rebuild2d, 1628, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6ttcrpy_5tmesh_Mesh3d) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6ttcrpy_5tmesh_Mesh3d.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ttcrpy_5tmesh_Mesh3d.tp_dictoffset && __pyx_type_6ttcrpy_5tmesh_Mesh3d.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ttcrpy_5tmesh_Mesh3d.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Mesh3d, (PyObject *)&__pyx_type_6ttcrpy_5tmesh_Mesh3d) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_ptype_6ttcrpy_5tmesh_Mesh3d = &__pyx_type_6ttcrpy_5tmesh_Mesh3d;
  if (PyType_Ready(&__pyx_type_6ttcrpy_5tmesh_Mesh2d) < 0) __PYX_ERR(0, 1026, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6ttcrpy_5tmesh_Mesh2d.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ttcrpy_5tmesh_Mesh2d.tp_dictoffset && __pyx_type_6ttcrpy_5tmesh_Mesh2d.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ttcrpy_5tmesh_Mesh2d.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Mesh2d, (PyObject *)&__pyx_type_6ttcrpy_5tmesh_Mesh2d) < 0) __PYX_ERR(0, 1026, __pyx_L1_error)
  __pyx_ptype_6ttcrpy_5tmesh_Mesh2d = &__pyx_type_6ttcrpy_5tmesh_Mesh2d;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC inittmesh(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC inittmesh(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_tmesh(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_tmesh(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_tmesh(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'tmesh' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_tmesh(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("tmesh", __pyx_methods, __pyx_k_Raytracing_on_unstructured_tria, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ttcrpy__tmesh) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ttcrpy.tmesh")) {
      if (unlikely(PyDict_SetItemString(modules, "ttcrpy.tmesh", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ttcrpy/tmesh.pyx":24
 * # distutils: language = c++
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * import scipy.sparse as sp
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":26
 * import numpy as np
 * cimport numpy as np
 * import scipy.sparse as sp             # <<<<<<<<<<<<<<
 * 
 * import vtk
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__36);
  __Pyx_GIVEREF(__pyx_n_s__36);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__36);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_sparse, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sp, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":28
 * import scipy.sparse as sp
 * 
 * import vtk             # <<<<<<<<<<<<<<
 * from vtk.util import numpy_support
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_vtk, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtk, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ttcrpy/tmesh.pyx":29
 * 
 * import vtk
 * from vtk.util import numpy_support             # <<<<<<<<<<<<<<
 * 
 * from ttcrpy.tmesh cimport Grid3D, Grid3Ducfs, Grid3Ducsp, Grid3Ducdsp, \
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_numpy_support);
  __Pyx_GIVEREF(__pyx_n_s_numpy_support);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_numpy_support);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_vtk_util, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_numpy_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy_support, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":41
 *     int build_matrix_siv[T](size_t, size_t, vector[vector[siv[T]]]&, object)
 * 
 * def set_verbose(v):             # <<<<<<<<<<<<<<
 *     """Set verbosity level for C++ code
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5tmesh_1set_verbose, NULL, __pyx_n_s_ttcrpy_tmesh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_set_verbose, __pyx_t_1) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":958
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM', int gradient_method=1,
 *                 bool tt_from_rp=1, bool interp_vel=0,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5tmesh_6Mesh3d_33builder, NULL, __pyx_n_s_ttcrpy_tmesh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh3d->tp_dict, __pyx_n_s_builder, __pyx_t_1) < 0) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5tmesh_Mesh3d);

  /* "ttcrpy/tmesh.pyx":957
 * 
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,
 *                 str method='FSM', int gradient_method=1,
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh3d, __pyx_n_s_builder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh3d->tp_dict, __pyx_n_s_builder, __pyx_t_2) < 0) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5tmesh_Mesh3d);

  /* "ttcrpy/tmesh.pyx":1553
 * 
 *     @staticmethod
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,             # <<<<<<<<<<<<<<
 *                 str method='FSM',double eps=1.e-15, int maxit=20,
 *                 bool process_obtuse=1,uint32_t n_secondary=5):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5tmesh_6Mesh2d_23builder, NULL, __pyx_n_s_ttcrpy_tmesh); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh2d->tp_dict, __pyx_n_s_builder, __pyx_t_2) < 0) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5tmesh_Mesh2d);

  /* "ttcrpy/tmesh.pyx":1552
 *         writer.Update()
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def builder(filename, size_t n_threads=1, bool cell_slowness=1,
 *                 str method='FSM',double eps=1.e-15, int maxit=20,
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh2d, __pyx_n_s_builder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6ttcrpy_5tmesh_Mesh2d->tp_dict, __pyx_n_s_builder, __pyx_t_1) < 0) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6ttcrpy_5tmesh_Mesh2d);

  /* "ttcrpy/tmesh.pyx":1618
 * 
 * 
 * def _rebuild3d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
 *      maxit, gradient_method, min_dist, n_secondary, n_tertiary,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5tmesh_3_rebuild3d, NULL, __pyx_n_s_ttcrpy_tmesh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rebuild3d, __pyx_t_1) < 0) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1628
 *     return g
 * 
 * def _rebuild2d(constructor_params):             # <<<<<<<<<<<<<<
 *     (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
 *      process_obtuse, n_secondary) = constructor_params
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6ttcrpy_5tmesh_5_rebuild2d, NULL, __pyx_n_s_ttcrpy_tmesh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rebuild2d, __pyx_t_1) < 0) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ttcrpy/tmesh.pyx":1
 * # -*- coding: utf-8 -*-             # <<<<<<<<<<<<<<
 * """
 * Raytracing on unstructured triangular and tetrahedral meshes
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ttcrpy.tmesh", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ttcrpy.tmesh");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* BytesEquals */
  static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
  static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(operand) if (unlikely(zerodivision_check && ((operand) == 0))) {\
    PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");\
    return NULL;\
}
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(0) break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        }
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntCompare */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntCompare */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
    static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetNameInClass */
    static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


    /* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* Declarations */
    #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
    #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
    #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
    #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
    static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
